#include "SeedGenDlg.h"
//#include "resource.h"
#include <windowsx.h>
#include <time.h>
#include <assert.h>
#include <stdio.h>

#ifndef GWLP_USERDATA
#define GWLP_USERDATA GWL_USERDATA
#define SetWindowLongPtr SetWindowLong
#define GetWindowLongPtr GetWindowLong
#define GWLP_WNDPROC GWL_WNDPROC
#define LONG_PTR LONG
#define DWLP_USER DWL_USER
#endif

typedef LRESULT prevWindowProc( HWND hWnd, UINT Msg, WPARAM wParam, LPARAM lParam );

LRESULT CALLBACK kbInputWindowProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	cGenerateSeedDlg *dlg = (cGenerateSeedDlg *)(LONG_PTR)GetWindowLongPtr( GetParent( hWnd), DWLP_USER );

	if ( dlg == NULL )
	{
		MessageBoxA( hWnd, "NULL", "kbInputWindowProc", MB_OK );
		return TRUE;
	}

	if ( uMsg == WM_GETDLGCODE )
		return DLGC_WANTALLKEYS;
	else
	if ( ( uMsg == WM_KEYDOWN ) || ( uMsg == WM_SYSKEYDOWN ) )
	{
		BYTE bt = BYTE(wParam) ^ BYTE( clock() );

		dlg->updateSeed( &bt, 1 );
		return TRUE;
	}

	if ( dlg->m_prevKbInputWindowProc != NULL )
	{
		return CallWindowProc( dlg->m_prevKbInputWindowProc, hWnd, uMsg, wParam, lParam);
	}

	return FALSE;
}

static HWND getProgressItem( HWND hDlg, const char *className )
{
	cFindDlgControl fc( hDlg, className );

	return fc.m_hCtrl;
}

LRESULT cGenerateSeedDlg::RealDlgProc( UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	if ( uMsg == WM_INITDIALOG )
	{
		RECT rect;
		GetWindowRect( m_hDlg, &rect );
		if ( rect.left == 0 && rect.top == 0 )
		{
			// Move the window relatively mouse position

			POINT point;

			GetCursorPos( &point );

			SetWindowPos( m_hDlg, HWND_TOPMOST, point.x - 100, point.y - 100, 0, 0, SWP_NOSIZE );
		}

		//m_hProgress = GetDlgItem( m_hDlg, IDC_Progress );
		m_hProgress = getProgressItem( m_hDlg, "msctls_progress" );

		SendMessage( m_hProgress, PBM_SETPOS, 0, 0 );

		m_curPosition = 0;

		EnableWindow( GetDlgItem( m_hDlg, IDOK ), FALSE );

		SetFocus( m_hProgress );

		m_prevKbInputWindowProc =
			(WNDPROC)(LONG_PTR)SetWindowLongPtrT( m_hProgress, GWLP_WNDPROC, (LONG_PTR)kbInputWindowProc ); 
	}
	else
	if ( uMsg == WM_MOUSEMOVE )
	{
		if ( m_curPosition >= m_bufferSize )
		{
			EnableWindow( GetDlgItem( m_hDlg, IDOK ), TRUE );
			EnableWindow( GetDlgItem( m_hDlg, IDCANCEL ), FALSE );
		}

		int xPos  = GET_X_LPARAM(lParam); 
		int yPos  = GET_Y_LPARAM(lParam);
		int ticks = (int)clock();

		if ( ticks == -1 )
		{
			assert( 0 );
		}

		BYTE buff[1];

		buff[0] = (BYTE)ticks;
		buff[0] ^= (BYTE)xPos << 2;
		buff[0] ^= (BYTE)yPos << 5;

		updateSeed( buff, sizeof( buff ) );
	}
	else
	if ( uMsg == WM_COMMAND )
	{
		EndDialog( m_hDlg, LOWORD(wParam) );
	}
	else
	if ( uMsg == WM_DESTROY )
	{
		if ( m_prevKbInputWindowProc != NULL )
			SetWindowLongPtrT( m_hProgress, GWLP_WNDPROC, (LONG_PTR)m_prevKbInputWindowProc );
	}

	return FALSE;
}

int cGenerateSeedDlg::updateSeed( BYTE *buff, int size )
{
	while ( size-- )
	{
		if ( m_curPosition < m_bufferSize )
			m_buffer[ m_curPosition ++ ] = buff[ size ];
	}

	if ( m_curPosition >= m_bufferSize )
	{
		EnableWindow( GetDlgItem( m_hDlg, IDOK ), TRUE );
		SetFocus( GetDlgItem( m_hDlg, IDOK ) );
	}

	SendMessage( m_hProgress, PBM_SETPOS, ( m_curPosition * 100 ) / m_bufferSize , 0 );

	return 1;
}

static BYTE xorBytes( BYTE *bf, int size )
{
	BYTE bt = bf[0];
	while( --size )
		bt ^= bf[size];

	return bt;
}

/////////////////////////////////
//
// Silent seed generation...
//	1. Starts infinite walk aroud directory on drive C:
//	2. Sleeps unpredictable and xors current WIN32_FIND_DATAW to single seed buffer byte.
//

static bool walkDirectory( const wchar_t *dir, WIN32_FIND_DATAW *fd, BOOL *pStop )
{
	cWCStr wcsPath; // wchar_t path[ MAX_PATH ];
	wcsPath.Set( dir ); // wcscpy_s( path, dir );
	int len = (int)wcsLen( wcsPath.m_szBody ); // (int)wcslen( path );

	if ( len == 0 )
		return false;

	//if ( path[len - 1] != L'\\' )
	//	path[ len++ ] = L'\\';

	if ( wcsPath.m_szBody[ len - 1 ] != L'\\' )
	{
		if ( wcsPath.Add( L"\\" ) == 0 )
			return false;

		len++;
	}

	//wcscpy_s( path + len, maxLen( path ) - len, L"*" );
	if ( wcsPath.Add( L"*" ) == 0 )
		return false;

	HANDLE hFind = FindFirstFileT( wcsPath.m_szBody, fd ); // path, fd );

	if ( hFind == INVALID_HANDLE_VALUE )
		return false;

	do
	{
		if ( *pStop )
		{
			break;
		}

		if ( fd->dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY )
		{
			if ( wcscmp( fd->cFileName, L"." ) == 0 ) continue;
			if ( wcscmp( fd->cFileName, L".." ) == 0 ) continue;
		}

		if ( fd->dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY )
		{
			//wcscpy_s( path + len, maxLen( path ) - len, fd->cFileName );

			cWCStr wcsDir( wcsPath.m_szBody, len );
			if ( wcsDir.Add( fd->cFileName ) == 0 )
				return false;

			//walkDirectory( path, fd, pStop );
			walkDirectory( wcsDir.m_szBody, fd, pStop );
		}
	}
	while( FindNextFileT( hFind, fd ) );

	FindClose( hFind );

	return true;
}

#pragma pack(1)

struct walkBlock
{
	BOOL stop;
	BOOL stopped;
	WIN32_FIND_DATAW fd;

	walkBlock()
	{
		memset( this, 0, sizeof( *this ) );
	}
};

#pragma pack()

static DWORD WINAPI walkProc( LPVOID lpParameter )
{
	walkBlock *wb = ( walkBlock * )lpParameter;
	
	wb->stopped = false;
	wb->stop = false;

	while ( !wb->stop )
		walkDirectory( L"C:\\", &(wb->fd), &(wb->stop) );

	wb->stopped = true;

	return 0;
}

//static
int cGenerateSeedDlg::GetSilentSeed( BYTE *buffer, int bufferSize )
{
	ULARGE_INTEGER freeBytesAvailableToCaller;
    ULARGE_INTEGER totalNumberOfBytes;
    ULARGE_INTEGER totalNumberOfFreeBytesuliFree;

	BOOL ok = GetDiskFreeSpaceExT( L"C:\\", &freeBytesAvailableToCaller, &totalNumberOfBytes, &totalNumberOfFreeBytesuliFree );

	BYTE bt0 = xorBytes( (BYTE*)&(totalNumberOfFreeBytesuliFree.QuadPart), sizeof( totalNumberOfFreeBytesuliFree.QuadPart ) );

	walkBlock wb;
	DWORD dwID;

	// Starts walk directory loop 
	HANDLE hThread = CreateThread( NULL, 0, (LPTHREAD_START_ROUTINE)walkProc, &wb, 0, &dwID );

	if ( hThread == NULL )
	{
		assert( hThread != NULL );
		return -1;
	}

	int i;

	for( i = 10; i--;  )
	{
		// Check was filled
		if ( wb.fd.cFileName[0] != 0 )
			break;

		assert( i > 0 );
		Sleep( 100 );
	}

	for ( i = 0; i < bufferSize; i++ )
	{
		// Sleep unpredictable to get a different files.
		DWORD ms = bt0 & 0x7F;

		if ( ms == 0 )
			ms = 1;

		Sleep( ms );

		// Mix time here
		clock_t clk = clock() ^ (clock_t)time(0);
		BYTE bt1 = xorBytes( (BYTE*)&clk, sizeof( clk ) );
		buffer[i] = bt0 ^ bt1;

		// Mix current directory info.
		bt0 = xorBytes( (BYTE*)&wb, sizeof( wb ) );
	}

	// Lets thread to be self stopped ( All handles have to be properly closed ).

	wb.stop = true;
	
	for( i = 10; i--; )
	{
		Sleep( 100 );

		if ( wb.stopped )
			break;

		assert( false );
	}
	
	CloseHandle( hThread );

	return 0;
}
//
//	End of silent seed generation.
//
////////////