#pragma once

//
//	Public recipient list
//
//	The Key Manager functions create and delete the list.
//
//		cPublicRecipientList *UpdatePublicRecipientList( cPublicRecipientList *prl = NULL );
//		void FreePublicRecipientList( cPublicRecipientList *prl );
//

#include <windows.h>
#include "resource.h"

class cPublicRecipient
{
public:
	wchar_t *m_name;
	wchar_t *m_eMail;
	BYTE *m_keyID;
	int m_keyIDSize;
	cPublicRecipient *m_next;
	
	cPublicRecipient( wchar_t *name, wchar_t *eMail, BYTE *keyID, int keyIDSize )
	{
		memset( this, 0, sizeof( *this ) );

		if ( ( name != NULL ) && ( *name != L'\0' ) )
			m_name = _wcsdup( name );
		else
		{
			// Forms name from email common name 'John Smith <JohnSmith@jetico.com>'
			if ( eMail != NULL )
			{
				m_name = _wcsdup( eMail );

				if ( m_name == NULL )
					return;

				for( wchar_t *p = m_name; *p; p++ )
				{
					if ( *p == L'<' )
					{
						while( *(p-1) == L' ' )
						{
							p--;
						}

						*p = L'\0';

						break;
					}

					// Here m_name is 'John Smith'
				}

			}
			else
				m_name = _wcsdup( _LoadIDS( IDS_szUndefined ) );
		}

		if ( eMail != NULL )
			m_eMail = _wcsdup( eMail );
		else
			m_eMail = _wcsdup( _LoadIDS( IDS_szUndefined ) );

		m_keyIDSize = keyIDSize;
		if ( m_keyIDSize == 0 )
			return;

		m_keyID = new BYTE[ m_keyIDSize ];
		if ( m_keyID == NULL )
		{
			m_keyIDSize = 0;
			return;
		}

		memcpy( m_keyID, keyID, m_keyIDSize );
	}
	
	~cPublicRecipient()
	{
		if ( m_name != NULL ) free( m_name );
		if ( m_eMail != NULL ) free( m_eMail );
		if ( m_keyID != NULL ) delete m_keyID;
	}
};

class cPublicRecipientList
{
	cPublicRecipient *m_head;

public:
	cPublicRecipientList()
	{
		m_head = NULL;
	}

	bool Empty()
	{
		if ( m_head == NULL )
			return false;

		while( m_head )
		{
			cPublicRecipient *cur = m_head; 
			m_head = cur->m_next;
			delete cur;
		}

		return true;
	}

	~cPublicRecipientList()
	{
		Empty();
	}

	int GetCount()
	{
		int count = 0;

		for( cPublicRecipient *cur = m_head; cur; cur = cur->m_next )	count ++;

		return count;
	}

	cPublicRecipient *GetAt( int pos )
	{
		int count = 0;

		for( cPublicRecipient *cur = m_head; cur; cur = cur->m_next )
			if ( count++ == pos )
				return cur;

		return NULL;
	}

	bool Add( cPublicRecipient *pr )
	{
		pr->m_next = m_head;
		m_head = pr;

		return true;
	}

	bool Add( wchar_t *name, wchar_t *eMail, BYTE *keyID, int keyIDSize )
	{
		cPublicRecipient *pr = new cPublicRecipient( name, eMail, keyID, keyIDSize );

		if ( pr == NULL )
			return false;

		return Add( pr );
	}
};
