#include "resource.h"
#include "DlgAskWriteConfirmation.h"

#include "..\Includes\getFileVersion.h"
#include "..\Includes\UnicodeFuncs.h"

#include <stdio.h>
#include <assert.h>
#include <windowsx.h>

cDlgAskWriteConfirmation::cDlgAskWriteConfirmation( const wchar_t *pDstName, const wchar_t *pSrcName )
{
	askSrcName = pSrcName;
	askDstName = pDstName;

	srcVersion[0] = '\0';
	dstVersion[0] = '\0';

	DWORD dstVerSize = sizeof( dstVersion );
	DWORD srcVerSize = sizeof( srcVersion );

	getFileVersionT( askDstName, dstVersion, &dstVerSize);
	getFileVersionT( askSrcName, srcVersion, &srcVerSize);

	getFileInfoT( askDstName, dstDate, sizeof( dstDate ), dstSize, sizeof( dstSize ) );
	getFileInfoT( askSrcName, srcDate, sizeof( srcDate ), srcSize, sizeof( srcSize ) );

	hDstIcon = NULL;
	hSrcIcon = NULL;
}

static void DrawFileIcon( HWND hwnd, HDC dc, HICON hIcon, int id )
{
	if ( !hIcon )
		return;

	RECT r;
	GetWindowRect( GetDlgItem( hwnd, id ), &r );
	POINT p = { r.left, r.top };
	ScreenToClient( hwnd, &p );
	DrawIcon( dc, p.x, p.y, hIcon );

}

BOOL cDlgAskWriteConfirmation::InitDialog()
{
	wchar_t str[MAX_PATH], pattern[MAX_PATH];

// Set destination file information...
	GetDlgItemTextT( IDC_DstFileText, pattern, MAX_PATH );
	::FormatT( str, maxLen( str ), pattern,
				askDstName,
				dstVersion,
				dstDate,
				dstSize
				);

	SetDlgItemTextT( IDC_DstFileText, str );

// Set source (distributive) file information...

	GetDlgItemTextT( IDC_SrcFileText, pattern, MAX_PATH );
	::FormatT( str, maxLen( str ), pattern,
				srcVersion,
				srcDate,
				srcSize
				);

	SetDlgItemTextT( IDC_SrcFileText, str);

	UINT uFlags = SHGFI_ICON;

	SHFILEINFOW sfi;

	if ( SHGetFileInfoT( askDstName, 0, &sfi, sizeof(sfi), uFlags | SHGFI_LARGEICON))
		hDstIcon = sfi.hIcon;

	if ( SHGetFileInfoT( askSrcName, 0, &sfi, sizeof(sfi), uFlags | SHGFI_LARGEICON))
		hSrcIcon = sfi.hIcon;

	bool overwriteExising = true;

	if ( CompareVersionsT( srcVersion, dstVersion ) < 0 )
		overwriteExising = false;

	HWND hCheckedRB = GetDlgItem( m_hDlg, IDC_RADIO1 );

	if ( !overwriteExising )
	{
		Button_SetCheck( hCheckedRB, 0 ); // Uncheck default first.
		hCheckedRB = GetDlgItem( m_hDlg, IDC_RADIO3 );
	}

	Button_SetCheck( hCheckedRB, 1 );

	if ( hCheckedRB )
	{
		wchar_t text[ 256 ];
		GetWindowTextT( hCheckedRB, text, sizeof( text ) );
		wcscat_s( text, _LoadIDS( IDS_szRecommended ) );
		SetWindowTextT( hCheckedRB, text );
	}

	return 0;
}

LRESULT cDlgAskWriteConfirmation::RealDlgProc( UINT uMsg, WPARAM wParam, LPARAM lParam )
{
    switch (uMsg)
	{ 
 		case WM_COMMAND:
		{
			switch (LOWORD(wParam))
			{
				case IDCANCEL:
				{
					EndDialog( m_hDlg, idCancelled );
					break;
				}
				case IDOK:
				{
					int res = 1;

					if		( Button_GetCheck( GetDlgItem( m_hDlg, IDC_RADIO1 ))) res = idOverwrite;
					else if ( Button_GetCheck( GetDlgItem( m_hDlg, IDC_RADIO2 ))) res = idOverwriteAll;
					else if ( Button_GetCheck( GetDlgItem( m_hDlg, IDC_RADIO3 ))) res = idSkip;
					else if ( Button_GetCheck( GetDlgItem( m_hDlg, IDC_RADIO4 ))) res = idSkipAll;

					EndDialog( m_hDlg, res );
					break;
				}
			}
			return 1;
		}
		case WM_INITDIALOG:
		{
			return InitDialog();
		}
		case WM_PAINT :
		{
			HDC dc;
			PAINTSTRUCT ps;

			if ((HDC) wParam != NULL)
				dc = (HDC) wParam;
			else
				dc = BeginPaint(m_hDlg,&ps);

			DrawFileIcon( m_hDlg, dc, hSrcIcon, IDC_IconSrc );
			DrawFileIcon( m_hDlg, dc, hDstIcon, IDC_IconDst );

			if ((HDC) wParam == NULL)
				EndPaint(m_hDlg,&ps);

			break;
		}
		case WM_DESTROY:
			if ( hSrcIcon ) DestroyIcon( hSrcIcon );
			if ( hDstIcon ) DestroyIcon( hDstIcon );
			break;
		default: return 0;
	}
	return 0;	
}

int DlgAskWriteConfirmation( HINSTANCE hInst, HWND hWnd, const wchar_t *dst, const wchar_t *src )
{
	cDlgAskWriteConfirmation dlg( dst, src );

	int res = dlg.DoModal( hInst, IDD_DlgAskConf, hWnd );

	return res;
}