#include "resource.h"
#include "DlgFindParams.h"
#include "BCKeyMan.h"

#include "..\CertificateLib\SMemory\SBlocks.h"

#include "..\Includes\EquStrings.h"
#include "..\Includes\UnicodeFuncs.h"
#include "..\Includes\ptbRegs.h"
#include "..\Includes\menuItem.h"

#include <assert.h>
#include <windowsx.h>

class cStrListItem
{
public:
	wchar_t *m_body;
	cStrListItem *m_next;

	cStrListItem( wchar_t *ws, int len )
	{
		memset( this, 0, sizeof( *this ) );
		m_body = new wchar_t[ len + 1 ];

		if ( m_body != NULL )
		{
			wcsncpy( m_body, ws, len );
			m_body[ len ] = L'\0';
		}
	}

	~cStrListItem()
	{
		if ( m_body ) delete m_body;
	}
};

bool cStrList::Add( wchar_t *ws, int len )
{
	cStrListItem *l = Find( ws, len );

	if ( l != NULL )
		Remove( l );

	l = new cStrListItem( ws, len );

	if ( ( l == NULL ) || ( l->m_body == NULL ) )
		return false;

	l->m_next = m_head;
	m_head = l;

	return true;
}

cStrList::~cStrList()
{
	while ( m_head )
	{
		cStrListItem *l = m_head;
		m_head = l->m_next;
		delete l;
	}
}

cStrListItem *cStrList::Find( wchar_t *ws, int len )
{
	cStrListItem *l;

	for ( l = m_head; l; l = l->m_next )
	{
		if ( wcsncmp( ws, l->m_body, len ) != 0 )
			continue;

		if ( wcslen( l->m_body ) == len )
			break;
	}

	return l;
}

bool cStrList::Remove( cStrListItem *r )
{
	if ( r == m_head )
	{
		m_head = r->m_next;
		delete r;
		return true;
	}

	for ( cStrListItem *l = m_head; l->m_next; l = l->m_next )
	{
		if ( l->m_next == r )
		{
			l->m_next = r->m_next;
			delete r;
			return true;
		}
	}

	assert( false );
	return false;
}

///////////////////////////////////////////
//
// cFindParams
//

int cFindParams::GetCurSel( int *pCount )
{
	int count = ListView_GetItemCount( m_hListView );
	int pos = ListView_GetNextItem( m_hListView, -1, LVNI_FOCUSED );

	if ( m_bSearchUp && ( pos == -1 ) )
		pos = count;

	if ( pCount != NULL )
		*pCount = count;

	return pos;
}

bool cFindParams::Add( wchar_t *ws, int len, HWND hCmb )
{
	m_walkCout = ListView_GetItemCount( m_hListView );

	bool res = false;
	
	if ( len != 0 )
		res = m_recentList.Add( ws, len );

	// Update combo box 
	if ( hCmb != NULL )
	{
		ComboBox_ResetContent( hCmb );

		for ( cStrListItem *sli = m_recentList.m_head; sli; sli = sli->m_next )
		{
			// :( standard combo box has no wide char implementation
			char bf[ 1024 ];
			wcstombs( bf, sli->m_body, sizeof( bf ) );
			ComboBox_AddString( hCmb, bf );
//			ComboBox_AddString( hCmb, sli->m_body );
		}

		ComboBox_SetCurSel( hCmb, 0 );
	}

	return res;
}

bool cFindParams::FindNext( HWND hWnd )
{
	int findRes = FindNext();

	if ( findRes == fnNotFound )
	{
		MessageBoxT( hWnd, _LoadIDS( IDS_NoMoreItemFound ), _LoadIDS( IDS_AppName ), MB_ICONINFORMATION );
		return false;
	}
	else
	if ( findRes == fnLoop )
	{
		MessageBoxT( hWnd, _LoadIDS( IDS_StartPointReached ), _LoadIDS( IDS_AppName ), MB_ICONINFORMATION );
		return true;
	}
	else
		return true;

	return false;
}

int cFindParams::FindNext()
{
	if ( ( m_hListView == NULL )
		|| ( m_columnQ == 0 )
		|| ( m_recentList.m_head == NULL ) )
	{
		assert( false );
		return fnError;
	}

	//
	//	Chooses string to find from on recent top nad prepares find pattern
	//
	wchar_t findPattern[ 1024 ];
	wcscpy_s( findPattern, L"*" );
	wcscat_s( findPattern, m_recentList.m_head->m_body );
	wcscat_s( findPattern, L"*" );

	//
	//	Walks all list view items and their subitems to compare with pattern
	//

	//
	// Gets cur selected.
	//
	int count = 0;
	int pos = GetCurSel( &count );

	if ( count == 0 )
		return fnNotFound;

	if ( m_walkCout == -1 ) // Was not inited yet
	{
//		assert( false );
		m_walkCout = count;
	}

	int res = fnNotFound;
	
	while( res == fnNotFound )
	{
		if ( m_bSearchUp )	{ if ( --pos < 0 )		pos = count - 1; }
		else				{ if ( ++pos >= count ) pos = 0; }

		for( int i =0; i < m_columnQ; i++ )
		{
			wchar_t text[ 1024 ] = L"\0";
			ListView_GetItemTextT( m_hListView, pos, i, text, maxLen( text ) );

			if ( EquStrings( text, findPattern, m_bMatchCase ) )
			{
				// Deselect all selected.
				int iSelected = -1;
				while( true )
				{
					iSelected = ListView_GetNextItem( m_hListView, iSelected, LVNI_SELECTED );
					if ( iSelected == -1 )
						break;

					ListView_SetItemState( m_hListView, iSelected, 0, LVIS_FOCUSED | LVIS_SELECTED );
				}

				// Select the new found.
				ListView_SetItemState( m_hListView, pos, LVIS_FOCUSED | LVIS_SELECTED, LVIS_FOCUSED | LVIS_SELECTED );
				res = fnFound;
				break;
			}
		}

		if ( --m_walkCout <= 0 )
		{
			m_walkCout = ListView_GetItemCount( m_hListView );
			if ( res == fnFound )
				res = fnLoop;
			
			break;
		}
	}

	return res;
}

cFindParams::cFindParams()
{
	m_bSearchUp = false;
	m_bMatchCase = true;
	m_wString = NULL;

	m_hListView = NULL;
	m_columnQ = 0;
	m_walkCout = -1;

	RegGet();
}

cFindParams::~cFindParams()
{
	if ( m_wString != NULL )
		delete m_wString;

	RegSave();
}

//static HKEY hKeyRoot = HKEY_CURRENT_USER;
//static char *szBCKeyManKey = "Software\\Jetico\\BCKeyMan";
static wchar_t *szFindParams = L"FPARAMs";

bool cFindParams::RegSave()
{
	cPlainTextBlock ptb;

	// Sets flags
	DWORD flags =  0;
	if ( m_bMatchCase ) flags |= eMatchCase;
	if ( m_bSearchUp  ) flags |= eSearchUp;

	ptb.WriteBytes( (BYTE *)&flags, sizeof( flags ) );

	// Sets stored strings.
	for ( cStrListItem *l =	m_recentList.m_head; l; l = l->m_next )
	{
		int len = wcslen( l->m_body );
		if ( len == 0 )
			continue;

		ptb.WriteBytes( (BYTE *)&len, sizeof( len ) );
		ptb.WriteBytes( (BYTE *)l->m_body, len * sizeof( wchar_t ) );
	}

	return writeToRegistry( &ptb, hKeyRoot, szBCKeyManKey, szFindParams );
}

bool cFindParams::RegGet()
{
	cPlainTextBlock ptb;

	if ( !readFromRegistry( &ptb, hKeyRoot, szBCKeyManKey, szFindParams ) )
		return false;

	if ( ptb.bufferSize < sizeof( DWORD ) )
		return false;

	// Gets flags
	DWORD flags = *( DWORD *)ptb.buffer;
	m_bMatchCase = ( flags & eMatchCase ) != 0;
	m_bSearchUp =  ( flags & eSearchUp ) != 0;

	// Gets stored strings.
	BYTE *pkt = ptb.buffer + sizeof( DWORD );
	int size = ptb.bufferSize - sizeof( DWORD );

	while ( size > 0 )
	{
		int len = *( int * )pkt;
		pkt += sizeof( len );
		
		wchar_t *ws = ( wchar_t * ) pkt;
		pkt += len * sizeof( wchar_t );
		size -= sizeof( len ) + len * sizeof( wchar_t );

		m_recentList.Add( ws, len );
	}

	return true;
}

bool cFindParams::IsEmpty()
{
	return ( m_wString == NULL ) || ( m_wString[0] == L'\0' );
}

//////////////////////////
//
//	cDlgFindParams
//

cDlgFindParams::cDlgFindParams( cFindParams *params )
{
	m_params = params;
}
/*
#ifndef CB_GETCOMBOBOXINFO

typedef struct tagCOMBOBOXINFO {
  DWORD cbSize;
  RECT  rcItem;
  RECT  rcButton;
  DWORD stateButton;
  HWND  hwndCombo;
  HWND  hwndItem;
  HWND  hwndList;
} COMBOBOXINFO, *PCOMBOBOXINFO, *LPCOMBOBOXINFO;

#define CB_GETCOMBOBOXINFO          0x0164
#define GetComboBoxInfo( hCmb, pcbi )  SendMessage( hCmb, CB_GETCOMBOBOXINFO, 0, (LPARAM)pcbi )

#endif
*/
static wchar_t *Cmb_GetFirstString( HWND hCmb, wchar_t *findStr, int wcSize )
{
	//COMBOBOXINFO cbi;
	//memset( &cbi, 0, sizeof( cbi ) );
	//cbi.cbSize = sizeof( cbi );
	//BOOL ok = GetComboBoxInfo( hCmb, &cbi );

	//assert( ok );

	//GetWindowTextT( cbi.hwndItem, findStr, wcSize ); // maxLen( findStr ) );

	GetWindowTextT( hCmb, findStr, wcSize );
	return findStr;
}

LRESULT cDlgFindParams::RealDlgProc( UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	LRESULT result;
	
	if ( cMenuItem::DlgProc( &result, m_hDlg, uMsg, wParam, lParam ) )
		return result;

	switch (uMsg)
	{
		case WM_INITDIALOG:
		{
			HWND hCmb = GetDlgItem( m_hDlg, IDC_cmbString );
			m_params->Add( NULL, 0, hCmb );
			Button_SetCheck( GetDlgItem( m_hDlg, IDC_chkSearchUp ), m_params->m_bSearchUp ? BST_CHECKED : BST_UNCHECKED );
			Button_SetCheck( GetDlgItem( m_hDlg, IDC_chkMatchCase ), m_params->m_bMatchCase ? BST_CHECKED : BST_UNCHECKED );
			return true;
		}
 		case WM_COMMAND:
		{
			switch (LOWORD(wParam))
			{
				case IDCANCEL:
					EndDialog( m_hDlg, IDCANCEL );
					return true;
				case IDOK:
				{
					wchar_t findStr[ 1024 ];
					HWND hCmb = GetDlgItem( m_hDlg, IDC_cmbString );
					Cmb_GetFirstString( hCmb, findStr, maxLen( findStr ) );
					m_params->Add( findStr, wcslen( findStr ), hCmb );

					if ( !OnFindNext() )
						return false;

					EndDialog( m_hDlg, IDOK );
					return true;
				}
				case IDC_chkMatchCase:
					m_params->m_bMatchCase = Button_GetCheck( GetDlgItem( m_hDlg, IDC_chkMatchCase ) ) == BST_CHECKED;
					break;
				case IDC_chkSearchUp:
					m_params->m_bSearchUp = Button_GetCheck( GetDlgItem( m_hDlg, IDC_chkSearchUp ) ) == BST_CHECKED;
					break;
				default:
					return false;
			}
		}
//		case WM_KILLFOCUS:
		default:
			return false;
	}
}

BOOL cDlgFindParams::OnFindNext()
{
	return m_params->FindNext( m_hDlg );
}
