#include "DlgViewLocalDatabase.h"
#include "BCDB_Functions.h"
#include "cWizardBlock.h"
#include ".\nwclient\http_get.h"

#include "..\CertificateLib\PGPPkts\pktdefs.h"
#include "..\CertificateLib\PGPPkts\PGPPkts.h"
#include "..\CertificateLib\PGPPkts\Packets.h"

#include "..\Includes\ptbRegs.h"
#include "..\Includes\TranslateDll.h"

#include <commctrl.h>
//
//	List view init functions and values
//

static UINT ColumnLabelIDS[] = 
{
	IDS_szUserName, IDS_szAlgorithm, IDS_LocalKeyID, IDS_CrtThumbprint, IDS_szFileName
};

const int fileNamePos = 4;

const int NUM_COLUMNS = sizeof( ColumnLabelIDS ) / sizeof( UINT );

static int _gnColumnFmt[ NUM_COLUMNS ] =
{
	LVCFMT_LEFT, LVCFMT_LEFT, LVCFMT_LEFT, LVCFMT_LEFT, LVCFMT_LEFT
};

static int _gnColumnWidth[NUM_COLUMNS]=
{
	223, 100, 80, 80, 70
};

static bool InitListViewColumns( HINSTANCE hInst, HWND hList )
{
	int i;
	LV_COLUMNW lvc;

	memset( &lvc, 0, sizeof( lvc ) );

	lvc.mask = LVCF_WIDTH;

	//
	//	Check if already initialised...
	//
	if ( ListView_GetColumnT( hList, 0, &lvc ) )
		return TRUE;

	lvc.mask = LVCF_FMT | LVCF_WIDTH | LVCF_TEXT | LVCF_SUBITEM;

	wchar_t text[ 80 ];

	int res;

	for(i=0; i<NUM_COLUMNS; i++)
	{
		LoadStringT( hInst, ColumnLabelIDS[i], text, maxLen( text ) );

		lvc.iSubItem=i;
		lvc.pszText = text;
		lvc.cx = _gnColumnWidth[i];
		lvc.fmt=_gnColumnFmt[i];

		res = ListView_InsertColumnT( hList, i, &lvc );

		if ( res == -1 )
		{
			MessageBoxT( hList, text, _LoadIDS( IDS_ErrorInAddLVColumn ), MB_OK );
			return FALSE;
		}
	}

	DWORD dwStyle = LVS_EX_FULLROWSELECT; // | LVS_EX_GRIDLINES;
	ListView_SetExtendedListViewStyleEx( hList, dwStyle, dwStyle );

	return TRUE;
}

static int LV_SetKeyID( HWND hList, int item, BYTE *keyID )
{
	BYTE *keyIDCopy = ( BYTE * )ListView_GetItemParam( hList, item );

	if ( keyIDCopy == NULL )
		keyIDCopy = new BYTE[ pgp_keyIDSize ];

	if ( keyIDCopy == NULL )
		return 0;

	memcpy( keyIDCopy, keyID, pgp_keyIDSize );

	ListView_SetItemParam( hList, item, (DWORD)keyIDCopy );

	return pgp_keyIDSize;
}

int lvGetSelectedItem( HWND hLV )
{
	if ( ListView_GetSelectedCount( hLV ) != 1 )
		return -1;
	else
	{
		int item = ListView_GetNextItem( hLV, -1, LVNI_SELECTED );
		return item;
	}
}

BYTE *LV_GetKeyID(
		HWND hList,
		int item // -1 means selected item
	)
{
	if ( item == -1 )
	{
		item = lvGetSelectedItem( hList );

		if ( item == -1 )
			return NULL;
	}

	return ( BYTE * )ListView_GetItemParam( hList, item );
}

static void LV_FreeKeyID( HWND hList, int item )
{
	BYTE *keyIDCopy = ( BYTE * )ListView_GetItemParam( hList, item );

	if ( keyIDCopy != NULL )
		delete keyIDCopy;
}

static void LV_DeleteAllItems( HWND hList )
{
	for ( int i = ListView_GetNextItem( hList, -1, LVNI_ALL ); i != -1; i = ListView_GetNextItem( hList, i, LVNI_ALL ) )
	{
		LV_FreeKeyID( hList, i );
	}

	ListView_DeleteAllItems( hList );
}

static int setPGPItemValues( HWND hList, int item, cpgp_keyInfo *keyInfo, int iImage )
{
	// Sets item image [+] if has subkeys, empty if not or |__ if subkey
	LVITEMW lvi;
	memset( &lvi, 0, sizeof( lvi ) );
	lvi.iItem = item;
	lvi.mask = LVIF_IMAGE;
	lvi.iImage = iImage; 
	ListView_SetItemT( hList, &lvi );

	// Sets user ID

	int subItem = 0;

	if ( keyInfo->userID != NULL )
	{
		//const char szEnd[1] = "";
		//cPTB ptbUserID;
		cWCStr wcsUserID;

		//if ( !ptbUserID.WriteBytes( keyInfo->userID->body, keyInfo->userID->size ) || 
		//	 !ptbUserID.WriteBytes( (BYTE*)szEnd, 1 ) )
		if ( wcsUserID.Set( (char *)keyInfo->userID->body, keyInfo->userID->size, CP_UTF8 ) == 0 )
		{
			MessageBoxT(
					hList,
					_LoadIDS( IDS_ErrorInAllocatingMemory ),
					_LoadIDS( IDS_BrowsePKUser ),
					MB_ICONSTOP
				);

			return ppe_memoryAllocation;
		}

		//ListView_SetItemTextT( hList, item, subItem++, getWCStr( (char *)ptbUserID.buffer ) );
		ListView_SetItemTextT( hList, item, subItem++, wcsUserID.m_szBody  );
	}
	else
	{
		if ( iImage == 3 ) // Subkey
		{
			ListView_SetItemTextT( hList, item, subItem, _LoadIDS( IDS_Subkey ) );
		}

		subItem++;
	}

	//
	//	Sets algorithm description
	//
	BYTE version = keyInfo->keyPkt[0];

	BYTE algID = -1;
	DWORD bitSize = -1;

	switch ( version )
	{
		case 2:
		case 3:
		{
			sPublicKeyV3 *pk = (sPublicKeyV3 *)keyInfo->keyPkt;
			algID = pk->algID;
			bitSize = ((DWORD)pk->mpi[0] << 8 ) | pk->mpi[1];

			break;
		}
		case 4:
		{
			sPublicKey *pk = (sPublicKey *)keyInfo->keyPkt;
			algID = pk->algID;
			bitSize = ((DWORD)pk->mpi[0] << 8 ) | pk->mpi[1];

			break;
		}
		default:
			return ppe_invalidPacketVersion;
	}

	wchar_t algDesc[ 128 ];
	FormatT( algDesc, maxLen( algDesc ), _LoadIDS( IDS_NameAndBits_SD ), cWCStr( PKT_GetPublicKeyAlgName( algID ) ).m_szBody, bitSize );

	ListView_SetItemTextT( hList, item, subItem++, algDesc );

	// Sets key ID string

	//char keyIDStr[ 1024 ];
	//int size = sizeof( keyIDStr );
	//bytes2string( keyIDStr, &size, keyInfo->keyID, pgp_keyIDSize );
	//ListView_SetItemTextA( hList, item, subItem++, keyIDStr );

	cWCStr wcsKeyId;
	bytes2string( &wcsKeyId, keyInfo->keyID, pgp_keyIDSize );
	ListView_SetItemTextT( hList, item, subItem++, wcsKeyId.m_szBody );

	subItem++; // skip thumbprint subitem

	// Sets key ring path
	wchar_t path[ MAX_PATH ];
	bool ok = BCDB_GetLocalKeyRingPath( path, maxLen( path ), false );

	ListView_SetItemTextT( hList, item, subItem++, path );

	//
	//	Set item parameter as key ID, it alocates new buffer for the key ID copies it, 
	//	the allocated buffer has to be deleted 
	//
	LV_SetKeyID( hList, item, keyInfo->keyID );

	return ppe_ok;
}

int cDlgViewLocalDatabase::SetPGPItemValues( BYTE *krPkt, int krSize, int item, BYTE *keyID )
{
	//
	//	1. Gets key info
	//

	cpgp_keyInfo keyInfo;

	int res = pgp_FindKeyByIDEx(
				&keyInfo,
				keyID,
				krPkt, krSize,
				m_secret );

	if ( res != ppe_ok )
	{
		MessageBoxT( m_hDlg, _LoadIDS( IDS_szErrorInKeyReading ), _LoadIDS( IDS_BrowsePKUser ), MB_ICONSTOP );
		return res;
	}

	//
	//	2. Set the key values.
	//

	setPGPItemValues( 
		m_hList,
		item,
		&keyInfo,
		( keyInfo.next ? 1 : 0 ) // If has subkeys ? collapsed : empty 
	);

	return ppe_ok;
}

static void addSubKey( HWND hList, int item, cpgp_keyInfo *keyInfo )
{
	LVITEMW lvi;
	memset( &lvi, 0, sizeof( lvi ) );
	lvi.iItem = item;

	item = ListView_InsertItemT( hList, &lvi );

	if ( item != -1 )
	{
		setPGPItemValues( hList, item, keyInfo, 3 );
	}
}

//
//	Expands PGP subkeys
//	
void cDlgViewLocalDatabase::expandItem( int item )
{
	LVITEMW lvi;
	memset( &lvi, 0, sizeof( lvi ) );
	lvi.iItem = item;
	lvi.mask = LVIF_IMAGE;
	lvi.iImage = 2; // Expanded
	ListView_SetItemT( m_hList, &lvi );

	cpgp_keyInfo keyInfo;

	BYTE *keyID = LV_GetKeyID( m_hList, item );
	if ( keyID == NULL )
	{
		assert( false );
		return;
	}

	int res = pgp_FindKeyByIDEx(
				&keyInfo,
				keyID,
				m_keyPkt, m_keyPktSize,
				m_secret );

	if ( res != ppe_ok )
		return;

	for ( cpgp_keyInfo *subKey = keyInfo.next; subKey; subKey = subKey->next )
	{
		addSubKey( m_hList, item + 1, subKey );
	}
}

int deleteSubKeys( HWND hList, int item )
{
	LVITEMW lvi;
	memset( &lvi, 0, sizeof( lvi ) );
	lvi.mask = LVIF_IMAGE;

	int count = 0;

	item++; // gets next subkey;

	while ( item < ListView_GetItemCount( hList ) ) 
	{
		lvi.iItem = item;
		ListView_GetItem( hList, &lvi );
		if ( lvi.iImage != 3 ) // 3 - Sub key image.
			break;

		count++;

		LV_FreeKeyID( hList, item );
		ListView_DeleteItem( hList, item );
	}

	return count;
}

static void collapseItem( HWND hList, int item )
{
	LVITEMW lvi;
	memset( &lvi, 0, sizeof( lvi ) );
	lvi.iItem = item;
	lvi.mask = LVIF_IMAGE;
	lvi.iImage = 1; // Collapsed
	ListView_SetItemT( hList, &lvi );

	deleteSubKeys( hList, item );
}

BOOL cDlgViewLocalDatabase::checkHitPlus( POINT pt )
{
	// Skip hittest of certificate and p12 secret keys
	if ( m_viewType != vt_pgpPublicKey )
		return FALSE;

	if ( pt.x > 32 )
		return FALSE;

	LVHITTESTINFO hti;

	memset( &hti, 0, sizeof( hti ) );

	hti.flags = LVHT_ONITEMICON;
	hti.pt = pt;

	int item = ListView_HitTest( m_hList, &hti );

	if ( item == -1 )
		return FALSE;

	LVITEMW lvi;
	memset( &lvi, 0, sizeof( lvi ) );
	lvi.iItem = item;
	lvi.mask = LVIF_IMAGE;

	ListView_GetItem( m_hList, &lvi );

	if ( lvi.iImage == 0 ) // Empty
		return TRUE;

	if ( lvi.iImage == 3 ) // Subkey
		return TRUE;

	if ( lvi.iImage == 1 ) // Collapsed
		expandItem( item );
	else
		collapseItem( m_hList, item );

	return TRUE;
}

int cDlgViewLocalDatabase::FillListView_PGP( bool secret )
{
	ListView_SetImageList( m_hList, m_himlPGP, LVSIL_SMALL );
	ListView_SetImageList( m_hList, m_himlPGPLarge, LVSIL_NORMAL );

	m_secret = secret;
	m_keyPkt = NULL;
	m_keyPktSize = 0;

	wchar_t path[ MAX_PATH ];
	bool ok = BCDB_GetLocalKeyRingPath( path, maxLen( path ), secret );

	if ( !ok )
	{
//		assert( false );
		return ppe_internalError;
	}

	if ( !m_fvKeyRing.Open( path ) )
		return ppe_internalError;

	if ( m_krImage.WriteBytes( m_fvKeyRing.m_ptr, (int)m_fvKeyRing.m_size.QuadPart, true ) != m_fvKeyRing.m_size.QuadPart )
		return ppe_internalError;

	m_keyPkt = m_krImage.buffer;
	m_keyPktSize = m_krImage.bufferSize;
	
	m_fvKeyRing.Close();

	if ( m_keyPktSize == 0 )
		return ppe_ok;

	int size = 0;

	int res = pgp_EnumKeyIDs( NULL, &size, m_keyPkt, m_keyPktSize, m_secret );

	if ( res == ppe_bufferTooSmall )
	{
		if ( m_keyIDtbl.Allocate( size ) == NULL )
			return ppe_memoryAllocation;

		res = pgp_EnumKeyIDs( m_keyIDtbl.buffer, &size, m_keyPkt, m_keyPktSize, m_secret );
	}

	if ( res != ppe_ok )
	{
		MessageBoxT( m_hDlg, _LoadIDS( IDS_KRIsInvalid ), _LoadIDS( IDS_BrowsePKUser ), MB_ICONSTOP );
		return res;
	}

	LVITEMW lvi;
	memset( &lvi, 0, sizeof( lvi ) );
	lvi.mask = LVIF_IMAGE;
	lvi.iImage = 1;

	int pos;
	for( pos = 0; pos < m_keyIDtbl.bufferSize; pos += pgp_keyIDSize )
	{
		ListView_InsertItemT( m_hList, &lvi );
	}

	int item = 0;
	for( pos = 0; pos < m_keyIDtbl.bufferSize; pos += pgp_keyIDSize )
	{
		SetPGPItemValues( m_keyPkt, m_keyPktSize, item, m_keyIDtbl.buffer + pos );
		item++;
	}

	return ppe_ok;
}

//
//	End of PGP list view service
//	
/////////////////////////

////////////////////////
//
//	Certificate and PKCS - #12 list view service
//

static void SetItemValues( HWND hList, int item, wchar_t **subItemT, int subItemQ )
{
	for( int subItem = subItemQ; subItem; subItem -- )
		ListView_SetItemTextT( hList, item, subItem, subItemT[ subItem - 1 ] ); 
}

const int subItemQ = 4;

int cDlgViewLocalDatabase::AddLVItem( const wchar_t *path, WIN32_FIND_DATAW *findData, int itemNum )
{
	LVITEMW lvi;
	memset( &lvi, 0, sizeof( lvi ) );
	lvi.mask = LVIF_IMAGE | LVIF_TEXT;
	lvi.iItem = itemNum;

	wchar_t *subItemT[ subItemQ ];
	wchar_t userName[ MAX_PATH ] = L"";
	wchar_t algDesc[ MAX_PATH ] = L"";
	wchar_t keyID[ MAX_PATH ] = L"";
	wchar_t thumbprint[ MAX_PATH ] = L"";
	wchar_t fileName[ 1024 ];

	subItemT[0] = algDesc;
	subItemT[1] = keyID;
	subItemT[2] = thumbprint;
	subItemT[3] = fileName;

	wcscpy_s( fileName, findData->cFileName );

	lvi.iImage = m_iconCache.GetIconNum( path, NULL );

	if ( ( findData->dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) == 0 )
	{
		cFileView fv( path );
		cUniversalString friendlyName;
		cPTB ptbLocalKeyID;
		cPTB ptbThumbprint;
		wchar_t errStr[ 1024 ];
		LoadStringT( m_hInst, IDS_NoError, errStr, maxLen( errStr ) );

		HPKA pka = NULL;

		//
		//	19.11.2015
		//	New version of PFX PDU packed contains thumbprint instead of public key local key ID. We'll treat it always as thumbrint.
		//	If certificate is found so BCDB_GetCerPublicKeyInfo recalculated both this values corectly.
		//
		bool res = BCDB_GetP12SecretKeyInfo(
			m_hInst, m_hDlg,
			fv.m_ptr, (int)fv.m_size.QuadPart,
			NULL,
			&friendlyName, &ptbThumbprint, //ptbLocalKeyID,
			errStr, maxLen( errStr )
			);

		if ( res )
		{
			// The secret key file was succefully read, so we try to find corresponded
			//	certificate file to get algorithm descryption ( secret key is encrypted and has no
			//	opened descryption )

			cPTB ptbCrt;

			bool lRes = BCDB_GetX509KeyByID( &ptbCrt, ptbThumbprint.buffer, ptbThumbprint.bufferSize, false, NULL );

			if ( lRes )
			{
				if ( !BCDB_GetCerPublicKeyInfo(
						m_hInst, m_hDlg,
						ptbCrt.buffer, ptbCrt.bufferSize,
						&pka, NULL,
						&ptbLocalKeyID, &ptbThumbprint )
					)
						wcscpy_s( algDesc, _LoadIDS( IDS_Undefined ) );
			}
		}
		else // Tries to read the file as certificate
		{
			res = BCDB_GetCerPublicKeyInfo(
				m_hInst, m_hDlg,
				fv.m_ptr, (int)fv.m_size.QuadPart,
				&pka, &friendlyName,
				&ptbLocalKeyID, &ptbThumbprint );
		}

		//
		//	Gets algorithm description from public key algorithm handle.
		//
		if ( pka != NULL )
		{
			int algID = pka_getAlgID( pka );
			int bitSize = pka_getModuleSize( pka ) * 8;
			pka_free( pka );

			FormatT( algDesc, maxLen( algDesc ), _LoadIDS( IDS_NameAndBits_SD ), cWCStr( PKT_GetPublicKeyAlgName( (BYTE)algID ) ).m_szBody, bitSize );
		}

		if ( !res )
		{
			SHFILEINFOW sfi;
			if ( SHGetFileInfoT( path, 0, &sfi, sizeof( sfi ), SHGFI_DISPLAYNAME | SHGFI_TYPENAME ) != 0 )
			{
				wcsncpy( userName, sfi.szTypeName, maxLen( sfi.szTypeName ) );
			}
			else
				wcsncpy( userName, errStr, maxLen( userName )  );
		}
		else
			wcsncpy( userName, friendlyName.wcBuffer, maxLen( userName ) );

		int size = maxLen( keyID );
		bytes2string( keyID, &size, ptbLocalKeyID.buffer, ptbLocalKeyID.bufferSize );
		
		size = maxLen( thumbprint );
		bytes2string( thumbprint, &size, ptbThumbprint.buffer, ptbThumbprint.bufferSize );
	}
	
	lvi.pszText = userName;

	int i = ListView_InsertItemT( m_hList, &lvi );

	SetItemValues( m_hList, i, subItemT, subItemQ );

	return i;
}

int cDlgViewLocalDatabase::FillListView( bool secret )
{
	m_iconCache.SetImageList( m_hList );
	m_secret = secret;

	wchar_t path[ MAX_PATH ];
	if ( !BCDB_GetDataBasePath( path, maxLen( path ) ) )
		return x509e_internalError;

	int len = wcslen( path );
	if ( path[ len - 1 ] != L'\\')
	{
		path[ len++ ] = L'\\';
		path[ len ] = L'\0';
	}

	// Fills directory first
	wcsncpy( path + len, m_secret ? L"*.p12" : L"*.cer", maxLen( path ) - len );

	WIN32_FIND_DATAW  findData;
	memset( &findData, 0, sizeof( findData ) );

	HANDLE hFind = FindFirstFileT( path, &findData );
	if ( hFind == INVALID_HANDLE_VALUE )
		return x509e_internalError;

	int lastDirItem = 0;

	do
	{
		wcsncpy( path + len, findData.cFileName, maxLen( path ) - len );

		AddLVItem( path, &findData, 0x7FFFffff ); // Add to the list end
	}
	while( FindNextFileT( hFind, &findData ) != 0 );

	FindClose( hFind );

	return x509e_ok;
}

//
//	End of Certificate and PKCS - #12 list view service
//	
/////////////////////////

int cDlgViewLocalDatabase::RefreshView( eViewType viewType )
{
	m_viewType = viewType;
	return FillListView();
}

int cDlgViewLocalDatabase::SelectTreeView( eViewType viewType )
{
	m_viewType = viewType;

	HTREEITEM hItem = TreeView_GetRoot( m_hTree );

	for( ; hItem; hItem = TreeView_GetNextSibling( m_hTree, hItem ) ) 
	{
		wchar_t text[ 100 ];

		TVITEMW tvi;
		memset( &tvi, 0, sizeof( tvi ) );
		tvi.mask = TVIF_TEXT | TVIF_PARAM;
		tvi.hItem = hItem;
		tvi.pszText = text;
		tvi.cchTextMax = maxLen( text );

		TreeView_GetItemT( m_hTree, &tvi );
		if ( tvi.lParam == viewType )
		{
			const wchar_t *caption = m_caption;
			cWCStr wcs;
			if ( !m_useTree )
				caption = wcs.FormatT( L"%s : %s", m_caption, tvi.pszText );

			SetWindowTextT( m_hDlg, caption );

			TreeView_SelectItem( m_hTree, hItem );
			return 1;
		}
	}

	return 0;
}

int cDlgViewLocalDatabase::FillListView()
{
	int res = FillListView_internal();

	ToolBarUpdateButtons();

	return res;
}

int cDlgViewLocalDatabase::FillListView_internal()
{
	LV_DeleteAllItems( m_hList );

	SelectTreeView( m_viewType );

	switch ( m_viewType )
	{
		case vt_crtPublicKey: return FillListView( false );
		case vt_p12SecretKey: return FillListView( true );
		case vt_pgpPublicKey: return FillListView_PGP( false );
		case vt_pgpSecretKey: return FillListView_PGP( true );
	}

	return ppe_internalError;
}

DWORD cDlgViewLocalDatabase::GetIDSbyCmd( int cmd )
{
	switch ( cmd )
	{
		case ID_KEY_DELETE:				return IDS_szDelete;
		case IDM_Find :					return IDS_szFind;
		case IDM_VIEW :					return IDS_szViews;
		case ID_KEY_GENERATENEWPUBLIC:	return IDS_szGenerateNew;
		case ID_KEY_SENDPUBLICKEYTOE:	return IDS_szSendPublicKeyToEMailRecipient;
		case ID_KEY_PROPERTIES:			return IDS_szKeyProperties;
		case ID_HELP_ABOUT:				return IDS_szAbout;

		default : return NULL;
	}
}

static sImageInfo siiT[ 6 ];

static void setImageInfo( int tPos, int iCmd, HIMAGELIST hIml, int iImage )
{
	if ( maxLen( siiT ) < tPos )
	{
		assert( false );
		return;
	}

	siiT[ tPos ].iCmd = iCmd;
	siiT[ tPos ].hIml = hIml;
	siiT[ tPos ].iImage = iImage;
}

static void tbAddButtons( HWND hToolbar, int imageWidth = 16 )
{
	if ( hToolbar == NULL )
		return;

	//
	//	We have to set the extended style for toolbar...
	//
	SendMessage( hToolbar, TB_SETEXTENDEDSTYLE, 0 , TBSTYLE_EX_DRAWDDARROWS | WS_EX_TRANSPARENT );
	SendMessage( hToolbar, TB_BUTTONSTRUCTSIZE, sizeof( sizeof(TBBUTTON) ), 0 );
	SendMessage( hToolbar, TB_SETBITMAPSIZE, 0, (LPARAM) MAKELONG ( imageWidth, imageWidth ) );

	// Add the bitmap containing button images to the toolbar. 

	TBADDBITMAP tbab;
	memset( &tbab, 0, sizeof( tbab ) );

	//tbab.hInst = HINST_COMMCTRL ; 
	//tbab.nID   = imageWidth == 16 ? IDB_STD_SMALL_COLOR : IDB_STD_LARGE_COLOR; 
	//int idbStd = SendMessage( hToolbar, TB_ADDBITMAP, (WPARAM) 0, (WPARAM) &tbab); 

	tbab.hInst = HINST_COMMCTRL ; 
	tbab.nID   = imageWidth == 16 ? IDB_VIEW_SMALL_COLOR : IDB_VIEW_LARGE_COLOR; 
	int idbView = SendMessage( hToolbar, TB_ADDBITMAP, (WPARAM) 0, (WPARAM) &tbab); 

	tbab.hInst = HINST_COMMCTRL ; 
	tbab.nID   = imageWidth == 16 ? IDB_HIST_SMALL_COLOR : IDB_HIST_LARGE_COLOR; 
	int idbHist = SendMessage( hToolbar, TB_ADDBITMAP, (WPARAM) 0, (WPARAM) &tbab); 

	//
	//	Tool bar button notification commands...
	//
	ToolBar_AddButton( hToolbar, 0						  , ID_KEY_GENERATENEWPUBLIC );
//	ToolBar_AddButton( hToolbar, 1						  , IDM_Find	, TBSTYLE_DROPDOWN );
	ToolBar_AddButton( hToolbar, 2						  , ID_KEY_SENDPUBLICKEYTOE );
	ToolBar_AddButton( hToolbar, 5						  , ID_KEY_PROPERTIES ); // idbStd + STD_PROPERTIES, IDM_PROPERTIES  );
	ToolBar_AddSep( hToolbar );
	ToolBar_AddButton( hToolbar, 4					      , ID_KEY_DELETE ); 
	ToolBar_AddSep( hToolbar );
	ToolBar_AddButton( hToolbar, idbView + VIEW_DETAILS   , cDlgViewLocalDatabase::IDM_VIEW	, TBSTYLE_DROPDOWN);
	ToolBar_AddButton( hToolbar, 3				          , ID_HELP_ABOUT  );
}

int cDlgViewLocalDatabase::InitToolBar()
{
	ODSFunc;

	CSysDll comctrlDll( false, "comctl32.dll" );

	bool useTransparentBMP = comctrlDll.version.dwMajorVersion >= 6;

#ifdef _DEBUG
//	useTransparentBMP = false;
#endif

	DWORD dwStyle = TBSTYLE_FLAT | 
			TBSTYLE_TOOLTIPS |
			TBSTYLE_TRANSPARENT |
			CCS_ADJUSTABLE |
//			CCS_NODIVIDER |
			WS_CHILD | WS_VISIBLE;
	
	int iconQ = 6; // number of tools inside the bitmap

	m_hToolbar16 = CreateToolbarEx( m_hDlg, dwStyle, IDC_TOOLBAR, iconQ, m_hInst, useTransparentBMP ? IDB_ToolbarXP : IDB_Toolbar, NULL, 0,0,0,0,0, sizeof(TBBUTTON)) ;
	if ( m_hToolbar16 == NULL )
		return false;

	Toolbar_SetUnicodeFormat( m_hToolbar16 );

	m_hToolbar24 = CreateToolbarEx( m_hDlg, dwStyle, IDC_TOOLBAR, iconQ, m_hInst, IDB_ToolbarXP_24, NULL, 0,0,0,0,0, sizeof(TBBUTTON)) ;

	if ( m_hToolbar24 == NULL )
		return false;

	Toolbar_SetUnicodeFormat( m_hToolbar24 );

	tbAddButtons( m_hToolbar24, 24 );
	tbAddButtons( m_hToolbar16, 16 );

	HIMAGELIST hImlTB = NULL;
	if ( m_hToolbar16 != NULL )
		hImlTB = Toolbar_GetImageList( m_hToolbar16 );

	if ( hImlTB != NULL )
	{
		int i = 0;

		setImageInfo( i++, ID_KEY_GENERATENEWPUBLIC, hImlTB, 0 );
		setImageInfo( i++, ID_KEY_SENDPUBLICKEYTOE, hImlTB, 2 );
		setImageInfo( i++, ID_HELP_ABOUT, hImlTB, 3 );
		setImageInfo( i++, ID_KEY_PROPERTIES, hImlTB, 5 );
		setImageInfo( i++, ID_KEY_DELETE, hImlTB, 4 );

		m_siiT = siiT;
		m_siiQ = i;
	}

	ToolBarUpdateSize();
	ToolBarUpdateButtons();

	return TRUE;
}

void cDlgViewLocalDatabase::ToolBarUpdateSize()
{
	// Moves list views below toolbar.

	bool bUseTB24 = m_bLargeToolbar;
	m_hToolbar = bUseTB24 ? m_hToolbar24 : m_hToolbar16;

	ShowWindow( m_hToolbar16, SW_HIDE );
	ShowWindow( m_hToolbar24, SW_HIDE );
	ShowWindow( m_hToolbar, SW_SHOW );

	RECT rect;
	GetWindowRect( m_hToolbar, &rect );
	int tbSize = rect.bottom - rect.top;
	
	if ( bUseTB24 )
		tbSize = 36;

	GetDlgItemRect( m_hTree, &rect );
	rect.top = tbSize;
	SetWindowPos( m_hTree, NULL, rect.left, rect.top, rect.right - rect.left, rect.bottom - rect.top, SWP_NOZORDER|SWP_SHOWWINDOW );

	GetDlgItemRect( m_hList, &rect );
	rect.top = tbSize;
	SetWindowPos( m_hList, NULL, rect.left, rect.top, rect.right - rect.left, rect.bottom - rect.top, SWP_NOZORDER|SWP_SHOWWINDOW );

	if ( m_splitter != NULL )
	{
		cRect r;
		GetDlgItemRect( m_splitter->m_hWnd, &r );
		int dy = r.top - tbSize;
		//GetWindowRect( m_splitter->m_hWnd, &r );
		r.top -= dy;
		SetWindowPos( m_splitter->m_hWnd, NULL, r.left, r.top, r.Width(), r.Height(), SWP_NOZORDER );
	}
}

static void tbEnableButtons( HWND hToolbar, bool bEnable )
{
	ToolBar_EnableButton( hToolbar, ID_KEY_DELETE, bEnable );
	ToolBar_EnableButton( hToolbar, ID_KEY_PROPERTIES, bEnable );
	ToolBar_EnableButton( hToolbar, ID_KEY_SENDPUBLICKEYTOE, bEnable );
}

void cDlgViewLocalDatabase::ToolBarUpdateButtons()
{
	bool enable = ( ListView_GetSelectedCount( m_hList ) != 0 );
	tbEnableButtons( m_hToolbar, enable );
}

static const wchar_t *szDBView = L"DBView";

static DWORD GetLVValues( HWND hList, BYTE *pkt, int size )
{
	if ( size < 0 )
		return LVS_REPORT;

	LONG styles = *( ( int * )pkt ); pkt += sizeof( int ); size -= sizeof( int );

	int columnSize;
	int col = 0;

	for( ; size > 0 ; col++ )
	{
		columnSize = *( ( int * )pkt ); pkt += sizeof( int ); size -= sizeof( int );
		ListView_SetColumnWidth( hList, col, columnSize );
	}

	LV_setStyle( hList, styles );

	return styles;
}

struct tvTabItem { int ids; int cmd; };

tvTabItem tvT[] = {
	IDS_szP12, cDlgViewLocalDatabase::vt_p12SecretKey,
	IDS_szX509, cDlgViewLocalDatabase::vt_crtPublicKey,
	IDS_szPGP_PK, cDlgViewLocalDatabase::vt_pgpPublicKey
};

void cDlgViewLocalDatabase::ShowTreeView()
{
	SelectTreeView( m_viewType );

	RECT r, rSplitter;

	GetDlgItemRect( m_hList, &r );
	GetDlgItemRect( m_splitter->m_hWnd, &rSplitter );

	if ( !m_useTree )
	{
		ShowWindow( m_hTree, SW_HIDE );
		ShowWindow( m_splitter->m_hWnd, SW_HIDE );
		r.left = 0;
		SetWindowPos( m_hList, 0, r.left, r.top, r.right - r.left, r.bottom - r.top, SWP_NOZORDER | SWP_SHOWWINDOW );
		return;
	}

	ShowWindow( m_hTree, SW_SHOW );
	ShowWindow( m_splitter->m_hWnd, SW_SHOW );
	r.left = rSplitter.right;
	SetWindowPos( m_hList, 0, r.left, r.top, r.right - r.left, r.bottom - r.top, SWP_NOZORDER | SWP_SHOWWINDOW );
}

BOOL cDlgViewLocalDatabase::FillTreeView()
{

//	PKCS #12 Secret Keys 
//	X.509 Certificates Public Keys
//	PGP Public Keys

	TVINSERTSTRUCTW is;
	memset( &is, 0, sizeof( is ) );

	is.hParent = TVI_ROOT;
	is.hInsertAfter = TVI_FIRST;
	is.item.mask = TVIF_TEXT | TVIF_PARAM;

	for ( int i = sizeof( tvT ) / sizeof( tvT[0] ); i--; )
	{
		cWCStr wcs;
		wcs.LoadStringT( IH.m_hInst, tvT[i].ids );
		is.item.pszText = wcs.m_szBody;
		is.item.lParam  = tvT[i].cmd;
		TreeView_InsertItemT( m_hTree, &is );
	}

	return FALSE;
}

class cDlgViewValues : public cRegDlgValues
{
public:
	int viewType;
	bool useTree;
	bool useSmallToolbar; // The large toolbar is preferable, but due to packing the place was reserved and filde as zero.

	cDlgViewValues( HKEY hRoot, const wchar_t *keyName, const wchar_t *valueName, int totalSubClassSize, int columns = 0 ):
		cRegDlgValues( hRoot, keyName, valueName, totalSubClassSize, columns )
	{
		useSmallToolbar = false;
	}

} dlgVals( hKeyRoot, szBCKeyManKey, szDBView, sizeof(cDlgViewValues), NUM_COLUMNS ); 

static HICON LoadIconEx( HINSTANCE hInstance, LPCTSTR lpIconName, int cx, int cy )
{
	return (HICON)LoadImage( hInstance, lpIconName, IMAGE_ICON, cx, cy, LR_DEFAULTCOLOR );
}

BOOL cDlgViewLocalDatabase::InitDialog()
{
	ODSFunc;

	SendMessage( m_hDlg, WM_SETICON, ICON_SMALL, (LPARAM)LoadIconEx( m_hInst, MAKEINTRESOURCE( IDI_Main ), 16, 16 ) );
	SendMessage( m_hDlg, WM_SETICON, ICON_BIG,   (LPARAM)LoadIconEx( m_hInst, MAKEINTRESOURCE( IDI_Main ), 32, 32 ) );

	GetWindowTextT( m_hDlg, m_caption, maxLen( m_caption ) );

	m_minSize.x = 120;
	m_minSize.y = 20;

	m_hList = GetDlgItem( m_hDlg, IDC_List );
	m_hTree = GetDlgItem( m_hDlg, IDC_Tree );

	InitListViewColumns( m_hInst, m_hList );

	m_columnQ = NUM_COLUMNS;

	CSysDll shell32( false, "shell32.dll" );
	bool useXPIcon = ( shell32.version.dwMajorVersion >= 6 );

	int xm = GetSystemMetrics(SM_CXSMICON) * 2;
	int ym = GetSystemMetrics(SM_CYSMICON);

	m_himlPGP = ImageList_Create( xm, ym, ILC_MASK | ( useXPIcon ? ILC_COLOR32 : ILC_COLOR16 ), 4, 10 );
	ImageList_AddIcon( m_himlPGP, LoadIconEx( m_hInst, MAKEINTRESOURCE( IDI_PKUserEmpty ), xm, ym ) );
	ImageList_AddIcon( m_himlPGP, LoadIconEx( m_hInst, MAKEINTRESOURCE( IDI_PkUserCollapsed ), xm, ym ) );
	ImageList_AddIcon( m_himlPGP, LoadIconEx( m_hInst, MAKEINTRESOURCE( IDI_PKUserExpanded ), xm, ym ) );
	ImageList_AddIcon( m_himlPGP, LoadIconEx( m_hInst, MAKEINTRESOURCE( IDI_SubKey ), xm, ym ) );


	m_himlPGPLarge = ImageList_Create(
			GetSystemMetrics(SM_CXICON),
			GetSystemMetrics(SM_CXICON),
			ILC_MASK | ( useXPIcon ? ILC_COLOR32 : ILC_COLOR16 ) ,
			4, 10 );

	enum { id_users = 220, id_xpUsers = 269 };

	if ( shell32.error )
	{
		assert( false );
	}
	else
	{	
		ImageList_AddIcon( m_himlPGPLarge, LoadIcon( shell32.hInst, MAKEINTRESOURCE(id_users) ) );
		ImageList_AddIcon( m_himlPGPLarge, LoadIcon( shell32.hInst, MAKEINTRESOURCE(id_users) ) );
		ImageList_AddIcon( m_himlPGPLarge, LoadIcon( shell32.hInst, MAKEINTRESOURCE(id_users) ) );
	}

	m_iconCache.Init( m_hInst, m_hList, 0, 0 );

	InitToolBar();

	m_splitter = new cSplitter( m_hInst, hKeyRoot, szBCKeyManKey );
	assert( m_splitter->m_hInst == m_hInst );

	m_splitter->Init( m_hDlg, m_hTree, m_hList );
	
	m_sizedCtrls.AddCtrl( IDC_List, cSizedCtrls::ptSDR );
	m_sizedCtrls.AddCtrl( IDC_Tree, cSizedCtrls::ptSD );
	m_sizedCtrls.AddCtrl( m_splitter->m_hWnd, cSizedCtrls::ptSD );
	m_sizedCtrls.AddCtrl( m_hToolbar16, cSizedCtrls::ptSR );
	m_sizedCtrls.AddCtrl( m_hToolbar24, cSizedCtrls::ptSR );

	// Read saved values

	m_viewType = vt_crtPublicKey;

	dlgVals.Restore( m_hDlg, m_hList ); //, (BYTE *)&m_viewType, sizeof( m_viewType ) );
	m_viewType = (eViewType)dlgVals.viewType;
	m_useTree = dlgVals.useTree;

	m_bLargeToolbar = !dlgVals.useSmallToolbar;

	ToolBarUpdateSize();

	if ( m_pKFF != NULL )
	{
		if ( *m_pKFF == kffPGP )
			m_viewType = vt_pgpPublicKey;
		else
		if ( GetKeyOnly() && ( *m_pKFF == kffX509 ) && ( ( m_outFormat & vld_flag_viewPublic ) == 0 ) )
			m_viewType = vt_p12SecretKey;
	}

    
	FillTreeView();

	FillListView();

	ShowTreeView();

//	SetWindowPos( m_splitter->m_hWnd, HWND_TOP, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_SHOWWINDOW );
//	ShowWindow( m_hList, SW_HIDE );
//	ShowWindow( m_hTree, SW_HIDE );

	return TRUE;
}

void cDlgViewLocalDatabase::DeleteDialog()
{
	dlgVals.viewType = m_viewType;
	dlgVals.useTree = m_useTree;
	dlgVals.useSmallToolbar = !m_bLargeToolbar;
	dlgVals.Save( m_hDlg, m_hList ); //, (BYTE *)&m_viewType, sizeof( m_viewType ) );

	LV_DeleteAllItems( m_hList );
	// Version 1.03.6 - 12.10.2011
	// m_iconCache.DestroySharedImageLists();

	if ( m_himlPGP      ) ImageList_Destroy( m_himlPGP );
	if ( m_himlPGPLarge ) ImageList_Destroy( m_himlPGPLarge );

	if ( GetKeyOnly() ) // See Note :: GetKeyOnly 4.
		return;

	if ( m_splitter )
		delete m_splitter;
}

bool cDlgViewLocalDatabase::GetKeyOnly()
{
	return ( m_ptbKeyOut != NULL ) && ( m_pKFF != NULL );
}

int cDlgViewLocalDatabase::OnGetKey()
{
	assert( m_pKFF != NULL );

	if ( m_pKFF == NULL )
		return false;

	bool pgpView = m_viewType == vt_pgpPublicKey;

	*m_pKFF = pgpView ? kffPGP : kffX509 ;

	if ( !GetSelected( m_ptbKeyOut, m_wcsUserID ) )
		return false;

	//
	//	Find corresponded public key for choosen secret key
	//

	bool secretKeyRequiered = ( m_outFormat & vld_flag_viewSecret ) != 0;

	if ( ( m_viewType != vt_p12SecretKey ) || secretKeyRequiered  ) 
		return true;

	cUniversalString friendlyName;
	cPTB ptbLocalKeyID;
	wchar_t errStr[ 1024 ];
	LoadStringT( m_hInst, IDS_NoError, errStr, maxLen( errStr ) );

	HPKA pka = NULL;

	bool lRes = BCDB_GetP12SecretKeyInfo(
		m_hInst, m_hDlg,
		m_ptbKeyOut->buffer, m_ptbKeyOut->bufferSize,
		NULL,
		&friendlyName, &ptbLocalKeyID,
		errStr, maxLen( errStr )
		);

	if ( !lRes )
	{
		DlgMessageBoxT( errStr, MB_ICONSTOP );
		return false;
	}
	// The secret key file was succefully read, so we try to find corresponded
	//	certificate file to get algorithm descryption ( secret key is encrypted and has no
	//	opened descryption )

	wchar_t filePath[ MAX_PATH ];

	lRes = BCDB_GetX509KeyByID(
			m_ptbKeyOut,
			ptbLocalKeyID.buffer, ptbLocalKeyID.bufferSize,
			false, NULL,
			filePath
		);

	if ( !lRes )
	{
		DlgMessageBoxT( IDS_CorrespondedPKCertificateNotFound, MB_ICONSTOP );
		return false;
	}

	cFileView fv( filePath );

	if ( fv.m_err )
	{
		DlgMessageBoxT( IDS_szErrorInKeyReading, MB_ICONSTOP );
		return false;
	}
	
	m_ptbKeyOut->WriteBytes( fv.m_ptr, (int)fv.m_size.QuadPart, true );

	return true;
}

int cDlgViewLocalDatabase::TreeViewNotify( NMHDR *pnmh )
{
	switch( pnmh->code )
	{
		case TVN_SELCHANGEDA :
		{
			NMTREEVIEWA *nmtv = (NMTREEVIEWA *)pnmh;
			m_viewType = (eViewType)(nmtv->itemNew.lParam);
			if ( nmtv->action != TVC_UNKNOWN )
				FillListView();
			return 1;
		}
		case TVN_SELCHANGEDW :
		{
			NMTREEVIEWW *nmtv = (NMTREEVIEWW *)pnmh;
			m_viewType = (eViewType)nmtv->itemNew.lParam;
			if ( nmtv->action != TVC_UNKNOWN )
				FillListView();
			return 1;
		}
		default:
			break;
	}

	return 0;
}

int cDlgViewLocalDatabase::ListViewNotify( NMHDR *pnmh )
{
	switch( pnmh->code )
	{
		case LVN_ITEMCHANGED:
		case LVN_ITEMACTIVATE :
			ToolBarUpdateButtons();
			return 0;	//lvSelChanged();
		case LVN_KEYDOWN:
		{
			NMLVKEYDOWN *kd = ( NMLVKEYDOWN * )pnmh;

			if ( kd->wVKey == VK_DELETE )
				OnCommand( ID_KEY_DELETE, 0 );

			break;
		}
		case NM_DBLCLK:
			if ( GetKeyOnly() )
			{
				if ( OnGetKey() )
					EndDialog( m_hDlg, IDOK );

				return true;
			}
			else
				return OnKeyProperties();

		case NM_CLICK:
		{
			NMITEMACTIVATE *nmia = ( NMITEMACTIVATE * )pnmh;
			return checkHitPlus( nmia->ptAction );
		}
		case NM_RCLICK:
		{
			NMITEMACTIVATE *nmia = ( NMITEMACTIVATE * )pnmh;

			POINT pt;
			GetCursorPos( &pt );
			MapWindowPoints( HWND_DESKTOP, m_hList, (LPPOINT)&pt, 1 );

			LVHITTESTINFO hti;
			memset( &hti, 0, sizeof( hti ) );

			hti.flags = LVHT_ONITEMICON;
			hti.pt = pt;

			int item = ListView_HitTest( m_hList, &hti );

			if ( item == -1 )
				return FALSE;

			MapWindowPoints( m_hList, HWND_DESKTOP, (LPPOINT)&hti.pt, 1 );

			HMENU hPopupMenu;

			int menuID = 2;
			hPopupMenu = cMenuItem::LoadPopupMenu( m_hInst, IDR_MENU, menuID, m_siiT, m_siiQ );

			SetMenuDefaultItem( hPopupMenu, ID_KEY_PROPERTIES, FALSE );

			cMenuItem::TrackPopupMenu(
				hPopupMenu,
				TPM_LEFTALIGN|TPM_LEFTBUTTON|TPM_VERTICAL,
				hti.pt.x,
				hti.pt.y,
				m_hDlg,
				NULL);

			break;
		}
		default : break;
	}

	return 0;
}

int cDlgViewLocalDatabase::ToolBarNotify( NMHDR *pnmh )
{
	LPNMHDR lpnm  = (LPNMHDR)pnmh;
	LPNMTOOLBAR lpnmTB = (LPNMTOOLBAR)pnmh;

	switch(lpnm->code)
	{
		//case TTN_GETDISPINFO:
		//{
		//	lpnmtdi->uFlags = TTF_DI_SETITEM; // 0;
		//	lpnmtdi->hinst = m_hInst;
		//	lpnmtdi->lpszText = MAKEINTRESOURCE( GetIDSbyCmd( lpnmtdi->hdr.idFrom ) );

		//	return TRUE;
		//}
		case TTN_GETDISPINFOW:
			return GetDispInfo( m_hInst, (LPNMTTDISPINFOW)pnmh, GetIDSbyCmd( pnmh->idFrom ) );

		case TTN_GETDISPINFOA:
			return GetDispInfo( m_hInst, (LPNMTTDISPINFOA)pnmh, GetIDSbyCmd( pnmh->idFrom ) );

		case NM_CLICK:
			if ( lpnmTB->iItem != IDM_Find )
			{
				break;
			}
			// else continue as dropdown 

		case TBN_DROPDOWN:
		{
			int menuID = -1;

			RECT rc;
			SendMessage( lpnmTB->hdr.hwndFrom, TB_GETRECT, (WPARAM)lpnmTB->iItem, (LPARAM)&rc );

			MapWindowPoints( lpnmTB->hdr.hwndFrom, HWND_DESKTOP, (LPPOINT)&rc, 2 );

			HMENU hPopupMenu;

			if ( lpnmTB->iItem == IDM_VIEW )
			{
				menuID = 1;
				hPopupMenu = cMenuItem::LoadPopupMenu( m_hInst, IDR_MENU, menuID, m_siiT, m_siiQ );
			}
			else
			if ( lpnmTB->iItem == IDM_Find )
			{ // Loads the popup menu from Main menu->Keys->Add an existing public keys
				HMENU hmParent = cMenuItem::LoadPopupMenu( m_hInst, IDR_BCDB_MENU, 1, m_siiT, m_siiQ );
				hPopupMenu = GetSubMenu(hmParent, 1 );
				RemoveMenu(hmParent, 1, MF_BYPOSITION);
				DestroyMenu(hmParent);
			}
			else
				return false;

			cMenuItem::TrackPopupMenu(
				hPopupMenu,
				TPM_LEFTALIGN|TPM_LEFTBUTTON|TPM_VERTICAL,
				rc.left,
				rc.bottom,
				m_hDlg,
				NULL);

			return (FALSE);
		}
		default :
			return FALSE;
	}

	return FALSE;
}

int cDlgViewLocalDatabase::LV_SetStyle( HWND hList, DWORD idCmd )
{
	DWORD dwlStyle = GetWindowLong( hList, GWL_STYLE);

	int style = dwlStyle & ~LVS_TYPEMASK;

	DWORD viewStyle = dwlStyle & LVS_TYPEMASK;

	switch ( idCmd )
	{
		case ID_ViewIcon:		style |= LVS_ICON;		break;
//		case ID_ViewSmallIcon:	style |= LVS_SMALLICON;	break;
		case ID_ViewDetails:	style |= LVS_REPORT;	break;
		case ID_ViewList:		style |= LVS_LIST;		break;

		case IDM_VIEW:	style |= ( ( viewStyle + 1 ) & LVS_TYPEMASK );
						break;

		default : return FALSE;
	}

	return SetWindowLong( hList, GWL_STYLE, style );
}

static bool LV_hasSelection( HWND hList )
{
	bool enable = ( hList != NULL ) && ( ListView_GetSelectedCount( hList ) != 0 );
	return enable;
}

BOOL cDlgViewLocalDatabase::OnInitMenuPopup( HMENU hmInit, int nIndex, BOOL fSystemMenu )
{
	//{
	//	cWSABlock wsa;
	//	int err = wsa.getError();

	//	bool enable = ( err == 0 );

	//	_EnableMenuItem( hmInit, ID_ADDEXISTINGPUBLICKEY_SEARCHINTHEINTERNET, enable );
	//}

	HMENU m_cmCur = hmInit;

	if (!m_cmCur)
	{
		return(TRUE);
	}
//
// GetMenuItemInfo does not work for Windows 95
//
//	MENUITEMINFO mii;
//	memset( &mii, 0, sizeof( mii ) );
//	mii.cbSize = sizeof(MENUITEMINFO);
//	mii.fMask = MIIM_ID;
//
//	nIndex = 0;
//	while (GetMenuItemInfo(m_cmCur, nIndex++, TRUE, &mii)  )
//	{
//		UINT wID = mii.wID;

	nIndex = GetMenuItemCount( m_cmCur );

	while ( nIndex )
	{
		UINT wID = GetMenuItemID( m_cmCur, --nIndex );
		switch ( wID )
		{
			case ID_KEY_DELETE :
			case ID_KEY_PROPERTIES :
			case ID_KEY_SENDPUBLICKEYTOE:
			case ID_KEY_COPYPUBLICKEYTOCLIPBOARD:
				_EnableMenuItem( hmInit, wID, LV_hasSelection( m_hList ) );
				break;
			case ID_ADDEXISTINGPUBLICKEY_SEARCHINTHEINTERNET:
			{
				// Version 1.05.2.7 - 24.10.2018
				//		2. Bug #4169 Remove "add public key from the internet" option

				RemoveMenu( hmInit, ID_ADDEXISTINGPUBLICKEY_SEARCHINTHEINTERNET, MF_BYCOMMAND );

/*
				cWSABlock wsa;
				int err = wsa.getError();

				bool enable = ( err == 0 );

				_EnableMenuItem( hmInit, wID, enable );
*/
				break;
			}
			case ID_KEY_FINDNEXT:
				_EnableMenuItem( hmInit, wID, IsFindParamsEmpty() );
				break;

			case ID_ViewIcon:		_CheckMenuItem( hmInit, wID, ( LV_getStyle( m_hList ) == LVS_ICON ) );		break;
			case ID_ViewDetails:	_CheckMenuItem( hmInit, wID, ( LV_getStyle( m_hList ) == LVS_REPORT ) );	break;
			case ID_ViewList:		_CheckMenuItem( hmInit, wID, ( LV_getStyle( m_hList ) == LVS_LIST ) );		break;
			case ID_LargeToolbar:	_CheckMenuItem( hmInit, wID, m_bLargeToolbar );		break;

			case ID_VIEW_PGP_PUBLICKEYS:
				_CheckMenuItem( hmInit, wID, ( m_viewType == vt_pgpPublicKey ) );
				_EnableMenuItem( hmInit, wID, !GetKeyOnly() || ( m_outFormat & vld_flag_viewPublic ) );
				break;
			case ID_VIEW_PUBLICKEYS:
				_CheckMenuItem( hmInit, wID, ( m_viewType == vt_crtPublicKey ) );
				_EnableMenuItem( hmInit, wID, !GetKeyOnly() || ( m_outFormat & vld_flag_viewPublic ) );
				break;
			case ID_VIEW_SECRETKEYS:
				_CheckMenuItem( hmInit, wID, ( m_viewType == vt_p12SecretKey ) );
				_EnableMenuItem( hmInit, wID, !GetKeyOnly() || ( ( m_outFormat & vld_flag_viewSecret ) != 0 ) );

				// See Note :: GetKeyOnly 1.
//				_EnableMenuItem( hmInit, wID, !GetKeyOnly() || ( ( m_outFormat & vld_flag_viewPublic ) == 0 ) );
				break;
			case ID_VIEW_SHOWTYPEVIEW:
				_CheckMenuItem( hmInit, wID, m_useTree );
				if ( m_useTree )
				{
					RemoveMenu( hmInit, ID_VIEW_PGP_PUBLICKEYS, MF_BYCOMMAND );
					RemoveMenu( hmInit, ID_VIEW_PUBLICKEYS, MF_BYCOMMAND );
					RemoveMenu( hmInit, ID_VIEW_SECRETKEYS, MF_BYCOMMAND );
				}
				break;
			case ID_RBCLICK_GETTHEKEY:
				// See Note :: GetKeyOnly 3.
				if ( GetKeyOnly() )
					SetMenuDefaultItem( hmInit, wID, MF_BYCOMMAND );
				else
					RemoveMenu( hmInit, wID, MF_BYCOMMAND );
				break;

			case ID_ADDEXISTINGPUBLICKEY_PASTEFROMCLIPBOARD:
				_EnableMenuItem( hmInit, wID, !isClipboardEmpty() );
				break;


			default:
				break;
		}
	}

	return 0L;
}

LRESULT cDlgViewLocalDatabase::RealDlgProc( UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	LRESULT result;
	
	if ( ( uMsg != WM_INITDIALOG )
		&& ( cMenuItem::DlgProc( &result, m_hDlg, uMsg, wParam, lParam ) ) )
		return result;

//	result = m_splitter->RealWndProc( uMsg, wParam, lParam );

	switch(uMsg)
	{
		case WM_INITDIALOG:
		{
			InitDialog();
//#ifndef _DEBUG

			//HMENU hMenu = cMenuItem::findMenuByID( GetMenu( m_hDlg ), ID_SignCertificate );
			//if ( hMenu != NULL )
			//	DeleteMenu( hMenu, ID_SignCertificate, MF_BYCOMMAND );

//#endif
			cMenuItem::ExchangeWindowMenu( m_hDlg, m_siiT, m_siiQ );
			break;
		}
		case WM_COMMAND:
			OnCommand( wParam, lParam );
			break;

		case WM_KEYDOWN: // 0x100
		{
			if ( wParam == VK_DELETE )
				OnCommand( ID_KEY_DELETE, 0 );

			break;
		}

		case WM_NOTIFY:
		{
			NMHDR *hdr = (NMHDR *)lParam;

			if ( ( hdr->code == TTN_GETDISPINFOA ) || ( hdr->code == TTN_GETDISPINFOW ) )
				ToolBarNotify( hdr );
			else
			if ( wParam == IDC_List )
				return ListViewNotify( (LPNMHDR) lParam );
			else
			if ( wParam == IDC_Tree )
				return TreeViewNotify( (LPNMHDR) lParam );
			else
			if ( ( hdr->hwndFrom == m_hToolbar ) ||
				( GetDlgCtrlID( m_hToolbar ) == wParam ) )
					return ToolBarNotify( hdr );

			return false;
		}
		case WM_INITMENUPOPUP:
			OnInitMenuPopup((HMENU)wParam, LOWORD(lParam), HIWORD(lParam));
			break;

		case WM_DESTROY:
			DeleteDialog();
			return true;

		case WM_LBUTTONDBLCLK:

		default:
			return false;
	}

	return true;
}

bool ViewLocalDatabaseT(
		HWND hWnd,
		DWORD viewFlags,
		cPTB *ptbKey,
		type_kff *pKFF,
		cWCStr *wcsUserID 
		//cPTB			// [OUT] Pgp subkey does not have an user ID ( master key has only )
		//	*ptbUserID// = NULL	// but we have to ask password for secret key with a name.
								// The BrowseKeyRingForUser generates here the subkey name as 
								// as "Subkey of ( master key user ID )" or  key user ID if the key has it

	)
{
	ODSFunc;

	cDlgViewLocalDatabase dlg( viewFlags, ptbKey, pKFF, wcsUserID );

	int res = dlg.DoModal( hMainInst, dlg.IDD, hWnd );

	return res == IDOK;
}