#include "DlgViewLocalDatabase.h"
#include "PublicKeyProps.h"
#include "cWizardDlgs.h"
#include "BCDB_Functions.h"
#include "DlgAskWriteConfirmation.h"
#include "..\CertificateLib\PGPPkts\pktdefs.h"
#include "..\CertificateLib\x509pkts\base64funcs.h"
#include "..\CertificateLib\PGPPkts\PGPPkts.h"
#include ".\PKParse.h"
#include "..\Includes\Mail.h"
#include "..\Includes\OpenFiles.h"
#include "..\Includes\TranslateDll.h"
#include "DlgAbout.h"
//#include "..\BCArchive\Help\HELPIDS.h"
#include <io.h>

static bool GetKeyUserID( HWND hDlg, cWCStr *wcsUserID, BYTE *keyID, BYTE *kr, int krSize, bool secret )
{
	cpgp_keyInfo keyInfo;
	cpgp_keyInfo *pKeyInfo = NULL;

	int res = pgp_FindKeyByIDEx(
				&keyInfo,
				keyID,
				kr, krSize,
				secret,
				&pKeyInfo 
				);

	if ( pKeyInfo == NULL )
		pKeyInfo = &keyInfo; // Master key used

	if ( res != ppe_ok )
	{
		DlgMessageBoxT( hMainInst, hDlg, IDS_szErrorInKeyReading, MB_ICONSTOP );
		return false;
	}

	if ( wcsUserID == NULL )
		return true;

	cPTB ptb;

	if ( !pgp_GetKeyUserID( &ptb, &keyInfo, pKeyInfo ) )
		wcsUserID->Add( _LoadIDS( IDS_Undefined ) );
	else
		wcsUserID->Add( (char*)ptb.buffer, ptb.bufferSize );

	return true;
}

BOOL cDlgViewLocalDatabase::OnKeyProperties( bool secret )
{
	cWizardBlock wBlock( m_hInst, m_hDlg );

	wchar_t fileName[ MAX_PATH ];

	int i = lvGetSelectedItem( m_hList );

	if ( i == -1 )
		return FALSE;

	wchar_t path[ MAX_PATH ];
	if ( !BCDB_GetDataBasePath( path, maxLen( path ) ) )
		return x509e_internalError;

	ListView_GetItemTextT( m_hList, i, fileNamePos, fileName, maxLen( fileName ) );

	wcscat_s( path, fileName );

	if ( secret )
	{
		wBlock.LoadSecretKey( path );
		cWizardDlg_SecKeyPkt dlg( &wBlock, m_hInst, m_hDlg );
		dlg.DoModal( m_hInst, IDD_CreateCertificate, m_hDlg );
	}
	else
	{
		wBlock.LoadPublicKey( path );
		cWizardDlg_PubKeyCrt dlg( &wBlock, m_hInst, m_hDlg );
		dlg.m_editState = esAllowEdition;
		int res = dlg.DoModal( m_hInst, IDD_CreateCertificate, m_hDlg );
		//if ( res == IDOK )
			OnRefresh();
	}

	return TRUE;
}

BOOL cDlgViewLocalDatabase::OnKeyProperties_PGP( bool secret )
{
	BYTE *keyID = LV_GetKeyID( m_hList );
	if ( keyID == NULL )
	{
		assert( false );
		return FALSE;
	}

	return ShowKeyPropertiesEx( m_hInst, m_hDlg, keyID, m_keyPkt, m_keyPktSize, secret );
}

BOOL cDlgViewLocalDatabase::OnKeyProperties()
{
	switch ( m_viewType )
	{
		case vt_crtPublicKey: return OnKeyProperties( false );
		case vt_p12SecretKey: return OnKeyProperties( true );
		case vt_pgpPublicKey: return OnKeyProperties_PGP( false );
		case vt_pgpSecretKey: return OnKeyProperties_PGP( true );
	}

	return false;
}

//
//	GetSelected_PGP
//	Subkey is required if it has been choosen ( No entire key )
//
bool cDlgViewLocalDatabase::GetSelected_PGP( cPTB *ptb, cWCStr *wcsUserID, bool secret )
{
	int res;

	BYTE *keyID = LV_GetKeyID( m_hList );
	if ( keyID == NULL )
	{
		assert( false );
		return FALSE;
	}

	const bool useEntireKeyPacket = false;
	
	if ( useEntireKeyPacket )
	{
		BYTE *entireKeyPos;
		int entireKeySize;

		// Gets entire PGP key with all subpackets and subkeys

		res = pgp_GetEntireKeyByID(
			&entireKeyPos,
			&entireKeySize,
			m_keyPkt, m_keyPktSize,
			keyID,
			secret
		);

		if ( res != ppe_ok )
		{
			assert( res == ppe_ok );
			return false;
		}

		return ptb->WriteBytes( entireKeyPos, entireKeySize, true ) == entireKeySize;
	}
	else
	{
		cpgp_keyInfo keyInfo;
		cpgp_keyInfo *pKeyInfo = NULL;

		int res = pgp_FindKeyByIDEx(
					&keyInfo,
					keyID,
					m_keyPkt, m_keyPktSize,
					secret,
					&pKeyInfo 
				);


		if ( pKeyInfo == NULL )
			pKeyInfo = &keyInfo; // Master key used

		res = IsEncryptionKey( (sPublicKey *)pKeyInfo->keyPkt, pKeyInfo->keyPktSize );

		if ( res == ppe_invalidAlgorithmID )
		{
			// Checks if has one encryption subkey only.
			if ( ( pKeyInfo->next == NULL ) ||
				 ( ppe_ok != IsEncryptionKey( (sPublicKey *)pKeyInfo->next->keyPkt, pKeyInfo->next->keyPktSize ) ) ||
				 ( pKeyInfo->next->next != NULL )
				 )
			{
				DlgMessageBoxT( IDS_SignatureAlgForEncryption, MB_ICONSTOP );
				return false;
			}
			else
			{
				pKeyInfo = pKeyInfo->next;
			}
		}
		else	
		if ( res != ppe_ok )
		{
			DlgMessageBoxT( IDS_szErrorInKeyReading, MB_ICONSTOP );
			return false;
		}

		if ( wcsUserID != NULL )
		{
			BCDB_GetPGPKeyUserIDT( wcsUserID, m_keyPkt, m_keyPktSize, keyID, secret );
		}

		return ptb->WriteBytes( pKeyInfo->keyPkt, pKeyInfo->keyPktSize, true ) == pKeyInfo->keyPktSize;
	}
} 

bool cDlgViewLocalDatabase::GetSelected( cPTB *ptb, cWCStr *wcsUserID, bool secret )
{
	wchar_t fileName[ MAX_PATH ];
	wchar_t filePath[ MAX_PATH ];

	int iSel = ListView_GetNextItem( m_hList, -1, LVNI_FOCUSED );

	if ( !secret )
	{
		// Gets selected certificate file name

		ListView_GetItemTextT( m_hList, iSel, fileNamePos, fileName, maxLen( fileName ) );

		// Forms certificate file path

		if ( !BCDB_GetDataBasePath( filePath, maxLen( filePath ) ) )
		{
			assert( false );
			return false;
		}

		wcscat_s( filePath, fileName );
	}
	else
	{
		// Gets corresponded certificate file name for selected secret key

		wchar_t secKeyFileName[ MAX_PATH ];
		ListView_GetItemTextT( m_hList, iSel, fileNamePos, secKeyFileName, maxLen( secKeyFileName ) );

		if ( !BCDB_GetDataBasePath( filePath, maxLen( filePath ) ) )
		{
			assert( false );
			return false;
		}

		wcscat_s( filePath, secKeyFileName );
	}

	cFileView fv( filePath );

	if ( fv.m_err )
	{
		assert( false );
		return false;
	}

	if ( wcsUserID != NULL )
	{
		wchar_t friendlyName[ 1024 ];
		ListView_GetItemTextT( m_hList, iSel, 0, friendlyName, maxLen( friendlyName ) );
		
		wcsUserID->Add( friendlyName );
	}

	return ptb->WriteBytes( fv.m_ptr, (int)fv.m_size.QuadPart, true ) == fv.m_size.QuadPart;
}

bool cDlgViewLocalDatabase::GetSelected( cPTB *ptb, cWCStr *wcsUserID )
{
	switch ( m_viewType )
	{
		case vt_crtPublicKey: return GetSelected( ptb, wcsUserID, false );
		case vt_p12SecretKey: return GetSelected( ptb, wcsUserID, true );
		case vt_pgpPublicKey: return GetSelected_PGP( ptb, wcsUserID, false );
		case vt_pgpSecretKey: return GetSelected_PGP( ptb, wcsUserID, true );
	}

	return false;
}

BOOL cDlgViewLocalDatabase::DeleteSelected()
{
	waRes res = wa_Yes;

	for ( int iItem = -1; ( iItem = ListView_GetNextItem( m_hList, iItem, LVNI_SELECTED ) ) != -1 ; )
	{
		res = OnDeleteItem( iItem, res );

		if ( ( res == wa_Cancel ) 
			|| ( res == wa_NoToAll ) )
				break;

		if ( res == wa_Yes ) 
			iItem--;
	}

	return res != wa_Cancel;
}

cDlgViewLocalDatabase::waRes cDlgViewLocalDatabase::OnDeleteItem( int item, cDlgViewLocalDatabase::waRes prevRes )
{
	if ( ( prevRes == wa_NoToAll ) ||  ( prevRes == wa_Cancel ) )
		return prevRes;

	if ( item == -1 )
	{
		item = ListView_GetNextItem( m_hList, -1, LVNI_SELECTED );
		if ( item == -1 )
			return wa_Cancel;
	}

	bool secret = true;

	switch ( m_viewType )
	{
		case vt_crtPublicKey:
			secret = false;

		case vt_p12SecretKey:
		{
			wchar_t fileName[ MAX_PATH ];
			wchar_t filePath[ MAX_PATH ];

			ListView_GetItemTextT( m_hList, item, fileNamePos, fileName, maxLen( fileName ) );

			if ( !BCDB_GetDataBasePath( filePath, maxLen( filePath ) ) )
			{
				assert( false );
				return wa_Cancel;
			}

			wcscat_s( filePath, fileName );

			wchar_t userName[ 1024 ];
			ListView_GetItemTextT( m_hList, item, 0, userName, maxLen( userName ) );

			wchar_t keyIDStr[ 1024 ];
			ListView_GetItemTextT( m_hList, item, 2, keyIDStr, maxLen( keyIDStr ) );

			cWCStr msg;
			msg.FormatT( 
				_LoadIDS( IDS_DoYouWantToDeleteKey_SSS_Q ),
				_LoadIDS( secret ? IDS_szSecret : IDS_szPublic ),
				userName,
				keyIDStr
				);

			int res = prevRes;
			if ( res != wa_YesToAll )
			{
				res = MessageBoxT( m_hDlg, msg.m_szBody, _LoadIDS( IDS_AppName ), MB_ICONQUESTION | MB_YESNOCANCEL );

				if ( res == IDCANCEL )
					return wa_Cancel;
				if ( res == IDNO )
					return wa_No;
			}

			if ( DeleteFileT( filePath ) )
			{
				ListView_DeleteItem( m_hList, item );
				return wa_Yes;
			}
			
			MessageBoxT( m_hDlg, filePath, _LoadIDS( IDS_ErrorInDeletingFile ), MB_ICONSTOP );

			return wa_Cancel;
		}

		case vt_pgpPublicKey:
			secret = false;
		case vt_pgpSecretKey:
		{
			cPTB ptb;
			if ( ptb.WriteBytes( m_keyPkt, m_keyPktSize ) != m_keyPktSize )
			{
				DlgMessageBoxT( IDS_ErrorInAllocatingMemory, MB_ICONSTOP );
				assert( false );
				return wa_Cancel;
			}

			BYTE *keyID = LV_GetKeyID( m_hList, item );

			//cPTB ptbUserID;
			cWCStr wcsUserID;
			if ( !GetKeyUserID( m_hDlg, &wcsUserID, keyID, m_keyPkt, m_keyPktSize, secret ) )
				return wa_Cancel;

			wchar_t message[ 1024 ];
			FormatT( message, maxLen( message ),
				_LoadIDS( IDS_AreYouSureYouWantToDeleteKey_SS_Q ),
				_LoadIDS( secret ? IDS_szSecret : IDS_szPublic ),
				wcsUserID.m_szBody );

			int res = prevRes;
			if ( res != wa_YesToAll )
			{
				res = DlgMessageBoxT( message, MB_ICONQUESTION | MB_YESNOCANCEL );

				if ( res == IDCANCEL )
					return wa_Cancel;
				if ( res == IDNO )
					return wa_No;
			}

			res = pgp_RemoveKeyByID( m_keyPkt, &m_keyPktSize, keyID, secret );

			if ( res != ppe_ok )
			{
				DlgMessageBoxT( IDS_ErrorInRemovingKey, MB_ICONSTOP );
				return wa_Cancel;
			}

			wchar_t krPath[ MAX_PATH ];
			bool ok = BCDB_GetLocalKeyRingPath( krPath, maxLen( krPath ), secret );

			if ( ptb.bufferSize == 0 )
			{
				if ( !DeleteFileT( krPath ) )
				{
					DlgMessageBoxT( IDS_ErrorInDeletingKRFile, MB_ICONSTOP );
					return wa_Cancel;
				}
			}
			else
			if ( WriteToFile( krPath, m_keyPkt, m_keyPktSize ) != m_keyPktSize )
			{
				assert( false );
				DlgMessageBoxT( IDS_ErrorInWrittingKRFile, MB_ICONSTOP );
				return wa_Cancel;
			}

			// If the deleted item is expanded it has to be collapsed before
			//	Otherwise the subkeys items are not deleted

			LVITEMW lvi;
			memset( &lvi, 0, sizeof( lvi ) );
			lvi.iItem = item;
			lvi.mask = LVIF_IMAGE;

			ListView_GetItemT( m_hList, &lvi );

			if ( lvi.iImage == 2 ) // Expanded
				deleteSubKeys( m_hList, item );

			ListView_DeleteItem( m_hList, item );

			return wa_Yes;
		}
	}

	return wa_Yes;
}

// Version 1.04.2 - 11.09.2014  Bug #1622: Key Manager loses menu items.
void cDlgViewLocalDatabase::ReloadMenu()
{
	HMENU hMenu = GetMenu( m_hDlg );
	DestroyMenu( hMenu );
	hMenu = LoadMenu( IH.m_hInst, MAKEINTRESOURCE(IDR_BCDB_MENU) );
	SetMenu( m_hDlg, hMenu );
	cMenuItem::ExchangeWindowMenu( m_hDlg, m_siiT, m_siiQ );
}

BOOL cDlgViewLocalDatabase::OnCommand( WPARAM wParam, LPARAM lParam )
{
	if ( HIWORD( wParam ) == 0 )
	{
		switch ( LOWORD( wParam ) )
		{
			case IDOK:

				if ( m_ptbKeyOut == NULL )
					return false;

				if ( !GetSelected( m_ptbKeyOut, m_wcsUserID ) )
					break;

				EndDialog( m_hDlg, IDOK );
				return true;

			case IDCANCEL:
			case ID_KEY_EXIT:
				EndDialog( m_hDlg, IDCANCEL );
				return true;
			case ID_ViewIcon:
		//		case ID_VIEW_SMALLICON:
			case ID_ViewDetails:
			case ID_ViewList:
			case IDM_VIEW:
				return LV_SetStyle( m_hList, wParam );
			case ID_LargeToolbar:
				m_bLargeToolbar = !m_bLargeToolbar;
				ToolBarUpdateSize();
				ToolBarUpdateButtons();
				return true;

			case ID_KEY_DELETE:
				return DeleteSelected();

			case ID_VIEW_PGP_PUBLICKEYS:
				m_viewType = vt_pgpPublicKey;
				return FillListView();
			case ID_VIEW_PUBLICKEYS:
				m_viewType = vt_crtPublicKey;
				return FillListView();
			case ID_VIEW_SECRETKEYS:
				m_viewType = vt_p12SecretKey;
				return FillListView();
			case ID_VIEW_SHOWTYPEVIEW:
				m_useTree = !m_useTree;
				ReloadMenu();
				ShowTreeView();
				break;

			case ID_DATABASE_REFRESH:
			case ID_VIEW_REFRESH:
				return OnRefresh();

			case ID_KEY_PROPERTIES:
				return OnKeyProperties();

			case ID_DATABASE_EXPORT:
				return OnExportDatabase();
			case ID_DATABASE_IMPORT:
				return OnImportDatabase();
			case ID_DATABASE_CHOOSESTOREFOLDER:
				return OnChooseStoreFolder();

			case ID_KEY_GENERATENEWPUBLIC:
				return OnGenerateKeyPair();

			case ID_ADDEXISTINGPUBLICKEY_SEARCHINTHEINTERNET:
				return OnAddPK_SearchInNet();

			case ID_ADDEXISTINGPUBLICKEY_BROWSEFORFILES:
				return OnAddPK_BrowseLocaly();

			case ID_ADDEXISTINGPUBLICKEY_PASTEFROMCLIPBOARD:
				return ReadPKFromClipboard();

			case ID_KEY_FIND:
				if ( !GetFindParams() )
					break;
			case ID_KEY_FINDNEXT:
				return FindAndSelect();
			case ID_KEY_SENDPUBLICKEYTOE:
				return SendPKToEMailRecipient();
			case ID_KEY_COPYPUBLICKEYTOCLIPBOARD:
				return CopyPKToClipboard();
			case ID_HELP_ABOUT:
			{
				cDlgAbout dlg;
				dlg.DoModal( m_hInst, IDD_dlgAbout, m_hDlg );
				return true;
			}
			case ID_HELP_CONTENT:
				OnHelp();
				return true;

			case ID_RBCLICK_GETTHEKEY:
				if ( !GetKeyOnly() )
				{
					assert( false );
				}
				else
				if ( OnGetKey() )
				{
					EndDialog( m_hDlg, IDOK );
					return true;
				}
				break;

			default:
				break;
		}
	}

	return false;
}

bool cDlgViewLocalDatabase::IsFindParamsEmpty()
{
	return m_findParams.IsEmpty();
}

bool cDlgViewLocalDatabase::GetFindParams()
{
	m_findParams.m_hListView = m_hList;
	m_findParams.m_columnQ = m_columnQ;

	cDlgFindParams dlg( &m_findParams );

	int res = dlg.DoModal( m_hInst, IDD_DlgFindParams, m_hDlg );
	return res == IDOK;
}

bool cDlgViewLocalDatabase::FindAndSelect()
{
	m_findParams.m_hListView = m_hList;
	m_findParams.m_columnQ = m_columnQ;

	return m_findParams.FindNext( m_hDlg );
}

bool cDlgViewLocalDatabase::OnAddPK_SearchInNet()
{
	cPTB ptbKey;
	int res = SearchPKUserT( m_hInst, m_hDlg, NULL, NULL, NULL, &ptbKey );

	if ( ptbKey.bufferSize == 0 )
		return false;

	BYTE keyID[ pgp_keyIDSize ];

	res = ExtractDataAndVerifyPGPPublicKeyBlock( ptbKey.buffer, ptbKey.bufferSize,
		NULL, 0, NULL /*&pka */, NULL, NULL, keyID, NULL, true, NULL );

	if ( res != ppe_ok )
		return false;

//	BCDB_GenerateLocalKeyID( &ptbKeyID, pka, false );
//	pka_free( pka );

	m_viewType = vt_pgpPublicKey;
	OnRefresh();

	int i = FindItem( keyID, sizeof( keyID ) );

	if ( i == -1 )
		return false;

	ListView_SetItemState( m_hList, i, LVIS_SELECTED, LVIS_SELECTED );

	return true;
}

int cDlgViewLocalDatabase::FindItem( BYTE *keyID, int keyIDSize )
{
	//char keyIDStr[ 1024 ];
	//int size = sizeof( keyIDStr );
	//bytes2string( keyIDStr, &size, keyID, keyIDSize );
	//char buff[ 1024 ];

	wchar_t buff[ 1024 ];
	cWCStr wcsKeyId;
	bytes2string( &wcsKeyId, keyID, keyIDSize );

	int hItem;

	for ( hItem = -1; ( hItem = ListView_GetNextItem( m_hList, hItem, LVNI_ALL ) ) != -1 ; )
	{
		buff[ 0 ] = L'\0';

		ListView_GetItemTextT( m_hList, hItem, 2, buff, maxLen( buff ) );

		// if ( strcmp( buff, keyIDStr ) == 0 )
		if ( wcsKeyId.equ( buff ) )
			break;
	}

	return hItem;
}

static UINT FindCompatibleFormat()
{
	UINT cbf = 0;

	while ( cbf = EnumClipboardFormats( cbf ) )
	{
		switch ( cbf )
		{
			case CF_LOCALE:
			case CF_OEMTEXT:
			case CF_TEXT:
			case CF_UNICODETEXT:
			case CF_DSPTEXT:
				return cbf;
			default:
				break;
		}
	}

	return 0;
}

static const char *strPGPPKStart	= "-----BEGIN PGP PUBLIC KEY BLOCK-----";
static const char *strPGPPKEnd		= "-----END PGP PUBLIC KEY BLOCK-----";

static const char *strX509CertificateStart = "-----BEGIN X509 CERTIFICATE-----";
static const char *strX509CertificateEnd   = "-----END X509 CERTIFICATE-----";

static const char *strCertificateStart = "-----BEGIN CERTIFICATE-----";
static const char *strCertificateEnd   = "-----END CERTIFICATE-----";

bool cDlgViewLocalDatabase::ReadASCIIArmoredBlock( const char *asciiBlock )
{
	const BYTE *block;
	int blockSize;

	eViewType viewType;

	int len = strlen( asciiBlock );

	if ( len == 0 )
		return false;

	// Finds ascii armored block

	if ( ( findEncodedBlock( &block, &blockSize, ( BYTE * )asciiBlock, len, strCertificateStart, strCertificateEnd ) == ppe_ok ) ||
		 ( findEncodedBlock( &block, &blockSize, ( BYTE * )asciiBlock, len, strX509CertificateStart, strX509CertificateEnd ) == ppe_ok )
		)
	{
		viewType = vt_crtPublicKey;
	}
	else
	if ( findEncodedBlock( &block, &blockSize, ( BYTE * )asciiBlock, len, strPGPPKStart, strPGPPKEnd ) == ppe_ok )
	{
		viewType = vt_pgpPublicKey;
	}
	else
	{
		DlgMessageBoxT( IDS_ASCIIArmoredBlockNotFound, MB_ICONSTOP );
		return false;
	}

	// Decodes it

	int size = 0;
	int res = base64Decode( NULL, &size, block, blockSize );

	if ( res != ppe_bufferTooSmall )
	{
		assert( false );
		return false;
	}

	cPTB ptbDecoded;
	if ( ptbDecoded.Allocate( size ) == NULL )
	{
		assert( false );
		return false;
	}

	res = base64Decode( ptbDecoded.buffer, &size, block, blockSize );

	if ( res != ppe_ok )
	{
		assert( false );
		return false;
	}

	if ( viewType == vt_pgpPublicKey )
	{
		int resIDS = BCDB_AddPGPKeyToLocalKeyRing( m_hDlg, ptbDecoded.buffer, ptbDecoded.bufferSize, false );
		if ( resIDS != NULL  )
		{
			DlgMessageBoxT( resIDS, MB_ICONSTOP );
			return false;
		}
	}
	else
	if ( viewType == vt_crtPublicKey )
	{
		cWCStr wcs;

		bool res = BCDB_AddCertificateToStorage( ptbDecoded.buffer, ptbDecoded.bufferSize, &wcs );
		if ( !res )
		{
			
			MessageBoxT( m_hDlg, wcs.m_szBody, _LoadIDS( IDS_AddCrtToStorage ), MB_ICONSTOP );
			return false;
		}
	}
	else
	{
		assert( false );
		return false;
	}

	RefreshView( viewType );
	return true;
}

bool cDlgViewLocalDatabase::isClipboardEmpty()
{
	bool res = OpenClipboard( m_hDlg ) == TRUE;

	if ( !res )
		return true;

	UINT cbf = CF_TEXT;

	if ( !IsClipboardFormatAvailable( CF_TEXT ) )
	{
		cbf = FindCompatibleFormat();

		if ( cbf == 0 )
		{
			CloseClipboard();
			return true;
		}
	}

	HANDLE hcbData = GetClipboardData( cbf );

	res = true;

	if ( hcbData != NULL )
	{
		char *lptstr = (char *)GlobalLock( hcbData ); 
		if (lptstr != NULL) 
		{
			if ( strlen( lptstr ) != 0 )
				res = false;

			GlobalUnlock( hcbData );
		}
	}
            
	CloseClipboard();

	return res;
}

//
//	ReadPKFromClipboard
//		1. Finds ascii armored PGP or Certificate blocks
//		2. Adds the public key to local database.
//
bool cDlgViewLocalDatabase::ReadPKFromClipboard()
{
	bool res = OpenClipboard( m_hDlg ) == TRUE;

	if ( !res )
		return FALSE;

	UINT cbf = CF_TEXT;

	if ( !IsClipboardFormatAvailable( CF_TEXT ) )
	{
		cbf = FindCompatibleFormat();

		if ( cbf == 0 )
		{
			CloseClipboard();

			DlgMessageBoxT( IDS_szClipboardIsEmpty, MB_ICONINFORMATION );

			return FALSE;
		}
	}

	HANDLE hcbData = GetClipboardData( cbf );

	res = FALSE;

	if ( hcbData != NULL )
	{
		char *lptstr = (char *)GlobalLock( hcbData ); 
		if (lptstr != NULL) 
		{
			if ( strlen( lptstr ) == 0 )
			{
				MessageBoxT( m_hDlg, L"Clipboard is empty", _LoadIDS( IDS_AppName ), MB_ICONINFORMATION );
			}
			else
				res = ReadASCIIArmoredBlock( lptstr );

			GlobalUnlock( hcbData );
		}
	}
            
	CloseClipboard();

	return res;
}

//
//	OnAddPK_BrowseLocaly
//		1. Chooses key block on local environment.
//		2. Gets key ID of added key.
//		3. Set current view type as format of added key.
//		4. Finds the added key by ID in the view and selects it.
//
bool cDlgViewLocalDatabase::OnAddPK_BrowseLocaly()
{
	type_kff kff;
	cPTB ptbKey;
	cPTB ptbEntirePkt;

	bool res = BrowseForPKBlockT( m_hInst, m_hDlg, L"C:\\", false, &ptbKey, &kff, NULL, &ptbEntirePkt );

	wchar_t pkPath[ MAX_PATH ];

	if ( res )
	{
		//
		//	v.1.05.1.2 - 02.12.2015
		//	BrowseForPKBlockT returns pure public key from PGP key ring. But to add key we need all it's information.
		//	So we uses here entire key ring.
		//
		if ( ( kff == kffPGP ) && ( ptbEntirePkt.bufferSize != 0 ) )
		{
			res = AddKeyToStorageEx( m_hDlg, ptbEntirePkt.buffer, ptbEntirePkt.bufferSize, kff, false, pkPath );
			ptbKey.Swap( &ptbEntirePkt );
		}
		else
		if ( ptbKey.bufferSize != 0 )
			res = AddKeyToStorageEx( m_hDlg, ptbKey.buffer, ptbKey.bufferSize, kff, false, pkPath );
	}

	if ( !res )
		return false;

	// Gets key ID, the key is to be selected after refreshing.

	HPKA pka = NULL;

	cPTB ptbKeyID;

	if ( kff == kffPGP )
	{
		m_viewType = vt_pgpPublicKey;

		int res = ExtractDataAndVerifyPGPPublicKeyBlock( ptbKey.buffer, ptbKey.bufferSize,
			NULL, 0, &pka, NULL, NULL, NULL, NULL, true, NULL );

		if ( res != ppe_ok )
		{
			DlgMessageBoxT( IDS_ErrorInReadingKey, MB_ICONSTOP );
			return false;
		}
	}
	else
	{
		BYTE *pkt;
		int size;

		// The public key certificate may be extracted from pfx signedData and written to file
		// so we have to read the file here.

		cFileView fv;
		if ( fv.Open( pkPath ) )
		{
			pkt = fv.m_ptr;
			size = (int)fv.m_size.QuadPart;
		}
		else
		{
			pkt = ptbKey.buffer;
			size = ptbKey.bufferSize;
		}

		m_viewType = vt_crtPublicKey;

		x509_certificate crt;

		int res = crt.parse( pkt, size  );
		if ( res != x509e_ok )
			return false;
		
//		int size = 0;
//		BYTE *keyID;

//		res = crt.tbsCertificate.getSubjectKeyID( &keyID, &size );

//		if ( res != x509e_ok )
//			return false;

//		if ( ptbKeyID.WriteBytes( keyID, size ) != size )
//			return false;

		res = pka_createAlgorithm_X509( &pka, &crt.tbsCertificate.subjectPKInfo );
		if ( res != x509e_ok )
			return false;
	}

	BCDB_GenerateLocalKeyID( &ptbKeyID, pka, false );
	pka_free( pka );

	// Refreshs
	
	
	if ( !OnRefresh() )
		return false;

	// Selects added item

//	BCDB_GenerateLocalKeyID( &ptbKeyID, pka, false );
//	pka_free( pka );

	int i = FindItem( ptbKeyID.buffer, ptbKeyID.bufferSize );

	if ( i == -1 )
		return false;

	ListView_SetItemState( m_hList, i, LVIS_SELECTED, LVIS_SELECTED );

	return true;
}

bool cDlgViewLocalDatabase::OnGenerateKeyPair()
{
	if ( GenerateKeyPair( m_hInst, m_hDlg ) == x509e_ok )
		return OnRefresh();

	return false;
}

bool cDlgViewLocalDatabase::OnRefresh()
{
	return FillListView() == x509e_ok;
}

int CALLBACK BrowseCallbackProc(
	HWND hwnd,
    UINT uMsg,
    LPARAM lParam,
    LPARAM lpData
)
{
	if ( uMsg == BFFM_INITIALIZED )
	{
		wchar_t curPath[ MAX_PATH ];
		BCDB_GetDataBasePath( curPath, MAX_PATH );

		SendMessageT( hwnd, BFFM_SETSELECTIONW, BFFM_SETSELECTIONA, TRUE, curPath );
		SetWindowTextT( hwnd, _LoadIDS( IDS_AppName ) );
	}
	else
	if ( uMsg == BFFM_SELCHANGED )
	{
		LPITEMIDLIST pidl = (LPITEMIDLIST)lParam;

		if ( pidl != NULL )
		{
			wchar_t newPath[ MAX_PATH ];
			BOOL bSuccess = SHGetPathFromIDListT( pidl, newPath );

			wchar_t text[ 1024 ] = L"";

			if ( bSuccess )
				FormatT( text, maxLen( text ), _LoadIDS( IDS_SelectedPath_S ), newPath );

			SendMessageT( hwnd, BFFM_SETSTATUSTEXTW, BFFM_SETSTATUSTEXTA, 0, text );
		}
	}

	return 0;
}

static int copyDBFile( HINSTANCE hInst, HWND hWnd, const wchar_t *dstFolder, const wchar_t *srcFolder )
{
	if ( !FileExistsT( dstFolder ) )
		CreateDirectoryT( dstFolder, NULL );

	wchar_t srcPath[ MAX_PATH ];
	wcscpy_s( srcPath, srcFolder );
	int srcLen = wcslen( srcPath );
	if ( srcPath[ srcLen - 1 ] != L'\\' )
		srcPath[ srcLen++ ] = L'\\';

	wchar_t dstPath[ MAX_PATH ];
	wcscpy_s( dstPath, dstFolder );
	int dstLen = wcslen( dstPath );
	if ( dstPath[ dstLen - 1 ] != L'\\' )
		dstPath[ dstLen++ ] = L'\\';

	wcsncpy( srcPath + srcLen, L"*", maxLen( srcPath ) - srcLen );

	WIN32_FIND_DATAW findData;

	HANDLE hFind = FindFirstFileT( srcPath, &findData );

	if ( hFind == INVALID_HANDLE_VALUE )
		return 0;

	int count = 0;

	bool bYesToAll = false;

	do
	{
		if ( ( findData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) != 0 )
			continue;

		wcsncpy( dstPath + dstLen, findData.cFileName, maxLen( dstPath ) - dstLen );
		wcsncpy( srcPath + srcLen, findData.cFileName, maxLen( srcPath ) - srcLen );

		if ( !bYesToAll && FileExistsT( dstPath ) )
		{
			int res = DlgAskWriteConfirmation( hInst, hWnd, dstPath, srcPath );

			if ( res == cDlgAskWriteConfirmation::idOverwriteAll )
			{
				bYesToAll = true;
			}
			else
			if ( ( res == cDlgAskWriteConfirmation::idSkipAll )
					|| ( res == cDlgAskWriteConfirmation::idCancelled ) )
			{
				break;
			}
			else
			if ( res == cDlgAskWriteConfirmation::idSkip )
			{
				continue;
			}
			else
			if ( res != cDlgAskWriteConfirmation::idOverwrite )
			{
				assert( false );
			}
		}

		while( true )
		{
			BOOL ok = CopyFileT( srcPath, dstPath, FALSE );

			if ( ok )
			{
				count++;
				break;
			}

			wchar_t msg[ 1024 ];
			FormatT( msg, maxLen( msg ),
				_LoadIDS( IDS_ErrorInCopingFile_SS ),
				srcPath, dstPath );

			int res = MessageBoxT( 0,  msg, _LoadIDS( IDS_AppName ), MB_ICONSTOP | MB_ABORTRETRYIGNORE );

			if ( res == IDABORT )
			{
				FindClose( hFind );
				return IDCANCEL;
			}
			else
			if ( res == IDRETRY )
				continue;
			else
			if ( res == IDIGNORE )
				break;
			else
			{
				assert( false );
			}
		}
	}
	while( FindNextFileT( hFind, &findData ) );

	FindClose( hFind );

	return count;
}

static int countFolderItems( const wchar_t *dir )
{
	wchar_t path[ MAX_PATH ];
	wcscpy_s( path, dir );
	int len = wcslen( path );
	if ( path[ len - 1 ] != L'\\' )
		path[ len++ ] = L'\\';

	wcsncpy( path + len, L"*", maxLen( path ) - len );

	WIN32_FIND_DATAW findData;

	HANDLE hFind = FindFirstFileT( path, &findData );

	if ( hFind == INVALID_HANDLE_VALUE )
		return 0;

	int count = 0;

	do
	{
		if ( ( findData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) != 0 )
			continue;

		count++;
	}
	while( FindNextFileT( hFind, &findData ) );

	FindClose( hFind );

	return count;
}

bool cDlgViewLocalDatabase::OnChooseStoreFolder()
{
	HRESULT hr = OleInitialize( 0 );

	
	wchar_t curPath[ MAX_PATH ];
	if ( !BCDB_GetDataBasePath( curPath, maxLen( curPath ) ) )
		return false;

	int len = wcslen( curPath );
	if ( curPath[ len - 1 ] == L'\\' )
		curPath[ --len ] = L'\0';

	wchar_t szTitle[ 1024 ];
	FormatT( szTitle, maxLen( szTitle ), _LoadIDS( IDS_ChooseStoreFolder ), curPath );

	LPITEMIDLIST myCompPidl = NULL;

	const bool startFromDesktop = true;

#ifdef SHGetFolderLocation
	if ( !startFromDesktop )
		SHGetFolderLocation( m_hDlg, CSIDL_DRIVES, NULL, 0, &myCompPidl );
#else
#define BIF_NEWDIALOGSTYLE 0x0040
#endif

	BROWSEINFOW bi;
	memset( &bi, 0, sizeof( bi ) );
	bi.hwndOwner = m_hDlg;
	bi.pidlRoot = myCompPidl;
	bi.pszDisplayName = NULL;
	bi.lpszTitle = szTitle;
	bi.ulFlags = 0
//		| BIF_DONTGOBELOWDOMAIN
		| BIF_NEWDIALOGSTYLE
		| BIF_RETURNONLYFSDIRS
//		| BIF_EDITBOX
//		| BIF_STATUSTEXT
		;

	bi.lpfn = BrowseCallbackProc;

	BYTE buff[ 1024 ];
	bi.lParam = (LPARAM)buff;

	wchar_t newPath[ MAX_PATH ];

	while( true )
	{
		LPITEMIDLIST pidl = SHBrowseForFolderT(&bi);

		if ( ( hr == S_OK ) || ( hr == S_FALSE ) )
			OleUninitialize();

		if (!pidl)
		{
			return FALSE;
		}

		BOOL bSuccess = SHGetPathFromIDListT( pidl, newPath );

		// Frees pid list
		static LPMALLOC pMalloc = NULL;	if ( pMalloc == NULL ) SHGetMalloc( &pMalloc );
		if ( pMalloc )
		{
			pMalloc->Free(pidl);
			if ( myCompPidl != NULL )
				pMalloc->Free(myCompPidl);
		}

		if (!bSuccess)
		{
			return FALSE;
		}

		bool foldersAreSame = _wcsicmp( newPath, curPath ) == 0;

		if ( !foldersAreSame )
			break;
		
		wchar_t message[ 1024 ];
		FormatT( message, maxLen( message ), _LoadIDS( IDS_STRING172 ),	curPath );

		DlgMessageBoxT( message, MB_ICONEXCLAMATION );
	}

	// Copy existing current items to new database

	int items = countFolderItems( curPath );

	if ( items != NULL )
	{
		wchar_t message[ 1024 ];
		FormatT( message, maxLen( message ), _LoadIDS( IDS_STRING173 ), newPath, curPath, items );

		int res = DlgMessageBoxT( message, MB_ICONQUESTION | MB_YESNO );

		if ( res == IDYES )
		{
			copyDBFile( m_hInst, m_hDlg, newPath, curPath );
		}
	}

	bool ok = BCDB_SetDataBasePath( newPath );

	OnRefresh();

	return ok;
}

bool cDlgViewLocalDatabase::OnExportDatabase()
{
	return BCDB_Port( m_hDlg, true );
}

bool cDlgViewLocalDatabase::OnImportDatabase()
{
	bool res = BCDB_Port( m_hDlg, false );

	if ( res )
		res = OnRefresh();

	return res;
}

static int compareTails( const wchar_t *s0, const wchar_t *s1 )
{
	int l0 = (int)wcslen( s0 );
	int l1 = (int)wcslen( s1 );

	if ( l0 > l1 )
		s0 += l0 - l1;
	else
		s1 += l1 - l0;

	return _wcsicmp( s0, s1 );
}

//
// cDlgViewLocalDatabase::GetSelectedPublicKey
//
//	1. Chooses selected certificate if x.509 crt is viewed.
//	2. Chooses corresponded certifiacte for selected secret key p12.
//	3. Chooses PGP key selected from local key ring.
//
bool cDlgViewLocalDatabase::GetSelectedPublicKey( cPTB *ptbBase64Encoded, cWCStr *pWCSUserID )
{
	//cPTB ptbBase64Encoded;
	cWCStr wcsUserID;

	int iSel = ListView_GetNextItem( m_hList, -1, LVNI_FOCUSED );

	if ( iSel == -1 )
	{
		assert( false );
		return false;
	}

	int res;
	BYTE *entireKeyPos;
	int entireKeySize;
	cFileView cfv; // cetificate file view

	const char *strStart;
	const char *strEnd;

	if ( m_viewType == vt_pgpPublicKey )
	{
		// Gets PGP key ID of selected item

		BYTE *keyID = LV_GetKeyID( m_hList, iSel );
		if ( keyID == NULL )
		{
			assert( keyID != NULL );
			return false;
		}

		// Gets entire PGP key with all subpackets and subkeys

		res = pgp_GetEntireKeyByID(
			&entireKeyPos,
			&entireKeySize,
			m_keyPkt, m_keyPktSize,
			keyID,
			false
		);

		if ( res != ppe_ok )
		{
			assert( res == ppe_ok );
			return false;
		}

		// Gets Key user ID

		if ( !GetKeyUserID( m_hDlg, &wcsUserID, keyID, entireKeyPos, entireKeySize, m_secret ) )
			return false;

		// Prepares base 64 encoded buffer data for PGP key

		strStart = strPGPPKStart;
		strEnd = strPGPPKEnd;
	}
	else
	{
		wchar_t fileName[ MAX_PATH ];
		wchar_t filePath[ MAX_PATH ];

		if ( m_viewType == vt_crtPublicKey )
		{
			// Gets selected certificate file name

			ListView_GetItemTextT( m_hList, iSel, fileNamePos, fileName, maxLen( fileName ) );

			// Forms certificate file path

			if ( !BCDB_GetDataBasePath( filePath, maxLen( filePath ) ) )
			{
				assert( false );
				return false;
			}

			wcscat_s( filePath, fileName );
		}
		else
		if ( m_viewType == vt_p12SecretKey )
		{
			// Gets corresponded certificate file name for selected secret key

			wchar_t secKeyFileName[ MAX_PATH ];
			ListView_GetItemTextT( m_hList, iSel, fileNamePos, secKeyFileName, maxLen( secKeyFileName ) );

			if ( !BCDB_GetDataBasePath( filePath, maxLen( filePath ) ) )
			{
				assert( false );
				return false;
			}

			wcscat_s( filePath, secKeyFileName );

			cFileView fv( filePath );

			cUniversalString friendlyName;
			cPTB ptbLocalKeyID;
			wchar_t errStr[ 1024 ];
			LoadStringT( m_hInst, IDS_NoError, errStr, maxLen( errStr ) );

			HPKA pka = NULL;

			bool lRes = BCDB_GetP12SecretKeyInfo(
				m_hInst, m_hDlg,
				fv.m_ptr, (int)fv.m_size.QuadPart,
				NULL,
				&friendlyName, &ptbLocalKeyID,
				errStr, maxLen( errStr )
				);

			if ( !lRes )
			{
				DlgMessageBoxT( errStr, MB_ICONSTOP );
				return false;
			}
			else
			{
				// The secret key file was succefully read, so we try to find corresponded
				//	certificate file to get algorithm descryption ( secret key is encrypted and has no
				//	opened descryption )

//				cPTB ptbCrt;

				lRes = BCDB_GetX509KeyByID(
						NULL,
						ptbLocalKeyID.buffer, ptbLocalKeyID.bufferSize,
						false, NULL,
						filePath
					);

				if ( !lRes )
				{
					DlgMessageBoxT( IDS_CorrespondedPKCertificateNotFound, MB_ICONSTOP );
					return false;
				}
			}
		}

		// Gets user ID ( certificate subject name )

		if ( !cfv.Open( filePath ) )
			return false;

		cUniversalString friendlyName;

		if ( !BCDB_GetCerPublicKeyInfo( m_hInst, m_hDlg, cfv.m_ptr, (int)cfv.m_size.QuadPart, NULL, &friendlyName ) )
			return false;

		wcsUserID.Add( friendlyName.buffer );

		// Prepares base 64 encoded buffer data for certificate 

		entireKeyPos = cfv.m_ptr;
		entireKeySize = (int)cfv.m_size.QuadPart;

		strStart = strX509CertificateStart;
		strEnd = strX509CertificateEnd;
	}

	if ( pWCSUserID != NULL )
		pWCSUserID->Set( wcsUserID.m_szBody );

	// Makes base 64 encoded buffer by prepared data

	if ( !createArmoredBlock( ptbBase64Encoded, entireKeyPos, entireKeySize, strStart, strEnd ) ) // , getMBStr( wcsUserID.m_szBody ) ) )
		return false;

	//int size = 0;

	//res = base64Encode( NULL, &size, entireKeyPos, entireKeySize, strStart, strEnd, wcsUserID.m_szBody );

	//if ( res != ppe_bufferTooSmall )
	//{
	//	assert( false );
	//	return false;
	//}

	//if ( ptbBase64Encoded->Allocate( size ) == NULL )
	//{
	//	assert( false );
	//	return false;
	//}

	//res = base64Encode( ptbBase64Encoded->buffer, &size, entireKeyPos, entireKeySize, strStart, strEnd );

	//if ( res != ppe_ok )
	//{
	//	assert( false );
	//	return false;
	//}
	//
	//ptbBase64Encoded->PutByte( 0 );

	return true;
}

static bool CopyToClipboard( HWND hWnd, const wchar_t *wc, int wcLen )
{
	BOOL res = OpenClipboard( hWnd );

	if ( !res )
		return FALSE;

	if ( !EmptyClipboard() )
	{
		CloseClipboard(); 
		return FALSE; 
	}

	if ( wcLen == -1 )
		wcLen = (int)wcslen( wc );

	int size = ( wcLen + 1 ) * 2;


	HGLOBAL hglbCopy = GlobalAlloc( GMEM_MOVEABLE,  size  ); 
	if (hglbCopy == NULL) 
	{ 
		CloseClipboard(); 
		return FALSE; 
	} 

	// Lock the handle and copy the text to the buffer. 

	wchar_t *lptstrCopy = (wchar_t *)GlobalLock(hglbCopy);

	bool bOk = ( lptstrCopy != NULL );

	if ( bOk )  // 24.10.2018 From BDK 9 Audit by Rechtspraak.
	{
		UINT uFormat = CF_UNICODETEXT;

		memcpy( lptstrCopy, wc, size );

		GlobalUnlock(hglbCopy); 

		// Place the handle on the clipboard. 

		SetClipboardData( uFormat, hglbCopy ); 
	}
            
	GlobalFree( hglbCopy );

	CloseClipboard();

	return bOk;
}

static bool GetRecipientNotes( cWCStr *wcsMsg, const wchar_t *wcsUserID )
{
	const char *szNotesBegin = 
		"Recipient notes:\r\n"
		"The ASCII armored block contains "; // Here is user ID

	cWCStr moduleName( MAX_PATH );
	int len = GetModuleFileNameT( NULL, moduleName.m_szBody, moduleName.m_len );

	const char *szNotesContinue_0 =
			" public key\r\n"
			"\r\n"
			"Please use following instructions to add the key to BCKeyManager local database:\r\n";

	const char *szNotesContinue_1 = NULL;

	if ( compareTails( moduleName.m_szBody, L"TextEncode.exe" ) == 0 )
		szNotesContinue_1 =
			"1. If you have no BestCrypt Text Encoder application on your computer,\r\n"
			"   please download it from http://www.jetico.com/bctextencoder.exe and run it.\r\n";
	else
		szNotesContinue_1 =
			"1. If BCArchive application is not installed on your computer,\r\n"
			"   please download it from http://www.jetico.com/bcarchive.exe and install it.\r\n";

	const char *szNotesContinue_2 =
			"2. Start the application and open Manage Key Database menu . \r\n"
			"   It brings up the BCKeyManager dialogue.\r\n"
			"3. Select entire the message text ( Ctrl+A ) and copy it to clipboard ( Crtl+C ).\r\n"
			"4. Open BCKeyManager->Key menu->Add Existing Public Key->Copy from Clipboard\r\n"
			"\r\n";

	wcsMsg->Add( szNotesBegin );
	wcsMsg->Add( wcsUserID );
	wcsMsg->Add( szNotesContinue_0 );
	wcsMsg->Add( szNotesContinue_1 );
	wcsMsg->Add( szNotesContinue_2 );

	return true;
}

//
//	Copy armored publc key text to clipboard
//
bool cDlgViewLocalDatabase::CopyPKToClipboard()
{
	cPTB ptbBase64Encoded;

	cWCStr wcsIserId;

	if ( !GetSelectedPublicKey( &ptbBase64Encoded, &wcsIserId ) )
		return false;
	
	cWCStr wcs;
//	GetRecipientNotes( &wcs, wcsIserId.m_szBody );

	wcs.AddF( L"The ASCII armored block contains public key.\r\nUser ID: %s\r\n\r\n", wcsIserId.m_szBody );

	wcs.Add( ptbBase64Encoded.buffer, ptbBase64Encoded.bufferSize );

	return CopyToClipboard( m_hDlg, wcs.m_szBody, wcs.m_len );
}

//
//	Send armored publc key by eMail
//	1. Sends selected certificate if x.509 crt is viewed.
//	2. Sends corresponded certifiacte for selected secret key p12.
//	3. Sends PGP key selected from local key ring.
//
bool cDlgViewLocalDatabase::SendPKToEMailRecipient()
{
	cPTB ptbBase64Encoded;
	//cPTB ptbUserID;
	cWCStr wcsUserID;

	if ( !GetSelectedPublicKey( &ptbBase64Encoded, &wcsUserID ) )
		return false;

	// Forms message text

	cWCStr wcs;

	GetRecipientNotes( &wcs, wcsUserID.m_szBody );

	const wchar_t *szSubject = L"Public Key Block ( formatted by BCKeyManager )";
	
	wcs.Add( (char *)ptbBase64Encoded.buffer, ptbBase64Encoded.bufferSize );

	int res = mapi_SendMail( m_hDlg, NULL, L"", L"", szSubject, wcs.m_szBody ); // (const char *)ptbMsg.buffer );

	return ( res == 0 );
}

wchar_t *cDlgViewLocalDatabase::GetHelpFileName( wchar_t *path, int wcSize )
{
	if ( ( pTranslateDll != NULL ) && ( pTranslateDll->GetHelpFileName( path, wcSize ) ) )
		return path;

	int len = GetModuleFileNameT( NULL, path, wcSize );

	const wchar_t *szBCArchive = L"BCArchive.exe";

	bool bBCArchivePath = _wcsicmp( path + len - wcsLen(szBCArchive), szBCArchive ) == 0;

	wcsncpy( path + len - 3, L"CHM", MAX_PATH - ( len - 3 ) );

	if ( !FileExistsT( path ) ) // Does not exist
	{
		if ( IDNO == MessageBoxT( m_hDlg, _LoadIDS( IDS_HelpFileNotFound_BrowseIt_Q ), _LoadIDS( IDS_AppName ), MB_YESNO | MB_ICONQUESTION ) )
			return NULL;

		if ( !BrowseFiles( m_hDlg, path, wcSize, _LoadIDS( IDS_HelpFilters ), _LoadIDS( IDS_ChooseHelpFile ), false ) )
		{
			assert( false );
			return NULL;
		}
	}

	if ( bBCArchivePath )
		wcsncpy( path + len, L"::/HTML_Help/html/06_local_public_key_database/01_local_public_key_database.htm", wcSize - len );
	else
		wcsncpy( path + len, L"::/HTML_Help/html/14_BestCrypt_utilities/05_1_public_key_manager.htm", wcSize - len );

	return path;
}

void cDlgViewLocalDatabase::OnHelp(void)
{
	wchar_t path[ 1024 ];

	if ( GetHelpFileName( path, maxLen( path ) ) == NULL )
		return;

	HWND hWnd = HtmlHelpT( GetDesktopWindow(), path, HH_DISPLAY_TOPIC, NULL );
}
