#include "KM_Funcs.h"
#include "..\CertificateLib\crtLibErrs.h"

CKeyManDll& CKeyManDll::getStatic(  bool bShowMessage )
{
	//ODSFunc;
	static CKeyManDll dll( false, L"BCKeyMan.dll", bShowMessage );

#ifdef _ODS
	if ( dll.error )
		ODS_f( L"Assertion" );
#endif
//	assert( !dll.error );

	return dll;
}

#define ppe_ok 0

typedef int typeSearchPKUserT(
		HINSTANCE hInst,
		HWND hWnd,
		cPTB *ptbKey,// = NULL, // If NULL the dialog is used for saving the key to BCDB.
		BYTE *keyID,// = NULL, // If key id specified the key ID filed is filled and then locked.
		cWCStr *wcsUserID,// = NULL,
		//cPTB *ptbUserID = NULL,// Chosen key block may not contain user ID,
										  // f.e. if it is a subkey. So it is useful to
										  // fill the value here.
		cPTB *ptbLastSavedKey// = NULL // If user saves key to BCDB it is copied to the ptb.
	);

int KM_SearchPKUser(
		HWND hWnd,
		cPTB *ptbKey, // = NULL -> if null the function is used to search key and save it to BCDB
		BYTE *keyID,// = NULL // If key id specified the key ID filed is filled and then locked.
		cWCStr *wcsUserID//cPTB *ptbUserID
	)
{
	static CKeyManDll &dll = CKeyManDll::getStatic( true );

	if ( dll.error != CDll::errNo )
	{
		ODS_fls( L"Assertion! Error in loading CKeyManDll" );
		return -1;
	}

	dll.GetProcAddress( "SearchPKUserT" );

	if ( dll.error )
		return -1;

	typeSearchPKUserT *proc = (typeSearchPKUserT *)dll.pProc;

	return proc( dll.hInst, hWnd, ptbKey, keyID, wcsUserID, NULL ); // ptbUserID, NULL );
}

typedef bool typeBrowseForPKBlockT(
		HINSTANCE hInst,
		HWND hWnd,
		const wchar_t *initialPath, // if initialPath == NULL it runs in BCDB
		bool secret,
		cPTB *ptbKey,	// [OUT] key in x509 or PGP format
		type_kff *pKeyPktFormat,	// [OUT] Format of key packet
		cWCStr *wcsUserID,			// [OUT] Pgp subkey does not have an user ID ( master key has only )
									// but we have to ask password for secret key with a name.
									// The BrowseKeyRingForUser generates here the subkey name as 
									// as "Subkey of ( master key user ID )" or  key user ID if the key has it
		cPTB *entirePacket
	);

bool KM_BrowseForPKBlock(
		HWND hWnd,
		const wchar_t *initialPath,// if initialPath == NULL it runs in BCDB
		bool secret,
		cPTB *ptbKey,// [OUT] key in x509 or PGP format
		type_kff *pKeyPktFormat,// [OUT] Format of key packet
		cWCStr			// [OUT] Pgp subkey does not have an user ID ( master key has only )
			*wcsUserID// = NULL	// but we have to ask password for secret key with a name.
								// The BrowseKeyRingForUser generates here the subkey name as 
								// as "Subkey of ( master key user ID )" or  key user ID if the key has it
	)
{
	static CKeyManDll &dll = CKeyManDll::getStatic( true );

	if ( ( initialPath == NULL )
		 )
			return KM_ViewLocalDatabase(
				hWnd,
				secret ? vld_flag_viewSecret : vld_flag_viewPublic,
				ptbKey,
				pKeyPktFormat,
				wcsUserID );

	if ( dll.error != CDll::errNo )
	{
		ODS_fls( L"Assertion! Error in loading CKeyManDll" );
		return false;
	}

	dll.GetProcAddress( "BrowseForPKBlockT" );

	if ( dll.error )
		return false;

	typeBrowseForPKBlockT *proc = (typeBrowseForPKBlockT *)dll.pProc;

	return proc( dll.hInst, hWnd, initialPath, secret, ptbKey, pKeyPktFormat, wcsUserID, NULL );
}

typedef bool typeGetPKBlockByIDT(
		HINSTANCE hInst,
		HWND hWnd,
		BYTE *keyID,
		int keyIDSize,
		bool secret,
		cPTB *ptbKey,// [OUT] key in x509 or PGP format
		type_kff *pKeyPktFormat,// [OUT] Format of key packet
		cWCStr *wcsUserID
	);

bool KM_GetPKBlockByID(
		HWND hWnd,
		BYTE *keyID,
		int keyIDSize,
		bool secret,
		cPTB *ptbKey,// [OUT] key in x509 or PGP format
		type_kff *pKeyPktFormat,// [OUT] Format of key packet
		cWCStr *wcsUserID
	)
{
	CKeyManDll &km = CKeyManDll::getStatic( true );
		
	if ( km.error != CDll::errNo )
	{
		ODS_fls( L"Assertion! Error in loading CKeyManDll" );
		return false;
	}

	if ( !km.GetProcAddress( "GetPKBlockByIDT" ) )
		return false;

	typeGetPKBlockByIDT *proc = (typeGetPKBlockByIDT *)km.pProc;

	return proc( km.hInst, hWnd, keyID, keyIDSize, secret, ptbKey, pKeyPktFormat, wcsUserID );
}

//
//	Function:
//		GenerateKeyPair
//	1. Generates key pair
//	2. Saves private key to file pfx_pdu format with .p12 extension
//		Adds the file to BCStorage.
//	3. Saves public key to certificate file with .cer extension.
//		The PK may be signed by its private key
//			Note, private key has to be added to BCStorage because the function takes
//			secret key attributes by ska local key ID
//		Or any other specified secret from BCStorage ( or may be from system storage WinNT )
//		Adds the file to BCStorage.
//	4. Fills localKeyID of the key been generated.
//
typedef int typeGenerateKeyPair(
		HINSTANCE hInst,
		HWND hWnd,
		BYTE *localKeyID,
		int *pLocalKeyIDBuffSize
	);

int KM_GenerateKeyPair(
		HINSTANCE hInst,
		HWND hWnd,
		BYTE *localKeyID, // == NULL
		int *pLocalKeyIDBuffSize // == NULL
	)
{
	static CKeyManDll &dll = CKeyManDll::getStatic( true );

	if ( dll.error != CDll::errNo )
	{
		ODS_fls( L"Assertion! Error in loading CKeyManDll" );
		return false;
	}

	dll.GetProcAddress( "GenerateKeyPair" );

	if ( dll.error )
		return -1;

	typeGenerateKeyPair *proc = (typeGenerateKeyPair *)dll.pProc;

	return proc( dll.hInst, hWnd, localKeyID, pLocalKeyIDBuffSize );
}

//
//	AddPublicKeyToStorage
//		Adds PGP public key to local key ring
//		Writes x509 certificate file to BCDB
//
bool KM_AddKeyToStorage( HWND hWnd, BYTE *pkt, int pktSize, type_kff kff, bool secret )
{
	static CKeyManDll &dll = CKeyManDll::getStatic( true );

	typedef bool typeAddKeyToStorage( HWND hWnd, BYTE *pkt, int pktSize, type_kff kff, bool secret );
	
	if ( dll.error != CDll::errNo )
	{
		ODS_fls( L"Assertion! Error in loading CKeyManDll" );
		return false;
	}

	dll.GetProcAddress( "AddKeyToStorage" );
	if ( dll.error )
	{
		dll.Unload();

		if ( secret )
			return false;

		return KM_AddPublicKeyToStorage( hWnd, pkt, pktSize, kff );
	}

	typeAddKeyToStorage *proc = (typeAddKeyToStorage *)dll.pProc;

	return proc( hWnd, pkt, pktSize, kff, secret );
}

bool KM_AddPublicKeyToStorage( HWND hWnd, BYTE *pkt, int pktSize, type_kff kff )
{
	static CKeyManDll &dll = CKeyManDll::getStatic( true );

	typedef bool typeAddPublicKeyToStorage( HWND hWnd, BYTE *pkt, int pktSize, type_kff kff );
	
	if ( dll.error != CDll::errNo )
	{
		ODS_fls( L"Assertion! Error in loading CKeyManDll" );
		return false;
	}

	dll.GetProcAddress( "AddPublicKeyToStorage" );
	if ( dll.error )
		return false;

	typeAddPublicKeyToStorage *proc = (typeAddPublicKeyToStorage *)dll.pProc;

	return proc( hWnd, pkt, pktSize, kff );
}

bool KM_BrowseForSecretKey(
		BYTE *keyID,
		int keyIDSize,
		cPTB *ptbKey,// [OUT] key in x509 or PGP format
		type_kff *pKeyPktFormat, // [OUT] Format of key packet
		cWCStr *wcsUserID // = NULL
	)
{
	static CKeyManDll &dll = CKeyManDll::getStatic( true );

	typedef bool typeBrowseForSecretKey(
		BYTE *keyID,
		int keyIDSize,
		cPTB *ptbKey,// [OUT] key in x509 or PGP format
		type_kff *pKeyPktFormat, // [OUT] Format of key packet
		cWCStr *wcsUserID // = NULL
	);

	if ( dll.error != CDll::errNo )
	{
		ODS_fls( L"Assertion! Error in loading CKeyManDll" );
		return false;
	}

	dll.GetProcAddress( "BrowseForSecretKey" );
	if ( dll.error )
		return false;

	typeBrowseForSecretKey *proc = (typeBrowseForSecretKey *)dll.pProc;

	return proc( keyID, keyIDSize, ptbKey, pKeyPktFormat, wcsUserID );
}


//
//	BrowseKeyFiles
//		Browses key file in BCDB ( default if pathBuff is empty ) or an other directory
//		Uses detailed view of p12 or cer files.
//		Returns true if key file is selected and fills pathBuff by file path name.
//
bool KM_BrowseKeyFiles(
		HWND hWnd,
		wchar_t *pathBuff, int buffSize,
		const wchar_t *filters,
		const wchar_t *boldCaptionText,
		const wchar_t *notes,
		const wchar_t *OKBtnText,
		DWORD dwFlags
	)
{
	static CKeyManDll &dll = CKeyManDll::getStatic( true );

	typedef bool typeBrowseKeyFiles(
			HWND hWnd,
			wchar_t *pathBuff, int buffSize,
			const wchar_t *filters,
			const wchar_t *boldCaptionText,
			const wchar_t *notes,
			const wchar_t *OKBtnText,
			DWORD dwFlags
		);
	
	if ( dll.error != CDll::errNo )
	{
		ODS_fls( L"Assertion! Error in loading CKeyManDll" );
		return false;
	}

	dll.GetProcAddress( "BrowseKeyFiles" );
	if ( dll.error )
		return false;

	typeBrowseKeyFiles *proc = (typeBrowseKeyFiles *)dll.pProc;

	return proc( hWnd, pathBuff, buffSize, filters, boldCaptionText, notes, OKBtnText, dwFlags );
}

//
//	KM_BrowseKeyRingForUser
//		Visual interfase for choosing a public or secret key from PGP key ring ( *.pkr, *.skr, *.gpg )
//		It allows to dynamically change key file path, so krOut contains the new path
//		ptbUserID is 
//			UserID for PGP key packet,
//			Friendly name for P12 secret key packet,
//			Subject RDN for certificate
//
int KM_BrowseKeyRingForUser(
		HWND hWnd,
		cPTB *ptbKey, // [OUT] 
		const char *keyRingPath, // [IN ]
		char *krOut,		 // [OUT] if krOut != NULL it is to be MAX_PATH size buffer.
		bool secret,
		cWCStr *wcsUserID
				// [OUT] Pgp subkey does not have an user ID ( master key has only )
				// but we have to ask password for secret key with a name.
				// The BrowseKeyRingForUser generates here the subkey name 
				// as "Subkey of ( master key user ID )" or  key user ID if the key has it
	)
{
	typedef bool typeBrowseKeyRingForUser(
			HINSTANCE hInst,
			HWND hWnd,
			cPTB *ptbKey, // [OUT] 
			const char *keyRingPath, // [IN ]
			char *krOut,		 // [OUT] if krOut != NULL it is to be MAX_PATH size buffer.
			bool secret,
			cWCStr *wcsUserID,
					// [OUT] Pgp subkey does not have an user ID ( master key has only )
					// but we have to ask password for secret key with a name.
					// The BrowseKeyRingForUser generates here the subkey name 
					// as "Subkey of ( master key user ID )" or  key user ID if the key has it
			cPTB *ptbKeyBlock // = NULL
					// Entire key packet - master key with subpackets including subkeys packets
		);
	
	static CKeyManDll &dll = CKeyManDll::getStatic( true );

	if ( dll.error != CDll::errNo )
	{
		ODS_fls( L"Assertion! Error in loading CKeyManDll" );
		return false;
	}

	dll.GetProcAddress( "BrowseKeyRingForUserT" );
	if ( dll.error )
		return false;

	typeBrowseKeyRingForUser *proc = (typeBrowseKeyRingForUser *)dll.pProc;

	return proc( dll.hInst, hWnd, ptbKey, keyRingPath, krOut, secret, wcsUserID, NULL );
}

//
//	KM_GetDataBasePath
//		Returns path to local database.
//
bool KM_GetDataBasePath( wchar_t *buff, int buffLen, DWORD dwFlag )
{
	static CKeyManDll &dll = CKeyManDll::getStatic( true );

	typedef bool typeGetDataBasePath( wchar_t *buff, int buffLen, DWORD dwFlag );
	
	if ( dll.error != CDll::errNo )
	{
		ODS_fls( L"Assertion! Error in loading CKeyManDll" );
		return false;
	}

	dll.GetProcAddress( "GetDataBasePath" );
	if ( dll.error )
		return false;

	typeGetDataBasePath *proc = (typeGetDataBasePath *)dll.pProc;

	return proc( buff, buffLen, dwFlag );
}

cPublicRecipientList *KM_UpdatePublicRecipientList( cPublicRecipientList *prl )
{
	static CKeyManDll &dll = CKeyManDll::getStatic( true );

	typedef cPublicRecipientList *typeUpdatePublicRecipientList( cPublicRecipientList *prl );
	
	if ( dll.error != CDll::errNo )
	{
		ODS_fls( L"Assertion! Error in loading CKeyManDll" );
		return false;
	}

	dll.GetProcAddress( "UpdatePublicRecipientList" );
	if ( dll.error )
		return false;

	typeUpdatePublicRecipientList *proc = (typeUpdatePublicRecipientList *)dll.pProc;

	return proc( prl );
}

void KM_FreePublicRecipientList( cPublicRecipientList *prl )
{
	static CKeyManDll &dll = CKeyManDll::getStatic( true );

	typedef void typeFreePublicRecipientList( cPublicRecipientList *prl );
	
	if ( dll.error != CDll::errNo )
	{
		ODS_fls( L"Assertion! Error in loading CKeyManDll" );
		return;
	}

	dll.GetProcAddress( "FreePublicRecipientList" );
	if ( dll.error )
		return;

	typeFreePublicRecipientList *proc = (typeFreePublicRecipientList *)dll.pProc;

	proc( prl );
}

//
//	ViewLocalDatabase
//		1. Shows and manages keys in local data base.
//		2. If paramters ptbKey and pKFF are specified it allows to choose
//			key from local database.
//

bool KM_ViewLocalDatabase(
		HWND hWnd,
		DWORD viewFlags, // = vld_flag_viewSecret | vld_flag_viewPublic,
		cPTB *ptbKey, // = NULL,
		type_kff *pKFF,// = NULL
		cWCStr			// [OUT] Pgp subkey does not have an user ID ( master key has only )
			*wcsUserID// = NULL	// but we have to ask password for secret key with a name.
								// The BrowseKeyRingForUser generates here the subkey name as 
								// as "Subkey of ( master key user ID )" or  key user ID if the key has it
	)
{
	ODSFuncS( _FormatT( L"%X, %p, %p, %p", viewFlags, ptbKey, pKFF, wcsUserID )  );

	static CKeyManDll &dll = CKeyManDll::getStatic( true );

	typedef bool typeViewLocalDatabaseT(
			HWND hWnd,
			DWORD viewFlags,// = vld_flag_viewSecret | vld_flag_viewPublic,
			cPTB *ptbKey, // = NULL,
			type_kff *pKFF,// = NULL,
			cWCStr *wcsUserID// = NULL
		);
	
	if ( dll.error != CDll::errNo )
	{
		ODS_fls( L"Assertion! Error in loading CKeyManDll" );
		return false;
	}

	dll.GetProcAddress( "ViewLocalDatabaseT" );
	if ( dll.error )
		return false;

	typeViewLocalDatabaseT *proc = (typeViewLocalDatabaseT *)dll.pProc;

	ODS_fls( _FormatT( L"proc == %p", proc ) );

	bool ok = proc( hWnd, viewFlags, ptbKey, pKFF, wcsUserID );

	ODSFuncRes( ok );

	return ok;
}

//
//	KM_EnumKeys uses BCDB_EnumPGPKeysBf and/or BCDB_EnumX509KeysBf functions according kff type.
//	Or BCDB_EnumKeysBf for both types.
//
//	Sample:
//  123456789, <qqq@qqqq.qqq>
//
//	Note, the key id may be used for extracting key from BCDB by GetPKBlockByID
//
bool KM_EnumKeys( cWCStr *wcsText, bool secret, type_kff kff ) //  kffNone )
{
	static CKeyManDll &dll = CKeyManDll::getStatic( true );

	const char *procName = "BCDB_EnumKeysBf";

	if ( kff == kffX509 )
		procName = "BCDB_EnumX509KeysBf";
	if ( kff == kffPGP )
		procName = "BCDB_EnumPGPKeysBf";

	if ( dll.error != CDll::errNo )
	{
		ODS_fls( L"Assertion! Error in loading CKeyManDll" );
		return false;
	}

	dll.GetProcAddress( procName );
	if ( dll.error )
		return false;

	typedef bool typeEnumKeysBf( wchar_t *wc, size_t *pLen, bool secret );
	typeEnumKeysBf *proc = (typeEnumKeysBf *)dll.pProc;

	size_t len = 0;

	int res = proc( NULL, &len, secret );

	if ( res == ppe_ok ) // Empty list.
		return true;

	if ( res != ppe_bufferTooSmall )
		return false;

	int oldPos = wcsText->m_len;

	if ( wcsText->Add( (int)len ) == 0 )
	{
		ODS_fls( L"Assertion!" );
		return false;
	}

	wchar_t *wc = wcsText->m_szBody + oldPos;

	res = proc( wc, &len, secret );

	return res == ppe_ok;
}