#include "BCKeyMan.h"
#include "PKSearch.h"
#include "proxySrv.h"
//#include "..\certificatelib\wincrypt\certstore.h" // For X509Crt verification.
#include "PKParse.h"
#include <stdio.h>
#include <assert.h>
#include "..\Includes\configFile.h"

//#include <certpol.h>
//#include <certsrv.h>
//#include <certcli.h> CryptExtAddCER

// http://wwwkeys.pgp.net:11371/pks/lookup?op=get&search=0x630A6EC4

//
//	With finger print request
//	Is:
//		http://pks.gpg.cz:11371/pks/lookup?op=index&search=ser@iae&fingerprint=on
//	Returns:
//	Public Key Server -- Index ``ser@iae ''
//
//	Type  bits/keyID    Date       User ID
//	--------------------------------------------------------------------------------
//	pub  1024D/630A6EC4 2001-07-16 Ser <ser@iae.nsk.su>
//	     Key fingerprint = C7EE 6C05 F70A 04E7 0EC5  F9D6 A29C A0CC 630A 6EC4
//
//
//	https://pgp.mit.edu/pks/lookup?search=%3Cv.saknys%40gmail.com%3E&op=index
//

int PgpLookupGetInfo(
		sPGP_URL *url,
		cPlainTextBlock *ptbOut,
		char *searchStr,
		bool index,
		bool skipVerification,
		bool decode,
		bool *pCancelled
	)
{
	char *pattern = index ? url->getIndexPattern : url->getKeyPattern; 
	//
	//	Prepare requiest string.
	//
	char *ptr = new char[ strlen( pattern ) + strlen( searchStr ) + 1];
	if ( ptr == NULL ) return ppe_memoryAllocation;

	sprintf( ptr, pattern, searchStr );

	//
	//	Set network connection 
	//
	HNCC hc;
	int res = ncInit( &hc );

	if ( res == nce_ok )
		res = ncConnect( hc, (sURL *)url, cProxySrv::getProxyURL( url->pType ) );

	if ( ( res == nce_ok ) && ( url->pType == ncpt_HTTPS ) )
	{
		//
		//	Get and verify SSL certificate.
		//

		int crtSize = 0;
		res = ncGetCertificate( hc, NULL, &crtSize );

		cPlainTextBlock crtPtb;
		if ( res == nce_bufferTooSmall )
		{
			if ( crtPtb.Allocate( crtSize ) == NULL )
				res = nce_memoryAllocation;
			else
			{
				res = ncGetCertificate( hc, crtPtb.buffer, &crtSize );
				if ( !skipVerification )
				{
					assert( false );
					//res = verifyX509Certificate( crtPtb.buffer, crtPtb.bufferSize );

					if ( res != ppe_ok )
					{
						delete ptr;
						return res;
					}
				}
			}
		}
	}

	cPlainTextBlock ptbMessage;
	if ( res == nce_ok )
		res = ncGetData( hc, ptr, (ncWriteFuncType *)cPlainTextBlock::WriteBytesStatic, &ptbMessage, pCancelled );

	ncShutDown( hc );

	delete ptr;

	if ( res != nce_ok )
		return res;

	if ( decode )
	{
		int parsed = 0;
		return DecodePGPPublicKeyBlock( ptbMessage.buffer, ptbMessage.bufferSize, ptbOut, &parsed );
	}
	else
	{
		if ( ptbOut->WriteBytes( ptbMessage.buffer, ptbMessage.bufferSize ) == 0 )
			res = pkae_memoryAllocation;

		return res;
	}
}

sPGP_URL pgpSrv1 = 
{
	ncpt_HTTP,
	"wwwkeys.pgp.net",
	11371,
	NULL,
	NULL,
	"/pks/lookup?op=index&search=%s",
	"/pks/lookup?op=get&search=%s"
};

sPGP_URL pgpSrv2 = 
{
	ncpt_HTTP,
	"keys.pgp.com",
	11371,
	NULL,
	NULL,
	"/pks/lookup?op=index&search=%s",
	"/pks/lookup?op=get&search=%s"
};

sPGP_URL pgpSrv3 = 
{
	ncpt_HTTP,
	"wwwkeys.us.pgp.net",
	11371,
	NULL,
	NULL,
	"/pks/lookup?op=index&search=%s",
	"/pks/lookup?op=get&search=%s"
};

sPGP_URL pgpSrv4 = 
{
	ncpt_HTTP,
	"wwwkeys.nl.pgp.net",
	11371,
	NULL,
	NULL,
	"/pks/lookup?op=index&search=%s",
	"/pks/lookup?op=get&search=%s"
};

sPGP_URL pgpSrv5 = 
{
	ncpt_HTTP,
	"wwwkeys.ch.pgp.net",
	11371,
	NULL,
	NULL,
	"/pks/lookup?op=index&search=%s",
	"/pks/lookup?op=get&search=%s"
};

sPGP_URL pgpSrv6 = 
{
	ncpt_HTTP,
	"wwwkeys.cz.pgp.net",
	11371,
	NULL,
	NULL,
	"/pks/lookup?op=index&search=%s",
	"/pks/lookup?op=get&search=%s"
};

sPGP_URL pgpSrv7 = 
{
	ncpt_HTTPS,
	"pgp.mit.edu",
	11371,
	NULL,
	NULL,
	"/pks/lookup?op=index&search=%s",
	"/pks/lookup?op=get&search=%s"
};

static sPGP_URL *pgpSrvs_internal[] = {
	&pgpSrv1,
	&pgpSrv2,
	&pgpSrv3,
	&pgpSrv4,
	&pgpSrv5,
	&pgpSrv6,
	&pgpSrv7,
	NULL
};

sPGP_URL **pgpSrvs = NULL;

static char *allocateNewString( const wchar_t *str )
{
	return newString( _getMbs( str ) );
}

static void FreePGPServicesList( sPGP_URL **pListT )
{
	for( int i = 0; pListT[i] != NULL; i++ )
	{
		if ( pListT[i]->addr != NULL ) delete pListT[i]->addr;
		if ( pListT[i]->getIndexPattern != NULL ) delete pListT[i]->getIndexPattern;
		if ( pListT[i]->getKeyPattern != NULL ) delete pListT[i]->getKeyPattern;
		if ( pListT[i]->pwd != NULL ) delete pListT[i]->pwd;
		if ( pListT[i]->usr != NULL ) delete pListT[i]->usr;

		delete pListT[i];
	}

	delete pListT;
}

bool FreePGPServicesList()
{
	if ( pgpSrvs == NULL )
		return false;

	if ( pgpSrvs == pgpSrvs_internal )
	{
		pgpSrvs = NULL;
		return false;
	}

	FreePGPServicesList( pgpSrvs );
		
	pgpSrvs = NULL;

	return true;
};

static sPGP_URL **readCFGFile( const wchar_t *path )
{
	if ( !FileExistsT( path ) )
		return NULL;

	cConfigFile cfgFile( path );

	const wchar_t *szUrlPattern = L"[URL%s]";

	// Get count

	int count = 0;

	for( ;; count++ )
	{
		wchar_t szNum[64];
		_itow( count, szNum, 10 );

		if ( !cfgFile.FindString( cWCStr().FormatT( szUrlPattern, szNum ) ) )
			break;
	}

	if ( count == 0 )
		return NULL;

	sPGP_URL_ptr *urlTab = new sPGP_URL_ptr[count + 1 ]; // NULL terminated buffer.

	if( urlTab == NULL )
		return NULL;

	urlTab[ count ] = NULL;

	for( int i = count; i--; )
	{
		wchar_t szNum[64];
		
		_itow( i, szNum, 10 );

		cWCStr wcs;
		const wchar_t *szUrlGroup = wcs.FormatT( szUrlPattern, szNum );

		wchar_t buff[ 1024 ];

		if ( !cfgFile.FindString( szUrlGroup, L"Address", buff, maxLen( buff ) ) )
			return NULL;

		urlTab[i] = new sPGP_URL;

		if ( urlTab[i] == NULL )
			return NULL;

		memset( urlTab[i], 0, sizeof( sPGP_URL ) );
		
		urlTab[i]->addr = allocateNewString( buff );

		if ( cfgFile.FindString( szUrlGroup, L"Type", buff, maxLen( buff ) ) )
				urlTab[i]->pType = getUrlType( getMBStr( buff ) );
		else	urlTab[i]->pType = ncpt_HTTP;

		if ( cfgFile.FindString( szUrlGroup, L"Port", buff, maxLen( buff ) ) )
				urlTab[i]->port = _wtoi( buff );
		else	urlTab[i]->port = 80;

		if ( !cfgFile.FindString( szUrlGroup, L"IndexPattern", buff, maxLen( buff ) ) )
			return NULL;

		urlTab[i]->getIndexPattern = allocateNewString( buff );

		if ( !cfgFile.FindString( szUrlGroup, L"GetPattern", buff, maxLen( buff ) ) )
			return NULL;

		urlTab[i]->getKeyPattern = allocateNewString( buff );
	}

	return urlTab;
}

bool GetPGPServicesList( HINSTANCE hInst )
{
	if ( pgpSrvs != NULL )
		FreePGPServicesList();

	wchar_t path[MAX_PATH];
	GetModuleFileNameT( hInst, path, maxLen( path ) );
	int pos = splitPath( path );
	wcsncpy( path + pos, L"keyServerList.txt", maxLen( path ) - pos );

	sPGP_URL **tmpPGPSrvs = readCFGFile( path );

	if ( tmpPGPSrvs != NULL )
	{
		pgpSrvs = tmpPGPSrvs;
		return true;
	}
	
	pgpSrvs = pgpSrvs_internal;

	return false;
}


//
//	getKeyByID
//		Seach key on PGP server.
//
static
 int getKeyByID( 
		HWND hWnd,
		char *srvName,
		BYTE *keyID,
		cPlainTextBlock *ptbSearchResult
	)
{
	
}

//
//	SearchPKUserByID
//		Queries specified key on server. 
//		Fills ptbKey block if key found and returns ppe_ok.
//		Otherwise returns a error code.
//
int SearchPKUserByID(
		HINSTANCE hInst,
		HWND hWnd,
		cPlainTextBlock *ptbKey,
		BYTE *keyID
	)
{
	return 0;
}

//
//	SearchPKUserIDs
//		Queries list of users with specified info. 
//		Fills ptbUser block with query result.
//
int SearchPKUserIDs(
		HINSTANCE hInst,
		HWND hWnd,
		cPlainTextBlock *ptbUsers,
		int urlNum,
		char *userInfo,
		bool *pCancelled
	)
{
	assert( pgpSrvs != NULL );

	if ( pgpSrvs == NULL )
		return ppe_internalError;

	return PgpLookupGetInfo( pgpSrvs[ urlNum ], ptbUsers, userInfo, true, true, false, pCancelled );
}

