#include "resource.h"
#include "BCKeyMan.h"
#include "PKSearch.h"
#include "PublicKeyProps.h"
#include "PKParse.h"
#include "BCDB_Functions.h"
#include "pksearchdlg.h"
#include ".\nwclient\http_get.h"

#include "..\CertificateLib\PGPPkts\pktdefs.h"
#include "..\CertificateLib\PGPPkts\pgppkts.h"

#include "..\Includes\cDlg.h"
#include "..\Includes\ptbRegs.h"
#include "..\Includes\menuItem.h"

#include <commctrl.h>
#include <windowsx.h>
#include <stdio.h>

class cDlgPKSearch : public cColoredDlg
{
//	cSizedCtrls sizedCtrls;
	cPlainTextBlock *m_ptbKey;	  // The block will by filled by internal key packet without 
								  //	main header, ( a subkey may be chosen ). 
	cWCStr *m_wcsUserID;
	//cPlainTextBlock *m_ptbUserID; // User ID may not present in subkey, so is is useful to
								  // fill the value here.
	cPlainTextBlock *m_ptbLastSavedKey; // If user saves key to BCDB it is copied to the ptb.

	BYTE *m_keyID; // If key id specified the key ID field is filled and then locked.

	HWND m_hList;
	bool m_cancelled;
	HANDLE m_searchThread;

 public:

	enum { ID = IDD_DlgPKSearch };
	cDlgPKSearch( cPlainTextBlock *ptbKey, BYTE *keyID = NULL, cWCStr *wcsUserID = NULL,
		cPlainTextBlock *ptbLastSavedKey = NULL );
	~cDlgPKSearch();

	bool Search();

 private:

	LRESULT RealDlgProc( UINT uMsg, WPARAM wParam, LPARAM lParam );
	BOOL InitDialog();
	bool OnNotify( NMHDR *hdr );
	void DeleteDialog();
	bool OnProperties();
	bool OnStartSearch();
	bool EnableButtons( bool enable );
//	bool OnStopSearch();
	bool GetCurrentKey( cPlainTextBlock *ptbOut );

	DWORD m_timeStart;
	DWORD m_timerID;
	static void CALLBACK SetTimeValue( HWND hwnd, UINT uMsg, UINT_PTR idEvent, DWORD dwTime );
	void InitTimer();
};


cDlgPKSearch::cDlgPKSearch( cPlainTextBlock *ptbKey, BYTE *keyID, cWCStr *wcsUserID, cPlainTextBlock *ptbLastSavedKey )
{
	m_ptbKey = ptbKey;
	m_keyID = keyID;
	m_wcsUserID = wcsUserID;
	m_ptbLastSavedKey = ptbLastSavedKey;

	SetRCValues( IDC_Line, IDC_STATIC_BOLD, IDC_Text );
}

cDlgPKSearch::~cDlgPKSearch()
{
}

static UINT ColumnLabelIDS[] = 
{
	IDS_szKeyID, IDS_szKeySize, IDS_szDate, IDS_szUserID
};

const int NUM_COLUMNS = sizeof( ColumnLabelIDS ) / sizeof( UINT );

static int _gnColumnFmt[ NUM_COLUMNS ] =
{
	LVCFMT_RIGHT, LVCFMT_RIGHT, LVCFMT_CENTER, LVCFMT_CENTER
};

static int _gnColumnWidth[NUM_COLUMNS]=
{
	70, 60, 100, 230
};

static BOOL SetItemValues( HWND hLView, int itemNum, char *id, char *bits, char *date, char *userID )
{
	int subItem = 0;

	ListView_SetItemTextT( hLView, itemNum,   subItem, getWCStr( id ) );
	ListView_SetItemTextT( hLView, itemNum, ++subItem, getWCStr( bits ) );
	ListView_SetItemTextT( hLView, itemNum, ++subItem, getWCStr( date ) );
	ListView_SetItemTextT( hLView, itemNum, ++subItem, getWCStr( userID ) );

	return TRUE;
}

BOOL InitListViewColunms( HINSTANCE hInst, HWND hLView )
{
	int i;
	LV_COLUMNW lvc;

	memset( &lvc, 0, sizeof( lvc ) );

	lvc.mask = LVCF_WIDTH;

	//
	//	Check if already initialised...
	//
	if ( ListView_GetColumnT( hLView, 0, &lvc ) )
		return TRUE;

	lvc.mask = LVCF_FMT | LVCF_WIDTH | LVCF_TEXT | LVCF_SUBITEM;

	wchar_t text[ 80 ];

	int res;

	for(i=0; i<NUM_COLUMNS; i++)
	{
		LoadStringT( hInst, ColumnLabelIDS[i], text, maxLen( text ) );

		lvc.iSubItem=i;
		lvc.pszText = text; // _gszColumnLabel[i];
		lvc.cx = _gnColumnWidth[i];
		lvc.fmt=_gnColumnFmt[i];

		res = ListView_InsertColumnT( hLView, i, &lvc );

		if ( res == -1 )
		{
			MessageBoxT( 0, text, _LoadIDS( IDS_ErrorInAddLVColumn ), MB_OK );
			return FALSE;
		}
	}

	DWORD dwStyle = LVS_EX_FULLROWSELECT; // | LVS_EX_GRIDLINES;
	ListView_SetExtendedListViewStyleEx( hLView, dwStyle, dwStyle );

	return TRUE;
}

int ListViewInit( HINSTANCE hInst, HWND hList )
{
	InitListViewColunms( hInst, hList );
	return 0;
}

BYTE *findStringInBuffer( BYTE *beg, BYTE *end, const char *str )
{ 
	int len = strlen( str );
	for(; beg<end; beg++ )
		if ( memcmp( beg, str, len ) == 0 )
			break;

	return beg;
}

BYTE *findStringInBuffer_End( BYTE *beg, BYTE *end, const char *str )
{ 
	int len = strlen( str );
	for(; beg<end; beg++ )
	{
		if ( memcmp( beg, str, len ) == 0 )
		{
			beg += len;
			break;
		}
	}

	return beg;
}

int AddLVUint( HWND hList, char *id, char *bits, char *data, char *userID, int num )
{
	LVITEMW lvi;

	memset( &lvi, 0, sizeof( lvi ) );

	lvi.mask = 0; // LVIF_IMAGE;
	lvi.iItem = num;

	int item = ListView_InsertItemT( hList, &lvi );

	SetItemValues( hList, item, id, bits, data, userID );

	return item;
}

int copyStr( char *bf, int size, const BYTE *start, const BYTE *finish )
{
	if ( size <= 0 )
		return 0;

	int i;

	for ( i = 0;  ; i++ )
	{
		if ( i >= size - 1 ) break;
		if ( start + i >= finish ) break;

		bf[i] = (char)start[i];
	}

	bf[i] = '\0';

	return i;
}

// pub  1024/<a href="/pks/lookup?op=get&search=0x630A6EC4">630A6EC4</a> 2001/07/16 Ser &lt;<a href="/pks/lookup?op=get&search=0x630A6EC4">ser@iae.nsk.su</a>&gt;
// pub  1024/<a href="/pks/lookup?op=get&search=0x9808FF12">9808FF12</a> 2001/07/12 Sergei Kashpirko &lt;<a href="/pks/lookup?op=get&search=0x9808FF12">ser-kas@mail.ru</a>&gt;
// pub  3072R/<a href="/pks/lookup?op=get&amp;search=0x27F118F767F3EBED">67F3EBED</a> 2013-09-14 <a href="/pks/lookup?op=vindex&amp;search=0x27F118F767F3EBED">Vykintas Saknys &lt;v.saknys@gmail.com&gt;</a>

int addPKUser( HWND hList, BYTE *ptr, BYTE *end, int *pCount )
{
	char bits[ 20 ];
	char ID[ 40 ];
//	char ref[ 1024 ];
	char userID[ 1024 ];
	char date[ 80 ];

	end = findStringInBuffer( ptr + 2, end, "\n" ); // "\r\n" );

	if ( end == ptr )
		return 0;

	if ( memcmp( ptr, "pub ", 4 ) != 0 )
		return 0;

	ptr += 4;

	while( *ptr == ' ' ) ptr++;

	BYTE *lEnd = findStringInBuffer( ptr, end, "/<a" );
	if ( end == lEnd )
		return 0;

	copyStr( bits, sizeof( bits ), ptr, lEnd );
	
	ptr = findStringInBuffer_End( lEnd, end, "search=0x" );
	if ( ptr == end )
		return 0;

	lEnd = findStringInBuffer( ptr, end, "\">" );
	if ( end == lEnd )
		return 0;

	copyStr( ID, sizeof( ID ), ptr, lEnd );

	ptr = findStringInBuffer_End( ptr, end, "</a> " );
	if ( ptr == end ) return 0;
	while( *ptr == ' ' ) ptr++;

	lEnd = findStringInBuffer( ptr, end, " " );
	if ( lEnd == end ) return 0;

	copyStr( date, sizeof( date ), ptr, lEnd );

	ptr = lEnd + 1;
	if ( memcmp( ptr, "<a", 2 ) == 0 ) // Skips available reference
	{
		ptr = findStringInBuffer_End( ptr, end, ">" );
		if ( ptr == end ) return 0;
	}

	lEnd = findStringInBuffer( ptr, end, "&lt;" ); // html symbol of '<'

	BYTE *leftTagPos = NULL;
	if ( lEnd != end )
		leftTagPos = lEnd;

	while( *(lEnd - 1) == ' ' ) lEnd--;
	
	int len = copyStr( userID, sizeof( userID ), ptr, lEnd );

	if ( leftTagPos != NULL )
	{
		ptr = leftTagPos + 4;

		if ( memcmp( ptr, "<a", 2 ) == 0 ) // Skips available reference
		{
			ptr = findStringInBuffer_End( ptr, end, ">" );
			if ( ptr == end ) return 0;
			lEnd = findStringInBuffer( ptr, end, "</a>&gt;" );
		}
		else
			lEnd = findStringInBuffer( ptr, end, "&gt;" );

		if ( lEnd == end ) return 0;

		strncpy( userID + len, " <", sizeof( userID ) - len );
		len += 2;
		len += copyStr( userID + len, sizeof( userID ) - len, ptr, lEnd );
		strncpy( userID + len, ">", sizeof( userID ) - len );
	}

	AddLVUint( hList, ID, bits, date, userID, (*pCount)++ );
	
	return 1;
}

static void	LV_DeleteAllItems( HWND hList )
{
	for ( int hItem = -1; ( hItem = ListView_GetNextItem( hList, hItem, LVNI_ALL ) ) != -1 ; )
	{
		cPlainTextBlock *ptb = (cPlainTextBlock *)ListView_GetItemParam( hList, hItem );
		if ( ptb != NULL )
			delete ptb;
	}
	
	ListView_DeleteAllItems( hList );
}

#ifdef _AAAAA_
Sample:
<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd" >
<html xmlns="http://www.w3.org/1999/xhtml">
<head>
<title>Search results for 'v saknys gmail com'</title>
<meta http-equiv="Content-Type" content="text/html;charset=utf-8" />
<style type="text/css">
/*<![CDATA[*/
 .uid { color: green; text-decoration: underline; }
 .warn { color: red; font-weight: bold; }
/*]]>*/
</style></head><body><h1>Search results for 'v saknys gmail com'</h1><pre>Type bits/keyID     Date       User ID
</pre><hr /><pre>
pub  3072R/<a href="/pks/lookup?op=get&amp;search=0x27F118F767F3EBED">67F3EBED</a> 2013-09-14 <a href="/pks/lookup?op=vindex&amp;search=0x27F118F767F3EBED">Vykintas Saknys &lt;v.saknys@gmail.com&gt;</a>
</pre><hr /><pre>
pub  4096R/<a href="/pks/lookup?op=get&amp;search=0xCEC4D2D515B48677">15B48677</a> 2011-06-26 <a href="/pks/lookup?op=vindex&amp;search=0xCEC4D2D515B48677">Vykintas Saknys &lt;v.saknys@gmail.com&gt;</a>
</pre></body></html>
#endif 


int ParseDataAndFillListView( HWND hList, cPlainTextBlock *ptb )
{
	int selected = ListView_GetNextItem( hList, -1, LVNI_SELECTED );

	LV_DeleteAllItems( hList );
	BYTE *end = ptb->buffer + ptb->bufferSize;
	BYTE *ptr = ptb->buffer;

	int count = 0;

	int bits = 0;

	while( true )
	{
		ptr = findStringInBuffer( ptr, end, "<pre>" );
		if ( ptr >= end )
			break;

		ptr+=5;

		while( strchr( "\r\n \t", *ptr ) != NULL )
		{
			if ( ptr++ >= end ) break;
		}

		if ( ptr >= end ) break;

		if ( strncmp( (char*)ptr, "pub", 3 ) != 0 )
			continue;

		count += addPKUser( hList, ptr, end, &count );
	}

	if ( count == 0 )
	{
		HDC hdc = GetDC( hList );
		RECT r;
		GetWindowRect( hList, &r );
		TextOutT( hdc, r.left, r.right, _LoadIDS( IDS_NoOneUserFound ), 10 );
	}

	if ( count >= selected )
		selected = 0;

	ListView_SetItemState( hList, selected, LVNI_SELECTED | LVNI_FOCUSED, LVNI_SELECTED | LVNI_FOCUSED );
	SetFocus( hList );

	return count;
}

int ParseDataAndFillListView_OLD( HWND hList, cPlainTextBlock *ptb )
{
	int selected = ListView_GetNextItem( hList, -1, LVNI_SELECTED );

	LV_DeleteAllItems( hList );
	BYTE *end = ptb->buffer + ptb->bufferSize;
	BYTE *ptr = NULL;

	//ptr = findStringInBuffer( ptb->buffer, end, "Type bits /keyID" );
	//if ( ptr == end )
	//	ptr = findStringInBuffer( ptb->buffer, end, "Type bits/keyID" );
	//if ( ptr == end )
	//	ptr = findStringInBuffer( ptb->buffer, end, "Type  bits/keyID" );
	//if ( ptr == end )
	//	return 0;

	ptr = findStringInBuffer( ptb->buffer, end, "<pre>" );
	if ( ptr == end )
		return 0;

	end = findStringInBuffer( ptr, end, "</pre>" );

	int count = 0;

	int bits = 0;

	while( true )
	{
		ptr = findStringInBuffer( ptr, end, "\npub " ); // "\r\n" );
		if ( ptr + 5 >= end ) break;

		ptr += 1; // 2; //start from the "pub "

		count += addPKUser( hList, ptr, end, &count );
	}

	if ( count == 0 )
	{
		HDC hdc = GetDC( hList );
		RECT r;
		GetWindowRect( hList, &r );
		TextOutT( hdc, r.left, r.right, _LoadIDS( IDS_NoOneUserFound ), 10 );
	}

	if ( count >= selected )
		selected = 0;

	ListView_SetItemState( hList, selected, LVNI_SELECTED | LVNI_FOCUSED, LVNI_SELECTED | LVNI_FOCUSED );
	SetFocus( hList );

	return count;
}
static char hex2char( BYTE hex )
{
	hex &= 0x0F;

	if ( hex < 0xA )
		return '0' + hex;
	else
		return 'A' + hex - 0xA;
}

static void makeKeyIDStr( char *buff, BYTE *keyID )
{
	int i;

	*(buff++) = '0';
	*(buff++) = 'x';

	keyID += 4; // last 4 bytes a required for sear

	for( i = 0; i < 4; i++ ) 
	{
		buff[2*i]		= hex2char( keyID[i] >> 4 );
		buff[2*i + 1 ]	= hex2char( keyID[i] );
	}

	buff[ 2*i ] = 0;
}

class cPKSearchValues : public cRegDlgValues
{
public:
	int m_lookupID;

	cPKSearchValues( HKEY hRoot, const wchar_t *keyName, const wchar_t *valueName, int columns )
		: cRegDlgValues( hRoot, keyName, valueName, sizeof( *this ), columns )
	{
	}

	bool Save( HWND dlg, HWND listView )
	{
		HWND hCmb = GetDlgItem( dlg, IDC_COMBO );
		m_lookupID = ComboBox_GetCurSel( hCmb );
		return cRegDlgValues::Save( dlg, listView );
	}
};

cPKSearchValues dlgVals( hKeyRoot, szBCKeyManKey, L"PKSearch", NUM_COLUMNS ); 

const int aNumber = 25; 

BOOL cDlgPKSearch::InitDialog()
{
	//
	// (WM_USER + 1) allows to catch WM_COMMAND with the number if VK_RETURN pressed
	//
	SendMessage( m_hDlg, WM_USER + 1, aNumber, 0 );

//	CenterPosition();

	SendMessage( m_hDlg, WM_SETICON, 0, (LPARAM)LoadIcon( m_hInst, MAKEINTRESOURCE( IDI_Main ) ) );

	//
	//	Hide Get It button if m_ptbKey is undefined -> the dialog called to search and save public key in local database.
	//
	if ( m_ptbKey == NULL )
	{
		HWND btnGet = GetDlgItem( m_hDlg, IDC_btnGet );

		RECT r;
		GetWindowRect( btnGet, &r );
		//POINT pos = { r.left, r.top };
		//ScreenToClient( m_hDlg, &pos );
		MapWindowPoints( NULL, m_hDlg, (LPPOINT)&r, 2 );
		ShowWindow( btnGet, SW_HIDE );

		// Move button Properties to the button Get It place

		HWND btnProperties = GetDlgItem( m_hDlg, IDC_btnProperties );
//		SetWindowPos( btnProperties, NULL, pos.x, pos.y, 0, 0, SWP_NOSIZE );
		SetWindowPos( btnProperties, NULL, r.left, r.top, 0, 0, SWP_NOSIZE );

		SetDlgItemTextT( IDCANCEL, _LoadIDS( IDS_szBntClose ) );
	}

//	m_sizedCtrls.Init( m_hDlg );
	
	m_sizedCtrls.AddCtrl( IDC_btnSearch,	cSizedCtrls::ptPR );
	m_sizedCtrls.AddCtrl( IDC_TimeBox,		cSizedCtrls::ptPR );
	m_sizedCtrls.AddCtrl( IDC_COMBO,		cSizedCtrls::ptSR );
	m_sizedCtrls.AddCtrl( IDC_EDIT,			cSizedCtrls::ptSR );
	m_sizedCtrls.AddCtrl( IDC_LISTVIEW,		cSizedCtrls::ptSDR );
	m_sizedCtrls.AddCtrl( IDC_btnSave,		cSizedCtrls::ptPDR );
	m_sizedCtrls.AddCtrl( IDC_btnProperties,cSizedCtrls::ptPDR );
	m_sizedCtrls.AddCtrl( IDCANCEL,			cSizedCtrls::ptPDR );
	m_sizedCtrls.AddCtrl( IDC_Line,			cSizedCtrls::ptSR );
	m_sizedCtrls.AddCtrl( IDC_bmpHeader,	cSizedCtrls::ptPR );
	m_sizedCtrls.AddCtrl( IDC_Description,	cSizedCtrls::ptSR );
	m_sizedCtrls.AddCtrl( IDC_LINE_STATIC,	cSizedCtrls::ptPDSR );

	//
	//	It will be hidden
	//
	if ( m_ptbKey != NULL )
		m_sizedCtrls.AddCtrl( IDC_btnGet,			cSizedCtrls::ptPDR );

	HWND hCmb = GetDlgItem( m_hDlg, IDC_COMBO );

	if ( ( hCmb != NULL ) && ( pgpSrvs != NULL ) )
	{
		for ( int i = 0; pgpSrvs[i]; i++ )
			ComboBox_InsertStringT( hCmb, i, _getWcs( pgpSrvs[i]->addr ) );

		ComboBox_SetCurSel( hCmb, 0 );
	}

	m_hList = GetDlgItem( m_hDlg, IDC_LISTVIEW );
	ListViewInit( m_hInst, GetDlgItem( m_hDlg, IDC_LISTVIEW ) );

	if ( m_keyID != NULL )
	{
		char keyIDText[ 100 ];

		makeKeyIDStr( keyIDText, m_keyID );

		SetDlgItemTextA( m_hDlg, IDC_EDIT, keyIDText );
		Edit_SetReadOnly( GetDlgItem( m_hDlg, IDC_EDIT ), TRUE );
	}

	m_searchThread = NULL;
	m_timerID = NULL;

	dlgVals.Restore( m_hDlg, m_hList );
	ComboBox_SetCurSel( hCmb, dlgVals.m_lookupID );

	return TRUE;
}

bool cDlgPKSearch::GetCurrentKey( cPlainTextBlock *ptbOut )
{
	int urlNum = ComboBox_GetCurSel( GetDlgItem( m_hDlg, IDC_COMBO ) );
	
	if ( pgpSrvs == NULL )
	{
		assert( false );
		return false;
	}

	bool found = false;

	//
	//	Walks all selected items, gets theirs key by index to ptb
	//

	for ( int hItem = -1; ( hItem = ListView_GetNextItem( m_hList, hItem, LVNI_SELECTED ) ) != -1 ; )
	{
		cPlainTextBlock *ptb = (cPlainTextBlock *)ListView_GetItemParam( m_hList, hItem );

		if ( ptb == NULL )
		{
			ptb = new cPlainTextBlock;

			wchar_t IDStr [ 50 ];
			wcscpy_s( IDStr, L"0x" );
			
			ListView_GetItemTextT( m_hList, hItem, 0, IDStr + 2, maxLen( IDStr ) - 2 );

			if ( PgpLookupGetInfo( pgpSrvs[ urlNum ], ptb, (char*)getMBStr( IDStr ), false, true, true, &m_cancelled ) == ppe_ok )
			{
				ListView_SetItemParam( m_hList, hItem, (DWORD)ptb );
			}
		}

		ptbOut->WriteBytes( ptb->buffer, ptb->bufferSize, true );
	}

	return ptbOut->bufferSize != 0;
}

DWORD WINAPI searchThreadFunc(
	  LPVOID lpParameter
	)
{
	cDlgPKSearch *pDlg = (cDlgPKSearch *)lpParameter;
	pDlg->Search();
	return 0;
}

void cDlgPKSearch::InitTimer()
{
	m_timeStart = GetTickCount();

	if ( m_timerID == NULL )
		m_timerID = SetTimer( m_hDlg, NULL, 1000, (TIMERPROC)SetTimeValue );
}

static void SetTimeBoxValue( HWND hDlg, DWORD totalSec )
{
	assert( GetDlgItem( hDlg, IDC_TimeBox ) != NULL );

	char str[ 20 ];
	int hour = totalSec / 3600;
	int min  = ( totalSec % 3600 ) / 60;
	int sec  = totalSec % 60;

	sprintf( str, "%02d:%02d:%02d", hour, min, sec );

	SetDlgItemTextA( hDlg, IDC_TimeBox, str );
}

void CALLBACK cDlgPKSearch::SetTimeValue(
  HWND hDlg,         // handle to window
  UINT uMsg,         // WM_TIMER message
  UINT_PTR idEvent,  // timer identifier
  DWORD dwTime       // current system time
)
{
	cDlgPKSearch *pThis = (cDlgPKSearch *)GetWindowLong(hDlg, DWL_USER);
	assert( pThis != NULL );

	DWORD totalSec = ( dwTime - pThis->m_timeStart ) / 1000;

	SetTimeBoxValue( hDlg, totalSec );
}

bool cDlgPKSearch::OnStartSearch()
{
//	assert( false );

	InitTimer();

	m_cancelled = false;
	DWORD id;
	m_searchThread = CreateThread( NULL, 0, searchThreadFunc, this, 0, &id );
	return m_searchThread != NULL;
}

bool cDlgPKSearch::EnableButtons( bool enable )
{
	SetTimeBoxValue( m_hDlg, 0 );
	ShowWindow( GetDlgItem( m_hDlg, IDC_TimeBox ), enable ? SW_HIDE : SW_SHOW );

	EnableWindow( GetDlgItem( m_hDlg, IDC_btnSearch ), enable );

	int selected = ListView_GetNextItem( m_hList, -1, LVNI_SELECTED );
	enable &= ( selected != -1 );
	EnableWindow( GetDlgItem( m_hDlg, IDC_btnGet ), enable );
	EnableWindow( GetDlgItem( m_hDlg, IDC_btnSave ), enable );
	EnableWindow( GetDlgItem( m_hDlg, IDC_btnProperties ), enable );

	EnableWindow( GetDlgItem( m_hDlg, IDCANCEL ), true );

	return true;
}

bool cDlgPKSearch::Search()
{
	int urlNum = ComboBox_GetCurSel( GetDlgItem( m_hDlg, IDC_COMBO ) );

	HWND btnSearch = GetDlgItem( m_hDlg, IDC_btnSearch );
//	char btnText[ 20 ];
//	GetWindowText( btnSearch, btnText, sizeof( btnText ) );
//	SetWindowText( btnSearch, "&Stop" );

	EnableButtons( false );

	char text[ 1024 ];
	cPlainTextBlock ptb;
	GetWindowTextA( GetDlgItem( m_hDlg, IDC_EDIT ), text, sizeof(text) );
	int res = SearchPKUserIDs( m_hInst, m_hDlg, &ptb, urlNum, text, &m_cancelled );

	if ( ( res != ppe_ok ) && ( res != nce_cancelled ) )
	{
		wchar_t message[ 256 ];
		if ( res == nce_invalidConnection )
			LoadStringT( hMainInst, IDS_InvalidConnection , message, maxLen( message ) );
		else
			FormatT( message, maxLen( message), _LoadIDS( IDS_ErrorNumber_X ), res );

		MessageBoxT( m_hDlg, message, _LoadIDS( IDS_ErrorInSearchigPK ), MB_ICONSTOP );
	}
	else
		ParseDataAndFillListView( m_hList, &ptb );

//	SetWindowText( btnSearch, btnText );
	
	m_searchThread = NULL;

	EnableButtons( true );

	return res == ppe_ok;
}

LRESULT cDlgPKSearch::RealDlgProc( UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	LRESULT result;
	
	if ( cMenuItem::DlgProc( &result, m_hDlg, uMsg, wParam, lParam ) )
		return result;

	switch(uMsg)
	{
		case WM_INITDIALOG:
			if ( !InitDialog() )
			{
				EndDialog( m_hDlg, IDCANCEL );
			}
			break;

		case WM_COMMAND:
		{
			if ( wParam == aNumber ) // VK_RETURN
			{
				if ( m_searchThread != NULL )
					break;

				HWND hFocus = GetFocus();
				if ( hFocus == GetDlgItem( m_hDlg, IDC_EDIT ) )
				{
					OnStartSearch();
				}
				else
				if ( hFocus == GetDlgItem( m_hDlg, IDC_LISTVIEW ) )
				{
					int selected = ListView_GetNextItem( m_hList, -1, LVNI_SELECTED );
					if ( selected != -1 )
					{
						OnProperties();
					}
				}

				break;
			}

			if ( HIWORD( wParam ) == 0 )
			{
				switch ( LOWORD( wParam ) )
				{
					case IDC_btnSearch:
					{
						if ( m_searchThread == NULL )
							OnStartSearch();
						else
							m_cancelled = true;

						break;
					}
					case IDC_btnGet:
					{
						assert( m_ptbKey != NULL );
						if ( m_ptbKey == NULL )
							break;

						EnableWindow( GetDlgItem( m_hDlg, IDC_btnGet ), false );

						cPlainTextBlock ptb;

						if ( m_keyID != NULL ) // If key id specified and Get button pressed we found directly the key.
						{
							if ( GetCurrentKey( m_ptbKey ) )
							{
								EndDialog( m_hDlg, IDOK );
							}
						}
						else
						{
							if ( GetCurrentKey( &ptb ) )
							{
								if ( ChooseKeyOrSubkeyT( m_hInst, m_hDlg, ptb.buffer, ptb.bufferSize, m_ptbKey, m_wcsUserID ) )
									EndDialog( m_hDlg, IDOK );
							}
						}
						
						EnableWindow( GetDlgItem( m_hDlg, IDC_btnGet ), true );

						break;
					}
					case IDC_btnSave:
					{
						cPlainTextBlock ptb;
						EnableWindow( GetDlgItem( m_hDlg, IDC_btnSave ), false );
						if ( GetCurrentKey( &ptb ) )
						{
							int errIDS = BCDB_AddPGPKeyToLocalKeyRing( m_hDlg, ptb.buffer, ptb.bufferSize, false );
							if ( errIDS != 0 )
							{
								DlgMessageBoxT( errIDS, MB_ICONSTOP, _LoadIDS( IDS_szSavePublicKeyError ) );
							}
							else
							if ( m_ptbLastSavedKey != NULL )
							{
								assert( m_ptbLastSavedKey->checkSig() );
								m_ptbLastSavedKey->WriteBytes( ptb.buffer, ptb.bufferSize );
							}
						}

						EnableWindow( GetDlgItem( m_hDlg, IDC_btnSave ), true );
						break;
					}
					case IDC_btnProperties:
						OnProperties();
						break;

					case IDOK:
						EndDialog( m_hDlg, LOWORD( wParam ) );
						return true;

					case IDCANCEL:

						if ( m_searchThread == NULL )
							EndDialog( m_hDlg, LOWORD( wParam ) );
						else
						{
							m_cancelled = true;
							EnableWindow( GetDlgItem( m_hDlg, IDCANCEL ), false );
						}

						return true;
					default:
						return false;
				}
			}

			break;
		}
//		case WM_SIZE:
//			sizedCtrls.OnSize( LOWORD( lParam ), HIWORD( lParam ) );
//			break;
		case WM_PAINT:
		{
//			OnPaint();
			return 0;
		}
		case WM_NOTIFY:
			return OnNotify( (NMHDR *)lParam );
		case WM_DESTROY:
			DeleteDialog();
			break;
		default:
			return false;
	}

	return true;
}

void cDlgPKSearch::DeleteDialog()
{
	HWND hList = GetDlgItem( m_hDlg, IDC_LISTVIEW );
	LV_DeleteAllItems( hList );

	if ( m_timerID )
		KillTimer( m_hDlg, m_timerID );

	dlgVals.Save( m_hDlg, m_hList );
}

bool cDlgPKSearch::OnProperties()
{
	if ( m_searchThread != NULL )
		return false;

	if ( pgpSrvs == NULL )
	{
		assert( false );
		return false;
	}

	for ( int hItem = -1; ( hItem = ListView_GetNextItem( m_hList, hItem, LVNI_SELECTED ) ) != -1 ; )
	{
		cPlainTextBlock *ptb = (cPlainTextBlock *)ListView_GetItemParam( m_hList, hItem );

		if ( ptb == NULL )
		{
			ptb = new cPlainTextBlock;

			wchar_t IDStr [ 50 ];
			wcscpy_s( IDStr, L"0x" );
			
			ListView_GetItemTextT( m_hList, hItem, 0, IDStr + 2, maxLen( IDStr ) - 2 );

			int urlNum = ComboBox_GetCurSel( GetDlgItem( m_hDlg, IDC_COMBO ) );
			if ( PgpLookupGetInfo( pgpSrvs[ urlNum ], ptb, (char*)getMBStr(IDStr), false, true, true ) != ppe_ok )
			{
				delete ptb;
				return false;
			}

			ListView_SetItemParam( m_hList, hItem, (DWORD)ptb );
		}

		int res = ppe_ok;

//		int parsed = 0;
//		cPlainTextBlock ptbKey;
//		res = DecodePGPPublicKeyBlock( ptb->buffer, ptb->bufferSize, &ptbKey, &parsed );

		BYTE *userID;
		int userIDSize;
		BYTE keyID[8];
		BYTE issuerKeyID[8];

		if ( res == ppe_ok )
			res = ExtractDataAndVerifyPGPPublicKeyBlock( ptb->buffer, ptb->bufferSize,0,0,0, &userID, &userIDSize, keyID, issuerKeyID, true ); // true means do not verify

		if ( res == ppe_ok )
			ShowKeyPropertiesEx( m_hInst, m_hDlg, keyID, ptb->buffer, ptb->bufferSize, false );
	}

	return true;
}

bool cDlgPKSearch::OnNotify( NMHDR *hdr )
{
	switch( hdr->code )
	{
		case LVN_ITEMCHANGED: //LVN_ITEMACTIVATE: // LVN_ITEMCHANGING:
		{
			int selected = ListView_GetNextItem( m_hList, -1, LVNI_SELECTED );
			bool enable = selected != -1;
			EnableWindow( GetDlgItem( m_hDlg, IDC_btnGet ), enable );
			EnableWindow( GetDlgItem( m_hDlg, IDC_btnSave ), enable );
			EnableWindow( GetDlgItem( m_hDlg, IDC_btnProperties ), enable );
			break;
		}
		case NM_DBLCLK:
			OnProperties();
			break;
		default:
			break;
	}

	return false;
}


//
//	SearchPKUser
//		1. Opens window with search options
//		2. Fills view window by search results.
//		3. Saves a selected key to local key ring.
//		4. Fills ptbKey block by a selected key.
//		
//	Returns ppe_ok if ptbKey is successfully filled, or ppe_keyNotFound. 
//	Otherwise returns a error code.
//
//	See also :
//		DecodePGPPublicKeyBlock
//		ExtractDataAndVerifyPGPPublicKeyBlock
//
int SearchPKUserT(
		HINSTANCE hInst,
		HWND hWnd,
		cPlainTextBlock *ptbKey,
		BYTE *keyID,// = NULL, If key id specified the key ID field is filled and then locked.
		cWCStr *wcsUserID,
		cPlainTextBlock *ptbLastSavedKey // If user saves key to BCDB it is copied to the ptb.
	)
{
	cWSABlock wsa;

	if ( wsa.getError() != 0 )
	{
		MessageBoxT( hWnd, L"Error in loading Ws2_32.dll", _LoadIDS( IDS_AppName ), MB_ICONSTOP );
		return pkae_cancelled;
	}

	if ( hInst == NULL )
		hInst = hMainInst; // GetModuleHandle( "BCKeyMan.dll" );
	 
	cDlgPKSearch dlg( ptbKey, keyID, wcsUserID, ptbLastSavedKey );

	int res = dlg.DoModal( hInst, dlg.ID, hWnd );

	if ( res == IDCANCEL )
		return pkae_cancelled;
	else
	if ( ( res == IDOK ) && ptbKey->bufferSize )
		return ppe_ok;
	else
		return ppe_keyNotFound;
}
