#include "BCKeyMan.h"
#include "PublicKeyProps.h"
#include "BCDB_Functions.h"
#include "cDlgBrowseKeyFile.h"
#include "PKParse.h"
#include "resource.h"

#include "..\CertificateLib\PGPPkts\PGPPkts.h"
#include "..\CertificateLib\PGPPkts\Packets.h"
#include "..\CertificateLib\x509Pkts\base64funcs.h"

#include "..\Includes\FileView.h"
#include "..\Includes\cDlg.h"
#include "..\Includes\OpenFiles.h"
#include "..\Includes\UnicodeFuncs.h"
#include "..\Includes\menuItem.h"
#include "..\Includes\TranslateDll.h"

#include <stdio.h>
#include <io.h>
#include <commctrl.h>

int IsEncryptionKey( sPublicKey *pk, int pkSize )
{
	BYTE algID = -1;

	int res = PKT_PublicKey_Parse( pk, pkSize, NULL, &algID );
	if ( res != ppe_ok ) return res;

	if ( isPKEncryptionAlgorithm( algID ) )
		return ppe_ok;

	return ppe_invalidAlgorithmID;
}

static int LV_SetKeyID( HWND hList, int item, BYTE *keyID )
{
	BYTE *keyIDCopy = ( BYTE * )ListView_GetItemParam( hList, item );

	if ( keyIDCopy == NULL )
		keyIDCopy = new BYTE[ pgp_keyIDSize ];

	if ( keyIDCopy == NULL )
		return 0;

	memcpy( keyIDCopy, keyID, pgp_keyIDSize );

	ListView_SetItemParam( hList, item, (DWORD)keyIDCopy );

	return pgp_keyIDSize;
}

static int lvGetSelectedItem( HWND hLV )
{
	if ( ListView_GetSelectedCount( hLV ) != 1 )
		return -1;
	else
	{
		int item = ListView_GetNextItem( hLV, -1, LVNI_SELECTED );
		return item;
	}
}

static BYTE *LV_GetKeyID(
		HWND hList,
		int item = -1 // -1 means selected item
	)
{
	if ( item == -1 )
	{
		item = lvGetSelectedItem( hList );

		if ( item == -1 )
			return NULL;
	}

	return ( BYTE * )ListView_GetItemParam( hList, item );
}

static void LV_FreeKeyID( HWND hList, int item )
{
	BYTE *keyIDCopy = ( BYTE * )ListView_GetItemParam( hList, item );

	if ( keyIDCopy != NULL )
		delete keyIDCopy;
}

static void LV_DeleteAllItems( HWND hList )
{
	for ( int i = ListView_GetNextItem( hList, -1, LVNI_ALL ); i != -1; i = ListView_GetNextItem( hList, i, LVNI_ALL ) )
	{
		LV_FreeKeyID( hList, i );
	}

	ListView_DeleteAllItems( hList );
}

class cDlgBrowsePKUser : public cColoredDlg
{
	cFileView *m_kr;
	cPlainTextBlock m_ptbDecoded; // Ascii armored key is decoded to the ptb.
	bool m_secret;
	//cPlainTextBlock *m_ptbUserID;
	cWCStr *m_wcsUserID;
	cPlainTextBlock m_keyIDtbl;

	cSizedCtrls sizedCtrls;
	HWND m_hToolBar;

 public:

	enum { IDD = IDD_DlgBrowseForPGPUser };

	cDlgBrowsePKUser(
		cPlainTextBlock *ptbKey,  // [OUT] - pure key packet in sPublicKey or sSecretKey format as declared in Packets.h
		const wchar_t *keyRingPath,
		bool secret = false,
		//cPlainTextBlock *ptbUserID = NULL,
		cWCStr *wcsUserID = NULL,
		cPlainTextBlock *ptbKeyBlock = NULL
				// Entire key packet - master key with subpackets including subkeys packets
	);

	BYTE *m_keyPkt;
	int m_keyPktSize;
	bool m_checkIfKeyMayBeUsedForEncryption;

	cDlgBrowsePKUser(
		BYTE *keyPkt, int keyPktSize,
		cPlainTextBlock *ptbKey,
		//cPlainTextBlock *ptbUserID = NULL,
		cWCStr *wcsUserID = NULL,
		bool checkIfKeyMayBeUsedForEncryption = true
	);

	~cDlgBrowsePKUser();

	wchar_t m_krPath[ MAX_PATH ];

 private:	
	cPlainTextBlock *m_ptbKey, *m_ptbKeyBlock;
	
	HIMAGELIST m_himl;
	RECT m_rect;
	HWND m_hListView;

	void expandItem( int item );
	BOOL checkHitPlus( POINT pt );
	int SetItemValues( BYTE *krPkt, int krSize, int item, BYTE *keyID );
	int FillListView( const wchar_t *krPath );
	int FillListView( BYTE *keyPkt, int keyPktSize );
	int BrowseForKeyRing();
	bool OnCommand( WPARAM wParam, LPARAM lParam );
	bool OnOK( int idCmd );
	bool OnDelete();

	//
	//	Initialization and destruction functions.
	//
	void InitDialog();
	void DeleteDialog();
	HWND CreateToolBar();

	//	
	//	List view functions...
	//
	BOOL lvSelChanged();
	void UpdateButtons();
	BOOL lvNotify( NMHDR *hdr  );
	BOOL tbNotify( NMHDR *pnmh );
	BOOL ShowCurrentItemProperties();
	bool dblClick();

	//
	//	Virtual dlgProc...
	//
	LRESULT RealDlgProc( UINT uMsg, WPARAM wParam, LPARAM lParam );
public:
	int DeleteKey( cpgp_keyInfo * pKeyInfo, const char *text );
};

cDlgBrowsePKUser::cDlgBrowsePKUser(
	BYTE *keyPkt, int keyPktSize,
	cPlainTextBlock *ptbKey,
	//cPlainTextBlock *ptbUserID,
	cWCStr *wcsUserID,
	bool checkIfKeyMayBeUsedForEncryption
)
{
	m_secret = false;
	m_wcsUserID = wcsUserID;

	INITCOMMONCONTROLSEX commCtrl;

	commCtrl.dwSize = sizeof( INITCOMMONCONTROLSEX );
	commCtrl.dwICC = 
		ICC_BAR_CLASSES |
		ICC_LISTVIEW_CLASSES |
		ICC_PROGRESS_CLASS |
		ICC_TAB_CLASSES |
		ICC_TREEVIEW_CLASSES;

	InitCommonControlsEx( &commCtrl );

	m_hListView = NULL;

	m_ptbKey = ptbKey;
	m_ptbKeyBlock = NULL;

	m_krPath[0] = L'\0';
	m_kr = NULL;
	m_keyPkt = keyPkt;
	m_keyPktSize = keyPktSize;
	m_checkIfKeyMayBeUsedForEncryption = checkIfKeyMayBeUsedForEncryption;
}

cDlgBrowsePKUser::cDlgBrowsePKUser(
		cPlainTextBlock *ptbKey,
		const wchar_t *keyRingPath,
		bool secret,
		//cPlainTextBlock *ptbUserID,
		cWCStr *wcsUserID,
		cPlainTextBlock *ptbKeyBlock
				// Entire key packet - master key with subpackets including subkeys packets
	)
{
	m_secret = secret;
	m_wcsUserID = wcsUserID;

	INITCOMMONCONTROLSEX commCtrl;

	commCtrl.dwSize = sizeof( INITCOMMONCONTROLSEX );
	commCtrl.dwICC = 
		ICC_BAR_CLASSES |
		ICC_LISTVIEW_CLASSES |
		ICC_PROGRESS_CLASS |
		ICC_TAB_CLASSES |
		ICC_TREEVIEW_CLASSES;

	InitCommonControlsEx( &commCtrl );

	m_hListView = NULL;
	m_ptbKey = ptbKey;
	m_ptbKeyBlock = ptbKeyBlock;

	if ( keyRingPath )
		wcscpy_s( m_krPath, keyRingPath );
	else
		m_krPath[0] = L'\0';

	m_kr = NULL;
	m_keyPkt = NULL;
	m_keyPktSize = 0;

	m_checkIfKeyMayBeUsedForEncryption = false;
}

cDlgBrowsePKUser::~cDlgBrowsePKUser()
{
	if ( m_kr )
		delete m_kr;

	OleUninitialize( );
}

int cDlgBrowsePKUser::BrowseForKeyRing()
{
	wchar_t szFiltersPublic[ 1024 ];
	LoadStringT( m_hInst, IDS_szFiltersPublicPGP, szFiltersPublic, maxLen( szFiltersPublic ) );		

	wchar_t path[ MAX_PATH ];
	if ( !BrowseFiles( m_hDlg, path, maxLen( path ), szFiltersPublic, _LoadIDS( IDS_OpenKRFile ), false ) )
		return ppe_ok;

	return FillListView( path );
}

static UINT ColumnLabelIDS[] = 
{
	IDS_szName, IDS_szAlgorithm, IDS_szKeySize, IDS_szKeyID
};

const int NUM_COLUMNS = sizeof( ColumnLabelIDS ) / sizeof( UINT );

static int _gnColumnFmt[ NUM_COLUMNS ]=
{
	LVCFMT_LEFT, LVCFMT_CENTER, LVCFMT_CENTER, LVCFMT_LEFT
};

static int _gnColumnWidth[NUM_COLUMNS]=
{
	270, 70, 70, 120
};

BOOL InitListViewColumns( HINSTANCE hInst, HWND hLView )
{
	int i;
	LV_COLUMNW lvc;

	memset( &lvc, 0, sizeof( lvc ) );

	lvc.mask = LVCF_WIDTH;

	//
	//	Check if already initialised...
	//
	if ( ListView_GetColumnT( hLView, 0, &lvc ) )
		return TRUE;

	lvc.mask = LVCF_FMT | LVCF_WIDTH | LVCF_TEXT | LVCF_SUBITEM;// | LVCF_IMAGE;

	wchar_t text[ 80 ];

	DWORD mask = LVS_EX_FULLROWSELECT; //  | LVS_EX_CHECKBOXES ; // | LVS_EX_SUBITEMIMAGES;

	ListView_SetExtendedListViewStyleEx( hLView, mask, mask );

	int res;

	for(i=0; i<NUM_COLUMNS; i++)
	{
		LoadStringT( hInst, ColumnLabelIDS[i], text, maxLen( text ) );

		lvc.iSubItem=i;
		lvc.pszText = text; // _gszColumnLabel[i];
		lvc.cx = _gnColumnWidth[i];
		lvc.fmt=_gnColumnFmt[i];
//		lvc.iImage = 0;

		res = ListView_InsertColumnT( hLView, i, &lvc );

		if ( res == -1 )
		{
			MessageBoxT( 0, text, _LoadIDS( IDS_ErrorInAddLVColumn ), MB_OK );
			return FALSE;
		}
	}

	return TRUE;
}

static int setItemValues( HWND hList, int item, cpgp_keyInfo *keyInfo, int iImage )
{
	LVITEMW lvi;
	memset( &lvi, 0, sizeof( lvi ) );
	lvi.iItem = item;
	lvi.mask = LVIF_IMAGE;
	lvi.iImage = iImage; 
	ListView_SetItem( hList, &lvi );

	int subItem = 0;

	if ( keyInfo->userID != NULL )
	{
		//const char szEnd[1] = "";
		//cPlainTextBlock ptbUserID;

		//if ( !ptbUserID.WriteBytes( keyInfo->userID->body, keyInfo->userID->size ) || 
		//	 !ptbUserID.WriteBytes( (BYTE*)szEnd, 1 ) )

		cWCStr wcsUserID( (char*)keyInfo->userID->body, keyInfo->userID->size, false, CP_UTF8 );

		if ( wcsUserID.isEmpty() )
		{
			MessageBoxT(
					hList,
					_LoadIDS( IDS_szMemoryAllocationError ),
					_LoadIDS( IDS_BrowsePKUser ),
					MB_ICONSTOP );

			return ppe_memoryAllocation;
		}

		//ListView_SetItemText( hList, item, subItem++, (char *)ptbUserID.buffer );
		ListView_SetItemTextT( hList, item, subItem++, wcsUserID.m_szBody );
	}
	else
	{
		if ( iImage == 3 ) // Subkey
		{
			wchar_t subkey[100] = L"";
			LoadStringT( hMainInst, IDS_Subkey, subkey, maxLen( subkey ) );
			ListView_SetItemTextT( hList, item, subItem, subkey );
		}

		subItem++;
	}

	BYTE version = keyInfo->keyPkt[0];

	BYTE algID = -1;
	DWORD bitSize = -1;

	switch ( version )
	{
		case 2:
		case 3:
		{
			sPublicKeyV3 *pk = (sPublicKeyV3 *)keyInfo->keyPkt;
			algID = pk->algID;
			bitSize = ((DWORD)pk->mpi[0] << 8 ) | pk->mpi[1];

			break;
		}
		case 4:
		{
			sPublicKey *pk = (sPublicKey *)keyInfo->keyPkt;
			algID = pk->algID;
			bitSize = ((DWORD)pk->mpi[0] << 8 ) | pk->mpi[1];

			break;
		}
		default:
			return ppe_invalidPacketVersion;
	}

	ListView_SetItemTextT( hList, item, subItem++, getWCStr( PKT_GetPublicKeyAlgName( algID ) ) );

	wchar_t size[ 80 ];

	if ( bitSize == -1 )
		wcscpy_s( size, L"X" );
	else
		FormatT( size, maxLen( size ), L"%d", bitSize ); 

	ListView_SetItemTextT( hList, item, subItem++, size );

	LV_SetKeyID( hList, item, keyInfo->keyID );

	wchar_t idStr[ 256 ];
	BCDB_bytes2string( idStr, maxLen( idStr ), keyInfo->keyID, sizeof( keyInfo->keyID ) );

	ListView_SetItemTextT( hList, item, subItem++, idStr );

	return ppe_ok;
}

int cDlgBrowsePKUser::SetItemValues( BYTE *krPkt, int krSize, int item, BYTE *keyID )
{
	//
	//	1. Gets key info
	//

	cpgp_keyInfo keyInfo;

	int res = pgp_FindKeyByIDEx(
				&keyInfo,
				keyID,
				krPkt, krSize,
				m_secret );

	if ( res != ppe_ok )
	{
		MessageBoxT( m_hDlg, _LoadIDS( IDS_szErrorInKeyReading ), _LoadIDS( IDS_BrowsePKUser ), MB_ICONSTOP );
		return res;
	}

	setItemValues( m_hListView, item, &keyInfo, ( keyInfo.next ? 1 : 0 ) ); // If has subkeys ? collapsed : empty 

	return ppe_ok;
}

int cDlgBrowsePKUser::FillListView( BYTE *keyPkt, int keyPktSize )
{
	if ( keyPktSize != 0 )
	{
		int size = 0;

		int res = pgp_EnumKeyIDs( NULL, &size, keyPkt, keyPktSize, m_secret );

		if ( res == ppe_bufferTooSmall )
		{
			if ( m_keyIDtbl.Allocate( size ) == NULL )
				return ppe_memoryAllocation;

			res = pgp_EnumKeyIDs( m_keyIDtbl.buffer, &size, keyPkt, keyPktSize, m_secret );
		}

		if ( res != ppe_ok )
		{
			MessageBoxT( m_hDlg, _LoadIDS( IDS_KRIsInvalid ), _LoadIDS( IDS_BrowsePKUser ), MB_ICONSTOP );
			return res;
		}

		LVITEMW lvi;
		memset( &lvi, 0, sizeof( lvi ) );
		lvi.mask = LVIF_IMAGE;
		lvi.iImage = 1;

		int pos;
		for( pos = 0; pos < m_keyIDtbl.bufferSize; pos += pgp_keyIDSize )
		{
			ListView_InsertItemT( m_hListView, &lvi );
		}

		int item = 0;
		for( pos = 0; pos < m_keyIDtbl.bufferSize; pos += pgp_keyIDSize )
		{
			SetItemValues( keyPkt, keyPktSize, item, m_keyIDtbl.buffer + pos );
			item++;
		}
	}

	UpdateButtons();

	return ppe_ok;
}

int cDlgBrowsePKUser::FillListView( const wchar_t *keyRingPath )
{
	LV_DeleteAllItems( m_hListView );

	if ( m_kr )
		delete m_kr;

	m_kr = new cFileView ( keyRingPath );

	if ( m_kr->m_err && ( FileExistsT( keyRingPath ) ) )
	{
		SetDlgItemTextT( IDC_edtKeyRing, IDS_Undefined );
		return -1;
	}
	
	type_kff kff = getKeyFileFormat( keyRingPath );

	if ( kff == kffPGPAscii )
	{
		int res = DecodePGPKeyBlock( m_kr->m_ptr, (int)m_kr->m_size.QuadPart, &m_ptbDecoded, NULL, NULL, m_secret );

		if ( ( res == ppe_invalidBlock ) && !m_secret )
			res = DecodePGPKeyBlock( m_kr->m_ptr, (int)m_kr->m_size.QuadPart, &m_ptbDecoded, NULL, NULL, true );

		if ( res != ppe_ok )
			return res;

		m_keyPkt = m_ptbDecoded.buffer;
		m_keyPktSize = m_ptbDecoded.bufferSize;
	}
	else
	{
		m_keyPkt = m_kr->m_ptr;
		m_keyPktSize = (int)m_kr->m_size.QuadPart;
	}


	int res = FillListView( m_keyPkt, m_keyPktSize );

	if ( res != ppe_ok )
		return res;

	if ( m_krPath != keyRingPath )
		wcscpy_s( m_krPath, keyRingPath );

	SetDlgItemTextT( IDC_edtKeyRing, m_krPath );

	return res;
}

enum { IDM_Browse = 200, IDM_Delete };

static DWORD GetIDSbyCmd( int cmd )
{
	switch ( cmd )
	{
		case IDM_Browse : return IDS_szBrowse;
		case IDM_Delete : return IDS_szDelete;
		default : return NULL;
	}
}

HWND cDlgBrowsePKUser::CreateToolBar()
{
	INITCOMMONCONTROLSEX commCtrl;

	commCtrl.dwSize = sizeof( INITCOMMONCONTROLSEX );
	commCtrl.dwICC = 
		ICC_BAR_CLASSES |
		ICC_LISTVIEW_CLASSES |
		ICC_PROGRESS_CLASS |
		ICC_TAB_CLASSES |
		ICC_TREEVIEW_CLASSES;

	InitCommonControlsEx( &commCtrl );

	HWND hPath = GetDlgItem( m_hDlg, IDC_edtKeyRing );

	RECT r;
	GetWindowRect( hPath, &r );
	POINT ptTBPos = { r.right + 2, r.top };
	ScreenToClient( m_hDlg, &ptTBPos );

	m_hToolBar = CreateWindowExT( 
			WS_EX_TRANSPARENT |
			0 ,
			TOOLBARCLASSNAMEW, L"",
			CCS_NOPARENTALIGN |
			CCS_ADJUSTABLE |
			CCS_NOMOVEY |
			CCS_NODIVIDER |
			CCS_TOP |
			CCS_NORESIZE |
			TBSTYLE_FLAT |
			TBSTYLE_TOOLTIPS |
			TBSTYLE_TRANSPARENT |
			TBSTYLE_AUTOSIZE |
			WS_CHILD | WS_VISIBLE ,
			ptTBPos.x, ptTBPos.y, 150, r.bottom - r.top + 1,
			m_hDlg, NULL, m_hInst, 0 );


	if ( m_hToolBar == NULL )
		return NULL;

	SendMessage( m_hToolBar, TB_SETEXTENDEDSTYLE, 0 , TBSTYLE_EX_DRAWDDARROWS | WS_EX_TRANSPARENT );
	SendMessage( m_hToolBar, TB_BUTTONSTRUCTSIZE, (WPARAM) sizeof(TBBUTTON), 0);
	Toolbar_SetUnicodeFormat( m_hToolBar );

	// Add the bitmap containing button images to the toolbar. 

	TBADDBITMAP tbab;
	memset( &tbab, 0, sizeof( tbab ) );

	tbab.hInst = HINST_COMMCTRL ; 
	tbab.nID   = IDB_STD_SMALL_COLOR; 
	int idbStd = SendMessage( m_hToolBar, TB_ADDBITMAP, (WPARAM) 0, (WPARAM) &tbab); 

	tbab.hInst = HINST_COMMCTRL ; 
	tbab.nID   = IDB_VIEW_SMALL_COLOR; 
	int idbView = SendMessage( m_hToolBar, TB_ADDBITMAP, (WPARAM) 0, (WPARAM) &tbab); 

	tbab.hInst = HINST_COMMCTRL ; 
	tbab.nID   = IDB_HIST_SMALL_COLOR; 
	int idbHist = SendMessage( m_hToolBar, TB_ADDBITMAP, (WPARAM) 0, (WPARAM) &tbab); 

	tbab.hInst = m_hInst; 
	tbab.nID   = IDB_ToolbarXP; 
	int idbInternal = SendMessage( m_hToolBar, TB_ADDBITMAP, (WPARAM) 7, (WPARAM) &tbab); // 3

	ToolBar_AddButton( m_hToolBar, idbInternal + 6  , IDM_Browse );
	ToolBar_AddButton( m_hToolBar, 0, 0, TBSTYLE_SEP );
	ToolBar_AddButton( m_hToolBar,idbStd + STD_DELETE      , IDM_Delete );

	return m_hToolBar;
}

void cDlgBrowsePKUser::InitDialog()
{
	SetRCValues(IDC_Line, IDC_STATIC_BOLD, 0, IDI_Main );

	SetDlgItemTextT( IDC_STATIC_BOLD, m_secret ? IDS_PGP_SKR : IDS_PGP_PKR );

	m_sizedCtrls.AddCtrl( IDOK, cSizedCtrls::ptPDR );
	m_sizedCtrls.AddCtrl( IDCANCEL, cSizedCtrls::ptPDR );
	m_sizedCtrls.AddCtrl( IDC_btnProperties, cSizedCtrls::ptPDR );

	m_sizedCtrls.AddCtrl( IDC_bmpHeader, cSizedCtrls::ptPR );
	m_sizedCtrls.AddCtrl( IDC_Line, cSizedCtrls::ptSR );
	m_sizedCtrls.AddCtrl( IDC_edtKeyRing, cSizedCtrls::ptSR );

	if ( GetItem( IDC_btnBrowse ) != NULL )
		m_sizedCtrls.AddCtrl( IDC_btnBrowse, cSizedCtrls::ptPR );

	m_sizedCtrls.AddCtrl( IDC_LISTVIEW, cSizedCtrls::ptSDR );
	m_sizedCtrls.AddCtrl( IDC_Line2, cSizedCtrls::ptPDSR );
	m_sizedCtrls.AddCtrl( IDC_Notes, cSizedCtrls::ptSR );

	if ( ( m_ptbKey == NULL )
		&& ( m_ptbKeyBlock == NULL ) ) //  the function is used for viewing and deleting a key
	{
//		SetDlgItemText( m_hDlg, IDOK, "&Delete" );
		SetDlgItemTextT( IDC_Notes, _LoadIDS( IDS_STRING183 ) );
	}
	else
		SetDlgItemTextT( IDC_Notes, L"" );

	CreateToolBar();
	if ( m_hToolBar != NULL )
		m_sizedCtrls.AddCtrl( m_hToolBar, cSizedCtrls::ptPR );


	SendMessage( m_hDlg, WM_SETICON, 0, (LPARAM)LoadIcon( m_hInst, MAKEINTRESOURCE( IDI_Main ) ) );

	if ( m_secret )
		SetWindowTextT( m_hDlg, _LoadIDS( IDS_ChooseSecretKey ) );

	//	
	//	Initialize views...
	//

	m_hListView = GetDlgItem( m_hDlg, IDC_LISTVIEW );

	InitListViewColumns( m_hInst, m_hListView );

	int cx = GetSystemMetrics(SM_CXSMICON) * 2;
	int cy = GetSystemMetrics(SM_CYSMICON);

	m_himl = ImageList_Create( cx, cy, ILC_MASK | ILC_COLOR16 , 4, 10 ); //
	ImageList_AddIcon( m_himl, (HICON)LoadImage( m_hInst, MAKEINTRESOURCE( IDI_PKUserEmpty ), IMAGE_ICON, cx, cy, LR_DEFAULTCOLOR ) );
	ImageList_AddIcon( m_himl, (HICON)LoadImage( m_hInst, MAKEINTRESOURCE( IDI_PkUserCollapsed ), IMAGE_ICON, cx, cy, LR_DEFAULTCOLOR ) );
	ImageList_AddIcon( m_himl, (HICON)LoadImage( m_hInst, MAKEINTRESOURCE( IDI_PKUserExpanded ), IMAGE_ICON, cx, cy, LR_DEFAULTCOLOR ) );
	ImageList_AddIcon( m_himl, (HICON)LoadImage( m_hInst, MAKEINTRESOURCE( IDI_SubKey ), IMAGE_ICON, cx, cy, LR_DEFAULTCOLOR ) );

	ListView_SetImageList( m_hListView, m_himl, LVSIL_SMALL );

	if ( wcslen( m_krPath ) == 0 )
	{
		if ( m_keyPktSize != 0 )
		{
			FillListView( m_keyPkt, m_keyPktSize );
			SetDlgItemTextT( IDC_edtKeyRing, _LoadIDS( IDS_DownloadedKeyPkt ) );
		}
		else
			BrowseForKeyRing();
	}
	else
		FillListView( m_krPath );
}

BOOL cDlgBrowsePKUser::lvSelChanged()
{
	UpdateButtons();

	return true;
}

static void addSubKey( HWND hList, int item, cpgp_keyInfo *keyInfo )
{
	LVITEMW lvi;
	memset( &lvi, 0, sizeof( lvi ) );
	lvi.iItem = item;

	item = ListView_InsertItemT( hList, &lvi );

	if ( item != -1 )
	{
		setItemValues( hList, item, keyInfo, 3 );
	}
}

void cDlgBrowsePKUser::expandItem( int item )
{
	LVITEMW lvi;
	memset( &lvi, 0, sizeof( lvi ) );
	lvi.iItem = item;
	lvi.mask = LVIF_IMAGE;
	lvi.iImage = 2; // Expanded
	ListView_SetItem( m_hListView, &lvi );

	cpgp_keyInfo keyInfo;

//	char text[ 1024 ];
//	ListView_GetItemText( hList, item, 0, text, sizeof( text ) );

	BYTE *keyID = LV_GetKeyID( m_hListView, item );
	if ( keyID == NULL )
	{
		assert( false );
		return;
	}

//	int res = pgp_FindKeyByUserID( &keyInfo, text, m_kr->m_ptr, m_kr->m_size, m_secret );
	int res = pgp_FindKeyByIDEx(
				&keyInfo,
				keyID,
				m_keyPkt, m_keyPktSize,
				m_secret );

	if ( res != ppe_ok )
		return;

	for ( cpgp_keyInfo *subKey = keyInfo.next; subKey; subKey = subKey->next )
	{
		addSubKey( m_hListView, item + 1, subKey );
	}
}

static int deleteSubKeys( HWND hList, int item )
{
	LVITEMW lvi;
	memset( &lvi, 0, sizeof( lvi ) );
	lvi.mask = LVIF_IMAGE;

	int count = 0;

	item++; // gets next subkey;

	while ( item < ListView_GetItemCount( hList ) ) 
	{
		lvi.iItem = item;
		ListView_GetItem( hList, &lvi );
		if ( lvi.iImage != 3 ) // 3 - Sub key image.
			break;

		count++;

		LV_FreeKeyID( hList, item );
		ListView_DeleteItem( hList, item );
	}

	return count;
}

static void collapseItem( HWND hList, int item )
{
	LVITEMW lvi;
	memset( &lvi, 0, sizeof( lvi ) );
	lvi.iItem = item;
	lvi.mask = LVIF_IMAGE;
	lvi.iImage = 1; // Collapsed
	ListView_SetItem( hList, &lvi );

	deleteSubKeys( hList, item );
}

BOOL cDlgBrowsePKUser::checkHitPlus( POINT pt )
{
	if ( pt.x > 32 )
		return FALSE;

	LVHITTESTINFO hti;

	memset( &hti, 0, sizeof( hti ) );

	hti.flags = LVHT_ONITEMICON;
	hti.pt = pt;

	int item = ListView_HitTest( m_hListView, &hti );

	if ( item == -1 )
		return FALSE;

	LVITEMW lvi;
	memset( &lvi, 0, sizeof( lvi ) );
	lvi.iItem = item;
	lvi.mask = LVIF_IMAGE;

	ListView_GetItem( m_hListView, &lvi );

	if ( lvi.iImage == 0 ) // Empty
		return TRUE;

	if ( lvi.iImage == 3 ) // Subkey
		return TRUE;

	if ( lvi.iImage == 1 ) // Collapsed
		expandItem( item );
	else
		collapseItem( m_hListView, item );

	return TRUE;
}

bool cDlgBrowsePKUser::dblClick()
{
	int item = lvGetSelectedItem( m_hListView );
	if ( item == -1 )
		return false;

	return OnOK( IDOK );
	//return ShowCurrentItemProperties()
}

BOOL cDlgBrowsePKUser::lvNotify( NMHDR *hdr  )
{
	switch( hdr->code )
	{
		case LVN_ITEMCHANGED:
		case LVN_ITEMACTIVATE :
			return lvSelChanged();
		case NM_DBLCLK:
			return dblClick();
		case NM_CLICK:
		{
			NMITEMACTIVATE *nmia = ( NMITEMACTIVATE * )hdr;
			return checkHitPlus( nmia->ptAction );
		}
		default : break;
	}

	return FALSE;
}

int getMasterKey( HWND hList, int item )
{
	LVITEMW lvi;
	memset( &lvi, 0, sizeof( lvi ) );
	lvi.mask = LVIF_IMAGE;
	lvi.iItem = item;
	ListView_GetItem( hList, &lvi );
	if ( lvi.iImage != 3 ) // 3 - Sub key image.
		return item;

	if ( item == 0 )
		return -1;

	return getMasterKey( hList, item - 1 );
}

BOOL cDlgBrowsePKUser::ShowCurrentItemProperties()
{
	BYTE *keyID = LV_GetKeyID( m_hListView );
	if ( keyID == NULL )
	{
		assert( false );
		return false;
	}
	return ShowKeyPropertiesEx( m_hInst, m_hDlg, keyID, m_keyPkt, m_keyPktSize, m_secret );
}

bool cDlgBrowsePKUser::OnOK( int idCmd )
{
	int item = lvGetSelectedItem( m_hListView );
	if ( item == -1 )
	{
		EndDialog( m_hDlg, idCmd );
		return false;
	}

	//wchar_t text[ 1024 ];
	//ListView_GetItemTextT( m_hListView, item, 0, text, maxLen( text ) );

	BYTE *keyID = LV_GetKeyID( m_hListView, item );
	if ( keyID == NULL )
	{
		assert( false );
		return false;
	}

	cpgp_keyInfo keyInfo;
	cpgp_keyInfo *pKeyInfo = NULL;

	int res = pgp_FindKeyByIDEx(
				&keyInfo,
				keyID,
				m_keyPkt, m_keyPktSize,
				m_secret,
				&pKeyInfo 
				);


	if ( pKeyInfo == NULL )
		pKeyInfo = &keyInfo; // Master key used

	if ( res != ppe_ok )
	{
		DlgMessageBoxT( IDS_szErrorInKeyReading, MB_ICONSTOP );
		return false;
	}
/*
	if ( ( m_ptbKey == NULL ) && ( m_ptbKeyBlock == NULL ) )
	{
		cPlainTextBlock ptb;

		pgp_GetKeyUserID( &ptb, &keyInfo, pKeyInfo );
		ptb.PutByte( 0 );

		DeleteKey( pKeyInfo, (char *)ptb.buffer ); // &keyInfo );
		return false;
	}
*/
	if ( m_checkIfKeyMayBeUsedForEncryption )
	{	
		int res = IsEncryptionKey( (sPublicKey *)pKeyInfo->keyPkt, pKeyInfo->keyPktSize );

		if ( res == ppe_invalidAlgorithmID )
		{
			DlgMessageBoxT( IDS_SignatureAlgForEncryption, MB_ICONSTOP, _LoadIDS( IDS_ChoosePKUser ) );
			return false;
		}
		else
		if ( res != ppe_ok )
		{
			DlgMessageBoxT( IDS_szErrorInKeyReading, MB_ICONSTOP );
			return false;
		}
	}

	if ( m_ptbKey != NULL )
		m_ptbKey->WriteBytes( pKeyInfo->keyPkt, pKeyInfo->keyPktSize, true );

	if ( m_ptbKeyBlock != NULL )
	{
		BYTE *entireKeyPos;
		int entireKeySize;

		res = pgp_GetEntireKeyByID(
				&entireKeyPos, &entireKeySize, 
				m_keyPkt, m_keyPktSize,
				keyID,
				m_secret );

		if ( res != ppe_ok )
		{
			DlgMessageBoxT( IDS_szErrorInKeyReading, MB_ICONSTOP );
			return false;
		}

		m_ptbKeyBlock->WriteBytes( entireKeyPos, entireKeySize, true );
	}

	if ( m_wcsUserID != NULL )
	{
		cPlainTextBlock ptbUserID;
		pgp_GetKeyUserID( &ptbUserID, &keyInfo, pKeyInfo );
		m_wcsUserID->Add( (char *)ptbUserID.buffer, ptbUserID.bufferSize );
	}

	EndDialog( m_hDlg, idCmd );

	return true;
}

bool cDlgBrowsePKUser::OnDelete()
{
	BYTE *keyID = LV_GetKeyID( m_hListView );
	if ( keyID == NULL )
	{
		assert( false );
		return false;
	}

	cpgp_keyInfo keyInfo;
	cpgp_keyInfo *pKeyInfo = NULL;

	int res = pgp_FindKeyByIDEx(
				&keyInfo,
				keyID,
				m_keyPkt, m_keyPktSize,
				m_secret,
				&pKeyInfo 
				);


	if ( pKeyInfo == NULL )
		pKeyInfo = &keyInfo; // Master key used

	if ( res != ppe_ok )
	{
		DlgMessageBoxT( IDS_szErrorInKeyReading, MB_ICONSTOP );
		return false;
	}

	cPlainTextBlock ptb;

	pgp_GetKeyUserID( &ptb, &keyInfo, pKeyInfo );
	ptb.PutByte( 0 );

	return DeleteKey( pKeyInfo, (char *)ptb.buffer ) == ppe_ok;
}

bool cDlgBrowsePKUser::OnCommand( WPARAM wParam, LPARAM lParam )
{
	if ( (HWND)lParam == m_hToolBar )
	{
		if ( wParam == IDM_Browse )
		{
			BrowseForKeyRing();
			return true;
		}
		else
		if ( wParam == IDM_Delete )
		{
			return OnDelete();
		}

		return true;
	}

	//
	//	Check if menu command...
	//
	if ( HIWORD( wParam ) == 0 )
	{
		switch ( LOWORD( wParam ) )
		{
			case IDC_btnBrowse:
				BrowseForKeyRing();
				return true;
			case IDOK:
				return OnOK( LOWORD( wParam ) );
			case IDCANCEL:
				EndDialog( m_hDlg, LOWORD( wParam ) );
				return true;
			case IDC_btnProperties:
				ShowCurrentItemProperties( );
				return true;
			default:
				return false;
		}
	}

	return false;
}

void cDlgBrowsePKUser::UpdateButtons()
{
	BOOL enable = lvGetSelectedItem( m_hListView ) != -1;
	EnableWindow( GetDlgItem( m_hDlg, IDOK ), enable );
	EnableWindow( GetDlgItem( m_hDlg, IDC_btnProperties ), enable );

	enable &= ( ( m_ptbKey == NULL ) && ( m_ptbKeyBlock == NULL ) );
	ToolBar_EnableButton( m_hToolBar, IDM_Delete, enable );
}

BOOL cDlgBrowsePKUser::tbNotify( NMHDR *pnmh )
{
	LPNMHDR lpnm  = (LPNMHDR)pnmh;
	LPNMTOOLBAR lpnmTB = (LPNMTOOLBAR)pnmh;

	switch(lpnm->code)
	{
		//case TTN_GETDISPINFOW:
		//{
		//	LPNMTTDISPINFOW lpnmtdiW = (LPNMTTDISPINFOW)pnmh;

		//	// stores the info
		//	lpnmtdiW->uFlags = TTF_DI_SETITEM; //0;

		//	if ( pTranslateDll != NULL )
		//	{
		//		if ( pTranslateDll->m_flipDialog )
		//			lpnmtdiW->uFlags |= TTF_RTLREADING;

		//		lpnmtdiW->hinst = NULL;
		//		LoadStringT( m_hInst, GetIDSbyCmd( lpnmtdiW->hdr.idFrom ), lpnmtdiW->szText, 80 );
		//	}
		//	else
		//	{
		//		lpnmtdiW->hinst = m_hInst;
		//		lpnmtdiW->lpszText = MAKEINTRESOURCEW( GetIDSbyCmd( lpnmtdiW->hdr.idFrom ) );
		//	}

		//	return TRUE;
		//}
		//case TTN_GETDISPINFOA:
		//{
		//	LPNMTTDISPINFOA lpnmtdi = (LPNMTTDISPINFOA)pnmh;
		//	// stores the info
		//	lpnmtdi->uFlags = TTF_DI_SETITEM; // 0;

		//	if ( pTranslateDll != NULL )
		//	{
		//		if ( pTranslateDll->m_flipDialog )
		//			lpnmtdi->uFlags |= TTF_RTLREADING;

		//		lpnmtdi->hinst = NULL;
		//		LoadStringTA( m_hInst, GetIDSbyCmd( lpnmtdi->hdr.idFrom ), lpnmtdi->szText, 80 );
		//	}
		//	else
		//	{
		//		lpnmtdi->hinst = m_hInst;
		//		lpnmtdi->lpszText = MAKEINTRESOURCEA( GetIDSbyCmd( lpnmtdi->hdr.idFrom ) );
		//	}

		//	return TRUE;
		//}
		case TTN_GETDISPINFOW:
			return GetDispInfo( m_hInst, (LPNMTTDISPINFOW)pnmh, GetIDSbyCmd( pnmh->idFrom ) );

		case TTN_GETDISPINFOA:
			return GetDispInfo( m_hInst, (LPNMTTDISPINFOA)pnmh, GetIDSbyCmd( pnmh->idFrom ) );

		default :
			return FALSE;
	}

	return FALSE;
}

LRESULT cDlgBrowsePKUser::RealDlgProc(UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	LRESULT result;
	
	if ( cMenuItem::DlgProc( &result, m_hDlg, uMsg, wParam, lParam ) )
		return result;

	switch(uMsg)
	{
		case WM_INITDIALOG:
			InitDialog();
			break;

		case WM_DESTROY:
			DeleteDialog();
			break;

		case WM_COMMAND:
			return OnCommand( wParam, lParam );

		case WM_NOTIFY:
		{
			NMHDR *hdr = (NMHDR *)lParam;

			if ( ( hdr->code == TTN_GETDISPINFOA ) || ( hdr->code == TTN_GETDISPINFOW ) )
				tbNotify( hdr );
			else
			if ( wParam == IDC_LISTVIEW )
				return lvNotify( (LPNMHDR) lParam );
			else
			if ( ( hdr->hwndFrom == m_hToolBar ) ||
				( GetDlgCtrlID( m_hToolBar ) == wParam ) )
				return tbNotify( hdr );

			return false;
		}

		default:
			return false;
	}

	return true;
}
/*
void cDlgBrowsePKUser::OnSize( int x, int y, bool init )
{
	int i;

	if ( init )
	{
		pdrHWnd[0] = GetDlgItem( m_hDlg, IDOK );
		pdrHWnd[1] = GetDlgItem( m_hDlg, IDCANCEL );
		pdrHWnd[2] = GetDlgItem( m_hDlg, IDC_btnProperties);

		for( i = sizeof(pdrHWnd) / sizeof( HWND ); i--; )
		{
			RECT r;
			GetWindowRect( pdrHWnd[i], &r );
			POINT pos = { r.left, r.top };
			ScreenToClient( m_hDlg, &pos );

			pdrDx[i] = x - pos.x;
			pdrDy[i] = y - pos.y;
		}

		sdrHWnd[0] = m_hListView;

		for( i = sizeof(sdrHWnd) / sizeof( HWND ); i--; )
		{
			RECT r;
			GetWindowRect( sdrHWnd[i], &r );

			sdrDx[i] = x - ( r.right - r.left );
			sdrDy[i] = y - ( r.bottom - r.top );
		}

		prHWnd[0] = GetDlgItem( m_hDlg, IDC_btnBrowse );

		for( i = sizeof(prHWnd) / sizeof( HWND ); i--; )
		{
			RECT r;
			GetWindowRect( prHWnd[i], &r );
			POINT pos = { r.left, r.top };
			ScreenToClient( m_hDlg, &pos );

			prDx[i] = x - pos.x;
			prDy[i] = pos.y;
		}

		srHWnd[0] = GetDlgItem( m_hDlg, IDC_edtKeyRing );

		for( i = sizeof(srHWnd) / sizeof( HWND ); i--; )
		{
			RECT r;
			GetWindowRect( srHWnd[i], &r );

			srDx[i] = x - ( r.right - r.left );
			srDy[i] = r.bottom - r.top;
		}
	}
	else
	{
		const SWP_Flags = SWP_NOZORDER;

		for( i = sizeof(pdrHWnd) / sizeof( HWND ); i--; )
		{
			SetWindowPos(
				pdrHWnd[i],
				NULL,
				x - pdrDx[i],
				y - pdrDy[i],
				0,
				0,
				SWP_Flags | SWP_NOSIZE );
		}

		for( i = sizeof(sdrHWnd) / sizeof( HWND ); i--; )
		{
			SetWindowPos(
				sdrHWnd[i],
				NULL,
				0,
				0,
				x - sdrDx[i],
				y - sdrDy[i],
				SWP_Flags | SWP_NOMOVE );
		}

		for( i = sizeof(prHWnd) / sizeof( HWND ); i--; )
		{
			SetWindowPos(
				prHWnd[i],
				NULL,
				x - prDx[i],
				prDy[i],
				0,
				0,
				SWP_Flags | SWP_NOSIZE );
		}

		for( i = sizeof(srHWnd) / sizeof( HWND ); i--; )
		{
			SetWindowPos(
				srHWnd[i],
				NULL,
				0,
				0,
				x - srDx[i],
				srDy[i],
				SWP_Flags | SWP_NOMOVE );
		}

		ShowWindow( m_hDlg, SW_SHOW );
	}
}
*/

void cDlgBrowsePKUser::DeleteDialog()
{
	LV_DeleteAllItems( m_hListView );
	DestroyWindow( m_hToolBar );
}

//
//	Choose public key or subkey for encryption
//
//	1. Parses key packet with subkeys.
//	2. Retures ppe_ok if the key has no subkeys and it may be used for encryption
//	( RSA or ElGamal algorithm, never DSA ).
//	3. If subkeys present 
//		3.1 Uses cDlgBrowsePKUser without key ring values
//		3.2 Allows to choose directly the key or it's subkeys
//	4. It does not allows to choose a key what may be used for signature only.
//	Returns true if key was chosen otherwise returns false
//
bool ChooseKeyOrSubkeyT(
		HINSTANCE hInst,
		HWND hWnd,
		BYTE *keyPkt, int keyPktSize,	// Common key packet
		cPlainTextBlock *ptbKey,			// Choosen direct key or it's subkey
		//cPlainTextBlock *ptbUserID		// User ID of choosen packet
		cWCStr *wcsUserID
	)
{
	// 2. 
	int size = 0;
	int res = pgp_EnumKeyIDs( NULL, &size, keyPkt, keyPktSize, false );
	if ( size == pgp_keyIDSize ) // The common key packet does not content subkeys
	{
		BYTE tag;
		int dataSize;
		int hdrSize;
		bool partial = 0;

		int res;
		int recCnt = 1;

		int parsed = 0;

		//
		//	Parses public key packet.
		//
		res = pgp_PacketHeader_Parse( &tag, &hdrSize, &dataSize, &partial, keyPkt, keyPktSize );
		if ( ( res == ppe_ok ) &&
			 ( tag == pkt_PublicKey )
			 )
		{
			sPublicKey *pk = (sPublicKey *)(keyPkt + parsed + hdrSize);
			int pkSize = dataSize;

			if ( IsEncryptionKey( pk, pkSize ) == ppe_ok )
			{
				bool res = true;
	//			if ( res ) res = ptbKeyID->Allocate( pgp_keyIDSize ) != NULL;
	//			if ( res ) res = pgp_EnumKeyIDs( ptbKeyID->buffer, &size, keyPkt, keyPktSize, false ) == ppe_ok;
	//			if ( res ) res = ptbUserID->WriteBytes( keyPkt, keyPktSize, true ) != 0;
//				if ( res ) res = ptbKey->WriteBytes( keyPkt, keyPktSize ) != 0;
				if ( res ) res = ptbKey->WriteBytes( (BYTE *)pk, pkSize ) != 0;
				if ( res ) 
				{
					if ( wcsUserID == NULL )
						return true;

					parsed += dataSize + hdrSize;

					//
					//	Parses user ID
					//	

					int iRes = pgp_PacketHeader_Parse( &tag, &hdrSize, &dataSize, &partial, keyPkt + parsed, keyPktSize - parsed );
					if ( iRes != ppe_ok ) res = false;
					if ( tag != pkt_UserID ) res = false;

					if ( res )
					{
						BYTE *userID = ( keyPkt + parsed + hdrSize );
						int  userIDSize = dataSize;
						wcsUserID->Add( (char *)userID, userIDSize );
						return true;
					}
				}
			}
		}
	}

	cDlgBrowsePKUser dlg( keyPkt, keyPktSize, ptbKey, wcsUserID, true );

	res = dlg.DoModal( hInst, dlg.IDD, hWnd );

	return res == IDOK;
}

//
//	BrowseKeyRingForUser
//		Visual interfase for choosing a public or secret key from PGP key ring ( *.pkr, *.skr, *.gpg )
//		It allows to dynamically change key file path, so krOut contains the new path
//		ptbUserID is 
//			UserID for PGP key packet,
//			Friendly name for P12 secret key packet,
//			Subject RDN for certificate
//
//		If both ptbKey and ptbKeyBlock are NULL the function is used for viewing and deleting a key
//		OK button in corresponded dialogue is changed as Delete and does not close the dialog.
//
//int BrowseKeyRingForUser(
//		HINSTANCE hInst,
//		HWND hWnd,
//		cPlainTextBlock *ptbKey, // [OUT] - pure key packet in sPublicKey or sSecretKey format as declared in Packets.h
//		const wchar_t *keyRingPath,
//		char *krOut, // = NULL // if krOut != NULL it is to be MAX_PATH size buffer.
//		bool secret,
//		cPlainTextBlock *ptbUserID,
//		cPlainTextBlock *ptbKeyBlock
//				// [OUT] Entire key packet - master key with subpackets including subkeys packets
//	)
//{
//	wchar_t *krOutW = NULL;
//
//	if ( krOut != NULL )
//	{
//		krOutW = new wchar_t[ MAX_PATH ];
//		krOutW[0] = L'\0';
//	}
//
//	int res = BrowseKeyRingForUserT( hInst, hWnd, ptbKey, keyRingPath, krOutW, secret, wcsUserID, ptbKeyBlock );
//
//	if ( krOutW != NULL )
//		_wcstombs( krOut, krOutW, MAX_PATH );
//
//	return res;
//}

int BrowseKeyRingForUserT(
		HINSTANCE hInst,
		HWND hWnd,
		cPlainTextBlock *ptbKey, // [OUT] - pure key packet in sPublicKey or sSecretKey format as declared in Packets.h
		const wchar_t *keyRingPath,
		wchar_t *krOut, // = NULL // if krOut != NULL it is to be MAX_PATH size buffer.
		bool secret,
		cWCStr *wcsUserID, // cPlainTextBlock *ptbUserID,
		cPlainTextBlock *ptbKeyBlock
				// [OUT] Entire key packet - master key with subpackets including subkeys packets
	)
{
	if ( hInst == NULL )
		hInst = hMainInst; // GetModuleHandle( "BCKeyMan.dll" );
	 
	cDlgBrowsePKUser dlg( ptbKey, keyRingPath, secret, wcsUserID, ptbKeyBlock );

	int res = dlg.DoModal( hInst, dlg.IDD, hWnd );

	if ( krOut )
		wcsncpy( krOut, dlg.m_krPath, MAX_PATH );

	return res;
}

bool BrowseForPKBlockT(
		HINSTANCE hInst,
		HWND hWnd,
		const wchar_t *initialPath, // if initialPath == NULL it runs in BCDB
		bool secret,
		cPlainTextBlock *ptbKey,	// [OUT] key in x509 or PGP format
		type_kff *pKeyPktFormat,	// [OUT] Format of key packet
		cWCStr *wcsUserID,			// [OUT] Pgp subkey does not have an user ID ( master key has only )
									// but we have to ask password for secret key with a name.
									// The BrowseKeyRingForUser generates here the subkey name as 
									// as "Subkey of ( master key user ID )" or  key user ID if the key has it
		cPTB *entirePacket			// [OUT] entire key ring.
	)
{
	if ( hInst == NULL )
		hInst = hMainInst;

	wchar_t pathW[ MAX_PATH ];
	pathW[0] = L'\0';

	if ( initialPath != NULL )
		wcscpy_s( pathW, initialPath );

	//static const wchar_t *szFiltersPublic = 
	//	L"X.509 certificates ( *.cer )\0*.cer\0"
	//	L"PGP public keys ( *.pkr,*.gpg,*.asc )\0*.pkr;*.gpg;*.asc\0"
	//	L"All files ( *.* )\0*.*\0";

	//static const wchar_t *szFiltersSecret = 
	//	L"PKCS 12 ( *.p12 )\0*.p12\0"
	//	L"PGP secret keys ( *.skr,*.gpg )\0*.skr;*.gpg\0"
	//	L"All files ( *.* )\0*.*\0";	

	wchar_t szFilters[ 1024 ];
	LoadStringT( hInst, secret ? IDS_szFiltersSecret : IDS_szFiltersPublic, szFilters, maxLen( szFilters ) );

	wchar_t text[ 1024 ];
	LoadStringT( hInst, secret ? IDS_szChooseSecretKeyAndPressGet : IDS_szChoosePublicKeyAndPressGet, text, maxLen( text ) );

	bool res = BrowseKeyFiles( 
			hWnd,
			pathW, maxLen( pathW ),
			szFilters,
			_LoadIDS( secret ? IDS_ChooseSecretKey : IDS_ChoosePublicKey ),
			text,
			_LoadIDS( IDS_szGetItBtnText )
		);

	if ( !res )
		return 1;

	cUniversalString usFriendlyName;

	type_kff kff = getKeyFileFormat( pathW );

	if ( pKeyPktFormat != NULL )
		*pKeyPktFormat = kff;

	if ( kff == kffX509 )
	{
		cFileView cfv( pathW );

		if ( cfv.m_err )
			return false;

		return ptbKey->WriteBytes( cfv.m_ptr, (int)cfv.m_size.QuadPart ) == cfv.m_size.QuadPart;
	}
	else
	if ( ( kff == kffPGP ) || ( kff == kffPGPAscii ) )
	{
		cWCStr wcsFriendlyName;
		cPTB keyRing;
		//
		//	v. 1.05.1.1 - 25.11.2015
		//	Bug #2388 Containers encrypted with PGP keys
		//	Prev version does not extract required key from key ring it just fill entireky key ring and parse it as secret key.
		//
		res = BrowseKeyRingForUserT( hInst, hWnd, ptbKey, pathW, pathW, secret, &wcsFriendlyName, &keyRing ) == IDOK;

		if ( !res )
			return false;

		if ( pKeyPktFormat != NULL )
			*pKeyPktFormat = kffPGP;

		if ( wcsUserID != NULL )
			return ( wcsUserID->Add( wcsFriendlyName.m_szBody ) != 0 );

		if ( entirePacket != NULL )
			keyRing.Swap( entirePacket );

		return true;
	}
	else
	{
		DlgMessageBoxT( hInst, hWnd, IDS_UndefinedKeyFileFormat, MB_ICONSTOP );
		return false;
	}
}

//
//	BrowseForPublicKey
//		Uses OpenFile to choose a public or secret key file from BCDB ( *.p12, *.cer ) 
//		or from PGP key ring by means of BrowseKeyRingForUser ( *.pkr, *.skr, *.gpg )
//		password is prompted for secret key decryption
//
bool BrowseForPublicKey(
		HINSTANCE hInst,
		HWND hWnd,
		const char *initialPath,
		bool secret,
		HPKA *pHPKA
	)
{
	return BrowseForPublicKeyT( hInst, hWnd, getWCStr( initialPath ), secret, pHPKA );
}

bool BrowseForPublicKeyT(
		HINSTANCE hInst,
		HWND hWnd,
		const wchar_t *initialPath,
		bool secret,
		HPKA *pHPKA
	)
{
	if ( hInst == NULL )
		hInst = hMainInst;

	wchar_t path[ MAX_PATH ];
	path[0] = L'\0';

	if ( initialPath != NULL )
		wcscpy_s( path, initialPath );

	if ( wcslen( path ) == 0 )
		BCDB_GetDataBasePath( path, maxLen( path ) );

	//static const wchar_t *szFiltersPublic = 
	//	L"PKCS 12 ( *.cer )\0*.cer\0"
	//	L"PGP public keys ( *.pkr,*.gpg )\0*.pkr;*.gpg\0"
	//	L"All files ( *.* )\0*.*\0";

	//static const wchar_t *szFiltersSecret = 
	//	L"PKCS 12 ( *.p12 )\0*.p12\0"
	//	L"PGP secret keys ( *.skr,*.gpg )\0*.skr;*.gpg\0"
	//	L"All files ( *.* )\0*.*\0";

	//const wchar_t *szFilters = secret ? szFiltersSecret : szFiltersPublic;

	wchar_t szFilters[ 1024 ];
	LoadStringT( hInst, secret ? IDS_szFiltersSecret : IDS_szFiltersPublic, szFilters, maxLen( szFilters ) );

	bool res = BrowseFiles( hWnd, path, maxLen( path ), szFilters, _LoadIDS( secret ? IDS_ChooseSecretKey : IDS_ChoosePublicKey ), false ) == TRUE;
	if ( !res )
		return 1;

	cUniversalString usFriendlyName;

	type_kff kff = getKeyFileFormat( path );

	if ( kff == kffX509 )
	{
		cFileView cfv( path );

		if ( cfv.m_err )
			return false;

		if ( secret )
			return BCDB_GetP12SecretKeyInfo( hInst, hWnd, cfv.m_ptr, (int)cfv.m_size.QuadPart, pHPKA, &usFriendlyName );
		else
			return BCDB_GetCerPublicKeyInfo( hInst, hWnd, cfv.m_ptr, (int)cfv.m_size.QuadPart, pHPKA );
	}
	else
	if ( ( kff == kffPGP ) || ( kff == kffPGPAscii ) )
	{
		cPlainTextBlock ptbKey; // , ptbFriendlyName;
		cWCStr wcsFriendlyName;

		res = BrowseKeyRingForUserT( hInst, hWnd, &ptbKey, path, path, secret, &wcsFriendlyName ) == IDOK;
		if ( !res )
			return false;

		if ( secret )
		{
			//ptbFriendlyName.PutByte( 0 );
			wcsFriendlyName.Add( L"" );

			return BCDB_GetPGPSecretKeyInfo( 
				hInst, hWnd,
				ptbKey.buffer, ptbKey.bufferSize, 
				pHPKA,
				wcsFriendlyName.m_szBody );
				//ptbFriendlyName.buffer );
		}
		else
		{
			return BCDB_GetPGPPublicKeyInfo( hInst, hWnd, ptbKey.buffer, ptbKey.bufferSize, pHPKA );
		}
	}
	else
	{
		DlgMessageBoxT( hInst, hWnd, IDS_UndefinedKeyFileFormat, MB_ICONSTOP );
		return false;
	}
}

bool BrowseForSecretKey(
		BYTE *keyID,
		int keyIDSize,
		cPlainTextBlock *ptbKey,// [OUT] key in x509 or PGP format
		type_kff *pKeyPktFormat, // [OUT] Format of key packet
		cWCStr *wcsUserID // = NULL
	)
{
	wchar_t idStr[ 256 ];
	BCDB_bytes2string( idStr, maxLen( idStr ), keyID, keyIDSize );

	wchar_t message[ 1024 ];
	FormatT( message, maxLen( message ), _LoadIDS ( IDS_ChooseSecretKeyWithID_S ), idStr );

	wchar_t path[ 1024 ];
	path[0] = L'\0';

	// PKCS 12 ( *.p12 )\000*.p12\000PGP secret keys ( *.skr,*.gpg )\000*.skr;*.gpg\000All files ( *.* )\000*.*\000

	if ( !BrowseFiles( 0, path, maxLen( path ), _LoadIDS( IDS_szFiltersSecret ), message, false ) )
		return false;
	
	type_kff kff = getKeyFileFormat( path );

	cPlainTextBlock ptbDecoded;
	cFileView fv( path );

	if ( fv.m_err )
		return false;

	BYTE *keyPtr = fv.m_ptr;
	int keySize = (int)fv.m_size.QuadPart;

	if ( kff == kffPGPAscii )
	{
		const char *szPGP_B64_Begin = "-----BEGIN PGP PRIVATE KEY BLOCK-----";
		const char *szPGP_B64_End   = "-----END PGP PRIVATE KEY BLOCK-----";

		if ( readArmoredBlock( &ptbDecoded, (const char *)fv.m_ptr, (int)fv.m_size.QuadPart, szPGP_B64_Begin, szPGP_B64_End ) )
		{
			kff = kffPGP;
			keyPtr = ptbDecoded.buffer;
			keySize = ptbDecoded.bufferSize;
		}
	}

	if ( pKeyPktFormat ) 
		*pKeyPktFormat = kff;

	if ( kff == kffX509 )
	{
		if ( wcsUserID != NULL )
		{
			cPlainTextBlock ptb, ptbAttrSet;
			P12_attrSet attrSet;

			int res = P12_PFX_PDU_parseEx( keyPtr, keySize, &ptbAttrSet, NULL, &ptb, NULL );
			if ( ( res == x509e_ok ) || ( res == x509e_emptyParameters ) || ( res == ppe_invalidPassword ) )
			{
				res = attrSet.parse( ptbAttrSet.buffer, ptbAttrSet.bufferSize, NULL, NULL, asn1tag_CONSTRUCTED_SET );
				if ( res == x509e_ok )
				{
					x509_struct sLocalKeyId;
					res = attrSet.getLocalKeyId( &sLocalKeyId );
					if ( res == x509e_ok )
					{
						bool keyIDIsEmpty = true;
						for( int i = pgp_keyIDSize; i--; )
						{
							if ( keyID[i] != 0 )
							{
								keyIDIsEmpty = false;
								break;
							}
						}

						if ( !keyIDIsEmpty && 
							  (
								( keyIDSize != sLocalKeyId.L ) || 
								( memcmp( keyID, sLocalKeyId.B, keyIDSize ) != 0 )
							  )
							)
						{
							// MessageBoxT( 0, L"Key ID is not compatible"... )
							return false;
						}

						x509_struct sFriendlyName;
						attrSet.getFriendlyName( &sFriendlyName );

						if ( res == x509e_ok )
						{
							cUniversalString us;
							// Really sFriendlyName is defenitelly big-endian
							us.SetUnicode_CheckBigEndianInput( sFriendlyName.B, sFriendlyName.L );
							us.MakeFromUnicodeString();
							wcsUserID->Add( us.wcBuffer );
						}
					}
				}
			}

			if ( wcsUserID->isEmpty() )
				wcsUserID->Add( path );
		}

		return 	( ptbKey->WriteBytes( keyPtr, keySize, true ) != 0 );
	}

	cpgp_keyInfo keyInfo, *pki;

	int res = pgp_FindKeyByIDEx(
				&keyInfo,
				keyID,
				keyPtr, keySize,
				true,
				&pki
			);

	if ( res != ppe_ok )
		return false;

	if ( pki == NULL )
		pki = &keyInfo;

	if ( memcmp( pki->keyID, keyID, keyIDSize ) != 0 )
		return false;

	if ( wcsUserID != NULL )
	{
		cPlainTextBlock ptb;
		pgp_GetKeyUserID( &ptb, &keyInfo, pki );
		wcsUserID->Add( (const char *)ptb.buffer, ptb.bufferSize );
	}

	return ptbKey->WriteBytes( pki->keyPkt, pki->keyPktSize ) == pki->keyPktSize;
}


///////////////////////////////////////////////////////////////////
//
//	GetPKBlockByID
//	Tries to find key with specified ID in BCDB
//	Uses ( .cer ) files for public  or ( .p12 ) files for secret key
//	Choose public ( in local.pkr ) or secret ( in local.skr ) key from BCDB
//
//	If specified key is found, returns true and fills ptbKey buffer 
//	Note, the packet is not parsed.
//
//	Usable function to extract corresponded key from the package
//
//		Public kffX509 - x509_certificate::parse	to get x509_publicKeyInfo as value tbsCertificate.subjectPKInfo 
//													and then pka_createAlgorithmX509 to create key handle
//		Secret kffX509 - P12_PFX_PDU_parse			to get x509_privateKeyInfo in ptbInfo.buffer 
//													and then ska_createAlgorithmX509 to create keu handle
//
//		Public kffPGP  - PKT_CreatePublicKeyAlgFromPacket to create public key handle
//
//		Secret kffPGP  - PKT_CreateSecretKeyAlgFromPacket to create secret key handle
//
//	Returns true if specified key is found
//
bool GetPKBlockByIDT(
		HINSTANCE hInst,
		HWND hWnd,
		BYTE *keyID,
		int keyIDSize,
		bool secret,
		cPlainTextBlock *ptbKey,// [OUT] key in x509 or PGP format
		type_kff *pKeyPktFormat, // [OUT] Format of key packet
		cWCStr *wcsUserID //cPlainTextBlock *ptbUserID
	)
{
	if ( hInst == NULL )
		hInst = hMainInst;

	const int lkidSize = 20;

	if ( keyIDSize >= lkidSize )
	{
		if ( BCDB_GetX509KeyByID( ptbKey, keyID, keyIDSize, secret, wcsUserID )) // ptbUserID ) )
		{
			if ( pKeyPktFormat ) *pKeyPktFormat = kffX509;
				return true;
		}

		//
		// Version 1.05.1
		//	- 20.11.2015 5.
		//	Local key id at secret key packet is certificate thumbprint.
		//	 a. Finds public key byt the ID.
		//	 b. Makes thumbprint
		//	 c. Finds secret key by the thumbprint again.
		//
		if ( secret )
		{
			cPTB ptbCrt;
			BYTE thumbPrint[ lkidSize ];
			if ( BCDB_GetX509KeyByID( &ptbCrt, keyID, lkidSize, false ) &&
				( x509_certificate::generateThumbprint( ptbCrt.buffer, ptbCrt.bufferSize, thumbPrint ) == x509e_ok ) &&
				BCDB_GetX509KeyByID( ptbKey, thumbPrint, lkidSize, true, wcsUserID ) )
			{
				if ( pKeyPktFormat ) *pKeyPktFormat = kffX509;
				return true;
			}
		}
	}

	wchar_t path[ MAX_PATH ];
	if ( !BCDB_GetLocalKeyRingPath( path, maxLen( path ), secret ) )
		return false;

	if ( keyIDSize != pgp_keyIDSize )
		return false;

	cFileView cfv( path );

	if ( cfv.m_err )
		return false;

	const char *szPGPKeyStart	= "-----BEGIN PGP PRIVATE KEY BLOCK-----";
	const char *szPGPKeyEnd		= "-----END PGP PRIVATE KEY BLOCK-----";

	if ( !secret )
	{
		szPGPKeyStart	= "-----BEGIN PGP PUBLIC KEY BLOCK-----";
		szPGPKeyEnd		= "-----END PGP PUBLIC KEY BLOCK-----";
	}

	BYTE *keyBf = cfv.m_ptr;
	int keyBfSize = (int)cfv.m_size.QuadPart;

	cPlainTextBlock ptbBf;

	if ( readArmoredBlock( &ptbBf, (const char *)keyBf, keyBfSize, szPGPKeyStart, szPGPKeyEnd ) )
	{
		keyBf = ptbBf.buffer;
		keyBfSize = ptbBf.bufferSize;
	}
	
	cpgp_keyInfo keyInfo, *pki;

	int res = pgp_FindKeyByIDEx(
				&keyInfo,
				keyID,
				keyBf, keyBfSize,
				secret,
				&pki
			);

	if ( res != ppe_ok )
		return false;

	if ( pki == NULL )
		pki = &keyInfo;

	if ( wcsUserID ) // ptbUserID )
	{
		cPlainTextBlock ptb;
		pgp_GetKeyUserID( &ptb, &keyInfo, pki );
		wcsUserID->Add( (const char *)ptb.buffer, ptb.bufferSize );
	}

	if ( memcmp( pki->keyID, keyID, keyIDSize ) != 0 )
		return false;

	if ( pKeyPktFormat ) *pKeyPktFormat = kffPGP;

	return ptbKey->WriteBytes( pki->keyPkt, pki->keyPktSize ) == pki->keyPktSize;
}

bool BrowseKeyFiles(
		HWND hWnd,
		wchar_t *pathBuff, int buffSize,
		const wchar_t *filters,
		const wchar_t *caption,
		const wchar_t *notes,
		const wchar_t *OKBtnText,
		DWORD dwFlags
	)
{
	if ( pathBuff[0] == L'\0' )
		BCDB_GetDataBasePath( pathBuff, buffSize );

	cDlgBrowseKeyFile dlg( caption, pathBuff, filters, notes, OKBtnText, dwFlags );

	int res = dlg.DoModal( hMainInst, IDD_BrowseKeyFile, hWnd );

	if ( res == IDOK )
	{
		if ( (int)wcslen( dlg.m_path ) > buffSize )
		{
			assert( false );
			return false;
		}

		wcsncpy( pathBuff, dlg.m_path, buffSize );
		return true;
	}

	return false;
}

int cDlgBrowsePKUser::DeleteKey( cpgp_keyInfo * pKeyInfo, const char *text )
{
	if ( !FileExistsT( m_krPath ) )
	{
		MessageBoxT( m_hDlg, m_krPath, _LoadIDS( IDS_KRFileDoesNotExist ), MB_ICONSTOP );
		assert( false );
		return -1;
	}
	else
	if ( m_kr == NULL )
	{
		assert( false );
		return -1;
	}

	type_kff kff = getKeyFileFormat( m_krPath );

	if ( kff == kffPGPAscii )
	{
		DlgMessageBoxT( IDS_ASCIIFileFormat, MB_ICONSTOP );
		assert( false );
		return -1;
	}

	cPlainTextBlock ptb;
	if ( ptb.WriteBytes( m_kr->m_ptr, (int)m_kr->m_size.QuadPart ) != m_kr->m_size.QuadPart )
	{
		DlgMessageBoxT( IDS_ErrorInAllocatingMemory, MB_ICONSTOP );
		assert( false );
		return -1;
	}

	wchar_t message[ 1024 ];
	FormatT( message, maxLen( message ),
		_LoadIDS( IDS_AreYouSureYouWantToDeleteKey_SS_Q ),
		L"", // _LoadIDS( secret ? IDS_szSecret : IDS_szPublic ),
		getWCStr( text ) );

	int res = DlgMessageBoxT( message, MB_ICONQUESTION | MB_YESNO );
	if ( res != IDYES )
		return -1;

	res = pgp_RemoveKeyByID( ptb.buffer, &ptb.bufferSize, pKeyInfo->keyID, m_secret );

	if ( res != ppe_ok )
	{
		assert( false );
		return res;
	}

	delete( m_kr );
	m_kr = NULL;

	if ( ptb.bufferSize == 0 )
	{
		if ( !DeleteFileT( m_krPath ) )
			return -1;
	}
	else
	if ( WriteToFile( m_krPath, ptb.buffer, ptb.bufferSize ) != ptb.bufferSize )
	{
		assert( false );
		return -1;
	}

	res = FillListView( m_krPath );

	if ( res != ppe_ok )
		assert( false );

	return res;
}

bool GetDataBasePath( wchar_t *buff, int buffLen, DWORD dwFlags )
{
	if ( dwFlags == GetDataBasePath_Flag_PublicKeyRing )
		return BCDB_GetLocalKeyRingPath( buff, buffLen, false );
	else
	if ( dwFlags == GetDataBasePath_Flag_SecretKeyRing )
		return BCDB_GetLocalKeyRingPath( buff, buffLen, true );
	else
		return BCDB_GetDataBasePath( buff, buffLen );
}