#ifndef _cDlgBrowseKeyFile_
#define _cDlgBrowseKeyFile_

#include "..\Includes\cDlg.h"
#include "..\Includes\IconCache.h"

class cDlgBrowseKeyFile : public cColoredDlg
{
	enum { flag_UseAddKeyButton = 1, flag_UseNewKeyButton = 2, flag_DisableBrowseButtons = 4 };

	cIconCache m_iconCache;
	HWND m_hList;
	wchar_t *m_caption;
	wchar_t *m_notes;

	// "Secret key PKCS-12 format ( *.p12) \0*.p12\0All files\0*.*\0\0"
	const wchar_t *m_filters;
	wchar_t *m_filtersBuff; // Buffer to be deleted.
	wchar_t *m_OKBtnText;

	// Fills filter combobox
	void InitFilters();
	// Returns pointer to current selected filter string f.e. "*.p12\0...."
	const wchar_t *GetCurrentFilter();

 public:

	wchar_t *m_path;

	cDlgBrowseKeyFile( const wchar_t *caption, const wchar_t *path, const wchar_t *filters, const wchar_t *notes, const wchar_t *OKBtnText, DWORD dwFlags )
	{
		m_caption = newString( caption );

		m_path = new wchar_t[ MAX_PATH ];
		if ( m_path ) wcsncpy( m_path, path, MAX_PATH );

		m_notes = newString( notes );

		m_filtersBuff = NULL;

		int len = wcslen( filters );
		m_filters = filters;
		if ( ( filters != NULL ) && ( len != 0 ) )
		{
			if ( filters[ len - 1 ] == L'%' ) // save filter. MS Visual Studio 8 loses zero separation at resource.
			{
				int filterLen = wcslen( filters ) + 1;
				m_filtersBuff = new wchar_t [ filterLen ];

				if ( m_filtersBuff )
				{
					wcsncpy( m_filtersBuff, filters, filterLen );

					for ( wchar_t *s = m_filtersBuff; *s; s++ )
					{
						if ( *s == L'%' )
							*s = L'\0';
					}

					m_filters = m_filtersBuff;
				}
			}
		}
		
		
		m_OKBtnText = newString( OKBtnText );

		m_dwFlags = dwFlags;
	}
	~cDlgBrowseKeyFile()
	{
		if ( m_caption ) delete m_caption;
		if ( m_path ) delete m_path;
		if ( m_notes ) delete m_notes;
		if ( m_OKBtnText ) delete m_OKBtnText;
		if ( m_filtersBuff ) delete m_filtersBuff;
	}

 private:

	DWORD m_dwFlags;

	LRESULT RealDlgProc( UINT uMsg, WPARAM wParam, LPARAM lParam );

	HWND m_hToolbar;
	int InitToolBar();
	bool InitDialog();
	bool OnCommand( WPARAM wParam, LPARAM lParam );
	bool OnLVNotify( NMHDR *pNMHdr );
	int OnGetMinMaxInfo( MINMAXINFO *pMMI );
	bool UpdateListView( const wchar_t *folder, const wchar_t *pattern );
	bool InitListViewColumns();
	bool SetCurrentFolder( const wchar_t *folder, const wchar_t *pattern );
	int AddLVItem( const wchar_t *path, WIN32_FIND_DATAW *findData, int itemNum = 0 );
	int ToolBarNotify( NMHDR *pnmh );
	int UpdateButtons();
	// Asks confirmation and delete all selected items.
	int DeleteSelected(void);
	bool OnOK( int iItem = -1 );
};

#endif // _cDlgBrowseKeyFile_