#include "cWizardDlgs.h"
#include "BCKeyMan.h"
#include "BCDB_Functions.h"
#include "PKParse.h"

#include <windowsx.h>
#include <io.h>
#include <stdio.h>

#include "..\Includes\stddlgs.h"
#include "..\CertificateLib\PGPPkts\Packets.h"
#include "..\CertificateLib\x509pkts\pkcs12_funcs.h"
#include "..\CertificateLib\x509pkts\base64funcs.h"
#include "..\CertificateLib\PGPPkts\PGPPkts.h"
#include "..\CertificateLib\PEM\cPEMCtx.h"

#include "..\Includes\OpenFiles.h"
#include "..\Includes\FileView.h"

static void FillListView( HWND hList, const wchar_t *ptr )
{
	cPlainTextBlock ptbCopy;
	ptbCopy.WriteString( ptr, true );

	wchar_t *begin = (wchar_t *)ptbCopy.buffer;

	while( *begin != L'\0' )
	{
		wchar_t *end = begin;
		for( ; *end; end++ )
		{
			if ( *end == L'\n' )
			{
				*end = L'\0';
				end++;
				break;
			}
		}

		AddString( hList, begin );
		begin = end;
	}
}

bool cWizardDlg_FinalInfo::OnInitDialog()
{
	HWND hList = GetDlgItem( m_hDlg, IDC_List );
	FillListView( hList, m_finalInfo->m_szBody );
	return true;
}

bool cWizardDlg_Welcome::OnInitDialog()
{
	Button_SetCheck( GetDlgItem( m_hDlg, IDC_rbGenerateNewKey ), m_pBlock->generateNew ? BST_CHECKED : BST_UNCHECKED );
	Button_SetCheck( GetDlgItem( m_hDlg, IDC_rbUseExistingKey ), !m_pBlock->generateNew ? BST_CHECKED : BST_UNCHECKED );

	SetExistentKeyInfo();

	Button_SetCheck( GetDlgItem( m_hDlg, IDC_chkCreateSecretKeyPacket ), m_pBlock->generateSecKey ? BST_CHECKED : BST_UNCHECKED );
	Button_SetCheck( GetDlgItem( m_hDlg, IDC_chkCreatePublicKeyPacket ), m_pBlock->generateCertificate ? BST_CHECKED : BST_UNCHECKED );

	Button_SetCheck( GetDlgItem( m_hDlg, IDC_chkSaveToBCStorage ), m_pBlock->saveToBCStorage ? BST_CHECKED : BST_UNCHECKED );
	Button_SetCheck( GetDlgItem( m_hDlg, IDC_chkSaveToFile ), m_pBlock->saveToFiles ? BST_CHECKED : BST_UNCHECKED );

	UpdateControls();

	return true;
}

static DWORD MenuHelpIDs[] = 
{
	IDC_rbGenerateNewKey,			IDS_rbGenerateNewKey,
	IDC_rbUseExistingKey,			IDS_rbUseExistingKey,
	IDC_chkCreateSecretKeyPacket,	IDS_chkCreateSecretKeyPacket,
	IDC_chkCreatePublicKeyPacket,	IDS_chkCreatePublicKeyPacket,
	IDC_edtPath,					IDS_edtPath,
	IDC_chkSaveToBCStorage,			IDS_chkSaveToBCStorage,
	IDC_chkSaveToFile,				IDS_chkSaveToFile,
	0,0
};

void cWizardDlg_Welcome::OnCtxHelp( HELPINFO *hInfo )
{
	m_useCtxAsIDS = true;
	OnHelpInfo( hInfo, MenuHelpIDs );
}

bool cWizardDlg_Welcome::OnCommand( WPARAM wParam, LPARAM lParam )
{
	if ( !m_wasInited )
	{
//		assert( false );
		return false;
	}

	if ( OnStdButton( wParam ) )
	{
		BOOL success = FALSE;
		UINT size = GetDlgItemInt( m_hDlg, IDC_edtLength, &success, 0 );
		if ( success )
			m_pBlock->keyBitSize = size;

//		m_pBlock->algPos = ComboBox_GetCurSel( GetDlgItem( m_hDlg, IDC_cmbAlgorithm ) );
//		m_pBlock->generateNew = Button_GetCheck( GetDlgItem( m_hDlg, IDC_rbGenerateNewKey ) ) == BST_CHECKED;

		return true;
	}
	else
	{
		switch ( LOWORD(wParam) )
		{
			case IDC_chkSaveToBCStorage:
				m_pBlock->saveToBCStorage = Button_GetCheck( GetDlgItem( m_hDlg, IDC_chkSaveToBCStorage ) ) == BST_CHECKED;
				m_pBlock->wasGenerated = false;
				UpdateControls();
				break;
			case IDC_chkSaveToFile:
				m_pBlock->saveToFiles = Button_GetCheck( GetDlgItem( m_hDlg, IDC_chkSaveToFile ) ) == BST_CHECKED;
				m_pBlock->wasGenerated = false;
				UpdateControls();
				break;
			case IDC_rbUseExistingKey:
			{
				bool checked = Button_GetCheck( GetDlgItem( m_hDlg, IDC_rbUseExistingKey ) ) == BST_CHECKED;

				if ( checked )
				{
					if ( OnBrowseKeyFile() )
					{
						m_pBlock->wasGenerated = false;
						UpdateControls();
					}
				}
//				else
//					Button_SetCheck( GetDlgItem( m_hDlg, IDC_rbUseExistingKey ), BST_UNCHECKED );
				break;
			}
			case IDC_rbGenerateNewKey:

				if ( ( m_pBlock->ska != NULL ) && m_pBlock->wasGenerated )
				{
					ska_free( m_pBlock->ska );
					m_pBlock->ska = NULL;

					m_pBlock->friendlyName[0] = 0;
					m_pBlock->rdnSubject.freeList();
					m_pBlock->path[0] = '\0';
				}

				m_pBlock->wasGenerated = false;
				UpdateControls();
				break;
			case IDC_btBrowse:
				if ( OnBrowseKeyFile() )
					m_pBlock->wasGenerated = false;
				break;
			case IDC_chkCreateSecretKeyPacket:
				m_pBlock->generateSecKey = Button_GetCheck( GetDlgItem( m_hDlg, IDC_chkCreateSecretKeyPacket ) ) == BST_CHECKED;
				m_pBlock->wasGenerated = false;
				UpdateControls();
				break;
			case IDC_chkCreatePublicKeyPacket:
				m_pBlock->generateCertificate = Button_GetCheck( GetDlgItem( m_hDlg, IDC_chkCreatePublicKeyPacket ) ) == BST_CHECKED;
				m_pBlock->wasGenerated = false;
				UpdateControls();
				break;

			default:
				return false;
		}
		return true;
	}
}

static int rsaCreateSecretKeyFromASN1Buffer( 
		HPKA *hSka,
		const BYTE *ptr,
		int size
	)
{
	x509_privateKeyInfo pki;
	pki.privateKey.B = (BYTE*)ptr;
	pki.privateKey.L = size;

	tUnit tu = {0};
	tu.pAlg = pgppkaRSA;

	pki.algOID.tu = &tu;

	return ska_createAlgorithm_X509( hSka, &pki );
}

bool cWizardDlg_Welcome::ReadSecretKeyFile( wchar_t *path )
{
	cFileView cfv( path );

	if ( cfv.m_err )
		return false;

	BYTE *ptr = cfv.m_ptr;
	int size = (int)cfv.m_size.QuadPart;

	if ( size != cfv.m_size.QuadPart ) // Overflow..
	{ ODS_fls( L"Assertion!" );	return false; }

	type_kff kff = getKeyFileFormat( path );

	cPTB b64ptb;
	if ( readArmoredBlock( &b64ptb, (const char *)ptr, size, "-----BEGIN ENCRYPTED PRIVATE KEY-----", "-----END ENCRYPTED PRIVATE KEY-----" ) )
	{
		ptr = b64ptb.buffer;
		size = b64ptb.bufferSize;
		kff = kffX509;
	}

	cPlainTextBlock ptbKey;
	cWCStr wcsFriendlyName;
	int res = ppe_ok;

	cUniversalString usFriendlyName;

	if ( kff == kffPEM )
	{
		cPTB ptb;
		cPTB ptbDecodedKey;

		const char *armorHeaders;

		if ( cPEMCtx::decodeBase64Block( ptb, ptr, size, &armorHeaders ) != ppe_ok  )
			return false;
		
		cPEMCtx ctx;

		if ( ctx.parseHeaders( armorHeaders ) != ppe_ok )
			return false;

		BYTE *ptr = ptb.buffer;
		int size = ptb.bufferSize;

		if ( !ctx.isPlainText() )
		{
			class cPasswordDlgLoop_CtxDecrypt : public cPasswordDlgLoop
			{
				cPTB &ptbDecodedKey;
				cPEMCtx &ctx;
				BYTE *ptr;
				int size;

			public:
				cPasswordDlgLoop_CtxDecrypt(
					const wchar_t *dlgText, int IDText, int IDPwd, const wchar_t *caption,
					int idsInvalidPassword, int idsErrorInProcessing,
					cPEMCtx &_ctx, cPTB &_ptbDecodedKey, BYTE *_ptr, int _size

					)
					: cPasswordDlgLoop( dlgText, IDText, IDPwd, caption, idsInvalidPassword, idsErrorInProcessing ),
					  ptbDecodedKey( _ptbDecodedKey ),
					  ctx( _ctx )
				{
					ptbDecodedKey = _ptbDecodedKey;
					ptr = _ptr;
					size = _size;
				}

				virtual int pwdActor()
				{
					int res = ctx.decrypt( ptbDecodedKey, m_pwd, ptr, size );
					switch ( res )
					{
						case ppe_ok : return errNo;
						case ppe_invalidPassword: return errInvalidPassword;
						default: return errInProcessing;
					}
				}

			} pwDlg(
				_FormatT( L"File %s", path + splitPath( path ) ), IDC_Text, IDC_Password, _LoadIDS( IDS_pwdDlgCaption ),
				IDS_InvalidPassword, IDS_ErrorInReadingKeyPacket,
				ctx, ptbDecodedKey, ptr, size );

			if ( pwDlg.DoModalLoop( IH.m_hInst, IDD_DlgPassword, NULL ) != IDOK )
				return false;

			ptr = ptbDecodedKey.buffer;
			size = ptbDecodedKey.bufferSize;
		}

		if ( m_pBlock->ska != NULL )
		{
			ska_free( m_pBlock->ska );
			m_pBlock->ska = NULL;
		}

		if ( rsaCreateSecretKeyFromASN1Buffer( &m_pBlock->ska, ptr, size ) != ppe_ok )
			return false;
	}
	else
	if ( kff == kffX509 )
	{
		if ( m_pBlock->ska != NULL )
		{
			ska_free( m_pBlock->ska );
			m_pBlock->ska = NULL;
		}

		if ( !BCDB_GetP12SecretKeyInfo( m_hInst, m_hDlg, ptr, size, &m_pBlock->ska, &usFriendlyName ) )
			return false;
	}
	else
	if ( kff == kffPGP )
	{
		cPTB ptbKeyRing;

		res = BrowseKeyRingForUserT( m_hInst, m_hDlg, &ptbKey, path, path, true, &wcsFriendlyName, &ptbKeyRing ); // ptbFriendlyName );
		if ( res != IDOK )
			return false;

		if ( !BCDB_GetPGPSecretKeyInfo( m_hInst, m_hDlg, ptbKey.buffer, ptbKey.bufferSize, &m_pBlock->ska, wcsFriendlyName.m_szBody ) ) // ( const char * )ptbFriendlyName.buffer ) )
			return false;

		usFriendlyName.Set( wcsFriendlyName.m_szBody, wcsFriendlyName.m_len );
		usFriendlyName.GenerateUnicodeString();
		res = 0;
	}
	else
	if ( kff == kffPGPAscii ) // Version 1.03.5 - 29.12.2010
	{
		const char *szPGP_B64_Begin = "-----BEGIN PGP PRIVATE KEY BLOCK-----";
		const char *szPGP_B64_End   = "-----END PGP PRIVATE KEY BLOCK-----";
		
		if ( !readArmoredBlock( &ptbKey, (const char *)ptr, size, szPGP_B64_Begin, szPGP_B64_End ) )
			return false;

		int size = 0;

		int res = pgp_EnumKeyIDs( NULL, &size, ptbKey.buffer, ptbKey.bufferSize, true );

		cPlainTextBlock keyIDtbl;

		if ( res == ppe_bufferTooSmall )
		{
			if ( keyIDtbl.Allocate( size ) == NULL )
				return false;

			res = pgp_EnumKeyIDs( keyIDtbl.buffer, &size, ptbKey.buffer, ptbKey.bufferSize, true );
		}

		if ( res != ppe_ok )
			return false;

		cpgp_keyInfo keyInfo;
		cpgp_keyInfo *pKeyInfo = NULL;
		
		BYTE *keyID = keyIDtbl.buffer;

		res = pgp_FindKeyByIDEx(
					&keyInfo,
					keyID,
					ptbKey.buffer, ptbKey.bufferSize,
					true,
					&pKeyInfo 
				);

		if ( res != ppe_ok )
			return false;

		if ( pKeyInfo == NULL )
			pKeyInfo = &keyInfo; // Master key used

		res = IsEncryptionKey( (sPublicKey *)pKeyInfo->keyPkt, pKeyInfo->keyPktSize );

		if ( res == ppe_invalidAlgorithmID )
		{
			// Checks if has one encryption subkey only.
			if ( ( pKeyInfo->next == NULL ) ||
				 ( ppe_ok != IsEncryptionKey( (sPublicKey *)pKeyInfo->next->keyPkt, pKeyInfo->next->keyPktSize ) ) ||
				 ( pKeyInfo->next->next != NULL )
				 )
			{
				DlgMessageBoxT( IDS_SignatureAlgForEncryption, MB_ICONSTOP );
				return false;
			}
			else
			{
				pKeyInfo = pKeyInfo->next;
			}
		}
		
		if ( pKeyInfo->userID != NULL )
			wcsFriendlyName.Set( (char *)pKeyInfo->userID->body, pKeyInfo->userID->size, CP_UTF8 );

		if ( !BCDB_GetPGPSecretKeyInfo( m_hInst, m_hDlg, pKeyInfo->keyPkt, pKeyInfo->keyPktSize, &m_pBlock->ska, wcsFriendlyName.m_szBody ) ) // ( const char * )ptbFriendlyName.buffer ) )
			return false;

		usFriendlyName.Set( wcsFriendlyName.m_szBody, wcsFriendlyName.m_len );
		usFriendlyName.GenerateUnicodeString();
	}
	else
	{
		DlgMessageBoxT( IDS_UndefinedKeyFileFormat, MB_ICONSTOP );
		return false;
	}


	usFriendlyName.unicodeBuff2wstr( m_pBlock->friendlyName, sizeof( m_pBlock->friendlyName ) );
	
	//
	//	We may bravely fills subject name by friendly name fields
	//
	m_pBlock->rdnSubject.freeList();
	friendlyNameToRDN( &usFriendlyName, &m_pBlock->rdnSubject );

	if ( res != ppe_ok )
		return false;

	wcscpy_s( m_pBlock->path, path );
	m_pBlock->keyBitSize = ska_getModuleSize( m_pBlock->ska ) * 8;
	return true;
}

bool cWizardDlg_Welcome::UpdateControls()
{
	m_pBlock->generateNew = Button_GetCheck( GetDlgItem( m_hDlg, IDC_rbGenerateNewKey ) ) == BST_CHECKED;

	//
	//	Force creation of secret key packet if new key is generated.
	//
	if ( m_pBlock->generateNew )
	{
		m_pBlock->generateSecKey = true;
		Button_SetCheck( GetDlgItem( m_hDlg, IDC_chkCreateSecretKeyPacket ), m_pBlock->generateSecKey ? BST_CHECKED : BST_UNCHECKED );
	}

	bool enable = m_pBlock->generateNew;

//	EnableWindow( GetDlgItem( m_hDlg, IDC_stAlgorithm ), enable );
//	EnableWindow( GetDlgItem( m_hDlg, IDC_stSize ), enable );
//	EnableWindow( GetDlgItem( m_hDlg, IDC_cmbAlgorithm ), enable );
//	EnableWindow( GetDlgItem( m_hDlg, IDC_edtLength ), enable );
//	EnableWindow( GetDlgItem( m_hDlg, IDC_spinSize ), enable );

	enable = !enable;

	EnableWindow( GetDlgItem( m_hDlg, IDC_edtPath ), enable );
	EnableWindow( GetDlgItem( m_hDlg, IDC_btBrowse ), enable );
	EnableWindow( GetDlgItem( m_hDlg, IDC_chkCreateSecretKeyPacket ), enable );
//	ShowWindow( GetDlgItem( m_hDlg, IDC_stUsedAlgorithm ), enable? SW_SHOW : SW_HIDE );
//	ShowWindow( GetDlgItem( m_hDlg, IDC_stUsedKeyID ), enable ? SW_SHOW : SW_HIDE );
//	ShowWindow( GetDlgItem( m_hDlg, IDC_usedAlgorithm ), enable ? SW_SHOW : SW_HIDE );
//	ShowWindow( GetDlgItem( m_hDlg, IDC_usedKeyID ), enable ? SW_SHOW : SW_HIDE );

	enable = ( m_pBlock->generateSecKey || m_pBlock->generateCertificate );
	EnableWindow( GetDlgItem( m_hDlg, IDC_btNext ), enable );

	if ( !m_pBlock->saveToFiles )
	{	m_pBlock->saveToBCStorage = true;
		Button_SetCheck( GetDlgItem( m_hDlg, IDC_chkSaveToBCStorage) , BST_CHECKED );
	}

	enable = m_pBlock->saveToFiles;
	EnableWindow( GetDlgItem( m_hDlg, IDC_chkSaveToBCStorage ), enable );

	// UpdateSTDButtons

	bool enableSKGeneration = m_pBlock->isSKReady();
	bool enablePKGeneration = m_pBlock->isPKReady();

	if ( m_pBlock->wasGenerated )
		EnableWindow( GetDlgItem( m_hDlg, IDOK ), false );
	else
	if ( m_pBlock->generateCertificate )
		EnableWindow( GetDlgItem( m_hDlg, IDOK ), enableSKGeneration &&  enablePKGeneration );
	else
		EnableWindow( GetDlgItem( m_hDlg, IDOK ), enableSKGeneration );

	return false;
}

static int bytes2string( char *str, int strSize, BYTE *bf, int bfSize )
{
	int len = 0;
	int pos = 0;

	for(; pos < bfSize; pos++ )
	{
		if ( len + 3 > strSize )
			break;

		sprintf( str + len, "%02X ", (DWORD)bf[pos] );
		len += 3;
	}

	return pos;
}

bool cWizardDlg_Welcome::OnBrowseKeyFile()
{
	wchar_t path[ MAX_PATH ];
	wcscpy_s( path, m_pBlock->path );

	//static const wchar_t *szFilters = 
	//	L"PKCS 12 ( *.p12 )\0*.p12\0"
	//	L"PGP secret keys ( *.skr,*.gpg,*.pgp )\0*.skr;*.gpg;*.pgp\0"
	//	L"All files ( *.* )\0*.*\0";

	wchar_t szFilters[ 1024 ];
	LoadStringT( hMainInst, IDS_szFiltersSecret, szFilters, maxLen( szFilters ) );

	wchar_t text[ 1024 ];
	LoadStringT( hMainInst, IDS_szChooseSecretKeyAndPressGet, text, maxLen( text ) );
	wcscat_s( text, _LoadIDS( IDS_TheSKWillBeUsedForNewPairGeneration ) );

	bool res = BrowseKeyFiles( 
			m_hDlg,
			path, maxLen( path ),
			szFilters,
			_LoadIDS( IDS_ChooseSecretKey ),
			text,
			_LoadIDS( IDS_szGetItBtnText )
		);

	if ( res )
		res = ReadSecretKeyFile( path );

	if ( res )
		SetExistentKeyInfo();
	else
	{
		//
		//	Returns back to generate new key position.
		//
		if ( m_pBlock->path[0] == '\0' )
		{
			m_pBlock->generateNew = true;
			Button_SetCheck( GetDlgItem( m_hDlg, IDC_rbUseExistingKey ), BST_UNCHECKED );
			Button_SetCheck( GetDlgItem( m_hDlg, IDC_rbGenerateNewKey ), BST_CHECKED );
		}
	}

	return res;
}

bool cWizardDlg_Welcome::OnNotify( NMHDR *hdr )
{
	switch( hdr->code )
	{
		case UDN_DELTAPOS:
			if ( hdr->idFrom == IDC_spinSize )
				return UDNDeltapos( (NMUPDOWN *)hdr, IDC_edtLength, 512, 8 * 1024, 512 );
		default : break;
	}

	return FALSE;
}

bool cWizardDlg_Welcome::SetExistentKeyInfo()
{
	SetDlgItemTextT( IDC_edtPath, m_pBlock->path );
/*
	if ( m_pBlock->ska != NULL )
	{
		int id = ska_getAlgID( m_pBlock->ska );
		int length = ska_getModuleSize( m_pBlock->ska );
		char buff[ 1024 ];
		sprintf( buff, "%s ( %d bits )", PKT_GetPublicKeyAlgName( id ), length * 8 );
		SetDlgItemText( m_hDlg, IDC_usedAlgorithm, buff );

		int size = 0;
		cPlainTextBlock ptbAuthKeyID;
		int ires = ska_generateLocalKeyID( m_pBlock->ska, NULL, &size );

		if ( ires != x509e_bufferTooSmall ) { assert( false ); }
		else
		{
			if ( ptbAuthKeyID.Allocate( size ) == NULL ) return false;
			ires = ska_generateLocalKeyID( m_pBlock->ska, ptbAuthKeyID.buffer, &size );

		    if ( ires != x509e_ok ) return false;
			bytes2string( buff, sizeof( buff ), ptbAuthKeyID.buffer, size );

			SetDlgItemText( m_hDlg, IDC_usedKeyID, buff );
		}
	}
	else
	{
		SetDlgItemText( m_hDlg, IDC_usedAlgorithm, "" );
		SetDlgItemText( m_hDlg, IDC_usedKeyID, "" );
	}
*/
	return true;
}