//
// Network Client library.
//
// Copyright (c) 2001 Jetico, Inc., Finland. All rights reserved.
//
// File: networkclient.h
// Revision:
// Created:		17.07.2001	
// Description: Network client high level function declaritions.
//

#ifndef __NetworkClient__
#define __NetworkClient__


#include "..\..\CertificateLib\crtLibErrs.h"

//
//	Network Cilent Errors : nce_
//
enum {
	nce_ok = 0,
	nce_bufferTooSmall,
	nce_memoryAllocation = ppe_last,
	nce_invalidConnection,
	nce_invalidServerAddress,
	nce_recieveError,
	nce_errorInWritting,
	nce_internalError,
	nce_cancelled
};

//
//	NC protocol types : ncpt_
//
enum
{
	ncpt_HTTP	= 0,
	ncpt_HTTPS	= 1,
	ncpt_LDAP	= 2,
	ncpt_LDAPS	= 3,
	ncpt_Unknown = -1
};

int getUrlType( const char *strType );
const char *getUrlStr( int type );

struct sURL
{
	//
	//	Protocol type, see enum above.
	//
	unsigned char pType;
	char *addr;
	char *dir;
	int  port;
	int  service;
	char *usr;
	char *pwd;

	sURL() { memset( this, 0, sizeof( *this ) ); }
	~sURL()
	{
		if ( addr != NULL ) delete addr;
		if ( dir  != NULL ) delete dir;
		if ( usr  != NULL ) delete usr;
		if ( pwd  != NULL ) delete pwd;

		memset( this, 0, sizeof( *this ) );
	}
};

//
//	createUrlStr
//		returns new allocated string if success, otherwise return NULL
//		caller function is reponsible for the string deleting
//
char *createUrlStr( sURL *url, const char *path = NULL );

struct sPGP_URL
{
	//
	//	Protocol type, see enum above.
	//
	unsigned char pType;
	char *addr;
	int  port;
	char *usr;
	char *pwd;

	char *getIndexPattern;  // Pattern to get indexes of specified key
	// http://wwwkeys.pgp.net:11371/pks/lookup?op=index&search=ser@iae.nsk.su
	char *getKeyPattern;	// Pattern to get key.
	// http://wwwkeys.pgp.net:11371/pks/lookup?op=get&search=0x630A6EC4
};

typedef sPGP_URL *sPGP_URL_ptr; 

extern sPGP_URL **pgpSrvs;

//
//	Handle to newtork client connection.
//
typedef void * HNCC; 

//
//	Function:
//		ncInit()
//	Purpouse: 
//		Allocates and initializes internal structures.
//		Use ncShutdown to free allocated data.
//	If successful returns nce_ok.
//
int ncInit(
		HNCC *hc // pointer to connection handler
	);

//
//	Function:
//		ncConnect()
//	Purpouse: 
//		Establishes connection with remote server. 
//	Returns nce_invalidConnection or nce_invalidServerAddress if no connection established.
//	If successful returns nce_ok.
//

int ncConnect(
		HNCC hc, // connection handler
		sURL *dstSrv, // Destination server URL
		sURL *proxySrv = 0 // Proxy server URL
	);

//
//	Function:
//		ncGetCertificate()
//	Purpouse: 
//		Gets server certificate for SSL connection ( HTTPS or LDAPS protoco types ).
//		Validity of certificate has to be checked by caller.
//	If successful returns nce_ok.
//
int ncGetCertificate(
		HNCC hc, // connection handler
		unsigned char *crtBuff, // Allocated buffer for certificate.
		int *pCrtBuffSize	// [IN] Pointer to size of allocated buffer.
							//	If allocated size is less then required 
							//	returns nce_bufferTooSmall
							// [OUT] Really written size,
	);


//
//	Handle to NC write context
//
typedef void *HNCWCTX;

//
//	Definitions of NC write function
//
typedef int ncWriteFuncType( HNCWCTX hctx, unsigned char *buff, int buffSize );


//
//	Function:
//		ncGetData()
//	Purpouse: 
//		Sends data request defined by path.
//	If successful returns nce_ok.
//
int ncGetData(
		HNCC hc,					// connection handler
		const char *path,					// String with path or script on destination server.
		ncWriteFuncType *ncWrite,	// Writing function, see typedefs above,
		HNCWCTX hwctx,				// The write context, it is used with ncWrite( hwctx, buff, size )
		bool *pCancelled
	);

//
//	Function:
//		ncShutdown()
//	Purpouse: 
//		Clears and frees all allocated data. 
//	Returns nce_ok.
//
int ncShutDown( HNCC hc );

#endif // __NetworkClient__