/////////////////////////////////////////////////////////////////
//
//      Copyright (c) 2017 Jetico, Inc., Finland
//      All rights reserved.
//
//      File: camellia.h
//      Revision:
//      Created:	22.07.2017
//      Description:
//			RFC 3713             Camellia Encryption Algorithm            April 2004
//		Realisation:
//			Calculations are based on 64bit limbs ( effective for 64bit system )
//
/////////////////////////////////////////////////////////////////

#pragma once

#define SERVICE_NAME			"CAMELLIA"
#define DISPLAY_NAME			"CAMELLIA-256"
#define ALGORITHM_ID			0xCA // 202
#define BLOCK_SIZE				0x10
#define MAXIMUM_KEYSIZE_BYTES	(256 / 8)

#include "bcintsize.h"

#ifdef  __cplusplus
extern "C" {
#endif

typedef struct _sKeyCtx sKeyCtx;

enum eCamelliaError {
	eCamelliaError_ok = 0,
	eCamelliaError_invalidKeySize,
	eCamelliaError_internal,
	eCamelliaError_memoryAllocation,
	eCamelliaError_test,
	eCamelliaError_last
};

int Camellia_GenerateKey( sKeyCtx **pCtx, const bc_u8 *key, const bc_u32 keyBitSize );
int Camellia_FreeKey( sKeyCtx **pCtx ); 

void Camellia_Encrypt_ECB( const sKeyCtx *ctx, bc_u8 *dst, const bc_u8 *src );
void Camellia_Decrypt_ECB( const sKeyCtx *ctx, bc_u8 *dst, const bc_u8 *src );

int Camellia_SelfTest();

#ifdef  __cplusplus
}
#endif