#include "cAlgCamellia.h"
#include "Camellia/camellia.h"

const char *cAlgCamellia::getDisplayName()
{
	return DISPLAY_NAME;
}

cAlgCamellia::cAlgCamellia()
{
	m_ctx = NULL;
	m_keySize = 0;
}


cAlgCamellia::~cAlgCamellia()
{
	if ( m_ctx != NULL )
		Camellia_FreeKey( (sKeyCtx **)(&m_ctx) );
}


BOOL cAlgCamellia::getAlgConstants( int *blockSize, int *keySize )
{
	*blockSize = BLOCK_SIZE;
	*keySize   = m_keySize;
	return TRUE;
}


int cAlgCamellia::getBlockSize()
{
	return BLOCK_SIZE;
}

int cAlgCamellia::getKeySize( int *pMax )
{
	if (pMax)
		*pMax = MAXIMUM_KEYSIZE_BYTES;

	return MAXIMUM_KEYSIZE_BYTES;
}

cSymAlg *cAlgCamellia::_createAlg( cAlgCamellia *alg, BYTE *key, int keySize ) // keySize in bytes
{
	if ( alg == NULL )
		return NULL;

	int res = Camellia_GenerateKey( (sKeyCtx **)&(alg->m_ctx), key, keySize * 8 );
	if ( res != eCamelliaError_ok )
		return NULL;

	alg->m_keySize = keySize;

	return alg;
}

cSymAlg *cAlgCamellia::createAlg( BYTE *key, int keySize )
{
	return _createAlg( new cAlgCamellia(), key, keySize );
}

cSymAlg *cAlgCamellia128::createAlg( BYTE *key, int keySize )
{
	return _createAlg( new cAlgCamellia128(), key, keySize );
}

cSymAlg *cAlgCamellia192::createAlg( BYTE *key, int keySize )
{
	return _createAlg( new cAlgCamellia192(), key, keySize );
}

int cAlgCamellia::encryptECB(
		const BYTE *inBf,		// Input plaintext buffer. 
		BYTE *outBf,	// Output encrypted buffer
		int  bfSize		// In/out buffers' size.
		)
{
	int i;

	if ((bfSize % BLOCK_SIZE) != 0)
		return sae_invalidBlockSize;

	for( i = 0; i < bfSize; i += BLOCK_SIZE)
	{
		Camellia_Encrypt_ECB( (sKeyCtx*)m_ctx, &(outBf[i]), &(inBf[i]) );
	}

	return sae_ok;
}

int cAlgCamellia::decryptECB(
		const BYTE *inBf,		// Input encrypted buffer.
		BYTE *outBf,	// Output plaintext buffer.
		int  bfSize		// In/out buffers' size.
		)
{
	int i;

	if ((bfSize % BLOCK_SIZE) != 0)
		return sae_invalidBlockSize;

	for( i = 0; i < bfSize; i += BLOCK_SIZE)
	{
		Camellia_Decrypt_ECB( (sKeyCtx*)m_ctx, &(outBf[i]), &(inBf[i]) );
	}

	return sae_ok;
}

