/*---------------------------------------------------------------------------*\

      Copyright (c) 2018 Jetico, Inc., Finland
      All rights reserved.

      File: secretKey.h
      Revision:
      Created:
      Description:

\*---------------------------------------------------------------------------*/

#pragma once

#include "../PGPPkts/PGPPkts.h"
#include "../publicKeyAlgs/publicKeyAlgs.h"

struct sSkaItem
{
	HPKA m_ska;
	const char *m_name;
	BYTE m_pgpKeyID[ pgp_keyIDSize ];

	sSkaItem()
	{
		m_ska = NULL;
		m_name = NULL;
	}

	const char *getName()
	{
		if ( m_name == NULL )
			return "";

		return m_name;
	}

	const char *setName( const char *name )
	{
		if ( m_name != NULL )
		{
			delete m_name;
			m_name = NULL;
		}

		size_t len = ( name == NULL ? 0 : strlen( name ) );

		if ( len == 0 )
			return "";

		char *newName = new char[ len + 1 ];
		if ( newName == NULL )
			return "";

		strncpy( newName, name, len + 1 );
		
		m_name = newName;

		return m_name;
	}

	~sSkaItem()
	{
		Free();
	}

	void Free()
	{
		if ( m_ska != NULL )
			ska_free( m_ska );
		
		m_ska = NULL;

		setName( NULL );

		memset( m_pgpKeyID, 0, sizeof( m_pgpKeyID ) );
	}
};

class cSecretKey
{
	cPTB ptb;

public:

	sSkaItem **skaT;
	int skaQ;

	cSecretKey()
	{
		skaT = NULL;
		skaQ = 0;
	}

	bool findKey( BYTE *keyID, int *pKeyN = NULL )
	{
		int i = skaQ;

		while( i-- )
		{
			if ( memcmp( skaT[i]->m_pgpKeyID, keyID, sizeof( skaT[i]->m_pgpKeyID ) ) == 0 )
				break;
		}

		if ( pKeyN != NULL )
			*pKeyN = i;

		return i >= 0;
	}

	void AddKey( HPKA ska, const char *keyName, BYTE *keyID )
	{
		sSkaItem *si = new sSkaItem();

		if ( si == NULL )
		{
			assert( false );
			return;
		}

		si->m_ska = ska;
		si->setName( keyName );

		if ( keyID != NULL )
			memcpy( si->m_pgpKeyID, keyID, sizeof( si->m_pgpKeyID ) );
		else
			memset( si->m_pgpKeyID, 0, sizeof( si->m_pgpKeyID ) );

		skaT = ( sSkaItem ** )ptb.ReAllocate( ( skaQ + 1 )*sizeof( sSkaItem * ) );
		if ( skaT == NULL )
			skaQ = 0;
		else 
		{
			skaT[skaQ++] = si;
		}
	}
/*
	enum type_kff {
		kffNone = 0,
		kffX509 = 1,
		kffPGP = 2,
		kffPGPAscii = 3
	};

	bool readKey( BYTE *pkt, int totalPktSize, type_kff kff, const char *keyName, BYTE *pgpKeyId )
	{
		bool ok = false;

		if ( kff == kffX509 )
			return readKeyX509( pkt, totalPktSize, keyName );

		return readKeyPGP( pkt, totalPktSize, keyName, pgpKeyId );
	}

	bool readKeyX509( BYTE *pkt, int totalPktSize, const char *keyName, cPassword *pPwd = NULL )
	{
		HPKA ska = NULL;

		bool ok = getP12SecretKeyInfo( IH.m_hInst, IH.m_hMainWnd, pkt, totalPktSize, &ska, NULL, NULL, NULL, 0, pPwd );

		if ( ok )
			AddKey( ska, keyName );

		return ok;
	}

	bool readKeyPGP( BYTE *pkt, int totalPktSize, const char *keyName, BYTE *pgpKeyId )
	{
		HPKA ska = NULL;

		if ( !GetPGPSecretKeyInfo( IH.m_hInst, IH.m_hMainWnd, pkt, totalPktSize, &ska, _getWcs( keyName ) ) )
		{
			cpgp_keyInfo keyInfo;
			cpgp_keyInfo *subKeyInfo = NULL;

			int res = pgp_FindKeyByIDEx(
						&keyInfo,
						keyID,
						pkt, totalPktSize,
						true,
						&subKeyInfo );

			if ( res != ppe_ok )
				return false;

			cpgp_keyInfo *ki = &keyInfo;
			if ( subKeyInfo != NULL ) // Subkey is found!
				ki = subKeyInfo;
			
			if ( !GetPGPSecretKeyInfo( IH.m_hInst, IH.m_hMainWnd, (BYTE*)ki->keyPkt, ki->keyPktSize, &ska, _getWcs( keyName ) ) )
				return false;
		}

		AddKey( ska, keyName, keyID );

		return true;
	}
*/

	bool openSecretKey( BYTE *secKeyID )
	{
		for( int i = skaQ; i--; )
		{
			if ( memcmp( skaT[i]->m_pgpKeyID, secKeyID, pgp_keyIDSize ) == 0 )
				return true;
		}

		return false;
	}
	
	void Close()
	{

		for ( int i = skaQ; i--; )
		{
			delete( skaT[i] );
		}

		skaT = NULL;
	}

	~cSecretKey()
	{
		Close();
	}

	//
	//	GetSecKeyByPassword
	//		Enumerates keys in BCDB.
	//		If the list is not empty gets password and tries to implement it to enuerated keys.
	//
	//bool GetSecKeyByPassword( void *pwdParam ); // HWND hWnd );
};
