#include <windows.h>
#include "AdvAPI.h"
#include "token_info.h"
#include <assert.h>

#include "ODS.h"
#ifdef _ODS
#include "UnicodeFuncs.h"
#endif


#include "../Setups/SetupIncludesT/cDriverUninstaller.h"
//#include "FileView.h"

//CAdvAPIDll advapi;

static const wchar_t *dllName = L"advapi32.dll";


bool CAdvAPIDll::Load( bool showError )
{
	//ODSFunc;

	if ( m_loaded )
	{
		return !error;
	}

	m_loaded = true;

	if ( !CDll::Load( false, dllName, NULL, showError ) )
	{
		ODS_fl;
		return false;
	}

	error = errGetProcAddress;

	m_intErr = 1000;

	if ( !GetProcAddress( "OpenSCManagerA", (void**)&procOpenSCManager ) )   { ODS_fl; return false; }

	if ( !GetProcAddress( "CreateServiceA", (void**)&procCreateService   ) )   { ODS_fl; return false; }
	if ( !GetProcAddress( "StartServiceA", (void**)&procStartService    ) )    { ODS_fl; return false; }
	if ( !GetProcAddress( "OpenServiceA", (void**)&procOpenService     ) )     { ODS_fl; return false; }
	if ( !GetProcAddress( "SetEntriesInAclA", (void**)&procSetEntriesInAcl ) ) { ODS_fl; return false; }

	if ( !GetProcAddress( "CloseServiceHandle", (void**)&procCloseServiceHandle ) ) { ODS_fl; return false; }
	if ( !GetProcAddress( "AllocateAndInitializeSid", (void**)&procAllocateAndInitializeSid ) ) { ODS_fl; return false; }
	if ( !GetProcAddress( "InitializeSecurityDescriptor", (void**)&procInitializeSecurityDescriptor ) ) { ODS_fl; return false; }

	if ( !GetProcAddress( "SetSecurityDescriptorDacl", (void**)&procSetSecurityDescriptorDacl ) ) { ODS_fl; return false; }

	if ( !GetProcAddress( "DeleteServiceA", (void**)&procDeleteService ) )
		GetProcAddress( "DeleteService", (void**)&procDeleteService );

	if ( !GetProcAddress( "ControlServiceA", (void**)&procControlService ) )
		GetProcAddress( "ControlService", (void**)&procControlService );
	
	GetProcAddress( "QueryServiceStatus", (void**)&procQueryServiceStatus );
	GetProcAddress( "OpenSCManagerW", (void**)&procOpenSCManagerW   );
	GetProcAddress( "CreateServiceW", (void**)&procCreateServiceW   );
	GetProcAddress( "StartServiceW", (void**)&procStartServiceW    );
	GetProcAddress( "OpenServiceW", (void**)&procOpenServiceW     );
	GetProcAddress( "SetEntriesInAclW", (void**)&procSetEntriesInAclW );

	GetProcAddress( "RegDeleteKeyExA", (void**)&procRegDeleteKeyEx );
	GetProcAddress( "RegDeleteKeyExW", (void**)&procRegDeleteKeyExW );

	GetProcAddress( "RevertToSelf", (void**)&procRevertToSelf );

	//
	//	Different version of avdapi32.dll have functions with 'A' and without
	//
	if (!GetProcAddress( "LookupPrivilegeValueA", (void**)&procLookupPrivilegeValue ) )
		GetProcAddress( "LookupPrivilegeValue", (void**)&procLookupPrivilegeValue );

	GetProcAddress( "LookupPrivilegeValueW", (void**)&procLookupPrivilegeValueW );

	GetProcAddress( "LogonUserW", (void**)&procLogonUserW );
	GetProcAddress( "LogonUserA", (void**)&procLogonUserA );

	GetProcAddress( "ImpersonateNamedPipeClient", (void**)&procImpersonateNamedPipeClient	);
	GetProcAddress( "ImpersonateLoggedOnUser", (void**)&procImpersonateLoggedOnUser		);
	GetProcAddress( "CreateProcessWithLogonW", (void**)&procCreateProcessWithLogonW		);
	
	GetProcAddress( "QueryServiceConfigW", (void**)&procQueryServiceConfig			);
	GetProcAddress( "ChangeServiceConfigW", (void**)&procChangeServiceConfig		);

	GetProcAddress( "OpenProcessToken", (void**)&procOpenProcessToken			);
	GetProcAddress( "GetTokenInformation", (void**)&procGetTokenInformation		);
	GetProcAddress( "LogonUserExW", (void**)&procLogonUserEx					);

	m_hScManager = procOpenSCManager(0, 0, SC_MANAGER_ALL_ACCESS);

	if ( m_hScManager == NULL ) 
	{
		m_intErr = ( 0x80000000 | GetLastError() );
	//	ODS_fls( _FormatT( L"Error in opening SCManager %d", m_intErr ) );
	//	return false; 
	}

	error = errNo;
	m_intErr = -2;

	return error == errNo;
}

BOOL CAdvAPIDll::LogonUserEx( LPCWSTR lpszUsername, LPCWSTR lpszDomain, LPCWSTR lpszPassword, DWORD dwLogonType, DWORD dwLogonProvider, PHANDLE phToken, PSID  *ppLogonSid, PVOID *ppProfileBuffer, LPDWORD pdwProfileLength, PQUOTA_LIMITS pQuotaLimits )
{
	if ( !Load( m_showMessage ) || ( !checkFunction_FL( __FUNCTION__, __LINE__, (void*)procLogonUserEx, m_showMessage ) ) )
		return FALSE;

	return procLogonUserEx( lpszUsername, lpszDomain, lpszPassword, dwLogonType, dwLogonProvider, phToken, ppLogonSid, ppProfileBuffer, pdwProfileLength, pQuotaLimits );
}

BOOL CAdvAPIDll::OpenProcessToken( IN HANDLE ProcessHandle, IN DWORD DesiredAccess, OUT PHANDLE pTokenHandle )
{
	if ( !Load( m_showMessage ) || ( !checkFunction_FL( __FUNCTION__, __LINE__, (void*)procOpenProcessToken, m_showMessage ) ) )
		return FALSE;

	return procOpenProcessToken( ProcessHandle, DesiredAccess, pTokenHandle );
}

BOOL CAdvAPIDll::GetTokenInformation( IN HANDLE TokenHandle, IN TOKEN_INFORMATION_CLASS TokenInformationClass, OUT LPVOID TokenInformation, IN DWORD TokenInformationLength, OUT PDWORD ReturnLength )
{
	if ( !Load( m_showMessage ) || ( !checkFunction_FL( __FUNCTION__, __LINE__, (void*)procGetTokenInformation, m_showMessage ) ) )
		return FALSE;

	return procGetTokenInformation( TokenHandle, TokenInformationClass, TokenInformation, TokenInformationLength, ReturnLength );
}

BOOL CAdvAPIDll::QueryServiceConfig( SC_HANDLE hService, LPQUERY_SERVICE_CONFIGW lpServiceConfig, DWORD cbBufSize, LPDWORD pcbBytesNeeded )
{
	if ( !Load( m_showMessage ) || ( !checkFunction_FL( __FUNCTION__, __LINE__, (void*)procQueryServiceConfig, m_showMessage ) ) )
		return FALSE;

	return procQueryServiceConfig( hService, lpServiceConfig, cbBufSize, pcbBytesNeeded );
}

BOOL CAdvAPIDll::ChangeServiceConfig(
    SC_HANDLE    hService,
    DWORD        dwServiceType,
    DWORD        dwStartType,
    DWORD        dwErrorControl,
    LPCWSTR     lpBinaryPathName,
    LPCWSTR     lpLoadOrderGroup,
    LPDWORD      lpdwTagId,
    LPCWSTR     lpDependencies,
    LPCWSTR     lpServiceStartName,
    LPCWSTR     lpPassword,
    LPCWSTR     lpDisplayName
    )
{
	if ( !Load( m_showMessage ) || ( !checkFunction_FL( __FUNCTION__, __LINE__, (void*)procChangeServiceConfig, m_showMessage ) ) )
		return FALSE;
	
	return procChangeServiceConfig( hService, dwServiceType, dwStartType, dwErrorControl, lpBinaryPathName, lpLoadOrderGroup, lpdwTagId, lpDependencies, lpServiceStartName, lpPassword, lpDisplayName );
}

BOOL CAdvAPIDll::ChangeServiceConfig( SC_HANDLE hService, LPQUERY_SERVICE_CONFIGW lpc )
{
	return ChangeServiceConfig( hService,
				lpc->dwServiceType,
				lpc->dwStartType,
				SERVICE_ERROR_NORMAL,
				lpc->lpBinaryPathName,
				lpc->lpLoadOrderGroup, 
				&(lpc->dwTagId),
				lpc->lpDependencies,
				lpc->lpServiceStartName,
				NULL,
				lpc->lpDisplayName );
}

CAdvAPIDll::CAdvAPIDll( bool loadNow, bool showError )
{
	memset( this, 0, sizeof( *this ) );

	error = TRUE;

	m_loaded = false;

	if ( loadNow )
		Load( showError );
}

CAdvAPIDll::~CAdvAPIDll()
{
	if ( m_hScManager ) 
		procCloseServiceHandle( m_hScManager );

	//if ( hInst )
	//	FreeLibrary( hInst );
}

BOOL CAdvAPIDll::RemoveDriver( IN LPCSTR DriverName )
{
	return RemoveDriver( getWCStr( DriverName ) );

	//if ( !CheckAdministratorPrivileges() )
	//return false;

	//SC_HANDLE  schService;
	//BOOL       ret;

	//schService = procOpenService( m_hScManager,
	//						  DriverName,
	//						  SERVICE_ALL_ACCESS
	//						  );

	//if ( schService == NULL )
	//	return FALSE;

	//ret = procDeleteService( schService );

	//procCloseServiceHandle( schService );

	//return ret;
}

BOOL CAdvAPIDll::RemoveDriver( IN LPCWSTR DriverName )
{
	if ( !CheckAdministratorPrivileges() )
		return false;

    SC_HANDLE  schService;
    BOOL       ret;

    schService = procOpenServiceW( m_hScManager,
                              DriverName,
                              SERVICE_ALL_ACCESS
                              );

    if ( schService == NULL )
        return FALSE;

    ret = procDeleteService( schService );
	if ( !ret )
	{
		int err = GetLastError();

		if ( err == ERROR_SERVICE_MARKED_FOR_DELETE )
		{
			//
			// V. 4.00.4 - 31.05.2012.
			// It is not error, the driver rather has been marked for deleting by UninstallDriverInternal ( see CAdvAPIDll::UninstallDriver );
			//

			ODS_fls( _FormatT( L"Error %s in DeleteService( %s )", L"ERROR_SERVICE_MARKED_FOR_DELETE", DriverName ) );
			ret = true;
		}
		else
			ODS_fls( _FormatT( L"Error %d in DeleteService( %s )", err, DriverName ) );
	}

    procCloseServiceHandle( schService );

    return ret;
}

BOOL CAdvAPIDll::StartDriver( IN LPCWSTR serviceName )
{
	SC_HANDLE hService = procOpenServiceW( m_hScManager, serviceName, SERVICE_ALL_ACCESS );

	BOOL ok = false;

	if ( hService == NULL )
	{
		int err = GetLastError();
		ODS_fls( _FormatT( L"Error %d in OpenService( %s )", err, serviceName ) );
	}
	else
	{
		ok = StartService( hService );

		if ( !ok )
		{
			int err = GetLastError();
			if ( err == ERROR_SERVICE_ALREADY_RUNNING )
				ok = true;
		
			if ( !ok )
			{
				ODS_fls( _FormatT( L"Error %d in StartService( %s )", err, serviceName ) );
				ok = true;
			}
		}
		
		procCloseServiceHandle( hService );
	}

	return ok;
}

BOOL CAdvAPIDll::StopDriver( IN LPCSTR DriverName )
{
	return StopDriver( getWCStr( DriverName ) );
}

class cSvcHandle
{
	SC_HANDLE	handle;
public:

	cSvcHandle(SC_HANDLE h) : handle(h)
	{
	}

	~cSvcHandle()
	{
		if ( handle != NULL )
			CAdvAPIDll::getStatic().CloseServiceHandle(handle);
	}

	operator SC_HANDLE ()
	{
		return handle;
	}
};

bool CAdvAPIDll::StopService( bool bWait, const wchar_t *serviceName, DWORD maxWait )
{
	ODSFunc;

	CAdvAPIDll &advapi = CAdvAPIDll::getStatic();

	if ( !advapi.Load( false ) )
		return false;

	cSvcHandle	hScm( advapi.procOpenSCManagerW( NULL, NULL, SC_MANAGER_ALL_ACCESS ) );
	if (NULL == hScm)
		return false;

	cSvcHandle	hSvc( advapi.procOpenServiceW( hScm, serviceName, SERVICE_ALL_ACCESS ) );
	if (NULL == hSvc)
	{
		DWORD dwErr = GetLastError();
		if ( dwErr == ERROR_SERVICE_DOES_NOT_EXIST )
			return true;

		ODS_fls( _FormatT( L"Error '%s' in Opening service", _GetSystemError( dwErr ) ) );
		return false;
	}

	BOOL	ret;
	SERVICE_STATUS	ss;

	ret = advapi.QueryServiceStatus(hSvc, &ss);
	if (!ret)
		return false;
	
	switch (ss.dwCurrentState)
	{
		case SERVICE_STOPPED:
		case SERVICE_STOP_PENDING:
			break;
		default:
			ret = advapi.ControlService(hSvc, SERVICE_CONTROL_STOP, &ss);
			if (!ret)
				return false;
	}

	if (!bWait)
		return true;

	DWORD period = 1000;
	int	retry = maxWait / period;

	while (SERVICE_STOPPED != ss.dwCurrentState)
	{
		Sleep(period);
		ret = advapi.QueryServiceStatus(hSvc, &ss);
		if (!ret)
			return false;

		if ( ( maxWait != INFINITE ) && (retry-- < 0) )
		{
			ODS_fls( _FormatT( L"Timeout in waiting service %s", serviceName ) );
			return false;
		}
	}

	return true;
}

BOOL CAdvAPIDll::StopDriver( IN LPCWSTR DriverName )
{
	DWORD dwErr = 0;
	BOOL ok = ControlService( DriverName, SERVICE_CONTROL_STOP, &dwErr );

	if ( ok )
		return true;

	if (
		( dwErr == ERROR_SERVICE_NOT_ACTIVE ) || // Is already stopped.
		( dwErr == ERROR_INVALID_SERVICE_CONTROL ) || // Was never installed.
		( dwErr == ERROR_SERVICE_DOES_NOT_EXIST )
		)
			return true;

	ODS_fls( _FormatT( L"Error '%s' in Stopping service %s", _GetSystemError(dwErr), DriverName ) );

	return ok;
}

BOOL CAdvAPIDll::ControlService( IN LPCWSTR DriverName, DWORD dwControl, DWORD *pErr )
{
	if ( !CheckAdministratorPrivileges() )
		return false;

	SC_HANDLE       schService;
	BOOL            ret;
	SERVICE_STATUS  serviceStatus;

	DWORD err;

	if ( pErr == NULL )
		pErr = &err;

	schService = procOpenServiceW( m_hScManager, DriverName, SERVICE_ALL_ACCESS );

	*pErr = GetLastError();

	if ( schService == NULL )
	{
		if ( *pErr == ERROR_SERVICE_DOES_NOT_EXIST )
			return true;

		ODS_fls( _FormatT( L"StopDriver:: Error in opening service ( err %d )", err ) );

		return false;
	}

	ret = procControlService( schService, dwControl, &serviceStatus );

	*pErr = GetLastError();

	procCloseServiceHandle( schService );

	return ret;
}

bool CAdvAPIDll::UninstallDriverInternal( const wchar_t *serviceName )
{
	// Realisation of driver uninstallig procedure - Uninstall v.3.06, 10.09.08.

	ODSFuncS( serviceName );

	ODSFuncRes( false );

	SC_HANDLE hsrv;
	SERVICE_STATUS status;
	BOOL rc;
	CDriverUninstaller du; // class code below

	if ( du.m_SetupapiDll.error )
	{
		ODS( L"Error in loading SetpAPI.dll" );
		return false;
	}

	hsrv = procOpenServiceW( m_hScManager, serviceName, SERVICE_ALL_ACCESS );

	ODS_fls( _FormatT( L"OpenServiceW %X", hsrv ) );

	if (!hsrv)
		return false;

	rc = procControlService( hsrv, SERVICE_CONTROL_STOP, &status); // stop service

	ODS_fls( _FormatT( L"ControlService %X", rc ) );

	bool ok = du.Uninstall( serviceName ) == TRUE; // cleanup PnP manager info

	BOOL OK = procDeleteService(hsrv); // delete service regardless of ControlService return code

	ODS_fls( _FormatT( L"DeleteService %X", OK ) );

	procCloseServiceHandle(hsrv);

	ODSFuncRes( ok );
	return ok;
}

bool CAdvAPIDll::UninstallDriver( const wchar_t *serviceName )
{
	ODSFuncS( serviceName );

	if ( !CheckAdministratorPrivileges() )
	{
		ODSFuncRes( L"No Administrator Privileges" );
		return false;
	}

	if ( UninstallDriverInternal( serviceName )  )
	{
		ODSFuncRes( true );
		return true;
	}

	ODS( L"Stop and Remove Service" );

	StopDriver( serviceName );
	bool ok = RemoveDriver( serviceName ) == TRUE;

	ODSFuncRes( ok );

	return ok;
}

LONG CAdvAPIDll::RegDeleteKeyEx( HKEY hKey, LPCSTR lpSubKey, REGSAM samDesired, DWORD Reserved )
{
	if ( Load( m_showMessage ) && ( procRegDeleteKeyEx != NULL ) )
		return procRegDeleteKeyEx( hKey, lpSubKey, samDesired, Reserved );

	return RegDeleteKeyA( hKey, lpSubKey );
}

// static
bool CAdvAPIDll::CheckAdministratorPrivileges( bool bForced )
{
	ODSFunc;

	static bool isNT = ( GetVersion() < 0x80000000 );

	if ( !isNT )
		return true;
	
	CAdvAPIDll &dll = getStatic();

	if ( !dll.Load( false ) )
		return false;

	if ( !dll.checkFunction_FL( __FUNCTION__, __LINE__, dll.procOpenSCManager, true ) )
		return false;

	bool ok = ( dll.m_hScManager != NULL );

	if ( !ok && bForced )
	{
		dll.m_hScManager = dll.procOpenSCManager(0, 0, SC_MANAGER_ALL_ACCESS);
		ODS_fls( _FormatT( L"bForced, m_hScManager = 0x%X", dll.m_hScManager ) );
		ok = ( dll.m_hScManager != NULL );
	}

	ODSFuncRes( ok );

	return ok;
}

LONG CAdvAPIDll::RegDeleteKeyExW( HKEY hKey, LPCWSTR lpSubKey, REGSAM samDesired, DWORD Reserved )
{
	ODSFuncS( _FormatT( L"%X, %s, %X", hKey, lpSubKey, samDesired ) );

	// WinXP do not have this function.
	LONG lRes;

	bool isNT = ( GetVersion() < 0x80000000 );

	if ( isNT && Load( m_showMessage ) && ( procRegDeleteKeyEx != NULL ) )
	{
		lRes = procRegDeleteKeyExW( hKey, lpSubKey, samDesired, Reserved );
		ODSFuncRes( (DWORD)lRes );
		return lRes;
	}

	cWCS parent, key;
	int pos = splitPath( lpSubKey, &parent, &key );
	if ( pos == 0 )
	{
		lRes = RegDeleteKeyT( hKey, lpSubKey );
		ODS_fls( _FormatT( L"RegDeleteKeyT( %s ) = %X", lpSubKey, (DWORD)lRes ) );
		return lRes;
	}

	HKEY hParent;
	lRes = RegOpenKeyExW( hKey, parent.m_szBody, 0, samDesired | KEY_SET_VALUE, &hParent );
	if ( lRes != ERROR_SUCCESS )
	{
		ODS_fls( _FormatT( L"RegOpenKeyExW( %s ) = %X", parent.m_szBody, (DWORD)lRes ) );
		return lRes;
	}

	lRes = RegDeleteKeyT( hParent, key.m_szBody );

	ODS_fls( _FormatT( L"RegDeleteKeyT( %s ) = %X", key.m_szBody, (DWORD)lRes ) );

	RegCloseKey( hParent );

	return lRes;
}

bool CAdvAPIDll::checkFunction_FL( const char *function, int line, void *ptr, bool showError )
{
	if ( ptr != NULL )
		return true;

//	assert( false );

	if ( showError )
	{
		cWCStr moduleName( MAX_PATH );
		GetModuleFileNameT( NULL, moduleName.m_szBody, moduleName.m_len );
		cWCStr appName(  moduleName.m_szBody + splitPath( moduleName.m_szBody ) );
		
		GetModuleFileNameT( hInst, moduleName.m_szBody, moduleName.m_len );

		cWCStr wcsBody;
		
		wcsBody.FormatT( 
			L"%s - Invalid function is detected!\n"
			L"Function: %s\n"
			L"Line: %d\n"
			L"%s\n"
			L"Dll path: %s",
			dllName,
			getWCStr( function ),
			line,
			IH.GetOSVersionStr(),
			moduleName.m_szBody
		);
	
		MessageBoxT( 0, wcsBody.m_szBody, appName.m_szBody, MB_ICONSTOP | MB_TOPMOST );
	}

	return false;
}

bool CAdvAPIDll::checkFunction( void *ptr, bool showError )
{
	if ( ptr != NULL )
		return true;

	//assert( false );

	if ( showError )
		MessageBoxT( 0, L"Invalid function!", dllName, MB_ICONSTOP | MB_TOPMOST );

	return false;
}

BOOL CAdvAPIDll::ImpersonateNamedPipeClient( IN HANDLE hNamedPipe )
{
	if ( !Load( m_showMessage ) || ( !checkFunction_FL( __FUNCTION__, __LINE__, (void*)procImpersonateNamedPipeClient, m_showMessage ) ) )
		return false;

	return procImpersonateNamedPipeClient( hNamedPipe );
}

BOOL CAdvAPIDll::ImpersonateLoggedOnUser( IN HANDLE  hToken )
{
	if ( !Load( m_showMessage ) || ( !checkFunction_FL( __FUNCTION__, __LINE__, (void*)procImpersonateLoggedOnUser, m_showMessage ) ) )
		return false;

	return procImpersonateLoggedOnUser( hToken );
}

BOOL CAdvAPIDll::CreateProcessWithLogonW(
      LPCWSTR lpUsername,
      LPCWSTR lpDomain,
      LPCWSTR lpPassword,
      DWORD   dwLogonFlags,
      LPCWSTR lpApplicationName,
      LPWSTR lpCommandLine,
      DWORD dwCreationFlags,
      LPVOID lpEnvironment,
      LPCWSTR lpCurrentDirectory,
      LPSTARTUPINFOW lpStartupInfo,
      LPPROCESS_INFORMATION lpProcessInformation
      )
{
	if ( !Load( m_showMessage ) || ( !checkFunction_FL( __FUNCTION__, __LINE__, (void*)procCreateProcessWithLogonW, m_showMessage ) ) )
		return false;

	return procCreateProcessWithLogonW( lpUsername, lpDomain, lpPassword, dwLogonFlags, lpApplicationName, lpCommandLine, dwCreationFlags, lpEnvironment, lpCurrentDirectory, lpStartupInfo, lpProcessInformation );
}

BOOL CAdvAPIDll::RevertToSelf()
{
	if ( !Load( m_showMessage ) || ( !checkFunction_FL( __FUNCTION__, __LINE__, (void*)procRevertToSelf, m_showMessage ) ) )
		return false;
	
	return procRevertToSelf();
}

BOOL CAdvAPIDll::LookupPrivilegeValue( IN LPCSTR lpSystemName, IN LPCSTR lpName, OUT PLUID lpLuid )
{
	if ( !Load( m_showMessage ) || ( !checkFunction_FL( __FUNCTION__, __LINE__, (void*)procLookupPrivilegeValue, m_showMessage ) ) )
		return NULL;
	
	return procLookupPrivilegeValue( lpSystemName, lpName, lpLuid );
}

BOOL CAdvAPIDll::LookupPrivilegeValue( IN LPCWSTR lpSystemName, IN LPCWSTR lpName, OUT PLUID lpLuid )
{
	if ( !Load( m_showMessage ) || ( !checkFunction_FL( __FUNCTION__, __LINE__, (void*)procLookupPrivilegeValueW, m_showMessage ) ) )
		return NULL;
	
	return procLookupPrivilegeValueW( lpSystemName, lpName, lpLuid );
}

BOOL CAdvAPIDll::LogonUser( 
		IN LPCWSTR lpszUsername,
		IN LPCWSTR lpszDomain,
		IN LPCWSTR lpszPassword,
		IN DWORD dwLogonType,
		IN DWORD dwLogonProvider,
		OUT PHANDLE phToken
	)
{
	if ( !Load( m_showMessage ) || ( !checkFunction_FL( __FUNCTION__, __LINE__, (void*)procLogonUserW, m_showMessage ) ) )
		return NULL;
	
	return procLogonUserW( lpszUsername, lpszDomain, lpszPassword, dwLogonType, dwLogonProvider, phToken );
}

BOOL CAdvAPIDll::LogonUser( 
		IN LPCSTR lpszUsername,
		IN LPCSTR lpszDomain,
		IN LPCSTR lpszPassword,
		IN DWORD dwLogonType,
		IN DWORD dwLogonProvider,
		OUT PHANDLE phToken
	)
{
	if ( !Load( m_showMessage ) || ( !checkFunction_FL( __FUNCTION__, __LINE__, (void*)procLogonUserA, m_showMessage ) ) )
		return NULL;
	
	return procLogonUserA( lpszUsername, lpszDomain, lpszPassword, dwLogonType, dwLogonProvider, phToken );
}

SC_HANDLE CAdvAPIDll::OpenSCManager( const wchar_t *lpMachineName, const wchar_t* lpDatabaseName, DWORD   dwDesiredAccess )
{
	if ( !Load( m_showMessage ) || ( !checkFunction_FL( __FUNCTION__, __LINE__, (void*)procOpenSCManagerW, m_showMessage ) ) )
		return NULL;
	
	return procOpenSCManagerW( lpMachineName, lpDatabaseName, dwDesiredAccess );
}

BOOL CAdvAPIDll::QueryServiceStatus( SC_HANDLE hService, LPSERVICE_STATUS lpServiceStatus )
{
	if ( !Load( m_showMessage ) || ( !checkFunction_FL( __FUNCTION__, __LINE__, (void*)procQueryServiceStatus, m_showMessage ) ) )
		return NULL;
	
	return procQueryServiceStatus( hService, lpServiceStatus );
}


SC_HANDLE CAdvAPIDll::OpenSCManager( const char *lpMachineName, const char* lpDatabaseName, DWORD   dwDesiredAccess )
{
	if ( !Load( m_showMessage ) || ( !checkFunction_FL( __FUNCTION__, __LINE__, (void*)procOpenSCManager, m_showMessage ) ) )
		return NULL;
	
	return procOpenSCManager( lpMachineName, lpDatabaseName, dwDesiredAccess );
}

BOOL CAdvAPIDll::CloseServiceHandle( SC_HANDLE   hSCObject )
{
	if ( !Load( m_showMessage ) || ( !checkFunction_FL( __FUNCTION__, __LINE__, (void*)procCloseServiceHandle, m_showMessage ) ) )
		return false;
	
	return procCloseServiceHandle( hSCObject );
}

SC_HANDLE CAdvAPIDll::CreateService( 
		SC_HANDLE    hSCManager,
		LPCWSTR     lpServiceName,
		LPCWSTR     lpDisplayName,
		DWORD        dwDesiredAccess,
		DWORD        dwServiceType,
		DWORD        dwStartType,
		DWORD        dwErrorControl,
		LPCWSTR     lpBinaryPathName,
		LPCWSTR     lpLoadOrderGroup,
		LPDWORD      lpdwTagId,
		LPCWSTR     lpDependencies,
		LPCWSTR     lpServiceStartName,
		LPCWSTR     lpPassword
		)
{
	if ( !Load( m_showMessage ) || ( !checkFunction_FL( __FUNCTION__, __LINE__, (void*)procCreateServiceW, m_showMessage ) ) )
		return false;
	
	return procCreateServiceW( hSCManager, lpServiceName, lpDisplayName, dwDesiredAccess, dwServiceType, dwStartType, dwErrorControl, lpBinaryPathName, lpLoadOrderGroup, lpdwTagId, lpDependencies, lpServiceStartName, lpPassword );
}

SC_HANDLE CAdvAPIDll::CreateService(
		SC_HANDLE    hSCManager,
		LPCSTR     lpServiceName,
		LPCSTR     lpDisplayName,
		DWORD        dwDesiredAccess,
		DWORD        dwServiceType,
		DWORD        dwStartType,
		DWORD        dwErrorControl,
		LPCSTR     lpBinaryPathName,
		LPCSTR     lpLoadOrderGroup,
		LPDWORD      lpdwTagId,
		LPCSTR     lpDependencies,
		LPCSTR     lpServiceStartName,
		LPCSTR     lpPassword
		)
{
	if ( !Load( m_showMessage ) || ( !checkFunction_FL( __FUNCTION__, __LINE__, (void*)procCreateService, m_showMessage ) ) )
		return NULL;
	
	return procCreateService( hSCManager, lpServiceName, lpDisplayName, dwDesiredAccess, dwServiceType, dwStartType, dwErrorControl, lpBinaryPathName, lpLoadOrderGroup, lpdwTagId, lpDependencies, lpServiceStartName, lpPassword );
}

BOOL CAdvAPIDll::StartService( SC_HANDLE hService, DWORD dwNumServiceArgs, LPCSTR *lpServiceArgVectors )
{
	if ( !Load( m_showMessage ) || ( !checkFunction_FL( __FUNCTION__, __LINE__, (void*)procStartService, m_showMessage ) ) )
	{
		ODS_fls( L"Assertion!" );
		return false;
	}
	
	if ( procStartService( hService, dwNumServiceArgs, lpServiceArgVectors ) )
		return true;

	ODS_fls( _FormatT( L"Assertion! Error %s in starting service %x", _GetSystemError(-1), hService ) );
	return false;
}

BOOL CAdvAPIDll::StartService( SC_HANDLE hService, DWORD dwNumServiceArgs, LPCWSTR *lpServiceArgVectors )
{
	if ( !Load( m_showMessage ) || ( !checkFunction_FL( __FUNCTION__, __LINE__, (void*)procStartServiceW, m_showMessage ) ) )
		return false;
	
	if ( procStartServiceW( hService, dwNumServiceArgs, lpServiceArgVectors ) )
		return true;

	m_dwLastErr = GetLastError();

	ODS_fls( _FormatT( L"Assertion! Error %s in starting service %x", _GetSystemError(m_dwLastErr), hService ) );
	return false;
}

SC_HANDLE CAdvAPIDll::OpenService( SC_HANDLE hSCManager, LPCSTR lpServiceName, DWORD dwDesiredAccess )
{
	if ( !Load( m_showMessage ) || ( !checkFunction_FL( __FUNCTION__, __LINE__, (void*)procOpenService, m_showMessage ) ) )
		return NULL;
	
	SC_HANDLE handle = procOpenService( hSCManager, lpServiceName, dwDesiredAccess );

	if ( handle == NULL )
		m_dwLastErr = GetLastError();

	return handle;
}

SC_HANDLE CAdvAPIDll::OpenService( SC_HANDLE hSCManager, LPCWSTR lpServiceName, DWORD dwDesiredAccess )
{
	if ( !Load( m_showMessage ) || ( !checkFunction_FL( __FUNCTION__, __LINE__, (void*)procOpenServiceW, m_showMessage ) ) )
		return NULL;
	
	SC_HANDLE handle = procOpenServiceW( hSCManager, lpServiceName, dwDesiredAccess );

	if ( handle == NULL )
		m_dwLastErr = GetLastError();

	return handle;
}

BOOL CAdvAPIDll::DeleteService( SC_HANDLE hService )
{
	if ( !Load( m_showMessage ) || ( !checkFunction_FL( __FUNCTION__, __LINE__, (void*)procDeleteService, m_showMessage ) ) )
		return false;
	
	return procDeleteService( hService );
}

BOOL CAdvAPIDll::ControlService( SC_HANDLE hService, DWORD dwControl, LPSERVICE_STATUS lpServiceStatus )
{
	if ( !Load( m_showMessage ) || ( !checkFunction_FL( __FUNCTION__, __LINE__, (void*)procControlService, m_showMessage ) ) )
		return false;
	
	return procControlService( hService, dwControl, lpServiceStatus );
}

BOOL CAdvAPIDll::AllocateAndInitializeSid(
		IN PSID_IDENTIFIER_AUTHORITY pIdentifierAuthority,
		IN BYTE nSubAuthorityCount,
		IN DWORD nSubAuthority0,
		IN DWORD nSubAuthority1,
		IN DWORD nSubAuthority2,
		IN DWORD nSubAuthority3,
		IN DWORD nSubAuthority4,
		IN DWORD nSubAuthority5,
		IN DWORD nSubAuthority6,
		IN DWORD nSubAuthority7,
		OUT PSID *pSid
	)
{
	if ( !Load( m_showMessage ) || ( !checkFunction_FL( __FUNCTION__, __LINE__, (void*)procAllocateAndInitializeSid, m_showMessage ) ) )
		return false;
	
	return procAllocateAndInitializeSid( 
			pIdentifierAuthority, nSubAuthorityCount, nSubAuthority0, nSubAuthority1, nSubAuthority2, nSubAuthority3, nSubAuthority4, nSubAuthority5, nSubAuthority6, nSubAuthority7, pSid );
}

DWORD CAdvAPIDll::SetEntriesInAcl( IN ULONG cCountOfExplicitEntries, IN PEXPLICIT_ACCESS_A pListOfExplicitEntries, IN PACL OldAcl, OUT PACL *NewAcl	)
{
	if ( !Load( m_showMessage ) || ( !checkFunction_FL( __FUNCTION__, __LINE__, (void*)procSetEntriesInAcl, m_showMessage ) ) )
		return -1;
	
	return procSetEntriesInAcl( cCountOfExplicitEntries, pListOfExplicitEntries, OldAcl, NewAcl );
}

DWORD CAdvAPIDll::SetEntriesInAcl( IN ULONG cCountOfExplicitEntries, IN PEXPLICIT_ACCESS_W pListOfExplicitEntries, IN PACL OldAcl, OUT PACL *NewAcl	)
{
	if ( !Load( m_showMessage ) || ( !checkFunction_FL( __FUNCTION__, __LINE__, (void*)procSetEntriesInAclW, m_showMessage ) ) )
		return -1;
	
	return procSetEntriesInAclW( cCountOfExplicitEntries, pListOfExplicitEntries, OldAcl, NewAcl );
}

BOOL CAdvAPIDll::InitializeSecurityDescriptor( OUT PSECURITY_DESCRIPTOR pSecurityDescriptor, IN DWORD dwRevision )
{
	if ( !Load( m_showMessage ) || ( !checkFunction_FL( __FUNCTION__, __LINE__, (void*)procInitializeSecurityDescriptor, m_showMessage ) ) )
		return false;
	
	return procInitializeSecurityDescriptor( pSecurityDescriptor, dwRevision );
}

BOOL CAdvAPIDll::SetSecurityDescriptorDacl( IN OUT PSECURITY_DESCRIPTOR pSecurityDescriptor, IN BOOL bDaclPresent, IN PACL pDacl, IN BOOL bDaclDefaulted )
{
	if ( !Load( m_showMessage ) || ( !checkFunction_FL( __FUNCTION__, __LINE__, (void*)procSetSecurityDescriptorDacl, m_showMessage ) ) )
		return false;
	
	return procSetSecurityDescriptorDacl( pSecurityDescriptor, bDaclPresent, pDacl, bDaclDefaulted );
}


BOOL CAdvAPIDll::ConvertSidToStringSid( IN  PSID Sid, OUT LPWSTR  *StringSid )
{
	if ( !Load( m_showMessage ) )
		return false;

	typedef WINADVAPI BOOL WINAPI typeConvertSidToStringSidW( IN  PSID Sid, OUT LPWSTR  *StringSid );

	typeConvertSidToStringSidW *proc = NULL;

	if ( !GetProcAddress( "ConvertSidToStringSidW", (void**)&proc ) )
		return false;

	return proc( Sid, StringSid );
}

BOOL CAdvAPIDll::ConvertStringSidToSid( LPCWSTR StringSid, PSID* Sid )
{
	if ( !Load( m_showMessage ) )
		return false;

	typedef WINADVAPI BOOL WINAPI typeConvertStringSidToSidW( LPCWSTR StringSid, PSID* Sid );

	typeConvertStringSidToSidW *proc = NULL;

	if ( !GetProcAddress( "ConvertStringSidToSidW", (void**)&proc ) )
		return false;

	return proc( StringSid, Sid );
}

PVOID CAdvAPIDll::FreeSid( PSID pSid )
{
	if ( !Load( m_showMessage ) ) return NULL;

	typedef WINADVAPI PVOID WINAPI typeFreeSid( __in PSID pSid );
	
	typeFreeSid *proc = NULL;

	if ( !GetProcAddress( "FreeSid", (void**)&proc ) )
		return NULL;

	return proc( pSid );
}

DWORD CAdvAPIDll::SetNamedSecurityInfo( LPCWSTR pObjectName, SE_OBJECT_TYPE ObjectType, SECURITY_INFORMATION SecurityInfo, PSID psidOwner, PSID psidGroup, PACL pDacl, PACL pSacl )
{
	if ( !Load( m_showMessage ) ) return (DWORD)-1;

	typedef WINADVAPI DWORD WINAPI typeSetNamedSecurityInfoW(
		__in LPCWSTR               pObjectName, // LPWSTR
		__in SE_OBJECT_TYPE        ObjectType,
		__in SECURITY_INFORMATION  SecurityInfo,
		__in_opt PSID              psidOwner,
		__in_opt PSID              psidGroup,
		__in_opt PACL              pDacl,
		__in_opt PACL              pSacl
	);

	typeSetNamedSecurityInfoW *proc = NULL;

	if ( !GetProcAddress( "SetNamedSecurityInfoW", (void**)&proc ) )
		return (DWORD)-1;

	return proc( pObjectName, ObjectType, SecurityInfo, psidOwner, psidGroup, pDacl, pSacl );
}

bool CAdvAPIDll::AdjustTokenPrivileges( HANDLE TokenHandle, BOOL DisableAllPrivileges, PTOKEN_PRIVILEGES NewState, DWORD BufferLength,
									   PTOKEN_PRIVILEGES PreviousState,  PDWORD ReturnLength  )
{
	if ( !Load( m_showMessage ) ) return false;

	typedef WINADVAPI
	BOOL
	WINAPI
	typeAdjustTokenPrivileges(
		__in      HANDLE TokenHandle,
		__in      BOOL DisableAllPrivileges,
		__in_opt  PTOKEN_PRIVILEGES NewState,
		__in      DWORD BufferLength,
		__out_bcount_part_opt(BufferLength, *ReturnLength) PTOKEN_PRIVILEGES PreviousState,
		__out_opt PDWORD ReturnLength
		);
	
	typeAdjustTokenPrivileges *proc = NULL;

	if ( !GetProcAddress( "AdjustTokenPrivileges", (void**)&proc ) )
		return false;

	return proc( TokenHandle, DisableAllPrivileges, NewState, BufferLength, PreviousState, ReturnLength ) == TRUE;
}


//-------------------------------------
// get service status
//
// return status code if successful
// -1 if not
//
// return codes:
//   SERVICE_STOPPED
//   SERVICE_RUNNING
//   SERVICE_PAUSED
//
// following return codes are used to indicate that the service is in the middle of getting to one
// of the above states:
//   SERVICE_START_PENDING
//   SERVICE_STOP_PENDING
//   SERVICE_CONTINUE_PENDING
//   SERVICE_PAUSE_PENDING
//
// sMachine:
//   machine name, ie: \SERVER
//   empty = local machine
//
// sService
//   service name, ie: "BCWipeSvc"
//
DWORD CAdvAPIDll::ServiceGetStatus( const wchar_t *sService, const wchar_t *sMachine  )
{
	CAdvAPIDll &advapi = CAdvAPIDll::getStatic();

	SC_HANDLE
		schm, // service control manager handle
		schs; // service handle

	SERVICE_STATUS ss; // service status
	DWORD dwStat = -1; // current service status

	// connect to the service control manager 
	
	if ( !advapi.Load( false ) )
		return dwStat;

	schm = advapi.OpenSCManager( sMachine, NULL, SC_MANAGER_CONNECT );

	// if successful...
	if( schm > 0 )
	{
		// open a handle to the specified service
		schs = advapi.OpenService( schm, sService, SERVICE_QUERY_STATUS );

		// if successful...
		if ( schs > 0 )
		{
			// retrieve the current status of the specified service    
			if( advapi.QueryServiceStatus( schs, &ss ) )
			{
				dwStat = ss.dwCurrentState;
			}

			// close service handle
			advapi.CloseServiceHandle( schs );
		}

		// close service control manager handle
		advapi.CloseServiceHandle( schm );
	}

	return dwStat;
}

//static 
bool CAdvAPIDll::SetPrivilege( const wchar_t *lpszPrivilege, bool bEnablePrivilege )
{
	HANDLE hToken;  

	// Get a token for this process. 

	if ( !CAdvAPIDll::getStatic().OpenProcessToken( GetCurrentProcess(), TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY, &hToken ) ) 
	{
		ODS_fls( L"Assertion! OpenProcessToken return error" );
		return false;
	}

	bool ok = CAdvAPIDll::getStatic().SetPrivilege( hToken, lpszPrivilege, bEnablePrivilege );

	CloseHandle( hToken );

	return ok;
}

bool CAdvAPIDll::SetPrivilege(
    HANDLE hToken,          // access token handle
    const wchar_t *lpszPrivilege,  // name of privilege to enable/disable
    bool bEnablePrivilege   // to enable or disable privilege
    ) 
{
	TOKEN_PRIVILEGES tp;
	LUID luid;

	if ( !LookupPrivilegeValue( 
			NULL,            // lookup privilege on local system
			lpszPrivilege,   // privilege to lookup 
			&luid ) )        // receives LUID of privilege
	{
		ODS_fls( _FormatT( L"LookupPrivilegeValue error: 0x%X", GetLastError() ) );
		return false; 
	}

	tp.PrivilegeCount = 1;
	tp.Privileges[0].Luid = luid;
	if (bEnablePrivilege)
		tp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;
	else
		tp.Privileges[0].Attributes = 0;

	// Enable the privilege or disable all privileges.

	if ( !AdjustTokenPrivileges(
		   hToken, 
		   false, 
		   &tp, 
		   sizeof(TOKEN_PRIVILEGES), 
		   (PTOKEN_PRIVILEGES) NULL, 
		   (PDWORD) NULL) )
	{ 
		  ODS_fls( _FormatT( L"AdjustTokenPrivileges error: 0x%X", GetLastError() ) ); 
		  return false; 
	} 

	if (GetLastError() == ERROR_NOT_ALL_ASSIGNED)
	{
		  ODS_fls( L"The token does not have the specified privilege. " );
		  return false;
	} 

	return true;
}

bool CAdvAPIDll::GetProcAddress( const char *procName, void **ppProc )
{
//	ODSFuncS( getWCStr( procName ) );

	m_intErr ++;

	if ( ppProc )
		*ppProc = NULL;

	if ( hInst == NULL )
	{
		//cWCStr wcs( "GetProcAddress:: " );
		//wcs.Add( procName );
		//wcs.Add( ", hInst == NULL \r\n"  );

		//WriteToFile( L"C:\\ODS\\Error.txt", (BYTE*)wcs.m_szBody, wcs.m_len * 2, true );

		return FALSE;
	}


	pProc = (void *)::GetProcAddress( hInst, procName );

	error = (pProc == NULL ? errGetProcAddress : errNo );

	//if ( error )
	//{
	//	cWCStr wcs( "GetProcAddress:: " );
	//	wcs.Add( procName );
	//	wcs.Add( "\r\n"  );

	//	WriteToFile( L"C:\\ODS\\Error.txt", (BYTE*)wcs.m_szBody, wcs.m_len * 2, true );
	//}

	if ( error && m_showMessage )
	{
		ErrorOut( "Can not get procedure address -", procName  );
	}
	else
	if ( ppProc )
		*ppProc = pProc;

	return error == errNo;
}

bool CAdvAPIDll::GetSidByProcessId( LPWSTR  *StringSid, DWORD pid ) 
{
	BOOL	rc = FALSE;
	HANDLE	hProc, hTok;
	DWORD	err;
//	TOKEN_USER	*pu;

	// open process by pid
	hProc = OpenProcess( PROCESS_QUERY_INFORMATION, FALSE, pid);
	err = GetLastError();

	if ( NULL == hProc )
	{
		ODS_fls( _FormatT( L"Assertion! Error %d in opening thread", err ) );
		return false;
	}
	
	rc = OpenProcessToken( hProc, TOKEN_QUERY, &hTok );
	err = GetLastError();
	CloseHandle(hProc);

	if ( !rc )
		return false;

	rc = GetSidByToken( StringSid, hTok );
	
	CloseHandle(hTok);

	return rc == TRUE;

	//// get sid of user identified by pid
	//pu = GetTokenUser(hTok);
	//CloseHandle(hTok);

	//if ( NULL == pu )
	//	return false;

	//// save user SID
	//rc = ConvertSidToStringSid( pu->User.Sid, StringSid );

	//free(pu);

	//if ( !rc )
	//	return false;

	//return true;
}

//
//	GetSidByToken
//
bool CAdvAPIDll::GetSidByToken( LPWSTR *StringSid, HANDLE hTok ) 
{
	BOOL	rc = FALSE;
	TOKEN_USER	*pu;


	// get sid of user identified by tid
	pu = GetTokenUser(hTok);

	if ( NULL == pu )
		return false;

	// save user SID
	rc = ConvertSidToStringSid( pu->User.Sid, StringSid );

	free(pu);

	if ( !rc )
		return false;

	return true;
}

TOKEN_USER *CAdvAPIDll::GetTokenUser( HANDLE hToken ) 
{
	DWORD		sz, err;
	TOKEN_USER	*ptu = NULL;
	BOOL		rc;

	// get required buffer size
	GetTokenInformation( hToken, TokenUser, NULL, 0, &sz );
	err = GetLastError();
	if (0 == sz || 0x10000 < sz) {
		return NULL;
	}
	ptu = (TOKEN_USER *)malloc(sz);
	if (NULL == ptu) {
		return NULL;
	}

	// get user sid
	rc = GetTokenInformation( hToken, TokenUser, ptu, sz, &sz );
	err = GetLastError();
	if (!rc) {
		free(ptu);
		return NULL;
	}
	return ptu;
}

//
//	CreateWritablePipe
//		creates named pipe writable by 'Everyone' group
//		so any logged user can communicate with service
//
HANDLE CreateWritablePipe(
		const wchar_t *szPipeName,
		DWORD openMode,		// = PIPE_ACCESS_DUPLEX | FILE_FLAG_OVERLAPPED,
		DWORD pipeMode,		// = PIPE_TYPE_BYTE,
		DWORD instances,	// = 1,
		DWORD outBuffSize,	// = sizeof( DWORD ),
		DWORD inBuffSize,	// = sizeof(DWORD),
		DWORD defTimeout	// = 1000
	)
{
	SECURITY_ATTRIBUTES		sa;
	HANDLE					hpipe = NULL;
	DWORD					err;

	PSID pEveryoneSID = NULL;
    PACL pACL = NULL;
    PSECURITY_DESCRIPTOR pSD = NULL;
    EXPLICIT_ACCESS ea;
    SID_IDENTIFIER_AUTHORITY SIDAuthWorld = SECURITY_WORLD_SID_AUTHORITY;

	CAdvAPIDll &advapi = CAdvAPIDll::getStatic();

    // Create a well-known SID for the Everyone group.
    if (!advapi.AllocateAndInitializeSid(&SIDAuthWorld, 1, SECURITY_WORLD_RID, 0, 0, 0, 0, 0, 0, 0, &pEveryoneSID)) {
        goto Cleanup;
    }

    // Initialize an EXPLICIT_ACCESS structure for an ACE.
    // The ACE will allow Everyone read access to the key.
    ZeroMemory(&ea, sizeof(EXPLICIT_ACCESS));
    ea.grfAccessPermissions = MAXIMUM_ALLOWED | SPECIFIC_RIGHTS_ALL | GENERIC_WRITE | GENERIC_READ; // was 0x801F0002;
    ea.grfAccessMode = SET_ACCESS;
    ea.grfInheritance= CONTAINER_INHERIT_ACE | OBJECT_INHERIT_ACE;
    ea.Trustee.TrusteeForm = TRUSTEE_IS_SID;
    ea.Trustee.TrusteeType = TRUSTEE_IS_WELL_KNOWN_GROUP;
    ea.Trustee.ptstrName  = (LPTSTR) pEveryoneSID;

    // Create a new ACL that contains the new ACEs.
    err = advapi.SetEntriesInAcl(1, &ea, NULL, &pACL);
    if (ERROR_SUCCESS != err) {
        goto Cleanup;
    }

    // Initialize a security descriptor.  
    pSD = (PSECURITY_DESCRIPTOR) LocalAlloc(LPTR, SECURITY_DESCRIPTOR_MIN_LENGTH); 
    if (NULL == pSD) { 
        goto Cleanup; 
    } 
 
    if (!advapi.InitializeSecurityDescriptor(pSD, SECURITY_DESCRIPTOR_REVISION)) {  
        goto Cleanup; 
    } 
 
    // Add the ACL to the security descriptor. 
    if (!advapi.SetSecurityDescriptorDacl(pSD, 
									TRUE,     // bDaclPresent flag   
									pACL, 
									FALSE))   // not a default DACL 
    {  
        goto Cleanup; 
    } 

    // Initialize a security attributes structure.
    sa.nLength = sizeof (SECURITY_ATTRIBUTES);
    sa.lpSecurityDescriptor = pSD;
    sa.bInheritHandle = FALSE;

    // Use the security attributes to set the security descriptor 
    // when you create a key.
	hpipe = CreateNamedPipeW(
				szPipeName,			// name
				openMode,			// mode
				pipeMode,			// type
				instances,			// instances
				outBuffSize,		// out buffer size					
				inBuffSize,			// in buffer size
				defTimeout,			// timeout
				&sa);				// security attributes

Cleanup:
    if (pEveryoneSID) 
        FreeSid(pEveryoneSID);
    if (pACL) 
        LocalFree(pACL);
    if (pSD) 
        LocalFree(pSD);

	return hpipe;
}

bool GetSidByProcessId( cWCStr *wcsSid, DWORD pid ) 
{
	wchar_t *szSID = NULL;

	if ( !CAdvAPIDll::getStatic().GetSidByProcessId( &szSID, pid ) )
		return false;

	if ( szSID != NULL )
	{
		wcsSid->Set( szSID );
		LocalFree( szSID );
	}

	return true;
}

bool GetCurrentThreadSid( cWCStr *wcsSid )
{
	wchar_t *szSid = NULL;

	HANDLE hTokenI = GetProcessImpersonatedToken_ODS();

	bool ok;

	//
	//	The thread has impersonated token
	//
	if ( hTokenI != NULL )
	{
		ok = CAdvAPIDll::getStatic().GetSidByToken( &szSid, hTokenI );
		CloseHandle( hTokenI );

		if ( !ok )
		{
			ODS_fls(  L"Assertion!" );
			return false;
		}
	}
	else
	{
		//
		//	The thread has no impersonated token
		//
		ok = CAdvAPIDll::getStatic().GetSidByProcessId( &szSid, GetCurrentProcessId() );

		if ( !ok )
		{
			ODS_fls(  L"Assertion!" );
			return false;
		}
	}

	wcsSid->Set( szSid );

	LocalFree( szSid );

	return ok;
}

//HANDLE CreateSecureSemaphore(
//    LONG lInitialCount,
//    LONG lMaximumCount,
//    LPCWSTR lpName,
//	int *pErr )
//{
//	HANDLE					hSemaphore		= NULL;
//    PSID					pEveryoneSID	= NULL;
//    PACL					pACL			= NULL;
//    PSECURITY_DESCRIPTOR	pSD				= NULL;
//
//	int						err;
//	if ( pErr == NULL )
//		pErr = &err;
//
//    EXPLICIT_ACCESS ea;
//    SID_IDENTIFIER_AUTHORITY SIDAuthWorld = SECURITY_WORLD_SID_AUTHORITY;
//
//	SECURITY_ATTRIBUTES sa;
//	ZeroMemory( &sa, sizeof( sa ) );
//	sa.nLength = sizeof( sa );
//
//	CAdvAPIDll &advapi = CAdvAPIDll::getStatic();
//
//    // Create a well-known SID for the Everyone group.
//    if (!advapi.AllocateAndInitializeSid(&SIDAuthWorld, 1,
//                     SECURITY_WORLD_RID, 0, 0, 0, 0, 0, 0, 0,
//                     &pEveryoneSID))
//	{
//		*pErr = GetLastError();
//
////		if ( advapi.error != errNo )
//		//{
//		//	cWCStr wcs;
//		//	wcs.FormatT( L"advapi.error %d, %d, %d\r\n", advapi.error, advapi.m_intErr, *pErr );
//		//	WriteToFile( L"C:\\ODS\\Error.txt", (BYTE*)wcs.m_szBody, wcs.m_len * 2, true );
//		//}
//
//		ODS_fls( _FormatT( L"AllocateAndInitializeSid Error 0x%X", *pErr ) );
//		return NULL;
//    }
//
//	// Initialize an EXPLICIT_ACCESS structure for an ACE.
//	// The ACE will allow Everyone read access to the key.
//	ZeroMemory( &ea, sizeof( EXPLICIT_ACCESS ) );
////	ea.grfAccessPermissions = MAXIMUM_ALLOWED | SPECIFIC_RIGHTS_ALL | GENERIC_WRITE;
//	ea.grfAccessPermissions = 0x801F0002;
//	ea.grfAccessMode = SET_ACCESS;
//	ea.grfInheritance= CONTAINER_INHERIT_ACE | OBJECT_INHERIT_ACE;
//	ea.Trustee.TrusteeForm = TRUSTEE_IS_SID;
//	ea.Trustee.TrusteeType = TRUSTEE_IS_WELL_KNOWN_GROUP;
//	ea.Trustee.ptstrName  = (LPTSTR) pEveryoneSID;
//
//	// Create a new ACL that contains the new ACEs.
//	err = advapi.SetEntriesInAcl( 1, &ea, NULL, &pACL );
//
//	if ( ERROR_SUCCESS != err )
//	{
//		*pErr = GetLastError();
//		ODS_fls( _FormatT( L"SetEntriesInAcl Error 0x%X", *pErr ) );
//		if (pEveryoneSID)	advapi.FreeSid( pEveryoneSID );
//		return NULL;
//	}
//
//	// Initialize a security descriptor.  
//	pSD = (PSECURITY_DESCRIPTOR) LocalAlloc(LPTR, SECURITY_DESCRIPTOR_MIN_LENGTH); 
//	if ( NULL == pSD )
//	{
//		*pErr = GetLastError();
//		ODS_fls( _FormatT( L"LocalAlloc Error 0x%X", *pErr ) );
//	}
//	else 
//	if (!advapi.InitializeSecurityDescriptor(pSD, SECURITY_DESCRIPTOR_REVISION))  
//	{
//		*pErr = GetLastError();
//		ODS_fls( _FormatT( L"InitializeSecurityDescriptor Error 0x%X", *pErr ) );
//	}
//	else
//	// Add the ACL to the security descriptor. 
//	if (!advapi.SetSecurityDescriptorDacl( pSD, 
//									TRUE,     // bDaclPresent flag   
//									pACL, 
//									FALSE))   // not a default DACL 
//	{
//		*pErr = GetLastError();
//		ODS_fls( _FormatT( L"SetSecurityDescriptorDacl Error 0x%X", *pErr ) );
//	}
//	else
//	{
//		// Initialize a security attributes structure.
//		sa.nLength = sizeof (SECURITY_ATTRIBUTES);
//		sa.lpSecurityDescriptor = pSD;
//		sa.bInheritHandle = FALSE;
//
//		// Use the security attributes to set the security descriptor 
//		// when you create a semaphore.
//		hSemaphore = CreateSemaphoreT( &sa, lInitialCount, lMaximumCount, lpName );
//
//		*pErr = GetLastError();
//
//		if ( hSemaphore == NULL )
//		{
//			ODS_fls( _FormatT( L"Error in creating semaphore %d", err ) );
//		}
//	}
//
//	if (pEveryoneSID)	advapi.FreeSid(pEveryoneSID);
//	if (pACL)			LocalFree(pACL);
//	if (pSD)			LocalFree(pSD);
//
//	return hSemaphore;
//}

class cSAObject
{
    PSID					pEveryoneSID;
    PACL					pACL;
    PSECURITY_DESCRIPTOR	pSD;
	int						err;

public:

	SECURITY_ATTRIBUTES		sa;
	int						*pErr;
	
	~cSAObject()
	{
		CAdvAPIDll &advapi = CAdvAPIDll::getStatic();

		if (pEveryoneSID)	advapi.FreeSid(pEveryoneSID);
		if (pACL)			LocalFree(pACL);
		if (pSD)			LocalFree(pSD);
	}

	cSAObject( int *pErrParam )
	{
		ZeroMemory( this, sizeof( *this ) );
		sa.nLength = sizeof( sa );

		if ( pErrParam == NULL )
			pErrParam = &err;

		pErr = pErrParam;

		EXPLICIT_ACCESS ea;
		SID_IDENTIFIER_AUTHORITY SIDAuthWorld = SECURITY_WORLD_SID_AUTHORITY;

		CAdvAPIDll &advapi = CAdvAPIDll::getStatic();

		// Create a well-known SID for the Everyone group.
		if (!advapi.AllocateAndInitializeSid(&SIDAuthWorld, 1,
						 SECURITY_WORLD_RID, 0, 0, 0, 0, 0, 0, 0,
						 &pEveryoneSID))
		{
			*pErr = GetLastError();

			ODS_fls( _FormatT( L"AllocateAndInitializeSid Error 0x%X", *pErr ) );
			return;
		}

		// Initialize an EXPLICIT_ACCESS structure for an ACE.
		// The ACE will allow Everyone read access to the key.
		ZeroMemory( &ea, sizeof( ea ) );
	//	ea.grfAccessPermissions = MAXIMUM_ALLOWED | SPECIFIC_RIGHTS_ALL | GENERIC_WRITE;
		ea.grfAccessPermissions = 0x801F0002;
		ea.grfAccessMode = SET_ACCESS;
		ea.grfInheritance= CONTAINER_INHERIT_ACE | OBJECT_INHERIT_ACE;
		ea.Trustee.TrusteeForm = TRUSTEE_IS_SID;
		ea.Trustee.TrusteeType = TRUSTEE_IS_WELL_KNOWN_GROUP;
		ea.Trustee.ptstrName  = (LPTSTR) pEveryoneSID;

		// Create a new ACL that contains the new ACEs.
		*pErr = advapi.SetEntriesInAcl( 1, &ea, NULL, &pACL );

		if ( ERROR_SUCCESS != *pErr )
		{
			ODS_fls( _FormatT( L"SetEntriesInAcl Error 0x%X", *pErr ) );
			return;
		}

		// Initialize a security descriptor.  
		pSD = (PSECURITY_DESCRIPTOR) LocalAlloc(LPTR, SECURITY_DESCRIPTOR_MIN_LENGTH); 
		if ( NULL == pSD )
		{
			*pErr = GetLastError();
			ODS_fls( _FormatT( L"LocalAlloc Error 0x%X", *pErr ) );
		}
		else 
		if (!advapi.InitializeSecurityDescriptor(pSD, SECURITY_DESCRIPTOR_REVISION))  
		{
			*pErr = GetLastError();
			ODS_fls( _FormatT( L"InitializeSecurityDescriptor Error 0x%X", *pErr ) );
		}
		else
		// Add the ACL to the security descriptor. 
		if (!advapi.SetSecurityDescriptorDacl( pSD, 
										TRUE,     // bDaclPresent flag   
										pACL, 
										FALSE))   // not a default DACL 
		{
			*pErr = GetLastError();
			ODS_fls( _FormatT( L"SetSecurityDescriptorDacl Error 0x%X", *pErr ) );
		}
		else
		{
			*pErr = ERROR_SUCCESS;

			// Initialize a security attributes structure.
			sa.nLength = sizeof (SECURITY_ATTRIBUTES);
			sa.lpSecurityDescriptor = pSD;
			sa.bInheritHandle = FALSE;
		}
	}
};

HANDLE CreateSecureMutex(
	BOOL bInitialOwner,
	LPCWSTR lpName,
	int *pErr )
{
	cSAObject sao( pErr );

	HANDLE	hMutex	= NULL;

	if ( *(sao.pErr) == ERROR_SUCCESS )
	{
		hMutex = CreateMutexW( &sao.sa, bInitialOwner, lpName );

		*(sao.pErr) = GetLastError();

		if ( hMutex == NULL )
		{
			ODS_fls( _FormatT( L"Error in creating mutex %d", *(sao.pErr) ) );
		}
	}

	return hMutex;
}

HANDLE CreateSecureSemaphore(
    LONG lInitialCount,
    LONG lMaximumCount,
    LPCWSTR lpName,
	int *pErr )
{
	cSAObject sao( pErr );

	HANDLE hSemaphore = NULL;

	if ( *(sao.pErr) == ERROR_SUCCESS )
	{
		hSemaphore = CreateSemaphoreT( &sao.sa, lInitialCount, lMaximumCount, lpName );

		*pErr = GetLastError();

		if ( hSemaphore == NULL )
		{
			ODS_fls( _FormatT( L"Error in creating semaphore %d", *(sao.pErr) ) );
		}
	}

	return hSemaphore;
}



//
//	SetFileValidDataT
//	Note 1.
//		The SE_MANAGE_VOLUME_NAME privilege is required for SetFileValidData function.
//		It has to be assigned before file opening.
//		Use CAdvAPIDll::SetPrivilege( SE_MANAGE_VOLUME_NAME, true );
//	Note 2.
//		The file cannot be on a network drive, or be compressed, sparse, or transacted.
//
bool SetFileValidDataT( HANDLE hFile, DWORDLONG fileSize )
{
	// Windows 2000 and 95 do not have SetFileValidData function.
	typedef BOOL WINAPI typeSetFileValidData(  __in  HANDLE hFile, __in  LONGLONG ValidDataLength );

	static CDll kernel32dll( false, "KERNEL32.dll", "SetFileValidData", false );

	if ( kernel32dll.error )
	{
		ODS_fls( L"Assertion! SetFileValidData function is not found" );
		return false;
	}

	//HANDLE hToken = NULL;
	//bool ok = true;
	//int res = 0;


	//if ( !CAdvAPIDll::getStatic().OpenProcessToken( GetCurrentProcess(), TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY, &hToken ) ) 
	//{
	//	ODS_fls( _FormatT( L"Assertion! OpenProcessToken return %d", GetLastError() ) );
	//	return false;
	//}

	//ok = CAdvAPIDll::getStatic().SetPrivilege( hToken, SE_MANAGE_VOLUME_NAME, true );

	//res = GetLastError();
	//CloseHandle( hToken );

	//if ( !ok )
	//{
	//	ODS_fls( _FormatT( L"Assertion! SetPrivilege return %d", res ) );
	//	return false;
	//}

	bool ok = ((typeSetFileValidData *)kernel32dll.pProc)( hFile, fileSize ) == TRUE;

	if ( !ok )
	{
		int res = GetLastError();
		ODS_fls( _FormatT( L"Assertion! SetFileValidData( %X, %s ) return %d", hFile, cWCStr().GetSizeText( fileSize ), res ) );
	}

	return ok;
}

//
//	Note! The SE_MANAGE_VOLUME_NAME privilege is required for SetFileValidData function.
//	It has to be assigned before file opening.
//	Use CAdvAPIDll::SetPrivilege( SE_MANAGE_VOLUME_NAME, true );
//
bool SetFileSizeAndValidData( const wchar_t *filePath, DWORDLONG fileSize, bool bFlushBuffers )
{
	HANDLE hFile = CreateFileT( filePath, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_ALWAYS, 0, NULL );

	if ( hFile == INVALID_HANDLE_VALUE )
	{
		ODS_fls( _FormatT( L"Assertion! CreateFileT return %d", GetLastError() ) );
		return false;
	}

	bool ok = SetFileSize( hFile, fileSize );

	if ( !ok )
	{
		int res = GetLastError();
		ODS_fls( _FormatT( L"Assertion! SetFileSize return %d", res ) );
	}
	else
		ok = SetFileValidDataT( hFile, fileSize );

	if ( ok && bFlushBuffers )
		FlushFileBuffers( hFile );

	CloseHandle( hFile );

	return ok;
}

//
//	Disables 'Program Compatibility Assistant Service'
//	It allows to skip annoing message 'Program might not have installed correctly'
//
void RestartPCA()
{
	ODSFunc;

	CAdvAPIDll &dll = CAdvAPIDll::getStatic();

	if ( dll.StopDriver( L"PcaSvc" ) )
	{
		ODS_fls( L"Service was stopped" );
		dll.StartDriver( L"PcaSvc" );
	}
	else
		ODS_fls( L"Service was NOT stopped" );
}

