/*
 *	Copyright  2016 Jetico Inc., Finland
 *	AdvApi_MakeHash.h
 *	Description:
 *		Calculate HASH 
 *
 *	Modified:	04.03.2015
 *	Author:		Alexander Pichuev.
 */

#pragma once

#include <windows.h>
#include <WinCrypt.h> // For CALG_SHA1
#include "ODS.h"

class cAdvApiForHash
{
	bool m_bOk;
	HMODULE m_hLib;

	typedef WINADVAPI BOOL WINAPI typeCryptAcquireContextW( HCRYPTPROV  *phProv, LPCWSTR    szContainer, LPCWSTR    szProvider, DWORD       dwProvType, DWORD       dwFlags  );
	typedef WINADVAPI BOOL WINAPI typeCryptCreateHash( HCRYPTPROV  hProv, ALG_ID      Algid, HCRYPTKEY   hKey, DWORD       dwFlags, HCRYPTHASH  *phHash );
	typedef WINADVAPI BOOL WINAPI typeCryptHashData( HCRYPTHASH  hHash, CONST BYTE  *pbData, DWORD   dwDataLen, DWORD   dwFlags );
	typedef WINADVAPI BOOL WINAPI typeCryptGetHashParam( HCRYPTHASH  hHash, DWORD   dwParam, BYTE    *pbData, DWORD   *pdwDataLen, DWORD   dwFlags );
	typedef WINADVAPI BOOL WINAPI typeCryptDestroyHash( HCRYPTHASH  hHash );
	typedef WINADVAPI BOOL WINAPI typeCryptReleaseContext( HCRYPTPROV  hProv, DWORD       dwFlags );

	typeCryptAcquireContextW	*m_CryptAcquireContext;
	typeCryptCreateHash			*m_CryptCreateHash;
	typeCryptHashData			*m_CryptHashData;
	typeCryptGetHashParam		*m_CryptGetHashParam;
	typeCryptDestroyHash		*m_CryptDestroyHash;
	typeCryptReleaseContext		*m_CryptReleaseContext;

	bool InitFuncs()
	{
		m_CryptAcquireContext	= (typeCryptAcquireContextW*)	::GetProcAddress( m_hLib, "CryptAcquireContextW" );	if ( m_CryptAcquireContext == NULL )	{ ODS_fls( L"No Function" ); return false; }
		m_CryptCreateHash		= (typeCryptCreateHash*)		::GetProcAddress( m_hLib, "CryptCreateHash" );		if ( m_CryptCreateHash == NULL )		{ ODS_fls( L"No Function" ); return false; }
		m_CryptHashData			= (typeCryptHashData*)			::GetProcAddress( m_hLib, "CryptHashData" );		if ( m_CryptHashData == NULL )			{ ODS_fls( L"No Function" ); return false; }
		m_CryptGetHashParam		= (typeCryptGetHashParam*)		::GetProcAddress( m_hLib, "CryptGetHashParam" );	if ( m_CryptGetHashParam == NULL )		{ ODS_fls( L"No Function" ); return false; }
		m_CryptDestroyHash		= (typeCryptDestroyHash*)		::GetProcAddress( m_hLib, "CryptDestroyHash" );		if ( m_CryptDestroyHash == NULL )		{ ODS_fls( L"No Function" ); return false; }
		m_CryptReleaseContext	= (typeCryptReleaseContext*)	::GetProcAddress( m_hLib, "CryptReleaseContext" );	if ( m_CryptReleaseContext == NULL )	{ ODS_fls( L"No Function" ); return false; }

		return true;
	}

public:

	cAdvApiForHash()
	{
		m_bOk = false;
		m_hLib = LoadLibraryW( L"advapi32.dll" );

		if ( m_hLib != NULL )
			m_bOk = InitFuncs();
	}

	~cAdvApiForHash()
	{
		if ( m_hLib != NULL )
			FreeLibrary( m_hLib );
	}

	bool cryptAcquireContext( HCRYPTPROV &hProv )
	{
		if(!m_CryptAcquireContext(&hProv, szContainer, szProvider, PROV_RSA_FULL, 0 ) )
		{
			// If error NTE_BAD_KEYSET so we have to create own new key set.
			DWORD dwErr = GetLastError();
			if ( dwErr == NTE_BAD_KEYSET )
			{
				const wchar_t *szContainer = L"ContainerXXX";
				const wchar_t *szProvider = NULL;

				if(!m_CryptAcquireContext(&hProv, szContainer, szProvider, PROV_RSA_FULL, 0 ) )
				{
					DWORD dwErr = GetLastError();
					if ( dwErr == NTE_BAD_KEYSET )
					{
						if (!m_CryptAcquireContext(&hProv, 
												 szContainer, 
												 szProvider, 
												 PROV_RSA_FULL, 
												 CRYPT_NEWKEYSET))
						{
							ODS_fls( _FormatT( L"Error %x!", GetLastError() ) );
							return false;
						}
						else
							ODS_fls( L"CRYPT_NEWKEYSET key is creazted!" );
					}
					else
					{
						ODS_fls( _FormatT( L"Error %x!", dwErr ) );
						return false;
					}
				}
			}
			else
			{
				ODS_fls( _FormatT( L"Error %x!", dwErr ) );
				return false;
			}
		}

		return true;
	}

	bool m_MakeHash( const BYTE* bf, DWORD bfLen, BYTE *hash, DWORD *pHashLen, ALG_ID hashId = CALG_SHA1 )
	{
		if ( !m_bOk )
			{ ODS_fls( L"Error!" ); return false; }

		HCRYPTPROV hProv = 0;
		HCRYPTHASH hHash = 0;
		DWORD dwHashLen;
		DWORD dwCount;

		if(!cryptAcquireContext( hProv ) )
			{ ODS_fls( L"Error!" ); return false; }

		if(!m_CryptCreateHash(hProv, hashId, 0, 0, &hHash))
			{ ODS_fls( _FormatT( L"Error %x!", GetLastError() ) ); return false; }

		BOOL ok = m_CryptHashData(hHash, bf, bfLen, 0);

		dwCount = sizeof(DWORD);

		if ( ok )
		{
			ok = m_CryptGetHashParam(hHash, HP_HASHSIZE, (BYTE *)&dwHashLen, &dwCount, 0);
			if ( !ok )
				ODS_fls( _FormatT( L"Error %x!", GetLastError() ) );
		}

		if ( ok )
		{
			if ( *pHashLen < dwHashLen )
			{
				ok = false;
				ODS_fls( _FormatT( L"Error %x < %x!", *pHashLen, dwHashLen ) );
			}

			*pHashLen = dwHashLen;
		}

		if ( ok )
		{
			ok = m_CryptGetHashParam(hHash, HP_HASHVAL, hash, &dwHashLen, 0);
			if ( !ok )
				ODS_fls( _FormatT( L"Error %x!", GetLastError() ) );
		}

		if (hHash) m_CryptDestroyHash(hHash);
		if (hProv) m_CryptReleaseContext(hProv,0);

		return ok == TRUE;
	}

	static bool MakeHash( const BYTE* bf, DWORD bfLen, BYTE *hash, DWORD *pHashLen, ALG_ID hashId = CALG_SHA1 )
	{
		static cAdvApiForHash dll;

		if ( !dll.m_bOk )
			{ ODS_fls( L"Error!" ); return false; }

		return dll.m_MakeHash( bf, bfLen, hash, pHashLen, hashId );
	}

	static bool MakeSHA1Hash( const BYTE* bf, DWORD bfLen, BYTE *hash, DWORD *pHashLen )
	{
		return MakeHash( bf, bfLen, hash, pHashLen, CALG_SHA1 );
	}
};