#include "FormatT.h"
#include "AutoDeallocate.h"
#include "ODS.h"

static int findCharPos( const wchar_t *str, wchar_t c, int len )
{
	for(int pos = 0;; pos++ )
	{
		if ( pos == len ) 		return -1;
		if ( str[ pos ] == 0 )	return -1;
		if ( str[ pos ] == c )	return pos;
	}
}

static int findCharPos( const wchar_t *str, int startPos, wchar_t c, int len )
{
	for(int pos = startPos;; pos++ )
	{
		if ( pos == len ) 		return -1;
		if ( str[ pos ] == 0 )	return -1;
		if ( str[ pos ] == c )	return pos;
	}
}

//
//	writeBuff
//		safely write string to buffer
//
static bool writeBuff( wchar_t *bf, int *bfPos, int bfSize, const wchar_t *s, int len = -1 )
{
	bool res = true;
	if ( len == -1 )
		len = (int)wcslen( s );

	if ( *bfPos >= bfSize )
		return false;

	if ( *bfPos + len >= bfSize )
	{
		len = bfSize - *bfPos - 1;
		res = false;
	}

	wcsncpy_s( bf + *bfPos, bfSize - *bfPos, s, len );
	*bfPos += len;
	bf[ *bfPos ] = L'\0';
	return res;
}

static bool getDigitChar( int *num, wchar_t wc )
{
	if ( ( wc >= L'0' ) && ( wc <= L'9' ) )
	{
		if ( num != NULL )
			*num = wc - L'0';

		return true;
	}
	else
	if ( wc == L'-' )
	{
		if ( num != NULL )
			*num = -1;

		return true;
	}

	return false;
}

#ifndef maxLen
#define maxLen( bf ) ( sizeof( bf ) / sizeof( *bf ) )
#endif

static bool isValidFlag( wchar_t wc )
{
	wchar_t validFlagT[] = { L'-', L'+', L'0', L' ', L'#' };

	for( int i = maxLen( validFlagT ); i--; )
	{
		if ( validFlagT[i] == wc )
			return true;
	}

	return false;
}

static bool isValidWidth( wchar_t wc )
{
	return ( wc >= L'0' ) && ( wc <= L'9' );
}

static int isValidPrefix( const wchar_t *ws, bool *pIsInt64 = NULL )
{
	const wchar_t *validPrefxT[] = { L"h", L"l", L"I", L"I32", L"I64" };

	int len = 0;
	bool isInt64 = false;

	for( int i = maxLen( validPrefxT ); i--; )
	{
		int lLen = (int)wcslen( validPrefxT[i] );
		if ( wcsncmp( ws, validPrefxT[i], lLen ) == 0 )
		{
			isInt64 = ( i == 4 );
			len = lLen;
			break;
		}
	}

	if ( pIsInt64 != NULL )
		*pIsInt64 = isInt64;

	return len;
}

static int getFmtSize( const wchar_t *fmtS )
{
	int pos = 0;

	if ( fmtS[ pos ] != L'%' )
		return 0;

	pos ++;

	if ( isValidFlag( fmtS[ pos ] ) )
		pos++;

	while( isValidWidth( fmtS[ pos ] ) ) pos++;

	if ( fmtS[ pos ] == L'.' ) // float precision
	{
		pos++;
		while( isValidWidth( fmtS[ pos ] ) ) pos++;
	}

	pos += isValidPrefix( fmtS + pos );

	pos++;

	return pos;
}

static bool isDouble( wchar_t wcType )
{
	const wchar_t *doubleTypes = L"eEfgG";

	for( const wchar_t *wsT = doubleTypes; *wsT; wsT++ )
	{
		if ( *wsT == wcType )
			return true;
	}

	return false;
}

static bool getFormatInfo( int *pValueSize, char *ptA, int ptASize, const wchar_t *fmtS )
{
	int pos = 0;

	if ( fmtS[ pos ] != L'%' )
		return false;

	pos++;

	if ( isValidFlag( fmtS[ pos ] ) )
		pos++;

	while( isValidWidth( fmtS[ pos ] ) ) pos++;

	if ( fmtS[ pos ] == L'.' ) // float precision
	{
		pos++;
		while( isValidWidth( fmtS[ pos ] ) ) pos++;
	}

	bool isInt64 = false;

	pos += isValidPrefix( fmtS + pos, &isInt64 );
	
	if ( isDouble( fmtS[ pos ] ) )
		*pValueSize = sizeof( double );
	else
	if ( isInt64 )
		*pValueSize = sizeof( __int64 );
	else
		*pValueSize = sizeof( int );

	pos++;

	if ( pos > ptASize )
		return false;

	size_t num;
	wcstombs_s( &num, ptA, ptASize, fmtS, pos );
	ptA[ pos ] = '\0';

	return true;
}

//static wchar_t *asciiConvert( wchar_t *bf, int bfSize, int *fmtSize, const wchar_t *fmtS, const cVAList *vl )
//{
//	bf[0] = L'\0';
//
//	*fmtSize = getFmtSize( fmtS );
//	if ( *fmtSize <= 1 )
//	{
//		assert( false );
//		*fmtSize = 1;
//		return bf;
//	}
//
//	cMBStr mbsBf( NULL, bfSize );
//
//	cMBStr mbsFmt( fmtS, *fmtSize + 1 );
//	
//	mbsFmt.m_szBody[ *fmtSize ] = '\0';
//
//	char wcType = mbsFmt.m_szBody[ *fmtSize - 1 ];
//
//	va_list marker = vl->marker;
//
//	int res = vsprintf( mbsBf.m_szBody, mbsFmt.m_szBody, marker );
//
//	const wchar_t *S = va_arg( marker, LPCWSTR );
//	
//	if ( res == -1 )
//	{
//		assert( false );
//		return bf;
//	}
//
//	mbstowcs( bf, mbsBf.m_szBody, bfSize );
//
//	return bf;
//}
//
//static const wchar_t *asciiConvert( wchar_t *bf, int bfSize, int *fmtSize, const wchar_t *fmtS, const wchar_t *val )
//{
//	bf[0] = L'\0';
//
//	*fmtSize = getFmtSize( fmtS );
//	if ( *fmtSize <= 1 )
//	{
//		assert( false );
//		*fmtSize = 1;
//		return bf;
//	}
//
//	cMBStr mbsBf( NULL, bfSize );
//
//	cMBStr mbsFmt( fmtS, *fmtSize + 1 );
//	
//	mbsFmt.m_szBody[ *fmtSize ] = '\0';
//
//	int res = sprintf( mbsBf.m_szBody, mbsFmt.m_szBody, val );
//	
//	if ( res == -1 )
//	{
//		assert( false );
//		return bf;
//	}
//
//	mbstowcs( bf, mbsBf.m_szBody, bfSize );
//
//	return bf;
//}

static bool isArgSizePrefix( const wchar_t *pos, int *argSize, int *skipSize )
{
	struct sPrefix{
		const wchar_t *type;
		int size;
	};

	sPrefix T[] = 
	{
		{ L"h", sizeof( short int ) },
		{ L"l", sizeof( long unsigned int ) },
		{ L"I", sizeof( int ) },
		{ L"I32", sizeof( __int32 ) },
		{ L"I64", sizeof( __int64 ) }
	};

	for ( int i = maxLen( T ); i--; )
	{
		int len = (int)wcslen( T[i].type );
		if ( wcsncmp( T[i].type, pos, len ) == 0 )
		{
			*argSize = T[i].size;
			*skipSize = len;
			return true;
		}
	}

	return false;
}

static bool isValidFormat( wchar_t fmt )
{
	const wchar_t *fmts = L"cCdiouxXeEfgGnpsS";
	for ( const wchar_t *s = fmts; *s; s++ )
	{
		if ( *s == fmt )
			return true;
	}

	return false;
}

wchar_t *FormatT( wchar_t *buff, int buffMaxLen, const wchar_t *pattern, ... )
{
	cVAList vl;
	va_start( vl.marker, pattern );
	return FormatT( buff, buffMaxLen, &vl, pattern );
}

const int maxFormatLen = 2048;

//wchar_t *FormatOnceT( const wchar_t *pattern, ... )
//{
//	static cAutoDeallocate < wchar_t > ad;
//
//	if ( ad.Allocate( maxFormatLen ) == NULL ) return L"<NULL>";
//
//	cVAList vl;
//	va_start( vl.marker, pattern );
//	return FormatT( ad.m_ptr, ad.m_len, &vl, pattern );
//}

wchar_t *FormatT( wchar_t *buff, int buffMaxLen, cVAList *vl, const wchar_t *pattern )
{
	int startPos = 0;
	int buffPos = 0;
	int patternLen = (int)wcslen( pattern );

	cADBlock < const wchar_t * > adParamT( 0, true ); // For enumerated strings only.

	bool bOrderedString = false; // '%1 %2 ...'
	
	for(;;)
	{
		int num = -1;		

		int pos = findCharPos( pattern, startPos, L'%', patternLen );

		if ( pos == -1 ) 
		{
			// Writes whole pattern and exit
			writeBuff( buff, &buffPos, buffMaxLen, pattern + startPos );
			return buff;
		}

		// If buffer is completely filled we may bravely return here
		if ( !writeBuff( buff, &buffPos, buffMaxLen, pattern + startPos, pos - startPos ) )
			return buff;

		int fmtSize = 2;
		int formatWidth = 0;
		bool digitsFillZero = false;
		bool leftSideFormatted = false;

		wchar_t fmtType = pattern[ pos + 1 ];

		//
		//	Formats strings
		//		a. ordered string "QQQ %2 WWW %1", S1, S2
		//			Note it is one based number
		//		b. digital string like %04d or %-30s
		//
		int addNum;
		if ( getDigitChar( &addNum, fmtType ) )
		{
			if ( addNum == -1 ) // left side formatted string '%-' encountered
			{
				// we have to check this f.e. '%-30s'
				ODS_f( L"assert( false )" );

				leftSideFormatted = true;

				if ( !getDigitChar( &addNum, pattern[ pos + fmtSize++ ] ) )
					return false;
			}

			formatWidth = addNum;

			if ( formatWidth == 0 )
				digitsFillZero = true;

			while( getDigitChar( &addNum, pattern[ pos + fmtSize ] ) )
			{
				formatWidth = 10 * formatWidth + addNum;
				fmtSize ++;
			}

			int skip, argSize;

			if ( isArgSizePrefix( pattern + pos + fmtSize, &argSize, &skip ) )
			{
				fmtSize += skip;
			}

			fmtType = pattern[ pos + fmtSize ];

			if ( bOrderedString || !isValidFormat( fmtType ) ) //  == L'\"' ) || ( fmtType == L'\'' ) || ( fmtType == L' ' ) || ( fmtType == L'\0' ) || ( fmtType == L'\t' ) || ( fmtType == L'\n' ) || ( fmtType == L'\r' ) )
			{
				// If we find one ordered parameters all other have to be ordered too.
				bOrderedString = true;

				num = formatWidth - 1; // one based number

				if ( num < 0 )
				{
					ODS_f( L"assert( num >= 0 )" );
					break;
				}

				fmtType = L's';
			}
			else
			{
				fmtSize++;
				if ( num != -1 ) // Ordered strings have been detected
				{
					// So what, let to uses next formatted parameter 
					//assert( false );
					//return false;
					num = -1;
				}
			}
		}

		LPCWSTR S;
		wchar_t digitStr[ 512 ] = L"XX";

		if ( num != -1 ) // uses %1 %2 order based strings 
		{

			//
			//	Problem detected the va_end and va_start does not work in subroutine.
			//
			//adParamT.Resize( num );

			while(  num >= adParamT.m_Q )
			{
				S = va_arg( vl->marker, LPCWSTR );
				adParamT.Add( &S, 1 );
			}

			S = adParamT.m_T[ num ];

			// Sets marker to beginning
			//	va_end( vl->marker );
			//	va_start( vl->marker, pattern );

			// Gets required argument
			//	for( int i = 0; i <= num; i++ )
			//		S = va_arg( vl->marker, LPCWSTR );
		}
		else
		if ( fmtType == L'%' ) // Skips one % of double %%
			S = L"%";
		else
		if ( fmtType == L'P' ) // Fill probels according argumment.
		{
			int valInt;
			valInt = va_arg( vl->marker, int );

			if ( ( valInt < 0 ) || ( valInt >= maxLen( digitStr ) ) )
				S = L"'Invalid value'";
			else
			{
				for( digitStr[valInt] = L'\0'; valInt--; )
					digitStr[valInt] = L' ';
				S = digitStr;
			}
		}
		else
		if ( fmtType == L'p' ) // Pointer
		{
			void *valPtr = NULL;
			valPtr = va_arg( vl->marker, void* );
			swprintf_s( digitStr, L"0x%p", valPtr );
			S = digitStr;
		}
		else
		if ( fmtType == L's' )
		{
			S = va_arg( vl->marker, LPCWSTR );

			// It is wrong if ordered string has not ordered parameter.
			if ( bOrderedString )
				ODS_f( L"assert( !bOrderedString )" );
			// assert( !bOrderedString );

			// We add it to table for any case.
			adParamT.Add( &S, 1 );

			if ( S == NULL )
			{
				// assert( false );
				S = L"NULL";
				// MessageBoxT( 0, pattern, L"FormatT - String is NULL", MB_ICONSTOP );
				// break;
			}

			int len = (int)wcslen( S );

			//
			//	Checks and formats left or right oriented string '%-20s'
			//		'%-20s' -> 'theString         '
			//		'%20s'  -> '         theString'
			//
			if ( ( formatWidth > len ) && ( formatWidth < maxLen( digitStr ) ) )
			{
				if ( leftSideFormatted )
				{
					wcsncpy_s( digitStr, S, len );

					for( ; len < formatWidth; len++ )
						digitStr[ len ] = L' ';
				}
				else
				{
					int pos = 0;

					for( ; pos <  ( formatWidth - len ); pos++ )
						digitStr[ pos ] = L' ';

					wcsncpy_s( digitStr + pos, maxLen( digitStr ) - pos, S, formatWidth - pos );
				}
				
				digitStr[ formatWidth ] = L'\0';

				S = digitStr;
			}
		}
		else
		{
			char bfA[ 100 ];
			char ptA[ 100 ] = "%0X";
			int valueSize = 0;
			if ( !getFormatInfo( &valueSize, ptA, maxLen( ptA ), pattern + pos ) )
			{
				// Skips the invalid format or "%%"
				wcsncpy_s( digitStr, pattern + pos, 2 );
				digitStr[2] = L'\0';

				if ( digitStr[1] == L'%' )
					digitStr[1] = L'\0';

				fmtSize = 2;
				S = digitStr;
			}
			else
			{
				// vsprintf( bfA, ptA, vl->marker );
				// move to next argument is required, it strange but vsprintf does not make it automatically

				double valDouble;
				int valInt;

				// It is wrong if ordered string has not ordered parameter.
				if ( bOrderedString )
					ODS_f( L"assert( !bOrderedString )" );
				// assert( !bOrderedString );

				if ( valueSize == sizeof( double ) )
				{
					valDouble = va_arg( vl->marker, double );
					sprintf_s( bfA, ptA, valDouble );
				}
				else
				{
					valInt = va_arg( vl->marker, int );
					sprintf_s( bfA, ptA, valInt );
				}

				size_t num;
				mbstowcs_s( &num, digitStr, bfA, maxLen( digitStr ) );
				S = digitStr;

				fmtSize = (int)strlen( ptA );

				//S = asciiConvert( digitStr, maxLen( digitStr ), &fmtSize, pattern + pos, &vl );
			}
		}

		if ( S == NULL )
			S = L"<Null>";

		startPos = pos + fmtSize;

		if ( !writeBuff( buff, &buffPos, buffMaxLen, S ) )
			return buff;
	}

	return buff;
}
