#include "ODS.h"
#ifdef _ODS
#include "cSemaphore.h"
#include "cLongName.h"
#include <assert.h>

HANDLE
CreateFileODS( const wchar_t *szFunc, int iLine,
    LPCWSTR lpFileName,
    DWORD dwDesiredAccess,
    DWORD dwShareMode,
    LPSECURITY_ATTRIBUTES lpSecurityAttributes,
    DWORD dwCreationDisposition,
    DWORD dwFlagsAndAttributes,
    HANDLE hTemplateFile
    )
{
	HANDLE hFile = CreateFileW( lpFileName, dwDesiredAccess, dwShareMode, lpSecurityAttributes, dwCreationDisposition, dwFlagsAndAttributes, hTemplateFile );

	DWORD err = GetLastError();

	if ( hFile == INVALID_HANDLE_VALUE )
		ODS( szFunc, iLine, _FormatT( L"CreateFile( %s ) -> Error %x", lpFileName, err ) );
	else
		ODS( szFunc, iLine, _FormatT( L"CreateFile( %s ) = 0x%X", lpFileName, hFile ) );

	SetLastError( err );

	return hFile;
}

#include <time.h>

cODS::cODS( const wchar_t *wcsName, const wchar_t *s, bool bForce )
{
	clear();

	if ( s == (const wchar_t *)(-1) )
	{
		set( wcsName, L"", bForce );
		return;
	}

	if ( s == NULL )
		s = L"Null";

	set( wcsName, _FormatT( L"( %s )", s ), bForce );
}

#ifndef maxLen
#define maxLen( bf ) ( sizeof( bf ) / sizeof( *bf ) )
#endif

void cODS::setResult( int r )		{	setResult( _FormatT( L"%d", r ) ); }
void cODS::setResult( __int64 r )	{	setResult( _FormatT( L"0x%016I64X", r ) ); }
void cODS::setResult( DWORD r )		{	setResult( _FormatT( L"0x%X", r ) ); }
void cODS::setResult( bool r )		{	setResult( _FormatT( L"%s", r ? L"true" : L"false" ) ); }

//static wchar_t *getDateStr( wchar_t *str, int strMaxLen, SYSTEMTIME *systime = NULL )
//{
//	SYSTEMTIME _systime;
//
//	if ( systime == NULL )
//	{
//		systime = &_systime;
//		GetLocalTime( systime );
//	}
//
//	FormatT( str, strMaxLen,
//		L"%u/%u/%u",
//		systime->wYear, systime->wMonth, systime->wDay );
//
//	return str;
//}

static wchar_t *getTimeStr( wchar_t *str, int strMaxLen, SYSTEMTIME *systime = NULL )
{
	SYSTEMTIME _systime;

	if ( systime == NULL )
	{
		systime = &_systime;
		GetLocalTime( systime );
	}

	FormatT( str, strMaxLen,
		L"%02u:%02u:%02u:%03u",
		systime->wHour, systime->wMinute, systime->wSecond, systime->wMilliseconds);

	return str;
}

static wchar_t *getMinuteStr( wchar_t *str, int strMaxLen, SYSTEMTIME *systime = NULL )
{
	SYSTEMTIME _systime;

	if ( systime == NULL )
	{
		systime = &_systime;
		GetLocalTime( systime );
	}

	FormatT( str, strMaxLen,
		L"%02u:%02u",
		systime->wHour, systime->wMinute);

	return str;
}

static wchar_t *getSecondsStr( wchar_t *str, int strMaxLen, SYSTEMTIME *systime = NULL )
{
	SYSTEMTIME _systime;

	if ( systime == NULL )
	{
		systime = &_systime;
		GetLocalTime( systime );
	}

	FormatT( str, strMaxLen,
		L"%02u:%03u",
		systime->wSecond, systime->wMilliseconds);

	return str;
}

static wchar_t *getDateTimeStr( wchar_t *str, int strMaxLen, SYSTEMTIME *systime = NULL )
{
	SYSTEMTIME _systime;

	if ( systime == NULL )
	{
		systime = &_systime;
		GetLocalTime( systime );
	}

	FormatT( str, strMaxLen,
		L"%u/%u/%u  %02u:%02u:%02u:%03u",
		systime->wYear, systime->wMonth, systime->wDay,
		systime->wHour, systime->wMinute, systime->wSecond, systime->wMilliseconds);

	return str;
}

void ODSE( const wchar_t *s0, const wchar_t *s1 )
{
	ODS( s0, s1, true );
}

void ODSFE( const wchar_t *pattern, ... )
{
	cWCSF wcs( 1024 );
	if ( wcs.m_szBody == NULL )
		return;

	cVAList vl;
	va_start( vl.marker, pattern );

	FormatT( wcs.m_szBody, wcs.m_len, &vl, pattern );

	ODS( wcs.m_szBody, L"", true );
}

void ODSF( const wchar_t *pattern, ... )
{
	cWCSF wcs( 1024 );
	if ( wcs.m_szBody == NULL )
		return;

	cVAList vl;
	va_start( vl.marker, pattern );

	FormatT( wcs.m_szBody, wcs.m_len, &vl, pattern );

	ODS( wcs.m_szBody );
}

//static int count = 0;

#ifdef _ODSM
class cODSMProcess
{
	bool m_bWasCalled;

	//cODS m_ods;

public:

	void setFirstCall()
	{
		m_bWasCalled = false;
		//m_ods.set( L"Multi Process Mode" );
	}

	cODSMProcess()
	{
		// setFirstCall();
	}

	bool isFirstCall()
	{
		if ( m_bWasCalled )
			return false;

		m_bWasCalled = true;

		return true;
	}

	~cODSMProcess()
	{
//		ODS( L"Process STOP" );
	}
};

//static cODSMProcess processStat;
cODSMProcess *pProcessStat = NULL;

#endif

struct sThreadBlock
{
	DWORD id;
	cWCSF name;
	cWCSF originalName;
	cWCSF path;
	int recCount;
	bool m_bWasCalled;

	bool isFirstCall()
	{
		if ( m_bWasCalled )
			return false;

		m_bWasCalled = true;

		return true;
	}

	sThreadBlock()
		:
		originalName( cWCSF::maxFormatLength ),
		name( cWCSF::maxFormatLength ),
		path( cWCSF::maxFormatLength )
	{
		Clear();
	}

	void Clear()
	{
		id = 0;
		name.SetLength( 0 );
		recCount = 0;
		m_bWasCalled = false;
	}
};

static void writeToFile( const wchar_t *filePath, const wchar_t *msg )
{
	if ( msg == 0 )
		return;

	cWCS wcs( filePath );

	wchar_t *fileName = wcsrchr( wcs.m_szBody, L'\\' );
	if ( fileName == NULL )
		return;

	*fileName = L'\0';

	if ( _waccess( wcs.m_szBody, 0 ) != 0 )
		CreateDirectoryW( wcs.m_szBody, NULL );

	HANDLE hFile =
		CreateFileW( filePath,
			GENERIC_READ | GENERIC_WRITE,
			FILE_SHARE_READ,
			NULL,
			OPEN_ALWAYS,
			0,
			NULL );

	if ( hFile == NULL )
		return;

	SetFilePointer( hFile, 0, NULL, FILE_END );

	DWORD written;

	WriteFile( hFile, msg, (DWORD)wcslen( msg ) * sizeof( wchar_t ), &written, NULL );
	WriteFile( hFile, L"\r\n", 2 * sizeof( wchar_t ), &written, NULL );

	CloseHandle( hFile );
}


//
//	readODSDirName
//		[HKEY_LOCAL_MACHINE\SOFTWARE\Jetico]
//			@=""
//			"ODSDir"="H:\\ODS"
//
static const wchar_t *readODSDirName( cWCSF &wcs, DWORD wow64Flag )
{
	if ( wcs.m_len != 0 ) 
		return wcs.m_szBody;

	//cWCSF err;
	//const wchar_t *szODSDirErr = L"C:\\ODS\\ODS_DIR.txt";

	HKEY hKey;
	LONG lRes = RegOpenKeyExW( HKEY_LOCAL_MACHINE, L"SOFTWARE\\Jetico", 0, KEY_READ | wow64Flag, &hKey );

	if ( lRes != ERROR_SUCCESS )
	{
		//err.FormatT( L"Error %d in opening \"HKEY_LOCAL_MACHINE\\SOFTWARE\\Jetico\" %x", lRes, wow64Flag );
		//writeToFile( szODSDirErr, err.m_szBody );
		return NULL;
	}

	DWORD dwType = REG_SZ;
	DWORD dwSize = 0;
	
	const wchar_t *szValueName = L"ODSDir";

	lRes = RegQueryValueExW( hKey, szValueName, NULL, &dwType, NULL, &dwSize );

	if ( lRes == ERROR_SUCCESS )
	{
		wcs.SetLength( dwSize / sizeof( wchar_t ) );

		lRes = RegQueryValueExW( hKey, szValueName, NULL, &dwType, (LPBYTE)wcs.m_szBody, &dwSize );
	}
	
	RegCloseKey( hKey );

	if ( lRes != ERROR_SUCCESS )
	{
		//err.FormatT( L"Error %d in reading \"HKEY_LOCAL_MACHINE\\SOFTWARE\\Jetico\" %x, value \"%s\"", lRes, wow64Flag, szValueName );
		//writeToFile( szODSDirErr, err.m_szBody );
		wcs.Free();
	}
	else
	{
		if ( wcs.m_szBody[ wcs.m_len - 1 ] == L'\\' )
			wcs.m_szBody[ wcs.m_len - 1 ] = L'\0';
		else
		if ( wcs.m_szBody[ wcs.m_len - 2 ] == L'\\' )
			wcs.m_szBody[ wcs.m_len - 2 ] = L'\0';
		else
		if	( wcs.m_szBody[ wcs.m_len ] != L'\0' )
			wcs.Add( L"" );
	}

	//writeToFile( szODSDirErr, wcs.m_szBody );

	return wcs.m_szBody;
}

static const wchar_t *readODSDirName( cWCSF &wcs )
{
	const wchar_t *dir = readODSDirName( wcs, KEY_WOW64_64KEY );
	if ( dir == NULL )
		dir = readODSDirName( wcs, 0 );

	return dir;
}


static bool updatePath( wchar_t *path, const wchar_t *orig, const wchar_t *pattern )
{
	int lenOrig = (int)wcslen( orig );
	int lenPatt = (int)wcslen( pattern );

	if ( lenPatt > lenOrig )
		return false;

	int count = 0;

	for( int i = (int)wcslen( path ) - lenOrig; i >= 0 ; )
	{
		if ( wcsncmp( path + i, orig, lenOrig ) == 0 )
		{
			wcsncpy( path + i, pattern, lenPatt );
			memmove( path + i + lenPatt, path + i + lenOrig, ( wcslen( path + i + lenOrig ) + 1 )* sizeof(wchar_t) );
			i -= lenOrig;
			count ++;
		}
		else
			i--;
	}

	return count != 0;
}

void updatePathSymbols( wchar_t *path  )
{
	wchar_t temp[ MAX_PATH ] = L"";
	if ( GetTempPathW( maxLen( temp ), temp ) != 0 )
		updatePath( path, temp, L"TEMP\\" );

	wchar_t tempShort[ MAX_PATH ] = L"";
	if ( GetShortPathNameW( temp, tempShort, maxLen( tempShort ) ) != 0 )
		updatePath( path, tempShort, L"TEMP\\" );

	for( int i = 0; path[i] != L'\0'; i++ ) 
	{
			 if ( path[i] == L'*'  ) path[i] = L'x';
		else if ( path[i] == L'@'  ) path[i] = L'a';
		else if ( path[i] == L'\"' ) path[i] = L'Q';
		else if ( path[i] == L'\'' ) path[i] = L'Q';
		else if ( path[i] == L'\\' ) path[i] = L'$';
		else if ( path[i] == L':'  ) path[i] = L'd';
		else if ( path[i] == L'<'  ) path[i] = L'{';
		else if ( path[i] == L'>'  ) path[i] = L'}';
		else if ( path[i] == L'?'  ) path[i] = L'$';
	}
}

class cODSObject
{
public:

	HINSTANCE m_hODSInst;
	bool m_enable;
	bool m_verbose;

	sThreadBlock m_namedThreadT[50];
	int m_namedThreadQ;

	int m_defRecCount;

	cWCSF m_wcsModulePath;
	cWCSF m_wcsOriginalName;

	cODSObject()
		:
		m_wcsModulePath( cWCSF::maxFormatLength ),
		m_wcsOriginalName( cWCSF::maxFormatLength )
	{
		m_namedThreadQ = maxLen( m_namedThreadT );
		m_defRecCount = 0;
		m_hODSInst = NULL;
		m_enable = true;
		m_verbose = false;
	}

	static cODSObject *getStatic()
	{
		static cODSObject staticObj;
		return &staticObj;
	}

	static cODSObject *m_external;

	static cODSObject *getCurrent()
	{
		if ( m_external != NULL )
			return m_external;

		return getStatic();
	}

	static const wchar_t *getODSDir()
	{
		static cWCSF wcs;
		static const wchar_t *regName = readODSDirName( wcs );

		if ( regName != NULL )
			return regName;

		return L"C:\\ODS";
	}

	static int findThreadBlock_CS( DWORD dwId = 0, const wchar_t *nameToSet = NULL, bool bClear = false )
	{
		static cSysCriticalSection cs;
		//static CRITICAL_SECTION cs;

		int i;

		EnterCriticalSection( &cs );

			i = cODSObject::getCurrent()->findThreadBlock( dwId, nameToSet );

			if ( ( i >= 0 ) && bClear )
				cODSObject::getCurrent()->m_namedThreadT[i].Clear();

		LeaveCriticalSection( &cs );

		if ( ( i < 0 ) && ( nameToSet != NULL ) )
			ODS_fls( _FormatT( L"Assertion: thread (0x%04X) number %d is acceeded", dwId, cODSObject::getCurrent()->m_namedThreadQ ) );

		return i;
	}

	int findThreadBlock( DWORD dwId = 0, const wchar_t *nameToSet = NULL )
	{
		int i = m_namedThreadQ;
		int iEmpty = -1;

		for( ; i-- ; )
		{
			if ( m_namedThreadT[i].id == 0 )
				iEmpty = i;

			if ( m_namedThreadT[i].id == dwId )
				break;
		}

		if ( nameToSet == NULL ) // No set name is required.
			return i;

		bool bFound = ( i > 0 );

		if ( bFound )
		{
			//ODS_fls( _FormatT( L"Assertion: thread %s, %x found at setting name", nameToSet, dwId ) );
			return i;
		}

		if ( iEmpty == -1 )
		{
			//ODS_fls( L"Assertion: thread number is acceeded" );
			return -1;
		}

		m_namedThreadT[ iEmpty ].id = dwId;
		m_namedThreadT[ iEmpty ].name.FormatT( nameToSet );
		updatePathSymbols( m_namedThreadT[ iEmpty ].name.m_szBody );
		m_namedThreadT[ iEmpty ].recCount = 0;

		//ODS_fls( _FormatT( L"%0x04X, %s", dwId, m_namedThreadT[ iEmpty ].name.m_szBody ) );

		return iEmpty;
	}

};

bool isCurObjVerbose()
{
	return cODSObject::getCurrent()->m_verbose;
}


#define DllExport	__declspec( dllexport )
extern "C" DllExport void ODS_SetExternalObject( cODSObject *obj );

void ODS_SetExternalObject( cODSObject *obj )
{
	cODSObject::m_external = obj;
}

bool ODS_SetExternalObject( HMODULE hInst )
{
	typedef void type_ODS_SetExternalObject( cODSObject *obj );
	type_ODS_SetExternalObject *proc = (type_ODS_SetExternalObject *)GetProcAddress( hInst, "ODS_SetExternalObject" );
	if ( proc == NULL )
		return false;

	proc( cODSObject::getCurrent() );

	return true;
}


//static
cODSObject *cODSObject::m_external = NULL;
const wchar_t *getODSDir()
{
	return cODSObject::getODSDir();
}


//static HINSTANCE hODSInst = NULL;
//bool cODS::m_enable = true;
//bool cODS::m_verbose = false;

//static sThreadBlock *namedThreadT = NULL;
//static int namedThreadQ = 0;
//static int defRecCount = 0;

//static cWCSF wcsModulePath;
//static cWCSF wcsOriginalName;


cODSThread::cODSThread(const wchar_t *name)
{
	DWORD dwId = GetCurrentThreadId();
	
	ODS_fls( _FormatT( L"%s, 0x%04X", name, dwId ) );

	// static sThreadBlock staticThreadNameT[50];
	//namedThreadT = &(staticThreadNameT[0]);
	//namedThreadQ = maxLen( staticThreadNameT );

	//if ( cODSObject::m_external != NULL )
	//	ODS_fls( L"External block disables separate thread name" );
	//else
		cODSObject::findThreadBlock_CS( dwId, name );
}

cODSThread::~cODSThread()
{
	cODSObject::findThreadBlock_CS( GetCurrentThreadId(), NULL, true );
}

static int getNamedThread()
{
	return cODSObject::findThreadBlock_CS( GetCurrentThreadId() );
}

static int *getRecCountPtr()
{
	int tn = getNamedThread();
	if ( tn != -1 )
		return &( cODSObject::getCurrent()->m_namedThreadT[ tn ].recCount );
	else
		return &( cODSObject::getCurrent()->m_defRecCount );
}

void updateFileName( wchar_t *path )
{
	wchar_t *fileName = wcsrchr( path, L'\\' );
	if ( fileName == NULL )
		return;

	fileName++;
	if ( wcslen( fileName ) < MAX_PATH )
		return;

	static int count = 0;

	do
	{
		FormatT( fileName + MAX_PATH - 20, 10, L"_%04X.txt", count++ );
	}
	while( _waccess( path, 0 ) == 0 );
}

bool ODS_SetFileName( const wchar_t *sParam )
{
	cODSObject *obj = cODSObject::getCurrent();

	CreateDirectoryW( obj->getODSDir(), NULL );

	wchar_t bf[ MAX_PATH ] = L"";
	GetModuleFileNameW( obj->m_hODSInst, bf, maxLen( bf ) );
	//int pos = splitPath( bf );

	wchar_t *moduleName = wcsrchr( bf, L'\\' );
	if ( moduleName == NULL )
		moduleName = bf;
	else
		moduleName++;

	if ( sParam == NULL )
		obj->m_wcsModulePath.FormatT( L"%s\\%s.txt", obj->getODSDir(), moduleName );
	else
	{
		cWCSF wcs;
		wcs.FormatT( L"%s-%s", moduleName, sParam );
		obj->m_wcsOriginalName.FormatT( wcs.m_szBody );

		updatePathSymbols( wcs.m_szBody );
		obj->m_wcsModulePath.FormatT( L"%s\\%s.txt", obj->getODSDir(), wcs.m_szBody );
	}

	updateFileName( obj->m_wcsModulePath.m_szBody );

	//(*getRecCountPtr()) = 0;
	
#ifdef _ODSM
	if ( pProcessStat != NULL )
		pProcessStat->setFirstCall();
#endif

	bool bExist = _waccess( obj->m_wcsModulePath.m_szBody, 0 ) == 0;

	return bExist;
}

void ODS_CreateNewFile( bool enable )
{
	cODSObject *obj = cODSObject::getCurrent();

	if ( enable )
	{
		cWCSF old;
		old.FormatT( L"Old-%s", obj->m_wcsModulePath.m_szBody );
		MoveFileExW( obj->m_wcsModulePath.m_szBody, old.m_szBody, MOVEFILE_REPLACE_EXISTING );
		obj->m_wcsModulePath.SetLength(0);
		ODS_SetFileName( NULL );
		ODS( L"ODS_CreateNewFile" );
	}
}

// static
const wchar_t *cODS::getTimeString( cWCSF &wcs, __int64 mSec )
{
	if ( mSec < 1000 )
		return wcs.FormatT( L"%dms", (int)mSec );

	__int64 sec = mSec / 1000;
	
	if ( sec < 60 )
	{
		if ( sec < 10 )
			return wcs.FormatT( L"%ds %dms", (int)sec, (int)(mSec % 1000) );
		else
			return wcs.FormatT( L"%ds", (int)sec );
	}

	int min = (int)(sec / 60);
	if ( min < 60 )
	{
		if ( min < 10 )
			return wcs.FormatT( L"%dm %ds", min, (int)(sec % 60) );
		else
			return wcs.FormatT( L"%dm", min );
	}

	int hour = min / 60;

	if ( hour < 10 )
		return wcs.FormatT( L"%dh %dm", hour, min % 60 );

	return wcs.FormatT( L"%dh", hour );
}

void cODS::set( const wchar_t *name, const wchar_t *params, bool bForce ) 
{
	m_bForce = bForce;
	m_szRes = NULL;
	m_wcsName.Set( name );
	outString( true, params );

	(*getRecCountPtr())++;
}

cODS::~cODS()
{
	(*getRecCountPtr())--;

	outString( false );
}

void ODS( const wchar_t *s0, int i, const wchar_t *s1, bool forceEnable )
{
	cWCSF wcs( s0 );

	if ( i != -1 )
		wcs.AddF( L":%d", i );

	ODS( wcs.m_szBody, s1, forceEnable );
}

class cDll
{
	HINSTANCE m_hInst;

public:

	void *m_pProc;

	cDll( const char *dllName, const char* procName )
	{
		m_hInst = LoadLibraryA( dllName );
		m_pProc =
			m_hInst == NULL ?
				NULL :
				GetProcAddress( m_hInst, procName );
	}

	~cDll()
	{
		if ( m_hInst != NULL )
			FreeLibrary( m_hInst );
	}
};

//
//	Veracode reports
//	CWE-73: External Control of File Name or Path
//	CWE-22: Improper Limitation of a Pathname to a Restricted Directory ('Path Traversal')
//	So we check path validity before CreateFileW call.
//
static bool checkValidPath( const wchar_t *path )
{
	if ( wcsstr( path, L"\\.." ) != NULL ) return false;
	if ( wcsstr( path, L"/.." ) != NULL ) return false;

	return true;
}

static void odsCS_tc( const wchar_t *s0, const wchar_t *s1, bool forceEnable )
{
	cODSObject *obj = cODSObject::getCurrent();

	if ( !obj->m_enable && !forceEnable )
		return;

	wchar_t msg[ 1024 ];

	if ( obj->m_wcsModulePath.isEmpty() )
		ODS_SetFileName( NULL );

	const wchar_t *szProcessID = L"";

	static bool bCreateNew = true;  // not multithread create new file per session. The flag is set false afte creation.

#ifdef _ODSM
	static cODSMProcess processStat;
	pProcessStat = &processStat;

	cWCSF wcsPID;
	szProcessID = wcsPID.FormatT( L"%04X:%04X ", GetCurrentProcessId(), GetCurrentThreadId() );

	bCreateNew = false; // Multithread mode continues write to existing file.
#endif

	cWCSF threadFileName( obj->m_wcsModulePath.m_szBody );

	int nameThreadN = getNamedThread();

	if ( nameThreadN != -1 )
	{
		if ( obj->m_namedThreadT[ nameThreadN ].path.isEmpty() )
		{
			threadFileName.m_len -= 4;
			threadFileName.AddF( L"-Thread-%s", obj->m_namedThreadT[ nameThreadN ].name.m_szBody );
			wchar_t *moduleName = wcsrchr( threadFileName.m_szBody, L'\\' );

			if ( moduleName != NULL )
			{
				moduleName++;
				obj->m_namedThreadT[ nameThreadN ].originalName.FormatT( moduleName );
				updatePathSymbols( moduleName );
			}

			obj->m_namedThreadT[ nameThreadN ].path.FormatT( L"%s.txt", threadFileName.m_szBody );

			updateFileName( obj->m_namedThreadT[ nameThreadN ].path.m_szBody );
		}

		threadFileName.Set( obj->m_namedThreadT[ nameThreadN ].path.m_szBody );			
	}

	if ( !checkValidPath( threadFileName.m_szBody ) )
	{
		assert( false );
		return;
	}

	static int cnt = 0;
	cnt++;

	int tryCount;
	DWORD err = 0;

	HANDLE hFile = INVALID_HANDLE_VALUE;

	// Try to wait.
	for( tryCount = 3; tryCount--; )
	{
		hFile = CreateFileW( 
			threadFileName.m_szBody,
			GENERIC_READ | GENERIC_WRITE,
			FILE_SHARE_READ,
			NULL,
			bCreateNew ? CREATE_ALWAYS : OPEN_ALWAYS, //count++ ? OPEN_ALWAYS : CREATE_ALWAYS,
			0,
			NULL );

		if ( hFile != INVALID_HANDLE_VALUE )
			break;

		size_t len = wcslen( threadFileName.m_szBody );
		if ( len >= MAX_PATH )
		{
			cLongName ln;
			ln.generate( threadFileName.m_szBody );
			hFile = CreateFileW( 
				ln.m_szBody,
				GENERIC_READ | GENERIC_WRITE,
				FILE_SHARE_READ,
				NULL,
				bCreateNew ? CREATE_ALWAYS : OPEN_ALWAYS, //count++ ? OPEN_ALWAYS : CREATE_ALWAYS,
				0,
				NULL );
		}
		
		if ( hFile != INVALID_HANDLE_VALUE )
			break;

		err = GetLastError();

		Sleep( 100 );
	}

	// Try to another files.
	for( tryCount = 0; tryCount < 5; tryCount++ )
	{
		if ( hFile != INVALID_HANDLE_VALUE )
			break;
		
		cWCSF name;
		name.FormatT( L"%s_%d", threadFileName.m_szBody, tryCount );

		if ( !checkValidPath( threadFileName.m_szBody ) )
		{
			assert( false );
			return;
		}

		hFile = CreateFileW( 
			cLongName().generate( name.m_szBody ),
			GENERIC_READ | GENERIC_WRITE,
			FILE_SHARE_READ,
			NULL,
			bCreateNew ? CREATE_ALWAYS : OPEN_ALWAYS, //count++ ? OPEN_ALWAYS : CREATE_ALWAYS,
			0,
			NULL );

		if ( hFile != INVALID_HANDLE_VALUE )
			break;

		err = GetLastError();

		Sleep( 100 );
	}

	if ( hFile == INVALID_HANDLE_VALUE )
	{
		static bool asked = false;
		if ( asked )
			return;

		asked = true;

		//MessageBoxW( 0,
		//	_FormatT(
		//		L"Error  - %d,\n"
		//		L"File name - %s",
		//		err, threadFileName.m_szBody ),
		//	L"Error in opening ODS file",
		//	MB_ICONSTOP );

		return;
	}
	
	bCreateNew = false;

	wchar_t timeBf[ 100 ];

#define __fullTime__

#ifdef __fullTime__
	getTimeStr( timeBf, maxLen( timeBf ) );
#else
	getSecondsStr( timeBf, maxLen( timeBf ) );
#endif

	wchar_t prefix[ 100 ];
	int i = 0;
	for ( int q = *getRecCountPtr(); i < q; i++ )
	{
		if ( i >= sizeof( prefix ) / sizeof( *prefix ) - 1 )
			break;
		
		prefix[i] = L' ';
	}

	prefix[i] = L'\0';

	FormatT( msg, maxLen( msg ), L"%s%s%s %s %s\r\n", szProcessID, timeBf, prefix, s0, s1 );

	DWORD dwWritten = 0;

	static wchar_t staticTimeBf[ 100 ] = L"";

	DWORD dwPos = SetFilePointer( hFile, 0, NULL, FILE_END );

	if ( 
#ifdef _ODSM
		( ( nameThreadN != -1 ) && obj->m_namedThreadT[ nameThreadN ].isFirstCall() ) ||
		( ( nameThreadN == -1 ) && processStat.isFirstCall() && ( cODSObject::m_external == NULL ) )
#else
		( dwPos == 0 ) 
#endif
		)
	{	
		if ( dwPos == 0 )
		{
			BYTE unicodePrefix[] = { 0xFF, 0xFE };
			WriteFile( hFile, unicodePrefix, sizeof( unicodePrefix ), &dwWritten, NULL );

			OSVERSIONINFOEXW osvi;
			osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFOEX);

			BOOL bOsVersionInfoEx = GetVersionExW( (OSVERSIONINFOW *)&osvi );

			cWCSF wcs;

			if ( !obj->m_wcsOriginalName.isEmpty() )
				wcs.AddF( L"Original name: %s", obj->m_wcsOriginalName.m_szBody );

			int nameThreadN = getNamedThread();

			if ( nameThreadN != -1 )
			{
				if ( !obj->m_namedThreadT[nameThreadN].originalName.isEmpty() )
					wcs.AddF( L"\\%s", obj->m_namedThreadT[nameThreadN].originalName.m_szBody );
			}
			
			if ( !wcs.isEmpty() )
				wcs.Add( L"\r\n" );				 

			if ( bOsVersionInfoEx )
				wcs.AddF( L"System: %d.%d (build %d) %s, Platform ID %X\r\n", osvi.dwMajorVersion, osvi.dwMinorVersion, osvi.dwBuildNumber, osvi.szCSDVersion, osvi.dwPlatformId );
			else
			{
				osvi.dwOSVersionInfoSize = sizeof (OSVERSIONINFO);
				if ( !GetVersionEx ( (OSVERSIONINFO *) &osvi) ) 
					wcs.Add( L"Error in getting windows version\r\n" );
				else
					wcs.AddF( L"System: %d.%d (build %d), Platform ID %X\r\n", osvi.dwMajorVersion, osvi.dwMinorVersion, osvi.dwBuildNumber, osvi.dwPlatformId );
			}
			
			WriteFile( hFile, wcs.m_szBody, (DWORD)wcslen( wcs.m_szBody ) * sizeof( wchar_t ), &dwWritten, NULL );
		}

		getDateTimeStr( staticTimeBf, maxLen( staticTimeBf ) );

		wchar_t moduleName[ MAX_PATH ] = L"";
		wchar_t userName[ MAX_PATH ] = L"";
		DWORD dwLen = maxLen( userName );
		
#ifdef _UnicodeFuncs_
		GetUserNameExT( 12, userName, &dwLen );
#else

		static cDll ui( "secur32.dll", "GetUserNameExW" );

		BOOL ok = false;
		DWORD lastErr = 0;

		if ( ui.m_pProc != NULL )
		{
			typedef BOOLEAN WINAPI typeGetUserNameExW( int nameFormat, LPWSTR lpNameBuffer, PULONG nSize  );
			ok = ((typeGetUserNameExW * )ui.m_pProc)( 12, userName, &dwLen );
			if ( !ok )
				lastErr = GetLastError();
		}

		if ( !ok )
		{
			//
			//	Windows 8 reports error ERROR_NO_SUCH_DOMAIN - 1355
			//
			dwLen = maxLen( userName );
			ok = GetUserNameW( userName, &dwLen );
		}
#endif
		GetModuleFileNameW( obj->m_hODSInst, moduleName, maxLen( moduleName ) );
		cWCSF path;
		path.FormatT(
			L"\r\n%s Started - '%s', User - '%s', %s\r\n",
			szProcessID, moduleName, userName, staticTimeBf );

		getMinuteStr( staticTimeBf, maxLen( staticTimeBf ) );

		int len = (int)wcslen( path );
		dwWritten = 0;
		WriteFile( hFile, path, len * sizeof( wchar_t ), &dwWritten, NULL );
	}
	else
		SetFilePointer( hFile, 0, NULL, FILE_END );

//	if ( cODS::m_enable || forceEnable )
	{
#ifndef __fullTime__
		getMinuteStr( timeBf, maxLen( timeBf ) );
		if ( wcscmp( timeBf, staticTimeBf ) != 0 )
		{
			wcscpy_s( staticTimeBf, timeBf );

			getDateTimeStr( timeBf, maxLen( timeBf ) );
			wchar_t date[ 512 ];
			FormatT( date, maxLen( date ), L"Date %s\r\n", timeBf );
			int len = (int)wcslen( date );
			dwWritten = 0;
			WriteFile( hFile, date, len * sizeof( *date ), &dwWritten, NULL );
		}
#endif

		int len = (int)wcslen( msg );
		dwWritten = 0;

		WriteFile( hFile, msg, len * sizeof( *msg ), &dwWritten, NULL );
		OutputDebugStringW( msg );
	}

	CloseHandle( hFile );
}

static void odsCS( const wchar_t *s0, const wchar_t *s1, bool forceEnable )
{
	try
	{
		odsCS_tc( s0, s1, forceEnable );
	}
	catch(...)
	{
		assert( false );
	}
}


static DWORD _GetThreadId( HANDLE hThread )
{
	typedef DWORD WINAPI typeGetThreadId( __in HANDLE Thread );

	static typeGetThreadId *proc = (typeGetThreadId *)GetProcAddress( LoadLibraryA( "kernel32.dll" ), "GetThreadId" );

	if ( proc == NULL )
		return (DWORD)-1;

	return (proc)( hThread );
}

DWORD ODS_waitForSingleObject( const wchar_t *sFunc, int iLine, HANDLE hHandle, DWORD dwMilliseconds, const wchar_t *name )
{
	iLine;
	sFunc;

	const DWORD dwODSTime = 10 * 1000; // 10 ms
	DWORD timeLeft = dwMilliseconds;

	DWORD dwRes = WAIT_OBJECT_0;
	DWORD dwId = _GetThreadId( hHandle );

	if ( dwId == 0 )
		ODS( sFunc, iLine, _FormatT( L"Assertion! Thread 0x%04X identifier is not detected, error %d", hHandle, GetLastError() ) );

	while( true )
	{
		DWORD time = min( dwODSTime, timeLeft );

		dwRes = WaitForSingleObject( hHandle, time );

		if ( timeLeft != INFINITE )
			timeLeft -= time;

		if ( dwRes != WAIT_TIMEOUT )
			break;

		if (  timeLeft == 0 )
			break;

		ODS( sFunc, iLine, _FormatT( L"Timeout in waiting handle 0x%04X ( ID = 0x%04X ) %s", hHandle, dwId, ( name ? name : L"" ) ) );
	}

	return dwRes;
}

void ODS( const wchar_t *s0, const wchar_t *s1, bool forceEnable )
{
	//if ( !cODS::m_enable && !forceEnable )
	//	return;

	static cSysCriticalSection cs; // CRITICAL_SECTION cs;

	EnterCriticalSection( &cs );

	odsCS( s0, s1, forceEnable );

	LeaveCriticalSection( &cs );
}

void ODS_SetInstance( HINSTANCE hInst )
{
	cODSObject::getCurrent()->m_hODSInst = hInst;
	cODSObject::getCurrent()->m_wcsModulePath.SetLength(0);
}

bool ODS_Enable( bool enable )
{
	bool bOld = cODSObject::getCurrent()->m_enable;
	cODSObject::getCurrent()->m_enable = enable;
	return bOld;
}

bool ODS_Verbose( bool bVerbose )
{
	bool bOld = cODSObject::getCurrent()->m_verbose;
	cODSObject::getCurrent()->m_verbose = bVerbose;
	return bOld;
}

#endif // _ODS

