#pragma once

#include <windows.h>
#include <time.h>

// Output debug strings
//#ifdef _DEBUG
//#	ifndef _ODS
//#		define _ODS
//#	endif
//#endif

// _ODSM allows to write module handle in ODS function for multy module debuging.	
#ifdef _ODSM
#define _ODS
#endif

#define ODS_fls( X )ODS( (_CRT_WIDE( __FUNCTION__)), __LINE__, (X) )
#define ODS_fl		ODS( (_CRT_WIDE( __FUNCTION__)), __LINE__ )
#define ODS_f( X )	ODS( (_CRT_WIDE( __FUNCTION__)), -1, (X) )

#define _FL (_CRT_WIDE( __FUNCTION__ )), __LINE__
#ifdef _ODS

extern const wchar_t *getODSDir();

#include "cWCSF.h"


HANDLE
CreateFileODS( const wchar_t *szFunc, int iLine,
    LPCWSTR lpFileName,
    DWORD dwDesiredAccess,
    DWORD dwShareMode,
    LPSECURITY_ATTRIBUTES lpSecurityAttributes,
    DWORD dwCreationDisposition,
    DWORD dwFlagsAndAttributes,
    HANDLE hTemplateFile
    );

//#include "UnicodeFuncs.h"

#ifndef _CRT_SECURE_NO_WARNINGS
#pragma warning(disable : 4996)
#endif


#ifndef _WIN32
//
//	Windows compiler does not allow formal parameters for macros...
//
#define ODSFL(format, args ...) \
	ODSF( L"%s::%d " #format, _W( __FUNCTION__ ), __LINE__, ##args);
#endif

void ODS( const wchar_t *s0, const wchar_t *s1 = L"", bool forceEnable = false );
void ODS_SetInstance( HINSTANCE hInst );
void ODSF( const wchar_t *pattern, ... );
void ODSE( const wchar_t *s0, const wchar_t *s1 = L"" );
void ODSFE( const wchar_t *pattern, ... );
void ODS( const wchar_t *s0, int i, const wchar_t *s1 = L"", bool forceEnable = false );
//void ODS_SetThreadName( const wchar_t *name );
#define ODS_SetThreadName( _X ) cODSThread __odsThread( _X )

bool ODS_SetExternalObject( HMODULE hInst );

class cODSThread
{
public:
	cODSThread( const wchar_t *name );
	~cODSThread();
};

//
//	ODS_CreateNewFile
//	If the function is called with true new file is created
//	If the function is called with false at application start it disable creating new file
//
void ODS_CreateNewFile( bool enable = true );
//bool ODS_GetFileName( wchar_t *path, int len );

//
//	ODS_SetFileName
//		Allows to change ame of ODS file on fly.
//		It uses 'executable name'-'sParam'
//
bool ODS_SetFileName( const wchar_t *sParam );

//
//	Enable/disable output string,
//
bool ODS_Enable( bool enable = true );
bool ODS_Verbose( bool bVerbose = true );

class cODS
{
	//wchar_t m_name[ 1024 ];
	//wchar_t m_fmtRes[ 128 ];
	cWCS m_wcsName;
	cWCS m_wcsRes;
	const wchar_t *m_szRes;
	bool m_bForce;
	__int64 m_clocks;

public:


	//static bool m_enable;
	//static bool m_verbose;

	static const wchar_t *getTimeString( cWCSF &wcs, __int64 mSec );

	void outString( bool start, const wchar_t *params = L"" )
	{
		typedef void typeODSF( const wchar_t *pattern, ... );

		typeODSF *funcODSF = m_bForce ? ODSFE : ODSF;

		//if ( start )
		//	funcODSF( L"<Start> %s%s", m_name, params );
		//else
		//	funcODSF( L"<Stop>  %s%s%s", m_name, (m_szRes == 0 ? L"" : L" = "), (m_szRes == 0? L"" : m_szRes) );
		if ( start )
		{
			m_clocks = clock();
			funcODSF( L"<Start> %s%s", m_wcsName.m_szBody, params );
		}
		else
		{
			__int64 diff = clock() - m_clocks;
			cWCSF wcs;

			funcODSF( L"<Stop>  %s%s%s - %s", m_wcsName.m_szBody, (m_szRes == 0 ? L"" : L" = "), (m_szRes == 0? L"" : m_szRes), getTimeString( wcs, diff * 1000 / CLOCKS_PER_SEC ) );
		}
	}

	void set( const wchar_t *name, const wchar_t *params = L"", bool bForce = false );
	
	void clear()
	{
		//memset( this, 0, sizeof( *this ) );
		m_bForce = false;
		m_szRes = NULL;
	}

	cODS() { clear(); } // 'set' function has to be called.
	cODS( const wchar_t *name, bool bForce = false ) {  clear(); set( name, L"", bForce ); }
	cODS( const wchar_t *name, const wchar_t *s = (const wchar_t *)(-1), bool bForce = false);

	~cODS();
	
	void setResult( const wchar_t *s ) //{ m_szRes = s; }
	{
		//if ( s != m_fmtRes )
		//{
		//	int maxL = sizeof( m_fmtRes )/sizeof( *m_fmtRes ) - 1;
		//	wcsncpy( m_fmtRes, s, maxL );
		//	m_fmtRes[ maxL ] = L'\0';
		//}

		//m_szRes = m_fmtRes;

		m_wcsRes.Set( s );
		m_szRes = m_wcsRes.m_szBody;
	}

	void setResult( int r );
	void setResult( DWORD r );
	void setResult( bool r );
	void setResult( __int64 r );
};

class cODSEnableBlock
{
	bool m_bOld;

public:
	cODSEnableBlock( bool bEnable )
	{
		m_bOld = ODS_Enable( bEnable );
	}

	~cODSEnableBlock()
	{
		ODS_Enable( m_bOld );
	}
};

bool isCurObjVerbose();

#define ODS_vfl			{ if ( isCurObjVerbose() ) ODS_fl;}
#define ODS_vfls( X )	{ if ( isCurObjVerbose() ) ODS_fls( (X) ); }

#define ODS_EnableBlock( __x ) cODSEnableBlock __ODSEnableBlock( __x )

#define ODSFuncX( __i ) ODSFuncS( _FormatT( L"%X", (int)(__i) ) )
#define ODSFuncI( __i ) ODSFuncS( _FormatT( L"%d", (int)(__i) ) )
#define ODSFuncS( __x ) cODS ___ods( _CRT_WIDE( __FUNCTION__ ), __x, false )
#define ODSBlockS( __blockName, __str ) cODS ___ods( __blockName, __str, false )
#define ODSFunc cODS ___ods( _CRT_WIDE( __FUNCTION__ ), false )
#define ODSFuncRes( __xr ) ___ods.setResult( __xr )

#define ODSFuncSE( __x ) cODS ___ods( _CRT_WIDE( __FUNCTION__ ), __x, true )
#define ODSFuncE cODS ___ods( _CRT_WIDE( __FUNCTION__ ), L"", true )

DWORD ODS_waitForSingleObject( const wchar_t *sFunc, int iLine, HANDLE hHandle, DWORD dwMilliseconds, const wchar_t *name = NULL );
#define WaitForSingleObject_withODS( _hTread, _dwTimeout, _name ) ODS_waitForSingleObject( _CRT_WIDE( __FUNCTION__ ), __LINE__, _hTread, _dwTimeout, _name )

extern void dbgWrite( int &cnt, const wchar_t *pref, const wchar_t *szSid, const char *body, int size, int err = 0, __int64 clocks = 0 );

#else

#define WaitForSingleObject_withODS( _hTread, _dwTimeout, name ) WaitForSingleObject( _hTread, _dwTimeout )
#define ODS_SetExternalObject( _x ) ((void)0)
#define ODS_SetThreadName( __x )	((void)0)
#define ODSFuncX( __x )				((void)0)
#define ODSFuncI( __x )				((void)0)
#define ODSFuncS( __x )				((void)0)
#define ODSBlockS( __name, __str )	((void)0)
#define ODSFunc						((void)0)
#define ODSFuncRes( __xr )			((void)0)
#define ODS( ... )					((void)0)
#define ODSF( ___x, ... )			((void)0)
#define ODSE( ___x )				((void)0)
#define ODSFE( ___x )				((void)0)
#define ODS_SetInstance( ___x )		((void)0)
#define ODS_Enable( ____x )			((void)0)
#define ODS_Verbose( ____x )		((void)0)
#define ODS_CreateNewFile( ____x )	((void)0)
#define ODS_SetFileName( ____x )	((void)0)
#define ODS_EnableBlock( __x )		((void)0)
#define ODS_vfl						((void)0)
#define ODS_vfls( X )				((void)0)
#define dbgWrite( ... )				((void)0)

#endif // _ODS
