#include "OpenFiles.h"
#include <commdlg.h>
#include <io.h>
#include <assert.h>
#include "shell32Dll.h"

//
//	Initial Path assigment...
//
static wchar_t szInitialPath[ MAX_PATH ] = L"";

bool BF_SetInitialPath( wchar_t *buffer, int nFileOffset )
{
	if ( nFileOffset )
	{
		nFileOffset = (int)wcslen( buffer );
	}

	if ( nFileOffset >= maxLen( szInitialPath ) )
		nFileOffset = maxLen( szInitialPath ) - 1;

	wcsncpy_s( szInitialPath, buffer, nFileOffset );
	szInitialPath[ nFileOffset ] = 0;

	return TRUE;
}

//
//	Note, index 0 specify using of lpstrCustomFilter
//	index 1 specify using first filter and so on.
//
static int lastFilterIndex = 0;

void BF_SetLastFilterIndex( int index )
{
	lastFilterIndex = index;
}

int BF_GetLastFilterIndex( const wchar_t *filterPairs )
{
	if ( lastFilterIndex > 0 )
	{
		if ( filterPairs == NULL )
			return lastFilterIndex;

		int count = 0;

		for( count = 0; count < 10; count++ )
		{
			if ( *filterPairs == '\0' )
				break;

			filterPairs += wcslen( filterPairs ) + 1;
			filterPairs += wcslen( filterPairs ) + 1;
		}

		if ( lastFilterIndex > count )
			return 0;
	}

	return lastFilterIndex;
}

wchar_t *BF_GetInitialPathW( HWND hWnd )
{
	if ( ( szInitialPath[0] == 0 ) || !FileExistsT( szInitialPath ) )
	{
		cShell32Dll &dll = cShell32Dll::getStatic();

		if ( dll.error != cShell32Dll::errNo )
			return NULL;

		if ( !dll.m_SHGetSpecialFolderPathW( hWnd, szInitialPath, CSIDL_PERSONAL, false ) )
			return NULL;
	}

	return szInitialPath;
}

static const wchar_t *findFilterExt( const wchar_t *filter, int index )
{
	const wchar_t *pos = filter;

	if ( pos == NULL )
		return NULL;

	while( index-- )
	{
		pos += wcslen( pos ) + 1;
		pos += wcslen( pos ) + 1;
	}

	while( *pos != L'.' )
	{
		if ( *pos == L'*' )
			return NULL;

		if ( filter > pos-- )
			return NULL;
	}

	return pos;
}

//
//	updateFiltersSring
//		change filter string f.e.
//			from "Text (*.txt)%*.TXT%All (*.*)%*.*%"
//			to	 "Text (*.txt)\0*.TXT\0All (*.*)\0*.*\0"
//

const wchar_t fiterSeparator = L'%';

static const wchar_t *updateFiltersSring( cWCStr *wcs, const wchar_t *filters )
{
	size_t size = 0;
	if ( filters == NULL )
		return filters;

	size = wcslen( filters ) + 1;

	if ( wcs->SetLength( (int)size ) == 0 )
		return filters;

	bool bFound = false;

	for( size_t i = size; i--; )
	{
		if ( filters[i] == fiterSeparator )
		{
			wcs->m_szBody[i] = L'\0';
			bFound = true;
		}
		else
			wcs->m_szBody[i] = filters[i];
	}

	if ( bFound )
		return wcs->m_szBody;
	else
		return filters;
}

bool BrowseFiles( HWND hWnd, wchar_t *buff, int buffSize, const wchar_t *filters, const wchar_t *szTitle, bool allowMultiSelect, DWORD additionalFlags )
{
	OPENFILENAMEW ofn;
	memset(&ofn,0,sizeof(ofn));

	ofn.lStructSize = sizeof(ofn);
	ofn.hwndOwner = hWnd;

	ofn.lpstrFileTitle = NULL;
	ofn.nMaxFileTitle = 0;

	wchar_t initialDir[ MAX_PATH ] = L"";

	if ( buff[0] != L'\0' )
	{
		int pos = splitPath( buff, initialDir );
		
		if ( pos != 0 )
		{
			BF_SetInitialPath( initialDir, (int)wcslen( initialDir ) );
			memmove( buff, (BYTE*)(buff + pos), (int)wcslen( buff + pos ) * sizeof( wchar_t ) );
		}

		ofn.lpstrInitialDir = initialDir;
	}

	if ( ( wcslen( initialDir ) == 0 ) || !FileExistsT( initialDir ) )
		ofn.lpstrInitialDir = BF_GetInitialPathW( hWnd );

	cWCStr wcsBf;

	filters = updateFiltersSring( &wcsBf, filters );

	ofn.lpstrFilter = filters;
	ofn.nFilterIndex = BF_GetLastFilterIndex( filters );

	ofn.lpstrFile = buff;
	buff[0] = L'\0'; // Otherwise W7 returns error FNERR_INVALIDFILENAME.

	ofn.lpstrTitle = szTitle; // F.e. "Create new archive"

	ofn.nMaxFile = buffSize;

	ofn.Flags = 
		OFN_EXPLORER |
//		OFN_HIDEREADONLY |
//		OFN_ENABLETEMPLATEHANDLE |
//		OFN_ENABLETEMPLATE |
//		OFN_ENABLEHOOK |
//		OFN_DONTADDTORECENT |
		OFN_NOCHANGEDIR |
		OFN_NONETWORKBUTTON |
		OFN_OVERWRITEPROMPT |
		OFN_ENABLESIZING |
		OFN_LONGNAMES |
		additionalFlags;

	if ( allowMultiSelect )
		ofn.Flags |= OFN_ALLOWMULTISELECT;

	bool ok;
	if ( additionalFlags & OFN_OVERWRITEPROMPT ) 
		ok = GetSaveFileNameW( &ofn ) == TRUE;
	else
		ok = GetOpenFileNameW( &ofn ) == TRUE;

	const char *errCaption = "Error in opening file";

	if ( !ok )
	{
		DWORD err = CommDlgExtendedError();
		
		switch( err )
		{
			case 0: // Canceled...
				return FALSE;

			case FNERR_BUFFERTOOSMALL:
				MessageBoxA( hWnd, "Buffer too small", errCaption, MB_ICONSTOP );
				break;
			default:
				MessageBoxA( hWnd, "Undefined error", errCaption, MB_ICONSTOP );
				break;
		}
	}

	if ( ofn.nFileExtension == 0 )
	{
		if ( !FileExistsT( buff ) && ( ofn.lpstrFilter != NULL ) ) //_waccess( buff, 0 ) != 0 ) // does not exist.
		{
			const wchar_t *ext = findFilterExt( ofn.lpstrFilter, ofn.nFilterIndex );

			if ( ext )
				wcscat_s( buff, buffSize, ext  );
		}
	}

	BF_SetInitialPath( buff, ofn.nFileOffset );
	BF_SetLastFilterIndex( ofn.nFilterIndex );

	return ok;
}

//
//	BF_isSinglePathBuffer
//		Just check szzBuffer ( double zero terminated string  ) according output buffer format ( see description 4. )
//
bool BF_isSinglePathBuffer( const wchar_t *szzBuff )
{
	if ( *szzBuff == L'\0' )
		return false;
	
	szzBuff = BF_getNextString( szzBuff );

	return ( *szzBuff == L'\0' );
}

//
//	BF_getNextString
//		Finds first zero and return pointer next symbol.
//		If the symbol is zero it is end of double zero terminated buffer.
//
const wchar_t *BF_getNextString( const wchar_t *szzStr )
{
	// Really invalid sting pointer upto first zero symbol.
	if ( IsBadStringPtrW( szzStr, (UINT_PTR)-1 ) )
	{
		ODS_fls( L"Assertion!" );
		return L"";
	}

	// Called process has to check this at previous call.
	if ( *szzStr == L'\0' )
	{
		ODS_fls( L"Assertion!" );
		return L"";
	}

	for( int i = 0;; )
	{
		if ( szzStr[i++] == L'\0' )
			return szzStr + i;
	}
}

//void convertMultiStringBuffer( char *buffA, const wchar_t *buffW, int buffSize )
//{
//	int len = 0;
//
//	char prev = -1;
//
//	while( buffSize )
//	{
//		wctomb_s( &len, buffA, buffSize, *buffW );
//		buffSize -= len;
//
//		if ( ( *buffA == 0 ) && ( prev == 0 ) )
//			return;
//		
//		prev = *buffA;
//		buffA++;
//		buffW++;
//	}
//
//	//int len;
//	//while( len = _wcstombs( buffA, buffW, buffSize ) )
//	//{
//	//	len++;
//	//	buffA = buffA + len;
//	//	buffW = buffW + len;
//	//	buffSize -= len;
//
//	//	if ( buffSize <=0 )
//	//	{
//	//		assert( false );
//	//		break;
//	//	}
//	//}
//}
//
//void convertMultiStringBuffer( wchar_t *buffW, const char *buffA, int buffSize )
//{
//	int len = 0;
//
//	char prev = -1;
//
//	while( buffSize-- )
//	{
//		mbtowc( buffW, buffA, 1 );
//
//		if ( ( *buffA == 0 ) && ( prev == 0 ) )
//			return;
//		
//		prev = *buffA;
//		buffA++;
//		buffW++;
//	}
//
//	//int len;
//	//while( len = mbstowcs( buffW, buffA, buffSize ) )
//	//{
//	//	len++;
//	//	buffA = buffA + len;
//	//	buffW = buffW + len;
//	//	buffSize -= len;
//
//	//	if ( buffSize <=0 )
//	//	{
//	//		assert( false );
//	//		break;
//	//	}
//	//}
//}