#include "OpenFilesA.h"
#include <commdlg.h>
#include <io.h>
#include <assert.h>
//#include "UnicodeFuncs.h"
//#include "AutoDeallocate.h"
//#include "shell32Dll.h"

const char *szStdFilters = 
		"Text files (*.txt)\0*.txt\0"
		"Documents (*.doc;*.xls;*.rtf)\0*.doc;*.xls;*.rtf\0"
		"Images (*.jpg;*.gif;*.bmp;*.tif)\0*.jpg;*.gif;*.bmp;*.tif\0"
		"Executable (*.exe;*.dll;*.com;*.bat;*.inf;*.lnk)\0*.exe;*.dll;*.com;*.bat;*.inf;*.lnk\0"
		"All files (*.*)\0*.*\0";

//BOOL BF_SetInitialPath( char *buffer, int nFileOffset )
//{
//	cWCStr wcs( buffer );
//	return BF_SetInitialPath( wcs.m_szBody, nFileOffset );
//}

static char szInitialPath[ MAX_PATH ] = "";

#ifndef maxLen
#define maxLen( buff ) ( sizeof( buff ) / sizeof( *buff ) )
#endif

BOOL BF_SetInitialPath( char *buffer, int nFileOffset )
{
	if ( nFileOffset )
		nFileOffset = (int)strlen( buffer );

	if ( nFileOffset >= maxLen( szInitialPath ) )
		nFileOffset = maxLen( szInitialPath ) - 1;

	strncpy_s( szInitialPath, buffer, nFileOffset );
	szInitialPath[ nFileOffset ] = 0;

	return TRUE;
}

static const char *findFilterExt( const char *filter, int index )
{
	const char *pos = filter;

	while( index-- )
	{
		pos += strlen( pos ) + 1;
		pos += strlen( pos ) + 1;
	}

	while( *pos != '.' )
	{
		if ( *pos == '*' )
			return NULL;

		if ( filter > pos-- )
			return NULL;
	}

	return pos;
}

static char *BF_GetInitialPath( )
{
	if ( szInitialPath[0] == 0 )
		return NULL;

	return szInitialPath;
}

BOOL BrowseFiles(HWND hWnd, char *buff, int buffSize, const char *filters, const char *szTitle, BOOL allowMultiSelect, DWORD additionalFlags )
{
	OPENFILENAMEA ofn;
	memset(&ofn,0,sizeof(ofn));

	ofn.lStructSize = sizeof(ofn);
	ofn.hwndOwner = hWnd;
//	ofn.hInstance = hInstance;

	ofn.lpstrFileTitle = NULL; // Buffer for name without path...
	ofn.nMaxFileTitle = 0;

	if ( buff[0] != '\0' )
		BF_SetInitialPath( buff, (int)strlen( buff ) );

	ofn.lpstrInitialDir = BF_GetInitialPath();

	ofn.lpstrFilter = filters;
	ofn.lpstrFile = buff;
	memset( buff, 0, buffSize ); // buff[0] = 0; // No Initialization required...

	ofn.lpstrTitle = szTitle; // F.e. "Create new archive"

	ofn.nMaxFile = buffSize;

	ofn.Flags = 
		OFN_EXPLORER |
//		OFN_HIDEREADONLY |
//		OFN_ENABLETEMPLATEHANDLE |
//		OFN_ENABLETEMPLATE |
		OFN_NOCHANGEDIR |
//		OFN_ENABLEHOOK |
		OFN_NONETWORKBUTTON |
//		OFN_DONTADDTORECENT |
		OFN_OVERWRITEPROMPT |
		OFN_ENABLESIZING |
		OFN_LONGNAMES;

	if ( allowMultiSelect )
		ofn.Flags |= OFN_ALLOWMULTISELECT;

//	ofn.lpTemplateName = MAKEINTRESOURCE(IDD_OpenDlg);
//	ofn.lpfnHook = (LPOFNHOOKPROC)OFNHookProc;

	BOOL ok;

	if ( additionalFlags & OFN_OVERWRITEPROMPT ) 
		ok = GetSaveFileNameA( &ofn );
	else
		ok = GetOpenFileNameA( &ofn );

	const char *errCaption = "Error in opening file";

	if ( !ok )
	{
		DWORD err = CommDlgExtendedError();

		switch( err )
		{
			case 0: // Canceled...
				return FALSE;

			case FNERR_BUFFERTOOSMALL:
				MessageBoxA( hWnd, "Buffer too small", errCaption, MB_ICONSTOP );
				break;
			default:
				MessageBoxA( hWnd, "Undefined error", errCaption, MB_ICONSTOP );
				break;
		}
	}

	if ( ofn.nFileExtension == 0 )
	{
		if ( _access( buff, 0 ) != 0 ) // does not exist.
		{
			const char *ext = findFilterExt( ofn.lpstrFilter, ofn.nFilterIndex );

			if ( ext )
				strcat_s( buff, buffSize, ext  );
		}
	}

	BF_SetInitialPath( buff, ofn.nFileOffset );

	return ok;
}



