#include "Shutdown.h"
#include "ODS.h"

//
// GetShutdownPrivileges.
//	One assigned it will never try to assigne again, simple returns TRUE.
//
BOOL GetShutdownPrivileges( HWND hWnd, BOOL silent )
{
	static BOOL assigned = FALSE;

	if ( assigned )
		return TRUE;

	HANDLE hToken;  
	TOKEN_PRIVILEGES tkp; 

	/* Get a token for this process. */ 

	if (!OpenProcessToken(GetCurrentProcess(), 
			TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY, &hToken)) 
	{
		ODS_fls( L"Assertion" );

		if ( !silent )
			MessageBoxA( hWnd ,"OpenProcessToken", "Shutdown error",MB_ICONSTOP | MB_TOPMOST ); 

		return FALSE;
	}

	/* Get the LUID for the shutdown privilege. */ 

	LookupPrivilegeValue(NULL, SE_SHUTDOWN_NAME, 
			&tkp.Privileges[0].Luid); 

	tkp.PrivilegeCount = 1;  /* one privilege to set    */ 
	tkp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED; 

	/* Get the shutdown privilege for this process. */ 

	AdjustTokenPrivileges(hToken, FALSE, &tkp, 0, 
			(PTOKEN_PRIVILEGES)NULL, 0); 

	/* Cannot test the return value of AdjustTokenPrivileges. */ 

	if (GetLastError() != ERROR_SUCCESS) 
	{
		ODS_fls( L"Assertion" );

		if ( !silent )
			MessageBoxA( hWnd, "AdjustTokenPrivileges", "Shutdown error", MB_ICONSTOP | MB_TOPMOST ); 
	}
	else
		assigned = TRUE;

	CloseHandle( hToken );

	return assigned;
}

/////////////////////////////////////////////////////////////////////////////////
// Function Shutdown();
// Take required privilege for NT and initiate shutdown.

BOOL Shutdown( HWND hWnd , BOOL isNT, int type ) // isNT == 1, type = EWX_SHUTDOWN )
{
	ODSFunc;

	if ( type == EWX_LOGOFF )
	{
		BOOL res = ExitWindowsEx( EWX_LOGOFF, 0 );
		if ( !res )
		{
			ODS_fls( L"Assertion!" );
		//	MessageBoxA( hWnd, "ExitWindowsEx", "Log Off error", MB_ICONSTOP);
		}

		return res;
	}

	if ( isNT == -1 )
		isNT = ( GetVersion() < 0x80000000 );

	if ( isNT )
	{
		GetShutdownPrivileges( hWnd, true );
	}
 
	/* Shut down the system and force all applications to close. */
	// InitiateSystemShutdown can be used here.
 
	if ( !ExitWindowsEx( type, 0 ) ) // EWX_REBOOT, 0)) 
	{
		ODS_fls( L"Assertion!" );
		//	MessageBoxA( hWnd, "ExitWindowsEx", "Shutdown error", MB_ICONSTOP );

		return FALSE;
	}

	return TRUE;
}

