/////////////////////////////////////////////////////////////////
//
//	Copyright  2009-2018 Jetico Inc., Finland
//	All rights reserved.
//
//	File: SignatureCheck.h
//
//	Revision:
//		29.09.2016
//			Anti hacker check.
//			It allows to check signed public key validity togather with certificate names.
//			See isValidKeyID and getRegDelay
//
//		22.03.2012 By Alexander Pichuev.
//		Added to Includes 
//
//	Created:
//		25.11.2009 By Nail Kaipov
//
//	Description: 
//		Verifies signature and checks if signer is Jetico.
//
/////////////////////////////////////////////////////////////////

#pragma once

#include <wintrust.h>

class CWinTrustAPI;

class CSignatureCheck
{
public:
	enum {
		eSuccess = 0, 
		eNotJetico = 1 // sinature is verified but signer is not Jetico.
	};

	//
	// Current status on of enum above or folloing:
	//	TRUST_E_NOSIGNATURE
	//	TRUST_E_EXPLICIT_DISTRUST
	//	TRUST_E_SUBJECT_NOT_TRUSTED
	//	CRYPT_E_SECURITY_SETTINGS
	//
	DWORD m_status;
	const TCHAR *m_statusStr;
	CSignatureCheck(void);
	virtual ~CSignatureCheck(void);

	BYTE m_subjectKeyID[20];

	BOOL	VerifyEmbeddedSignature(LPCTSTR Filename, LPCTSTR Catalog = NULL, HKEY hKey = NULL );
	static  bool getRegDelay();
	bool isValidKeyID();

	static inline bool IsTrustedModule( HMODULE hModule = NULL, HKEY hKey = NULL )
	{
#ifdef UNICODE
		wchar_t moduleName[ MAX_PATH ];
#else
		char moduleName[ MAX_PATH ];
#endif

		if ( !GetModuleFileName( hModule, moduleName, sizeof( moduleName ) / sizeof( *moduleName ) ) )
			return false;

		CSignatureCheck csc;

		bool ok = csc.VerifyEmbeddedSignature( moduleName, NULL, hKey ) == TRUE;

		bool bValidPublicKey = csc.isValidKeyID();

		if ( ok && !bValidPublicKey )
		{
			if ( csc.getRegDelay() )
				ok = false;
		}

		return ok;
	}
private:
	DWORD ReadFileSignatureText(LPCTSTR Filename, LPTSTR Buffer, DWORD &Length);
	DWORD ReadCatalogSignatureText(LPCTSTR Filename, LPCTSTR Catalog, LPTSTR Buffer, DWORD &Length);
	DWORD ExtractSignerText(WINTRUST_DATA *WinTrust, LPTSTR Buffer, DWORD &Length);

	DWORD saveSubjectId( CWinTrustAPI &wt, HANDLE hWVTStateData, LPTSTR Buffer, DWORD Length );
};
