/*
 *	Copyright 2014 Jetico, Inc. All rights reserved.
 *	cADPtrTable.h
 *	cADPtrTable template class description.
 *		The Free() function is called at this calss destructor.
 *		Standard destructor is called for every element at removing.
 *		
 *	Created:	23.04.2013
 *	Author:		Alexander Pichuev.
 */

#pragma once

template < class T > T round4K( T val )
{
	T _4K_minus1 = (2 << 12) - 1;
	val += _4K_minus1;
	val &= ~_4K_minus1;

	return val;
}

template < class T > class cADPtrTable
{
	size_t m_allocatedSize;
	
	T** allocate( size_t size, size_t &sizeAllocated )
	{
		size_t byteSize = size * sizeof( void * );

		byteSize = round4K( byteSize );
		
		void *p = malloc( byteSize );

		if ( p == NULL )
			return NULL;

		memset( p, 0, byteSize );

		sizeAllocated = byteSize / sizeof( void * );

		return (T**)p;
	}
	
	size_t Extend( size_t pos, size_t Q )
	{
		size_t newSize = m_Q + Q;

		if ( m_allocatedSize < newSize )
		{
			size_t allocatedSize;

			T **pT = allocate( newSize, allocatedSize );

			if ( pT == NULL )
				return (size_t)-1;
			
			if ( m_T != NULL )
			{
				memcpy( pT, m_T, pos * sizeof( void * ) );
				memcpy( pT + pos + Q, m_T + pos, (m_Q - pos) * sizeof( void * ) );
				free( m_T );
			}
		
			m_T = pT;
			m_allocatedSize = allocatedSize;
		}
		else
			memmove( m_T + pos + Q, m_T + pos, (m_Q - pos) * sizeof( void * ) );
		
		m_Q = newSize;

		return pos;
	}

	//
	//	It does not call destructors for eliments,
	//	just allocates new place for table ( and copy old pointers ), if size is increased otherwise decreases current m_Q.
	//	Returns old size.
	//
	size_t Resize( size_t newSize )
	{
		size_t oldSize = m_Q;

		if ( oldSize >= newSize )
		{
			memset( m_T + newSize, 0, (oldSize - newSize) * sizeof( void * ) );
		}
		else
		if ( Extend( m_Q, newSize - m_Q ) == (size_t)-1 )
			return (size_t)-1;

		m_Q = newSize;

		return oldSize;
	}

public:
	T **m_T;
	size_t m_Q;

	cADPtrTable()
	{
		m_T = NULL;
		m_Q = 0;
		m_allocatedSize = 0;
	}
	
	~cADPtrTable()
	{
		Free();
	}

	void Free()
	{
		if ( m_T == NULL )
			return;

		for( size_t i = 0; i < m_Q; i++ )
		{
			if ( m_T[i] == NULL )
				continue;
			
			delete m_T[i];
		}

		free( m_T );

		m_T = NULL;
		m_Q = 0;
		m_allocatedSize = 0;
	}

	int Add( T *in )
	{
		return Insert( in, (size_t)-1 );
	}

	int Insert( T *in, size_t iPos  )
	{
		if ( ( iPos == (size_t)-1 ) || ( iPos > m_Q ) )
			iPos = m_Q;

		size_t newSize = m_Q + 1;
		size_t addedPos = Extend( iPos, 1 );

		if ( addedPos == (size_t)-1 )
			return -1;

		m_T[ addedPos ] = in;

		return addedPos;
	}

	int Remove( size_t pos )
	{
		if ( pos >=  m_Q )
			return -1;

		if ( m_T[pos] != NULL )
			delete m_T[ pos ];

		m_T[pos] = NULL;

		for( size_t i = pos + 1; i < m_Q; i++ )
		{
			m_T[i-1] = m_T[i];
		}

		Resize( m_Q - 1 );

		return pos;
	}

	T* GetAt( size_t i )
	{
		if ( i >= m_Q )
			return NULL;

		return m_T[i];
	}
};
