/**************************************************************************************************************\
*	ATA device class 
*	Purpouse:	ATA device identification.
*
*	Started:	11.03.2011
*
*	Sources:
*	Uses smartmontools-5.40 codes as sample.
*
*	Calls IDENTIFY DEVICE with ATA8-ACS draft structure
*	Required fileds are:
*	 - (word 217) Nominal media rotation rate
*		SSD drive reports - 1
*	 - (word 83 ) Bit 2:	CFA Feature Set
*		SSD 40Gb drive with serial number "CO-ZR4IEQ81X5KL8001N"
*		retports:
*		CFA  bit - 0
*
*	Strange behaivior in W7. 
*		It does not allow to open logical drives for read/write access for non administrator ( it is normaly ).
*		It does not allow to implement functions ata_via_scsi_pass_through_ioctl and smart_get_version for volume or drive 
*		if it is opened for read access only ( it is normaly too ).
*		But it does not require adminstrator privileges for opening physical drive for read/write access.
*
*	Use the hasSSDDriveExtent function to find SDD drive in physical drive extent list of logical drive.
*
\**************************************************************************************************************/

//
//	Using sample:
//
//	ODSF( L"Physical drives:" );
//
//	for ( int phydrive = 0; phydrive <= 255; phydrive++ )
//	{
//		cATADevice ataDev;
//		if ( ataDev.Open( phydrive, 0 ) != 0 )
//			continue;
//
//		ata_identity ai;
//		if ( cATADevice::identifyDevice( ataDev.m_hDevice, &ai ) != 0 )
//			continue;
//		
//		if ( ataDev.isSSDDrive( &ai ) )
//		{
//			ODSF( L"Drive %s is SDD", ataDev.m_devpath );
//		}
//	}
//
//	ODSF( L"Logical drives:" );
//
//	for ( int logdrive = 2; logdrive < 26; logdrive++ )
//	{
//		cATADevice ataDev;
//		if ( ataDev.Open( -1, logdrive ) != 0 )
//			continue;
//
//		ata_identity ai;
//		if ( cATADevice::identifyDevice( ataDev.m_hDevice, &ai ) != 0 )
//			continue;
//		
//		if ( ataDev.isSSDDrive( &ai ) )
//		{
//			ODSF( L"Drive %s is SDD", ataDev.m_devpath );
//		}
//	}
//

#pragma once

#include <windows.h>
#include "AutoDeallocate.h"

#ifndef _WIN32_WINNT		// Allow use of features specific to Windows XP or later.                   
#define _WIN32_WINNT 0x0501	// Change this to the appropriate value to target other versions of Windows.
#endif						

#pragma pack(1)

/* ATA8-ACS draft structure. Based on ATA6 hd_driveid from linux/hdreg.h */
struct ata_identity  {
	unsigned short	config;		/* lots of obsolete bit flags */
	unsigned short	cyls;		/* Obsolete, "physical" cyls */
	unsigned short	reserved2;	/* reserved (word 2) */
	unsigned short	heads;		/* Obsolete, "physical" heads */
	unsigned short	track_bytes;	/* unformatted bytes per track */
	unsigned short	sector_bytes;	/* unformatted bytes per sector */
	unsigned short	sectors;	/* Obsolete, "physical" sectors per track */
	unsigned short	vendor0;	/* vendor unique */
	unsigned short	vendor1;	/* vendor unique */
	unsigned short	vendor2;	/* Retired vendor unique */
	char			serial_no[20];	/* 0 = not_specified */
	unsigned short	buf_type;	/* Retired */
	unsigned short	buf_size;	/* Retired, 512 byte increments
					 * 0 = not_specified
					 */
	unsigned short	ecc_bytes;	/* for r/w long cmds; 0 = not_specified */
	unsigned char	fw_rev[8];	/* 0 = not_specified */
	char			model[40];	/* 0 = not_specified */
	unsigned char	max_multsect;	/* 0=not_implemented */
	unsigned char	vendor3;	/* vendor unique */
	unsigned short	dword_io;	/* 0=not_implemented; 1=implemented */
	unsigned char	vendor4;	/* vendor unique */
	unsigned char	capability;	/* (upper byte of word 49)
					 *  3:	IORDYsup
					 *  2:	IORDYsw
					 *  1:	LBA
					 *  0:	DMA
					 */
	unsigned short	reserved50;	/* reserved (word 50) */
	unsigned char	vendor5;	/* Obsolete, vendor unique */
	unsigned char	tPIO;		/* Obsolete, 0=slow, 1=medium, 2=fast */
	unsigned char	vendor6;	/* Obsolete, vendor unique */
	unsigned char	tDMA;		/* Obsolete, 0=slow, 1=medium, 2=fast */
	unsigned short	field_valid;	/* (word 53)
					 *  2:	ultra_ok	word  88
					 *  1:	eide_ok		words 64-70
					 *  0:	cur_ok		words 54-58
					 */
	unsigned short	cur_cyls;	/* Obsolete, logical cylinders */
	unsigned short	cur_heads;	/* Obsolete, l heads */
	unsigned short	cur_sectors;	/* Obsolete, l sectors per track */
	unsigned short	cur_capacity0;	/* Obsolete, l total sectors on drive */
	unsigned short	cur_capacity1;	/* Obsolete, (2 words, misaligned int)     */
	unsigned char	multsect;	/* current multiple sector count */
	unsigned char	multsect_valid;	/* when (bit0==1) multsect is ok */
	unsigned int	lba_capacity;	/* Obsolete, total number of sectors */
	unsigned short	dma_1word;	/* Obsolete, single-word dma info */
	unsigned short	dma_mword;	/* multiple-word dma info */
	unsigned short  eide_pio_modes; /* bits 0:mode3 1:mode4 */
	unsigned short  eide_dma_min;	/* min mword dma cycle time (ns) */
	unsigned short  eide_dma_time;	/* recommended mword dma cycle time (ns) */
	unsigned short  eide_pio;       /* min cycle time (ns), no IORDY  */
	unsigned short  eide_pio_iordy; /* min cycle time (ns), with IORDY */
	unsigned short	words69_70[2];	/* reserved words 69-70
					 * future command overlap and queuing
					 */
	unsigned short	words71_74[4];	/* reserved words 71-74
					 * for IDENTIFY PACKET DEVICE command
					 */
	unsigned short  queue_depth;	/* (word 75)
					 * 15:5	reserved
					 *  4:0	Maximum queue depth -1
					 */
	unsigned short	sata_capable;	/* (word 76)
					* 0:    Zero
					* 1:    SATA-I
					* 2:    SATA-II
					* 7:3   reserved
					* 8:    NCQ support
					* 9:    Partial & Slumber PM states support
					* 10:   PHY event counter support
					* 15:11 reserved
					*/
	unsigned short	sata_reserved;	/* (word 77) */
	unsigned short	sata_command_set_1;	/* (word 78) */
	unsigned short	sata_command_set_2;	/* (word 79) */

	unsigned short  major_rev_num;	/* (word 80) */
	unsigned short  minor_rev_num;	/* (word 81) */
	unsigned short  command_set_1;	/* (word 82) supported
					 * 15:	Obsolete
					 * 14:	NOP command
					 * 13:	READ_BUFFER
					 * 12:	WRITE_BUFFER
					 * 11:	Obsolete
					 * 10:	Host Protected Area
					 *  9:	DEVICE Reset
					 *  8:	SERVICE Interrupt
					 *  7:	Release Interrupt
					 *  6:	look-ahead
					 *  5:	write cache
					 *  4:	PACKET Command
					 *  3:	Power Management Feature Set
					 *  2:	Removable Feature Set
					 *  1:	Security Feature Set
					 *  0:	SMART Feature Set
					 */
	unsigned short  command_set_2;	/* (word 83)
					 * 15:	Shall be ZERO
					 * 14:	Shall be ONE
					 * 13:	FLUSH CACHE EXT
					 * 12:	FLUSH CACHE
					 * 11:	Device Configuration Overlay
					 * 10:	48-bit Address Feature Set
					 *  9:	Automatic Acoustic Management
					 *  8:	SET MAX security
					 *  7:	reserved 1407DT PARTIES
					 *  6:	SetF sub-command Power-Up
					 *  5:	Power-Up in Standby Feature Set
					 *  4:	Removable Media Notification
					 *  3:	APM Feature Set
					 *  2:	CFA Feature Set
					 *  1:	READ/WRITE DMA QUEUED
					 *  0:	Download MicroCode
					 */
	unsigned short  cfsse;		/* (word 84)
					 * cmd set-feature supported extensions
					 * 15:	Shall be ZERO
					 * 14:	Shall be ONE
					 * 13:6	reserved
					 *  5:	General Purpose Logging
					 *  4:	Streaming Feature Set
					 *  3:	Media Card Pass Through
					 *  2:	Media Serial Number Valid
					 *  1:	SMART selt-test supported
					 *  0:	SMART error logging
					 */
	unsigned short  cfs_enable_1;	/* (word 85)
					 * command set-feature enabled
					 * 15:	Obsolete
					 * 14:	NOP command
					 * 13:	READ_BUFFER
					 * 12:	WRITE_BUFFER
					 * 11:	Obsolete
					 * 10:	Host Protected Area
					 *  9:	DEVICE Reset
					 *  8:	SERVICE Interrupt
					 *  7:	Release Interrupt
					 *  6:	look-ahead
					 *  5:	write cache
					 *  4:	PACKET Command
					 *  3:	Power Management Feature Set
					 *  2:	Removable Feature Set
					 *  1:	Security Feature Set
					 *  0:	SMART Feature Set
					 */
	unsigned short  cfs_enable_2;	/* (word 86)
					 * command set-feature enabled
					 * 15:	Shall be ZERO
					 * 14:	Shall be ONE
					 * 13:	FLUSH CACHE EXT
					 * 12:	FLUSH CACHE
					 * 11:	Device Configuration Overlay
					 * 10:	48-bit Address Feature Set
					 *  9:	Automatic Acoustic Management
					 *  8:	SET MAX security
					 *  7:	reserved 1407DT PARTIES
					 *  6:	SetF sub-command Power-Up
					 *  5:	Power-Up in Standby Feature Set
					 *  4:	Removable Media Notification
					 *  3:	APM Feature Set
					 *  2:	CFA Feature Set
					 *  1:	READ/WRITE DMA QUEUED
					 *  0:	Download MicroCode
					 */
	unsigned short  csf_default;	/* (word 87)
					 * command set-feature default
					 * 15:	Shall be ZERO
					 * 14:	Shall be ONE
					 * 13:6	reserved
					 *  5:	General Purpose Logging enabled
					 *  4:	Valid CONFIGURE STREAM executed
					 *  3:	Media Card Pass Through enabled
					 *  2:	Media Serial Number Valid
					 *  1:	SMART selt-test supported
					 *  0:	SMART error logging
					 */
	unsigned short  dma_ultra;	/* (word 88) */
	unsigned short	trseuc;		/* time required for security erase */
	unsigned short	trsEuc;		/* time required for enhanced erase */
	unsigned short	CurAPMvalues;	/* current APM values */
	unsigned short	mprc;		/* master password revision code */
	unsigned short	hw_config;	/* hardware config (word 93)
					 * 15:	Shall be ZERO
					 * 14:	Shall be ONE
					 * 13:
					 * 12:
					 * 11:
					 * 10:
					 *  9:
					 *  8:
					 *  7:
					 *  6:
					 *  5:
					 *  4:
					 *  3:
					 *  2:
					 *  1:
					 *  0:	Shall be ONE
					 */
	unsigned short	acoustic;	/* (word 94)
					 * 15:8	Vendor's recommended value
					 *  7:0	current value
					 */
	unsigned short	msrqs;		/* min stream request size */
	unsigned short	sxfert;		/* stream transfer time */
	unsigned short	sal;		/* stream access latency */
	unsigned int	spg;		/* stream performance granularity */
	unsigned long long lba_capacity_2;/* 48-bit total number of sectors */
	unsigned short	words104_125[22];/* reserved words 104-125 */
	unsigned short	last_lun;	/* (word 126) */
	unsigned short	word127;	/* (word 127) Feature Set
					 * Removable Media Notification
					 * 15:2	reserved
					 *  1:0	00 = not supported
					 *	01 = supported
					 *	10 = reserved
					 *	11 = reserved
					 */
	unsigned short	dlf;		/* (word 128)
					 * device lock function
					 * 15:9	reserved
					 *  8	security level 1:max 0:high
					 *  7:6	reserved
					 *  5	enhanced erase
					 *  4	expire
					 *  3	frozen
					 *  2	locked
					 *  1	en/disabled
					 *  0	capability
					 */
	unsigned short  csfo;		/*  (word 129)
					 * current set features options
					 * 15:4	reserved
					 *  3:	auto reassign
					 *  2:	reverting
					 *  1:	read-look-ahead
					 *  0:	write cache
					 */
	unsigned short	words130_155[26];/* reserved vendor words 130-155 */
	unsigned short	word156;	/* reserved vendor word 156 */
	unsigned short	words157_159[3];/* reserved vendor words 157-159 */
	unsigned short	cfa_power;	/* (word 160) CFA Power Mode
					 * 15 word 160 supported
					 * 14 reserved
					 * 13
					 * 12
					 * 11:0
					 */
	unsigned short	words161_167[7]; /* Reserved for CFA */
	unsigned short  form_factor;	 /* (word 168) device form factor, ACS-2 */
	unsigned short  data_set_mgmt;	 /* (word 169) data set management aka TRIM support ACS-2 */
	unsigned short  words170_175[6]; /* ATA identifier and reserved area */
	char            media_serial[40];/* Media Serial Number */
	char            manufacturer[20];/* Media? Manufacturer */
	unsigned short	sct_transport;	/* (word 206) SCT Command transport */
	unsigned short	word207;		/* (word 207) reserved for CE-ATA */
	unsigned short	word208;		/* (word 208) reserved for CE-ATA */
	unsigned short	lb_pb_algn;		/* (word 209) Alignment of logical blocks within a larger physical block */

	unsigned short	words210_213[4];	/* verify counts */

	unsigned short	nv_cache_caps;	/* (word 214) NV Cache Capabilities */
	unsigned short	nv_cache_size_lo;	/* (word 215) NV Cache Size in Logical Blocks (15:0) */
	unsigned short	nv_cache_size_hi;	/* (word 216) NV Cache Size in Logical Blocks (31:16) */

	unsigned short	rot_rate;		/* (word 217) Nominal media rotation rate */
	unsigned short	word218;		/* (word 218) reserved */

	unsigned short	nv_cache_opt1;	/* (word 219) NV Cache Options */
	unsigned short	nv_cache_opt2;	/* (word 220) NV Cache Options */

	unsigned short	word221;		/* (word 221) reserved */

	unsigned short	tr_major;		/* (word 222) Transport Major revision number */
	unsigned short	tr_minor;		/* (word 223) Transport Minor revision number */

	unsigned short	words224_233[10];	/* reserved for CE-ATA */
	unsigned short	min_ucode_size;		/* (word 234) Minimum number of 512 byte units per DOWNLOAD MICROCODE command */
	unsigned short	max_ucode_size;		/* (word 235) Maximum number of 512 byte units per DOWNLOAD MICROCODE command */

	unsigned short	words236_254[19];/* reserved words 236-254 */

	unsigned short	integrity_word;	/* (word 255)
					 * 15:8 Checksum
					 *  7:0 Signature
					 */
};

#pragma pack()

class cDevice
{
	bool m_bAdmin;

public:

	wchar_t m_devpath[30];

	HANDLE m_hDevice;
	
	cDevice()
	{
		m_bAdmin = false;
		m_hDevice = INVALID_HANDLE_VALUE;
	}

	int Open( int phydrive, int logdrive );
	int Open( const wchar_t *devpath );
	
	~cDevice()
	{
		Close();
	}

	void Close()
	{
		if ( m_hDevice != INVALID_HANDLE_VALUE )
			::CloseHandle( m_hDevice );
	}
};

class cATADevice : public cDevice
{
public:

	static int identifyDevice( HANDLE hDevice, ata_identity *ai );
	static bool isSSDDrive( ata_identity *ai );

	//
	//	isSupportCFA
	//	Returns true if the device support CompactFlash Association (CFA) feature set.
	//
	//	See D1699r4a-ATA8-ACS.pdf
	//	4.7 CompactFlash Association (CFA) feature set
	//	The optional CompactFlash Association (CFA) feature set provides support for devices that implement the
	//	CFA specifications. A device that implements the CFA feature set shall implement the following minimum set of
	//	commands:
	//	a) CFA REQUEST EXTENDED ERROR CODE
	//	b) CFA WRITE SECTORS WITHOUT ERASE
	//	c) CFA ERASE SECTORS
	//	d) CFA WRITE MULTIPLE WITHOUT ERASE
	//	e) CFA TRANSLATE SECTOR
	//	f) SET FEATURES Enable/Disable 8-bit transfer
	//	Devices reporting the value 848Ah in IDENTIFY DEVICE data
	//	
	//
	static bool isSupportCFA( ata_identity *ai );

	//
	//	getPhysDriveNum
	//		Returns (-1) if physical drive number is not detected.
	//		It uses DeviceIoControl with IOCTL_VOLUME_GET_VOLUME_DISK_EXTENTS.
	//		But logical drive may be placed on several physical drives.
	//		See hasSSDDriveExtent as sample, it parses all extents to find SDD drive in extent list.
	//
	static int getPhysDriveNum( int logdrive );
};

#pragma pack (1)

//
//	The structure allows to accumulate physical drive information at table.
//	The information may be saved to registry, f.e. after administrator once read it. See cSSDNotifier::SavePDInfoToRegistry()
//	Or allows to avoid several time asking.
//
struct sPhysDriveInfo
{
	DWORD	diskNumber;
	DWORD	dwFlags;
	//bool	typeSSD;
	//bool	supportCFA;

	enum eFlag { pdEmpty = -1, eTypeSDD = 0x0001, eStoragePool = 0x0002, eSupportCFA = 0x0100 };
	
	bool getFlag( eFlag flag )
	{
		return ( dwFlags & flag ) != 0;
	}

	void setFlag( eFlag flag, bool bSet )
	{
		if ( bSet )
			dwFlags |= flag;
		else
			dwFlags &= ~flag;
	}

	void setSupportCFA	( bool bSet )	{ setFlag( eSupportCFA,	bSet ); }
	void setTypeSSD		( bool bSet )	{ setFlag( eTypeSDD,	bSet ); }
	
	sPhysDriveInfo()
	{
		memset( this, 0, sizeof( *this ) );
		MarkAsFree();
	}

	void Free()
	{
		MarkAsFree();
	}

	void MarkAsFree()
	{
		diskNumber	= pdEmpty;
		//typeSSD		= false;
		//supportCFA	= false;
		dwFlags = 0;
	}

	static  sPhysDriveInfo *find( DWORD diskNum, sPhysDriveInfo *T, int Q )
	{
		for( int i = Q; i--; )
		{
			if ( T[i].diskNumber == diskNum )
				return T + i;
		}

		return NULL;
	}

	static bool add( DWORD diskNum, bool bSSD, bool bSupportCFA, sPhysDriveInfo *T, int Q )
	{
		sPhysDriveInfo *pdi = find( diskNum, T, Q );
		
		if ( pdi != NULL )
		{
			pdi->setSupportCFA( bSupportCFA );
			return true;
		}

		//
		//	Find first empty.
		//
		pdi = find( pdEmpty, T, Q );

		if ( pdi == NULL )
			return false;

		pdi->diskNumber = diskNum;
		pdi->setSupportCFA( bSupportCFA );	// pdi->supportCFA = bSupportCFA;
		pdi->setTypeSSD( bSSD );			// pdi->typeSSD = bSSD;

		return true;
	}

	static sPhysDriveInfo *allocateTable( int Q )
	{
		return new sPhysDriveInfo[Q];
	}

	void freeTable( sPhysDriveInfo **pT, int Q )
	{
		if ( *pT == NULL )
			return;

		delete (*pT );

		*pT = NULL;
	}
};

#pragma pack ()

//
//	hasSSDDriveExtent
//	Walks all extent of corresponded logical drive:
//		1. checks if the extent physical drive is SSD
//		2. sets pSupportCFA as false if at least one SDD drive does not support CFA functions
//	Returns:
//		-1 if error
//		0 - does not have extent with SDD drive.
//		otherwise number of extents with SDD drive.
//
int hasSSDDriveExtent( int logDrive, bool *pSupportCFA );
int hasSSDDriveExtent_actor( sPhysDriveInfo *pdi, void *param );

//
//	hasStoragePoolDriveExtent
//	Walks all extent of corresponded logical drive:
//		returns number of extents which support thin provision.
//
int hasStoragePoolDriveExtent( int logDrive, cADTable < sPhysDriveInfo > *pdiT = NULL );

typedef int typeWalkLogDriveActor( sPhysDriveInfo *pdi, void *param );
int walkLogDriveExtents( HANDLE hLogDrive, typeWalkLogDriveActor *actor, void *param, cADTable < sPhysDriveInfo > *pdiT );

//
//	getLogicalDriveNumber
//		Returns logical drive number by file path.
//
int getLogicalDriveNumber( const wchar_t *filePath );
bool getLogicalDriveName( wchar_t *bf, int bfSize, const wchar_t *filePath );

//
//	szPDInfo value is used to save physical drive info prepared at administrative mode.
//
extern const wchar_t *szPDInfo;

class cPhysDriveTable
{
public:
	//
	//	Administrator may read, update and save physical drive info by SavePDInfoToRegistry.
	//		Note, SSDNotifier::updateSSDInfo() called from BCWipeTM at startup always update the physical drive infor in registry.
	//	Non administrator uses ReadPDInfoFromRegistry.
	//
	static cADTable < sPhysDriveInfo > m_adt;
	static int ReadPDInfoFromRegistry( const wchar_t *szBCWipeKey );
	static int SavePDInfoToRegistry( const wchar_t *szBCWipeKey );
};
