#ifndef __cArgs__
#define __cArgs__

#include <string.h>

#ifdef _AutoDeallocate_h_
#ifdef _UnicodeFuncs_
#include "cMBStr.h"
#endif
#endif

class cArgs
{
	wchar_t *buffer;
	wchar_t *cmdRest;

public:
	int argc;
	wchar_t **argv;

	cArgs( const wchar_t *cmdLine = NULL )
	{
		memset( this, 0, sizeof( *this ) );
		if ( cmdLine != NULL )
			Init( cmdLine );
	}

	void Init( const wchar_t *cmdLine );
	void Init( const char *cmdLine );
	~cArgs();

	bool has( const wchar_t *value )
	{
		return Find( value ) != -1;
	}

	const wchar_t *arg( int num )
	{
		if ( num < argc )
			return argv[num];

		return NULL;
	}

#ifdef _AutoDeallocate_h_
#ifdef _UnicodeFuncs_
	//
	// It works if both the following files are included before the definition.
	//
	//#include "UnicodeFuncs.h"
	//#include "AutoDeallocate.h"
	const char *argA( int pos )
	{
		static cADArray < cMBStr > adT( 1 );

		if ( adT.m_Q <= pos )
			adT.Add( NULL, pos + 1 - adT.m_Q );

		if ( ( adT.m_Q <= pos ) ||
			 ( adT.m_T[ pos ].Set( arg( pos ), -1 ) == 0 ) )
				return "<NULL>";

		return adT.m_T[ pos ].m_szBody;
	}

#endif // _UnicodeFuncs_
#endif // _AutoDeallocate_h_

	const wchar_t *GetCmdLine();

	//
	//	Sample:
	//		name == -a, argument -aAnyValue, so *pValue = AnyValue
	//
	//	Note, the following arguments have same name and value.
	//		-aAnyValue
	//		-a AnyValue
	//		-a   AnyValue
	//		-a "AnyValue"
	//		-a   "AnyValue"
	//
	bool ReadAndRemoveArgumentWithValue( const wchar_t *name, const wchar_t **pValue, const wchar_t *szDefValue = NULL )
	{ return Get( name, pValue, szDefValue ); }

	bool ReadAndRemoveArgumentWithIntValue( const wchar_t *name, int *pValue, int defValue = 0 );

	const wchar_t *ReadAndRemoveArgumentWithValue( const wchar_t *name )
	{
		const wchar_t *value;
		if ( !ReadAndRemoveArgumentWithValue( name, &value ) )
			return NULL;
		else
			return value;
	}

	//
	//	Get functions:
	//	Read argument at specified position and removes it.
	//	Without pos it finds the argument in any position first.
	//
	bool Get( int pos,	const wchar_t *name );
	bool Get(			const wchar_t *name );
	bool Get(			const wchar_t *name, const wchar_t **pValue, const wchar_t *szDefValue = NULL );

	bool Remove( int i );

	bool ReadAndRemoveArgument( int pos,const wchar_t *name ) { return Get( pos,name ); }
	bool ReadAndRemoveArgument(			const wchar_t *name ) { return Get(		name ); }

	bool EQU( int pos, const wchar_t *name )
	{
		return (_wcsicmp( argv[pos], name ) == 0 );
	}

	//
	//	Find
	//		returns position of the item found, otherwise return -1.
	//
	int Find( const wchar_t *string );
	int FindPart( const wchar_t *string );

	static bool checkOnOffArg( const wchar_t *type, bool &bOn );
};

#endif // __cArgs__