/*
 *	Copyright 2014 Jetico, Inc. All rights reserved.
 *	cAutoFile.h
 *	Class cAutoFile ( cFile ) used for easy opening and normal closing file handle at destructor.
 *	Created:	25.09.2013
 *	Author:		Alexander Pichuev.
 */

#pragma once

class cAutoFile
{
public:
	HANDLE m_hFile;

	enum { 
		openMode_ReadWrite = GENERIC_READ | GENERIC_WRITE,
		shareMode_All = FILE_SHARE_READ | FILE_SHARE_DELETE | FILE_SHARE_WRITE
	};

	DWORD m_error;

	cAutoFile(
			LPCWSTR lpFileName = NULL,
			DWORD dwDesiredAccess = openMode_ReadWrite,
			DWORD dwShareMode = shareMode_All,
			LPSECURITY_ATTRIBUTES lpSecurityAttributes = NULL,
			DWORD dwCreationDisposition = OPEN_ALWAYS,
			DWORD dwFlagsAndAttributes = FILE_ATTRIBUTE_NORMAL
		)
	{
		m_hFile = INVALID_HANDLE_VALUE;
		m_error = ERROR_FILE_NOT_FOUND;

		if ( lpFileName != NULL )
			Open( lpFileName, dwDesiredAccess, dwShareMode, lpSecurityAttributes, dwCreationDisposition, dwFlagsAndAttributes );
	}

	HANDLE Open(
			LPCWSTR lpFileName,
			DWORD dwDesiredAccess,
			DWORD dwShareMode,
			LPSECURITY_ATTRIBUTES lpSecurityAttributes,
			DWORD dwCreationDisposition,
			DWORD dwFlagsAndAttributes
		)
	{
		Close();
		
		m_hFile = CreateFileT( lpFileName, dwDesiredAccess, dwShareMode, lpSecurityAttributes, dwCreationDisposition, dwFlagsAndAttributes, NULL );

		//
		//	v. 6.07.12	- 28.07.2014
		//		Prev version compres with NULL here, it arises Error: The operation completed successfully. (0x0)
		//
		if ( m_hFile != INVALID_HANDLE_VALUE )
			m_error = ERROR_SUCCESS;
		else
			m_error = GetLastError();

		return m_hFile;
	}

	bool wasError()
	{
		return ( ( m_hFile == INVALID_HANDLE_VALUE ) || ( m_error != ERROR_SUCCESS ) );
	}

	void Close()
	{
		if ( m_hFile != INVALID_HANDLE_VALUE )
		{
			CloseHandle(m_hFile);
			m_hFile = INVALID_HANDLE_VALUE;
		}
	}

	~cAutoFile()
	{
		Close();
	}
};