#include "cDlgCreateFile.h"

#define dwlErr (DWORDLONG(-1))
#define _4K ( 4 * 1024 )
#define RoundUp( value, blockSize ) ( ( (value) +  (blockSize) - 1 ) & ~((blockSize) - 1) )

static DWORDLONG setFilePointer( HANDLE hFile, DWORDLONG dwlCurPos, DWORD dwMoveMethod )
{
	ULARGE_INTEGER uli;
	uli.QuadPart = dwlCurPos;

	uli.LowPart = ::SetFilePointer( hFile, uli.LowPart, (PLONG)&(uli.HighPart), dwMoveMethod );

	if ( uli.LowPart == 0xFFFFFFFF )
	{
		int errorNumber = GetLastError();

		if ( errorNumber != NO_ERROR )
		{
			uli.QuadPart = dwlErr; // DWORDLONG(-1);
		}
	}

	return uli.QuadPart;
}

bool cDlgCreateFile::setFileLength( DWORDLONG length, int *pErrorNumber )
{
	ULARGE_INTEGER largeInt = { 0, 0 };

	if ( m_rngObj.generate )
	{
		largeInt.QuadPart = setFilePointer( m_hFile, 0, FILE_CURRENT );
		if ( largeInt.QuadPart == dwlErr  )
		{
			*pErrorNumber = GetLastError();

			if ( *pErrorNumber != NO_ERROR )
			{
				*pErrorNumber |= 0x50000000; // To detect the point
				return FALSE;
			}
		}

		bool ok = TRUE;

		int buffSize = m_rngObj.buffSize; // 1024 * 256; // 256K


		for( __int64 step = buffSize; ; largeInt.QuadPart += step )
		{
			if ( m_cancelled )
				break;

			__int64 diff = (length - largeInt.QuadPart);

			if ( diff <= 0 )
				break;

			if ( step > diff )
				step = RoundUp( diff, _4K );

			DWORD written = 0;
			if ( !m_rngObj.WriteToFile( m_hFile, (DWORD)step, &written ) )
			{
				*pErrorNumber = GetLastError();
				ok = FALSE;
				break;
			}
			else
			if	( written != (DWORD) step )
			{
				*pErrorNumber = GetLastError() | 0x20000000; // To detect the point
				ok = FALSE;
				break;
			}
		}
		
		return ok;
	}
	else // Does not fill by random
	{
		LARGE_INTEGER oldSize;
		oldSize.LowPart = GetFileSize( m_hFile, (DWORD *)&(oldSize.HighPart) );

		LONG size,hiSize;

		largeInt.QuadPart = length;
		size = largeInt.LowPart;
		hiSize = largeInt.HighPart;

		DWORD res = SetFilePointer( m_hFile, size, &hiSize, FILE_BEGIN );
		if ((res != (DWORD) size) || !SetEndOfFile( m_hFile ))
			return FALSE;

	//	v. 7.05.4 - 12.12.2001
	//	Creation of container without filling by random data did not really allocate disk space while file 
	//	was not closed. ( SetEndOfFile function does not do it in Win2k environment ). Now BestCrypt writes
	//	file length at the end of every block after length been set. Step of length setting is one percent 
	//	of required container length.

		const int buffSize = 16 * 1024; // 16K

		__int64 diff = length - oldSize.QuadPart;

		if ( ( GetVersion() < 0x80000000 ) && // if NT 
			 ( diff > buffSize ) )  // if total length > 16K * 100 > 1.5 M
		{

			BYTE buff[ buffSize ];
			memset( buff, 0, sizeof( buff ) );
			DWORDLONG *pDWL = (DWORDLONG *)( buff + sizeof( buff ) - sizeof( DWORDLONG ) );
			*pDWL = length;

			size = sizeof( buff );
			SetFilePointer( m_hFile, -size, NULL, FILE_END );
			WriteFile( m_hFile, buff, sizeof( buff ), &res, NULL );

			if ( res != sizeof( buff ) )
				return FALSE;
		}
	}

	return TRUE;
}

DWORD cDlgCreateFile::static_SetLengthThreadProc( cDlgCreateFile *pThis )
{
	if ( pThis == NULL )
	{
		assert( false );
		return 0;
	}

	return pThis->setLengthThreadProc();
}

DWORD cDlgCreateFile::setLengthThreadProc( )
{
	if ( m_length == 0 )
	{
		assert( false );
		PostMessage( m_hDlg, cDlgCreateFile::wmStopMessage, 1, 0 );
		return 0;
	}

	if ( m_hWndT[ enProgress ] == NULL )
	{
		assert( false );
		PostMessage( m_hDlg, cDlgCreateFile::wmStopMessage, 2, 0 );
		return 0;
	}

	if ( m_hWndT[ enPercent ] == NULL )
	{
		assert( false );
		PostMessage( m_hDlg, cDlgCreateFile::wmStopMessage, 3, 0 );
		return 0;
	}

	// Prepare step...
	
	DWORDLONG currentLength, step;
	const DWORDLONG minStepSize = 4 * 1024;
	const DWORDLONG startLength = 8 * 1024;
	currentLength = startLength;
	step = ((m_length / 100 ) + ( minStepSize - 1 ) ) & ~( minStepSize - 1 );

	setProgressPosition( 0 );

	// 
	//	Gets current length
	//	File position has to be the place after filling the hidden sector, key block and fist container sectors.
	//
	// // Set File Length Common Loop...
	// // It is enough to set file pointer to the file end once for the common loop
	// // Because no we will to write to file end only ( and file pointer has to be at the file end ).

	ULARGE_INTEGER largeInt;
	largeInt.QuadPart = setFilePointer( m_hFile, 0, FILE_CURRENT );

	wchar_t message[ 256 ];
	const wchar_t *errPattern = L"Set length error 0x%X";
	int lastError = 0;

	if ( largeInt.LowPart == 0xFFFFFFFF )
	{
		int lastError = GetLastError();

		if ( lastError!= NO_ERROR )
		{
			lastError |= 0x10000000; // To detect the point
			::FormatT( message, maxLen( message ), errPattern, lastError );
			DlgMessageBoxT( message, MB_ICONSTOP );
			return FALSE;
		}
	}

	currentLength = largeInt.QuadPart;

	for( ; ; )
	{
		if ( currentLength > m_length )
			currentLength = m_length;

		if ( !setFileLength( currentLength, &lastError ) )
		{
			::FormatT( message, maxLen( message ), errPattern, lastError );
			int res = DlgMessageBoxT( message, MB_ICONSTOP | MB_RETRYCANCEL );

			if ( res == IDRETRY )
				continue;

			m_cancelled = TRUE;
		}

		if ( m_cancelled )
			break;

		int progressPosition = (int)((currentLength * 100) / m_length );

		setProgressPosition( progressPosition );

		if ( currentLength == m_length )
			break;

		currentLength += step;
	}

	if ( m_rngObj.generate )
		m_rngObj.StopIt();

#ifdef _TestWaiting
	char message[ 1024 ];
	sprintf( message,
		"Test waiting:\n"
		"   fillWaiting  - %d sec\n"
		"   writeWaiting - %d sec\n"
		"   total time   - %d sec\n"
		"   step - %d Kb",
		m_rngObj.fillWaiting / CLOCKS_PER_SEC ,
		m_rngObj.writeWaiting / CLOCKS_PER_SEC,
		( clock() - m_rngObj.totalClocks ) / CLOCKS_PER_SEC,
		step / 1024 );

	MessageBox( message ); 
#endif

	if ( !m_cancelled )
	{
		if ( !m_flushText.isEmpty() )
			SetWindowTextT( m_hWndT[ enText ], m_flushText.m_szBody );

		FlushFileBuffers( m_hFile );
	}

	PostMessage( m_hDlg, WM_COMMAND, IDOK, 0 );

	return 1;
}

