#include "cDlgDownload.h"
#include "cDlg.h"
#include "cWinINetDll.h"
#include <time.h>

class cDlgDownload :
	public cDlgDownloadResources,
	public cDlgBase,
	public cDownloadActorBlock
{
	const wchar_t *m_dstDir;
	const wchar_t *m_dstName;
	const wchar_t *m_srcURL;
	const wchar_t *m_srcName;

	time_t m_startTime;

	time_t m_updateTime;
	DWORD m_downloadSize;
	DWORD m_totalSize;

	bool m_cancelled;

	cWCStr m_iconicPattern;

public:

	DWORD m_res;
	
	cDlgDownload(
			const wchar_t *dstDir,
			const wchar_t *dstName,
			const wchar_t *srcURL,
			const wchar_t *srcName,
			cDlgDownloadResources *rcIds )
				: cDlgDownloadResources( rcIds )
	{
		m_res = dfeOK;

		if ( m_hMainInst == NULL )
		{
			assert( false );
			m_res = dfeInternalError;
		}

		m_dstDir	= dstDir;
		m_dstName	= dstName;
		m_srcURL	= srcURL;
		m_srcName	= srcName;
		
		m_startTime = 0;
		m_updateTime = 0;
		m_totalSize = 0;
		m_downloadSize = 0;

		m_cancelled = false;
	}

	virtual LRESULT RealDlgProc( UINT uMsg, WPARAM wParam, LPARAM lParam )
	{
		if ( uMsg == WM_SIZE )
		{
			updateCaption();
		}

		return cDlgBase::RealDlgProc( uMsg, wParam, lParam );
	}

	bool InitDialog();
	
	void updateCaption();

	bool OnCommand( WORD ctrlId, HWND hCtrl, WORD isAccelerator )
	{
		if ( ctrlId == IDOK )
		{
			EndDialog( m_hDlg, m_res );
		}
		else
		if ( ctrlId == IDCANCEL )
		{
			m_cancelled = true;
			EnableWindow( GetDlgItem( m_hDlg, IDCANCEL ), false );
		}
		else
		{
			assert( false );
		}

		return true;
	}

	static DWORD WINAPI stDownloadFunc( LPVOID lpThreadParameter )
	{
		cDlgDownload *pThis = (cDlgDownload *)lpThreadParameter;

		if ( pThis == NULL )
		{
			assert( false );
			return -1;
		}

		pThis->m_res = pThis->downloadFunc();

		PostMessage( pThis->m_hDlg, WM_COMMAND, IDOK, 0 );

		return pThis->m_res;
	}

	DWORD downloadFunc();

	bool Actor( DWORD dwSize );
};

static const wchar_t *szDwnlName = L"BCTraveller_get";

bool cDlgDownload::Actor( DWORD dwSize )
{
	m_downloadSize += dwSize;

	time_t tmCur = time( NULL );

	DWORD diff = (DWORD)( tmCur - m_startTime );

	if ( diff > 3 )
	{
//		if ( !IsIconic( m_hDlg ) )
//			ShowWindow( m_hDlg, SW_SHOW );

		if ( m_updateTime != diff )
		{
			m_updateTime = diff;

			DWORDLONG bytesPerSecond = m_downloadSize / diff;

			wchar_t szSize[ 100 ];
			GetSizeTextT( szSize, maxLen( szSize ), bytesPerSecond );
			wcscat_s( szSize, L"/" );
			wcscat_s( szSize, cWCStr().Translate( "Sec" ) );
			SetDlgItemTextT( m_idcTransferRate, szSize );

			if ( ( bytesPerSecond > 0 ) && ( m_totalSize != 0 ) )
			{
				wchar_t szTotalSize[ 64 ];
				GetSizeTextT( szTotalSize, maxLen( szTotalSize ), m_totalSize );

				DWORDLONG timeRest = ( m_totalSize - m_downloadSize ) / bytesPerSecond;
				wchar_t timeRestStr[ 100 ];
				GetTimeTextT( timeRestStr, maxLen( timeRestStr ), (DWORD)timeRest );

				GetSizeTextT( szSize, maxLen( szSize ), m_downloadSize );

				SetDlgItemTextT( m_idcEstimationTime, cWCStr().FormatT( _LoadIDS( m_idsCopiedSSS ), timeRestStr, szSize, szTotalSize ) );
			}
		}
	}

	return !m_cancelled;
}

DWORD cDlgDownload::downloadFunc()
{
	m_startTime = time( NULL );

	cWCStr url;
	url.Add( m_srcURL );
	if ( m_srcName != NULL )
	{
		url.Add( L"/" );
		url.Add( m_srcName );
	}

	//
	//	url.m_szBody = "http://www.jetico.com/Translations/Traveller/Farsi/bctraveller_farsi.chm"
	//
	//	gets vaules
	//	 protocol		- http://
	//	 server name	- www.jetico.com
	//	 location path	- /Translations/Traveller/Farsi/bctraveller_farsi.chm
	//
	cWCStr srvName;
	cWCStr protocol;
	int pos = 0;
	int startPos = 0;
	for ( pos = 0; ; pos++ )
	{
		if ( url.m_szBody[ pos ] == L'\0' )
			return -1;

		if ( wcsncmp( url.m_szBody + pos, L"://", 3 ) == 0 ) 
		{
			pos += 3;
			startPos = pos;
			protocol.Add( url.m_szBody, pos );
			continue;
		}

		if ( url.m_szBody[ pos ] == L'/' )
		{
			srvName.Add( url.m_szBody + startPos, pos - startPos );
			break;
		}
	}

	cWinINet dll( szDwnlName, srvName.m_szBody, url.m_szBody + pos, protocol.m_szBody );

	if ( dll.error )
	{
		assert( false );
		return dfeInternalError;
	}

	const wchar_t *dstPath = m_dstDir;
	cWCStr wcs;
	
	if ( m_dstName != NULL )
		dstPath = wcs.FormatT( L"%s\\%s", m_dstDir, m_dstName );

	HANDLE hFile = CreateFileT( dstPath, GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL );

	if ( hFile == INVALID_HANDLE_VALUE )
	{
		assert( false );
		return dfeInternalError;
	}

	dll.getFileSize( &m_totalSize );

	bool ok = dll.downloadFile( hFile, this );

	CloseHandle( hFile );

	if ( !ok )
	{
		assert( false );
		DeleteFileT( dstPath );
		return dfeConnetionError;
	}

	if ( m_cancelled )
	{
		DeleteFileT( dstPath );
		return dfeCancelled;
	}

	return dfeOK;
}

void cDlgDownload::updateCaption()
{
	cWCStr wcsCaption;

	if ( IsIconic( m_hDlg ) && ( m_totalSize != 0 ) )
	{
		int percent = m_downloadSize * 100 / m_totalSize;
		wcsCaption.FormatT( m_iconicPattern.m_szBody, percent, m_srcName );
	}
	else
		wcsCaption.LoadStringT( IH.m_hInst, m_idsCaption );

	SetWindowTextT( m_hDlg, wcsCaption );
}

bool cDlgDownload::InitDialog()
{
	// Hides the window temporary, it will be shown after 3 seconds downloading.
	// See Actor() function.

	m_iconicPattern.Add( GetWindowTextLengthT( m_hDlg ) + 1 );
	GetWindowTextT( m_hDlg, m_iconicPattern.m_szBody, m_iconicPattern.m_len );
	
	ShowWindow( m_hDlg, SW_HIDE );
//	SetWindowPos( m_hDlg, 0, 0,0,0,0, SWP_HIDEWINDOW | SWP_NOSIZE | SWP_NOMOVE | SWP_NOZORDER );

	updateCaption();

	wchar_t bf[ 512 ];
	GetDlgItemTextT( m_idcFrom, bf, maxLen( bf ) );
	SetDlgItemTextT( m_idcFrom, cWCStr().FormatT( bf, m_srcName, m_srcURL ) );

	wcscpy_s( bf, m_dstDir );

	if ( m_dstName != NULL )
	{
		size_t len = wcslen( bf );
		if ( ( len > 0 ) && ( bf[ len - 1 ] != L'\\' ) )
			bf[ len++ ] = L'\\';

		wcscpy_s( bf + len, maxLen( bf ) - len, m_dstName );
	}
		
	SetDlgItemTextT( m_idcDownloadTo, bf );

	DWORD theadId = 0;

	HANDLE hThread = CreateThread( 0, 0, stDownloadFunc, this, 0, &theadId );

	if ( hThread == NULL )
	{
		assert( false );
		return false;
	}

	return true;
}

INT_PTR DownloadFile(
		const wchar_t *dstDir,
		const wchar_t *dstName,
		const wchar_t *srcURL,
		const wchar_t *srcName,
		cDlgDownloadResources *rcIds
	)
{
	cDlgDownload dlg( dstDir, dstName, srcURL, srcName, rcIds );

	if ( dlg.m_res != dfeOK )
		return dlg.m_res;

	INT_PTR idRes = dlg.DoModal( rcIds->m_hMainInst, rcIds->m_dlgId, rcIds->m_hMainWnd );

	return idRes;
}

