// ProgressMT.cpp : implementation file
//

#include "cDlgProgressMT.h"
#include <assert.h>

/////////////////////////////////////////////////////////////////////////////
// cDlgProgressMT dialog
// Requirements::
// 1) pObject is pointer to structure where first item is pointer to this Dlg
//		This parameter will assigned in constructor.
// 2) m_pObject will passed to TreadProc..
// 3) ThreadProc have to call this->OnOK when finished.


static BOOL CALLBACK enumFunc_UpdateTable( HWND hWnd, LPARAM lParam )
{
	cDlgProgressMT *pThis = (cDlgProgressMT *)lParam;

	if ( pThis == NULL )
		return true;

	return pThis->updateTable( hWnd );
}

bool cDlgProgressMT::updateTable( HWND hWnd )
{
	char szClass[ 16 ];
	char szText[ 64 ];

	if (!::GetClassNameA(hWnd, szClass, sizeof(szClass)))
		return true;

	if ( _stricmp( szClass, "msctls_progress" ) == 0 )
		m_hWndT[ enProgress ] = hWnd;
	else
	if ( _stricmp( szClass, "msctls_progress32" ) == 0 )
		m_hWndT[ enProgress ] = hWnd;
	else
	if ( _stricmp( szClass, "static" ) == 0 )
	{
		if (::GetWindowTextA( hWnd, szText, maxLen( szText ) ) )
		{
			if ( _stricmp( szText, "100%" ) == 0 )
				m_hWndT[ enPercent ] = hWnd;
			else
				m_hWndT[ enText ] = hWnd;
		}
	}

	return true;
}

void cDlgProgressMT::updateHandles()
{
	if ( m_idcT != NULL )
	{
		for( int i = maxLen( m_hWndT ); i--; )
		{
			m_hWndT[i] = GetDlgItem( m_hDlg, m_idcT[i] );
		}

		return;
	}


	EnumChildWindows( m_hDlg, enumFunc_UpdateTable, (LPARAM)this );
}

cDlgProgressMT::cDlgProgressMT( typeThreadProc *proc, int idsTerminateQ, const wchar_t *caption, const wchar_t *text, int *idcT )
{
	m_threadProc = proc;
	m_caption.Add( caption );
	m_text.Add( text );
	m_idsTerminateQ = idsTerminateQ;
	m_cancelled = false;
	m_idcT = idcT;
	m_pVoidParam = NULL;
	m_idIcon = 0;

	for ( int i = maxLen( m_hWndT ); i--; )
		m_hWndT[i] = 0;
}

cDlgProgressMT::~cDlgProgressMT()
{
}

void cDlgProgressMT::OnCancel() 
{
	//
	//	We have to call the cDlgProgressMT::OnCancel twice to really stop the dialog.
	//	Fistly it happens when the Cancels button was pressed
	//	Next time it happens if thread function detect cancelling.
	//
	if ( m_cancelled )
	{
		//
		//	Caller has to set to NULL the m_threadHandle value to avoid additional termination.
		//
		if ( m_threadHandle != NULL )
		{
			TerminateThread( m_threadHandle, 0 );
			DWORD exitCode;

			//
			// if the thread is STILL_ACTIVE we will not close the dialog.
			//
			if ( GetExitCodeThread( m_threadHandle, &exitCode ) && ( exitCode != 0 ) )
				return;
		}

		EndDialog( m_hDlg, IDCANCEL );
		return;
	}

	if ( ( m_idsTerminateQ != 0 ) &&
		 ( IDYES != DlgMessageBoxT( m_idsTerminateQ, MB_YESNO | MB_DEFBUTTON2) )  )
			return;

	m_cancelled = TRUE;
	EnableWindow( GetItem( IDCANCEL ), false );
}

DWORD WINAPI cDlgProgressMT::staticThreadProc(  void *param )
{
	cDlgProgressMT *pThis = ( cDlgProgressMT * )param;

	DWORD res = -1;

	if ( ( pThis == NULL ) || ( pThis->m_threadProc == NULL ) )
	{
		assert( false );
		res = -1;
	}
	else
	{
		res = pThis->m_threadProc( pThis );
	}

	PostMessage( pThis->m_hDlg, wmStopMessage, 0, (LPARAM)res );

	return res;
}

LRESULT cDlgProgressMT::RealDlgProc(UINT uMsg, WPARAM wParam, LPARAM lParam ) 
{
	switch ( uMsg )
	{
		case wmStopMessage :
		{
			if ( wParam == noError )
				EndDialog( m_hDlg, IDOK );
			else
			{
				cWCStr msg( 100 );
				DlgMessageBoxT( 
					::FormatT( msg.m_szBody, msg.m_len, L"Error in processing\nError number %d", (int) wParam ),
					MB_ICONSTOP );

				EndDialog( m_hDlg, IDCANCEL );
			}

			m_dwRes = (DWORD)lParam;

			return lParam;
		}

		case WM_INITDIALOG:
		{
			ODSFunc;

			if ( m_idIcon != 0 )
			{
				SetDlgIcon( m_idIcon, ICON_SMALL );
				SetDlgIcon( m_idIcon, ICON_BIG );
			}

			updateHandles();

			if ( !m_caption.isEmpty() )
				SetWindowTextT( m_hDlg, m_caption.m_szBody );

			if ( !m_text.isEmpty() )
				SetWindowTextT( m_hWndT[ enText ], m_text.m_szBody );

			SendMessage( m_hWndT[ enProgress ], PBM_SETRANGE, 0, MAKELPARAM( 0, 100 ) );

			setProgressPosition( 0 );

			void *pVoid = (void*)this;
			ODS_fls( _FormatT( L"%I64X", pVoid ) );

			m_threadHandle = CreateThread( NULL, 0, staticThreadProc, pVoid, 0, &m_threadID );

			assert( m_threadHandle != NULL );

			ShowWindow( m_hDlg, SW_SHOW );
			
			return true; 
		}

		case WM_COMMAND:
		{
			int cmd = LOWORD( wParam );

			if ( cmd == IDCANCEL )
				OnCancel();

			return true;
		}

		case WM_DESTROY:
		{
			if ( m_threadHandle )
				CloseHandle( m_threadHandle );
		}

		default:
			break;
	}

	return false;
}

void cDlgProgressMT::setProgressPosition( int position )
{
	assert ( position >= 0 );
	if ( position > 100 )
	{
		assert( FALSE );
		position = 100;
	}

	PostMessage( m_hWndT[ enProgress ], PBM_SETPOS, position, 0 );

	wchar_t text[ 20 ];
	::FormatT( text, maxLen( text ), L"%d%%", position );

	SetWindowTextT( m_hWndT[ enPercent ], text );
}

