#include "cDlgTabView.h"
#include "..\Includes\cStringBuffer.h"

static bool ReadValue( wchar_t *bf, int bfLen, const wchar_t *str, int num )
{
	//
	// Finds first symbol at tabbed string.
	//
	for( int n = num; n > 0 ; str++ )
	{
		if ( *str == L'\n' ) return false;
		if ( *str == L'\0' ) return false;
		if ( *str == L'\t' ) n--;
	}

	//
	// Copy upto next tabbed item
	//
	for( int i = 0; ; i++, str++ )
	{
		if ( i >= bfLen )
		{
			if ( bfLen == 0 )
			{
				assert( false );
				return false;
			}

			bf[ bfLen - 1 ] = L'\0';
			return false;
		}

		if ( ( *str != L'\n' ) && ( *str != L'\t' ) && ( *str != L'\0' ) )
		{
			bf[i] = *str;
			continue;
		}

		if ( *str == L'\t' )
			str++;

		bf[i] = L'\0';
		break;
	}

	return true;
}

static int ListView_InitColunms( HWND hLView, HINSTANCE hInst, const wchar_t *columnLabelStr )
{
	int i;
	LV_COLUMNW lvc;

	memset( &lvc, 0, sizeof( lvc ) );

	lvc.mask = LVCF_FMT | LVCF_WIDTH | LVCF_TEXT | LVCF_SUBITEM;

	wchar_t text[ 80 ];
	const wchar_t *str = columnLabelStr;

	for( i=0; ; i++ )
	{
		if ( !ReadValue( text, maxLen( text ), str, i ) )
			break;

		lvc.iSubItem = i;
		lvc.pszText = text; // _gszColumnLabel[i];
		lvc.cx = 120; //_gnColumnWidth[i];
		lvc.fmt= LVCFMT_CENTER; //_gnColumnFmt[i];

		ListView_InsertColumnT( hLView, i, &lvc );
	}

	return i;
}

bool FillListView( HWND hTab, cStringBuffer *sb )
{
	LVITEMW lvi;
	memset( &lvi, 0, sizeof( lvi ) );
	
	int item;

	wchar_t str[ 1024 ];
	wchar_t bf[ 256 ];

	for( int n = 0; sb->ReadStringT( str, maxLen( str ) ); n++ )
	{
		lvi.iItem = n;
		item = ListView_InsertItemT( hTab, &lvi );

		for( int i = 0; ReadValue( bf, maxLen( bf ), str, i ); i++ ) 
		{
			ListView_SetItemTextT( hTab, item, i, bf );
		}
	}

	return true;
}

bool cDlgTabView::InitDialog()
{
	if ( m_caption != NULL )
		SetWindowTextT( m_hDlg, m_caption );

	if ( m_idiIcon != 0 )
		SetDlgIcon( m_idiIcon );

	HWND hTab = GetDlgItem( m_hDlg, m_idcView );
	m_sizedCtrls.Init( m_hDlg );
	m_sizedCtrls.AddCtrl( hTab,	cSizedCtrls::ptSDR );

	ListView_SetExtendedListViewStyleEx( hTab, LVS_EX_GRIDLINES, LVS_EX_GRIDLINES );

	cStringBuffer sb( (BYTE*)m_tabbedText, (int)wcslen( m_tabbedText ) * sizeof( wchar_t ) );
	wchar_t columnNames[ 1024 ];
	if ( !sb.ReadStringT( columnNames, maxLen( columnNames ) ) )
		return false;

	m_iColumns = ListView_InitColunms( hTab, m_hInst, columnNames );

	bool ok = FillListView( hTab, &sb );

	m_rv.InitColumns( m_iColumns );
	m_rv.Restore( m_hDlg, GetItem( m_idcView ) );

	return ok;
}
