#pragma once

#include "UnicodeFuncs.h"

//
//	calss cEnumHardLinks
//		Enumerates hard link names.
//		Created:	11.12.2015
//
//	Usage:
//		cEnumHardLinks efn();
//
//		if ( !efn.findFirst( "AAA" ) )
//			return efn.m_err;
//		do
//		{
//			... efn.getLinkName()
//		}
//		while( efn.findNext() );
//		return efn.m_err;
//		

class cEnumHardLinks
{
	HANDLE m_hFindStream;
	cWCS m_linkName;
	cWCS m_path;
	int m_rootLen;

public:

	DWORD m_err;
	cEnumHardLinks( const wchar_t *hardLinkPath = NULL )
		: m_path( MAX_PATH )
	{
		m_err = ERROR_BAD_ENVIRONMENT;

		if ( hardLinkPath != NULL )
			findFirst( hardLinkPath );
	}

	~cEnumHardLinks()
	{
		if ( m_hFindStream != INVALID_HANDLE_VALUE )
			FindClose( m_hFindStream );
	}

	bool allocate( int len )
	{
		if ( m_linkName.SetLength( len ) == 0 )
		{
			ODS_fls( L"Assertion! allocation error" );
			m_err = ERROR_NOT_ENOUGH_MEMORY;
			return false;
		}

		return true;
	}

	const wchar_t *getLinkName()
	{
		m_path.SetLength( m_rootLen );
		 
		m_path.Add( m_linkName.m_szBody );
		return m_path.m_szBody;
	}

	bool findFirst( const wchar_t *hardLinkPath )
	{
		if (!GetRootDirT( m_path.m_szBody, hardLinkPath ) )
		{
			m_err = ERROR_DIR_NOT_ROOT;
			return false;
		}

		m_rootLen = m_path.GetLength();

		if (m_path.getAt(m_rootLen-1) == L'\\' )
			m_rootLen--;

		m_err = ERROR_SUCCESS;
		if ( !allocate( MAX_PATH ) )
			return false;

		DWORD len = m_linkName.m_allocatedLen;
		m_hFindStream = FindFirstFileNameT( hardLinkPath, 0, &len, m_linkName.m_szBody );
		
		if (m_hFindStream == INVALID_HANDLE_VALUE)
		{
			m_err = GetLastError();

			if ( m_err == ERROR_MORE_DATA )
			{
				if ( !allocate( len ) )
					return false;
				
				len = m_linkName.m_allocatedLen;
				m_hFindStream = FindFirstFileNameT( hardLinkPath, 0, &len, m_linkName.m_szBody );
			}

			if (m_hFindStream == INVALID_HANDLE_VALUE)
				m_err = GetLastError();
		}

		return m_err == ERROR_SUCCESS;
	}

	bool findNext()
	{
		DWORD len = m_linkName.m_allocatedLen;
		BOOL bRes = FindNextFileNameT( m_hFindStream, &len, m_linkName.m_szBody );

		if ( bRes )
			return true;

		m_err = GetLastError();
		if ( m_err == ERROR_HANDLE_EOF )
		{
			m_err = ERROR_SUCCESS;
			return false;
		}

		if ( m_err != ERROR_MORE_DATA )
			return false;

		if ( !allocate( len ) )
			return false;
			
		len = m_linkName.m_allocatedLen;

		bRes = FindNextFileNameT( m_hFindStream, &len, m_linkName.m_szBody );

		if ( bRes )
			return true;

		m_err = GetLastError();

		return false;
	}
};
