#pragma once

#include "UnicodeFuncs.h"

class cFindFile
{
public:
	WIN32_FIND_DATAW m_data;
	HANDLE m_hFind;

	cWCStr m_path;
	int m_fileNamePos;
	int m_err;

	inline DWORDLONG getSize()
	{
		return MAKEDWORDLONG( m_data.nFileSizeHigh, m_data.nFileSizeLow );
	}
	
	cFindFile( const wchar_t *dir, const wchar_t *pattern = NULL )
	{
		if ( pattern == NULL )
			pattern = L"*";

		memset( &m_data, 0, sizeof( m_data ) );
		m_err = NO_ERROR;

		m_hFind = INVALID_HANDLE_VALUE;
		m_path.Set( dir );

		m_fileNamePos = (int)wcslen( m_path.m_szBody );

		if ( m_fileNamePos == 0 )
		{
			ODS_fls( L"Assertion!" );
			m_err = ERROR_FILE_NOT_FOUND;
		}
		else
		if ( m_path.m_szBody[ m_fileNamePos - 1 ] != L'\\' )
		{
			m_path.Add( L"\\" );
			m_fileNamePos++;
		}

		m_path.SetLength( m_fileNamePos + MAX_PATH + 1 );
		
		copyString( m_path.m_szBody + m_fileNamePos, pattern, m_path.m_len - m_fileNamePos );
	}
	
	~cFindFile()
	{
		Close();
	}

	void Close()
	{
		if ( m_hFind == INVALID_HANDLE_VALUE )
			return;

		FindClose( m_hFind );
		m_hFind = INVALID_HANDLE_VALUE;
	}

	bool isFound()
	{
		return m_hFind != INVALID_HANDLE_VALUE;
	}

	bool findFirstFile()
	{
		m_hFind = FindFirstFileT( m_path.m_szBody, &m_data, &m_err );

		if ( !isFound() )
			return false;

		while( isServiceDir() )
		{
			if ( !findNextFile() )
				return false;
		}

		return true;
	}

	bool findNextFile()
	{
		if ( m_hFind == INVALID_HANDLE_VALUE )
		{
			ODS_fls( L"Assertion!" );
			return false;
		}

		while ( FindNextFileT( m_hFind, &m_data ) )
		{
			// 11.05.2016 
			// Strange behavior 'NTFS share' at FindFirstFile and FindNextFile reports '.' folder in unpredictable order ( other systems catches it at FindFirstFile only )
			if ( isServiceDir() )
				continue;

			return true;
		}

		m_err = GetLastError();

		Close();

		return false;
	}

	bool noMoreFiles()
	{
		return m_err == ERROR_NO_MORE_FILES;
	}

	const wchar_t *getCurrentPath()
	{
		copyString( m_path.m_szBody + m_fileNamePos, m_data.cFileName, m_path.m_len  - m_fileNamePos );
		return m_path.m_szBody;
	}

	DWORDLONG getCurrentSize()
	{
		ULARGE_INTEGER uli;
		uli.HighPart = m_data.nFileSizeHigh;
		uli.LowPart  = m_data.nFileSizeLow;

		return uli.QuadPart;
	}

	inline bool isServiceDir()
	{
		if ( !isDir() )
			return false;

		if ( wcscmp( m_data.cFileName, L"." ) == 0 ) return true;
		if ( wcscmp( m_data.cFileName, L".." ) == 0 ) return true;

		return false;
	}

	inline bool isDir()
	{	return ( ( m_data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) != 0 ); }

	static bool DirectoryIsEmpty( const wchar_t *dir )
	{
		cFindFile cff( dir );

		if ( cff.findFirstFile() )
			return false;

		return cff.noMoreFiles();
	}
};