#include "cMBStr.h"
#include <string.h>
#include <stdio.h> // For error fprintf only
#include <assert.h>

#ifdef _WIN32

//
//	Is reliease at UnicodeFuncs.cpp
//
extern int _wcstombs( char *mbs, int mbsLen, const wchar_t *wcs, int wcsLen, bool doubleZeroTerminated, UINT codePage );
extern int _mbstowcs( wchar_t *wcs, int wcsLen, const char *mbs, int mbsLen, bool doubleZeroTerminated, UINT codePage );

char *cMBStr::allocateAndCopyStr( const wchar_t *wcs, int *pAllocatedLen, int wcsLen,  bool doubleZeroTerminated, UINT codePage, int *pLen )
{
	if ( wcs == NULL )
		return NULL;

	int mbsLen = _wcstombs( NULL, 0, wcs, wcsLen, doubleZeroTerminated, codePage );

	if ( pLen != NULL )
		*pLen = mbsLen;

	char *mbs = allocateStr( ++mbsLen, pAllocatedLen );

	if ( mbs != NULL )
		_wcstombs( mbs, mbsLen, wcs, wcsLen, doubleZeroTerminated, codePage  );

	return mbs;
}


int cMBStr::Add( const wchar_t *wcs, int wcsLen, bool doubleZeroTerminated, UINT codePage )
{
	if ( wcs == NULL )
		return NULL;

	int mbsLen = _wcstombs( NULL, 0, wcs, wcsLen, doubleZeroTerminated, codePage );

	// mbsLen includes the byte for the terminating null character.

	if ( ( mbsLen + m_len + doubleZeroTerminated ) < m_allocatedLen )
	{
		_wcstombs( m_szBody + m_len - 1, mbsLen, wcs, wcsLen, doubleZeroTerminated, codePage  );
		m_len += mbsLen - 1;
		return m_len;
	}

	cMBStr mbs( wcs, wcsLen, doubleZeroTerminated, codePage );
	return Add( mbs.m_szBody, -1 );
}

bool cMBStr::Set( const wchar_t *ws, int len, bool doubleZeroTerminated, UINT codePage )
{
	if ( m_szBody != NULL )
	{
		m_len = 1;
		m_szBody[0] = '\0';
		return Add( ws, len, doubleZeroTerminated, codePage ) != 0;
	}

	m_szBody = allocateAndCopyStr( ws, &m_allocatedLen, len, doubleZeroTerminated, codePage, &m_len );
	if ( m_szBody == NULL )
		return false;

	return true;
}

int cMBStr::WriteTo( wchar_t *bf, int bfMaxLen, UINT codePage )
{
	return _mbstowcs( bf, bfMaxLen, m_szBody, -1, false, codePage );
}

cMBStr::cMBStr( const char *mbs, int len )
{
	m_szBody = NULL;
	m_len = 0;

	Set( mbs, len );
}

cMBStr::cMBStr( const wchar_t *ws, int len, bool doubleZeroTerminated, UINT codePage )
{
	m_szBody = NULL;
	m_len = 0;

	Set( ws, len, doubleZeroTerminated, codePage );
}

#endif // #ifdef _WIN32

void cMBStr::freeStr( char **pStr, int *pAllocatedLen )
{
	if ( pStr == NULL )
		return;

	//
	//	16.02.2011
	//	The error fixed 
	//	It was - 'if ( *pStr == NULL )'
	//

	if ( *pStr != NULL )
	{
		if ( pAllocatedLen != NULL )
		{
			memset( *pStr, 0, *pAllocatedLen );
			*pAllocatedLen = 0;
		}

		delete *pStr;
	}

	*pStr = NULL;
}

char *cMBStr::allocateStr( int strLen, int *pAllocatedLen )
{
	if ( pAllocatedLen == NULL )
		return NULL;

	const int allocationBlockSize = 4 * 1024;

	int blockCount = ( strLen + allocationBlockSize - 1 ) / allocationBlockSize;

	*pAllocatedLen = blockCount * allocationBlockSize;

    char *ptr = new char[ *pAllocatedLen ];
    if ( ptr != NULL )
        memset( ptr, 0, *pAllocatedLen );

    return ptr;
}

cMBStr::cMBStr( int len )
{
	m_szBody = allocateStr( ++len, &m_allocatedLen );

	if ( m_szBody != NULL )
		m_len = len;
	else
		m_len = 0;
}

bool cMBStr::Set( const char *mbs, int mbsLen )
{
	if ( mbsLen == -1 )
		mbsLen = (int)strlen( mbs );

	if ( m_szBody != NULL )
	{
		if ( m_allocatedLen < mbsLen + 1 )
			Free();
	}

	if ( m_szBody == NULL )
		m_szBody = allocateStr(  mbsLen + 1, &m_allocatedLen );

	if ( m_szBody == NULL )
		return false;

	memcpy( m_szBody, mbs, mbsLen );
	m_szBody[ mbsLen ] = '\0';
	m_len = mbsLen + 1;

	return true;
}

int cMBStr::Add( const char *mbs, int len )
{
	if ( m_szBody == NULL )
	{
		Set( mbs, len );
		return m_len;
	}

	if ( len == -1 )
	{
		if ( mbs == NULL )
			return 0;

		len = (int)strlen( mbs ); // while( mbs[len] != '\0' ) len++;
	}

	int newLen = m_len + len;

	if ( m_allocatedLen > newLen + 1 )
	{
		if ( mbs == NULL )
			memset( m_szBody + m_len - 1, 0, len );
		else
			memcpy( m_szBody + m_len - 1, mbs, len );

		m_len = newLen;
		m_szBody[ m_len - 1 ] = '\0';
		return true;
	}

	int tmp_len = newLen;
	int tmp_allocatedLen = 0;
	char *tmp_ptr = allocateStr( tmp_len, &tmp_allocatedLen );

	if ( tmp_ptr == NULL )
		return 0;

	//
	//	Note, m_len includes last zero byte
	//
	if ( m_len > 1 )
		memcpy( tmp_ptr, m_szBody, m_len );

	if ( mbs != NULL )
		memcpy( tmp_ptr + m_len - 1, mbs, len );

	if ( tmp_len < 1 )
	{		
        assert( tmp_len > 1 );
	    return 0;
	}
	
	tmp_ptr[ tmp_len - 1 ] = '\0';
		
	Free();

	m_szBody = tmp_ptr;
	m_len = tmp_len;
	m_allocatedLen = tmp_allocatedLen;

	return len;
}

void cMBStr::Free()
{
	freeStr( &m_szBody, &m_allocatedLen );
	m_len = 0;
}

cMBStr::~cMBStr()
{
	freeStr( &m_szBody, &m_allocatedLen );
}

const char *cMBStr::MakeLower()
{
	for( int i = m_len; i--; )
		m_szBody[i] = (char)tolower( m_szBody[i] );

	return m_szBody;
}

const char *cMBStr::MakeUpper()
{
	for( int i = m_len; i--; )
		m_szBody[i] = (char)toupper( m_szBody[i] );

	return m_szBody;
}
