/*
 *	Copyright 2015 Jetico, Inc. All rights reserved.
 *	OpenFilesList.h
 *	Description:
 *		Class cOpenFilesList
 *			Walks multy or single string result of BrowseFiles.
 *		Sample:
 *			cWCS wcs( 4 * 1024 );
 *			if ( !BrowseFiles( m_hDlg, wcs.m_szBody, wcs.m_allocatedLen, NULL, NULL, true ) )
 *				return false;
 *
 *			cOpenFilesList bfl( wcs.m_szBody, wcs.m_allocatedLen );
 *
 *			while( bfl.getNext() != NULL )
 *				.. Do somthing with bfl.getCur()
 *
 *	Modified:	04.03.2015
 *	Author:		Alexander Pichuev.
 */

#pragma once

#include "UnicodeFuncs.h"
#include "cFilesList.h"

class cOpenFilesList : public cFilesList
{
	bool m_bMultyString;
	bool m_bAddSlash;

	cWCS m_wcs;

public:
	
	cOpenFilesList( const wchar_t *bf, int bfSize = -1 )
		: cFilesList( bf, bfSize )
	{
		ODS_fls( _FormatT( L"'%s' %d", bf, bfSize ) );

		m_bMultyString = false;
		m_bAddSlash = false;
	}

	const wchar_t *getCur()
	{
		return m_wcs.m_szBody;
	}

	const wchar_t *getNext()
	{
		if ( m_bf == 0 )
			return NULL;

		if ( m_pos == NULL ) // first call
		{
			int len = (int)wcslen( m_bf );

			if ( len > m_bfSize )
				return false;

			if ( len == 0 )
				return false;

			m_bMultyString = ( m_bf[len+1] != L'\0' );
			m_bAddSlash = m_bf[len-1] != L'\\'; // Windows 8 does not add slash at directory string.

			if ( !m_bMultyString )
			{
				m_pos = m_bf;
				m_wcs.Set( m_bf );
				return m_wcs.m_szBody;
			}

			m_pos = m_bf + len + 1;
		}

		if ( !m_bMultyString )
			return NULL;

		if ( m_pos >= m_bf + m_bfSize )
			return NULL;

		if ( *m_pos == L'\0' )
			return NULL;

		m_wcs.Set( m_bf );

		if ( m_bAddSlash )
			m_wcs.Add( L"\\" );

		m_wcs.Add( m_pos );

		m_pos += wcslen( m_pos ) + 1;

		ODS_fls( m_wcs.m_szBody );

		return m_wcs.m_szBody;
	}
};
