//////////////////////////////////////
//
//	Created: 17.06.2010
//	Class cProfileTable
//	
//	Same as GetPrivateProfileStringT and GetProfileStringT
//	But it fills sections and values tables ones at cProfileTable::ReadFromBuffer
//
//	Following functions cProfileTable::FindSection and cProfileTableSection::FindValue allows quickly find 
//	required value at required section.
//
////////////////////////////////////

#pragma once

#include "UnicodeFuncs.h"
#include "autoDeallocate.h"
#include "cStringBuffer.h"

class cProfileTableValue
{
public:

	wchar_t *m_name;
	wchar_t *m_value;
	
	//
	//	Parameter may be used as attribute or pointer.
	//
	void *m_param;
	
	cProfileTableValue()
	{
		m_name = NULL;
		m_value = NULL;
		m_param = NULL;
	}
	
	~cProfileTableValue()
	{
		Free();
	}

	void Free()
	{
		deleteString( &m_name );
		deleteString( &m_value );
	}
};

class cProfileTableSection
{
public:
	cADTable < cProfileTableValue > m_valueT;
	wchar_t *m_name;
	void *m_param;
	
	cProfileTableSection()
	{
		m_name = NULL;
		m_param = NULL;
	}
	
	~cProfileTableSection()
	{
		Free();
	}

	void Free()
	{
		deleteString( &m_name );
		m_valueT.Free();
	}
	
	cProfileTableValue *FindValue( const wchar_t *szValue, int len = -1, bool bCreate = false );
	cProfileTableValue *AddValue( const wchar_t *szName, int nameLen = -1, const wchar_t *szValue = NULL, int valueLen = -1 );
	int GetValueNumber( const wchar_t *szValueName, int len );
	bool RemoveValue( const wchar_t *szName, int nameLen = -1 );
};

class cProfileTable
{
public:
	cADTable < cProfileTableSection > m_sectionT;

	cProfileTable( BYTE *bf = NULL, int bfSize = 0 )
	{
		if ( bfSize != NULL )
			ReadFromBuffer( bf, bfSize );
	}

	~cProfileTable()
	{
		Free();
	}

	void Free()
	{
		m_sectionT.Free();
	}

	bool ReadFromBuffer( BYTE *bf, int bfSize );
	bool ReadFromStringBuffer( cStringBuffer *sb );
	bool WriteToPtb( cPlainTextBlock *ptb, bool bUnicode = true ); // if not uinicode all strings are converted to multi byte string ( cMBStr ).
	
	cProfileTableSection *FindSection( const wchar_t *szSection, int len = -1, bool bCreate = true );
};