/*
 *	Copyright 2019 Jetico Inc. All rights reserved.
 *	cRebootRequest.h
 *	Created:	16.05.2019
 *	Author:		Alexander Pichuev.
 */

#pragma once

#include <windows.h>

//
//	Class cRebootRequest
//		Sets request for reboot to be checked in current session or after reboot.
//
class cRebootRequest
{
	static wchar_t *getPath( const wchar_t *szAppName )
	{
		static bool used = false;
		static wchar_t path[ MAX_PATH ];

		if ( used )
			return path;

		used = GetTempPathT( maxLen( path ), path ) != 0;

		if ( szAppName != NULL )
		{
			wcscat_s( path, szAppName );
			wcscat_s( path, L"_" );
		}

		wcscat_s( path, L"BCRebootRequest.tmp" );

		return path;
	}

public:

	//
	//	Set
	//	 Creates new file in TEMP folder and marks it to be removed after reboot.
	//
	static bool Set( const wchar_t *szAppName = NULL )
	{
		wchar_t *path = getPath( szAppName );

		if ( FileExistsT( path ) )
			return true;

		HANDLE hFile = CreateFileT( path, GENERIC_WRITE, FILE_SHARE_WRITE | FILE_SHARE_READ, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_HIDDEN, NULL );

		if ( hFile == INVALID_HANDLE_VALUE )
			return false;

		CloseHandle( hFile );

		bool renamed = MoveFileExT( path, NULL, MOVEFILE_DELAY_UNTIL_REBOOT) != 0;

		return renamed;
	}

	//
	//	Returns true if was reboot after last request, otherwise return false.
	//	Checks if file created by RebootRequest is removed, if so reboot has been initiated.
	//
	static bool WasReboot( const wchar_t *szAppName = NULL )
	{
		ODSFunc;
		bool exist = FileExistsT( getPath( szAppName ) ) == TRUE;

		ODSFuncRes( exist ? getPath( szAppName ) : L"Does not exist" );

		return !exist;
	}

};