/*
 *	Copyright 2014 Jetico Inc. All rights reserved.
 *	cRegFuncs.h
 *	Base class with registry functions.
 *	Created:	06.11.2014
 *	Author:		Alexander Pichuev.
 */

#pragma once

#include "cWCS.h"
#include <windows.h>
#include <winReg.h>

#ifndef KEY_WOW64_32KEY
#define KEY_WOW64_32KEY 0x200
#define KEY_WOW64_64KEY 0x100
#endif

class cRegFuncs
{
	cWCS &getStaticWcs()
	{
		static cWCS wcs;
		return wcs;
	}

public:
	enum { sveNative = 0, sve32 = KEY_WOW64_32KEY, sve64 = KEY_WOW64_64KEY };

	const wchar_t *m_szMainSection;
	void SetMainSection( const wchar_t *szMainSection );

	cRegFuncs( const wchar_t *szMainSection = NULL );

	static bool GetRegString( HKEY root, const wchar_t *section, const wchar_t *subKey, wchar_t *bf, int size );
	static bool GetRegString( HKEY root, const wchar_t *section, const wchar_t *subKey, cWCS &wcs );
	static bool SetRegString( HKEY root, const wchar_t *section, const wchar_t *subKey, const wchar_t *bf );

	DWORD GetRegDW( HKEY root, const wchar_t * key, const wchar_t * subKey, bool useBCSection = true, DWORD defValue = -1, int sveFlags = sveNative );
	bool SetRegDW( HKEY root, const wchar_t *key, const wchar_t *subKey, DWORD value, bool useBCSection = true, int sveFlags = sveNative );

	static bool SetRegDW( HKEY root, const wchar_t *key, const wchar_t *subKey, DWORD value, int sveFlags ); //  = sveNative );

	bool GetRegBytes( HKEY root, const wchar_t *section, const wchar_t *subKey, BYTE **pValue, int *pSize , bool bcSection = true );
	bool SetRegBytes( HKEY root, const wchar_t *section, const wchar_t *subKey, BYTE *value, int size, bool bcSection = true );

	//
	//	Get and Set function of predefined flags
	//	Note, GetFlag function returns true if flag has never set before.
	//
	bool GetFlag( DWORD flag );
	bool SetFlag( DWORD flag, bool set = true, HKEY hRoot = HKEY_CURRENT_USER );
};
