#pragma once
#include <WinReg.h>
#include "ODS.h"
	
class cRegKey
{
protected:
	HKEY m_hKey;

public:

	DWORD m_disp;
	
	cRegKey( HKEY hKey = NULL )
	{
		m_hKey = hKey;
	}

	cRegKey( HKEY hRoot, const wchar_t *lpSubKey, DWORD ulOptions, REGSAM samDesired )
	{
		m_hKey = NULL;
		LONG lRes = openKey( hRoot, lpSubKey, ulOptions, samDesired );

		if ( ( lRes == ERROR_FILE_NOT_FOUND ) && ( samDesired & KEY_CREATE_SUB_KEY ) )
			lRes = createKey( hRoot, lpSubKey, ulOptions, samDesired );
	}

	LONG createKey( HKEY hRoot, const wchar_t *lpSubKey, DWORD ulOptions, REGSAM samDesired )
	{
		closeKey();

		LONG lRes = RegCreateKeyExW( hRoot, lpSubKey, 0, NULL, ulOptions, samDesired, NULL, &m_hKey, &m_disp );

		return lRes;
	}

	LONG openKey( HKEY hRoot, const wchar_t *lpSubKey, DWORD ulOptions, REGSAM samDesired )
	{
		closeKey();

		LONG lRes = RegOpenKeyExW( hRoot, lpSubKey, ulOptions, samDesired, &m_hKey );

		return lRes;
	}

	HKEY getKey()
	{
		return m_hKey;
	}

	LONG closeKey()
	{
		LONG lRes = 0;

		if ( m_hKey != NULL )
			lRes = RegCloseKey( m_hKey );
		
		m_hKey = NULL;

		return lRes;
	}
	
	~cRegKey()
	{
		closeKey();
	}

	static DWORD getDWValue( HKEY hKey, const wchar_t *szSection, const wchar_t *szValue, DWORD defValue = 0 )
	{
		ODSFuncS( szSection );

		cRegKey rk( hKey, szSection, NULL, KEY_READ );

		if ( rk.getKey() == NULL )
		{
			ODS_fls( _FormatT( L"Error in opening key '%s'", szSection ) );
			return defValue;
		}

		DWORD dwRes = 0;
		DWORD dwSize = sizeof( dwRes );
		DWORD dwType;

		if ( ERROR_SUCCESS != RegQueryValueExT( rk.getKey(), szValue, NULL, &dwType, (BYTE *)&dwRes, &dwSize ) )
		{
			ODS_fls( _FormatT( L"Error in quering value '%s'", szValue ) );
			return defValue;
		}

		ODSFuncRes( dwRes );
		return dwRes;
	}

	static bool setDWValue( HKEY hKey, const wchar_t *szSection, const wchar_t *szValue, DWORD value )
	{
		ODSFuncS( szSection );

		cRegKey rk( hKey, szSection, NULL, KEY_WRITE );

		if ( rk.getKey() == NULL )
		{
			ODS_fls( _FormatT( L"Error in opening key '%s'", szSection ) );
			return false;
		}

		DWORD dwSize = sizeof( value );
		DWORD dwType = REG_DWORD;

		if ( ERROR_SUCCESS != RegSetValueExT( rk.getKey(), szValue, NULL, dwType, (BYTE *)&value, dwSize ) )
		{
			ODS_fls( _FormatT( L"Error in quering value '%s'", szValue ) );
			return false;
		}

		return true;
	}
};
