#include <windows.h>
#include "..\CertificateLib\SMemory\SBlocks.h"
#include <string.h>
#include <assert.h>
#include "cStringArray.h"

#ifndef maxLen
#define maxLen( _XX ) ( sizeof( _XX ) / sizeof( *_XX ) )
#endif


int cStringArray::findString( const char *str )
{
	int i = Q;

	while (  i-- )
		if ( _stricmp( str, T[i] ) == 0 )
			break;

    return i;
}

int cStringArray::findString( const char *str, int len, bool fromBeginning, const char **pTail )
{
	int i = Q;

	if ( len == -1 )
		len = strlen( str );

	if ( fromBeginning )
	{
		for( i = 0; i < Q; i++ )
		{
			if ( strncmp( str, T[i], len ) == 0 )
				break;
		}

		if ( i >= Q )
			return -1;
	}
	else
	{
		while (  i-- )
			if ( strncmp( str, T[i], len ) == 0 )
				break;
	}

	if ( ( i >= 0 ) && ( pTail != NULL ) )
		*pTail = T[i] + len;

    return i;
}

int cStringArray::addString( const char *str )
{
	int len = (int)strlen( str );

	return addString( str, len );
}

static bool isDoubleZeroTerminated( const char *str )
{
	// it simple checks if the string has the { '\', '0' } symbol sequence

	if ( str[0] == 0 )
		return false;

	for ( int i = 1; str[i] != 0; i++ )
	{
		if ( ( str[i] == '0' ) && ( str[i-1] == '\\' ) )
			return true;
	}

	return false;
}

int cStringArray::addString( const char *str, int len, bool checkDoubleZero )
{
	// checks if the string is double zero terminated
	// Reason: RCFile sometimes does not add the last '\0' for double zero terminated strings but sometimes does.
	// so to avoid the ambiguity we will add the '\0' for double zero terminated string if it was skipped.

	bool addZero = checkDoubleZero && isDoubleZeroTerminated( str );

	if ( addZero )
	{
		if ( len < 3 )
			addZero = false;
		else
		if ( strcmp( str + len - 1, "\"" ) != 0 )
			addZero = false;
		else
		if ( strcmp( str + len - 3, "\\0\"" ) == 0 )
			addZero = false;
	}

	char *newStr = new char[ len + 1 + ( addZero ? 2 : 0 ) ];

	if ( newStr == NULL )
	{
		freeTable();
		return 0;
	}

	memcpy( newStr, str, len );
	newStr[ len ] = '\0';

	if ( addZero )
		strcpy_s( newStr + len - 1, 3, "\\0\"" );
	
	if ( ptbTable.WriteBytes( NULL, sizeof( char * ) ) == 0  )
	{
		freeTable();
		delete newStr;
		return 0;
	}

	T = ( char **)ptbTable.buffer;

	T[Q] = newStr;

	Q++;

	return Q;
}

bool cStringArray::removeString( int num )
{
	if ( num >= Q )
		return false;

	delete( T[num] );

	Q--;

	for( int i = num; i < Q; i++ )
	{
		T[i] = T[i+1];
	}

	return true;
}

bool cStringArray::concatenateString( int num, const char *addStr )
{
	if ( num >= Q )
		return false;

	if ( addStr == NULL )
		return false;

	int addLen = (int)strlen( addStr ) + 1;

	const char *s = T[num];

	char *newStr = NULL;

	if ( s == NULL )
		s = "";

	addLen += (int)strlen( s );
	newStr = new char[ addLen ];
	if ( newStr == NULL )
		return false;

	strcpy_s( newStr, addLen, s );
	strcat_s( newStr, addLen, addStr );

	delete ( T[num] );

	T[num] = newStr;

	return true;
}

int cStringArray::createTable( BYTE *buff, int buffSize, const char *prefixStr )
{
	freeTable();

	if ( ( buff == NULL ) || ( buffSize == 0 ) )
		return 0;

	int pos = 0;

	char path[ 4096 ];

	if ( prefixStr != NULL )
	{
		pos = (int)strlen( prefixStr );
		if ( pos >= sizeof( path ) )
			pos = sizeof( path ) - 1;

		strncpy_s( path, prefixStr, pos );
	}

	path[ pos ] = '\0';

	int lPos = pos;

	for( int ti = 0, biStart = 0, biFinish = 0 ; ; biFinish++ )
	{
		bool bTab = ( buff[biFinish] == '\t' ) && m_bTabulationAsNewLine ;

		if ( ( buff[biFinish] == '\r' ) || 
			 ( bTab ) ||
			 ( biFinish == buffSize )
			 )
		{
			if ( biStart == -1 )
			{
				//assert( false );
				//return 0;
				continue;
			}
			
			int len = biFinish - biStart;

			if ( lPos + len >= sizeof( path ) )
				len = sizeof( path ) - lPos - 1;

			strncpy_s( path + lPos, maxLen( path ) - lPos, (char *)buff + biStart, len );
			path[ lPos + len ] = '\0';

			if ( m_treatBackSlashAsStringConcatenation &&
				( path[ lPos + len - 1 ] == '\\' ) )
			{
				lPos += len - 1;
			}
			else
			{
				if ( addString( path ) == 0 )
					return 0;

				lPos = pos;
			}

			if ( biFinish == buffSize )
				break;

			if ( bTab )
				biStart = biFinish + 1; 
			else
				biStart = -1;
		}
		else
		if ( buff[biFinish] == '\n' )
			biStart = biFinish + 1;
	}

	return Q;
}
