//////////////////////////////////////
//
//	Class String Array...
//
///
#pragma once

#include "..\CertificateLib\SMemory\SBlocks.h"
#include "ODS.h"

class cStringArray
{
	//cPlainTextBlock ptbBuff;
	cPlainTextBlock ptbTable;

public:
	char **T; // Pointer to table
	int    Q; // Item quantity in the table
	bool m_bTabulationAsNewLine;

	cStringArray( BYTE *buff = NULL, int buffSize = 0, const char *prefixStr = 0, bool useBckSlash = false )
	{
		m_treatBackSlashAsStringConcatenation = useBckSlash;
		T = NULL;
		Q = 0;
		m_bTabulationAsNewLine = false;
		createTable( buff, buffSize, prefixStr );
	}

	~cStringArray()
	{
		freeTable();
	}

	void freeTable()
	{
		while( Q-- )
		{
			delete T[Q];
		}
		
		T = NULL;
		Q = 0;

		ptbTable.Free();
	}

	//
	//	createTable
	//		parses buff, finds strings and adds it to table
	//		if a prefixStr is assigned it will be automatically concatenated with the found string and added to table.
	//
	bool m_treatBackSlashAsStringConcatenation;
	int createTable( BYTE *buff, int buffSize, const char *prefixStr = NULL);

	int addString( const char *str, int len, bool checkDoubleZero = false );
	int addString( const char *str );
	bool concatenateString( int num, const char *addStr );

	int findString( const char *str, int len, bool fromBeginning = false , const char **pTail = NULL );
	int findString( const char *str );

	//
	//	removeString
	//		frees string and moves tail down.
	//
	bool removeString( int num );
};

typedef wchar_t * wcharPtr;

class cStringArrayT
{
	cPlainTextBlock ptbTable;
	enum { sigVal = 0x72726173 }; // "sarr"
	UINT sig;

public:

	bool checkSig()	{ return sig == sigVal; }

	// Iterator
	int m_current;
	const wchar_t *getCurrent();

	const wchar_t *getFirst()
	{
		m_current = 0;
		return getCurrent();
	}

	const wchar_t *getNext()
	{
		m_current++;
		return getCurrent();
	}

	const wchar_t *getPrev()
	{
		m_current--;
		return getCurrent();
	}

	bool m_bTabulationAsNewLine;
	wcharPtr *T; // Pointer to table
	int    Q; // Item quantity in the table

	void commonInit( bool useBckSlash = false, bool bTabulationAsNewLine = false )
	{
		m_current = 0;
		m_treatBackSlashAsStringConcatenation = useBckSlash;
		T = NULL;
		Q = 0;
		m_bTabulationAsNewLine = bTabulationAsNewLine;
		sig = sigVal;
	}

	cStringArrayT( BYTE *buff = NULL, int buffSize = 0, const wchar_t *prefixStr = 0, bool useBckSlash = false, bool bTabulationAsNewLine = false )
	{
		commonInit( useBckSlash, bTabulationAsNewLine );
		createTable( buff, buffSize, prefixStr );
	}

	cStringArrayT( const wchar_t *szText, int len = -1, const wchar_t *prefixStr = 0 )
	{
		commonInit();
		createTable( szText, len, prefixStr );
	}

	~cStringArrayT()
	{	freeTable();	}

	void freeTable();
	static wchar_t *allocateStr( int len );
	static void freeStr( wchar_t *wcs );
	static wchar_t *createCopy( const wchar_t *str, int len, bool checkDoubleZero = false );

	//
	//	createTable
	//		parses buff, finds strings and adds it to table
	//		if a prefixStr is assigned it will be automatically concatenated with the found string and added to table.
	//
	bool m_treatBackSlashAsStringConcatenation;
	int createTable( const BYTE *buff, int buffSize, const wchar_t *prefixStr = NULL, bool bForceUnicode = false );
	int createTable( const wchar_t *wcs, int wcsLen = -1, const wchar_t *prefixStr = NULL )
	{
		if ( wcsLen == -1 )
			wcsLen = (int)wcslen( wcs );

		return createTable( (BYTE*)wcs, wcsLen * sizeof( wchar_t ), prefixStr, true );
	}

	int addTable( const wcharPtr *pT, int pQ, bool checkDoubleZero = false );
	bool copyFrom( const cStringArrayT *sa, bool bAppend = true, bool checkDoubleZero = false );

	int addStringPtr( wchar_t *ptr );
	int addString( const wchar_t *str, int len, bool checkDoubleZero = false );
	int addString( const wchar_t *str );
	int addStringOnce( const wchar_t *str, int len = -1 );

	int findString( const wchar_t *str, int len, const wchar_t **pTail = NULL ) const;
	int findString( const wchar_t *str ) const;

	bool updateString( int pos, const wchar_t *str, int len = -1, bool checkDoubleZero = false );

	//
	//	compare
	//		returns  1 if this array has at least one different string
	//		returns -1 if 'sa' array has at least one different string
	//		returns 0 if there are no different string.//		
	//
	int compare( cStringArrayT *sa );

	//
	//	Finds string completely included by str.
	//	F.e. returns valid i if T[i]= 'C:\' and str = 'C:\Tmp'
	//
	int findStringIncluded( const wchar_t *str, int len = -1 );

	//
	//	removeString
	//		frees string and moves tail down.
	//
	bool removeString( int num );

	const wchar_t *getAt( int pos ) const
	{
		if ( pos >= Q ) return L"NULL";
		if ( pos < 0 ) return L"NULL";
		return T[ pos ];
	}

	bool setAt( int pos, const wchar_t *str, int len = -1 )
	{
		if ( pos < 0 )
			return false;

		while( pos >= Q )
		{
			if ( addString( L"" ) == 0 )
				return false;
		}

		freeStr( T[pos] );

		T[pos] = createCopy( str, len );

		return T[ pos ] != NULL;
	}

	bool isEmpty() const
	{
		return Q == 0;
	}

	static int getCount( const cStringArrayT *sa )
	{
		if ( sa == NULL )
			return 0;
		else
			return sa->Q;
	}

	static bool isEmpty( const cStringArrayT *sa )
	{
		if ( sa == NULL )
			return true;
		else
			return sa->isEmpty();
	}

	static void ODSTable( const cStringArrayT *sa );
};

#ifdef _ODS
static inline void ODS_ShowStringArray( const wchar_t *szAction, const cStringArrayT &sa )
{
	ODS_fls( _FormatT( L"%s, Q = %d", szAction, sa.Q ) );
	for( int i = 0; i < sa.Q; i++ )
		ODS_fls( _FormatT( L"%d) '%s'", i, sa.T[i] ) );
}
#else
#define ODS_ShowStringArray(...) ((void)0)
#endif
