#pragma once
#include "windefs.h"
#include "../CertificateLib/SMemory/SBlocks.h"

class cStringBuffer
{
	//
	//	Buffer for converted text.
	//
	cPlainTextBlock m_ptbBuff;
	bool SetBuffer( cPlainTextBlock *ptb )
	{
		return SetBuffer( ptb->buffer, ptb->bufferSize, false );
	}

public:

	bool SetBuffer( const BYTE *bf, int bfSize, bool init = true )
	{
		int size = m_ptbBuff.WriteBytes( bf, bfSize, true );
		m_bf = m_ptbBuff.buffer;
		m_bfSize = m_ptbBuff.bufferSize;
		if ( size != bfSize )
			return false;

		if ( init )
			Init( m_bf, m_bfSize );

		return true;
	}

	const BYTE *m_bf;
	SIZE_T m_bfSize;

	bool m_skipEmptyString;

	SIZE_T m_pos, m_startPos;
	bool m_isUnicode;
	bool m_isUTF8;
	bool m_bTabulationAsNewLine;

	void Restart()
	{
		m_pos = m_startPos;
	}

	int GetStringQ();

	cStringBuffer( const wchar_t *szText, int len = -1 );

	cStringBuffer( const BYTE *bf, int bfSize, bool bTabulationAsNewLine = false )
	{
		m_skipEmptyString = true;
		m_bTabulationAsNewLine = bTabulationAsNewLine;
		Init( bf, bfSize );
	}

	cStringBuffer()
	{
		m_skipEmptyString = true;
		m_bTabulationAsNewLine = false;
		Init( NULL, 0 );
	}

	void InitA( BYTE *bf, SIZE_T bfSize );
	bool ReadStringA( char *buff, int buffSize );
	bool m_bCompleted; // Read op[eration complete flag, is thrue if 'end of string' or 'end of file' are found.

#ifndef _WIN32	
	
	bool ReadString( char *buff, int buffSize )
	{
	    return ReadStringA( buff, buffSize );
	}
	void Init( BYTE *bf, SIZE_T bfSize )
	{
	    InitA( bf, bfSize );
	}
	
#else // #ifndef _WIN32

	void Init( const BYTE *bf, SIZE_T bfSize );
	bool ReadString( char *buff, int buffSize );
	bool ReadStringW( wchar_t *buff, int buffSize );
	bool ReadStringT( wchar_t *buff, int buffSize );

#ifdef _UnicodeFuncs_

	bool ReadStringT( cWCStr *wcs )
	{
		cWCStr bf( 1024 * 4 ); // 4K

		if ( bf.m_szBody == NULL )
			return false;
		
		wcs->Free();

		bool bFirst = true;

		while( true )
		{
			if ( !ReadStringT( bf.m_szBody, bf.m_len ) )
				return false;
			
			if ( m_bCompleted )
			{
				if ( bFirst )
				{
					wcs->Swap( &bf );
					wcs->Truncate();
				}
				else
					wcs->Add( bf.m_szBody );

				break;
			}
			
			bFirst = false;
			wcs->Add( bf.m_szBody );			
		}

		if ( wcs->m_szBody == NULL )
			return false;

		return true;
	}
#endif

	virtual bool FindString( wchar_t *buff, int buffLen, const wchar_t *str, bool getTailOnly = true );

	// Profiles functions
	// [Section]
	//	Value = 'Data'
	bool GetProfileStringT( wchar_t *buff, int *pBuffLen, const wchar_t *section, const wchar_t *value );

	//
	//	The virtual function is required if the class is inherited by cStringFile 
	//	to move mapped window to one step forward if required.
	//
	virtual bool UpdateMapping() { return true; }

# ifdef _UnicodeFuncs_

	bool GetProfileStringA( char *buff, int *pBuffLen, const char *section, const char *value )
	{
		cWCStr wSection( section );
		cWCStr wValue( value );

		wchar_t *wBuff = NULL;
		int wBuffLen = *pBuffLen;
		
		if ( wBuffLen != 0 )
			wBuff = new wchar_t[ wBuffLen ];

		bool res = GetProfileStringT( wBuff, pBuffLen, wSection.m_szBody, wValue.m_szBody );

		if ( res )
			_wcstombs( buff, *pBuffLen, wBuff );

		if ( wBuff != NULL )
			delete [] wBuff;

		return res;
	}
# endif // #ifdef _UnicodeFuncs
#endif // #ifdef _WIN32
};

#ifdef _WIN32
//
//	Wide char conversion functions.
//
bool ConvertFromBigEndian( cPlainTextBlock *ptb, const BYTE *bf, int bfSize );
bool ConvertFromUTF8( cPlainTextBlock *ptb, const BYTE *bf, int bfSize );
bool ConvertFromACP( cPlainTextBlock *ptb, const BYTE *bf, int bfSize );
bool ConvertToMBS( UINT codePage, cPlainTextBlock *ptb, const wchar_t *wcs, int wcsLen = -1 );

#define ConvertToUTF8( ptb, wcs, wcsLen ) ConvertToMBS( CP_UTF8, (ptb), (wcs), (wcsLen) )
#define ConvertToUTF7( ptb, wcs, wcsLen ) ConvertToMBS( CP_UTF7, (ptb), (wcs), (wcsLen) )
#define ConvertToACP(  ptb, wcs, wcsLen ) ConvertToMBS( CP_ACP,  (ptb), (wcs), (wcsLen) )

//
//	bytes2string
//		Returns number of processed bytes
//
int bytes2string( wchar_t *str, int strSize, BYTE *bf, int bfSize, const wchar_t cSeparator = ' ' );

//
//	string2bytes
//		If success returns number of parsed bytes
//		Otherwise returns -1
//
int string2bytes( BYTE *bf, int bfSize, const wchar_t *str );

template < class T > T string2int( const wchar_t *str, int radix = 10, const wchar_t **pStr = NULL, const wchar_t separator = 0 )
{
	while( *str == L' ' ) str++;
	if ( ( str[0] == L'0' ) && ( str[1] == L'x' ) ) // skip '0x' prefix
	{
		str += 2;
		radix = 16;

		if ( wcsncmp( str, L"I64", 3 ) == 0 )
			str += 3;
	}

	T dw = 0;

	for( ; *str; str++ )
	{
		BYTE bt = 0;
		if ( ( *str >= L'0' ) && ( *str <= L'9' ) )
			bt = *str - L'0';
		else
		if ( radix != 0x10 )
			break;
		else
		if ( ( *str >= L'a' ) && ( *str <= L'f' ) )
			bt = *str - L'a' + 0xA;
		else
		if ( ( *str >= L'A' ) && ( *str <= L'F' ) )
			bt = *str - L'A' + 0xA;
		else
		{
			if ( ( separator == 0 ) ||
				 ( separator != L' ' ) && ( separator != L'\t' ) )
			{
				while( ( *str == L' ' ) || ( *str == L'\t' ) )
					str++;
			}

			if ( ( separator != 0 ) && ( *str != separator ) )
				return T(-1);

			break;
		}

		T DW = dw * radix;
		
		if ( dw > DW ) // overflow
			return T(-1);

		dw = DW + bt;
	}

	if ( pStr != NULL )
		*pStr = str;

	return dw;
}

#endif // #ifdef _WIN32