#pragma once

//
//	Class cWS2Dll
//		Dynamically loads Ws2_32.dll and allows to use it's functions
//		The dll may be loaded immediately or it is loaded at the first time any fuction is used.
//		It is very convenient to use the feature for static defined ws2dll value declared here.
//

#include "CDll.h"

class cWS2Dll : public CDll
{
	typedef int PASCAL FAR typeWSAStartup( IN WORD wVersionRequired, OUT LPWSADATA lpWSAData);
	typedef int PASCAL FAR typeWSACleanup(void);
	typedef int PASCAL FAR type_connect ( IN SOCKET s, IN const struct sockaddr FAR *name, IN int namelen );
	typedef USHORT PASCAL FAR type_htons (IN USHORT hostshort);
	typedef struct hostent FAR * PASCAL FAR type_gethostbyname(IN const char FAR * name);
	typedef SOCKET PASCAL FAR type_socket ( IN int af, IN int type, IN int protocol );
	typedef int PASCAL FAR type_recv ( IN SOCKET s, OUT char FAR * buf, IN int len, IN int flags );
	typedef int PASCAL FAR type_send (IN SOCKET s, IN const char FAR * buf, IN int len, IN int flags);
	typedef int PASCAL FAR type_closesocket ( IN SOCKET s);
	typedef int PASCAL FAR type_shutdown ( IN SOCKET s, IN int how);
	typedef int PASCAL FAR type_select ( IN int nfds, IN OUT fd_set FAR *readfds, IN OUT fd_set FAR *writefds, IN OUT fd_set FAR *exceptfds, IN const struct timeval FAR *timeout);
	typedef unsigned long PASCAL FAR type_inet_addr (IN const char FAR * cp);
	typedef struct hostent FAR * PASCAL FAR type_gethostbyaddr( IN const char FAR * addr, IN int len, IN int type );

	typeWSAStartup *m_WSAStartup;
	typeWSACleanup *m_WSACleanup;
	typeWSACleanup *m_WSAGetLastError;
	type_connect *m_connect;
	type_htons *m_htons;
	type_gethostbyname *m_gethostbyname;
	type_socket *m_socket;

	type_recv *m_recv;
	type_send *m_send;
	type_closesocket *m_closesocket;
	type_shutdown *m_shutdown;
	type_select *m_select;
	type_inet_addr *m_inet_addr;
	type_gethostbyaddr *m_gethostbyaddr;

public:

	bool m_loaded;

	bool Load( bool showError )
	{
		if ( m_loaded )
			return !error;

		m_loaded = true;

		if ( !CDll::Load( false, "Ws2_32.dll", NULL, false ) )
		{
			//
			// Windows 95 does not have Ws2_32.dll, but has Wsock32.dll instead
			//
			if ( !CDll::Load( false, "Wsock32.dll", NULL, false ) )
				if ( !CDll::Load( false, "Ws2_32.dll", NULL, showError ) )
					return false;
		}

		if ( !GetProcAddress( "WSAStartup",		showError, (void**)&m_WSAStartup		) ) return false;
		if ( !GetProcAddress( "WSACleanup",		showError, (void**)&m_WSACleanup		) ) return false;
		if ( !GetProcAddress( "WSAGetLastError",showError, (void**)&m_WSAGetLastError	) ) return false;
		if ( !GetProcAddress( "connect",		showError, (void**)&m_connect			) ) return false;
		if ( !GetProcAddress( "htons",			showError, (void**)&m_htons				) ) return false;
		if ( !GetProcAddress( "gethostbyname",	showError, (void**)&m_gethostbyname		) ) return false;
		if ( !GetProcAddress( "socket",			showError, (void**)&m_socket			) ) return false;
		if ( !GetProcAddress( "recv",			showError, (void**)&m_recv				) ) return false;
		if ( !GetProcAddress( "send",			showError, (void**)&m_send				) ) return false;
		if ( !GetProcAddress( "closesocket",	showError, (void**)&m_closesocket		) ) return false;
		if ( !GetProcAddress( "shutdown",		showError, (void**)&m_shutdown			) ) return false;
		if ( !GetProcAddress( "select",			showError, (void**)&m_select			) ) return false;
		if ( !GetProcAddress( "inet_addr",		showError, (void**)&m_inet_addr			) ) return false;
		if ( !GetProcAddress( "gethostbyaddr",	showError, (void**)&m_gethostbyaddr		) ) return false;

		return true;
	}

	cWS2Dll( bool loadNow = false, bool showError = true ) : CDll()
	{
		m_loaded = false;
		
		if ( loadNow )
			Load( showError );
	}

	int WSAStartup( IN WORD wVersionRequired, OUT LPWSADATA lpWSAData )
	{ return ( !Load( true ) ? -1 : m_WSAStartup( wVersionRequired, lpWSAData ) ); }

	int WSACleanup()
	{ return ( !Load( true ) ? -1 : m_WSACleanup() ); }

	int WSAGetLastError()
	{ return ( !Load( true ) ? -1 : m_WSAGetLastError() ); }

	int connect( IN SOCKET s, IN const struct sockaddr FAR *name, IN int namelen )
	{ return ( !Load( true ) ? -1 : m_connect( s, name, namelen ) ); }

	USHORT htons( USHORT hostshort )
	{ return ( !Load( true ) ? -1 : m_htons( hostshort ) ); }

	struct hostent *gethostbyname( const char * name )
	{ return ( !Load( true ) ? NULL : m_gethostbyname( name ) ); }

	SOCKET socket( int af, int type, int protocol )
	{
		if ( !Load( true ) )
			return -1;
		else
			return m_socket( af, type, protocol );
	}

	int recv( SOCKET s, char *buf, int len, int flags )
	{ return ( !Load( true ) ? -1 : m_recv( s, buf, len, flags ) ); }

	int send( SOCKET s, const char *buf, int len, int flags )
	{ return ( !Load( true ) ? -1 : m_send( s, buf, len, flags ) ); }

	int closesocket( SOCKET s )
	{ return ( !Load( true ) ? -1 : m_closesocket( s ) ); }

	int shutdown( SOCKET s, int how )
	{ return ( !Load( true ) ? -1 : m_shutdown( s, how ) ); }

	int select( int nfds, fd_set *readfds, fd_set *writefds, fd_set *exceptfds, const struct timeval *timeout )
	{ return ( !Load( true ) ? -1 : m_select( nfds, readfds, writefds, exceptfds, timeout ) ); }

	unsigned long inet_addr(IN const char FAR * cp)
	{ return ( !Load( true ) ? -1 : m_inet_addr( cp ) ); }

	struct hostent *gethostbyaddr( const char *addr, int len, int type )
	{ return ( !Load( true ) ? NULL : m_gethostbyaddr( addr, len, type ) ); }
};