/*
 *	Copyright 2016 Jetico, Inc. All rights reserved.
 *	Class cWaitRegistryKeys realisation
 *	Created:	09.06.2016
 *	Author:		Alexander Pichuev.
 */

#include "cWaitRegistryKeys.h"
#include "cSemaphore.h"

DWORD cWaitRegistryKeys::waitFunction()
{
	if ( m_szThreadName != NULL )
		ODS_SetThreadName( m_szThreadName );

	ODSFunc;

	cSemaphore csStop( 0, 1, m_szStopSemaphore );

	DWORD timeout = m_timeout;

	if ( csStop.m_err == ERROR_ALREADY_EXISTS )
	{
		ODS_fls( L"Assertion! Semaphore object already exists" );
		return DWORD( -1 );
	}
	else
	if ( csStop.m_err != 0 )
	{
		ODS_fls( _FormatT( L"Assertion! Error '%s' in creating semaphore", cWCStr().GetSystemError( csStop.m_err ) ) );
		return DWORD( -1 );
	}

	//cWCStr wcsKeyName;
	//wcsKeyName.FormatT( L"%s\\SwapFileStatus", szBCWipeKey );
	
	while( true )
	{
		int keyQ = m_saSubKeys.Q;
		cADBlock < cRegNotifier > adbNotifiers( keyQ, false );
		cADBlock < HANDLE > adbHandles( keyQ + 1 );

		if ( ( adbNotifiers.m_T == NULL ) || ( adbHandles.m_T == NULL ) )
		{ ODS_fls( L"Assertion!" ); return DWORD( -1 ); }

		for( int i = 0; i < keyQ; i++ )
		{
			if ( !adbNotifiers.m_T[i].create( m_hRoot, m_saSubKeys.T[i], m_bWow64 ) )
			{ ODS_fls( L"Error in notifiers creating." ); return DWORD( -1 ); }
			
			adbHandles.m_T[i] = adbNotifiers.m_T[i].m_handle;
		}

		adbHandles.m_T[keyQ] = csStop.m_hSemaphore;

		// Wait for an event to occur.
		DWORD dwRes = WAIT_TIMEOUT;
		
		while( true )
		{
			dwRes = WaitForMultipleObjects( adbHandles.m_Q, adbHandles.m_T, false, timeout );
			
			DWORD err = GetLastError();

			ODS_fls( 
				_FormatT( 
					L"Wait stop - %s",
						dwRes == WAIT_TIMEOUT ? L"Timeout" :
						dwRes == WAIT_OBJECT_0 + keyQ ? L"Stop semaphore is released" :
						getWCStr( dwRes ) 
				)
			);
			
			if ( ( dwRes != WAIT_TIMEOUT ) || !processTimeout() )
				break;
		}

		if ( dwRes == WAIT_FAILED )
		{
			ODS_fls( L"Error in WaitForMultipleObjects." );
		}

		if ( dwRes == ( WAIT_OBJECT_0 + keyQ ) ) // csStop semaphore is released.
			break;

		if ( !processNotificationResult( dwRes ) )
			break;
	}

	return 0;
}


bool cWaitRegistryKeys::startWaiting()
{
	m_hThread = CreateThread( NULL, 0, waitFunction_static, (void *)this, 0, NULL );

	if ( m_hThread == NULL )
	{
		ODS_fls( L"Assertion!" );
		return false;
	}
	
	return true;
}

DWORD cWaitRegistryKeys::stopWaiting()
{
	ODSFunc;

	if ( m_hThread == NULL )
	{
		ODS_fls( L"Assertion!" );
		return -1;
	}

	cSemaphore cs;
	
	if ( cs.openSemaphore( m_szStopSemaphore ) )
		cs.releaseSemaphore();
	else
		ODS_fls( _FormatT( L"Assertion! Error '%s' in opening semaphore", cWCStr().GetSystemError( cs.m_err ) ) );

	DWORD dwRes = WaitForSingleObject_withODS( m_hThread, 1000, NULL );

	if ( dwRes == WAIT_TIMEOUT )
	{
		ODS_fls( L"Assertion! Wait stopped thread timeout" );
		dwRes = -2;
	}
	else
	if ( !GetExitCodeThread( m_hThread, &dwRes ) )
	{
		ODS_fls( _FormatT( L"Assertion! GetExitCodeThread stopped thread error '%s'", cWCStr().GetSystemError() ) );
		dwRes = -3;
	}

	CloseHandle( m_hThread );
	m_hThread = NULL;

	return dwRes;
}

//static
bool cWaitRegistryKeys::checkValid( cWaitRegistryKeys *obj )
{
	if ( obj == NULL )
		ODS_fls( L"Assertion! Object is NULL!" );
	else
	if ( !obj->isValid() )
		ODS_fls( L"Assertion! Object is invalid!" );
	else
		return true;

	return false;
}

//static
DWORD WINAPI cWaitRegistryKeys::waitFunction_static( void *ptr )
{
	cWaitRegistryKeys *pThis = (cWaitRegistryKeys *)ptr;

	if ( !cWaitRegistryKeys::checkValid( pThis ) )
		return (DWORD)-1;

	return pThis->waitFunction();
}
