#include "cWalkVolumes.h"
#include "UnicodeFuncs.h"

static bool isValidDriveName( const wchar_t *drvName, int *pDrvNum )
{
	if ( drvName[0] > L'Z' ) return false;
	if ( drvName[0] < L'A' ) return false;

	if ( wcscmp( drvName + 1, L":\\" ) != 0 )
		return false;

	if ( pDrvNum != NULL )
		*pDrvNum = drvName[0] - L'A';

	return true;
}

int cWalkVolumes::Walk()
{
	int count = 0;

	wchar_t volumeName[ MAX_PATH ];

	HANDLE hFind = FindFirstVolumeT( volumeName, maxLen( volumeName ) );
		
	int volumeNum = 0; // Is used to save checked mask only.

	//
	// Vista does not enumerate by FindFirstVolumeT network drives.
	// So the following trick is used.
	// It clears mask of processed drives and walks the mask after volumes processing
	// The left drives will processed as Windows 95 does.
	//
	DWORD dwDrives = GetLogicalDrives();

	cSilentMode sm;

	if ( hFind != INVALID_HANDLE_VALUE )
	{
		do
		{
			volumeNum++;

			wchar_t mountPoint[ MAX_PATH * 3 ];
			wchar_t rootDir[ MAX_PATH * 3];

			DWORD dwLength = 0;
			memset( mountPoint, 0, sizeof( mountPoint ) );

			if ( !GetVolumePathNamesForVolumeNameT( volumeName, rootDir, maxLen( rootDir ), &dwLength ) )
				continue;

			if ( wcslen( rootDir ) == 0 ) // The volume has no path names...
				continue;

			int drvNum = 0;

			if ( isValidDriveName( rootDir, &drvNum ) )
				dwDrives &= ~( 1 << drvNum );

			int iRes = Actor( volumeNum, rootDir );

			if ( iRes == 0 )
				continue;

			count += iRes;

			HANDLE hFindMP = INVALID_HANDLE_VALUE;

			DWORD dwSysFlags = 0;
			wchar_t fileSysNameBuf[ 32 ];
			if ( !GetVolumeInformationT(
						rootDir, NULL, 0, NULL, NULL,
						&dwSysFlags, fileSysNameBuf, 
						maxLen( fileSysNameBuf ) ),
						true )
				continue;

			// Detect support for reparse points, and therefore for volume 
			// mount points, which are implemented using reparse points.

			if ( dwSysFlags & FILE_SUPPORTS_REPARSE_POINTS ) 
			{
				hFindMP = FindFirstVolumeMountPointT( volumeName, mountPoint, maxLen( mountPoint ) );

				if ( hFindMP != INVALID_HANDLE_VALUE )
				{
					int len = wcslen( rootDir );
					do
					{
						wcscpy( rootDir + len, mountPoint );

						int iRes = Actor( volumeNum, rootDir );

						if ( iRes == 0 )
							continue;

						count += iRes;
					}
					while( FindNextVolumeMountPointT( hFindMP, mountPoint, maxLen( mountPoint ) ) );

					FindVolumeMountPointCloseT( hFindMP );
				}
			}
		}
		while( FindNextVolumeT( hFind, volumeName, maxLen( volumeName ) ) );

		FindVolumeCloseT( hFind );
	}

	if ( dwDrives != 0 ) // if ( count == 0 )
	{
		// Windows 95 and NT4 does not have the FindFirstVolume function, old filling works here
		// Windows 2K has the function but the method does not work properly because 
		//	the following GetVolumePathNamesForVolumeName does not present at kernel32.dll

//		DWORD dwDrives = GetLogicalDrives( );

		for( int driveNum = 0; driveNum < 26; driveNum++ )
		{
			if ( ( dwDrives & ( 1 << driveNum ) ) == 0 )
				continue;

			wchar_t driveName[] = L"A:\\";
			driveName[0] = L'A' + driveNum;

			int iRes = Actor( ++volumeNum, driveName );

			if ( iRes == 0 )
				continue;

			count += iRes;
		}
	}

	return count;
}
