#pragma once

#include <windows.h>
#include "cDll.h"
#include "cXML.h"

#ifndef _WININET_

//
// flags common to open functions (not InternetOpen()):
//

#define INTERNET_FLAG_RELOAD            0x80000000  // retrieve the original item

//
// flags for InternetOpenUrl():
//

#define INTERNET_FLAG_RAW_DATA          0x40000000  // FTP/gopher find: receive the item as raw (structured) data
#define INTERNET_FLAG_EXISTING_CONNECT  0x20000000  // FTP: use existing InternetConnect handle for server if possible

//
// flags for InternetOpen():
//

#define INTERNET_FLAG_ASYNC             0x10000000  // this request is asynchronous (where supported)

//
// protocol-specific flags:
//

#define INTERNET_FLAG_PASSIVE           0x08000000  // used for FTP connections

//
// additional cache flags
//

#define INTERNET_FLAG_NO_CACHE_WRITE    0x04000000  // don't write this item to the cache
#define INTERNET_FLAG_DONT_CACHE        INTERNET_FLAG_NO_CACHE_WRITE
#define INTERNET_FLAG_MAKE_PERSISTENT   0x02000000  // make this item persistent in cache
#define INTERNET_FLAG_FROM_CACHE        0x01000000  // use offline semantics
#define INTERNET_FLAG_OFFLINE           INTERNET_FLAG_FROM_CACHE

//
// additional flags
//

#define INTERNET_FLAG_SECURE            0x00800000  // use PCT/SSL if applicable (HTTP)
#define INTERNET_FLAG_KEEP_CONNECTION   0x00400000  // use keep-alive semantics
#define INTERNET_FLAG_NO_AUTO_REDIRECT  0x00200000  // don't handle redirections automatically
#define INTERNET_FLAG_READ_PREFETCH     0x00100000  // do background read prefetch
#define INTERNET_FLAG_NO_COOKIES        0x00080000  // no automatic cookie handling
#define INTERNET_FLAG_NO_AUTH           0x00040000  // no automatic authentication handling
#define INTERNET_FLAG_RESTRICTED_ZONE   0x00020000  // apply restricted zone policies for cookies, auth
#define INTERNET_FLAG_CACHE_IF_NET_FAIL 0x00010000  // return cache file if net request fails

//
// Security Ignore Flags, Allow HttpOpenRequest to overide
//  Secure Channel (SSL/PCT) failures of the following types.
//

#define INTERNET_FLAG_IGNORE_REDIRECT_TO_HTTP   0x00008000 // ex: https:// to http://
#define INTERNET_FLAG_IGNORE_REDIRECT_TO_HTTPS  0x00004000 // ex: http:// to https://
#define INTERNET_FLAG_IGNORE_CERT_DATE_INVALID  0x00002000 // expired X509 Cert.
#define INTERNET_FLAG_IGNORE_CERT_CN_INVALID    0x00001000 // bad common name in X509 Cert.

// setable flags
#define SECURITY_FLAG_IGNORE_REVOCATION         0x00000080
#define SECURITY_FLAG_IGNORE_UNKNOWN_CA         0x00000100
#define SECURITY_FLAG_IGNORE_WRONG_USAGE        0x00000200

#define SECURITY_FLAG_IGNORE_CERT_CN_INVALID    INTERNET_FLAG_IGNORE_CERT_CN_INVALID
#define SECURITY_FLAG_IGNORE_CERT_DATE_INVALID  INTERNET_FLAG_IGNORE_CERT_DATE_INVALID

#define SECURITY_FLAG_IGNORE_REDIRECT_TO_HTTPS  INTERNET_FLAG_IGNORE_REDIRECT_TO_HTTPS
#define SECURITY_FLAG_IGNORE_REDIRECT_TO_HTTP   INTERNET_FLAG_IGNORE_REDIRECT_TO_HTTP

//
// more caching flags
//

#define INTERNET_FLAG_RESYNCHRONIZE     0x00000800  // asking wininet to update an item if it is newer
#define INTERNET_FLAG_HYPERLINK         0x00000400  // asking wininet to do hyperlinking semantic which works right for scripts
#define INTERNET_FLAG_NO_UI             0x00000200  // no cookie popup
#define INTERNET_FLAG_PRAGMA_NOCACHE    0x00000100  // asking wininet to add "pragma: no-cache"
#define INTERNET_FLAG_CACHE_ASYNC       0x00000080  // ok to perform lazy cache-write
#define INTERNET_FLAG_FORMS_SUBMIT      0x00000040  // this is a forms submit
#define INTERNET_FLAG_FWD_BACK          0x00000020  // fwd-back button op
#define INTERNET_FLAG_NEED_FILE         0x00000010  // need a file for this request
#define INTERNET_FLAG_MUST_CACHE_REQUEST INTERNET_FLAG_NEED_FILE


#define INTERNET_OPEN_TYPE_PRECONFIG                    0   // use registry configuration
#define INTERNET_OPEN_TYPE_DIRECT                       1   // direct to net
#define INTERNET_OPEN_TYPE_PROXY                        3   // via named proxy
#define INTERNET_OPEN_TYPE_PRECONFIG_WITH_NO_AUTOPROXY  4   // prevent using java/script/INS

#define INTERNET_INVALID_PORT_NUMBER    0           // use the protocol-specific default

#define INTERNET_DEFAULT_FTP_PORT       21          // default for FTP servers
#define INTERNET_DEFAULT_GOPHER_PORT    70          //    "     "  gopher "
#define INTERNET_DEFAULT_HTTP_PORT      80          //    "     "  HTTP   "
#define INTERNET_DEFAULT_HTTPS_PORT     443         //    "     "  HTTPS  "
#define INTERNET_DEFAULT_SOCKS_PORT     1080        // default for SOCKS firewall servers.

#define INTERNET_SERVICE_FTP    1
#define INTERNET_SERVICE_GOPHER 2
#define INTERNET_SERVICE_HTTP   3

#define HTTP_QUERY_MIME_VERSION                 0
#define HTTP_QUERY_CONTENT_TYPE                 1
#define HTTP_QUERY_CONTENT_TRANSFER_ENCODING    2
#define HTTP_QUERY_CONTENT_ID                   3
#define HTTP_QUERY_CONTENT_DESCRIPTION          4
#define HTTP_QUERY_CONTENT_LENGTH               5
#define HTTP_QUERY_CONTENT_LANGUAGE             6
#define HTTP_QUERY_ALLOW                        7
#define HTTP_QUERY_PUBLIC                       8
#define HTTP_QUERY_DATE                         9
#define HTTP_QUERY_EXPIRES                      10
#define HTTP_QUERY_LAST_MODIFIED                11
#define HTTP_QUERY_MESSAGE_ID                   12
#define HTTP_QUERY_URI                          13
#define HTTP_QUERY_DERIVED_FROM                 14
#define HTTP_QUERY_COST                         15
#define HTTP_QUERY_LINK                         16
#define HTTP_QUERY_PRAGMA                       17
#define HTTP_QUERY_VERSION                      18  // special: part of status line
#define HTTP_QUERY_STATUS_CODE                  19  // special: part of status line
#define HTTP_QUERY_STATUS_TEXT                  20  // special: part of status line
#define HTTP_QUERY_RAW_HEADERS                  21  // special: all headers as ASCIIZ
#define HTTP_QUERY_RAW_HEADERS_CRLF             22  // special: all headers
#define HTTP_QUERY_CONNECTION                   23
#define HTTP_QUERY_ACCEPT                       24
#define HTTP_QUERY_ACCEPT_CHARSET               25
#define HTTP_QUERY_ACCEPT_ENCODING              26
#define HTTP_QUERY_ACCEPT_LANGUAGE              27
#define HTTP_QUERY_AUTHORIZATION                28
#define HTTP_QUERY_CONTENT_ENCODING             29
#define HTTP_QUERY_FORWARDED                    30
#define HTTP_QUERY_FROM                         31
#define HTTP_QUERY_IF_MODIFIED_SINCE            32
#define HTTP_QUERY_LOCATION                     33
#define HTTP_QUERY_ORIG_URI                     34
#define HTTP_QUERY_REFERER                      35
#define HTTP_QUERY_RETRY_AFTER                  36
#define HTTP_QUERY_SERVER                       37
#define HTTP_QUERY_TITLE                        38
#define HTTP_QUERY_USER_AGENT                   39
#define HTTP_QUERY_WWW_AUTHENTICATE             40
#define HTTP_QUERY_PROXY_AUTHENTICATE           41
#define HTTP_QUERY_ACCEPT_RANGES                42
#define HTTP_QUERY_SET_COOKIE                   43
#define HTTP_QUERY_COOKIE                       44
#define HTTP_QUERY_REQUEST_METHOD               45  // special: GET/POST etc.
#define HTTP_QUERY_REFRESH                      46
#define HTTP_QUERY_CONTENT_DISPOSITION          47

#define INTERNET_FLAG_NEED_FILE         0x00000010  // need a file for this request

#define HTTP_QUERY_FLAG_NUMBER					0x20000000

#define ERROR_NO 0
#define ERROR_INTERNET_INVALID_URL				12005
#define ERROR_INTERNET_SEC_CERT_REV_FAILED		12057	// If the system time is set back to current time. HttpSendRequest can be called successfully.
														// Or if SECURITY_FLAG_IGNORE_REVOCATION is added to INTERNET_OPTION_SECURITY_FLAGS, this error can be ignored.	

#ifndef HTTP_CONNECT_ERROR
#define HTTP_CONNECT_ERROR -8
#endif

#endif

class cDownloadActorBlock
{
public:
	cDownloadActorBlock() {};

	//
	//	downloadFile uses the actor, if it returns false the download process is stopped.
	//
	virtual bool Actor( DWORD dwSize )
	{
		assert( false );
		return true;
	};
};

class cWinINet : public CDll
{
protected:

	HANDLE m_Initialize;
	HANDLE m_Connection;
	HANDLE m_File;

public:

	DWORD m_httpError;

	typedef HANDLE WINAPI typeInternetOpenW(
    __in_opt LPCWSTR lpszAgent,
    __in DWORD dwAccessType,
    __in_opt LPCWSTR lpszProxy,
    __in_opt LPCWSTR lpszProxyBypass,
    __in DWORD dwFlags
    );

	typeInternetOpenW *m_InternetOpenW;

	bool iOpen( const wchar_t *szType )
	{
		if ( error || ( m_InternetOpenW == NULL ) )
		{
			ODS_fls( L"Assertion!" );
			return false;
		}

		iCloseHandle( &m_Initialize );

//		m_Initialize = m_InternetOpenW( szType, INTERNET_OPEN_TYPE_DIRECT, NULL, NULL, 0 );
//		m_Initialize = m_InternetOpenW( szType, INTERNET_OPEN_TYPE_PROXY, L"proxy", L"<local>", 0 );
		m_Initialize = m_InternetOpenW( szType, INTERNET_OPEN_TYPE_PRECONFIG, NULL, NULL, 0 );

		error = ( m_Initialize == NULL );

		if ( error )
			ODS_fls( L"Assertion!" );

		return error == FALSE;
	}

	typedef HANDLE WINAPI typeInternetConnectW(
		__in HANDLE HANDLE,
		__in LPCWSTR lpszServerName,
		__in WORD nServerPort,
		__in_opt LPCWSTR lpszUserName,
		__in_opt LPCWSTR lpszPassword,
		__in DWORD dwService,
		__in DWORD dwFlags,
		__in_opt DWORD_PTR dwContext
		);

	typeInternetConnectW *m_InternetConnectW;

	WORD m_defPort;
	DWORD m_service;

	bool iConnect( const wchar_t *serverName )
	{
		if ( m_Connection != NULL )
			iCloseHandle( &m_Connection );

		if ( error )
		{ ODS_fls( L"Assertion!" ); return false; }

		if ( m_InternetConnectW == NULL )
		{ ODS_fls( L"Assertion!" ); return false; }

		if ( m_Initialize == NULL )
		{ ODS_fls( L"Assertion!" ); return false; }

		m_Connection = m_InternetConnectW( m_Initialize, serverName, m_defPort, NULL, NULL, m_service, 0, 0);

		error = ( m_Connection == NULL );
		if ( error )
		{ ODS_fls( L"Assertion!" ); return false; }

		return error == FALSE;
	}

	typedef HANDLE WINAPI typeHttpOpenRequestW(
    __in HANDLE hConnect,
    __in_opt LPCWSTR lpszVerb,
    __in_opt LPCWSTR lpszObjectName,
    __in_opt LPCWSTR lpszVersion,
    __in_opt LPCWSTR lpszReferrer,
    __in_opt LPCWSTR FAR * lplpszAcceptTypes,
    __in DWORD dwFlags,
    __in_opt DWORD_PTR dwContext
    );

	typeHttpOpenRequestW *m_HttpOpenRequestW;

	typedef BOOL WINAPI typeHttpSendRequestW(
    __in HANDLE hRequest,
    __in_ecount_opt(dwHeadersLength) LPCWSTR lpszHeaders,
    __in DWORD dwHeadersLength,
    __in_bcount_opt(dwOptionalLength) LPVOID lpOptional,
    __in DWORD dwOptionalLength
    );
	
	typeHttpSendRequestW *m_HttpSendRequestW;

	bool openRequest( const wchar_t *szObjectName, bool useCache = false, bool bSecure = false, bool bIgnoreCrtError = false )
	{
		m_httpError = HTTP_CONNECT_ERROR;

		if ( m_File != NULL )
			iCloseHandle( &m_File );

		if ( error )
		{ ODS_fls( L"Assertion!" ); return false; }

		if ( m_HttpOpenRequestW == NULL )
		{ ODS_fls( L"Assertion!" ); return false; }

		if ( m_Connection == NULL )
		{ ODS_fls( L"Assertion!" ); return false; }

		DWORD dwFlags = 
				INTERNET_FLAG_KEEP_CONNECTION |
				INTERNET_FLAG_NO_UI |
				0;

		if ( !useCache )
			dwFlags |=
				INTERNET_FLAG_NO_CACHE_WRITE |
				INTERNET_FLAG_RELOAD;

		if ( bIgnoreCrtError )
			dwFlags |=
				SECURITY_FLAG_IGNORE_REVOCATION |
				INTERNET_FLAG_IGNORE_CERT_CN_INVALID |
				INTERNET_FLAG_IGNORE_CERT_DATE_INVALID;

		if ( bSecure || (m_defPort == INTERNET_DEFAULT_HTTPS_PORT) )
			dwFlags |= INTERNET_FLAG_SECURE;


		m_File = m_HttpOpenRequestW( m_Connection, NULL, szObjectName, NULL, NULL, NULL, dwFlags, NULL );

		error = ( m_File == NULL );
		if ( error )
		{ ODS_fls( L"Assertion!" ); return false; }

		int err;

		if( !m_HttpSendRequestW( m_File, NULL, 0, NULL, 0 ) )
		{
			// ERROR_INTERNET_INVALID_URL 12005

			err = GetLastError();
			m_httpError = err;

			if ( err == ERROR_INTERNET_SEC_CERT_REV_FAILED )
				ODS_fls( _FormatT( L"Assertion! Error %s", L"ERROR_INTERNET_SEC_CERT_REV_FAILED" ) );
			else
			if ( err == ERROR_INTERNET_INVALID_URL )
				ODS_fls( _FormatT( L"Assertion! Error %s", L"ERROR_INTERNET_INVALID_URL" ) );
			else
				ODS_fls( _FormatT( L"Assertion! Error %s", _GetSystemError( err ) ) );

			iCloseHandle( &m_File );
			error = true;
			return false;
		}

		return true;
	}
	
	typedef BOOL WINAPI typeInternetCloseHandle( HANDLE HANDLE );
	(typeInternetCloseHandle *)m_InternetCloseHandle;

	bool iCloseHandle( HANDLE *pHandle )
	{
		if ( m_InternetCloseHandle == NULL )
			return false;

		if ( pHandle == NULL )
			return false;

		if ( !m_InternetCloseHandle( *pHandle ) )
			return false;

		*pHandle = NULL;

		return true;
	}

	typedef BOOL WINAPI typeInternetReadFile(
    __in HANDLE hFile,
    __out_bcount(dwNumberOfBytesToRead) LPVOID lpBuffer,
    __in DWORD dwNumberOfBytesToRead,
    __out LPDWORD lpdwNumberOfBytesRead
    );
	
	typeInternetReadFile *m_InternetReadFile;

	bool iReadFile( BYTE *bf, DWORD bfSize, DWORD *pDWBytes )
	{
		if ( m_InternetReadFile == NULL )
		{
			error = true;
			return false;
		}

		if ( m_File == NULL )
		{
			error = true;
			return false;
		}

		DWORD dwRead = 0;

		if ( pDWBytes == NULL )
			pDWBytes = &dwRead;
		
		return m_InternetReadFile( m_File, bf, bfSize, pDWBytes ) == TRUE && ( *pDWBytes != 0 );
	}
	

	typedef DWORD WINAPI typeInternetSetFilePointer(
		HANDLE hFile,
		LONG lDistanceToMove,
		PVOID pReserved,
		DWORD dwMoveMethod,
		DWORD_PTR dwContext
	);

	DWORD iSetFilePointer(
		HANDLE hFile,
		LONG lDistanceToMove = 0,
		PVOID pReserved = NULL,
		DWORD dwMoveMethod = FILE_BEGIN,
		DWORD_PTR dwContext = NULL
	)
	{
		typeInternetSetFilePointer *procSetFilePointer;
		
		procSetFilePointer = (typeInternetSetFilePointer *)::GetProcAddress( hInst, "InternetSetFilePointer" );

		if ( procSetFilePointer == NULL )
			return -1;

		return procSetFilePointer( hFile, lDistanceToMove, pReserved, dwMoveMethod, dwContext );
	}


	typedef BOOL WINAPI typeHttpQueryInfoW(
    __in HANDLE hRequest,
    __in DWORD dwInfoLevel,
    __inout_bcount_opt(*lpdwBufferLength) LPVOID lpBuffer,
    __inout LPDWORD lpdwBufferLength,
    __inout_opt LPDWORD lpdwIndex
    );

	typeHttpQueryInfoW *m_HttpQueryInfoW;

	typedef BOOL WINAPI typeInternetQueryDataAvailable(
    __in HANDLE hFile,
    __out_opt LPDWORD lpdwNumberOfBytesAvailable,
    __in DWORD dwFlags,
    __in_opt DWORD_PTR dwContext
    );
	
	typeInternetQueryDataAvailable *m_InternetQueryDataAvailable;

	bool getFileSize( DWORD *pSize )
	{
		//*pSize = iSetFilePointer( m_File, 0, 0, FILE_END );
		//iSetFilePointer( m_File );

		//if ( *pSize != -1 )
		//	return true;

		DWORD bfSize = sizeof( *pSize );
		
		bfSize = sizeof( *pSize );

		if ( !m_HttpQueryInfoW( m_File, HTTP_QUERY_FLAG_NUMBER | HTTP_QUERY_CONTENT_LENGTH, (void *)pSize, &bfSize, 0 ) )
			return false;

		return true;


		//if ( m_InternetQueryDataAvailable == NULL )
		//	return false;

		//if ( m_File == NULL )
		//	return false;

		//bool ok = m_InternetQueryDataAvailable( m_File, pSize, 0, 0 ) == TRUE;

		//int err = GetLastError();

		//return ok;

		//DWORD dwErr;
		//DWORD bfSize = sizeof( *pSize );


		//if ( !m_HttpQueryInfoW( m_File, HTTP_QUERY_CONTENT_LENGTH, (void *)pSize, &bfSize, &dwErr ) )
		//	return false;

		//return dwErr == ERROR_NO; // ERROR_INSUFFICIENT_BUFFER
	}

	DWORD getStatusCode()
	{
		DWORD status = -1;
		DWORD bfSize = sizeof( status );
		DWORD dwErr;

		if ( !m_HttpQueryInfoW( m_File, HTTP_QUERY_STATUS_CODE , (void *)&status, &bfSize, &dwErr ) )
		{
			int err = GetLastError();
			return -1;
		}
		
		if ( dwErr == ERROR_NO )
			return status;

		return -1;
	}

	typedef BOOL WINAPI typeInternetFindNextFileW( __in HANDLE hFind, __out LPVOID lpvFindData );
	typeInternetFindNextFileW *m_InternetFindNextFileW;

	bool iFindNextFile( HANDLE hFind, WIN32_FIND_DATAW *fd )
	{
		if ( m_InternetFindNextFileW != NULL )
			return false;

		return m_InternetFindNextFileW( hFind, fd ) == TRUE;
	}

	typedef HANDLE WINAPI typeFtpFindFirstFileW(
    __in HANDLE hConnect,
    __in_opt LPCWSTR lpszSearchFile,
    __out_opt LPWIN32_FIND_DATAW lpFindFileData,
    __in DWORD dwFlags,
    __in_opt DWORD_PTR dwContext
    );

	typeFtpFindFirstFileW *m_FtpFindFirstFileW;

	HANDLE iFindFirstFile( LPCWSTR lpszSearchFile, WIN32_FIND_DATAW *lpFindFileData, DWORD dwFlags, DWORD_PTR dwContext )
	{
		if ( ( m_FtpFindFirstFileW == NULL ) || ( m_Connection || NULL ) )
			return NULL;

		return m_FtpFindFirstFileW( m_Connection, lpszSearchFile, lpFindFileData, dwFlags, dwContext );
	}

public:

	cWinINet( const wchar_t *path = NULL )
		: CDll( false, "Wininet.dll" )
	{
		ODSFunc;

		cWinINet_Init();

		if ( path != NULL )
			openPath( path );

		ODSFuncRes( error );
	}

	bool checkPath( const wchar_t *path )
	{
		const wchar_t *pos = wcsstr( path, L"://" );
		if ( pos == NULL ) return false;

		const wchar_t *szType = parseProtocol( path );
		if ( szType == NULL ) return false;

		return true;
	}

	bool openPath( const wchar_t *path )
	{
		// https://drive.google.com/file/d/0ByGQUYXFre8oTVJ1NmtscV9XODA/view?usp=sharing

		const wchar_t *pos = wcsstr( path, L"://" );
		if ( pos == NULL )
		{ ODS_fls( L"Assertion!" ); return false; }

		const wchar_t *szType = parseProtocol( path );
		if ( szType == NULL )
		{ ODS_fls( L"Assertion!" ); return false; }

		wchar_t *szSite = _wcsdup( pos + 3 );
		if ( szSite == NULL )
		{ ODS_fls( L"Assertion!" ); return false; }

		wchar_t *szObject = wcschr( szSite, L'/' );
		bool ok = ( szObject != NULL );

		if ( ok )
			*(szObject++) = L'\0';
		else
			ODS_fls( L"Assertion!" );

		ok &= iOpen( szType );
		ok &= iConnect( szSite );
		ok &= openRequest( szObject, true );

		free( szSite );

		if ( !ok )
		{ ODS_fls( L"Assertion!" ); return false; }

		return true;
	}
	
	//
	//	Returns connection type
	//
	const wchar_t *parseProtocol( const wchar_t *szProtocol )
	{
		if ( _wcsnicmp( szProtocol, L"https", 5 ) == 0 )
		{
			m_defPort = INTERNET_DEFAULT_HTTPS_PORT;
			m_service = INTERNET_SERVICE_HTTP;
			return L"HTTPGET";
		}
		else
		if ( _wcsnicmp( szProtocol, L"ftp", 3 ) == 0 )
		{
			m_defPort = INTERNET_DEFAULT_FTP_PORT;
			m_service = INTERNET_SERVICE_FTP;
			return L"FTPGET";
		}
		else
		if ( _wcsnicmp( szProtocol, L"gopher", 6 ) == 0 )
		{
			m_defPort = INTERNET_DEFAULT_GOPHER_PORT;
			m_service = INTERNET_SERVICE_GOPHER;
			return L"GOPHERGET";
		}
		else
		if ( _wcsnicmp( szProtocol, L"http", 4 ) == 0 )
		{
			m_defPort = INTERNET_DEFAULT_HTTP_PORT;
			m_service = INTERNET_SERVICE_HTTP;
			return L"HTTPGET";
		}

		return NULL;
	}

	cWinINet(
		const wchar_t *szType, // L"HTTPGET"
		const wchar_t *szSite, // L"www.jetico.com"
		const wchar_t *szObject, // L"/index.htm" 
		const wchar_t *szProtocol // = L"http:"
		) : CDll( false, "Wininet.dll" )
	{
		cWinINet_Init();

		if ( parseProtocol( szProtocol ) == NULL )
			return;

		if ( szType != NULL )
			iOpen( szType );

		if ( szSite != NULL )
			iConnect( szSite );

		if ( szObject != NULL )
			openRequest( szObject, true );
	}
	
	void cWinINet_Init()
	{
		ODSFunc;

		m_defPort = -1;
		m_service = -1;
		m_Initialize = NULL;
		m_Connection = NULL;
		m_File = NULL;

		GetProcAddress( "InternetCloseHandle", false, (void**)&m_InternetCloseHandle ); 
		GetProcAddress( "InternetOpenW", false, (void**)&m_InternetOpenW );
		GetProcAddress( "InternetConnectW", false, (void**)&m_InternetConnectW );
		GetProcAddress( "HttpOpenRequestW", false, (void**)&m_HttpOpenRequestW );
		GetProcAddress( "HttpSendRequestW", false, (void**)&m_HttpSendRequestW );
		GetProcAddress( "InternetReadFile", false, (void**)&m_InternetReadFile );
		GetProcAddress( "HttpQueryInfoW", false, (void**)&m_HttpQueryInfoW );
		GetProcAddress( "InternetQueryDataAvailable", false, (void**)&m_InternetQueryDataAvailable );
		GetProcAddress( "InternetFindNextFileW", false, (void**)&m_InternetFindNextFileW );
		GetProcAddress( "FtpFindFirstFileW", false, (void**)&m_FtpFindFirstFileW );

		assert( !error );

		m_httpError = HTTP_CONNECT_ERROR;

		ODSFuncRes( error );
	}

	cWinINet(
		WORD port,
		DWORD service,
		const wchar_t *szType = NULL, // L"HTTPGET"
		const wchar_t *szSite = NULL, // L"www.jetico.com"
		const wchar_t *szObject = NULL // L"/index.htm" 
		) : CDll( false, "Wininet.dll" )
	{
		ODSFunc;

		cWinINet_Init();

		m_defPort = port;
		m_service = service;

		if ( szType != NULL )
			iOpen( szType );

		if ( szSite != NULL )
			iConnect( szSite );

		if ( szObject != NULL )
			openRequest( szObject, true );
	}

	~cWinINet()
	{
		close();
	}

	void close()
	{
		iCloseHandle( &m_File );
		iCloseHandle( &m_Connection );
		iCloseHandle( &m_Initialize );
	}

	enum { dnlBlockSize = 64 * 1024 };

	bool downloadFile( HANDLE hFile, cDownloadActorBlock *actorBlock = NULL )
	{

		BYTE *bf = new BYTE[ dnlBlockSize ];

		if ( bf == 0 )
		{ ODS_fls( L"Assertion!" ); return false; }

		DWORD dwSize = 0;

		bool ok = false;

		while( iReadFile( bf, dnlBlockSize, &dwSize ) )
		{
			if ( actorBlock != NULL )
			{
				if ( !actorBlock->Actor( dwSize ) )
					break;
			}

			if ( !ok ) // Definitelly first call
			{
				//
				// if a error is reportd, it look as HTML doccumment.
				// We just try to read the documment title.
				//
				//<!DOCTYPE HTML PUBLIC "-//IETF//DTD HTML 2.0//EN">
				//<html><head>
				//<title>401 Authorization Required</title>
				//</head><body>
				//<h1>Authorization Required</h1>

				cXMLBlock xml;

				if ( xml.parse( bf, dwSize ) )
				{
					if ( ( xml.bodyL > 7 ) &&
						 ( memcmp( xml.bodyT, "DOCTYPE", 7 ) == 0 )
						)
					{
						cXMLBlock html, head, title;
						
						if ( html.parse( xml.nextT, xml.nextL ) &&
							  head.parse( html.bodyT, html.bodyL ) &&
							   title.parse( head.bodyT, head.bodyL )
							)
						{
							char *errMessage = ( char * )title.bodyT;

							if ( m_showMessage )
							{
								errMessage[ title.bodyL ] = '\0';
								MessageBoxA( 0,errMessage, "Download error", MB_ICONSTOP );
							}

							int error = atoi( errMessage );
						}

						// We can read xml file with error.
						break;
					}
				}
			}

			DWORD dwWritten = 0;
			ok = WriteFile( hFile, bf, dwSize, &dwWritten, NULL ) == TRUE;

			if ( !ok || ( dwWritten != dwSize ) )
			{
				ODS_fls( _FormatT( L"Assertion! Error %s in writing file", _GetSystemError(-1) ) );
				ok = false;
				break;
			}
		}

//		DWORD status = getStatusCode();

		delete bf;

		return ok;
	}
};

