#include "clipboardFuncs.h"

static UINT findCompatibleFormat()
{
	UINT cbf = 0;

	while ( cbf = EnumClipboardFormats( cbf ) )
	{
		switch ( cbf )
		{
//			case CF_LOCALE:
			case CF_OEMTEXT:
			case CF_TEXT:
			case CF_UNICODETEXT:
//			case CF_DSPTEXT:
				return cbf;
			default:
				break;
		}
	}

	return 0;
}

bool CopyToClipboard( HWND hWnd, const wchar_t *wc, int wcLen )
{
	BOOL res = OpenClipboard( hWnd );

	if ( !res )
		return FALSE;

	if ( !EmptyClipboard() )
	{
		CloseClipboard(); 
		return FALSE; 
	}

	if ( wcLen == -1 )
		wcLen = (int)wcslen( wc );

	int size = ( wcLen + 1 ) * 2;


	HGLOBAL hglbCopy = GlobalAlloc( GMEM_MOVEABLE,  size  ); 
	if (hglbCopy == NULL) 
	{ 
		CloseClipboard(); 
		return FALSE; 
	} 

	// Lock the handle and copy the text to the buffer. 

	wchar_t *lptstrCopy = (wchar_t *)GlobalLock(hglbCopy);

	bool bOk = ( lptstrCopy != NULL );

	if ( bOk ) // 24.10.2018 From BDK 9 Audit by Rechtspraak.
	{
		UINT uFormat = CF_UNICODETEXT;

		memcpy( lptstrCopy, wc, size );

		GlobalUnlock(hglbCopy); 

		// Place the handle on the clipboard. 

		SetClipboardData( uFormat, hglbCopy ); 
	}
            
	GlobalFree( hglbCopy );

	CloseClipboard();

	return bOk;
}

bool ReadFromClipboard( HWND hWnd, cWCStr *wcsCB, bool silent )
{
	// wchar_t errMsg[ 256 ];

	BOOL res = OpenClipboard( hWnd );
	
	if ( !res )
		return false;
	
	UINT cbf = CF_UNICODETEXT;
	cWCStr tmpWCS;
	
	if ( !IsClipboardFormatAvailable( cbf ) )
	{
		cbf = findCompatibleFormat();
		
		if ( cbf == 0 )
		{	
			CloseClipboard();

			if ( !silent )
			{
				//LoadStringT( IH.m_hInst, IDS_ClipBoardIsEmpty, errMsg, maxLen( errMsg ) );
				//MessageBoxT( hWnd, errMsg, IH.LoadStringOnceT( IDS_Warning ), MB_ICONINFORMATION );
				MessageBoxT( hWnd, L"Clipboard is empty", L"Warning", MB_ICONINFORMATION );
			}
			
			return true;
		}
	}
	
	HANDLE hcbData = GetClipboardData( cbf );
	
	res = FALSE;
	
	if ( hcbData != NULL )
	{
		wchar_t *lptstr = (wchar_t *)GlobalLock( hcbData );

		if ( lptstr != NULL )
		{
			if ( wcslen( lptstr ) == 0 )
			{
				if ( !silent )
				{
					//LoadStringT( IH.m_hInst, IDS_ClipBoardIsEmpty, errMsg, maxLen( errMsg ) );
					//MessageBoxT( hWnd, errMsg, IH.LoadStringOnceT( IDS_Warning ), MB_ICONINFORMATION );
					MessageBoxT( hWnd, L"Clipboard is empty", L"Warning", MB_ICONINFORMATION );
				}
			}
			else
			{
				if ( cbf != CF_UNICODETEXT )
				{
					wcsCB->Add( (char*)lptstr );
				}
				else
					wcsCB->Add( lptstr );

				res = ( wcsCB->m_len != 0 );
			}

			GlobalUnlock( hcbData );
		}
	}
	
	CloseClipboard();
	
	return res == TRUE;
}

void ClearClipboard( HWND hWnd )
{
	BOOL res = OpenClipboard( hWnd );

	EmptyClipboard();

	CloseClipboard();
}

bool IsClipboardEmpty( HWND hWnd )
{
	BOOL res = OpenClipboard( hWnd );
	
	if ( !res )
		return false;

	bool bEmpty = true;

	UINT cbf = 0;

	int err = -1;

	while ( cbf = EnumClipboardFormats( cbf ) )
	{
		HANDLE hcbData = GetClipboardData( cbf );

		if ( hcbData == NULL )
		{
			err = GetLastError();
			continue;
		}
				
		if ( hcbData != NULL )
		{
			void *lpv = (void *)GlobalLock( hcbData );

			if ( lpv != NULL )
			{
				GlobalUnlock( hcbData );
				bEmpty = false;
				break;
			}

			err = GetLastError();
		}		
	}

	CloseClipboard();
		
	return bEmpty;
}

