#pragma once

#include "cSysCriticalSection.h"
#include "ODS.h"
#include <assert.h>

class cKernelDll
{
	HINSTANCE hInst;

public:

	HMODULE getInstance()
	{
		return hInst;
	}

	// Do not use the in global constructor, and avoid global contruction at all.
	static __declspec(noinline) cKernelDll &getStatic()
	{
		ODSFunc;

		static cSysCriticalSection cs;
		SYS_CRITICAL_SECTION_BLOCK( &cs );

		static cKernelDll dll;
#ifdef _ODS
		if ( dll.hInst == NULL  ) ODS_f( L"Assertion" );
#endif
		assert( dll.hInst != NULL );

		ODSFuncRes( _FormatT( L"%p", &cs ) );

		return dll;
	}

	cKernelDll()
	{
		hInst = LoadLibraryA( "kernel32.dll" );
	}

	~cKernelDll()
	{
		FreeLibrary( hInst );
	}

	bool IsWow64Process()
	{
		static BOOL isWow64Process = -1;
		
		if ( isWow64Process == -1 )
		{
			isWow64Process = 0;

			if ( !IsWow64Process( GetCurrentProcess(), &isWow64Process ) )
				return false;
		}

		return isWow64Process > 0;
	}

	BOOL IsWow64Process( HANDLE hProcess, BOOL *pIsWow64Process )
	{
		typedef BOOL WINAPI typeIsWow64Process( HANDLE hProcess, BOOL *isWow64Process );

		if ( hInst == NULL )
			return false;

		static typeIsWow64Process *proc = (typeIsWow64Process *)GetProcAddress( hInst, "IsWow64Process" );

		if ( proc == NULL )
		{
			*pIsWow64Process = false;
			return true;
		}

		return proc( hProcess, pIsWow64Process );
	}

	void GetNativeSystemInfo( LPSYSTEM_INFO lpSystemInfo )
	{
		typedef void WINAPI typeGetNativeSystemInfo( LPSYSTEM_INFO lpSystemInfo );

		memset( lpSystemInfo, 0, sizeof( SYSTEM_INFO ) );

		typeGetNativeSystemInfo *proc = (typeGetNativeSystemInfo *)GetProcAddress( hInst, "GetNativeSystemInfo" );

		if ( proc == NULL )
			GetSystemInfo( lpSystemInfo );
		else
			proc( lpSystemInfo );
	}
/*
	BOOL Wow64EnableWow64FsRedirection( BOOL enable )
	{
		typedef BOOL WINAPI typeWow64EnableWow64FsRedirection( BOOL enable );

		typeWow64EnableWow64FsRedirection *proc = (typeWow64EnableWow64FsRedirection *)GetProcAddress( hInst, "Wow64EnableWow64FsRedirection" );

		if ( proc == NULL )
			return FALSE;

		return proc( enable );
	}
*/
	BOOL Wow64DisableWow64FsRedirection( PVOID* pOldValue )
	{
		typedef BOOL WINAPI typeWow64DisableWow64FsRedirection( PVOID* OldValue );

		typeWow64DisableWow64FsRedirection *proc = (typeWow64DisableWow64FsRedirection *)GetProcAddress( hInst, "Wow64DisableWow64FsRedirection" );

		if ( proc == NULL )
			return FALSE;

		return proc( pOldValue );
	}

	BOOL Wow64RevertWow64FsRedirection( PVOID pOldValue )
	{
		typedef BOOL WINAPI typeWow64RevertWow64FsRedirection( PVOID OldValue );

		typeWow64RevertWow64FsRedirection *proc = (typeWow64RevertWow64FsRedirection *)GetProcAddress( hInst, "Wow64RevertWow64FsRedirection" );

		if ( proc == NULL )
			return FALSE;

		return proc( pOldValue );
	}

	UINT GetSystemWow64Directory( LPTSTR lpBuffer, UINT uSize )
	{
		typedef UINT typeGetSystemWow64Directory( LPTSTR lpBuffer, UINT uSize );

		typeGetSystemWow64Directory *proc = (typeGetSystemWow64Directory *)GetProcAddress( hInst, "GetSystemWow64DirectoryA" );

		if ( proc == NULL )
			return 0;

		return proc( lpBuffer, uSize );
	}

	//
	//	Enables visual styles without manifest but for curent thread.
	//
	ULONG_PTR EnableVisualStyles()
	{
#if (_WIN32_WINNT >= 0x0500) || (_WIN32_FUSION >= 0x0100) || ISOLATION_AWARE_ENABLED
	
		//
		// Google:: How to enable visual styles without a manifest?
		//
		//#pragma comment(linker,"\"/manifestdependency:type='win32' name='Microsoft.Windows.Common-Controls' version='6.0.0.0' processorArchitecture='*' publicKeyToken='6595b64144ccf1df' language='*'\"")

		typedef WINBASEAPI BOOL WINAPI typeActivateActCtx( __inout_opt HANDLE hActCtx, __out   ULONG_PTR *lpCookie );

		typeActivateActCtx *proc = (typeActivateActCtx *)GetProcAddress( hInst, "ActivateActCtx" );

		if ( proc == NULL )
			return 0;

		typedef HANDLE WINAPI typeCreateActCtxW( __in PCACTCTXW pActCtx );
		
		typeCreateActCtxW *procCreateActCtxW = (typeCreateActCtxW *)GetProcAddress( hInst, "CreateActCtxW" );

		if ( procCreateActCtxW == NULL )
			return 0;

		wchar_t dir[MAX_PATH];
		ULONG_PTR ulpActivationCookie = FALSE;
		ACTCTXW actCtx =
		{
			sizeof(actCtx),
			ACTCTX_FLAG_RESOURCE_NAME_VALID | ACTCTX_FLAG_SET_PROCESS_DEFAULT | ACTCTX_FLAG_ASSEMBLY_DIRECTORY_VALID,
			L"shell32.dll", 0, 0, dir, (const wchar_t *)124
		};

		UINT cch = GetSystemDirectoryW(dir, sizeof(dir) / sizeof(*dir));

		if (cch < sizeof(dir) / sizeof(*dir))
		{
			dir[cch] = TEXT('\0');
			proc( procCreateActCtxW(&actCtx), &ulpActivationCookie);
		}

		return ulpActivationCookie;
#else
		return 0;
#endif // (_WIN32_WINNT >= 0x0500) || (_WIN32_FUSION >= 0x0100) || ISOLATION_AWARE_ENABLED
	}

};
/*
class cDisableWow64Redirection
{
	bool m_disable;

public:
	cDisableWow64Redirection( bool disable )
	{
		m_disable = disable;

		if ( m_disable )
			cKernelDll::getStatic().Wow64EnableWow64FsRedirection( false );
	}

	~cDisableWow64Redirection()
	{
		if ( m_disable )
			cKernelDll::getStatic().Wow64EnableWow64FsRedirection( true );
	}
};
*/

// use cKernelDll::getStatic() instead
//extern cKernelDll kernelDll;

//
//	MSDN reports at Wow64EnableWow64FsRedirection docummentation:
//		This function may not work reliably when there are nested calls. Therefore, 
//		this function has been replaced by the Wow64DisableWow64FsRedirection and 
//		Wow64RevertWow64FsRedirection functions.
//
//	It is required to temporary stop disabling of redirection.
//
//	cDisableWow64Redirection.stopDisabling();
//	HANDLE hFile = CreateFile( ... );
//	cDisableWow64Redirection.continueDisabling();
//
//	Note 1. continueDisabling restore disabling if the previous call of startDisable sets the bRet value.
//	Note 2. manifest themes does not work if redirection is disabled. So we have to stop disabling before 
//		confirmation dialog start and continueDisabling when dialog is closed. Assertion message loads the
//		common control dlls and the messahe and all following dialogue look as manisfest file was not loaded.
//
class cDisableWow64Redirection
{
public:

	PVOID OldValue;
	BOOL bRet;

	cDisableWow64Redirection()
	{
		OldValue = NULL;
		bRet = FALSE;
	}

	BOOL startDisable( BOOL bDisable )
	{
		ODSFuncS( _FormatT( L"%x, %x", bRet, OldValue ) );

		if ( !bDisable )
			bRet = FALSE;
		else
			bRet = cKernelDll::getStatic().Wow64DisableWow64FsRedirection (&OldValue);

		ODSFuncRes( _FormatT( L"%x, %x", bRet, OldValue ) );

		return bRet;
	}

	cDisableWow64Redirection( bool bDisable )
	{
		startDisable( bDisable );
	}

	~cDisableWow64Redirection()
	{
		stopDisabling();
	}

	void stopDisabling()
	{
		ODSFuncS( _FormatT( L"%x, %x", bRet, OldValue ) );
		// Restore the previous WOW64 file system redirection value.
		if (bRet == TRUE) 
			cKernelDll::getStatic().Wow64RevertWow64FsRedirection (OldValue);

		ODSFuncRes( _FormatT( L"%x, %x", bRet, OldValue ) );
	}

	void continueDisabling()
	{
		//ODS_fl;
		// If bRet was not assigned before it will not restore disabling.
		startDisable( bRet );
	}

	void upload()
	{
		stopDisabling();
		bRet = false;
	}
};
