#include "Mail.h"
#include <MAPI.h>
#include "CDll.h"
#include "UnicodeFuncs.h"
#include <io.h>
#include <stdio.h>
#include "cMBStr.h"

/*
int mapi_LogOnAndSendMail(
		HWND hWnd,
		const wchar_t *filePath,
		const wchar_t *userName,
		const wchar_t *eMailAddress,
		const char *szSubject,
		const char *szNotes
	)
{

}
*/

int mapi_SendMail(
		HWND hWnd,
		const char *filePath,
		const char *userName,
		const char *eMailAddress,
		const char *szSubject,
		const char *szNotes,
		const char *originalFileName
	)
{
	CDll mapi32dll( false, "mapi32.dll", "MAPISendMail", true );
	if ( mapi32dll.error )
		return -1;

	MAPISENDMAIL *sendMail = ( MAPISENDMAIL * )mapi32dll.pProc;
///////////

	MapiRecipDesc ord;
	memset( &ord, 0, sizeof( ord ) );
	ord.ulRecipClass = MAPI_ORIG;
	ord.lpszName = NULL; // "Pav";
	ord.lpszAddress = NULL; // "SMTP:M@X.COM";
	ord.ulEIDSize = 0;
	ord.lpEntryID = NULL;

	MapiRecipDesc rec;

	bool empty = (eMailAddress == NULL) || ( strlen( eMailAddress ) == 0 );

	if ( !empty )
	{
		memset( &rec, 0, sizeof( rec ) );
		rec.ulRecipClass = MAPI_TO;
		rec.lpszName = ( char * )userName;
		rec.lpszAddress = ( char * )eMailAddress;
		rec.ulEIDSize = 0;
		rec.lpEntryID = NULL;
	}

	MapiFileDesc fdsc;
	int filesQ = 0;
	
	if ( ( filePath != NULL ) && ( strlen( filePath ) != 0 ) )
	{
		filesQ = 1;

		memset( &fdsc, 0, sizeof( fdsc ) );
		fdsc.flFlags = 0;
		fdsc.nPosition = -1; // ???

		fdsc.lpszPathName = ( char *)filePath; //  tmpFileName; // "C:\\qqq.bca";

		const char *fileName = originalFileName;

		if ( fileName == NULL )
			fileName = filePath + splitPath( filePath, NULL, NULL );

		fdsc.lpszFileName = ( char * )fileName;

		SHFILEINFOA sfi;
		memset( &sfi, 0, sizeof( sfi ) );
		SHGetFileInfoA( filePath, 0, &sfi, sizeof( sfi ), SHGFI_TYPENAME );

		fdsc.lpFileType = sfi.szTypeName;
	}

	MapiMessage msg;
	memset( &msg, 0, sizeof( msg ) );
	
	msg.lpszSubject = (char *)szSubject;
	msg.lpszNoteText = (char *)szNotes;
	msg.lpszMessageType = NULL;

	msg.lpszDateReceived = NULL; // "2004/01/01 00:00"; // The format is YYYY/MM/DD HH:MM, using a 24-hour clock.
	msg.lpszConversationID = 0;
	msg.flFlags = MAPI_SENT | MAPI_UNREAD;
	msg.lpOriginator = NULL; // &ord;
	msg.nRecipCount = empty ? 0 : 1;
	msg.lpRecips = empty ? NULL : &rec;
	msg.nFileCount = filesQ;
	msg.lpFiles = filesQ == 0 ? NULL : &fdsc;

//	int res = sendMail( 0, (ULONG_PTR)hWnd, &msg, MAPI_DIALOG /*| MAPI_ENVELOPE_ONLY */, 0 );
	int res = sendMail( 0, (ULONG)(ULONG_PTR)hWnd, &msg, MAPI_DIALOG /*| MAPI_ENVELOPE_ONLY */, 0 );

	if ( res != SUCCESS_SUCCESS )
	{
		// an error f.e. MAPI_E_ATTACHMENT_NOT_FOUND
		char errStr[ 100 ];
		switch( res )
		{
			case MAPI_E_FAILURE: 				sprintf_s( errStr, "E_FALURE" );			break;
			case MAPI_USER_ABORT: 				sprintf_s( errStr, "USER_ABORT" );			break;
			case MAPI_E_LOGON_FAILURE: 			sprintf_s( errStr, "LOGON_FAILURE" );			break;
			case MAPI_E_DISK_FULL:				sprintf_s( errStr, "DISK_FULL" );				break;
			case MAPI_E_ATTACHMENT_NOT_FOUND: 	sprintf_s( errStr, "ATTACHMENT_NOT_FOUND" );	break;
			default: 
				sprintf_s( errStr, "0x%X", res );
		}
		
		char message[ 1024 ];
		//sprintf_s( message, "Error %s\n", errStr );

		//if ( msg.lpRecips != NULL )
		//{
		//	size_t len = strlen( message );
		//	sprintf_s(
		//		message + len, sizeof( message ) - len,
		//		"User name : %s\n"
		//		"E-mail address : %s\n",
		//		msg.lpRecips->lpszName,
		//		msg.lpRecips->lpszAddress);
		//}

		//if ( msg.lpFiles != NULL )
		//{
		//	size_t len = strlen( message );
		//	sprintf_s(
		//		message + len, sizeof( message ) - len,
		//		"File name %s", 
		//		msg.lpFiles->lpszPathName);
		//}
		//MessageBoxA( hWnd, message, "Error in sending e-Mail", MB_OK );

		sprintf_s( message, "E-mail cannot be send (error %s)", errStr );
		MessageBoxA( hWnd, message, "Send e-Mail", MB_ICONSTOP );
	}

	return res;
}

int mapi_SendMail(
		HWND hWnd,
		const wchar_t *filePath,
		const wchar_t *userName,
		const wchar_t *eMailAddress,
		const wchar_t *szSubject,
		const wchar_t *szNotes,
		const wchar_t *originalFileName
	)
{
	cMBStr subj( szSubject );
	cMBStr notes( szNotes );
	cMBStr origFileName( originalFileName );

	return mapi_SendMail( hWnd, filePath, userName, eMailAddress, subj.m_szBody, notes.m_szBody, origFileName.m_szBody );
}

//struct smParam
//{
//	LHANDLE lhSession;
//	ULONG_PTR ulUIParam;
//	lpMapiMessage lpMessage;
//	FLAGS flFlags;
//	ULONG ulReserved;
//};
//
//DWORD WINAPI sendMailThreadP( void *param )
//{
//	smParam *sp = (smParam *)param;
//
//	CDll mapi32dll( false, "mapi32.dll", "MAPISendMail", true );
//	if ( mapi32dll.error )
//		return -1;
//
//	MAPISENDMAIL *sendMail = ( MAPISENDMAIL * )mapi32dll.pProc;
//
//	DWORD tlsIndex = TlsAlloc();
//
//	DWORD dwRes = sendMail( sp->lhSession, sp->ulUIParam, sp->lpMessage, sp->flFlags, sp->ulReserved );
//
//	TlsFree( tlsIndex );
//
//	return dwRes;
//}
//
//ULONG sendMailThread( 
//	LHANDLE lhSession,
//	ULONG_PTR ulUIParam,
//	lpMapiMessage lpMessage,
//	FLAGS flFlags,
//	ULONG ulReserved
//	)
//{
//	smParam sp;
//	sp.lhSession = lhSession;
//	sp.ulUIParam = ulUIParam;
//	sp.lpMessage = lpMessage;
//	sp.flFlags = flFlags;
//	sp.ulReserved = ulReserved;
//
//
//
//	HANDLE hThread = CreateThread( NULL, 0, sendMailThreadP, &sp, 0, NULL );
//
//	DWORD dwRes = WaitForSingleObject( hThread, INFINITE );
//
//	ULONG ulRes;
//
//	if ( !GetExitCodeThread( hThread, &ulRes ) )
//		ulRes = -1;
//
//	CloseHandle( hThread );
//
//	return ulRes;
//}

int mapi_SendMail(
		HWND hWnd,
		const wchar_t *filePath,
		const wchar_t *userName,
		const wchar_t *eMailAddress,
		const char *szSubject,
		const char *szNotes,
		const char *originalFileName
	)
{
	CDll mapi32dll( false, "mapi32.dll", "MAPISendMail", true );
	if ( mapi32dll.error )
		return -1;

	MAPISENDMAIL *sendMail = ( MAPISENDMAIL * )mapi32dll.pProc;
///////////

	MapiRecipDesc ord;
	memset( &ord, 0, sizeof( ord ) );
	ord.ulRecipClass = MAPI_ORIG;
	ord.lpszName = NULL; // "Pav";
	ord.lpszAddress = NULL; // "SMTP:M@X.COM";
	ord.ulEIDSize = 0;
	ord.lpEntryID = NULL;

	MapiRecipDesc rec;

	bool empty = (eMailAddress == NULL) || ( wcslen( eMailAddress ) == 0 );

	char eMailAddressA[ 1024 ];
	char userNameA[ 1024 ];

	if ( !empty )
	{
		memset( &rec, 0, sizeof( rec ) );
		rec.ulRecipClass = MAPI_TO;
		_wcstombs( userNameA, sizeof( userNameA ), userName );
		rec.lpszName = userNameA;

		strcpy_s( eMailAddressA, "SMTP:" );
		int len = (int)strlen(eMailAddressA);

		_wcstombs( eMailAddressA + len, sizeof( eMailAddressA ) - len, eMailAddress );
		rec.lpszAddress = eMailAddressA;
		rec.ulEIDSize = 0;
		rec.lpEntryID = NULL;
	}

	MapiFileDesc fdsc;
	int filesQ = 0;
	
	cMBStr filePathA;

	if ( ( filePath != NULL ) && ( wcslen( filePath ) != 0 ) )
	{
		filesQ = 1;
		memset( &fdsc, 0, sizeof( fdsc ) );
		fdsc.flFlags = 0;
		fdsc.nPosition = -1; // ???

		cWCStr filePathLong;

		GetLongPathNameT( filePath, &filePathLong );
		filePathA.Set( filePathLong.m_szBody );

		fdsc.lpszPathName = filePathA.m_szBody;

		const char *fileNameA = originalFileName;

		if ( fileNameA == NULL )
			fileNameA = filePathA.m_szBody + splitPath( filePathA.m_szBody, NULL, NULL );

		fdsc.lpszFileName = (char *)fileNameA;

		fdsc.lpFileType = NULL;
	}

	MapiMessage msg;
	memset( &msg, 0, sizeof( msg ) );
	
	msg.lpszSubject = (char *)szSubject;
	msg.lpszNoteText = (char *)szNotes;
	msg.lpszMessageType = NULL;

	msg.lpszDateReceived = NULL; // "2004/01/01 00:00"; // The format is YYYY/MM/DD HH:MM, using a 24-hour clock.
	msg.lpszConversationID = 0;
	msg.flFlags = MAPI_SENT | MAPI_UNREAD;
	msg.lpOriginator = NULL; // &ord;
	msg.nRecipCount = empty ? 0 : 1;
	msg.lpRecips = empty ? NULL : &rec;
	msg.nFileCount = filesQ;
	msg.lpFiles = filesQ == 0 ? NULL : &fdsc;

	//int res = sendMailThread( 0, (ULONG)(ULONG_PTR)hWnd, &msg, MAPI_DIALOG | MAPI_NEW_SESSION /*| MAPI_ENVELOPE_ONLY */, 0 );
	int res = sendMail( 0, (ULONG)(ULONG_PTR)hWnd, &msg, MAPI_DIALOG | MAPI_NEW_SESSION /*| MAPI_ENVELOPE_ONLY */, 0 );

	if ( res == MAPI_E_LOGON_FAILURE )
		res = sendMail( 0, (ULONG)(ULONG_PTR)hWnd, &msg, MAPI_DIALOG | MAPI_LOGON_UI, 0 );
		//res = sendMailThread( 0, (ULONG)(ULONG_PTR)hWnd, &msg, MAPI_DIALOG | MAPI_LOGON_UI, 0 );

	if ( res != SUCCESS_SUCCESS )
	{
		char message[ 1024 ] = "";
		char errStr[ 100 ];
		switch( res )
		{
			case MAPI_USER_ABORT: 				return res; // strcpy_s( errStr, "USER_ABORT" );				break;
			case MAPI_E_LOGON_FAILURE:			strcpy_s( errStr, "LOGON_FAILURE" );			break;
			case MAPI_E_DISK_FULL:				strcpy_s( errStr, "DISK_FULL" );				break;
			case MAPI_E_ATTACHMENT_NOT_FOUND: 	strcpy_s( errStr, "ATTACHMENT_NOT_FOUND" );	break;
			default: 
				sprintf_s( errStr, "number: 0x%X", res );
		}

		char *userName = ( msg.lpRecips == NULL? "NULL" : msg.lpRecips->lpszName);
		char *eMail = ( msg.lpRecips == NULL? "NULL" : msg.lpRecips->lpszAddress );
		char *path = ( msg.lpFiles  == NULL? "NULL" : msg.lpFiles->lpszPathName );
		
		//sprintf_s( message, 
		//	"Error %s\n"
		//	"User name : %s\n"
		//	"E-mail address : %s\n"
		//	"File name %s", 
		//	errStr, 
		//	userName,
		//	eMail,
		//	path );

		//MessageBoxA( hWnd, message, "Error in sending e-Mail", MB_OK );

		sprintf_s( message, "E-mail cannot be send (error %s)", errStr );
		MessageBoxA( hWnd, message, "Send e-Mail", MB_ICONSTOP );
	}

	return res;
}
