#include "UnicodeFuncs.h"
#include "TranslateDll.h"
#include "cStringArray.h"
#include "cdll.h"
#include <windowsx.h>
#include "menuItem.h"
#include <assert.h>
#include "ODS.h"

#ifndef MIIM_STRING
#define MIIM_STRING      0x00000040
#define MIIM_FTYPE       0x00000100
#define ODS_HOTLIGHT        0x0040
#endif

sImageInfo *g_siiT = NULL;
int g_siiQ = 0;

//
//	Class standard images
//

//const BYTE stdDot[] = { 0x6, 0xF, 0xF, 0xF, 0x6 }; // simple byte bitmap, 1=> bit on
//const BYTE stdChk[] = { 0x0, 0x1, 0xF, 0xe, 0x4 }; // simple byte bitmap, 1=> bit on
//#define stdWidth   4
//#define stdHeight  5

const BYTE stdDot[] = { 0x00, 0x00, 0x1C, 0x3e, 0x3e, 0x3e, 0x1c, 0x00 };
const BYTE stdSelDot[] = { 0x00, 0x1c, 0x22, 0x41, 0x41, 0x41, 0x22, 0x1c };
const BYTE stdChk[] = { 0x00, 0x01, 0x03, 0x47, 0x6e, 0x7c, 0x38, 0x10 };
#define stdWidth   8
#define stdHeight  8

HBITMAP CreateBmp( const BYTE *array )
{
	LONG lSize = ::GetMenuCheckMarkDimensions();
	SIZE size;
	size.cx = LOWORD( lSize );
	size.cy = HIWORD( lSize );

	if (size.cx > 32)
		size.cx = 32;
	int iwRow = (size.cx + 15) >> 4;    // # of WORDs per raster line
	int nShift = (size.cx - stdWidth) / 2;     // # of bits to shift over
	nShift += ((iwRow * 16) - size.cx); // padding for word alignment
	if (nShift > 16 - stdWidth)
		nShift = 16 - stdWidth;    // maximum shift for 1 word

	if (size.cy > 32)
		size.cy = 32;

	BYTE rgbBitmap[32 * 2 * sizeof(WORD)];
	memset(rgbBitmap, 0xff, sizeof(rgbBitmap));

	BYTE* pbOut = &rgbBitmap[iwRow * sizeof(WORD) *
							((size.cy - (stdHeight+1)) >> 1)];
	const BYTE* pbIn = array;
	for (int y = 0; y < stdHeight; y++)
	{
		WORD w = (WORD)~(((DWORD)*pbIn++) << nShift);
		// bitmaps are always hi-lo
		pbOut[0] = HIBYTE(w);
		pbOut[1] = LOBYTE(w);
		pbOut += iwRow * sizeof(WORD);
	}

	HBITMAP hBmp = ::CreateBitmap(size.cx, size.cy, 1, 1, (LPVOID)&rgbBitmap );

	return hBmp;
}


bool cSTDImage::CreateImages()
{
	if ( m_ok )
		return true;

	if ( m_dotBmp == NULL ) m_dotBmp = CreateBmp( stdDot );
	if ( m_chkBmp == NULL ) m_chkBmp = CreateBmp( stdChk );
	if ( m_dotSelectedBmp == NULL )
							m_dotSelectedBmp = CreateBmp( stdSelDot );

	m_ok = true;

	return m_ok;
}

cSTDImage::cSTDImage()
{
	m_ok = false;
	m_chkBmp = NULL;
	m_dotBmp = NULL;
}

cSTDImage::~cSTDImage()
{
	if ( m_chkBmp != NULL )
		DeleteObject( m_chkBmp );

	if ( m_dotBmp != NULL )
		DeleteObject( m_dotBmp );
	
	if( m_dotSelectedBmp != NULL )
		DeleteObject( m_dotSelectedBmp );
}

cSTDImage cMenuItem::m_stdImgs;

void cMenuItem::SetRadio( HMENU hMenu, int id )
{
	cMenuItem *miParam = (cMenuItem *)GetMenuItemParam( hMenu, id, false );
	if ( miParam == NULL )
		return;

	miParam->m_bRadio = true;
}

bool cMenuItem::SetText( const wchar_t *text )
{
	if ( m_text != NULL )
		delete m_text;

	m_text = NULL;

	int len = 0;
	if ( text != NULL )
		len = (int)wcslen( text );

	m_text = new wchar_t[ len + 1 ];
	if ( m_text != NULL )
	{
		if ( len == 0 )
			m_text[0] = L'\0';
		else
			wcscpy_s( m_text, len + 1, text );
	}
	
	m_cmdChar = 0;

	for( int i = len - 1; i-- > 0;  )
	{
		if ( text[i] == L'&' )
		{
			if ( ( i != 0 ) && ( text[i - 1] == L'&' ) )
				i--;
			else
			{
				m_cmdChar = text[i + 1];
				break;
			}
		}
	}
	
	return m_text != NULL;
}

cMenuItem::cMenuItem( HMENU hMenu, HIMAGELIST hIml, int iImage, const wchar_t *text, UINT wID, HIMAGELIST hImlDisabled )
{
	m_size = sizeof( cMenuItem );
	m_sig = eSig;

	m_hMenu = hMenu;
	m_hIml = hIml;
	m_hImlDisabled = hImlDisabled;
	m_iImage = iImage;
	m_bRadio = false;
	m_text = NULL;
	m_wID = wID;

	SetText( text );
}

cMenuItem::~cMenuItem()
{
	if ( m_text != NULL )
		delete m_text;
}

static HMENU FindSubmenuById( HMENU hMenu, UINT nID )
{
	UINT nItems = GetMenuItemCount( hMenu );
	for (int iItem = 0; iItem < (int)nItems; iItem++)
	{
		HMENU hPopup = GetSubMenu( hMenu, iItem);
		if ( hPopup != NULL)
		{
			// recurse to child popup
			hPopup = FindSubmenuById( hPopup, nID);
			// check popups on this popup
			if ( hPopup != NULL)
				return hPopup;
		}
		else
		{
			UINT id = GetMenuItemID( hMenu, iItem );
			if ( id == nID )
			{
				// it is a normal item inside our popup
				return hMenu;
			}
		}
	}

	// not found
	return NULL;
}

int cMenuItem::m_iconSize = 16;
HFONT cMenuItem::m_hFontMenu = NULL;

SIZE cMenuItem::GetMenuTextSize( HWND hWnd, UINT fType, UINT fState, const wchar_t *text )
{
	//ODSFuncS( text );

	if ( m_hFontMenu == NULL )
		GetSystemSettings();

	SIZE s = { 0, 0 };

	if ( m_hFontMenu == NULL )
	{
		ODS_fls( L"Assertion!  m_hFontMenu == NULL" );
		assert( false );
		return s;
	}

	if( fType & MFT_SEPARATOR )
	{
		s.cy = ::GetSystemMetrics(SM_CYMENU) / 2;
		s.cx = 0;
	}
	else
	{
		HDC hdc = GetDC( hWnd );
		HFONT hFontBold = NULL;
		HFONT hOldFont = NULL;

		LOGFONT lf = { 0 };
		::GetObject( m_hFontMenu, sizeof( LOGFONT ), &lf );

		if( fState & MFS_DEFAULT )
		{
			lf.lfWeight += 200;
			hFontBold = ::CreateFontIndirect( &lf );
			assert( hFontBold != NULL);
			hOldFont = SelectFont( hdc, hFontBold );
		}
		else
		{
			hOldFont = SelectFont( hdc, m_hFontMenu );
		}

		RECT rcText = { 0, 0, 0, 0 };
		int res = DrawTextT( hdc, text, -1, &rcText, DT_SINGLELINE | DT_LEFT | DT_VCENTER | DT_CALCRECT );
		s.cx = rcText.right - rcText.left;
		SelectFont( hdc, hOldFont );
	
		if ( hFontBold != NULL )
			DeleteObject( hFontBold );

		DeleteDC( hdc );

		s.cy = lf.lfHeight;
		if( s.cy < 0)
			s.cy = -s.cy;
		const int cyMargin = 8;
		s.cy += cyMargin;

		//// height of item is the bigger of these two
		//lpMeasureItemStruct->itemHeight = max(cy, (int)m_szButton.cy);

		//// width is width of text plus a bunch of stuff
		//cx += 2 * s_kcxTextMargin;   // L/R margin for readability
		//cx += s_kcxGap;              // space between button and menu text
		//cx += 2 * m_szButton.cx;     // button width (L=button; R=empty margin)
		//cx += m_cxExtraSpacing;      // extra between item text and accelerator keys

		//// Windows adds 1 to returned value
		//cx -= ::GetSystemMetrics(SM_CXMENUCHECK) - 1;
		//lpMeasureItemStruct->itemWidth = cx;   // done deal
	}

	//ODSFuncRes( s.cx );

	return s;
}

const int delta = 6;

static int selectedShift = 1;
const static bool useSelectedBorder = true;

static int rightIconBorder = 1;

static int iconPos = 1;

int cMenuItem::GetItemHeight( int textHeight )
{
	return max( m_iconSize + iconPos + 1 + selectedShift + useSelectedBorder, textHeight + 2 * selectedShift );
}

static int getTabSymbolPos( const wchar_t *text )
{
	for( int i = 0; text[i] != L'\0'; i++ )
	{
		if ( text[i] == L'\t' )
			return i;
	}

	return -1;
}

static int popupArrowSize = 10;

bool cMenuItem::MeasureItem( HWND hWnd, MEASUREITEMSTRUCT *pmi )
{
	//ODSFunc;
	
	popupArrowSize = m_iconSize - 6;

	if ( pmi->CtlType != ODT_MENU )
	{
		assert( false );
		return false;
	}

	cMenuItem *miParam = (cMenuItem *)pmi->itemData;

	pmi->itemWidth	= pmi->itemHeight + delta + rightIconBorder;

	if ( miParam == NULL )
	{
		// is separator
		pmi->itemHeight = 3;
		return true;
	}

	HMENU hMenu = miParam->m_hMenu;
	bool bMainMenu = false;

	if ( GetMenu( hWnd ) == hMenu )
	{
		bMainMenu = true;
	}
	else
	{
		hMenu = FindSubmenuById( hMenu, pmi->itemID );
	}

	if ( hMenu == NULL )
	{
		// Not main menu but menu with submenus.
		hMenu = miParam->m_hMenu;
//		return false;
	}

	MENUITEMINFOW mii;
	memset( &mii, 0, sizeof ( mii ) );

	mii.cbSize = MENUITEMINFO_SIZE_VERSION_400; // sizeof( mii );

	mii.fMask = MIIM_TYPE | MIIM_STATE | MIIM_SUBMENU;

	if ( !GetMenuItemInfoT( hMenu, pmi->itemID, false, &mii ) )
	{
		assert( false );
		ODSF( L"GetMenuItemInfoT error %d", GetLastError() );
		return false;
	}

	wchar_t bf[ 256 ];
	const wchar_t *text =  miParam->m_text;

	//
	//	Tabulation '\t' symbol means separated string f.e. >'Exit    Alt+X'<
	//
	int pos = getTabSymbolPos( text );
	if ( ( pos != -1 ) && ( pos < maxLen( bf ) ) )
	{
		wcsncpy_s( bf, maxLen( bf ), text, _TRUNCATE );
		bf[ pos ] = L' ';
		text = bf;
		
	}

	SIZE size = GetMenuTextSize( hWnd, mii.fType, mii.fState, text );

	pmi->itemHeight = GetItemHeight( size.cy );

	int textSize = 0;
	textSize = size.cx;

	if ( bMainMenu )
		pmi->itemWidth = delta + textSize;
	else
	{
		pmi->itemWidth += textSize;

//		if ( mii.hSubMenu != NULL ) // has submenu and has to contain '>'
			pmi->itemWidth += popupArrowSize;
	}

	return true;
}

#ifndef _NoComctl32Lib_
#define _UseImageList_
#endif

#ifdef _UseImageList_

#ifndef ILS_NORMAL
#define ILS_NORMAL              0x00000000 
#define ILS_GLOW                0x00000001
#define ILS_SHADOW              0x00000002
#define ILS_SATURATE            0x00000004
#endif

static BOOL _ImageList_Draw(HIMAGELIST himl, int i, HDC hdcDst, int x, int y, UINT fStyle)
{
	//if ( !cComCtl32Dll::getStatic().LoadOnce() )
	//	return false;

	typedef BOOL WINAPI typeImageList_Draw(HIMAGELIST himl, int i, HDC hdcDst, int x, int y, UINT fStyle);

	static typeImageList_Draw *proc;
	static bool ok = cComCtl32Dll::getStatic().GetProcAddress( "ImageList_Draw", false, (void**)&proc ) == TRUE;

	if ( !ok )
		return false;

	return proc( himl, i, hdcDst, x, y, fStyle );
}


bool DrawIndirect( HIMAGELIST hIml, int iImage, HDC hDC, int x, int y, UINT fStyle = ILD_NORMAL, DWORD fState = ILS_NORMAL )
{
	typedef WINCOMMCTRLAPI BOOL WINAPI typeImageList_DrawIndirect(IMAGELISTDRAWPARAMS* pimldp);

	static typeImageList_DrawIndirect *proc;
	static bool ok = cComCtl32Dll::getStatic().GetProcAddress( "ImageList_DrawIndirect", false, (void**)&proc ) == TRUE;

	if ( !ok )
		return false;

	// Sample draw saturated
	struct IMAGELISTDRAWPARAMS_V51
	{
		DWORD       cbSize;
		HIMAGELIST  himl;
		int         i;
		HDC         hdcDst;
		int         x;
		int         y;
		int         cx;
		int         cy;
		int         xBitmap;        // x offest from the upperleft of bitmap
		int         yBitmap;        // y offset from the upperleft of bitmap
		COLORREF    rgbBk;
		COLORREF    rgbFg;
		UINT        fStyle;
		DWORD       dwRop;
		DWORD       fState;
		DWORD       Frame;
		COLORREF    crEffect;
	};

	IMAGELISTDRAWPARAMS_V51 ildp;
	memset( &ildp, 0, sizeof( ildp ) );
	ildp.cbSize = sizeof( ildp );
	ildp.himl = hIml;
	ildp.i = iImage;
	ildp.hdcDst =hDC;
	ildp.x = x;
	ildp.y = y;
	ildp.rgbBk = RGB( 0, 0, 0 );// CLR_NONE;
	ildp.rgbFg = RGB( 0, 0, 0 );
	ildp.fStyle = ILD_TRANSPARENT | fStyle; // ILD_NORMAL;// | ILD_OVERLAYMASK;
	ildp.fState = fState; // ILS_SATURATE;//  | ILS_ALPHA;
	ildp.Frame = 100;
	ildp.crEffect = RGB( 0x0,0x0,0x0 );

	//return ImageList_DrawIndirect( (IMAGELISTDRAWPARAMS *)&ildp ) != FALSE;
	return proc( (IMAGELISTDRAWPARAMS *)&ildp ) != FALSE;
}

bool DrawSaturated( HIMAGELIST hIml, int iImage, HDC hDC, int x, int y )
{
	return DrawIndirect( hIml, iImage, hDC, x, y, ILD_NORMAL, ILS_SATURATE );
}

#endif // _UseImageList_

bool FillRect_WithGradient( HDC hdc,  RECT *r, COLORREF cr, DWORD dwMode = GRADIENT_FILL_RECT_H, bool fromTopToDown = false )
{
	static CDll msImgDll( false, "Msimg32.dll", "GradientFill", false );

	if ( msImgDll.error )
	{
		return false;
	}

	TRIVERTEX        vert[2] ;
	GRADIENT_RECT    gRect;
	vert [0] .x      = r->left;
	vert [0] .y      = r->top;
	vert [0] .Red    = 0xff00;
	vert [0] .Green  = 0xff00;
	vert [0] .Blue   = 0xff00;
	vert [0] .Alpha  = 0x0000;

	vert [1] .x      = r->right;
	vert [1] .y      = r->bottom; 
	vert [1] .Red    = GetRValue( cr ) << 8;
	vert [1] .Green  = GetGValue( cr ) << 8;
	vert [1] .Blue   = GetBValue( cr ) << 8; // 0xff00;
	vert [1] .Alpha  = 0x0000;

	if ( fromTopToDown )
	{
		// swap verts
		vert[0].x ^= vert[1].x;
		vert[1].x ^= vert[0].x;
		vert[0].x ^= vert[1].x;

		vert[0].y ^= vert[1].y;
		vert[1].y ^= vert[0].y;
		vert[0].y ^= vert[1].y;
	}

	gRect.UpperLeft  = 0;
	gRect.LowerRight = 1;

	typedef WINGDIAPI BOOL  WINAPI typeGradientFill( IN HDC, IN PTRIVERTEX, IN ULONG, IN PVOID, IN ULONG, IN ULONG);
	typeGradientFill *progGradientFill = ( typeGradientFill * )msImgDll.pProc;

	return progGradientFill( hdc, vert, 2, &gRect, 1, dwMode ) != FALSE;
}

bool DrawBitmap( HDC hDC, RECT *r, HBITMAP hBmp, DWORD dwRop = SRCAND )
{
	BITMAP bmpInfo;
	::GetObject( hBmp, sizeof(bmpInfo), &bmpInfo );

	HDC memDC = CreateCompatibleDC( hDC );

	// Select the bitmap into the in-memory DC
	HBITMAP hOldBitmap = (HBITMAP)::SelectObject( memDC, hBmp );

	// Find a centerpoint for the bitmap in the client area
	int nX = r->left + ( r->right - r->left - bmpInfo.bmWidth ) / 2;
	int nY = r->top + ( r->bottom - r->top - bmpInfo.bmHeight) / 2;

	// Copy the bits from the in-memory DC into the on-
	// screen DC to actually do the painting. Use the centerpoint
	// we computed for the target offset.
	BitBlt( hDC, nX, nY, bmpInfo.bmWidth, bmpInfo.bmHeight, memDC, 0, 0, dwRop );

	SelectObject( memDC, (HGDIOBJ)hOldBitmap );

	// ???
	//DeleteDC( memDC );
	//DeleteObject( hOldBitmap );

	return true;
}

static const int rowQ = 20;
struct { int left; int right; } rowT[ rowQ ];

//
//	Disables update rectangle of recently selected item. 
//	It allows to avoid error in detecting the last menu has to be extended up to right size of menu bar.
//	See updateRect
//
//static RECT recentlySelectedMenuRect = { 0,0,0,0 };

static int getNum( RECT *r )
{
	int size = r->bottom - r->top;
	if ( size == 0 )
		return 0;

	int n = r->top / size;

	if ( n < rowQ )
		return n;

	return 0;
}

//
//	addRectToTable
//		adds rectangel to row
//		returns number of the rectangle row.
//
static int addRectToTable( RECT *r, bool *pNewPosition )
{
	int num = getNum( r );

	bool newPosition = false;

	if ( ( rowT[num].right == 0 ) && ( rowT[num].left == 0 ) )
	{
		rowT[num].right = r->right;
		rowT[num].left = r->left;
		newPosition = true;
	}
	else
	{
		if ( rowT[num].right < r->right )
		{
			rowT[num].right = r->right;
			newPosition = true;
		}

		if ( rowT[num].left > r->left )
			rowT[num].left = r->left;
	}

	if ( pNewPosition )
		*pNewPosition = newPosition;

	return num;
}

static bool isTableEmpty()
{
	for ( int q = rowQ; q--; )
	{
		if ( rowT[q].right != 0 ) return false;
		if ( rowT[q].left  != 0 ) return false;
	}

	return true;
}

static void clearTable()
{
	memset( rowT, 0, sizeof( rowT ) );
}

static void flipRect( RECT *r )
{
	r->left	 ^= r->right;
	r->right ^= r->left;
	r->left  ^= r->right;
}

//void GetClient( RECT *r, HWND hWnd, HWND hParent )
//{
//	if ( hParent == NULL )
//		hParent = GetParent( hWnd );
//
//	RECT parentR, wndR;
//	GetWindowRect( hParent, &parentR );
//	GetWindowRect( hWnd, &wndR );
//	GetClient( r, &wndR, &parentR );
//}
//
void GetClient( RECT *r, const RECT *scrR, const RECT *parentR )
{
	r->top = scrR->top - parentR->top;
	r->bottom = scrR->bottom - parentR->top;

#ifndef _NoTranslation_
	if ( ( pTranslateDll != NULL ) && pTranslateDll->m_flipDialog )
	{
		r->left = parentR->right - scrR->right;
		r->right = parentR->right - scrR->left;
	}
	else
#endif // #ifndef _NoTranslation_

	{
		r->left = scrR->left - parentR->left;
		r->right = scrR->right - parentR->left;
	}
}


//
//	updateTable
//	1. Clears old rows state
//	2. Gets menu item's rectangles and add them to row tables
//
static void updateTable( HWND hWnd, HMENU hMenu )
{
	clearTable();

	int c = GetMenuItemCount( hMenu );

	RECT rMenu,rm;
	RECT r;
	
	GetWindowRect( hWnd, &r );

	while( c-- )
	{
		GetMenuItemRect( hWnd, hMenu, c, &rm );
		GetClient( &rMenu, &rm, &r );
		addRectToTable( &rMenu, NULL );

		//GetMenuItemRect( hWnd, hMenu, c, &r );
		//MapWindowPoints( NULL, hWnd, (POINT*)&r, 2 );
		//addRectToTable( &r, NULL );
	}
}

void cMenuItem::OnSize( HWND hWnd, HMENU hMenu, UINT nType, int cx, int cy )
{
	clearTable();

	if ( hMenu == NULL )
		return;
}

static bool isLastInRow( int num, RECT *r )
{
	return rowT[num].right == r->right;
}

//
//	Extends last menu item rectangle to be completely filled with gradient the main menu rectangle
//
static bool updateRect( HWND hWnd, HMENU hMenu, RECT *r ) 
{
	bool newPosition = false;
	int num = addRectToTable( r, &newPosition );

	//
	//	The 'newPosition==true' means that the menu item was never present at the row table.
	//	So it seems the menu size and row state was changed.
	//	The table is updated.
	//	Note, the clever to updated table once at OnSize, but Win2K sends WM_SIZE after updating
	//
	if ( newPosition )
	{
		updateTable( hWnd, hMenu );

		// Assertion if the rectangle 
		// 1. has same num
		assert( addRectToTable( r, &newPosition ) == num );
		// 2. was included by updateTable
		assert( !newPosition );
	}

	bool isLast = isLastInRow( num, r );

	if ( !isLast )
		return false;


	RECT wr;
	GetWindowRect( hWnd, &wr );
	r->right = wr.right - wr.left - rowT[num].left;

	return true;
}

bool cMenuItem::DrawItem( HWND hWnd, DRAWITEMSTRUCT *pdi )
{
//	ODSFunc;
	static bool isW95 = GetVersion() >= 0x80000000;
	RECT r = pdi->rcItem;
	COLORREF	bkColor, txtColor, bkIconColor, borderColor;
	
	bkIconColor = COLOR_BTNFACE; //COLOR_BTNSHADOW; // COLOR_GRAYTEXT;
	borderColor = COLOR_BTNSHADOW; //COLOR_BTNFACE; // COLOR_MENUHILIGHT;

	UINT ilStyle = ILD_TRANSPARENT;

	int shift = iconPos + selectedShift;
	bool selected = false;
	bool grayed = false;

	//if ( pdi->itemAction & ODA_DRAWENTIRE )
	//{
	//	static int i = 0;
	//	bkColor = (COLOR_MENU + i++ ) % 30;
	//}

	bool popupDisplaied = false;

	if ( pdi->itemState & ODS_GRAYED )
	{
		bkColor = COLOR_MENU;
		txtColor = ::GetSysColor( COLOR_GRAYTEXT );
		grayed = true;
		if ( isW95 )
			ilStyle = ILD_MASK;
		else
			ilStyle |= ILD_BLEND25; // 0;
	}
	else
	if ( pdi->itemState & ( ODS_FOCUS | ODS_SELECTED | ODS_HOTLIGHT ) )
	{
		bkColor = 28; //COLOR_GRADIENTINACTIVECAPTION;
		// non XP COLOR_HIGHLIGHT;
		txtColor = ::GetSysColor( COLOR_MENUTEXT );// ::GetSysColor( COLOR_HIGHLIGHTTEXT );
		ilStyle |= 0;
		bkIconColor = bkColor;

		popupDisplaied = ( pdi->itemState & ODS_SELECTED ) != 0;

		selected = true;
		
		shift = iconPos;
	}
	else
	{
		bkColor = COLOR_MENU;
		txtColor = ::GetSysColor( COLOR_MENUTEXT );
		ilStyle = 0;
	}

	cMenuItem *miParam = (cMenuItem *)pdi->itemData;

	int cxText = m_iconSize + selectedShift + iconPos + 1 + delta + rightIconBorder;

	bool bMainMenu = false;
	HMENU hMenu = NULL;

	if ( miParam != NULL )
	{
		if ( (DWORDLONG)miParam < 5 )
		{
			ODS_fls( L"Assertion!" );
			return false;
		}

		hMenu = miParam->m_hMenu;
	}

	if ( ( GetMenu( hWnd ) == hMenu ) && ( miParam != NULL ) )
	{
		bMainMenu = true;
	}

	if ( bMainMenu )
	{
		RECT rIcon = r;

		// Updates rectangle for last item. It has to be extended up to menu bar right side.
		if ( !selected )
		{
			updateRect( hWnd, hMenu, &rIcon ); 
		}

		cxText = delta;

		if ( useSelectedBorder && selected )
		{
			//
			// Fills as usual first
			//	( otherwise it looks contacted with caption rectangle )
			//
			if ( !FillRect_WithGradient( pdi->hDC, &rIcon, ::GetSysColor( COLOR_MENU ), GRADIENT_FILL_RECT_V ) ) 
				FillRect( pdi->hDC, &rIcon, ::GetSysColorBrush( COLOR_MENU ) );

			rIcon.top += 1;

			//
			//	Draws the border.
			//
//			if ( !FillRect_WithGradient( pdi->hDC, &rIcon, ::GetSysColor( borderColor ), GRADIENT_FILL_RECT_V, selected ) ) 
				FillRect( pdi->hDC, &rIcon, ::GetSysColorBrush( borderColor ) );

			//
			// Prepares for drawing body of seleced item
			//
			rIcon.top += 1;
			rIcon.left += 1;

			//
			//	It does not draw bottom border if popup menu is displaied
			//
			if ( !popupDisplaied )
				rIcon.bottom -= 1;

			rIcon.right -= 1;
		}

		//if ( ( rIcon.right - rIcon.left ) > 0x200 )
		//{
		//	rIcon.right = rIcon.left + 0x50;
		//}

		if ( !FillRect_WithGradient( pdi->hDC, &rIcon, ::GetSysColor( bkIconColor ), GRADIENT_FILL_RECT_V, selected ) ) 
			FillRect( pdi->hDC, &rIcon, ::GetSysColorBrush( bkIconColor ) );

		SetBkMode(pdi->hDC, TRANSPARENT );
		SetTextColor( pdi->hDC, txtColor );
	}
	else
	{
		RECT rIcon = r;
		rIcon.right = rIcon.left + m_iconSize + 1 + selectedShift + rightIconBorder;

		if ( useSelectedBorder && selected )
		{
			if ( !FillRect_WithGradient( pdi->hDC, &rIcon, ::GetSysColor( borderColor ) ) ) 
				FillRect( pdi->hDC, &rIcon, ::GetSysColorBrush( borderColor ) );

			rIcon.top += 1;
			rIcon.left += 0;
			rIcon.bottom -= 1;
			rIcon.right -= 0;
		}

		if ( !FillRect_WithGradient( pdi->hDC, &rIcon, ::GetSysColor( bkIconColor ) ) )
		{
			FillRect( pdi->hDC, &rIcon, ::GetSysColorBrush( bkIconColor ) );
		}

		// is separator
		if ( miParam == NULL )
		{
			r.left += cxText;
			r.bottom -= 1;
			r.top += 1;
			FillRect( pdi->hDC, &r, ::GetSysColorBrush( COLOR_GRAYTEXT ) );
			//ODS( L"Separator" );
			return true;
		}

		miParam->m_stdImgs.CreateImages();

		RECT rBK = r;
		rBK.left = rIcon.right;

		if ( useSelectedBorder && selected )
		{
			FillRect( pdi->hDC, &rBK, ::GetSysColorBrush( borderColor ) );//bkIconColor ) );
			rBK.top += 1;
			rBK.left += 0;
			rBK.bottom -= 1;
			rBK.right -= 1;
		}

		FillRect( pdi->hDC, &rBK, ::GetSysColorBrush( bkColor ) );

		SetBkMode(pdi->hDC, TRANSPARENT );
		SetTextColor( pdi->hDC, txtColor );

		bool checked = ( pdi->itemState & ODS_CHECKED ) != 0;

		if ( ( miParam->m_hIml != NULL ) && ( miParam->m_iImage != -1 ) )
		{
	//	Draws saturated shadow before 
	//		if ( selected )
	//		{
	//			DrawSaturated( miParam->m_hIml, miParam->m_iImage, pdi->hDC, r.left + iconPos + selectedShift, r.top + iconPos + selectedShift );
	//		}
	//
#ifdef _UseImageList_
			bool useShadow = false;
			if ( useShadow && selected )
			{
				if ( miParam->m_hImlDisabled != NULL )
					_ImageList_Draw( miParam->m_hImlDisabled, miParam->m_iImage, pdi->hDC, r.left + iconPos + selectedShift, r.top + iconPos + selectedShift + useSelectedBorder, 0 );
				else
					DrawSaturated( miParam->m_hIml, miParam->m_iImage, pdi->hDC, r.left + iconPos + selectedShift, r.top + iconPos + selectedShift + useSelectedBorder );
			}
	
			if ( !grayed ||
				 !(
				    ( miParam->m_hImlDisabled != NULL ) ?
					_ImageList_Draw( miParam->m_hImlDisabled, miParam->m_iImage, pdi->hDC, r.left + shift, r.top + shift + useSelectedBorder, 0 ):
				    DrawSaturated( miParam->m_hIml, miParam->m_iImage, pdi->hDC, r.left + shift, r.top + shift + useSelectedBorder )
				   )
				)
			{
				if ( !DrawIndirect( miParam->m_hIml, miParam->m_iImage, pdi->hDC, r.left + shift, r.top + shift + useSelectedBorder, ilStyle ) )
					_ImageList_Draw( miParam->m_hIml, miParam->m_iImage, pdi->hDC, r.left + shift, r.top + shift + useSelectedBorder, ilStyle );
			}
#endif // _UseImageList_
		}
		else
		if ( checked || ( miParam->m_bRadio && selected ) )
		{
			RECT r = rIcon;

			if ( miParam->m_bRadio )
			{
				r.left += selectedShift;
				r.top += selectedShift;
			}
			else
			{
				r.left += shift;
				r.top += shift + useSelectedBorder;
			}
			//r.right = r.left + 14;
			//r.bottom = r.top + 14;

			if ( miParam->m_bRadio && ( miParam->m_stdImgs.m_dotBmp != NULL ) )
			{
				if ( selected && !checked )
					DrawBitmap( pdi->hDC, &r, miParam->m_stdImgs.m_dotSelectedBmp );
				else
					DrawBitmap( pdi->hDC, &r, miParam->m_stdImgs.m_dotBmp );
			}
			else
			if ( miParam->m_stdImgs.m_chkBmp != NULL )
			{
				DrawBitmap( pdi->hDC, &r, miParam->m_stdImgs.m_chkBmp );
			}
			else
			{
				UINT state = DFCS_MENUCHECK;
				UINT type = DFC_MENU;

			//	UINT type = DFC_BUTTON;
			//	UINT state = DFCS_CHECKED;

				DrawFrameControl( pdi->hDC, &r, type, state );
			}
		}
	}

	r.left += cxText;

	const wchar_t *text =  miParam->m_text;

	int pos = getTabSymbolPos( text );

	int res = DrawTextT( pdi->hDC, text, pos, &r, DT_VCENTER | DT_SINGLELINE );
	//ODSF( L"DrawTextT( %s ) = %d", text, res );

	if ( pos != -1 )
	{
		wchar_t bf[ 256 ];
		wcsncpy_s( bf, maxLen( bf ), text + pos + 1, _TRUNCATE );
		wcscat_s( bf, maxLen( bf ), L" " );
		r.right -= popupArrowSize;
		DrawTextT( pdi->hDC, bf, -1, &r, DT_RIGHT | DT_VCENTER | DT_SINGLELINE );
//		ODSF( L"DrawTextT( %s, pos %d ) = %d", text, pos, res );
	}

	return true;
}

//
// WM_MENUCHAR message 
//
//LONG cMenuItem::MenuChar( HMENU hMenu, wchar_t wc )
//{
//	UINT nItems = GetMenuItemCount( hMenu );
//
//	MENUITEMINFOW mii;
//	memset( &mii, 0, sizeof( mii ) );
//	mii.cbSize = MENUITEMINFO_SIZE_VERSION_400; //sizeof( mii );
//	mii.fMask = MIIM_DATA | MIIM_SUBMENU;
//
//	for ( int iItem = 0; iItem < (int)nItems; iItem++ )
//	{
//		if ( !GetMenuItemInfoT( hMenu, iItem, true, &mii ) )
//			break;
//
//		cMenuItem *miParam = ( cMenuItem * )mii.dwItemData;
//
//		if ( miParam == NULL )
//			continue;
//
//		UINT nID = iItem;// GetMenuItemID( hMenu, iItem );
//
//		if ( towupper( miParam->m_cmdChar ) == towupper( wc ) )
//		{
//			//if ( mii.hSubMenu != NULL )
//			//{
//			//	return MAKELONG( nID, MNC_SELECT );
//			//}
//			//else
//				return MAKELONG( nID, MNC_EXECUTE );
//		}
//	}
//
//	return MAKELONG( 0, MNC_IGNORE );
//}
LONG cMenuItem::MenuChar( HMENU hMenu, wchar_t wc )
{
	UINT nItems = GetMenuItemCount( hMenu );

	for ( int iItem = 0; iItem < (int)nItems; iItem++ )
	{
		cMenuItem *miParam = (cMenuItem *)GetMenuItemParam( hMenu, iItem, true );
		if ( miParam == NULL )
			continue;

		if ( towupper( miParam->m_cmdChar ) == towupper( wc ) )
		{
			return MAKELONG( iItem, MNC_EXECUTE );

			//UINT nID = GetMenuItemID( hMenu, iItem );
			//if ( nID == UINT(-1) )
			//	return MAKELONG( iItem, MNC_SELECT );
			//else
			//	return MAKELONG( iItem, MNC_EXECUTE );
		}
	}

	return MAKELONG( 0, MNC_IGNORE );
}

void cMenuItem::DestroyMenu( HMENU hMenu, bool destroyHandle )
{
	MENUITEMINFOW mii;
	memset( &mii, 0, sizeof( mii ) );
	mii.cbSize = MENUITEMINFO_SIZE_VERSION_400; // sizeof( mii );
	mii.fMask = MIIM_DATA | MIIM_SUBMENU;

	UINT nItems = GetMenuItemCount( hMenu );
	for (int iItem = 0; iItem < (int)nItems; iItem++)
	{
		if ( GetMenuItemInfoT( hMenu, iItem, true, &mii ) )
		{
			cMenuItem *paramItem = ( cMenuItem * )mii.dwItemData;
			if ( paramItem != NULL )
			{
				assert( paramItem->m_hMenu == hMenu );
				if ( paramItem->m_hMenu == hMenu )
					delete paramItem;
			}
		}

		if ( mii.hSubMenu != NULL )
			DestroyMenu( mii.hSubMenu );
	}

	if ( destroyHandle )
		::DestroyMenu( hMenu );
}

bool cMenuItem::DeleteMenu( HMENU hMenu, UINT nPosition, UINT nFlags )
{
	if ( hMenu == NULL )
	{
		assert( false );
		return false;
	}

	MENUITEMINFOW mii;
	memset( &mii, 0, sizeof( mii ) );
	mii.cbSize = MENUITEMINFO_SIZE_VERSION_400; // sizeof( mii );
	mii.fMask = MIIM_DATA | MIIM_SUBMENU;

	if ( GetMenuItemInfoT( hMenu, nPosition, nFlags, &mii ) ) // & MF_BYPOSITION
	{
		cMenuItem *paramItem = ( cMenuItem * )mii.dwItemData;
		if ( paramItem != NULL )
		{
//			assert( paramItem->m_hMenu == hMenu );
//			if ( paramItem->m_hMenu == hMenu )
			
			if ( nFlags == MF_BYCOMMAND )
			{
				assert( paramItem->m_wID == nPosition );
				if ( paramItem->m_wID == nPosition )
					delete paramItem;
			}
			else
			{
				assert( paramItem->m_hMenu == hMenu );
				if ( paramItem->m_hMenu == hMenu )
					delete paramItem;
			}
		}
	}

	if ( mii.hSubMenu != NULL )
		cMenuItem::DestroyMenu( mii.hSubMenu, false );
	
	return ::DeleteMenu( hMenu, nPosition, nFlags ) == TRUE;
}

bool cMenuItem::ExchangeWindowMenu( HWND hWnd, sImageInfo *siiT, int siiQ )
{
	//ODSFunc;

	if ( !isValidOS() )
	{
		ODS_fls( L"!isValidOS()" );
		return false;
	}

	clearTable();

	if ( hWnd == NULL )
	{
		ODS_fls( L"hWnd == NULL" );
		return false;
	}

	HMENU hMenuCurrent = GetMenu( hWnd );

	if ( hMenuCurrent == NULL )
	{
		ODS_fls( L"hMenuCurrent == NULL" );
		return false;
	}

	bool destroySource = false;

	HMENU hMenu = CreateCopy( hMenuCurrent, destroySource, false, siiT, siiQ  );

	if ( hMenu == NULL )
	{
		ODS_fls( L"hMenu == NULL" );
		return false;
	}

	if ( !destroySource ) // Was not destoied at creation
	{
		::DestroyMenu( hMenuCurrent );
	}

	bool ok = SetMenu( hWnd, hMenu ) == TRUE;

	//ODSFuncRes( ok );

	return ok;
}

bool cMenuItem::isValidOS()
{
	return true;

	OSVERSIONINFO ovi = { sizeof(OSVERSIONINFO) };
	::GetVersionEx(&ovi);

	// Windows 2000 or later
	return ovi.dwMajorVersion >= 5;
}

HMENU cMenuItem::CreateCopy( HMENU hSrcMenu, bool destroySource, bool popup, sImageInfo *siiT, int siiQ )
{
	//ODSFunc;

	if ( !isValidOS() )
		return hSrcMenu;

#ifndef _NoTranslation_
	if ( pTranslateDll != NULL )
		pTranslateDll->TranslateMenu( hSrcMenu );
#endif

	HMENU hMenu;
	
	if ( popup )
			hMenu = ::CreatePopupMenu();
	else	hMenu = ::CreateMenu();

	if ( hMenu == NULL )
	{
		//ODSFuncRes( NULL );
		return NULL;
	}

	//ODS( FormatOnceT( L"Source menu count %d", GetMenuItemCount( hSrcMenu ) ) );

	bool ok = CopyItems( hMenu, hSrcMenu, false, siiT, siiQ );

	if ( destroySource )
 		::DestroyMenu( hSrcMenu );

	if ( ok )
	{
		//ODS( FormatOnceT( L"Menu count %d", GetMenuItemCount( hMenu ) ) );
		//ODSFuncRes( DWORD(hMenu) );
		return hMenu;
	}

	::DestroyMenu( hMenu );

	//ODSFuncRes( NULL );
	return NULL;
}

HMENU cMenuItem::LoadPopupMenu( HINSTANCE hInst, UINT id, UINT uSubMenu, sImageInfo *siiT, int siiQ  )
{
    HMENU hmParent = LoadMenu( hInst, MAKEINTRESOURCE(id) );

    if (!hmParent)
    {
		return(NULL);
	}

    HMENU hmPopup = GetSubMenu(hmParent, uSubMenu );
    RemoveMenu(hmParent, uSubMenu, MF_BYPOSITION);
    DestroyMenu(hmParent);

	return cMenuItem::CreateCopy( hmPopup, true, true, siiT, siiQ );
}

int cMenuItem::TrackPopupMenu( HMENU hMenu, UINT fuFlags, int x, int y, HWND hWnd, LPTPMPARAMS lptpm )
{
#ifndef _NoTranslation_
	if ( ( pTranslateDll != NULL ) && ( pTranslateDll->m_flipDialog ) )
		fuFlags |= TPM_RIGHTALIGN;
#endif

	int cmd = TrackPopupMenuEx( hMenu, fuFlags, x, y, hWnd, lptpm );

	cMenuItem::DestroyMenu( hMenu, true );

	return cmd;
}



static bool getImageInfo( HIMAGELIST *phIml, HIMAGELIST *phImlDisabled, int *piImage, int cmdID, sImageInfo *siiT, int siiQ )
{
	for ( int i = siiQ; i--; )
	{
		if ( siiT[i].iCmd == cmdID )
		{
			*phIml   = siiT[i].hIml;
			*piImage = siiT[i].iImage;

			if ( phImlDisabled != NULL )
			{
				*phImlDisabled = siiT[i].hImlDisabled;
			}

			return true;
		}
	}

	return false;
}

bool cMenuItem::InsertItem(
			HMENU hMenu,
			UINT dPos,
			UINT cmdID,
			const wchar_t *text,
			HIMAGELIST hIml, int iImage )
{
	MENUITEMINFOW mii;
	memset( &mii, 0, sizeof( mii ) );
	mii.cbSize = MENUITEMINFO_SIZE_VERSION_400; //sizeof( mii );

	if ( !isValidOS() )
	{
		mii.fType = MFT_STRING;
		mii.fMask = MIIM_ID | MIIM_TYPE | MIIM_DATA;
		mii.wID = cmdID;
		mii.dwTypeData = (LPWSTR)text;
		
		//mii.fMask = MIIM_ID | MIIM_STRING;
		//mii.wID = cmdID;
		//mii.dwItemData = text;

		return InsertMenuItemT( hMenu, dPos, true, &mii ) == TRUE;
	}

	mii.fMask = MIIM_ID | MIIM_TYPE | MIIM_DATA; // MIIM_STATE
	mii.fType = MFT_OWNERDRAW;

	cMenuItem *mi = new cMenuItem( hMenu, hIml, iImage, text, cmdID );
	assert( mi != NULL );

	mii.dwItemData = (dwItemDataType)mi;
	mii.wID = cmdID;

	if ( InsertMenuItemT( hMenu, dPos, true, &mii ) )
		return true;

	assert( false );
	return false;
}

BOOL
_SetMenuText(
    IN HMENU hMenu,
    IN UINT uPosition,
	bool byPosition,
	const wchar_t *text 
	)
{
	UINT nID = uPosition; 
	if ( byPosition )
		nID = GetMenuItemID( hMenu, uPosition );

	UINT nState = GetMenuState( hMenu, nID, MF_BYCOMMAND );
	nState &= ~(MF_BITMAP|MF_SEPARATOR|MF_BYPOSITION);

	nState |= MF_STRING;

	// set menu text
	return ModifyMenuT( hMenu, nID, nState, nID, text );
}

//
//	updateMenuT
//		Deletes old and creates new item with same values because : 
//		The SetMenuItemInfo does not call measure item.
//		So the item size is not chaned with text length.
//
//static
bool cMenuItem::updateMenuT( HMENU hMenu, UINT uPosition, bool byPosition )
{
	MENUITEMINFOW mii;

	memset( &mii, 0, sizeof( mii ) );
	mii.cbSize = MENUITEMINFO_SIZE_VERSION_400; // sizeof( mii );

	DWORD dwId = -1;
	UINT uPos = -1;
	HMENU hSubMenu = NULL;

	if ( byPosition )
	{
		hSubMenu = hMenu;

		mii.fMask = MIIM_ID;
		uPos = uPosition;
		if ( !GetMenuItemInfoT( hSubMenu, uPos, true, &mii ) )
			return false;

		dwId = mii.wID;
	}
	else
	{
		dwId = uPosition;

		hSubMenu = cMenuItem::findMenuByID( hMenu, dwId, NULL, &uPos );

		if ( ( hSubMenu == NULL ) || ( uPos == -1 ) )
			return false;
	}

	mii.fMask = 
		MIIM_TYPE |
		MIIM_ID |
		MIIM_STATE |
		MIIM_DATA |
		MIIM_SUBMENU;

	if ( !GetMenuItemInfoT( hSubMenu, dwId, false, &mii ) )
	{
		int err = GetLastError();
		ODSF( L"GetMenuItemInfoT( %d ) - error %d", dwId, err );
		return false;
	}

	if ( ( mii.fType & MFT_OWNERDRAW ) == 0 )
	{
		assert( false );
		return false;
	}

	bool bIsSeparator = ( mii.fType & MFT_SEPARATOR ) != 0;

	cMenuItem *mi = (cMenuItem *)mii.dwItemData;

	if ( !bIsSeparator && ( mi == NULL ) )
	{
		assert( false );
		return false;
	}

	//memset( &mii, 0, sizeof( mii ) );
	//mii.cbSize = MENUITEMINFO_SIZE_VERSION_400; //sizeof( mii );

	mii.fMask = MIIM_TYPE; // MIIM_STATE | 
	mii.fType = MFT_OWNERDRAW;
	mii.dwItemData = NULL;
		
	if ( bIsSeparator )
	{
		mii.fType |= MFT_SEPARATOR;
	}
	else
	{
		mii.fMask |= 
			MIIM_ID |
			MIIM_STATE |
			MIIM_DATA |
			MIIM_SUBMENU;

		mii.dwItemData = (dwItemDataType)mi;
		mii.wID = dwId;
	}

	::DeleteMenu( hSubMenu, dwId, MF_BYCOMMAND );

	if ( !InsertMenuItemT( hSubMenu, uPos, true, &mii ) )
	{
		assert( false );
		return false;
	}

	return true;
}

// static
bool cMenuItem::SetMenuText( HMENU hMenu, UINT uPosition, bool byPosition, const wchar_t *text )
{
	MENUITEMINFOW mii;
	memset( &mii, 0, sizeof( mii ) );
	mii.cbSize = sizeof( mii );
	mii.fMask = MIIM_DATA;

	if ( !GetMenuItemInfoT( hMenu, uPosition, byPosition, &mii ) )
		return false;

	cMenuItem *paramItem = ( cMenuItem * )mii.dwItemData;

	if ( cMenuItem::isValid( paramItem ) )
	{
		paramItem->SetText( text );

		if ( !SetMenuItemInfoT( hMenu, uPosition, byPosition, &mii ) )
			return false;

		return updateMenuT( hMenu, uPosition, byPosition );
	}

	return SetMenuTextT( hMenu, uPosition, byPosition, text ) == TRUE;
}
	
bool cMenuItem::CopyItems( HMENU dst, HMENU src, bool destroySource, sImageInfo *siiT, int siiQ )
{
	//ODSFunc;

	MENUITEMINFOW mii;
	wchar_t text[ 256 ];

	UINT Q = ::GetMenuItemCount( src );
	UINT i = 0;
	UINT dPos = 0;

	//ODS( FormatOnceT( L"Q %d", Q ) );

	for ( ; i < Q; i++, dPos++ )
	{
		memset( &mii, 0, sizeof( mii ) );
		mii.cbSize = MENUITEMINFO_SIZE_VERSION_400; // sizeof( mii );
		memset( text, 0, sizeof( text ));

		// Note, NT4 does not recognise MIIM_FTYPE flag.

		mii.fMask = 
			MIIM_TYPE |
			MIIM_ID |
			MIIM_STATE |
//			MIIM_STRING |
			MIIM_SUBMENU;

		mii.fType = MFT_STRING;

		mii.dwTypeData = text;
		mii.cch = maxLen( text );

		if ( !GetMenuItemInfoT( src, i, true, &mii ) )
		{
			int err = GetLastError();
			ODSF( L"GetMenuItemInfoT( %d ) - error %d", i, err );
			break;
		}

		if ( ( ( mii.fType & MFT_SEPARATOR ) == 0 ) && ( wcslen( text ) == 0 ) )
		{
			GetMenuStringT( src, i, text, maxLen( text ), MF_BYPOSITION );
		}

		if ( ( mii.fType & MFT_OWNERDRAW ) != 0 )
		{
			assert( false );
			return false;
		}

		bool bIsSeparator = ( mii.fType & MFT_SEPARATOR ) != 0;

		HMENU hSource = mii.hSubMenu;
		UINT curID = mii.wID;

		memset( &mii, 0, sizeof( mii ) );
		mii.cbSize = MENUITEMINFO_SIZE_VERSION_400; //sizeof( mii );

		mii.fMask = MIIM_TYPE; // MIIM_STATE | 
		mii.fType = MFT_OWNERDRAW;
		mii.dwItemData = NULL;
		
		if ( bIsSeparator )
		{
			mii.fType |= MFT_SEPARATOR;
		}
		else
		{
			mii.fMask |= MIIM_ID | MIIM_DATA;
			mii.wID = curID;
			HIMAGELIST hIml;
			HIMAGELIST hImlDisabled;
			int iImage;

			if ( !getImageInfo( &hIml, &hImlDisabled, &iImage, curID, siiT, siiQ ) )
			{
				hIml = NULL;
				iImage = -1;
			}
			
			cMenuItem *mi = new cMenuItem( dst, hIml, iImage, text, curID, hImlDisabled );
			assert( mi != NULL );

			mii.dwItemData = (dwItemDataType)mi;
		}

		bool skipCopy = false;

		if ( !skipCopy && ( hSource != NULL ) )
		{
			mii.hSubMenu = CreateCopy( hSource, false, false, siiT, siiQ );

			if ( mii.hSubMenu != NULL )
				mii.fMask |= MIIM_SUBMENU;

			if ( destroySource )
			{
				::DeleteMenu( src, i, MF_BYPOSITION );
				i--;
				Q--;
			}
		}

		if ( !InsertMenuItemT( dst, dPos, true, &mii ) )
		{
			assert( false );
			return false;
		}
	}

	return true;
}

//
//	cMenuItem::DlgProc
//		It has to be used before messages are cased in WndProc
//		Returns true if pResult has to be returned immediately, otherwise standard windows procedure works normaly 
//static 
bool cMenuItem::WndProc( LRESULT *pResult, HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	if ( pResult == NULL )
	{
		assert( false );
		return false;
	}

	switch ( uMsg )
	{
		case WM_MEASUREITEM:
		{
			//ODSFuncS( L"WM_MEASUREITEM" );
			MEASUREITEMSTRUCT *pmi = (MEASUREITEMSTRUCT *)lParam;
			*pResult = (LRESULT)cMenuItem::MeasureItem( hDlg, pmi );
			return true;
		}

		case WM_DRAWITEM :
		{
			//ODSFuncS( L"WM_DRAWITEM" );
			DRAWITEMSTRUCT *pdi = ( DRAWITEMSTRUCT * )lParam;
			*pResult = (LRESULT)cMenuItem::DrawItem( hDlg, pdi );
			return true;
		}

		case WM_MENUCHAR:
		{
			//ODSFuncS( L"WM_MENUCHAR" );
			HMENU hMenu = (HMENU)lParam;
			WORD wc = LOWORD( wParam );
			*pResult = (LRESULT)cMenuItem::MenuChar( hMenu, wc );
			return true;
		}

		default:
			break;
	}

	return false;
}

//
//	cMenuItem::DlgProc
//		It has to be used before messages are cased in RealDlgProc
//		Returns true if pResult has to be returned immediately, otherwise standard dialog procedure works normaly 
//
bool cMenuItem::DlgProc( LRESULT *pResult, HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	if ( pResult == NULL )
	{
		assert( false );
		return false;
	}

	if ( WndProc( pResult, hDlg, uMsg, wParam, lParam ) )
	{
		//SetWindowLong( hDlg, DWL_MSGRESULT, *pResult );
		SetWindowLongPtrT( hDlg, DWLP_MSGRESULT, *pResult );
		return true;
	}

	switch ( uMsg )
	{
		case WM_INITDIALOG:
			cMenuItem::ExchangeWindowMenu( hDlg );
			break;

		case WM_DESTROY:
			cMenuItem::DestroyWindowMenu( hDlg );
			break;

		default:
			break;
	}

	return false;
}

void cMenuItem::GetSystemSettings()
{
	cMenuItem::OnSize( NULL, NULL, 0, 0, 0 );

	NONCLIENTMETRICS info;
	memset( &info, 0, sizeof( info ) );

	//
	// 17.10.2011
	// windows XP does not work with the struct defide for (WINVER >= 0x0600)
	//
	//info.cbSize = sizeof(info);
	info.cbSize = sizeofAfter( NONCLIENTMETRICS, lfMessageFont );

	SetLastError( 0 );

	BOOL bRet = ::SystemParametersInfo( SPI_GETNONCLIENTMETRICS, sizeof(info), &info, 0 );
	if ( !bRet )
	{
		ODS_fls( _FormatT( L"Assertion! SystemParametersInfo returns false, lastError %s" , cWCStr().GetSystemError() ) );
		assert( false );
	}
	else
	{
		LOGFONT logfont = { 0 };

		if( m_hFontMenu != NULL )
			GetObject( m_hFontMenu, sizeof( LOGFONT ), &logfont );

		if(	logfont.lfHeight != info.lfMenuFont.lfHeight ||
			logfont.lfWidth != info.lfMenuFont.lfWidth ||
			logfont.lfEscapement != info.lfMenuFont.lfEscapement ||
			logfont.lfOrientation != info.lfMenuFont.lfOrientation ||
			logfont.lfWeight != info.lfMenuFont.lfWeight ||
			logfont.lfItalic != info.lfMenuFont.lfItalic ||
			logfont.lfUnderline != info.lfMenuFont.lfUnderline ||
			logfont.lfStrikeOut != info.lfMenuFont.lfStrikeOut ||
			logfont.lfCharSet != info.lfMenuFont.lfCharSet ||
			logfont.lfOutPrecision != info.lfMenuFont.lfOutPrecision ||
			logfont.lfClipPrecision != info.lfMenuFont.lfClipPrecision ||
			logfont.lfQuality != info.lfMenuFont.lfQuality ||
			logfont.lfPitchAndFamily != info.lfMenuFont.lfPitchAndFamily ||
			lstrcmp(logfont.lfFaceName, info.lfMenuFont.lfFaceName) != 0)
		{
			HFONT hFontMenu = ::CreateFontIndirect(&info.lfMenuFont);

			if ( hFontMenu == NULL )
			{
				ODS_fls( _FormatT( L"Assertion! CreateFontIndirect returns NULL, lastError %d" , GetLastError() ) );
				assert( hFontMenu != NULL );
			}
			else
			{
				if( m_hFontMenu != NULL)
					::DeleteObject( m_hFontMenu );

				m_hFontMenu = hFontMenu;
//				SetFont(m_fontMenu);
//				AddStrings(_T("NS\0"));   // for proper item height
//				AutoSize();
			}
		}
	}

	//// check if we need extra spacing for menu item text
	//CWindowDC dc(m_hWnd);
	//HFONT hFontOld = dc.SelectFont(m_fontMenu);
	//RECT rcText = { 0, 0, 0, 0 };
	//dc.DrawText(_T("\t"), -1, &rcText, DT_SINGLELINE | DT_LEFT | DT_VCENTER | DT_CALCRECT);
	//if((rcText.right - rcText.left) < 4)
	//{
	//	::SetRectEmpty(&rcText);
	//	dc.DrawText(_T("x"), -1, &rcText, DT_SINGLELINE | DT_LEFT | DT_VCENTER | DT_CALCRECT);
	//	m_cxExtraSpacing = rcText.right - rcText.left;
	//}
	//else
	//{
	//	m_cxExtraSpacing = 0;
	//}
	//dc.SelectFont(hFontOld);

//	// get Windows version
//	OSVERSIONINFO ovi = { sizeof(OSVERSIONINFO) };
//	::GetVersionEx(&ovi);
//
//	// query keyboard cues mode (Windows 2000 or later)
//	if(ovi.dwMajorVersion >= 5)
//	{
//#ifndef SPI_GETKEYBOARDCUES
//		const UINT SPI_GETKEYBOARDCUES = 0x100A;
//#endif //!SPI_GETKEYBOARDCUES
//		BOOL bRetVal = TRUE;
//		bRet = ::SystemParametersInfo(SPI_GETKEYBOARDCUES, 0, &bRetVal, 0);
//		m_bUseKeyboardCues = (bRet && !bRetVal);
//		m_bAllowKeyboardCues = true;
//		ShowKeyboardCues(!m_bUseKeyboardCues);
//	}
//
//	// query flat menu mode (Windows XP or later)
//	if((ovi.dwMajorVersion == 5 && ovi.dwMinorVersion >= 1) || (ovi.dwMajorVersion > 5))
//	{
//#ifndef SPI_GETFLATMENU
//		const UINT SPI_GETFLATMENU = 0x1022;
//#endif //!SPI_GETFLATMENU
//		BOOL bRetVal = FALSE;
//		bRet = ::SystemParametersInfo(SPI_GETFLATMENU, 0, &bRetVal, 0);
//		m_bFlatMenus = (bRet && bRetVal);
//	}
}

//
//	findMenuByID
//		Walks parent menu and finds the menu contains item with corresponded id.
//	If item has submenu so it's identifier is unpredictable.
//	The function allows to find the parent menu and the parent menu ID ( the ID may be used for icon detection ) 
//	by identifier of any submenu item.
//
HMENU cMenuItem::findMenuByID( HMENU hMenuParent, UINT idFind, UINT *pParentId, UINT *pPosition )
{
	MENUITEMINFOW mii;
	wchar_t text[ 256 ];

	UINT Q = ::GetMenuItemCount( hMenuParent );
	UINT i = 0;
	UINT dPos = 0;

	for ( ; i < Q; i++, dPos++ )
	{
		memset( &mii, 0, sizeof( mii ) );
		mii.cbSize = MENUITEMINFO_SIZE_VERSION_400; //sizeof( mii );
		memset( text, 0, sizeof( text ));

		mii.fMask = 
			MIIM_ID |
			MIIM_STRING |
			MIIM_DATA | // For debug purpouses..
			MIIM_SUBMENU;

		mii.dwTypeData = text;
		mii.cch = maxLen( text );

		if ( !GetMenuItemInfoT( hMenuParent, i, true, &mii ) )
		{
			if ( pPosition != NULL )
				*pPosition = -1;

			break;
		}

		if ( ( mii.fType & MFT_SEPARATOR ) != 0 )
			continue;

		if ( mii.wID == idFind )
		{
			if ( pPosition != NULL )
				*pPosition = dPos;

			return hMenuParent;
		}

		if ( mii.hSubMenu != NULL )
		{
			HMENU hMenu = findMenuByID( mii.hSubMenu, idFind, pParentId, pPosition );

			if ( hMenu != NULL )
			{
				if ( ( hMenu == mii.hSubMenu ) && ( pParentId != NULL ) )
					*pParentId = mii.wID;

				return hMenu;
			}
		}
	}

	return NULL;
}



//
//	findMenuBySubMenu
//	Walks parent menu and finds the menu contains item with corresponded submenu. 
//	If found we may use the pMenu and pId to modify the menu item.
//	Sample:
//		if menu with corresponded identefier is found by findMenuByID
//		the function may find the menu item has the submenu 
//
bool cMenuItem::findMenuBySubMenu( HMENU *pMenu, int *pId, HMENU hMenuParent, HMENU hSubmenu )
{
	MENUITEMINFOW mii;
	wchar_t text[ 256 ];

	UINT Q = ::GetMenuItemCount( hMenuParent );
	UINT i = 0;
	UINT dPos = 0;

	for ( ; i < Q; i++, dPos++ )
	{
		memset( &mii, 0, sizeof( mii ) );
		mii.cbSize = MENUITEMINFO_SIZE_VERSION_400; // sizeof( mii );
		memset( text, 0, sizeof( text ));

		mii.fMask = 
			MIIM_ID |
			MIIM_STRING |
			MIIM_SUBMENU;

		mii.dwTypeData = text;
		mii.cch = maxLen( text );

		if ( !GetMenuItemInfoT( hMenuParent, i, true, &mii ) )
			break;

		if ( ( mii.fType & MFT_SEPARATOR ) != 0 )
			continue;

		if ( mii.hSubMenu == hSubmenu )
		{
			if ( pMenu != NULL ) *pMenu = hMenuParent;
			if ( pId != NULL ) *pId = mii.wID;
			return true;
		}

		if ( mii.hSubMenu != NULL )
		{
			if ( findMenuBySubMenu( pMenu, pId, mii.hSubMenu, hSubmenu ) )
				return true;
		}
	}

	return false;
}

#ifndef _NoTranslation_

//
//	fillLanguageMenu
//	 1. Finds menu with corresponded id.
//	 2. Enumerates available translations and fills the language menu.
//		Uses alternatePath to get translattion instaled with main module.
//		F.e. BCArchive reads it's own translation and the BestCrypt translation as alternativePath.
//	 3. Checks on current language by updateLanguageMenu
//
//static 
bool cMenuItem::fillLanguageMenu( HMENU hMenuParent, int idLanguageFirst, cStringArrayT *saPaths, const wchar_t *alternatePath, const wchar_t *modulePath )
{
	// 1.
	HMENU hMenu = findMenuByID( hMenuParent, idLanguageFirst );

	if ( hMenu == NULL )
	{
//		assert( false );
		return false;
	}

	// 2.

	cWCStr wcsbf;

	if ( modulePath == NULL )
	{
		if ( wcsbf.AddLength( MAX_PATH ) == NULL )
		{
			assert( false );
			return false;
		}

		GetModuleFileNameT( NULL, wcsbf.m_szBody, wcsbf.m_len );
		modulePath = wcsbf.m_szBody;
	}

	int pos = splitPath( modulePath );
	const wchar_t *exeName = modulePath + pos;

	wchar_t folderName[ MAX_PATH ];
	cTranslateDll::getModuleTranslationFolder( folderName, modulePath );

	cStringArrayT saNames, saTmp;

	if ( saPaths == NULL )
		saPaths = &saTmp;
	else
		saPaths->freeTable();

	cTranslateDll::enumLangFiles( saPaths, &saNames, folderName );

	if ( ( alternatePath != NULL ) && ( wcslen( alternatePath ) != 0 ) && ( FileExistsT( alternatePath ) ) )
	{
		wchar_t altFolderName[ MAX_PATH ];

		if ( cTranslateDll::getModuleTranslationFolder( altFolderName, alternatePath ) )
		{
			if ( wcscmp( altFolderName, folderName ) != 0 )
				cTranslateDll::enumLangFiles( saPaths, &saNames, altFolderName, exeName );
		}
	}


	MENUITEMINFOW mii;
	memset( &mii, 0, sizeof( mii ) );
	mii.cbSize = MENUITEMINFO_SIZE_VERSION_400; // sizeof( mii );
	mii.fMask = MIIM_ID | MIIM_TYPE;
	mii.fType = MFT_RADIOCHECK | MFT_STRING;
	//mii.fMask = MIIM_ID | MIIM_STRING | MIIM_FTYPE;
	//mii.fType = MFT_RADIOCHECK;

	int err;
	int i = 0;

	for ( ; i < saNames.Q; i++ )
	{
		UINT pos = i + 1;
		mii.wID = idLanguageFirst + pos;
		mii.dwTypeData = saNames.T[ i ];

		InsertMenuItemT( hMenu, pos, true, &mii );

		err = GetLastError();
	}

//#ifdef _DEBUG
//	// Adds Choose translation folder
//	{
//		UINT pos = i + 1;
//		mii.wID = idLanguageFirst + cTranslateDll::maxLangs;
//		mii.dwTypeData = L"Choose folder";
//
//		InsertMenuItemT( hMenu, pos, true, &mii );
//	}
//#endif

	// 3. 
	//return updateLanguageMenu( hMenu, idLanguageFirst, idLanguageFirst );
	return true;
}

// static 
bool cMenuItem::updateLanguageMenu( HMENU hMenuParent, int idLanguageFirst, int idCur )
{
	// 1.
	HMENU hMenu = findMenuByID( hMenuParent, idLanguageFirst );

	if ( hMenu == NULL )
	{
//		assert( false );
		return false;
	}

	UINT Q = ::GetMenuItemCount( hMenu );
	UINT i = 0;

	for ( ; i < Q; i++ )
	{
		UINT id = GetMenuItemID( hMenu, i );

		cMenuItem::SetRadio( hMenu, id ); 

		UINT chk = ( id == idCur ? MF_CHECKED : MF_UNCHECKED );
		
		chk |= MF_BYCOMMAND;

		CheckMenuItem( hMenu, id, chk );
	}

	return true;
}

#endif // #ifndef _NoTranslation_
