#pragma once

#include <windows.h>
#include <commctrl.h>

//
//	Class cMenuItem
//		Allows to use owner draw menu
//
//	1. 	Initialization sample
//
//		MENUITEMINFOW mii;
//		memset( &mii, 0, sizeof( mii )); 
//		mii.cbSize = sizeof( mii );
//
//		mii.fMask = MIIM_ID | MIIM_TYPE | MIIM_DATA;
//		mii.fType = MFT_OWNERDRAW;
//		mii.wID = id;
//		mii.dwItemData = (ULONG_PTR)new cMenuItem( hMenu, hIml, 0, menuItemText );
//
//		bool res = InsertMenuItemT( hMenu, pos, TRUE, &mii ) != FALSE;
//
//		Separator sample:
//
//		mii.fMask = MIIM_TYPE;
//		mii.fType = MFT_OWNERDRAW | MFT_SEPARATOR;
//		bool res = InsertMenuItemT( hMenu, pos, TRUE, &mii ) != FALSE;
//
//
//	2. Required messages at main window
//
//		case WM_MEASUREITEM:
//		{
//			MEASUREITEMSTRUCT *pmi = (MEASUREITEMSTRUCT *)lParam;
//			return cMenuItem::MeasureItem( hwnd, pmi );
//		}
//
//		case WM_DRAWITEM :
//		{
//			DRAWITEMSTRUCT *pdi = ( DRAWITEMSTRUCT * )lParam;
//			return cMenuItem::DrawItem( hwnd, pdi );
//		}
//
//		case WM_MENUCHAR:
//		{
//			HMENU hMenu = (HMENU)lParam;
//			WORD wc = LOWORD( wParam );
//			ULONG res = cMenuItem::MenuChar( hMenu, wc );
//			if ( res != 0 )
//			{
//				SetWindowLong( hwnd, DWL_MSGRESULT, res );
//				return true;
//			}
//
//			return false;
//		}
//
//		Note, the messages may be intercepted by cMenuItem::DlgProc function
//
//	3. Destroy menu static function required for deallocation all data.
//
//		DestroyMenu( HMENU hMenu );
//		cMenuItem::DestroyWindowMenu( HWND hWnd );
//
//	4. Use the cMenuItem::ExchangeWindowMenu( HWND hWnd ) static function 
//		for exchenging existing main menu of a window
//
//	5. Use the cMenuItem::CreateCopy( HMENU hSrcMenu, bool destroySource = true, bool popup = true ) 
//		for temporary popup menus
//

class cStringArrayT;

class cSTDImage
{
public:

	bool m_ok;

	HBITMAP m_chkBmp;
	HBITMAP m_dotBmp;
	HBITMAP m_dotSelectedBmp;

	cSTDImage();
	~cSTDImage();

	bool CreateImages();
};

struct sImageInfo
{
	int iCmd;			// menu item cmdID
	HIMAGELIST hIml;	// existing image list
	HIMAGELIST hImlDisabled;	// existing disabled image list
	int iImage;			// number of image from the image list corresponded for the cmdID
	
	sImageInfo()
	{
		memset( this, 0, sizeof( *this ) );
	}

	void Set( int piCmd, HIMAGELIST phIml, int piImage, HIMAGELIST phImlDisabled = NULL )
	{
		iCmd = piCmd;
		hIml = phIml;
		hImlDisabled = phImlDisabled;
		iImage = piImage;
	}
};

class cMenuItem
{
public:

	int m_size;
	DWORD m_sig;
	enum { eSig = 0xBCAA };

	static bool isValid( cMenuItem *pThis )
	{
		if ( pThis == NULL ) return false;
		if ( pThis->m_sig != eSig ) return false;
		if ( pThis->m_size < sizeof( cMenuItem )  ) return false;

		return true;
	}

	static cSTDImage m_stdImgs;

	static int m_iconSize;
	static HFONT m_hFontMenu;
	static void GetSystemSettings();
	static SIZE GetMenuTextSize( HWND hWnd, UINT fType, UINT fState, const wchar_t *text );
	static int GetItemHeight( int textHeight = 0 );

	HMENU m_hMenu;
	HIMAGELIST m_hIml;
	HIMAGELIST m_hImlDisabled;
	int m_iImage;
	UINT m_wID;
	wchar_t *m_text;
	wchar_t m_cmdChar;
	bool m_bRadio;

	static LONG MenuChar( HMENU hMenu, wchar_t wc );
	static bool MeasureItem( HWND hWnd, MEASUREITEMSTRUCT *pmi );
	static bool DrawItem( HWND hWnd, DRAWITEMSTRUCT *pdi );
	static void DestroyMenu( HMENU hMenu, bool destroyHandle = true );
	static void SetRadio( HMENU hMenu, int id );

	static bool DeleteMenu( HMENU hMenu, UINT nPosition, UINT nFlags );

	//
	//	OnSize
	//		Required to update bound setings for last item in main menu
	//		The updating is required for main menu bar completely filling 
	//		in same style and collor as the menu items
	//
	static void OnSize( HWND hWnd, HMENU hMenu, UINT nType, int cx, int cy );
//	static void OnSize( UINT nType, int cx, int cy );
	//
	//	Allocates new space for and copy the text
	//
	bool SetText( const wchar_t *text );

	cMenuItem( HMENU hMenu, HIMAGELIST hIml, int iImage, const wchar_t *text, UINT wID = 0, HIMAGELIST hImlDisabled = NULL);
	~cMenuItem();

	//
	//	Functions required for exchanging existing menu
	//
	static bool ExchangeWindowMenu( HWND hWnd, sImageInfo *siiT = NULL, int siiQ = 0 );
	static HMENU LoadPopupMenu( HINSTANCE hInst, UINT id, UINT uSubMenu, sImageInfo *siiT = NULL, int siiQ = 0 );
	static int TrackPopupMenu( HMENU hmenu, UINT fuFlags, int x, int y, HWND hwnd, LPTPMPARAMS lptpm );
	static bool isValidOS();

	static HMENU CreateCopy( HMENU hSrcMenu, bool destroySource = true, bool popup = false, sImageInfo *siiT = NULL, int siiQ = 0  );
	static bool CopyItems( HMENU dst, HMENU src, bool destroySource, sImageInfo *siiT = NULL, int siiQ = 0  );

	static bool InsertItem(
			HMENU hMenu,
			UINT dPos,
			UINT cmdID,
			const wchar_t *text,
			HIMAGELIST hIml = NULL,
			int iImage = -1
		);

	static void DestroyWindowMenu( HWND hWnd )
	{
		DestroyMenu( GetMenu( hWnd ) );
		SetMenu( hWnd, NULL );
	}

	static bool SetMenuText( HMENU hMenu, UINT uPosition, bool byPosition, const wchar_t *text );

	//
	//	updateMenuT
	//		Deletes old and reates new item with same values because : 
	//		The SetMenuItemInfo does not call measure item.
	//		So the item size is not chaned with text length.
	//
	static bool updateMenuT( HMENU hMenu, UINT uPosition, bool byPosition );

	//
	//	cMenuItem::DlgProc
	//		It has to be used at the beginning of real DlgProc of processed dialog.
	//			1. If main dialog menu is accesed by GetMenu function, it will be exchanged to owner draw
	//				at WM_INITDIALOG and destroied at WM_DESTORY.
	//			2. It catchs and traslates all required owner draw menu messages
	//				WM_MEASUREITEM, WM_DRAWITEM, WM_MENUCHAR
	//
	//		Returns true if pResult has to be returned immediately,
	//		otherwise standard dialog procedure works normaly 
	//
	//	Usage sample:
	//		LRESULT cDlgView::RealDlgProc( UINT uMsg, WPARAM wParam, LPARAM lParam )
	//		{
	//			LRESULT result;
	//
	//			if ( cMenuItem::DlgProc( &result, m_hDlg, uMsg, wParam, lParam ) )
	//				return result;
	//
	//			switch(uMsg)
	//			{
	//				case WM_INITDIALOG:
	//				....
	//			}
	//		}
	//
	static bool DlgProc( LRESULT *pResult, HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam );
	static bool WndProc( LRESULT *pResult, HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam );

	// Language transaltion functions.

	//
	//	fillLanguageMenu
	//
	//	 1. Finds menu with corresponded id.
	//
	//	 2. Enumerates available translations to saPaths string table with translation folder names
	//		See !TranslationStrategy.txt
	//
	//	 3. Fills the language menu.
	//		Every new added item has:
	//			a. name as language neutral name,
	//			b. it's wID is idLanguageFirst + number in table + 1
	//				Note, the default English is really first and it's wID is idLanguageFirst.
	//
	//	 4. Checks on current language by updateLanguageMenu
	//
	static bool fillLanguageMenu( HMENU hMenu, int idLanguageFirst, cStringArrayT *saPaths = NULL, const wchar_t *alternatePath = NULL, const wchar_t *modulePath = NULL );

	//
	// updateLanguageMenu
	//	Checks on the radio button menu with current translation.
	//
	static bool updateLanguageMenu( HMENU hMenuParent, int idLanguageFirst, int idCur );

	//
	//	findMenuByID
	//		Walks parent menu and finds the menu contains item with corresponded id.
	//	If item has submenu so it's identifier is unpredictable.
	//	The function allows to find the parent menu and the parent menu ID ( the ID may be used for icon detection ) 
	//	by identifier of any submenu item.
	//	Also it fill the *pPosition as position of menu item with ID in parent menu.
	//
	static HMENU findMenuByID( HMENU hMenuParent, UINT idFind, UINT *pParentId = NULL, UINT *pPosition = NULL );
	//
	//	findMenuBySubMenu
	//	Walks parent menu and finds the menu contains item with corresponded submenu. 
	//	If found we may use the pMenu and pId to modify the menu item.
	//	Sample:
	//		if menu with corresponded identefier is found by findMenuByID
	//		the function may find the menu item has the submenu 
	//
	static bool findMenuBySubMenu( HMENU *pMenu, int *pId, HMENU hMenuParent, HMENU hSubmenu );
};

extern sImageInfo *g_siiT;
extern int g_siiQ;

#ifdef CDSIZEOF_STRUCT
#define MENUITEMINFO_SIZE_VERSION_400  CDSIZEOF_STRUCT(MENUITEMINFOA,cch)
#define dwItemDataType ULONG_PTR
#else
#define MENUITEMINFO_SIZE_VERSION_400  sizeof( MENUITEMINFOA )
#define dwItemDataType ULONG
#endif
