#include <windows.h>
#include <stdio.h>
//#include <psapi.h>
#include "processInfo.h"

static bool GetProcessName( DWORD processID, wchar_t *bf, int bfLen )
{
	// Get a handle to the process.

	HANDLE hProcess = OpenProcess( PROCESS_QUERY_INFORMATION |
								   PROCESS_VM_READ,
								   FALSE, processID );

	// Get the process name.

	if ( NULL == hProcess )
		return false;

	HMODULE hMod;
	DWORD cbNeeded;
	DWORD err = 0;

	DWORD dwRes = false;

	if ( cPSAPI_Dll::getStatic().EnumProcessModules( hProcess, &hMod, sizeof(hMod), &cbNeeded) )
	{
		dwRes = cPSAPI_Dll::getStatic().GetModuleBaseNameW( hProcess, hMod, bf, bfLen );
	}
	else
		err = GetLastError();

	CloseHandle( hProcess );

	return dwRes != 0;
}

bool GetProcessIdByName( const wchar_t *processName, DWORD *pDWID )
{
	if ( cPSAPI_Dll::getStatic().error )
		return false;

	// Get the list of process identifiers.

	DWORD aProcesses[1024], cbNeeded, cProcesses;
	unsigned int i;

	if ( !cPSAPI_Dll::getStatic().EnumProcesses( aProcesses, sizeof(aProcesses), &cbNeeded ) )
		return false;

	// Calculate how many process identifiers were returned.

	cProcesses = cbNeeded / sizeof(DWORD);

	// Finds proces by name.

	wchar_t bf[ MAX_PATH ];

	for ( i = 0; i < cProcesses; i++ )
	{
		if ( !GetProcessName( aProcesses[i], bf, sizeof( bf )/sizeof( wchar_t) ) )
			continue;

		if ( _wcsicmp( processName, bf ) == 0 )
		{
			if ( pDWID != NULL )
				*pDWID = aProcesses[i];

			return true;
		}
	}

	return false;
}

struct sGPIBNParam
{
	const wchar_t *processName;
	DWORD *pDWID;
};

DWORD WINAPI static_GetProcessIdByName( void *param )
{
	sGPIBNParam *sp = (sGPIBNParam *)param;

	return GetProcessIdByName( sp->processName, sp->pDWID );
}

bool GetProcessIdByNameI( const wchar_t *processName, DWORD *pDWID )
{
	sGPIBNParam  sp;
	sp.processName = processName;
	sp.pDWID = pDWID;

	HANDLE hThread = CreateThread( 0, 0, static_GetProcessIdByName, &sp, 0, NULL );

	if ( hThread == NULL )
		return false;

	DWORD dwRes = WaitForSingleObject( hThread, INFINITE );

	DWORD exitCode = 0;

	if ( dwRes == WAIT_OBJECT_0 )
	{
		GetExitCodeThread( hThread, &exitCode );
	}

	CloseHandle( hThread );

	return dwRes != 0;
}

BOOL CProcessInfo::FindModule( const wchar_t *str ) 
{
	if (!hProc) {
		return 0;
	}

	DWORD	rc, i, sz, ret, len, off, x;
	wchar_t	*path = NULL;
	HMODULE *hmod = NULL;

#ifdef  _ODSM
	ODSF(L"Looking for '%s'", str);
#endif	
	sz = 4096 * sizeof(HMODULE);
	hmod = (HMODULE *)malloc(sz);
	if (!hmod) {
		return FALSE;
	}

	rc = cPSAPI_Dll::getStatic().EnumProcessModules(hProc, hmod, sz, &ret);
	if (!rc) {
#ifdef  _ODSM
	DWORD err = GetLastError();
	ODSF(L"EnumProcessModules failed: %d, sz=%d(%d), ret=%d(%d)", err);
#endif	
		free(hmod);
		return FALSE;
	}

	path = (wchar_t *)malloc( MAX_PATH * sizeof( wchar_t ));
	if (!path) {
		free(hmod);
		return FALSE;
	}

	len = wcslen(str);
	ret = ret / sizeof(HMODULE);
	x = FALSE;
	for (i = 0; i < ret; i++) {
		rc = cPSAPI_Dll::getStatic().GetModuleFileNameExW(hProc, hmod[i], path, MAX_PATH);
		if (!rc) {
#ifdef  _ODSM
		off = GetLastError();
		ODSF(L"Error %d while getting module name: ", off);
#endif	
			continue;
		}
		off = wcslen(path);

#ifdef  _ODSM
		ODSF(L"Module: '%s'(%d) vs '%s' (%d)", path, off, str, len);
#endif			
		if (len > off) {
			continue;
		}
		off -= len;

		if ( _wcsicmp( path+off, str ) == 0 ) {
			x = TRUE;
			break;
		}
	}

	free(hmod);
	free(path);
	return x;
}

//int TerminateProcessByName( const wchar_t *processName, DWORD exitCode )
//{
//	DWORD dwId = -1;
//	if ( !GetProcessIdByName( processName, &dwId ) )
//		return ERROR_FILE_NOT_FOUND;
//
//	HANDLE hProcess = OpenProcess( PROCESS_TERMINATE, 0, dwId );
//
//	if ( hProcess == NULL )
//		return GetLastError();
//
//	BOOL ok = TerminateProcess( hProcess, exitCode );
//	
//	CloseHandle( hProcess );
//
//	if ( !ok )
//		return GetLastError();
//
//	return ERROR_SUCCESS;
//}