#include <windows.h>
#include <winnt.h>
//#include "processorTypes.h"
#include "kernelDll.h"

static int bitcount(unsigned int a)
{
   a = (a & 0x55555555) + ((a >>  1) & 0x55555555); // max 2
   a = (a & 0x33333333) + ((a >>  2) & 0x33333333); // max 4
   a = (a + (a >> 4)) & 0x0f0f0f0f; // max 8 per 4, now 8 bits
   a = (a + (a >> 8)); // max 16 per 8 bits
   a = (a + (a >> 16)); // max 32 per 8 bits
   return a & 0xff;
}

//
//	PT_getProcessorCount
//		Returns quantity of processors available for current process.
//
int PT_getProcessorCount()
{
	// MDSN Remarks
	//	A process affinity mask is a bit vector in which each bit represents the processors that a process is allowed to run on. 
	//	A system affinity mask is a bit vector in which each bit represents the processors that are configured into a system.

	unsigned long proc,sys;
	GetProcessAffinityMask(GetCurrentProcess(), &proc, &sys);
	return bitcount(proc);
}

DWORD PT_getSystemProcessorType( WORD *pArchitecture )
{
	static DWORD processorType = 0; // one of PROCESSOR_INTEL_386 ...
	static WORD processorArchitecture = 0; // one of PROCESSOR_ARCHITECTURE_INTEL ...

	if ( processorType == 0 )
	{
		SYSTEM_INFO sysInfo;
		cKernelDll::getStatic().GetNativeSystemInfo( &sysInfo );
		processorType = sysInfo.dwProcessorType;
		processorArchitecture = sysInfo.wProcessorArchitecture;
	}

	if ( pArchitecture != NULL )
		*pArchitecture = processorArchitecture;

	return processorType;
}


#ifndef IMAGE_FILE_MACHINE_AMD64
#define IMAGE_FILE_MACHINE_AMD64             0x8664  // AMD64 (K8)
#endif

bool PT_checkCompatibility( DWORD dwProcessorType, WORD wImageFileHeaderMachine )
{
	if ( ( wImageFileHeaderMachine == IMAGE_FILE_MACHINE_I386 ) ||
		 ( wImageFileHeaderMachine == IMAGE_FILE_MACHINE_UNKNOWN ) // w95 vxd file header has the machine type
		 )
	{
		return 
			( dwProcessorType == PROCESSOR_INTEL_386 ) ||
			( dwProcessorType == PROCESSOR_INTEL_486 ) ||
			( dwProcessorType == PROCESSOR_INTEL_PENTIUM );
	}
	else
	if ( wImageFileHeaderMachine == IMAGE_FILE_MACHINE_IA64 )
	{
		return ( dwProcessorType == PROCESSOR_INTEL_IA64 );
	}
	else
	if ( wImageFileHeaderMachine == IMAGE_FILE_MACHINE_AMD64 )
	{
		return ( dwProcessorType == PROCESSOR_AMD_X8664 );
	}

	return false;
}

static bool ReadBytes( HANDLE hFile, LPVOID buffer, DWORD  size )
{
    DWORD bytes = 0;

	if (!ReadFile(hFile, buffer, size, &bytes, NULL))
		return false;

	return (size == bytes);
}

WORD PT_getFileMachineType( HANDLE hImage )
{
	IMAGE_DOS_HEADER	image_dos_header;
	IMAGE_FILE_HEADER	image_file_header;

	if ( !ReadBytes( hImage, &image_dos_header, sizeof(IMAGE_DOS_HEADER)) )
		return 0;

    if ( IMAGE_DOS_SIGNATURE != image_dos_header.e_magic )
		return 0;

    DWORD newOffset = SetFilePointer( hImage, image_dos_header.e_lfanew, NULL, FILE_BEGIN );
	if ( newOffset == 0xFFFFFFFF )
		return 0;

	ULONG ntSignature;
    ReadBytes (hImage, &ntSignature, sizeof(ULONG));

    if ( ( IMAGE_NT_SIGNATURE != ntSignature) &&
		 ( IMAGE_VXD_SIGNATURE != ntSignature )
		)
		return 0;

	if ( !ReadBytes(hImage, &image_file_header, IMAGE_SIZEOF_FILE_HEADER ) ) 
		return 0;

	return image_file_header.Machine;
}

WORD PT_getFileMachineType( const char *fileName )
{
	HANDLE hImage = 
		CreateFileA(
				fileName,
				GENERIC_READ,
				FILE_SHARE_READ,
				NULL,
				OPEN_EXISTING,
				FILE_ATTRIBUTE_NORMAL,
				NULL);

    if (INVALID_HANDLE_VALUE == hImage)
		return -1;

	WORD type = PT_getFileMachineType( hImage );

	CloseHandle( hImage );

	return type;
}

