#include "ptbFuncs.h"
#include "cMBStr.h"

int readIntFromBuffer( BYTE *buff, int size, int *pPos )
{
	int val;

	if ( ( size - *pPos ) < sizeof ( val ) )
		return -1;

	val = *(int *)( buff + *pPos );
	*pPos += sizeof( val );

	return val;
}

bool readBytesFromBuffer( BYTE *buff, int size, int *pPos, BYTE *bfOut, int bfOutSize )
{
	int len = readIntFromBuffer( buff, size, pPos );

	if ( len != bfOutSize )
	{
		ODS_f( L"Assertion: len > bfOutSize" );
		ODS_fls( L"Assertion!" );
		return false;
	}

	if ( (size - *pPos ) < len )
	{
		ODS_f( L"Assertion: (size - *pPos ) < len" );
		ODS_fls( L"Assertion!" );
		return false;
	}

	memcpy( bfOut, buff + *pPos, len );
	*pPos += len;

	return true;
}

char *readStringFromBufferA( BYTE *buff, int size, int *pPos )
{
	if ( *pPos >= size )
		return NULL;

	int len = readIntFromBuffer( buff, size, pPos );

	if ( len < 0 )
	{
		ODS_fls( L"Assertion!" );
		return NULL;
	}

	if ( (size - *pPos ) < len )
		return NULL;

	char *str = new char[ len + 1 ];

	if ( str == NULL )
		return NULL;

	memcpy( str, buff + *pPos, len );
	str[ len ] = '\0';
	*pPos += len;

	return str;
}

wchar_t *readStringFromBuffer( BYTE *buff, int size, int *pPos )
{
	if ( *pPos >= size )
		return NULL;

	int len = readIntFromBuffer( buff, size, pPos );

	if ( len < 0 )
	{
		ODS_fls( L"Assertion!" );
		return NULL;
	}

	if ( (size - *pPos ) < len )
		return NULL;

	wchar_t wcs[ 1024 ] = L"";

	getWCStr( (const char*)buff + *pPos, wcs, maxLen( wcs ), CP_UTF8, len );

	*pPos += len;

	return newString( wcs );
}

bool writeBytesToPTB( cPlainTextBlock *ptb, const BYTE *bf, int size )
{
	if ( !writeIntToPTB( ptb, size ) )
		return false;

	return ptb->WriteBytes( bf, size ) == size;
}

bool writeStringToPTB( cPlainTextBlock *ptb, const char *str )
{
	int size = 0;
	
	if ( str != NULL )
		size = (int)strlen( str );

	if ( !writeIntToPTB( ptb, size ) )
		return false;

	return ptb->WriteBytes( (BYTE *)str, size ) == size;
}

bool writeStringToPTB( cPlainTextBlock *ptb, const wchar_t *str )
{
	cMBStr mbs( str, -1, false, CP_UTF8 );

	return writeStringToPTB( ptb, mbs.m_szBody ); 
}

bool writeIntToPTB( cPlainTextBlock *ptb, int val )
{
	BYTE *b = (BYTE *)( &val );
	return ptb->WriteBytes( b, sizeof( val ) ) == sizeof( val );
}
