#include "ptbRegs.h"
#include "UnicodeFuncs.h"
#include "sysRand.h"
#include "cDlg.h"

#include "..\CertificateLib\PGPPkts\pktdefs.h"
#include "..\CertificateLib\symmetricAlgs\cSymAlg.h"
#include "..\certificatelib\symmetricalgs\calgbfish.h"

#include <time.h>
#include <io.h>
#include <commctrl.h>

//static BYTE algID = pgpsaBlowfish;

static int getKeySize()
{
//	int keySize = sa_getKeySize( algID );
	int keySize = cAlgBfish::getKeySize();
	return keySize;
}

static int getBlockSize()
{
//	return sa_getBlockSize( algID );
	return cAlgBfish::getBlockSize();
}

static bool encryptBuffer( BYTE *key, BYTE *out, BYTE *in, int size )
{
	//HSA sa;
	//if ( sa_createAlgorithm( &sa, algID, key, getKeySize() ) != sae_ok )
	//	return false;

	//int res = sa_encrypt( sa, out, in, size );

	cSymAlg *sa = cAlgBfish::createAlg( key, getKeySize() );
	if ( sa == NULL )
		return false;

	int res = sa->encryptECB( in, out, size );

	delete sa;

	return res == sae_ok;
}

static bool decryptBuffer( BYTE *key, BYTE *out, BYTE *in, int size )
{
	//HSA sa;
	//if ( sa_createAlgorithm( &sa, algID, key, getKeySize() ) != sae_ok )
	//	return false;

	//int res = sa_decrypt( sa, out, in, size );

	cSymAlg *sa = cAlgBfish::createAlg( key, getKeySize() );
	if ( sa == NULL )
		return false;

	int res = sa->decryptECB( in, out, size );

	delete sa;

	return res == sae_ok;
}

bool readFromRegistry( cPlainTextBlock *ptb, HKEY hRegRoot, const wchar_t *regKeyName, const wchar_t *valueName )
{
	HKEY hKey;
	LONG res = RegOpenKeyExT( hRegRoot, regKeyName, 0, KEY_READ, &hKey );

	if ( res != ERROR_SUCCESS )
		return false;

	DWORD dwSize = 0;
	DWORD dwType = REG_BINARY;
	res = RegQueryValueExT( hKey, valueName, 0, &dwType, NULL, &dwSize );

	if ( dwSize == 0 )
	{
		RegCloseKey( hKey );
		return false;
	}

	cPlainTextBlock ptbReg;

	if ( ptbReg.Allocate( dwSize ) == NULL )
	{
		RegCloseKey( hKey );
		return false;
	}

	res = RegQueryValueExT( hKey, valueName, 0, &dwType, ptbReg.buffer, &dwSize );

	RegCloseKey( hKey );

	if ( res != ERROR_SUCCESS )
		return false;

	int keySize = getKeySize();
	int blockSize = getBlockSize();

	int dataSize = dwSize - keySize;

	if ( dataSize < blockSize )
		return false;

	if ( ( dataSize % blockSize ) != 0 ) // The encrypted block has to padded upto block size
		return false;

	if ( ptb->Allocate( dataSize ) == NULL )
		return false;

	if ( !decryptBuffer( ptbReg.buffer, ptb->buffer, ptbReg.buffer + keySize, dataSize ) )
		return false;

	// Checks padding

	int tail = ptb->buffer[ dataSize - 1 ];
	if ( tail > blockSize )
		return false;

	if ( tail == 0 )
		return false;

	for ( int i = tail; i; i-- )
		if ( ptb->buffer[ dataSize - i ] != tail )
			return false;

	//  Truncates the padded tail

	ptb->bufferSize -= tail;

	return true;
}

bool writeToRegistry( cPlainTextBlock *ptb, HKEY hRegRoot, const wchar_t *regKeyName, const wchar_t *valueName  )
{
	cPlainTextBlock ptbReg;

	if ( ptb != NULL )
	{
		// Padds input data
		int blockSize = getBlockSize();
		int tail = blockSize - ptb->bufferSize % blockSize;
		for( int t = tail; t--;  )
				ptb->PutByte( tail );

		// Generates key
		int keySize = getKeySize();
		if ( ptbReg.Allocate( keySize + ptb->bufferSize ) == NULL )
			return false;

		sRandCtx ctx;

		ctx.setSeed( time(0) ^ clock() );
		ctx.fillBuffer( ptbReg.buffer, keySize );

		// Encrypts input data

		if ( !encryptBuffer( ptbReg.buffer, ptbReg.buffer + keySize, ptb->buffer, ptb->bufferSize ) )
			return false;

//#ifdef _DEBUG
//
//		cPlainTextBlock tmp;
//		if ( tmp.Allocate( ptb->bufferSize ) == NULL )
//		{
//			assert( false );
//			return false;
//		}
//
//		bool decOk = decryptBuffer( ptbReg.buffer, tmp.buffer, ptbReg.buffer + keySize, tmp.bufferSize );
//
//		if ( !decOk || ( memcmp( tmp.buffer, ptb->buffer, tmp.bufferSize ) != 0 ) )
//		{
//			assert( false );
//			return false;
//		}
//
//#endif // _DEBUG
	}

	// Writes the encrypted block
	HKEY hKey;
	LONG res = RegCreateKeyExT( hRegRoot, regKeyName, 0, 0, 0, KEY_WRITE, 0, &hKey, 0 );

	if ( res != ERROR_SUCCESS )
		return false;

	if ( ptb == NULL )
		res = RegDeleteValueT( hKey, valueName );
	else
		res = RegSetValueExT( hKey, valueName, 0, REG_BINARY, ptbReg.buffer, ptbReg.bufferSize );

	RegCloseKey( hKey );

	if ( res != ERROR_SUCCESS )
		return false;

	return true;
}

cRegDlgValues::~cRegDlgValues()
{
	if ( m_colT ) delete m_colT;
	deleteString( &m_keyName );
	deleteString( &m_valueName );
}

cRegDlgValues::cRegDlgValues( HKEY hRoot, const char *keyName, const char *valueName, int classSize, int columns )
{
	Init( hRoot, cWCStr( keyName ).m_szBody, cWCStr( valueName ).m_szBody, classSize, columns );
}

cRegDlgValues::cRegDlgValues( HKEY hRoot, const wchar_t *keyName, const wchar_t *valueName, int classSize, int columns )
{
	Init( hRoot, keyName, valueName, classSize, columns );
}

void cRegDlgValues::InitColumns( int columns, int lvType )
{
	if ( m_colT != NULL )
	{
		delete m_colT;
		m_colT = NULL;
	}

	m_colQ = columns;

	if ( m_colQ != 0 )
	{
		m_colT = new int[ m_colQ ];
		if ( m_colT == NULL )
			m_colQ = 0;

		for ( int i = m_colQ; i--; ) m_colT[i] = 100;
	}

	m_lvType = lvType;
}

void cRegDlgValues::Init( HKEY hRoot, const wchar_t *keyName, const wchar_t *valueName, int classSize, int columns )
{
	memset( this, 0, classSize );
	m_hRoot = hRoot;
	m_keyName = newString( keyName );
	m_valueName = newString( valueName );
	m_size = classSize;
	InitColumns( columns, 0 );
}

//
//	checkValid
//		We can get error at saving so all columns are zero.
//
static bool checkValid( int *colT, int colQ )
{
	if ( colQ == 0 )
		return true;

	for( int i = colQ; i--; )
	{
		if ( colT[i] != 0 ) // Fond first non zero.
			return true;
	}

	return false;
}

bool cRegDlgValues::Restore( HWND hDlg, HWND hList, BYTE *buff, int buffSize, DWORD spFlags )
{
	if ( ( m_hRoot == NULL ) || ( m_keyName == NULL ) )
		return false;

	HKEY hKey;
	LONG res;
	
	res = RegOpenKeyExT( m_hRoot, m_keyName, 0, KEY_READ, &hKey );

	if ( res != ERROR_SUCCESS )
		return false;

	DWORD dwSize = 0;
	DWORD dwType = REG_BINARY;

	res = RegQueryValueExT( hKey, m_valueName, 0, &dwType, NULL, &dwSize );

	if ( res != ERROR_SUCCESS )
	{
		RegCloseKey( hKey );
		return false;
	}

	if ( ( dwType != REG_BINARY ) || ( dwSize == 0 ) )
	{
		RegCloseKey( hKey );
		return false;
	}

	cPlainTextBlock ptb;
	if ( ptb.Allocate( dwSize ) == NULL )
	{
		RegCloseKey( hKey );
		return false;
	}

	res = RegQueryValueExT( hKey, m_valueName, 0, &dwType, ptb.buffer, &dwSize );

	RegCloseKey( hKey );

	if ( res != ERROR_SUCCESS )
		return false;

	// Reads list view columns and type
	int colQ = *((int *)ptb.buffer);
	DWORD pos = sizeof( int );
	int *colT = (int *)( ptb.buffer + pos );

	pos += colQ * sizeof( int );
	if ( pos > dwSize ) return false;
	DWORD lvType = *(DWORD *)( ptb.buffer + pos );

	// Check all zero it rather was a error at saving this values
	if ( !checkValid( colT, colQ ) )
		return false;

	if ( m_colQ < colQ )
	{
		// It may be not inited yet.
		InitColumns( colQ, lvType );

		if ( m_colQ < colQ )
			return false; // Skip any initializations
	}

	memcpy( m_colT, colT, colQ * sizeof( int ) );

	m_lvType = lvType;
	pos +=  sizeof( m_lvType );
	if ( pos + sizeof( m_rect ) > dwSize )
		return false;
	
	// Reads dialogue rectangle
	memcpy( &m_rect, ptb.buffer + pos, sizeof( m_rect ) );
	pos += sizeof( m_rect );

	DWORD bodySize = m_size - sizeof( cRegDlgValues );
	bodySize = min( bodySize, dwSize - pos );

	// Reads body of subclass
	memcpy( (BYTE *)this + sizeof( cRegDlgValues ), ptb.buffer + pos, bodySize );
	pos += bodySize;

	buffSize = min( (DWORD)buffSize, dwSize - pos );
	if ( buffSize > 0 )
	{
		memcpy( buff, ptb.buffer + pos, buffSize );
		pos += buffSize;
	}

	// Sets restored positions.

	if ( hDlg != NULL )
	{
		DWORD flags = SWP_NOZORDER | spFlags;

		if ( cDlg::BelongToScreen( (POINT *)&m_rect ) )
			SetWindowPos( hDlg, 0, m_rect.left, m_rect.top, m_rect.right - m_rect.left, m_rect.bottom - m_rect.top, flags );
	}

	if ( hList == NULL )
		return true;

	DWORD dwStyle = ::GetWindowLong( hList, GWL_STYLE );
	DWORD dwNewStyle = (dwStyle & ~LVS_TYPEMASK) | m_lvType;
	::SetWindowLong( hList, GWL_STYLE, dwNewStyle);

	for ( int i = 0; i < m_colQ; i++ )
	{
		ListView_SetColumnWidth( hList, i, m_colT[i] );
	}

	return true;
}

bool cRegDlgValues::Save( BYTE *buff, int buffSize )
{
	if ( ( m_hRoot == NULL ) || ( m_keyName == NULL ) )
		return false;

	HKEY hKey;
	LONG res;

	res = RegOpenKeyExT( m_hRoot, m_keyName, 0, KEY_WRITE, &hKey );

	if ( res == ERROR_FILE_NOT_FOUND )
	{
		res = RegCreateKeyExT( m_hRoot, m_keyName, 0, NULL, 0, KEY_WRITE, NULL, &hKey, NULL );
	}

	if ( res != ERROR_SUCCESS )
		return false;

	cPlainTextBlock ptb;
	// Write list view columns and type
	ptb.WriteInt( m_colQ );
	ptb.WriteBytes( (BYTE *)m_colT, m_colQ * sizeof( int ) );
	ptb.WriteBytes( (BYTE *)&m_lvType, sizeof( m_lvType ) );

	// Write dialogue rectangle
	ptb.WriteBytes( (BYTE *)&m_rect, sizeof( m_rect ) );

	// Writes body of subclass
	ptb.WriteBytes( ((BYTE *)this) + sizeof( cRegDlgValues ), m_size - sizeof( cRegDlgValues ) );

	if ( buffSize != 0 )
		ptb.WriteBytes( buff, buffSize );

	res = RegSetValueExT( hKey, m_valueName, 0, REG_BINARY, ptb.buffer, ptb.bufferSize );

	RegCloseKey( hKey );

	return ( res == ERROR_SUCCESS );
}

bool cRegDlgValues::Save( HWND hDlg, HWND hList, BYTE *buff, int buffSize )
{
//	memset( this, 0, m_size );

	if ( IsIconic( hDlg ) )
	{
		ODS_fls( L"Warning! Saving is disabled for iconic window" );
		return false;
	}

	if ( hDlg != NULL )
		GetWindowRect( hDlg, &m_rect );

	m_lvType = GetWindowLong( hList, GWL_STYLE ) & LVS_TYPEMASK;

	for ( int i = 0; i < m_colQ; i++ )
	{
		m_colT[i] = ListView_GetColumnWidth( hList, i );
	}

	return Save( buff, buffSize );
}
