#pragma once

#include "UnicodeFuncs.h"
//
//	calss cSafeFileTime
//
//	Safely reades file time
//	Remembers  the time and file name
//	The time of the file may be restored when required or automatically at destructor.
//

class cSafeFileTime
{
	FILETIME CreationTime;
	FILETIME LastAccessTime;
	FILETIME LastWriteTime;

	wchar_t m_fileName[ MAX_PATH ];

public:

	bool m_updateSaved;

	static bool SafeGetFileTime(
		const wchar_t *fileName,
		LPFILETIME lpCreationTime,
		LPFILETIME lpLastAccessTime, 
		LPFILETIME lpLastWriteTime ) 
	{

		WIN32_FIND_DATAW findData;

		HANDLE findFile = FindFirstFileT( fileName, &findData );

		if ( findFile != INVALID_HANDLE_VALUE )
		{
			if (lpCreationTime)		*lpCreationTime		= findData.ftCreationTime;
			if (lpLastAccessTime)	*lpLastAccessTime	= findData.ftLastAccessTime;
			if (lpLastWriteTime)	*lpLastWriteTime	= findData.ftLastWriteTime;

			FindClose( findFile );

			return TRUE;
		}

		return FALSE;
	}

	bool ReadTime( const wchar_t *fileName )
	{
		m_updateSaved = false;

		wcscpy_s( m_fileName, MAX_PATH, fileName );

		if (!SafeGetFileTime( m_fileName, &CreationTime, &LastAccessTime, &LastWriteTime ))
			m_updateSaved = false;

		return true;
	}

	cSafeFileTime( const wchar_t *fileName = NULL )
	{
		if ( fileName != NULL )
			ReadTime( fileName );
	}

	virtual HANDLE OpenFile( LPCWSTR lpFileName )
	{
		return ::CreateFileT(
					lpFileName, 
					GENERIC_READ | GENERIC_WRITE,
					FILE_SHARE_READ,
					NULL,
					OPEN_EXISTING,
					FILE_ATTRIBUTE_NORMAL, //FILE_FLAG_NO_BUFFERING,
					NULL );
	}

	bool updateSaved( const wchar_t *fileName = NULL )
	{
		if ( fileName == NULL )
			fileName = m_fileName;

		HANDLE hFile = OpenFile( fileName );

		if ( hFile == INVALID_HANDLE_VALUE )
			return false;

		SetFileTime( hFile, &CreationTime, &LastAccessTime, &LastWriteTime );

		CloseHandle( hFile );

		return true;
	}

	~cSafeFileTime()
	{
		if ( !m_updateSaved )
			return;

		updateSaved();
	}
};
