#ifndef __Shell32Dll__
#define __Shell32Dll__

//
//	Shell 32 manual implementation.
//	Main reason:
//		Windows NT4 tries to loads SHGetSpecialFolderLocationA and SHGetSpecialFolderPathA
//		BUT shell32.dll does not have the functions. It has only
//		SHGetSpecialFolderLocation and SHGetSpecialFolderPath
//

#include <windows.h>
#include <shlobj.h>
#include "CDll.h"
#include "UnicodeFuncs.h"
#include "ODS.h"

typedef WINSHELLAPI HRESULT			WINAPI typeSHGetSpecialFolderLocation(HWND hwndOwner, int nFolder, LPITEMIDLIST * ppidl);
typedef WINSHELLAPI LPITEMIDLIST	WINAPI typeSHBrowseForFolder(LPBROWSEINFOA lpbi);
typedef WINSHELLAPI BOOL			WINAPI typeSHGetPathFromIDList(LPCITEMIDLIST pidl, LPSTR pszPath);
typedef WINSHELLAPI BOOL			WINAPI typeSHGetPathFromIDListW(LPCITEMIDLIST pidl, LPWSTR pszPath);
typedef WINSHELLAPI HRESULT			WINAPI typeSHGetMalloc(LPMALLOC * ppMalloc);
typedef WINSHELLAPI HINSTANCE	  APIENTRY typeFindExecutable(LPCSTR lpFile, LPCSTR lpDirectory, LPSTR lpResult);
typedef WINSHELLAPI BOOL			WINAPI typeSHGetSpecialFolderPath(HWND hwndOwner, LPSTR lpszPath, int nFolder, BOOL fCreate);
typedef WINSHELLAPI BOOL			WINAPI typeSHGetSpecialFolderPathW(HWND hwndOwner, LPWSTR lpszPath, int nFolder, BOOL fCreate);
//typedef SHSTDAPI_(DWORD)				   typeSHGetFileInfoA(LPCSTR pszPath, DWORD dwFileAttributes, SHFILEINFOA FAR *psfi, UINT cbFileInfo, UINT uFlags);
typedef WINSHELLAPI DWORD *	WINAPI typeSHGetFileInfo(LPCSTR pszPath, DWORD dwFileAttributes, SHFILEINFOA *psfi, UINT cbFileInfo, UINT uFlags);
typedef WINSHELLAPI HICON WINAPI typeExtractIconW(HINSTANCE hInst, LPCWSTR lpszExeFileName, UINT nIconIndex);
typedef WINSHELLAPI HICON WINAPI typeExtractIconA(HINSTANCE hInst, LPCSTR lpszExeFileName, UINT nIconIndex);
typedef WINSHELLAPI UINT WINAPI typeExtractIconExW( LPCWSTR lpszFile, int nIconIndex, HICON *phiconLarge, HICON *phiconSmall, UINT nIcons );
typedef WINSHELLAPI UINT WINAPI typeExtractIconExA( LPCSTR lpszFile, int nIconIndex, HICON *phiconLarge, HICON *phiconSmall, UINT nIcons );

enum {
	idShl32_Unknown			= 1,
	idShl32_ClosedFolder	= 4,
	idShl32_OpenedFolder	= 5,
	idShl32_DriveFloppy		= 7,
	idShl32_DriveRemovable	= 8,
	idShl32_DriveHard		= 9,
	idShl32_DriveNetwork	= 10,
	idShl32_DriveNetworkDiskonnected	= 11,
	idShl32_DriveCD			= 12,
	idShl32_VirtualMemory	= 13,
	idShl32_MyComputer		= 16,
	idShl32_Hibernate		= 26,
	idShl32_Delete			= 131
};

class cShell32Dll : public CDll
{
public:

	// Do not use the in global constructor, and avoid global contruction at all.
	static __declspec(noinline) cShell32Dll &getStatic()
	{
		static cSysCriticalSection cs;
		cSysCriticalSection_Block csb( &cs );

		static cShell32Dll dll;
		if ( dll.error ) ODS_fls( L"Assertion" );
		return dll;
	}

	typeExtractIconW *m_ExtractIconW;
	typeExtractIconA *m_ExtractIconA;

	typeExtractIconExW *m_ExtractIconExW;
	typeExtractIconExA *m_ExtractIconExA;

	typeSHBrowseForFolder *m_SHBrowseForFolder;
	typeSHGetPathFromIDList *m_SHGetPathFromIDList;
	typeSHGetPathFromIDListW *m_SHGetPathFromIDListW;
	typeSHGetMalloc *m_SHGetMalloc;
	typeSHGetSpecialFolderLocation *m_SHGetSpecialFolderLocation;
	typeFindExecutable *m_FindExecutable;
	typeSHGetFileInfo *m_SHGetFileInfo;
	
	typeSHGetSpecialFolderPath *m_SHGetSpecialFolderPath_R;
	typeSHGetSpecialFolderPathW *m_SHGetSpecialFolderPath_RW;

	BOOL m_SHGetSpecialFolderPath( HWND hwndOwner, LPSTR lpszPath, int nFolder, BOOL fCreate)
	{
		if ( m_SHGetSpecialFolderPath_R )
			return m_SHGetSpecialFolderPath_R( hwndOwner, lpszPath, nFolder, fCreate );

		LPITEMIDLIST pidl = NULL;

		if ( m_SHGetSpecialFolderLocation( hwndOwner, nFolder, &pidl ) != NOERROR )
		{
			//	NT4 does not recognise the folder ids but the folder 
			//	are located in Windows directory ( by experiments )

			switch ( nFolder )
			{
				//	CSIDL_INTERNET_CACHE 
				//	Version 4.72. File system directory that serves as a common 
				//	repository for temporary Internet files. A typical path is 
				//	C:\Documents and Settings\username\Temporary Internet Files. 
				case CSIDL_INTERNET_CACHE:
				{
					if ( GetWindowsDirectoryA( lpszPath, MAX_PATH ) == 0 )
						return false;

					strcat_s( lpszPath, MAX_PATH, "\\Profiles\\" );
					DWORD size = 100;
					if ( !GetUserNameA( lpszPath + strlen( lpszPath ), &size ) )
						return false;

					strcat_s( lpszPath, MAX_PATH, "\\Temporary Internet Files" );
					break;
				}

				//	CSIDL_COOKIES
				//	File system directory that serves as a common repository for 
				//	Internet cookies. A typical path is C:\Documents and Settings
				//	\username\Cookies.
				case CSIDL_COOKIES:
				{
					if ( GetWindowsDirectoryA( lpszPath, MAX_PATH ) == 0 )
						return false;

					strcat_s( lpszPath, MAX_PATH, "\\Profiles\\" );
					DWORD size = 100;

					if ( !GetUserNameA( lpszPath + strlen( lpszPath ), &size ) )
						return false;

					strcat_s( lpszPath, MAX_PATH, "\\Cookies" );
					break;
				}
				//	CSIDL_RECENT 
				//	File system directory that contains the user's most recently used 
				//	documents. A typical path is C:\Documents and Settings\username\Recent. 
				//	To create a shortcut in this folder, use SHAddToRecentDocs. In addition to 
				//	creating the shortcut, this function updates the Shell's list of recent 
				//	documents and adds the shortcut to the Documents submenu of the Start menu.
				case CSIDL_RECENT:
				{
					if ( GetWindowsDirectoryA( lpszPath, MAX_PATH ) == 0 )
						return false;

					strcat_s( lpszPath, MAX_PATH, "\\Profiles\\" );
					DWORD size = 100;
					if ( !GetUserNameA( lpszPath + strlen( lpszPath ), &size ) )
						return false;

					strcat_s( lpszPath, MAX_PATH, "\\Recent" );
					break;
				}
				case CSIDL_HISTORY:
				{
					if ( GetWindowsDirectoryA( lpszPath, MAX_PATH ) == 0 )
						return false;

					strcat_s( lpszPath, MAX_PATH, "\\Profiles\\" );
					DWORD size = 100;

					if ( !GetUserNameA( lpszPath + strlen( lpszPath ), &size ) )
						return false;

					strcat_s( lpszPath, MAX_PATH, "\\History" );
					break;
				}
				//	CSIDL_FAVORITES 
				//	File system directory that serves as a common repository for the user's favorite 
				//	items. A typical path is C:\Documents and Settings\username\Favorites
				case CSIDL_FAVORITES:
				{
					if ( GetWindowsDirectoryA( lpszPath, MAX_PATH ) == 0 )
						return false;

					strcat_s( lpszPath, MAX_PATH, "\\Profiles\\" );
					DWORD size = 100;
					if ( !GetUserNameA( lpszPath + strlen( lpszPath ), &size ) )
						return false;

					strcat_s( lpszPath, MAX_PATH, "\\Favorites" );
					break;
				}
				//	CSIDL_NETHOOD 
				//	A file system folder containing the link objects that may exist in the My 
				//	Network Places virtual folder. It is not the same as CSIDL_NETWORK, which 
				//	represents the network namespace root. A typical path is C:\Documents and 
				//	Settings\username\NetHood. 
				case CSIDL_NETHOOD:
				{
					if ( GetWindowsDirectoryA( lpszPath, MAX_PATH ) == 0 )
						return false;

					strcat_s( lpszPath, MAX_PATH, "\\Profiles\\" );
					DWORD size = 100;
					
					if ( !GetUserNameA( lpszPath + strlen( lpszPath ), &size ) )
						return false;

					strcat_s( lpszPath, MAX_PATH, "\\NetHood" );
					break;
				}
				default:
					return FALSE;
			}
		}
		else
		{
			BOOL bRes = m_SHGetPathFromIDList( pidl, lpszPath );

			LPMALLOC lpMalloc = NULL;

			if ( ( S_OK == m_SHGetMalloc(&lpMalloc) ) && (NULL != lpMalloc))
			{
				if (NULL != pidl)
				{
					lpMalloc->Free( pidl );
				}
			}

			if ( GetVersion() & 0x80000000 ) // Windows 95/98/ME
			{
				if ( bRes != TRUE )
					return FALSE;
			}
			else
			if ( bRes != NOERROR )
				return FALSE;
		}

		if ( fCreate )
			CreateDirectoryA( lpszPath, NULL );

		return TRUE;
	}

	BOOL m_SHGetSpecialFolderPathW( HWND hwndOwner, LPWSTR lpszPath, int nFolder, BOOL fCreate)
	{
		if ( m_SHGetSpecialFolderPath_RW )
			return m_SHGetSpecialFolderPath_RW( hwndOwner, lpszPath, nFolder, fCreate );

		LPITEMIDLIST pidl = NULL;
		
		if ( m_SHGetSpecialFolderLocation == NULL )
		{
			ODS_fls( L"Assertion!" );
			return false;
		}

		if ( m_SHGetSpecialFolderLocation( hwndOwner, nFolder, &pidl ) != NOERROR )
		{
			//	NT4 does not recognise the folder ids but the folder 
			//	are located in Windows directory ( by experiments )

			GetWindowsDirectoryT( lpszPath, MAX_PATH );
			wcscat_s( lpszPath, MAX_PATH, L"\\Profiles\\" );
			DWORD size = 100;
			GetUserNameT( lpszPath + wcslen( lpszPath ), &size );

			switch ( nFolder )
			{
				//	CSIDL_INTERNET_CACHE 
				//	Version 4.72. File system directory that serves as a common 
				//	repository for temporary Internet files. A typical path is 
				//	C:\Documents and Settings\username\Temporary Internet Files. 
				case CSIDL_INTERNET_CACHE:
				{
					wcscat_s( lpszPath, MAX_PATH, L"\\Temporary Internet Files" );
					break;
				}

				//	CSIDL_COOKIES
				//	File system directory that serves as a common repository for 
				//	Internet cookies. A typical path is C:\Documents and Settings
				//	\username\Cookies.
				case CSIDL_COOKIES:
				{
					wcscat_s( lpszPath, MAX_PATH, L"\\Cookies" );
					break;
				}
				//	CSIDL_RECENT 
				//	File system directory that contains the user's most recently used 
				//	documents. A typical path is C:\Documents and Settings\username\Recent. 
				//	To create a shortcut in this folder, use SHAddToRecentDocs. In addition to 
				//	creating the shortcut, this function updates the Shell's list of recent 
				//	documents and adds the shortcut to the Documents submenu of the Start menu.
				case CSIDL_RECENT:
				{
					wcscat_s( lpszPath, MAX_PATH, L"\\Recent" );
					break;
				}
				case CSIDL_HISTORY:
				{
					wcscat_s( lpszPath, MAX_PATH, L"\\History" );
					break;
				}
				//	CSIDL_FAVORITES 
				//	File system directory that serves as a common repository for the user's favorite 
				//	items. A typical path is C:\Documents and Settings\username\Favorites
				case CSIDL_FAVORITES:
				{
					wcscat_s( lpszPath, MAX_PATH, L"\\Favorites" );
					break;
				}
				//	CSIDL_NETHOOD 
				//	A file system folder containing the link objects that may exist in the My 
				//	Network Places virtual folder. It is not the same as CSIDL_NETWORK, which 
				//	represents the network namespace root. A typical path is C:\Documents and 
				//	Settings\username\NetHood. 
				case CSIDL_NETHOOD:
				{
					wcscat_s( lpszPath, MAX_PATH, L"\\NetHood" );
					break;
				}
				default:
					return FALSE;
			}
		}
		else
		{
			if ( m_SHGetPathFromIDListW == NULL )
				return FALSE;

			BOOL bRes = m_SHGetPathFromIDListW( pidl, lpszPath );

			LPMALLOC lpMalloc = NULL;

			if ( ( S_OK == m_SHGetMalloc(&lpMalloc) ) && (NULL != lpMalloc))
			{
				if (NULL != pidl)
				{
					lpMalloc->Free( pidl );
				}
			}

			if ( bRes != 1 )
				return FALSE;
		}

		if ( fCreate )
			CreateDirectoryT( lpszPath, NULL );

		return TRUE;
	}

	HICON hFolder;
	HICON hClosedFolder;
	enum { idShell32_ClosedFolder = 4, idShell32_OpenedFolder = 5 };

	cShell32Dll() : CDll( FALSE, "shell32.dll", NULL, TRUE )
	{
		hFolder = NULL;
		hClosedFolder = NULL;

		if ( error )
			return;

		// It does not show error, if the function does not exist ( NT4 )
		// the m_SHGetSpecialFolderPath manually implement it.
		GetProcAddress( "SHGetSpecialFolderPathA", FALSE, (void**)&m_SHGetSpecialFolderPath_R );
		GetProcAddress( "SHGetSpecialFolderPathW", FALSE, (void**)&m_SHGetSpecialFolderPath_RW );
		GetProcAddress( "SHGetPathFromIDListW", FALSE, (void**)&m_SHGetPathFromIDListW );
		error = FALSE;

		GetProcAddress( "SHGetSpecialFolderLocation", TRUE, (void**)&m_SHGetSpecialFolderLocation );
		GetProcAddress( "SHBrowseForFolder", TRUE, (void**)&m_SHBrowseForFolder );
		GetProcAddress( "SHGetPathFromIDList", TRUE, (void**)&m_SHGetPathFromIDList );
		GetProcAddress( "SHGetMalloc", TRUE, (void**)&m_SHGetMalloc );
		GetProcAddress( "FindExecutableA", TRUE, (void**)&m_FindExecutable );

		if ( !GetProcAddress( "ExtractIconW", TRUE, (void**)&m_ExtractIconW ) )
			error = false;

		GetProcAddress( "ExtractIconA", TRUE, (void**)&m_ExtractIconA );

		if ( !GetProcAddress( "ExtractIconExW", TRUE, (void**)&m_ExtractIconExW ) ) error = false;
		if ( !GetProcAddress( "ExtractIconExA", TRUE, (void**)&m_ExtractIconExA ) ) error = false;

		if ( !GetProcAddress( "SHGetFileInfoA", false, (void**)&m_SHGetFileInfo ) )
		{
			error = FALSE;
			GetProcAddress( "SHGetFileInfo", true, (void**)&m_SHGetFileInfo );
		}
	}

	bool setFolderIcons()
	{
		char path[ MAX_PATH ];

		GetTempPathA( sizeof( path ), path );

		SHFILEINFOA sfi;

		//if ( SHGetFileInfoA( path, 0, &sfi, sizeof(sfi), SHGFI_ICON | SHGFI_SMALLICON | SHGFI_DISPLAYNAME  ))
		if ( m_SHGetFileInfo( path, 0, &sfi, sizeof(sfi), SHGFI_ICON | SHGFI_SMALLICON | SHGFI_DISPLAYNAME  ))
		{
			hFolder = sfi.hIcon;
		}
		else
		{
			hFolder = LoadIcon( hInst, MAKEINTRESOURCE( idShell32_OpenedFolder ) );
		}

		hClosedFolder = LoadIcon( hInst, MAKEINTRESOURCE( idShell32_ClosedFolder ) );

		return hFolder != NULL;
	}

	HICON GetFolderIcon( bool opened = false )
	{
		if ( hFolder == NULL )
			setFolderIcons();

		return opened ? hFolder : hClosedFolder;
	}
};

#endif // __Shell32Dll__