//////////////////////////////////////////////////////
//
//	Random generator based on system used scheme.
//		Xn+1 = A*Xn + C
//		The constants A nd C used by Donald Knuth recomendation for 64bits
//
//	Purpouse:
//		The inline optimisation with 64bit operations sped up generation process in 10 times 
//		comparing with ordinary system rand().
//
///////////////////////////////////////////////////////

#pragma once
#include <time.h>

union uDWLData
{
	WORD word[4];
	DWORD dword[2];
	DWORDLONG dwl;
};

struct sRandCtx
{
	LARGE_INTEGER X;
	DWORDLONG XSaved;
	DWORDLONG A,C;

	// MMIX by Donald Knuth	64bit	A = 6364136223846793005	 C= 1442695040888963407	

	sRandCtx()
	{
		set( 0 );
	}

	void set( DWORDLONG x )
	{
		X.QuadPart = x;
		A = 6364136223846793005L;
		C = 1442695040888963407L;
	}

	inline DWORD getDW()
	{
		//
		// Microsoft Visual/Quick C/C++
		//
		// Xn+1 = A*Xn + C

		X.QuadPart = A*X.QuadPart + C;

		//
		//	Weak point if we show internal data.
		//	Intruder just enumerate the rest part and compare with next result and gets entire internal data.
		//
		// return DWORD( X >> 31 );
		return DWORD( ( X.QuadPart * X.HighPart ) >> 31 );
	}

	inline DWORDLONG getDWL()
	{
		uDWLData data;

		data.dword[0] = getDW();
		data.dword[1] = getDW();

		return data.dwl;
	}

	void setSeed( DWORDLONG seed )
	{
		set( seed );
		XSaved = X.QuadPart;
	}

	void setBlockSeed( DWORDLONG seed )
	{
		X.QuadPart = XSaved ^ seed;
	}

	void fillBuffer( BYTE *buff, DWORD buffSize, bool bComplement = false )
	{
		DWORDLONG *dwlBuff = (DWORDLONG*)buff;
		DWORD dwBuffSize = buffSize / sizeof( *dwlBuff );

		DWORD tail = dwBuffSize * sizeof( *dwlBuff );

		if ( bComplement )
		{
			for( DWORD i = 0; i < dwBuffSize; i++ ) { dwlBuff[ i ] = ~getDWL(); }
			for( ; tail < buffSize; tail++ ) { buff[ tail ] = ~((BYTE)getDW()); }
		}
		else
		{
			for( DWORD i = 0; i < dwBuffSize; i++ )	{ dwlBuff[ i ] = getDWL(); }
			for( ; tail < buffSize; tail++ ) { buff[ tail ] = (BYTE)getDW(); }
		}
	}

	static DWORDLONG getTimeSeed()
	{
		DWORDLONG dwl;
		GetSystemTimeAsFileTime( (FILETIME*)&dwl );
		clock_t cl = clock();

		dwl *= (DWORDLONG)cl;
		dwl ^= (DWORDLONG)cl;

		return dwl;
	}
};
