/*******************************************************************\
*	Copyright (c) 1994-2007 Jetico, Inc., Finland
*	All rights reserved.
*
*	File: alg.h
*	Revision: 
*	Created:
*	Description: interface with Algorithm Driver
*
\*******************************************************************/

#pragma once

#include "cSymAlgInfo.h"
#include "..\..\..\certificatelib\smemory\sblocks.h"

typedef DWORD KEY_HANDLE;

extern BOOL Alg_VerifyDriver( LPCSTR AlgServiceName, UCHAR *pVMajor = NULL, UCHAR *pVMinor = NULL );

extern BOOL 
	Alg_GetKeyHandler(
		LPCSTR AlgServiceName, 
        UCHAR  *Key,
		DWORD  KeyLength, // in bites
		BYTE   *Pool,
		DWORD  *KeyHandle
	);

extern BOOL 
	Alg_FreeKeyHandler( LPCSTR AlgServiceName, DWORD  KeyHandle );

extern BOOL
	Alg_Encrypt(
		LPCSTR AlgServiceName,
		DWORD  KeyHandle, 
		BYTE   *IVector, 
		BYTE   *Buffer, 
		DWORD  Length
	);

extern BOOL
	Alg_Decrypt( 
		LPCSTR AlgServiceName,
		DWORD  KeyHandle, 
		BYTE   *IVector, 
		BYTE   *Buffer, 
		DWORD  Length
	);

extern bool 
	Alg_GetAlgInfoByID(
		DWORD algID,
		DWORD encMode,
		char *algServiceNameBuff = NULL, int algServiceNameBuffSize = 0,
		DWORD *pKeyLength = NULL,
		DWORD *pAlgBlockSize = NULL,
		char *dispName = NULL,
		int dispNameSize = 0 );

extern int Alg_GetKeySize( DWORD algID, DWORD encModeId );

extern bool 
	Alg_GetModeInfo(
		LPCSTR AlgServiceName,
		int modeId,
		DWORD *pKeyBitLength = NULL
	);

class cIdTable
{
public:
	DWORD *T;	// Table with IDs
	int Q;		// IDs quantity

	bool checkValid( DWORD dwId )
	{
		if ( dwId == (DWORD)-1 )
			return false;

		bool bValid = false;

		for ( int i = 0; i<Q; i++ )
		{
			if ( T[i] == dwId )
			{
				bValid = true;
				break;
			}
		}

		return bValid;
	}

	void setFirst( DWORD dwFirst, bool bForward )
	{
		return setFirst( T, Q, dwFirst, bForward );
	}

	static void setFirst( DWORD *pT, int pQ, DWORD dwFirst, bool bForward )
	{
		if ( pQ  <= 1 )
			return;

		if ( dwFirst == (DWORD)-1 )
			dwFirst = pT[0];

		// if List is used from beginnig bForward is true, our first item has to be at 0 position.
		int firstPos = bForward ? 0 : (pQ - 1);

		for( int i = 0; i < pQ; i++ )
		{
			if ( pT[i] == dwFirst )
			{
				pT[i] = pT[firstPos];
				pT[firstPos] = dwFirst;
				return;
			}
		}
	}
};

//
//	class cEnumAlgIDs
//		Enumerates all available algorithm ID in registry
//
class cEnumAlgIDs : public cIdTable
{
public:

	static __declspec(noinline) cEnumAlgIDs &getStatic()
	{
		static cEnumAlgIDs algIDs;
		return algIDs;
	};

	cEnumAlgIDs( DWORD dwFirst = (DWORD)-1, bool bForward = true );
	~cEnumAlgIDs();

	int getAlgId( const char *algName );
};

//
//	class cEnumEncModes
//
struct sEncMode { int ids; DWORD id; };
class cEnumEncModes
{
public:

	static __declspec(noinline) cEnumEncModes &getStatic()
	{
		static cEnumEncModes encModes;
		return encModes;
	};

	sEncMode *T;	// Table with IDs and IDSs
	int Q;		// IDs quantity
	
	cEnumEncModes( DWORD dwFirst = (DWORD)-1 );
	~cEnumEncModes();

	int getAlgId( const wchar_t *algName, HINSTANCE hInst );
};

//
//	class cEnumHashAlgIDs
//		Enumerates all available hash algorithm ID in registry
//
class cEnumHashAlgIDs : public cIdTable
{
public:

	static __declspec(noinline) cEnumHashAlgIDs &getStatic()
	{
		static cEnumHashAlgIDs hashAlgs;
		return hashAlgs;
	};

	cEnumHashAlgIDs( DWORD dwFirst = (DWORD)-1, bool bForward = true );
	~cEnumHashAlgIDs();

	int getAlgId( const char *algName );
};

//
//	The algorithm information may be created inside the Key Generator, 
//	but the algorithm drivers has to be loaded before using.
//
extern bool SetSymAlgInfo( cSymAlgInfo *T, int Q );

class cSymAlgorithm : public cSymAlgInfo
{
public:

	KEY_HANDLE	m_keyHandle;
	bool m_initError;

	void Free();

	cSymAlgorithm();
	~cSymAlgorithm() { Free(); }

	bool Init( DWORD algId, DWORD encModeId );

	static const wchar_t *getModeName( int encMode );
	const wchar_t *getModeName() { return getModeName( m_encModeId ); }

	const wchar_t *getAlgName( wchar_t *bf, int bfLen );

	// Creates key handle
	static bool createKey( KEY_HANDLE *pKeyHandle, DWORD algId, DWORD encModeId, BYTE *keyBuff, int keyBuffSize, BYTE *pool );
	bool createKey( DWORD algId, DWORD encModeId, BYTE *keyBuff, int keyBuffSize, BYTE *pool );

	// Encrypt/decrypt operations.
	static bool encrypt( BYTE *iv, BYTE *bfOut, const BYTE *bfIn, int bfSize, KEY_HANDLE keyHandle, DWORD algId, DWORD encModeId );
	static bool decrypt( BYTE *iv, BYTE *bfOut, const BYTE *bfIn, int bfSize, KEY_HANDLE keyHandle, DWORD algId, DWORD encModeId );
	bool encrypt( BYTE *iv, BYTE *bfOut, const BYTE *bfIn, int bfSize );
	bool decrypt( BYTE *iv, BYTE *bfOut, const BYTE *bfIn, int bfSize );

	//
	//	delegateHandle
	//		Delegates intenal key handle to corresponded pointer
	//		The delegated handle is not freed at destructor.
	void delegateHandle( KEY_HANDLE *pKeyHandle );

	static int getByteSize( int bitSize ) { return ( ( bitSize + 7 ) / 8 ); }

	bool supportLRWMode();
	bool supportMode( DWORD mode );
};

class cPassword;

//
//	Class cPwdKeyV5
//		Saves generated key and its values for V5 key generator.
//		It uses one advantaged of V5, all slots have same salt, so if hash algorithm was not changed, so this password key will be compatible with all algorithms
//		if they key sizes are same.
//
class cPwdKeyV5
{
	//
	//	The PTB buffer consits of
	//		[ key ] + [ password digest ] + [ salt ]
	//
	cPlainTextBlock m_ptb;
	DWORD m_hashAlgId;
	DWORD m_iterations;

	void Reset()
	{
		m_key = NULL;
		m_keySize = 0;
		m_hashAlgId = -1;
		m_iterations = 0;
	}

	bool getPwdDigest( cPassword *pwd, DWORD hashAlgId, BYTE *digest );
	bool generateKey( int keySize, cPassword *pwd, DWORD hashAlgId, BYTE *salt, int saltSize, int iterations );
	bool isSameParams( int keySize, cPassword *pwd, DWORD hashAlgId, BYTE *salt, int saltSize, int iterations );

public:

	BYTE *m_key;
	int m_keySize;
	
	cPwdKeyV5()
	{
		Reset();
	}

	void Free()
	{
		m_ptb.Free();
	}
	
	~cPwdKeyV5()
	{
		Free();
	}

	bool getKey( int keySize, cPassword *pwd, DWORD hashAlgId, BYTE *salt, int saltSize, int iterations );
};

class cPwdAlgorithm : public cSymAlgorithm
{
public:
	cPwdKeyV5 *m_pwdKey;

	bool m_err;
	cPwdAlgorithm( DWORD algID = -1, DWORD encModeId = -1, cPassword *pwd = NULL, DWORD hashAlgId = -1, BYTE *salt = NULL, int saltSize = 0, int iteration = 0 );
	bool generateKey( DWORD algID, DWORD encModeId, cPassword *pwd, DWORD hashAlgId, BYTE *salt, int saltSize, int iteration );
};

