//
//  bc_ioctl.h - Control Codes for interface between Control Panel and drivers
//

#ifndef __BCIOCTL
#define __BCIOCTL

#include <windef.h> 
//#include "ntdefs.h"

//#define COMPILE_FOR_64_BIT 1

#define UNUSED_ALGORITHM_ID 0xFF

// MAX_KEY_LENGTH value in bites
#define MAX_KEY_LENGTH      512

/* Pool of random bytes */
#define POOL_SIZE_BYTES		512

// EX_ENCRYPT flags are used in EncryptEx commands
#define EX_ENCRYPT_LRW_MODE_FLAG  0xBC000001
#define EX_ENCRYPT_CBC_MODE_FLAG  0xBC000002
#define EX_ENCRYPT_XTS_MODE_FLAG  0xBC000004
#define EX_ENCRYPT_ECB_MODE_FLAG  0xBC000008

#define ECB_MODE_NAME  "ECB"
#define CBC_MODE_NAME  "CBC"
#define LRW_MODE_NAME  "LRW"
#define XTS_MODE_NAME  "XTS"

// NOT_EXISTING_KEY_HANDLE... can be sent in 
// IOCTL_PRIVATE_TEST_KEY_HANDLE. If result shows
// that the Key Handle is valid, hence, 
// the Algorithm Driver supports LRW encryption mode.
#define NOT_EXISTING_KEY_HANDLE_V8_SPECIFIC 0xABCD0123


// The following error is returned by BCDisk
// driver (Windows NT version) if a error 
// during opening file on Windows NT Server 
// occurs. Retrying of operation may solve 
// the problem.
#define BCSTATUS_ERR_RETRY_REQUIRED      0x12

typedef DWORD KEY_HANDLE;
 
typedef struct _BUFFER_HEADER { char   Signature[8];	// input: "LOCOS97",00 ; output: "LS06CXX",00
                                UCHAR   Status;
                              } BUFFER_HEADER, *PBUFFER_HEADER;

/*******************************************************************************
***************** IOCTLs for BestCrypt Disk Driver Service *********************
********************************************************************************/

// Get version command & structure definition

#define IOCTL_PRIVATE_GET_VERSION_INFO \
              CTL_CODE(IOCTL_DISK_BASE, 0x800, METHOD_IN_DIRECT, FILE_ANY_ACCESS)


typedef struct  _VERSION_BUFFER { char   Signature[8];	// input: "LOCOS97",00 ; output: "LS06CXX",00
                                  UCHAR   Status;
								  UCHAR   Major_version;
								  UCHAR   Minor_version;
								} VERSION_BUFFER, *PVERSION_BUFFER;

//
//  Control codes for the "program -> driver" interface
//
//  This control codes are private for this driver, hence, 
//  functions number must be >= 0x800
//
//  Functions numbers for Disk Driver      are 0x800 - 0x80F
//  


// Get Logical Disk Mask command & structure definition

#define IOCTL_PRIVATE_GET_BCDISK_MASK \
              CTL_CODE(IOCTL_DISK_BASE, 0x801, METHOD_IN_DIRECT, FILE_ANY_ACCESS)


typedef struct  _GETMASK_BUFFER {
	char   Signature[8];	// input: "LOCOS97",00 ; output: "LS06CXX",00
	UCHAR  Status;
	DWORD  Mask;

	//
	//	Mask of the drives available for transparent ecryption.
	//	To set transparent ecryption we will use same command as for usual mount
	//	 but for bcrmpart.vxd (or .sys) "BestCrypt Removable Partitions"
	//
	DWORD  hostsMask;
} GETMASK_BUFFER, *PGETMASK_BUFFER;


// Get Logical Disk Mask command & structure definition

#define IOCTL_PRIVATE_GET_BCDISK_INFO \
              CTL_CODE(IOCTL_DISK_BASE, 0x802, METHOD_IN_DIRECT, FILE_ANY_ACCESS)

#define DISKINFO_BUFFER_FILENAME_SIZE 256
#define MOUNT_BC_READONLY_NTFS 0x0000FFFF


#ifdef COMPILE_FOR_64_BIT

#pragma pack(push, bcioctl_h)
#pragma pack(4)

typedef struct  _DISKINFO_BUFFER { char   Signature[8];	// input: "LOCOS97",00 ; output: "LS06CXX",00
                                   UCHAR  Status;
                                   char   filename[DISKINFO_BUFFER_FILENAME_SIZE - 8];
								   HANDLE process;
                                   DWORD  diskNumber;
                                   DWORD  diskSizeLow;
                                   DWORD  diskSizeHigh;
                                   DWORD  keyHandle;
                                   char   algDevice[128];
                                   DWORD  algorithmID;
                                   DWORD  readOnly;
                                   HANDLE thread;
                                   DWORD  sectorSize;
                                   BOOL   fullDismounted;
                                   DWORD  dataOffset;
                                   DWORD  UserId;                // Ser added 25-Apr-2003
                                   DWORD  HarddiskVolumeNumber;  // Ser added 25-Apr-2003
                                 } DISKINFO_BUFFER, *PDISKINFO_BUFFER;


typedef struct  _DISKINFO_BUFFER_3 { char   Signature[8];	// input: "LOCOS97",00 ; output: "LS06CXX",00
                                   UCHAR  Status;
                                   char   filename[DISKINFO_BUFFER_FILENAME_SIZE - 8];
								   HANDLE process;
                                   DWORD  diskNumber;
                                   DWORD  diskSizeLow;
                                   DWORD  diskSizeHigh;
                                   DWORD  keyHandle;
                                   char   algDevice[128];
                                   DWORD  algorithmID;
                                   DWORD  readOnly;
                                   HANDLE thread;
                                   DWORD  sectorSize;
                                   BOOL   fullDismounted;
                                   DWORD  dataOffset;
                                   DWORD  UserId;                // Ser added 25-Apr-2003
                                   DWORD  HarddiskVolumeNumber;  // Ser added 25-Apr-2003
                                   DWORD  dataOffsetHigh;        // Ser added 02-Feb-2005
                                 } DISKINFO_BUFFER_3, *PDISKINFO_BUFFER_3;

#pragma pack(pop, bcioctl_h)

#else

typedef struct  _DISKINFO_BUFFER { char   Signature[8];	// input: "LOCOS97",00 ; output: "LS06CXX",00
                                   UCHAR  Status;
                                   char   filename[DISKINFO_BUFFER_FILENAME_SIZE];
                                   DWORD  diskNumber;
                                   DWORD  diskSizeLow;
                                   DWORD  diskSizeHigh;
                                   DWORD  keyHandle;
                                   char   algDevice[128];
                                   DWORD  algorithmID;
                                   DWORD  readOnly;
                                   HANDLE process;
                                   HANDLE thread;
                                   DWORD  sectorSize;
                                   BOOL   fullDismounted;
                                   DWORD  dataOffset;
                                   DWORD  UserId;                // Ser added 25-Apr-2003
                                   DWORD  HarddiskVolumeNumber;  // Ser added 25-Apr-2003
                                 } DISKINFO_BUFFER, *PDISKINFO_BUFFER;


typedef struct  _DISKINFO_BUFFER_3 { char   Signature[8];	// input: "LOCOS97",00 ; output: "LS06CXX",00
                                   UCHAR  Status;
                                   char   filename[DISKINFO_BUFFER_FILENAME_SIZE];
                                   DWORD  diskNumber;
                                   DWORD  diskSizeLow;
                                   DWORD  diskSizeHigh;
                                   DWORD  keyHandle;
                                   char   algDevice[128];
                                   DWORD  algorithmID;
                                   DWORD  readOnly;
                                   HANDLE process;
                                   HANDLE thread;
                                   DWORD  sectorSize;
                                   BOOL   fullDismounted;
                                   DWORD  dataOffset;
                                   DWORD  UserId;                // Ser added 25-Apr-2003
                                   DWORD  HarddiskVolumeNumber;  // Ser added 25-Apr-2003
                                   DWORD  dataOffsetHigh;        // Ser added 02-Feb-2005
                                 } DISKINFO_BUFFER_3, *PDISKINFO_BUFFER_3;
#endif


#pragma pack(4)

typedef struct _DI_OFFSET
{
	LONGLONG offset;
	LONGLONG length;
} DI_OFFSET;


enum eMASK 
{
		MASK_OrderNumber    = 0x00000001, 
		MASK_MountPoint     = 0x00000002, 
		MASK_ContainerName  = 0x00000004,
		MASK_InternalId     = 0x00000008,
		MASK_RemoveSecurity = 0x00000010  // The flag is valid for CMD_ChangeNtSecurity only. If set, we will remove security settings for the BestCrypt drive.
};

enum eCMD 
{
		CMD_GetQuantity = 0,  // if set, quantity of all mount points returned in 'order_number' field
		CMD_MapContainer,
		CMD_GetInformation,
		CMD_Disconnect,
		CMD_ChangeNtSecurity,
		CMD_ChangeNtSecurityObject, // For this command set V8_buf->mountPoint to "\\??\\PhysicalDriveN". Set MASK_RemoveSecurity to remove security, do NOT set it to restore security.
		CMD_DisconnectAll           // The command works only for BestCrypt Traveller driver, not for regular BestCrypt.
};

typedef struct  _DISKINFO_BUFFER_V8_0
{
	char       Signature[8];           // input: "LOCOS97",00 ; output: "LS06CXX",00
	UCHAR      Status;                 // return status, read bcstatus.h
	DWORD      structSize;             // size of the structure in bytes
	DWORD      structVersion;          // version of the structure, currently is 0 (version 8.0)

	enum eMASK validFieldsMask;        // can be set to one of the 'enum eMASK' numbers.
	enum eCMD  command;                // can be set to one of the 'enum eCMD' numbers.

	DWORD      orderNumber;            // 1st mounted virtual drive has orderNumber 0, 2nd - 1, 3rd - orderNumber 2 ...
	DWORD      internalId;             // number of mounted device, internal for driver. In BC v.7 the number was the same as DiskNumber: 0 for A:\, ..., 2 for C:\ ...

	WCHAR      fileName[ MAX_PATH ];   // name of container file
	WCHAR      mountPoint[ MAX_PATH ]; // drive letter ("C:\") or mount point ("H:\ntfs_empty_directory")
	DWORD      keyHandle;

	DWORD      containerVersion;       // 0   - use 16-bit IV; 
	                                   // 1-7 - use 64-bit IV; 
	                                   // 8   - use 64-bit IV and another header structure, where Computer Name placed on other offset and the header can be encrypted

	BOOL       fullDismounted;         // used in Disconnect request to instruct driver whether disconnect must be done, or it is just possible.

	WCHAR      algDevice[128];         // name of Encryption Algorithm Driver
	DWORD      algorithmId;
	DWORD      userId;
	DWORD      readOnly;               // == 1 if mount BC drive as read-only device
	
	DWORD      sectorSize;             // size of sector on hard drive where container file resides
	DWORD      harddiskVolumeNumber;   // Not used when mounted. Used when the driver returns information about its virtual drive.
	
	ULONGLONG  process;                // reserved for use by driver
	
	ULONGLONG  volumeOffsetOnPhysicalDisk;

	ULONGLONG  thread;                 // reserved for use by driver
	BYTE       reserved[108];          // reserved for future versions

	DWORD      ivModeFlag;             // contains EX_ENCRYPT_CBC_MODE_FLAG or EX_ENCRYPT_LRW_MODE_FLAG

	DWORD      offsetArrayNumber;
	DI_OFFSET  offsetArray[1];

} DISKINFO_BUFFER_V8_0, *PDISKINFO_BUFFER_V8_0;

#pragma pack()


// Map Logical Disk command definition

#define IOCTL_PRIVATE_MAP_BCDISK \
              CTL_CODE(IOCTL_DISK_BASE, 0x803, METHOD_IN_DIRECT, FILE_ANY_ACCESS)

// Disconnect BestCrypt Disk command & structure definition

#define IOCTL_PRIVATE_DISCONNECT_BCDISK \
              CTL_CODE(IOCTL_DISK_BASE, 0x804, METHOD_IN_DIRECT, FILE_ANY_ACCESS)


typedef struct  _DISCONNECT_BUFFER { char   Signature[8];	// input: "LOCOS97",00 ; output: "LS06CXX",00
                                     UCHAR  Status;
								     DWORD  DiskNumber;
									 BOOL   FullDismounted;
								   } DISCONNECT_BUFFER, *PDISCONNECT_BUFFER;


// Map Logical Disk Extended command - added highest 32-bit for header offset 
// to DISKINFO_BUFFER structure (2-Feb-2005).

#define IOCTL_PRIVATE_MAP_BCDISK_EXT \
              CTL_CODE(IOCTL_DISK_BASE, 0x806, METHOD_IN_DIRECT, FILE_ANY_ACCESS)


// Run one of the command specified in DISKINFO_BUFFER_V8_0 structure
// This command appears since BestCrypt v.8 and bcbus.sys v.4.01
#define IOCTL_PRIVATE_V8_COMMAND \
              CTL_CODE(IOCTL_DISK_BASE, 0x80A, METHOD_IN_DIRECT, FILE_ANY_ACCESS)


// Create or delete Symbolic Link for initialized Device Object

#define IOCTL_PRIVATE_CHANGE_SYMBOLIC_LINK \
              CTL_CODE(IOCTL_DISK_BASE, 0x807, METHOD_IN_DIRECT, FILE_ANY_ACCESS)

typedef struct  _CHANGELINK_BUFFER { char   Signature[8];	// input: "LOCOS97",00 ; output: "LS06CXX",00
                                     UCHAR  Status;
								     DWORD  DiskNumber;
									 BOOL   DeleteLink;
								   } CHANGELINK_BUFFER, *PCHANGELINK_BUFFER;


// Remove or restore NT security descriptor for the BestCrypt virtual 
// disk to allow the BestCrypt Control Panel make everything even if 
// user is not an administrator

#define IOCTL_PRIVATE_CHANGE_NT_SECURITY \
              CTL_CODE(IOCTL_DISK_BASE, 0x808, METHOD_IN_DIRECT, FILE_ANY_ACCESS)

typedef struct  _CHANGESEC_BUFFER { char   Signature[8];	// input: "LOCOS97",00 ; output: "LS06CXX",00
                                    UCHAR  Status;
							        DWORD  DiskNumber;
								    BOOL   RemoveSecurity;
							      } CHANGESEC_BUFFER, *PCHANGESEC_BUFFER;


// Request to BCBUS driver for Windows 2000/XP to ask about real disk number

#define IOCTL_PRIVATE_GET_DISK_NUMBER \
              CTL_CODE(IOCTL_DISK_BASE, 0x809, METHOD_IN_DIRECT, FILE_ANY_ACCESS)

typedef struct  _DISKNUMBER_BUFFER { char   Signature[8];	// input: "LOCOS97",00 ; output: "LS06CXX",00
                                    UCHAR  Status;
                                    DWORD  DiskNumber;
                                    DWORD  UserId;   // Ser added 25-Apr-2003
                                    DWORD  HarddiskVolumeNumber;
                                  } DISKNUMBER_BUFFER, *PDISKNUMBER_BUFFER;


// Ser added 25-Apr-2003: Request to BCBUS driver for Windows 2000/XP to save information about Mountmgr 
// number in the \Device\HarddiskVolumeN name

#define IOCTL_PRIVATE_SAVE_MOUNTMGR_DISK_NUMBER \
              CTL_CODE(IOCTL_DISK_BASE, 0x900, METHOD_IN_DIRECT, FILE_ANY_ACCESS)

typedef struct  _MOUNTMGR_DISKNUMBER_BUFFER { char   Signature[8];	// input: "LOCOS97",00 ; output: "LS06CXX",00
                                              UCHAR  Status;
                                              DWORD  DiskNumber;
                                              DWORD  HarddiskVolumeNumber;
                                            } MOUNTMGR_DISKNUMBER_BUFFER, *PMOUNTMGR_DISKNUMBER_BUFFER;


///////////////////////////////////////////////////////////


typedef struct  _BC_EJECT_BUFFER { char   Signature[8];	// input: "LOCOS97",00 ; output: "LS06CXX",00
                                   UCHAR  Status;
                                   DWORD  DiskNumber;
                                 } BC_EJECT_BUFFER, *PBC_EJECT_BUFFER;


#define IOCTL_PRIVATE_DISK_EJECT \
              CTL_CODE(IOCTL_DISK_BASE, 0x902, METHOD_IN_DIRECT, FILE_ANY_ACCESS)



///////////////////////////////////////////////////////////

#define IOCTL_PRIVATE_MODIFY_SYMBOLIC_LINK \
              CTL_CODE(IOCTL_DISK_BASE, 0x901, METHOD_IN_DIRECT, FILE_ANY_ACCESS)

typedef struct  _SYMBOLIC_LINK_BUFFER { char   Signature[8];	// input: "LOCOS97",00 ; output: "LS06CXX",00
                                        UCHAR  Status;
                                        DWORD  DiskNumber;
                                        BOOL   CreateSymbolicLink;
										WCHAR  SymbolicLinkName[128];
                                      } SYMBOLIC_LINK_BUFFER, *PSYMBOLIC_LINK_BUFFER;


///////////////////////////////////////////////////////////

// Ioctl is added to define numbers after Volume Arrival
// for given volume data (Signature + Offset) (Ser: 24-04-2005)

#define BCIOCTL_VOLUME_DATA_SIZE (sizeof(DWORD) + sizeof(LONGLONG))

#define IOCTL_PRIVATE_GET_DRIVE_NUMBERS \
              CTL_CODE(IOCTL_DISK_BASE, 0x908, METHOD_IN_DIRECT, FILE_ANY_ACCESS)

typedef struct  _GET_DRIVE_NUMBERS_BUFFER { char   Signature[8];	// input: "LOCOS97",00 ; output: "LS06CXX",00
                                            UCHAR  Status;
                                            DWORD  Size;
                                            BOOL   IsBestCryptDrive;
                                            DWORD  BcDriveNumber;
                                            DWORD  WindowsDriveNumber;
                                            DWORD  HarddiskVolumeNumber;
                                            DWORD  UserId;
                                            BYTE   DriveData[ BCIOCTL_VOLUME_DATA_SIZE ];
                                          } GET_DRIVE_NUMBERS_BUFFER, *PGET_DRIVE_NUMBERS_BUFFER;

//
//  Control codes for the "program -> algorithm driver" interface
//
//  This control codes are private for this driver, hence, 
//  functions number must be >= 0x800
//
//  Functions numbers for Algorithm Driver are 0x810 - 0x8FF
//  

// Expand key, save it, and return a handle for the given key

#define IOCTL_PRIVATE_CREATE_KEY_HANDLE \
              CTL_CODE(IOCTL_DISK_BASE, 0x810, METHOD_IN_DIRECT, FILE_ANY_ACCESS)


typedef struct  _CREATEKEY_BUFFER { char        Signature[8];	// input: "LOCOS97",00 ; output: "LS06CXX",00
                                    UCHAR       Status;
                                    DWORD       KeyLength;		// In bits...
                                    KEY_HANDLE  KeyHandle;
                                    UCHAR       Key[MAX_KEY_LENGTH / 8];
                                    UCHAR       Pool[POOL_SIZE_BYTES];
                                  } CREATEKEY_BUFFER, *PCREATEKEY_BUFFER;


// Test if the key for the given key handle exists

#define IOCTL_PRIVATE_TEST_KEY_HANDLE \
              CTL_CODE(IOCTL_DISK_BASE, 0x811, METHOD_IN_DIRECT, FILE_ANY_ACCESS)


typedef struct  _TESTKEY_BUFFER { char        Signature[8];	// input: "LOCOS97",00 ; output: "LS06CXX",00
                                  UCHAR        Status;
								  UCHAR        IsValid;
								  KEY_HANDLE  KeyHandle;
								} TESTKEY_BUFFER, *PTESTKEY_BUFFER;


// Free Key Handle command: uses a TESTKEY_BUFFER structure definition

#define IOCTL_PRIVATE_FREE_KEY_HANDLE \
              CTL_CODE(IOCTL_DISK_BASE, 0x812, METHOD_IN_DIRECT, FILE_ANY_ACCESS)


// Compare Keys command & structure definition

#define IOCTL_PRIVATE_COMPARE_KEYS \
              CTL_CODE(IOCTL_DISK_BASE, 0x813, METHOD_IN_DIRECT, FILE_ANY_ACCESS)


typedef struct  _COMPARE_KEYS_BUFFER { char        Signature[8];	// input: "LOCOS97",00 ; output: "LS06CXX",00
                                       UCHAR       Status;
                                       UCHAR       AreEqual;
                                       KEY_HANDLE  KeyHandle1;
                                       KEY_HANDLE  KeyHandle2;
                                     } COMPARE_KEYS_BUFFER, *PCOMPARE_KEYS_BUFFER;


// Encrypt command & structure definition

#define IOCTL_PRIVATE_ENCRYPT \
              CTL_CODE(IOCTL_DISK_BASE, 0x814, METHOD_IN_DIRECT, FILE_ANY_ACCESS)

              // Buffer for data is in the Output buffer of DeviceIoControl 

typedef struct  _ENCRYPT_BUFFER { char        Signature[8];	// input: "LOCOS97",00 ; output: "LS06CXX",00
                                  UCHAR       Status;
                                  KEY_HANDLE  KeyHandle;
                                  UCHAR       IVector[8];
                                  DWORD       Length;       // in bytes
                                } ENCRYPT_BUFFER, *PENCRYPT_BUFFER;

// Decrypt command definition

#define IOCTL_PRIVATE_DECRYPT \
              CTL_CODE(IOCTL_DISK_BASE, 0x815, METHOD_IN_DIRECT, FILE_ANY_ACCESS)


// Encrypt/Decrypt commands available since BestCrypt v.8

#define IOCTL_PRIVATE_ENCRYPT_DECRYPT_V8 \
              CTL_CODE(IOCTL_DISK_BASE, 0x81A, METHOD_IN_DIRECT, FILE_ANY_ACCESS)

              // Buffer for data is in the Output buffer of DeviceIoControl 

typedef struct  _ENCRYPT_BUFFER_V8 { char        Signature[8];     // input: "LOCOS97",00 ; output: "LS06CXX",00
                                     UCHAR       Status;
                                     DWORD       Size;             // Size of the structure 
                                     BOOL        EncryptOperation; // TRUE if encrypt, FALSE if decrypt
                                     KEY_HANDLE  KeyHandle;
                                     UCHAR       IVector[16];
                                     DWORD       Length;           // Length in bytes
                                     DWORD       EncryptionMode;   // EX_ENCRYPT_LRW_MODE_FLAG or EX_ENCRYPT_CBC_MODE_FLAG
                                   } ENCRYPT_BUFFER_V8, *PENCRYPT_BUFFER_V8;



// Get function address from Algorithm Driver
#define IOCTL_PRIVATE_GETFUNCTIONS \
              CTL_CODE(IOCTL_DISK_BASE, 0x816, METHOD_IN_DIRECT, FILE_ANY_ACCESS)

typedef struct  _GETFUNC_BUFFER { char        Signature[8];	// input: "LOCOS97",00 ; output: "LS06CXX",00
                                  UCHAR       Status;
                                  PVOID       FuncTestKey;
                                  PVOID       FuncFreeKey;
                                  PVOID       FuncEncrypt;
                                  PVOID       FuncDecrypt;
                                } GETFUNC_BUFFER, *PGETFUNC_BUFFER;

// Get version command & structure definition

#define IOCTL_PRIVATE_GET_INFO \
              CTL_CODE(IOCTL_DISK_BASE, 0x817, METHOD_IN_DIRECT, FILE_ANY_ACCESS)


typedef struct  _INFO_BUFFER { char   Signature[8];	// input: "LOCOS97",00 ; output: "LS06CXX",00
                               UCHAR  Status;
                               UCHAR  DisplayName[32];
                               DWORD  AlgorithmID;
                               DWORD  KeyLength;
} INFO_BUFFER, *PINFO_BUFFER;


// Check, whether the algorithm driver supports 64-bit sector number as Initial Vector 

#define IOCTL_PRIVATE_SUPPORTS_SYNCRO_64BITS \
              CTL_CODE(IOCTL_DISK_BASE, 0x818, METHOD_IN_DIRECT, FILE_ANY_ACCESS)


typedef struct  _SYNCRO64BITS_BUFFER { char   Signature[8];	// input: "LOCOS97",00 ; output: "LS06CXX",00
                                       UCHAR  Status;
                                       BOOL   SupportStatus;
} SYNCRO64BITS_BUFFER, *PSYNCRO64BITS_BUFFER;


// Get Extended Functions Addresses from Algorithm Driver.
// Uses the same GETFUNC_BUFFER.
#define IOCTL_PRIVATE_GETFUNCTIONS_EX \
              CTL_CODE(IOCTL_DISK_BASE, 0x819, METHOD_IN_DIRECT, FILE_ANY_ACCESS)


// Get information about Mode of Algorithm Driver.

#define IOCTL_PRIVATE_GET_MODE_INFO \
              CTL_CODE(IOCTL_DISK_BASE, 0x821, METHOD_IN_DIRECT, FILE_ANY_ACCESS)


typedef struct  _MODE_INFO_BUFFER { char   Signature[8];     // input: "LOCOS97",00 ; output: "LS06CXX",00
                                    UCHAR  Status;
                                    DWORD  Mode;             // Input parameter - one of EX_ENCRYPT_xxx_MODE_FLAG
                                    char   ModeName[32];
                                    DWORD  KeyLength;        // in bites! - for the mode

                                    DWORD  Reserved_1;
                                    DWORD  Reserved_2;       // Reserved_N is 0 if not used.
                                    DWORD  Reserved_3;
                                    DWORD  Reserved_4;
} MODE_INFO_BUFFER, *PMODE_INFO_BUFFER;



// Get information about keyboard + mouse + disk activity
#define IOCTL_PRIVATE_GET_AND_RESET_FLAG \
              CTL_CODE(IOCTL_DISK_BASE, 0x855, METHOD_IN_DIRECT, FILE_ANY_ACCESS)


typedef struct  _FLAG_BUFFER { char  Signature[8];	// input: "LOCOS97",00 ; output: "LS06CXX",00
                               UCHAR Status;
                               DWORD Flag;
                             } FLAG_BUFFER, *PFLAG_BUFFER;


// Set or Reset filesystem hooks

#define IOCTL_FSH_REFRESH_HOOK \
              CTL_CODE(IOCTL_DISK_BASE, 0x820, METHOD_IN_DIRECT, FILE_ANY_ACCESS)

typedef struct { char    Signature[8];	// input: "LOCOS97",00 ; output: "LS06CXX",00
                 UCHAR   Status;
                 ULONG   HookMask;
               } REFRESH_BUFFER, *PREFRESH_BUFFER;



// Get information about symbolic link
// This command appears since BestCrypt v.8 and bcbus.sys v.4.04 (19-Dec-2006)
#define IOCTL_PRIVATE_GET_SYMBOLIC_LINK_INFO \
              CTL_CODE(IOCTL_DISK_BASE, 0x80B, METHOD_IN_DIRECT, FILE_ANY_ACCESS)

typedef struct  _LINK_INFO_BUFFER { char   Signature[8];	// input: "LOCOS97",00 ; output: "LS06CXX",00
                                    UCHAR  Status;
                                    WCHAR  LinkNameInput[256];
                                    WCHAR  OutputName[256];
                                  } LINK_INFO_BUFFER, *PLINK_INFO_BUFFER;

#ifndef FSCTL_SET_REPARSE_POINT
#define FSCTL_SET_REPARSE_POINT         CTL_CODE(FILE_DEVICE_FILE_SYSTEM, 41, METHOD_BUFFERED, FILE_SPECIAL_ACCESS) // REPARSE_DATA_BUFFER,
#define FSCTL_GET_REPARSE_POINT         CTL_CODE(FILE_DEVICE_FILE_SYSTEM, 42, METHOD_BUFFERED, FILE_ANY_ACCESS) // REPARSE_DATA_BUFFER
#define FSCTL_DELETE_REPARSE_POINT      CTL_CODE(FILE_DEVICE_FILE_SYSTEM, 43, METHOD_BUFFERED, FILE_SPECIAL_ACCESS) // REPARSE_DATA_BUFFER,
#endif

#endif

