#include "cCertificateLibEncryptionKey.h"

cADBlock < cCertificateLibEncryptionKey > cCertificateLibEncryptionKeyManager::m_keyT;
DWORD cCertificateLibEncryptionKeyManager::m_lastId = 0;
HANDLE cCertificateLibEncryptionKeyManager::m_hSemaphore = NULL;

bool cCertificateLibEncryptionKeyManager::AddToTable( cCertificateLibEncryptionKey *alg )
{
	CheckSemaphore();

	KEY_HANDLE id = ++m_lastId;
	if ( !m_keyT.Add( alg ) )
		return false;

	alg->m_keyID = id;

	ReleaseSemaphore();

	return true;
}

cCertificateLibEncryptionKey *cCertificateLibEncryptionKeyManager::FindKeyById( DWORD keyId, bool bDelete )
{
	CheckSemaphore();

	int found = -1;

	DWORD maxId = 0;

	for( int i = 0; i < m_keyT.m_Q; i++ )
	{
		maxId = max( maxId, m_keyT.m_T[i].m_keyID );

		if ( m_keyT.m_T[i].m_keyID == keyId )
		{
			if ( found != -1 )
			{
				ODS_fls( L"Assertion!" );
				found = -1;
				break;
			}
		}
	}

	if ( bDelete && ( found != - 1 ) )
	{
		m_keyT.Remove( found );
		found = -1;
	}

	ReleaseSemaphore();

	if ( found == -1 )
		return NULL;

	return m_keyT.m_T + found;
}

const wchar_t *szSemaphoreName = L"Global\\EncryptionKeySemaphore";

void cCertificateLibEncryptionKeyManager::CheckSemaphore()
{
	if ( m_hSemaphore == NULL )
	{
		HANDLE hSemaphore = CreateSemaphoreT( NULL, 0, 1, szSemaphoreName );

		if ( hSemaphore == NULL )
		{ ODS_fls( L"Assertion!" ); return; }

		if ( GetLastError() == ERROR_ALREADY_EXISTS )
			CloseHandle( hSemaphore );
		else
		{
			m_hSemaphore = hSemaphore;
			LONG lPrev = 0;
			::ReleaseSemaphore( m_hSemaphore, 1, &lPrev ); // Allow to proceed
			static cAutoHandle ah( m_hSemaphore );
		}
	}

	DWORD dwRes = WaitForSingleObject( m_hSemaphore, INFINITE );
}

void cCertificateLibEncryptionKeyManager::ReleaseSemaphore()
{
	LONG lPrev = 0;
	::ReleaseSemaphore( m_hSemaphore, 1, &lPrev );
}
