/*******************************************************************
 *
 *	Copyright (c) 2007-2015 Jetico Inc., Finland
 *	All rights reserved.
 *
 *	File: kblock.cpp
 *	Revision: 
 *	Created:
 *	Description: implementation of procedures to work with 
 *		Data Block that stores encryption keys
 *
 *******************************************************************/

#include <windows.h>
#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <time.h>
#include <commctrl.h>
#include "UnicodeFuncs.h"
#include "kblock.h"
#include "..\..\Comm_def\Kg_defs.h"
#include "..\..\..\CertificateLib\crtLibErrs.h"
#include "..\..\..\CertificateLib\RNG\sha1random.h"
#include "..\..\..\CertificateLib\hashAlgs\hashAlgs.h"
#include "..\..\..\CertificateLib\PGPPkts\PGPPkts.h"
#include "..\..\..\CertificateLib\x509pkts\pkcs12_funcs.h"
#include "..\..\..\CertificateLib\symmetricAlgs\symmetricAlgs.h"
#include "alg.h"
#include "cKeyBlockParams.h"
#include "kgGhost.h"
#include "resource.h"
#include "ods.h"
#include "Bc_ioctl.h"
#include "..\..\..\Includes\cStringBuffer.h"
#include "..\..\..\Includes\sysRand.h"

//static const DWORD maxHPBlockSize = _4K;
//
//bool dbContInfo::isForHiddenPart()
//{
//	return length == maxHPBlockSize;
//}
//
//bool dbContInfo::isValid()
//{
//	return length >= maxHPBlockSize;
//}

//
//	If pHashAlgID is NULL or if *pHashAlgID is invalid ( f.e. -1 )
//	It enumerates all available hash algorithms and and tries to verify.
//
bool verifyDigest( DWORD *pHashAlgID, BYTE *digest, BYTE *bf, DWORD bfSize )
{
	DWORD hashAlgId = -1;

	if ( pHashAlgID != NULL )
		hashAlgId  = *pHashAlgID;

	cEnumHashAlgIDs &hashAlgs = cEnumHashAlgIDs::getStatic();

	// Checks validity.
	bool valid = hashAlgs.checkValid( hashAlgId );

	// Verifies with valid algorithm.
	if ( valid )
	{
		if ( VerifyDigest( hashAlgId, digest, bf, bfSize ) )
			return true;

#ifdef _DEBUG
		//
		//	Excludes enumeration for predefined hash algorithm.
		//	So we have to disable exchenging of hash algorithm if password is added or hidden part is created.
		//
		for ( int i = 0; i<hashAlgs.Q; i++ )
		{
			hashAlgId = hashAlgs.T[i];

			if ( VerifyDigest( hashAlgId, digest, bf, bfSize ) )
			{
				if ( pHashAlgID != NULL )
					*pHashAlgID = hashAlgId;
				
				assert( false );
			}
		}
#endif
		return false;
	}

	// Enumerates all avilable.
	for ( int i = 0; i<hashAlgs.Q; i++ )
	{
		hashAlgId = hashAlgs.T[i];

		if ( VerifyDigest( hashAlgId, digest, bf, bfSize ) )
		{
			if ( pHashAlgID != NULL )
				*pHashAlgID = hashAlgId;
			
			return true;
		}
	}

	return false;
}

BOOL CalculateDigest( DWORD hashAlgID, BYTE *digest, int digestSize, BYTE *buff_0, int buffSize_0, BYTE *buff_1, int buffSize_1 )
{
//	int res = ha_makeDigestEx( hashAlgID, Digest, Data, Length );
//	return ( res == hae_ok );

	HHA ha;
	int res = ha_createAlgorithm( &ha, hashAlgID );
	if ( res != hae_ok )
		return false;

	if ( ha_getDigestSize( ha ) > digestSize )
	{
		ha_free( ha );
		return false;
	}

	res = ha_init( ha );

	if ( res == hae_ok )
		res = ha_process( ha, buff_0, buffSize_0 );
	
	if ( ( res == hae_ok ) && ( buff_1 != NULL ) && ( buffSize_1 != 0 ) )
		res = ha_process( ha, buff_1, buffSize_1 );

	if ( res == hae_ok )
		res = ha_final( ha, digest );

	ha_free( ha );

	return ( res == hae_ok );
}

BOOL VerifyDigest( DWORD hashAlgID, BYTE *pDigest, BYTE *buff, int buffSize, BYTE *buff_1, int buffSize_1 )
{
	int hashSize = 0;
	if ( ha_getDigestSize( &hashSize, hashAlgID ) != hae_ok )
	{
		assert( false );
		return false;
	}

	cPTB ptbDigest;
	BYTE *digest = ptbDigest.Allocate( hashSize );

	if ( digest == NULL )
	{
		assert( false );
		return false;
	}

	if ( !CalculateDigest( hashAlgID, digest, hashSize, buff, buffSize, buff_1, buffSize_1 ) )
	{
		assert( false );
		return false;
	}

	if ( memcmp( pDigest, digest, hashSize ) != 0)
	{
		return FALSE;
	}

	return TRUE;
}

void ShredData( BYTE *data, int dataLength )
{
	//int i;

	if ( (!data) || (dataLength < 0) )
	{
		assert( false );
		return;
	}

	// 1. Initial variant
	//static DWORD dwOld = 0; // It happed that ShredData was called within one time-clock interval 

	//DWORD dwInit;
	//do {
	//	ULARGE_INTEGER uli;
	//	uli.QuadPart = (DWORDLONG)time( NULL ) * (DWORDLONG)clock();
	//	dwInit = uli.LowPart ^ uli.HighPart;
	//}
	//while( dwInit == dwOld );
	//
	//dwOld = dwInit;

	//srand( dwInit );

	//for( i = 0; i < dataLength; i++ )
	//	data[i] = (BYTE)rand();


	// 2. 12.04.2016
	DWORDLONG dwl = sRandCtx::getTimeSeed();

	static DWORD count = 0;
	count++;

	P12_producePseudoRandomBytes( data, dataLength, (BYTE*)&count, sizeof(count), 1, (BYTE*)&dwl, sizeof(dwl), P12_prgm4IV, pgphaSHA256 );

	// 3. Pepok sugdested to use CryptGenRandom

	//???
	//CryptGenRandom
}

BOOL DataBlockInit(
		DATA_BLOCK	*db,
		DWORD		dbSize,
		DWORD		algId,
		DWORD		encModeId,
		DWORD		hashAlgId,
		cKeyBlockParams	&kbParams,
		sSpaceUnitBase	*pContInfo
	)
{
	ODSFuncS( _FormatT( L"algId 0x%X, encModeId 0x%X, hashAlgId 0x%X", algId, encModeId, hashAlgId ) );

	ShredData( (BYTE *)db, dbSize ); // fills DataBlock by random garbage

	db->header.m_body.InitFirst();

	db->header.m_body.contAlgId = algId;
	db->header.m_body.encModeId = encModeId;
	db->header.m_body.hashAlgId = hashAlgId;
	db->header.m_body.wKeyGenId = KEYGEN_ID;
	db->header.m_body.setVersion( DB_FormatVersion );

	memset( db->header.m_body.description, 0, sizeof(db->header.m_body.description) );

	//copyEx < dbContInfo > ( &(db->header.m_body.contInfo), pContInfo );
	copyEx < sSpaceUnitBase > ( &(db->header.m_body.contInfo), pContInfo );

	for( int i = DB_maxKeyQ; i--; )
		db->header.m_body.keyMap[i].Init();

	if ( kbParams.kbPBE.isKeyGenV5() )
	{
		db->header.m_body.setKeyGenIdV5( kbParams.kbPBE.m_iterations );
		db->updateSalt( &kbParams, false, KEYGEN_ID_V5 );
	}

	return TRUE;
}

//
//	encryptHeader
//		1. Checks if header is encrypted
//			1.1 if it is encrypted with the key simple copies eencrypted db to encrypted Buffer
//			1.2 if the key does not decrypts the "encrypted" db returns false
//		2. Encrypts the plain text db to enc buffer
//
//	returns error code
//
static DWORD encryptHeader( DATA_BLOCK *dbOut, DATA_BLOCK *dbIn, sKeyInfo *keyInfo )
{
	if ( dbHeader::isEncrypted( dbIn->header.m_body ) )
	{
		// Simple check decryption by the key info
		if ( !dbIn->header.decryptBodyTo( NULL, keyInfo, dbIn->pool ) )
			return ERROR_DAMAGED_DATA_BLOCK;

		if ( dbOut != dbIn )
			memcpy( dbOut, dbIn, sizeof( dbOut->header ) );

		return ERROR_NO;
	}

	cPlainTextBlock ptb;
	if ( ptb.Allocate( sizeof( tcEncBlock < dbHeader > ) ) == NULL )
	{
		assert( false );
		return ERROR_INTERNAL_PROBLEM;
	}

	if ( !dbIn->header.encryptTo( ( tcEncBlock < dbHeader > *)ptb.buffer, keyInfo, dbIn->pool ) )
	{
		assert( false );
		return ERROR_INVALID_ALGORITHM;
	}

	memcpy( dbOut, ptb.buffer, ptb.bufferSize );

	return ERROR_NO;
}

////
////	decryptHeader
////		Checks if header is not encrypted
////		1. Verifies header and copies the plain text header to outHeader
////		2. Decrypts with verification the encrypted db to outHeader
////
////	Returns error code
////
//static DWORD decryptHeader( dbHeader *outHeader, DATA_BLOCK *dbIn, DWORD algId, DWORD encModeId, KEY_HANDLE keyHandle )
//{
//	if ( !dbIn->header.m_body.IsEncrypted() )
//	{
//		if ( outHeader != &(dbIn->header.m_body) )
//			memcpy( outHeader, &(dbIn->header.m_body), sizeof( dbHeader ) );
//
//		if ( !dbIn->header.verifyTail( algId, encModeId, keyHandle ) )
//			return ERROR_INCORRECT_BLOCK_SIGNATURE;
//	}
//	else
//		if ( !dbIn->header.decryptBodyTo( outHeader, algId, encModeId, keyHandle ) )
//			return ERROR_INCORRECT_BLOCK_SIGNATURE;		
//
//	return ERROR_NO;
//}
//
//static DWORD decryptHeader( dbHeader *outHeader, DATA_BLOCK *dbIn, sKeyInfo *keyInfo )
//{
//	cSymAlgorithm cAlg;
//	if ( !cAlg.createKey( keyInfo->keyAlgId, keyInfo->key, keyInfo->getKeySizeBytes(), dbIn->pool ) )
//		return ERROR_INVALID_ALGORITHM;
//
//	return decryptHeader( outHeader, dbIn, cAlg.m_algId, cAlg.m_keyHandle ); 
//}

bool EncryptHeader(
		HWND			hWnd,
		const wchar_t	*contPath,
		tcEncBlock < dbHeader >
					*db,
		DWORD		dbSize,
		DWORD		*pErrCode,
		DWORD		*pFlags,
		sCKHParams	*ps
		//dbContInfo	*mainContInfo,
		//dbContInfo	*contInfo,
		//DWORD		*pContAlgId,	// [out], if container is encrypted the contAlgId value is -1,
		//							// but we have to return real algorithm id back
		//DWORD		*pEncModeId,	// [out], returns here container encryption mode id ( LRW or CBC )
		//DWORD		*pHashAlgId	// [out], returns here id of the hash algorithm is used for key block verification,
	)
{
	KEY_HANDLE keyHandle;
	
	sCKHParams	psTmp;
	if ( ps == NULL )
		ps = &psTmp;

	bool isEncrypted = dbHeader::isEncrypted(db->m_body);

	bool res = 
		CreateHandle_MainPart( 
			hWnd,
			contPath,
			_LoadIDS( isEncrypted ? IDS_ENTER_PASSWORD_FOR_DECRYPTION : IDS_ENTER_PASSWORD_FOR_ENCRYPTION ),
			(DATA_BLOCK **)&db,
			&dbSize,
			NULL,
			&keyHandle,
			ps );

	if ( !res )
		return false;

	if ( isEncrypted )
	{
		res = decryptHeaderEx( 
				(DATA_BLOCK *)db,
				ps->contAlgId, ps->encModeId, &(ps->hashAlgId), keyHandle, 
				(tcEncBlock < dbHeader > *)db,
				pErrCode );

//		*pHashAlgId = hashAlgId;
	}
	else
		res = encryptHeaderEx( 
				(DATA_BLOCK *)db,
				ps->contAlgId, ps->encModeId, ps->hashAlgId, keyHandle, 
				(tcEncBlock < dbHeader > *)db,
				pErrCode,
				true );
	
	FreeKeyHandle( ps->contAlgId, keyHandle );

	return res;
}
/*
static bool decryptPool( DATA_BLOCK	*dataBlock, DWORD algId, KEY_HANDLE keyHandle, BYTE	*decryptedPool )
{
	if ( decryptedPool == NULL )
		return false;

	BYTE iv[ IVECTOR_LENGTH_DB ];
	memset( iv, 0, sizeof( iv ) );
	return cSymAlgorithm::decrypt( iv, decryptedPool, dataBlock->pool, DB_PoolSize, algId, keyHandle ) )
}

static bool decryptPool( DATA_BLOCK	*dataBlock, sKeyInfo *keyInfo, BYTE	*decryptedPool )
{
	cSymAlgorithm cAlg;
	if ( !cAlg.createKey( keyInfo->keyAlgId, keyInfo->key, keyInfo->getKeySizeBytes(), dataBlock->pool ) )
		return false;

	return decryptPool( dataBlock, decryptedPool, algId, keyHandle );
}
*/

//
//	verifyKeyHandle
//		Any create key handle function has a ready key handle.
//		The function allows to verify the handle before using or bata block updating.
//
DWORD verifyKeyHandle( DATA_BLOCK *db, KEY_HANDLE keyHandle, DWORD algId, DWORD encModeId, dbHeader *decHeaderOut )
{
	cPlainTextBlock ptb;

	tcEncBlock < dbHeader > *decHeader = (tcEncBlock < dbHeader > *)ptb.Allocate( sizeof( tcEncBlock < dbHeader > ) );

	if ( decHeader == NULL )
		return ERROR_BUFFER_TOO_SMALL;

	DWORD err = ERROR_NO;

	if ( !decryptHeaderEx( db, algId, encModeId, NULL, keyHandle, decHeader, &err ) )
		return err;

	copyEx( decHeaderOut, &(decHeader->m_body ) );

	return ERROR_NO;
}

//
//	decryptAndVerifyHeader
//
DWORD
	decryptAndVerifyHeader(
		DATA_BLOCK	*dataBlock,
		sKeyInfo	*keyInfo,
		tcEncBlock < dbHeader > 
					*decryptedHeader,
		KEY_HANDLE	*pDelegateKeyHandle		// = NULL, // key handle to be delegeted
	)
{
	cSymAlgorithm cAlg;
	if ( !cAlg.createKey( keyInfo->keyAlgId, keyInfo->encModeId, keyInfo->key, keyInfo->getKeySizeBytes(), dataBlock->pool ) )
	{
		assert( false );
		return ERROR_INVALID_ALGORITHM;
	}

	DWORD err;

	DWORD hashAlgId = keyInfo->hashAlgId;

	bool ok = decryptHeaderEx( dataBlock, cAlg.m_algId, cAlg.m_encModeId, &hashAlgId, cAlg.m_keyHandle, decryptedHeader, &err );

	bool bAllowDelegation = ok;
#ifdef _allowBadHeader
	if ( err == ERROR_INCORRECT_BLOCK_SIGNATURE )
		bAllowDelegation = true;
#endif
	if ( bAllowDelegation && ( pDelegateKeyHandle != NULL ) )
		cAlg.delegateHandle( pDelegateKeyHandle );

	if ( !ok )
		return err;

	return ERROR_NO;
}

//
//	isValid
//		simple checks size, hashAlgId and keyAlgId
//
bool sKeyInfo::isValid()
{
	bool validSize = ( size <= sizeof( *this ) ) && ( size >= ( sizeof( *this ) - sizeof ( key ) ) );

	if ( !validSize )
		return false;

	int hashSize = 0;
	if ( ha_getDigestSize( &hashSize, hashAlgId ) != hae_ok )
		return false;

	if ( !Alg_GetAlgInfoByID(  keyAlgId, encModeId ) )
		return false;

	return true;
}

int sKeyInfo::getBufferSize( int keySize )
{
	return ( sizeWithoutKey + keySize );
}

int sKeyInfo::getAlgKeySizeBytes( int contAlgId, DWORD encModeId )
{
	DWORD keyLengthInBites = -1;

	if ( !Alg_GetAlgInfoByID( contAlgId, encModeId, NULL, 0, &keyLengthInBites ) )
		return -1;

	return ( keyLengthInBites + 7 ) / 8;
}

int sKeyInfo::getAlgKeySizeBytes()
{
	return getAlgKeySizeBytes( keyAlgId, encModeId );
}

int sKeyInfo::getKeySizeBytes()
{
	return size - sizeWithoutKey;
}


// 
// DataBlockAllocate() - allocates memory for DATA_BLOCK structure 
//
// The procedure which calls DataBlockAllocate() is responsible for 
// releasing the memory with the DataBlockFree() function.
//
BOOL DataBlockAllocate( 
		DATA_BLOCK	**pDataBlock,
		DWORD		*pDataSize
	)
{
	if ( *pDataSize == 0 )
		*pDataSize = DB_MinSize;

	*pDataSize = DRound( *pDataSize, _4K );

	if ( *pDataSize > DB_MaxSize )
		*pDataSize = DB_MaxSize;
	
	*pDataBlock = (DATA_BLOCK *)VirtualAlloc( NULL, *pDataSize, MEM_COMMIT, PAGE_READWRITE );
	
	if (!(*pDataBlock)) return FALSE;

	ShredData( (BYTE *)*pDataBlock, *pDataSize );

	return TRUE;
}

BOOL DataBlockReAllocate( 
		DATA_BLOCK	**ppDB,
		DWORD oldSize,
		DWORD *pNewSize
	)
{
	*pNewSize = DRound( *pNewSize, _4K );

	if ( oldSize > *pNewSize )
		ShredData( (BYTE *)*ppDB  + *pNewSize, oldSize - *pNewSize );

	DATA_BLOCK	*pDBNew;

	if ( !DataBlockAllocate( &pDBNew, pNewSize ) )
		return FALSE;
	
	memcpy( pDBNew, *ppDB, oldSize );
//	ShredData( (BYTE *)*pDBNew  + oldSize, *pNewSize - oldSize );

	DataBlockFree( ppDB, &oldSize );
	
	*ppDB = pDBNew;
	
//	(*ppDB)->header.m_body.keyBlockSize = *pNewSize;

	return TRUE;
}

bool DATA_BLOCK::updateSalt( cKeyBlockParams *kbParams, bool bReadFromDBToKbParams, int kgID )
{
	int slotNum = DEFAULT_SALT_SLOT;
	
	if ( bReadFromDBToKbParams )
	{
		if ( kgID == KEYGEN_ID_V5 )
		{
			int size = header.m_body.keyMap[slotNum].kbSize;
			if ( kbParams->kbPBE.m_hashAlgId == (DWORD)-1 )
				size = sizeof( keyTab[slotNum] ); // Entire key tab slot for encrypted header.
			else
				ha_getDigestSize( &size, kbParams->kbPBE.m_hashAlgId );

			kbParams->kbPBE.setSalt( keyTab[slotNum], size );
		}
		else
		{
			kbParams->kbPBE.setSaltSize( SALT_SIZE );
			kbParams->kbPBE.updateSalt();
		}
	}
	else
	{
		if ( dbHeader::isEncrypted(header.m_body) ) return false;
		if ( header.m_body.wKeyGenId != KEYGEN_ID_V5 ) return false;

		if ( header.m_body.keyMap[slotNum].kbType == kbtSalt )
			header.m_body.freeKeyBlockChain( this, _4K, header.m_body.keyMap, slotNum );

		WORD size = (WORD)kbParams->kbPBE.getSaltSize();
		if ( size != kbParams->kbPBE.getSaltSize() )
		{ assert( false ); return false; }

		int pos = header.m_body.findEmptySpace( header.m_body.keyMap, size, slotNum );

		if ( pos != slotNum )
		{ assert( false ); return false; }

		if ( !header.m_body.MarkBusied( slotNum, -1, kbtSalt, size ) )
		{ assert( false ); return false; }

		memcpy( keyTab[slotNum], kbParams->kbPBE.m_salt.buffer, size );
	}

	return true;
}

									
/**********************************************************************
 * 
 * DataBlockFree() - wipes and frees DATA_BLOCK structure 
 *
 * The procedure which calls DataBlockAllocate() is responsible for 
 * releasing the memory with the DataBlockFree() function.
 *
 **********************************************************************/

BOOL DataBlockFree( DATA_BLOCK **DataBlock, DWORD *pDataSize )
{
	ODS_fls( _FormatT( L"0x%X", *DataBlock ) );

	if ( pDataSize != 0 )
	{
		ShredData( (BYTE *)*DataBlock, *pDataSize );
		VirtualFree( *DataBlock, *pDataSize, MEM_DECOMMIT );
	}

	VirtualFree( *DataBlock, 0, MEM_RELEASE);
	*DataBlock = NULL;
	return TRUE;
}

//
//	DataBlockCheck
//		Verifies that DATA_BLOCK structure is not damaged
//		and checks that DATA_BLOCK stores at least 
//		KeysCount keys.
//
DWORD DataBlockCheck(
		DATA_BLOCK *DataBlock,
		DWORD      DataSize,
		DWORD      AlgId
	)
{
	ODSFunc;

	if ( ( DataBlock == NULL ) || ( DataSize < sizeof( DataBlock->header ) ) )
		return ERROR_DAMAGED_DATA_BLOCK;
	
	if ( AlgId == -1 ) // Allows all available encryption algorithms
		return ERROR_NO;

	if ( strncmp( DataBlock->header.m_body.signature, dbSignature, sizeof( DataBlock->header.m_body.signature ) ) != 0 )
		return ERROR_NO; // Encrypted header has encrypted signature ERROR_INCORRECT_BLOCK_SIGNATURE;

	if ( DataBlock->header.m_body.getVersion() != DB_FormatVersion )
		return ERROR_INCORRECT_FORMAT_VERSION;

	if ( DataBlock->header.m_body.contAlgId != AlgId )
		return ERROR_INCORRECT_ENCR_ALGORITHM;

	return ERROR_NO;
}

/**********************************************************************
 * 
 * DataBlockWriteKey() - fills in a free KEY_BLOCK in the DataBlock.
 *
 **********************************************************************/

DWORD
	DataBlock_CheckFreeSpace(
		DATA_BLOCK	*dataBlock,
		DWORD		requiredSize,
		DWORD		*pMaxFreeSize,
		int			treatAsFree
	)
{
	return dbHeader::checkFreeSpace( dataBlock->header.m_body.keyMap, requiredSize, pMaxFreeSize, treatAsFree );
}


//
//	DataBlockWriteKey
//
//		Writes new key to key table.
//		Note, the data block may be extended so data block pointer and it's size may be changed!
//
//		1. Writes encoded from kbParams key to key table
//			1.1 If kbParams is NULL it does not write key block, 
//				it simple free the key block chain started form kbNumber.
//			1.2 If keyMapInfo.position == -1 it writes key to any empty location
//				otherwise it shreds specified key block number and writes the new one.
//
//		2. Updates key map table in dataBlock header
//
//		3. Calculates digest for entire data block header->m_body and encrypts info.
//
//			3.1 If bEncryptIt is FALSE
//				a. Calculates digest of DATA_BLOCK::dbHeader->m_body to DATA_BLOCK::dbHeader->m_digest field
//
//			3.2 If bEncryptIt is TRUE
//				a. Calculates digest of DATA_BLOCK::dbHeader->m_body to DATA_BLOCK::dbHeader->m_digest field
//				b. Encrypts DATA_BLOCK::dbHeader
//
DWORD DataBlockWriteKey(
		DATA_BLOCK			**ppDB,		// completely or partialy encrypted data block
		DWORD				*pDBSize,
		bool				bEncryptIt,	// usually the header tail is encrypted only
										// if the flag is true so entire header are completely encrypted.
		sKeyInfo			*keyInfo,	// = NULL, key for data block updating
		cKeyBlockParams		*kbParams,	// = NULL, key block parameters required to encode keyInfo to the key block has
										// to be written to output data block

		sKeyMapInfo			*keyMapInfo,// if is  NULL or has undefined position -1, writes new key to first empty position
										// if not NULL shreds existing position first, fills the key map info with new data.
		dbKeyId				*keyMapT,	// if not NULL it contains information even about hidden part locations 
										// ( if theirs passwords were accepted see AddNewKey ChangePassword )
		dbHeader			*decHeader,	// if not NULL copy here updated header info ( is decrypted )
		bool				bDataBlockIsDamaged // It is true at ChangePassword only.
	)
{
	ODSFunc;

	DATA_BLOCK *db = *ppDB;
	
	sKeyMapInfo kmiTmp;
	if ( keyMapInfo == NULL )
		keyMapInfo = &kmiTmp;

	if ( keyInfo == NULL )
	{
		assert( false );
		ODS_fls( L"Asertion!" );
		return ERROR_INTERNAL_PROBLEM;
	}

	// Allocates place for decrypted and temporary data.

	cPlainTextBlock ptb;

	int encodedKeySize = 0;

	if ( kbParams != NULL )
	{
		// Version 2.01.5.2 - 2.11.2016 
		//		Bug #3043 SSS: undefined key generator error
		//		We have to copy and reinit SSS parameters
		kbParams->kbSSS.SetEncIds( &(kbParams->kbPBE) );
		encodedKeySize = kbParams->getEncodedKeySize( keyInfo->size );
	}

	//
	//	Allocates maximum size datablock
	//
	DWORD dbSize = DB_MaxSize;

	if ( ptb.Allocate(
			dbSize + 
			encodedKeySize +
			sizeof( sKeyInfo ) +
			0 ) == NULL )
	{
		assert( false );
		ODS_fls( L"Asertion!" );
		return ERROR_INTERNAL_PROBLEM;
	}

	int pos = 0;
	DATA_BLOCK    *dbTmp = (DATA_BLOCK *)(ptb.buffer + pos); pos += dbSize;
	BYTE	  *kbEncoded =				 (ptb.buffer + pos); pos += encodedKeySize;
	sKeyInfo *keyInfoTmp = (sKeyInfo *)  (ptb.buffer + pos); pos += sizeof( sKeyInfo );

	assert( pos == ptb.bufferSize );

	// Decrypts cont info and copies it to temporary buffer

	memcpy( (BYTE *)dbTmp + sizeof( dbTmp->header ), (BYTE *)db + sizeof( db->header ), *pDBSize - sizeof( db->header ) );
	ShredData( (BYTE *)dbTmp + *pDBSize, dbSize - *pDBSize );
	ShredData( kbEncoded, encodedKeySize );

	DWORD dwRes = decryptAndVerifyHeader( db, keyInfo, &(dbTmp->header) );
	
	if ( ( dwRes != ERROR_NO ) &&
		( dwRes != ERROR_INCORRECT_BLOCK_SIGNATURE ) )
	{
		ODSFuncRes( dwRes );
		return dwRes;
	}

	// Finds empty KEY_BLOCK inside DataBlock

	int num = keyMapInfo->position;
	int req = dbKeyId::GetRequiredQ( encodedKeySize );

	// Frees existing key block ( shreds datablock and marks empty key map table )
	if ( keyMapInfo->position != -1 )
	{
		dbHeader::freeKeyBlockChain( dbTmp, dbSize, dbTmp->header.m_body.keyMap, keyMapInfo );
		keyMapInfo->setUndefined();
	}

	if ( kbParams != NULL )
	{
		if ( keyMapT == NULL )
			keyMapT = dbTmp->header.m_body.keyMap;

		if ( !bDataBlockIsDamaged )
			num = dbHeader::findEmptySpace( keyMapT, req );

		if ( num == -1 )
		{
			ODS_fls( L"Asertion!" );
			return ERROR_NOT_ENOUGH_SPACE_FOR_KEY;
		}

		if ( !kbParams->encodeKey( kbEncoded, encodedKeySize, keyInfo, sizeof( *keyInfo ) ) ) // keyInfo->size ) )
		{
			assert( false );
			ODS_fls( L"Asertion!" );
			return ERROR_INTERNAL_PROBLEM;
		}

		ODS_fls( _FormatT( L"Encoded key 0x%08I64X, size 0x%X, pos %d, type %d", *(DWORDLONG*)kbEncoded, encodedKeySize, num, kbParams->kbType ) );

#ifdef _DEBUG
		bool res = kbParams->kbType == kbtPKE; // public key cannot verify it's own encryption
		while( !res )
		{
			res = kbParams->decodeKey( keyInfoTmp, sizeof( sKeyInfo ), kbEncoded, encodedKeySize );
			if ( res )
				res = ( memcmp( keyInfo, keyInfoTmp, keyInfo->size ) == 0 );

			assert( res );
		}
#endif

		// Verify 
		if ( kbParams->kbType != kbtPKE ) // public key cannot verify it's own encryption
		{
			if ( !kbParams->decodeKey( keyInfoTmp, sizeof( sKeyInfo ), kbEncoded, encodedKeySize ) )
			{
				assert( false );
				ODS_fls( L"Asertion!" );
				return ERROR_INTERNAL_PROBLEM;
			}

			if ( memcmp( keyInfo, keyInfoTmp, keyInfo->size ) != 0 )
			{
				assert( false );
				ODS_fls( L"Asertion!" );
				return ERROR_INTERNAL_PROBLEM;
			}
		}

		BYTE *kb = dbTmp->keyTab[ num ];

		// Encode keyInfo by kbParams directly to the place in data block was found.

		if ( kb + encodedKeySize > ( BYTE* )dbTmp + dbSize )
		{
			assert( false );
			ODS_fls( L"Asertion!" );
			return ERROR_INTERNAL_PROBLEM;
		}

		memcpy( kb, kbEncoded, encodedKeySize );

		keyMapInfo->setInfo( num, 1, kbParams->kbType, encodedKeySize );

		dbHeader::MarkBusied( dbTmp->header.m_body.keyMap, keyMapInfo );
	}

#ifdef _allowBadHeader
	if ( bDataBlockIsDamaged )
	{
		if ( num >= DB_minKeyQ )
		{
			ODS_fls( L"Asertion! Damaged data block cannot be more then 4K" );
			return ERROR_NOT_ENOUGH_SPACE_FOR_KEY;
		}
		*pDBSize = DB_MinSize;
		memcpy( (*ppDB)->keyTab[ num ], dbTmp->keyTab[ num ], keyMapInfo->kbSize );
		return ERROR_NO;
	}
#endif

	// Gets realy busied size
	DWORD realSize = dbTmp->header.m_body.getRealSize();

	if ( realSize == DWORD(-1) )
	{
		assert( false );
		ODS_fls( L"Asertion!" );
		return ERROR_INTERNAL_PROBLEM;
	}

	DWORD rounded_4K_Size = DRound( realSize, _4K );

	if ( rounded_4K_Size > dbSize )
	{
		assert( false );
		ODS_fls( L"Asertion!" );
		return ERROR_INTERNAL_PROBLEM;
	}

	// Checks if new block has to be extended
	if ( rounded_4K_Size > *pDBSize )
	{
		DataBlockFree( ppDB, pDBSize );

#ifdef _UseSimpleAndExetendedBlockOnly_
		rounded_4K_Size = DB_MaxSize;
#endif

		*pDBSize = rounded_4K_Size;

		if ( !DataBlockAllocate( ppDB, pDBSize ) ||
			( *pDBSize != rounded_4K_Size ) )
		{
			assert( false );
			ODS_fls( L"Asertion!" );
			return ERROR_INTERNAL_PROBLEM;
		}
	}
	else
	if ( ( rounded_4K_Size < *pDBSize )
#ifdef _UseSimpleAndExetendedBlockOnly_
			&& ( rounded_4K_Size == _4K )
#endif
		)
	{
		cWCStr message;
		message.FormatT(
				_LoadIDS( IDS_HPTruncate_SSQ ), 
				_GetSizeText( rounded_4K_Size ),
				_GetSizeText( *pDBSize ) );

		//sprintf( message, 
		//		"Real size of data block is %d but original size is %d\n"
		//		"The data block may be truncated.\n"
		//		"Note, if you have hidden part for the container,\n"
		//		"so the hidden part key may be lost in truncating.\n"
		//		"\n"
		//		"Do you want to truncate the data block down to real size?",
		//		rounded_4K_Size, *pDBSize );

		int res = MessageBoxT( 0, message.m_szBody, _LoadIDS( IDS_KeyGen ), MB_YESNO | MB_ICONQUESTION | MB_DEFBUTTON2 );

		if ( res == IDYES )
		{
			// Shreds and truncates tail of existing.
			ShredData( (BYTE *)*ppDB + rounded_4K_Size, *pDBSize - rounded_4K_Size );
			*pDBSize = rounded_4K_Size;
		}
	}

	// Copies tmp data block to new or to old

	memcpy( *ppDB, dbTmp, *pDBSize );
	copyEx( decHeader, &(dbTmp->header.m_body ) );

	// Encrypts header or updates it's tail digest

	if ( bEncryptIt )
	{
		ODS_fls( L"bEncryptIt" );
		int res = encryptHeader( *ppDB, dbTmp, keyInfo );
		ODSFuncRes( res );
		return res;
	}
	else
	{
		if ( ! (*ppDB)->header.updateTail( keyInfo, (*ppDB)->pool ) )
		{
			assert( false );
			ODS_fls( L"Asertion!" );
			return ERROR_INTERNAL_PROBLEM;
		}

		ODSFuncRes( ERROR_NO );
		return ERROR_NO;
	}
}

//
//	Optimization for decode key with encrypted header.
//	Main reason do not check first the DEFAULT_SALT_SLOT it is zero.
//
static int getOrderPosition( bool bIsEncrypted, int pos )
{
	if ( !bIsEncrypted )
		return pos;

	if ( DEFAULT_SALT_SLOT != 0 )
		return pos;

	if ( pos > 1 )
		return pos;

	return pos == 0 ? 1 : 0;
}

//
//	findAndOpenSecretKey
//		if encrypted check all slots.
//		if not check pke market slots and empty for hidden.
//
static int findAndOpenSecretKey(
		DATA_BLOCK	*dataBlock,			// Data block have read directly from container
		DWORD		dbSize,
		cKeyBlockParams *kbParams		// Key block paramteres used for key encryption.
)
{
	//
	// 19.11.2015
	//	Proceses for PKE all suspected slots
	//
	DWORD maxKeyQ = DATA_BLOCK::getKeyQ( dbSize );

	bool isEncrypted = dbHeader::isEncrypted(dataBlock->header.m_body);

	for( DWORD i = 0; i < maxKeyQ; i++ )
	{
		bool bSuspected =
				isEncrypted ||
				( dataBlock->header.m_body.keyMap[i].kbType == kbtEmpty ) ||
				( dataBlock->header.m_body.keyMap[i].kbType == kbtPKE );

		if ( !bSuspected )
			continue;

		BYTE *kbEncrypted = &(dataBlock->keyTab[ i ][0]);

		// It is assigned if we close KM after key choosing.
		kbParams->kbPKE.bCancelled = false;

		if ( kbParams->kbPKE.openSecretKey( kbParams->kbSSS.hWnd, &kbParams->kbPBE.m_pwd, kbEncrypted, true ) )
			return i;
	}

	return -1;
}

static int createKeyHandle(
		DATA_BLOCK	*dataBlock,			// Data block have read directly from container
		DWORD		dbSize,
		cKeyBlockParams *kbParams,		// Key block paramteres used for key encryption.
		sKeyMapInfo *keyMapInfo,		// = NULL,
		KEY_HANDLE	*pKeyHandle,		// = NULL,	creates key handle if the parameter is not NULL 
		sKeyInfo	*pKeyInfoCopy,		// = NULL,
		dbHeaderBase	*pDecryptedHeader,	// = NULL - place for decrypted header copy
		sSpaceUnitBase	*pContInfo,			// = NULL - place for decrypted cont info for hidden or usual part corresponded to pKeyHandle

		BYTE		*kbEncrypted,
		DWORD		kbEncryptedSize,
		sKeyInfo	*pKeyInfo,
		sVerificationBlock *decVerifyBlock,
		DWORD		iPos, // &foundNum,
		tcEncBlock < dbHeader > *decHeader
	)
{
	int kbSize = kbParams->getEncodedKeySize( pKeyInfo->size );

	bool isForHiddenPart = dbContInfo::isForHiddenPart( &(pKeyInfo->contInfo) );

	int res = ERROR_NO;

	if ( isForHiddenPart )
	{
		//
		// Verifies the key by veryfy info
		//	See DataBlockWriteKey_Hidden for encoded key packing format
		//

		tcEncBlock < sVerificationBlock > *verifyBlock = ( tcEncBlock < sVerificationBlock > * )( kbEncrypted + kbSize );

		int verifyBlockSize = sizeof( *verifyBlock );

		if ( (DWORD)kbSize + verifyBlockSize > kbEncryptedSize )
		{
			assert( false );
			return ERROR_INCORRECT_PASSWORD;
		}

		cSymAlgorithm cAlg;
		if ( !cAlg.createKey( pKeyInfo->keyAlgId, pKeyInfo->encModeId, pKeyInfo->key, pKeyInfo->getKeySizeBytes(), dataBlock->pool ) )
		{
			assert( false );
			return ERROR_INCORRECT_PASSWORD;
		}

		if ( !verifyBlock->decryptBodyTo( decVerifyBlock, cAlg.m_algId, cAlg.m_encModeId, cAlg.m_keyHandle )
			|| !decVerifyBlock->check() ) // 08.11.2015
		{
			assert( false );
			return ERROR_INCORRECT_PASSWORD;
		}
		
		kbSize += verifyBlockSize;
		
		copyEx < dbHeaderBase > ( pDecryptedHeader, NULL );
		copyEx < sKeyInfo > ( pKeyInfoCopy, pKeyInfo );
		copyEx < sSpaceUnitBase > ( pContInfo, &(pKeyInfo->contInfo) );

		if ( keyMapInfo != NULL )
		{
			keyMapInfo->setInfo( iPos, 0, kbParams->kbType, kbSize );
			//keyMapInfo->setInfo( foundNum, 0, kbParams->kbType, kbSize );
		}

		if ( pKeyHandle != NULL )
			cAlg.delegateHandle( pKeyHandle );

		return res;
	}
	else
	{
		//ODS_Enable( true );

		res = decryptAndVerifyHeader( 
					dataBlock,
					pKeyInfo,
					decHeader,
					pKeyHandle
				);

		if ( ( res == ERROR_NO ) 
#ifdef _allowBadHeader
			  || ( res == ERROR_INCORRECT_BLOCK_SIGNATURE ) 
#endif
			)
		{
			if ( keyMapInfo != NULL )
			{
				//keyMapInfo->setInfo( foundNum, 0, kbParams->kbType, kbSize );
				keyMapInfo->setInfo( iPos, 0, kbParams->kbType, kbSize );
			}

			copyEx( pKeyInfoCopy, pKeyInfo );
			copyEx( pContInfo, &(pKeyInfo->contInfo) );
			if ( res == ERROR_NO ) 
				copyEx < dbHeaderBase > ( pDecryptedHeader, &(decHeader->m_body) );

			return res;
		}
	}

	return ERROR_INCORRECT_PASSWORD;
}

//
//	createKeyHandleWithPKE
//		Parses key map empty slots
//		Tries to find in BCDB compatible by identifier ( every PKE slot contains PGP_ID + X509 local key id at beginning).
//		Opens the compatible secret key by password.
//			Note! It implements password only for compatible secret keys.
//			So the operation does not require a lot of time.
//		Tries to decode the slot as PKE.
//		Create required key handle.
//
static int createKeyHandleWithPKE(
		DATA_BLOCK	*dataBlock,			// Data block have read directly from container
		DWORD		dbSize,
		cKeyBlockParams *kbParams,		// Key block paramteres used for key encryption.
		sKeyMapInfo *keyMapInfo,		// = NULL,
		KEY_HANDLE	*pKeyHandle,		// = NULL,	creates key handle if the parameter is not NULL 
		sKeyInfo	*pKeyInfoCopy,		// = NULL,
		dbHeaderBase	*pDecryptedHeader,	// = NULL - place for decrypted header copy
		sSpaceUnitBase	*pContInfo,		// = NULL - place for decrypted cont info for hidden or usual part corresponded to pKeyHandle

		sKeyInfo	*pKeyInfo,
		sVerificationBlock *decVerifyBlock,
//		DWORD		&foundNum,
		tcEncBlock < dbHeader > *decHeader
	)
{
	int pkePos = findAndOpenSecretKey( dataBlock, dbSize, kbParams );
	if ( pkePos < 0 )
		return ERROR_INCORRECT_PASSWORD;

	BYTE *kbEncrypted = &(dataBlock->keyTab[ pkePos ][0]);
	DWORD kbEncryptedSize = dbSize - ((DWORD)kbEncrypted - DWORD(dataBlock));

	if ( !kbParams->decodeKey( pKeyInfo, sizeof( sKeyInfo ), kbEncrypted, kbEncryptedSize ) )
		return ERROR_INCORRECT_PASSWORD;

	return
		createKeyHandle(
				dataBlock, dbSize, kbParams, keyMapInfo, pKeyHandle, pKeyInfoCopy, pDecryptedHeader, pContInfo,
				kbEncrypted, kbEncryptedSize, pKeyInfo, decVerifyBlock, pkePos, decHeader
//				kbEncrypted, kbEncryptedSize, pKeyInfo, decVerifyBlock, foundNum, decHeader
			);
}


//
//	DataBlockVerifyKeyBlockAndGetKey
//
//	1. Tries to decode key info, it walks key blocks and implement kbParams to 
//		PBE - decrypt key by key formed from password.
//		PKE - decrypt by public key
//		SSS - implement every share to the key block
//	 Notes
//		a. If kbParams type is undefined ( kbtEmpty ), so all available 
//			encoding shemes will be implemented to DATA_BLOCK
//		b. If pKeyHandle and pKeyInfoCopy both are NULL, so the function
//			checks if the password in kbParams.kbPBE.pwd is compatible with any key block.
//			It returns ERROR_NO if at least the password is compatible even with one pwd 
//			for block with kbtSSS type
//
//	2. If decoded the key info is for hidden part, it reads next bytes as verification block, 
//			See DataBlockWriteKey_Hidden
//
DWORD DataBlockVerifyKeyBlockAndGetKey(
		DATA_BLOCK	*dataBlock,			// Data block have read directly from container
		DWORD		dbSize,
		cKeyBlockParams *kbParams,		// Key block paramteres used for key encryption.
		sKeyMapInfo *keyMapInfo,		// = NULL,
		KEY_HANDLE	*pKeyHandle,		// = NULL,	creates key handle if the parameter is not NULL 
		sKeyInfo	*pKeyInfoCopy,		// = NULL,
		dbHeaderBase	*pDecryptedHeader,	// = NULL - place for decrypted header copy
		sSpaceUnitBase	*pContInfo,			// = NULL - place for decrypted cont info for hidden or usual part corresponded to pKeyHandle
		HWND		hProgress,			// = NULL
		bool		*pCancelled,
		bool		bFirst4K			// The parsing may be devided on two part for encrypted header emprovment. First4K  or other part.
										// If this flag is false so first 4K is skipped.
	)
{
	ODSFunc;
#ifdef _ODS
	cWCStr wcs( 1024 );
	bytes2string( wcs.m_szBody, wcs.m_len, (BYTE*)dataBlock, dbSize );
	ODS_fls( _FormatT( L"DB 0x%X: \n%s", dataBlock, wcs.m_szBody ) );
#endif
	//
	//	If the function is called without key handle and key info copy
	//	so it is used for pasowrd checking only.
	//	So it has to check compatibility of kbParams.kbPBE.password with all passowrds for 
	//	all existing key blocks  
	//
	bool checkPwdOnly = ( pKeyHandle == NULL ) && ( pKeyInfoCopy == NULL ) && //( pMainKeyInfo == NULL ) &&
						( pDecryptedHeader == NULL ) && ( pContInfo == NULL );

	if ( kbParams->kbSSS.pErrorCode == NULL )
	{ assert( false ); return ERROR_INTERNAL_PROBLEM; }

//	cSymAlgorithm cAlg;

	//
	// Walks key blocks and implement kbParams to 
	//		PBE - decrypt key by key formed from password.
	//		PKE - decrypt by public key
	//		SSS - implement every share to the key block
	//

	BYTE *kbEncrypted;
	int kbEncryptedSize;
	//DWORD foundNum = -1;

	// Decrypted plaintext buffers allocation

	cPlainTextBlock ptb;

	if ( ptb.Allocate( 
		sizeof( tcEncBlock < dbHeader > ) + 
		sizeof( sVerificationBlock ) +
		sizeof( sKeyInfo ) +
		0 )
			== NULL )
	{ assert( false ); return ERROR_INTERNAL_PROBLEM;	}

	int pos = 0;

	tcEncBlock < dbHeader > *decHeader	= (tcEncBlock < dbHeader > *)(ptb.buffer + pos); pos += sizeof( *decHeader );
	sVerificationBlock *decVerifyBlock = (sVerificationBlock*)(ptb.buffer + pos); pos += sizeof( sVerificationBlock );
	sKeyInfo	  *pKeyInfo	= (sKeyInfo   *)(ptb.buffer + pos); pos += sizeof( sKeyInfo );

	assert( pos == ptb.bufferSize );

	bool isEncrypted = dbHeader::isEncrypted(dataBlock->header.m_body);

	DWORD maxKeyQ = DATA_BLOCK::getKeyQ( dbSize );

	if ( hProgress != NULL )
		SendMessage( hProgress, PBM_SETRANGE, 0, MAKELPARAM( 0, maxKeyQ ) );

	bool pkeBlockWasFound = false;

	//
	//	30.06.2011
	//	1. For encrypted block it enumerates secret keys with pk identifier before algorithm processing.
	//		Prev version it does within algorithm enumeration process and without pk identifier.
	//		See cProgressDlg_EncHeaderParse::Run().
	//	2. If key block is not encrypted only the corresponded secret key from BCDB is loaded if PKE key block is encounterred ( see decodeKey ).
	//
	////
	////	Secret key notes.
	////
	////	1. If key block is encrypted the password is checked for all existing secret key first of all, to be implemented for every key block.
	////	2. If key block is not encrypted only the corresponded secret key from BCDB is loaded if PKE key block is encounterred ( see decodeKey ).
	////	3. If no one key block is decoded, but the PKE key was found ( pkeBlockWasFound == true )
	////		so secret key is loaded outside BCDB and the loop is started again with new secret key
	////

	////
	////	Tries to get secret key silently.
	////	If secret key is activated so enumeration looks too slow...
	////
	//if ( isEncrypted && ( kbParams->kbPKE.ska == NULL ) )
	//{
	//	kbParams->kbPKE.openSecretKey( kbParams->kbSSS.hWnd, &kbParams->kbPBE.m_pwd, NULL, checkPwdOnly ); // 27.06.2011 - true );
	//}

	//
	//	13.11.2014
	//	Public key encrypted hidden part now may be create, so we have to open the key before parsing
	//
	int res = 
		createKeyHandleWithPKE(
			dataBlock, dbSize, kbParams, keyMapInfo, pKeyHandle, pKeyInfoCopy, pDecryptedHeader, pContInfo,
			pKeyInfo, decVerifyBlock, decHeader
		);
			
	if ( res == ERROR_NO )
		return res;

	if ( !isEncrypted && ( dataBlock->header.m_body.wKeyGenId < KEYGEN_ID_V5 ) && kbParams->kbPBE.m_keyFileList.hasFiles() )
		MessageBoxT(
			NULL,
			_LoadIDS( IDS_szKeyFilesDoesNoEffectOnOldContainers ),
			_LoadIDS( IDS_WARNING ),
			MB_ICONEXCLAMATION  );


	WORD kbTypeSave = kbParams->kbType;

	DWORD i = bFirst4K ? 0 : DATA_BLOCK::getKeyQ( _4K );

	for (;;)
	{
		if ( hProgress != NULL )
			SendMessage( hProgress, PBM_SETPOS, i, 0 );

		if ( i >= maxKeyQ )
		{
			if ( pkeBlockWasFound &&
				!checkPwdOnly )
			{
				// See 'Secret key notes 3' above.
				if ( kbParams->kbPKE.ska == NULL ) // Was never loaded
				{
					//
					// Open secret key with browsing dialogue.
					// The decodeKey tries to open SK without browsing by default.
					//
					if ( kbParams->kbPKE.openSecretKey( kbParams->kbSSS.hWnd, &kbParams->kbPBE.m_pwd, NULL, false ) )
					{
						i = 0;
						continue;
					}
				}
			}

			break;
		}

		int iPos = getOrderPosition( isEncrypted, i );
		kbEncrypted = &(dataBlock->keyTab[ iPos ][0]);

#ifdef _ODS
		cWCStr wcs( 1024 );
		bytes2string( wcs.m_szBody, wcs.m_len, kbEncrypted, DB_MinEncryptedKeyBlockSize );
		ODS_fls( _FormatT( L"i = %d,\nkeyTab[%d] = %s", i, iPos, wcs.m_szBody ) );
#endif

		// Type of currently tryied key block
		WORD kbType;

		if ( isEncrypted )
		{
			kbEncryptedSize = ( maxKeyQ - iPos ) * DB_MinEncryptedKeyBlockSize;

			//
			//	Note, if kbParams type is undefined ( kbtEmpty ), so all available 
			//	encoding shemes are implemented to DATA_BLOCK,
			//	see kbParams->decodeKey
			//
			kbType = kbtEmpty;
			i++;
		}
		else
		{
			kbEncryptedSize = dataBlock->header.m_body.keyMap[i].kbSize;
			kbType = dataBlock->header.m_body.keyMap[i].kbType;

			ODS_fls( _FormatT( L"kbEncryptedSize = %d, kbType = %d", kbEncryptedSize, (int)kbType ) );
 
			// Hidden part may be saved here
			if ( kbType == kbtEmpty )
			{
				kbEncryptedSize = ( maxKeyQ - i ) * DB_MinEncryptedKeyBlockSize;
				i++;
			}
			else
			{
				// skip busied blocks
				i += dbKeyId::GetRequiredQ( kbEncryptedSize );
			}
		}

		//
		//	Checks if following decode function gets valid encrypted block 
		//
		if ( kbEncrypted > (BYTE*)dataBlock + dbSize - kbEncryptedSize )
		{
			assert( false );
			return ERROR_DAMAGED_DATA_BLOCK;
		}

		if ( kbType == kbtSalt )
			continue;

		if ( checkPwdOnly )
		{
			kbParams->kbType = kbtEmpty;
			*(kbParams->kbSSS.pErrorCode) = ERROR_INCORRECT_PASSWORD;

			bool decoded = kbParams->decodeKey( pKeyInfo, sizeof( sKeyInfo ), kbEncrypted, kbEncryptedSize, true ); // , false ); //true );

			if ( decoded )
			{
				if ( keyMapInfo != NULL )
					keyMapInfo->setInfo( iPos, 0, kbParams->kbType, kbParams->getEncodedKeySize( pKeyInfo->size ) );
//					keyMapInfo->setInfo( foundNum, 0, kbParams->kbType, kbParams->getEncodedKeySize( pKeyInfo->size ) );

				return ERROR_NO;
			}

			if ( *kbParams->kbSSS.pErrorCode == ERROR_INCORRECT_PASSWORD )
				continue;

			if ( *kbParams->kbSSS.pErrorCode != ERROR_NO )
				return *kbParams->kbSSS.pErrorCode;

			continue;
		}
		else
		{
			if ( kbType == kbtPKE )
				pkeBlockWasFound = true;

			if ( kbTypeSave != kbtEmpty ) // original kbType is defined
			{
				if (  ( kbType != kbTypeSave ) &&	// Determinated type skips incompatible not empty blocks
					  ( kbType != kbtEmpty )		// Hidden part may be saved here
					) 
						continue;

				kbParams->kbType = (ekbType)kbTypeSave;
			}
			else	// original kbType is undefined
			{
				kbParams->kbType = (ekbType)kbType;
			}

			*(kbParams->kbSSS.pErrorCode) = ERROR_INCORRECT_PASSWORD;

			if ( !kbParams->decodeKey( pKeyInfo, sizeof( sKeyInfo ), kbEncrypted, kbEncryptedSize ) )
			{
				if ( *(kbParams->kbSSS.pErrorCode) == ERROR_INCORRECT_PASSWORD )
					continue;

				if ( *(kbParams->kbSSS.pErrorCode) != ERROR_NO )
					return *kbParams->kbSSS.pErrorCode;

				continue;
			}

			//ODS( L"key is decoded" );
		}

		int res = 
			createKeyHandle(
				dataBlock, dbSize, kbParams, keyMapInfo, pKeyHandle, pKeyInfoCopy, pDecryptedHeader, pContInfo,
				kbEncrypted, kbEncryptedSize, pKeyInfo, decVerifyBlock, iPos, decHeader
			);
		
		if ( res == ERROR_INCORRECT_PASSWORD )
			continue;
			
		return res;
	}

	//
	//	No one compatible password was found
	//
//	if ( kbParams->kbType == kbtEmpty )
		kbParams->kbType = (ekbType)kbTypeSave;
	
	return ERROR_INCORRECT_PASSWORD;
}

DWORD GenerateKey(
		BYTE *key,
		int   keyLength,
		BYTE *salt,
		int	  saltSize,
		BYTE *seed,
		int   seedSize,
		BYTE  hashAlgID,
		bool isKey // determintes is genereted key or initial vector
	)
{
	int err = P12_producePseudoRandomBytes(
				key,  keyLength,
				salt, saltSize,
				1,
				seed, seedSize,
				isKey ? P12_prgm4key : P12_prgm4IV,
				hashAlgID
			);

	if ( err == x509e_ok )
		return ERROR_NO;

	if ( err == hae_invalidAlgorithmID )
		return ERROR_INCORRECT_HASH_ALGORITHM;

	assert( false );
	return ERROR_INTERNAL_PROBLEM;

}

DWORD DataBlockWriteToKeyTable(
		DATA_BLOCK	**ppDB,
		DWORD		*pDBSize,
		int			num,
		BYTE		*kb,
		DWORD		kbSize
	)
{
	bool isEncrypted = dbHeader::isEncrypted((*ppDB)->header.m_body);

	if ( !isEncrypted )
	{
		DWORD sizeRequired = (*ppDB)->header.m_body.getRealSize(); //(BYTE*)((*ppDB)->keyTab[ num ][ 0 ]) - (BYTE*)(*ppDB)

		if ( sizeRequired == DWORD(-1) )
		{
			assert( false );
			return ERROR_INTERNAL_PROBLEM;
		}
		
		sizeRequired += kbSize;

		if ( sizeRequired > *pDBSize )
		{
			if ( !DataBlockReAllocate( ppDB, *pDBSize, &sizeRequired ) )
			{
				assert( false );
				return ERROR_INTERNAL_PROBLEM;
			}

			*pDBSize = sizeRequired;
		}
	}

	memcpy( (BYTE*)(&((*ppDB)->keyTab[ num ][ 0 ])), kb, kbSize );

	return ERROR_NO;
}

bool updateKeyParams( sCKHParams *ps, sKeyInfo *keyInfo, dbHeader *decHeader )
{
	bool equ = true;

	ps->contInfo = keyInfo->contInfo;

	if ( ( decHeader != NULL ) && !dbContInfo::isForHiddenPart( &(keyInfo->contInfo) ) )
	{
		copyEx < dbHeaderBase > ( &(ps->decryptedHeader), decHeader );
		assert( keyInfo->hashAlgId == decHeader->hashAlgId );
	}

	if ( keyInfo->keyAlgId  != ps->contAlgId ) { assert( ps->contAlgId == -1 ); ps->contAlgId = keyInfo->keyAlgId;  equ = false; }
	if ( keyInfo->encModeId != ps->encModeId ) { assert( ps->encModeId == 0 ); ps->encModeId = keyInfo->encModeId; equ = false; }
	if ( keyInfo->hashAlgId != ps->hashAlgId ) { assert( ps->hashAlgId == -1 ); ps->hashAlgId = keyInfo->hashAlgId; equ = false; }

	return equ;
}


////
////	HP_GetMainPartContInfo
////		1. Finds position of encrypted main part info
////			See DataBlockWriteKey_Hidden.
////		2. Decryptes main part contInfo
////		3. Delegates main part key handle for further mounting
////
//bool HP_GetMainPartContInfo( DATA_BLOCK *db, KEY_HANDLE keyHandleHP, sCKHParams *psHP, KEY_HANDLE *pKeyHandleMain, sCKHParams *psMain ) )
//{
//	// 1.0 Prepares main key params
//	sCKHParams tmp;
//	if ( psMain == NULL )
//		psMain = &tmp;
//
//	memcpy( psMain, psHP, sizeof( *psMain ) );
//	psMain->keyMapInfo.setUndefined(); // It cannot be used in future.
//
//	// 1.1 Gets position of encrypted main part cont info inside data block,
//	//	see DataBlockWriteKey_Hidden.
//	WORD mainPartKBSize = sizeof( tcEncBlock < sKeyInfo > );
//
//	if ( ( psHP->keyMapInfo.position == -1 ) || ( psHP->keyMapInfo.kbSize < mainPartKBSize ) )
//		return false;
//
//	BYTE *pos = &(db->keyTab[ psHP->keyMapInfo.position ][0]) + mainPartKBSize;
//
//	tcEncBlock < sKeyInfo > *encKeyInfo = ( tcEncBlock < sKeyInfo > *)(pos);
//
//	// 2. Decryptes main cont info
//
//	if ( ptb.Allocate(
//			sizeof( sKeyInfo ) +
//			0 ) == NULL )
//				return ERROR_INTERNAL_PROBLEM;
//
//	int pos = 0;
//
//	// Main part key block position ( is encrypted by hidden part key ).
//	sKeyInfo *keyInfo = ( sKeyInfo* )ptb.buffer + pos; pos += mainPartKBSize;
//	assert( ptb.bufferSize == pos );
//
//	if ( !encKeyInfo->decryptBodyTo( keyInfo, psMain->contAlgId, psMain->encModeId, keyHandleHP ) )
//		return false;
//
//	updateKeyParams( psMain, keyInfo );
//
//	// Checks main key 
//	cSymAlgorithm cAlg;
//	if ( !cAlg.createKey( psMain->contAlgId, psMain->encModeId, keyInfo->key, keyInfo->size, db->pool ) )
//		return false;
//
//	psMain->contInfo = keyInfo->contInfo;
//
//	// 3. Delegates main part key handle
//	if ( pMainPartContInfo != NULL )
//	{
//		cAlg.delegateHandle( pMainPartContInfo );
//	}
//
//	return true;
//}

//
//	CreateHandle_MainPart
//		Main part password required.
//
bool CreateHandle_MainPart(

		HWND hWnd,

		const wchar_t *text,
		const wchar_t *caption,

		DATA_BLOCK			**ppDB,
		DWORD				*pDBSize,
		
		sKeyInfo			*keyInfo, // = NULL

		KEY_HANDLE			*pKeyHandle, // = NULL
		sCKHParams			*ps,		// = NULL
		dbKeyId				*keyMapT	// = NULL
	)
{
	if ( caption == NULL )
		caption = _LoadIDS( IDS_ENTER_EXISTING_PASSWORD );
	if ( text == NULL )
		text = _LoadIDS( IDS_KeyBlockUpdating );

	sCKHParams tmp;
	if ( ps == NULL )
		ps = &tmp;

	for(;;)
	{
		DWORD flags = 0;
		DWORD errCode = 0;

		KEY_HANDLE keyHandle;

		if ( !GetKeyHandle( 
				hWnd, text, caption,
				ppDB, pDBSize,
				&errCode, &flags,
				&keyHandle,
				//NULL,
				ps,
				keyInfo )
			)
				return false;

		if ( !dbContInfo::isForHiddenPart( &(ps->contInfo) ) )
		{
			//
			// Delegates key handle.
			//
			if ( pKeyHandle != NULL )
				*pKeyHandle = keyHandle;
			else
				FreeKeyHandle( ps->contAlgId, keyHandle );

			return true;
		}

		if ( keyMapT != NULL )
			dbHeader::MarkBusied( keyMapT, &(ps->keyMapInfo) );

		FreeKeyHandle( ps->contAlgId, keyHandle );

		MessageBoxT(
			hWnd,
			_LoadIDS( IDS_HP_PWD_CannotBeUsedForKeyBlockUpdating ),
			_LoadIDS( IDS_WARNING ),
			MB_ICONEXCLAMATION  );
	}
}

//
//	DataBlockWriteKey_Hidden
//
//		1. Writes hidden key info encrypted by kbParams
//		2. Writes simple verification block at the key tail encrypted by hidden key info
//
//	How the key read, see DataBlockVerifyKeyBlockAndGetKey
//
//	Note 1. The function does not extend key block for hidden part key. 
//			It will security nonsens if block is extended but key map reports that no one key presents in extended part.
//
//	Note 2. It shreds old key block information if keyMapInfo specified ( is not NULL or position != -1 )
//
DWORD 
	DataBlockWriteKey_Hidden(
		DATA_BLOCK			**ppDB,
		DWORD				*pDBSize,
		sKeyInfo			*hiddenKeyInfo,	//	= NULL	if null, so it does not required to save encoded key info
											//	the existing hidden part from started position is simple shreded
		cKeyBlockParams		*kbParams,		//	= NULL

		sKeyMapInfo			*keyMapInfo,	// = NULL,	See Note 2.

		dbKeyId				*keyMap,		// the map is used for detecting busied items.
											// Ifnormation about existing hidden parts positions may be accumulated by and sent to the
											// function create new hidden part. See CreateHiddenPart
		bool				bDataBlockIsDamaged // It is true at ChangePassword only.
	)
{
	if ( keyMap == NULL )
	{
		if ( hiddenKeyInfo != NULL )
		{	assert( false );
			return ERROR_INTERNAL_PROBLEM;
		}
	}

	int hiddenPartKBSize = DB_MinEncryptedKeyBlockSize;
	int verifyBlockSize = sizeof( tcEncBlock < sVerificationBlock > );
	
	if ( kbParams != NULL )
	{
		//
		// Version 2.01.7.4 - 24.10.2018
		//		Checks if hiddenKeyInfo is not NULL.
		//
		hiddenPartKBSize = ( hiddenKeyInfo == NULL ? DB_MinEncryptedKeyBlockSize : kbParams->getEncodedKeySize( hiddenKeyInfo->size ) );
	}

	int kbSize = hiddenPartKBSize + verifyBlockSize;

	// allocaters space for temporary plain text data
	cPlainTextBlock ptb;

	if ( ptb.Allocate(
			kbSize +
//			sizeof( sKeyInfo ) +
			sizeof( sKeyInfo ) +
//			sizeof( tcEncBlock < dbHeader > ) +
			0 ) == NULL )
	{
		assert( false );
		return ERROR_INTERNAL_PROBLEM;
	}

	int pos = 0;

	// Hidden part ket block position.
	BYTE *hpKB = ptb.buffer + pos; pos += hiddenPartKBSize;

	// Main part key block position ( will be encrypted by hidden part key ).
	tcEncBlock < sVerificationBlock > *verifyBlock = ( tcEncBlock < sVerificationBlock > *)(ptb.buffer + pos); pos += verifyBlockSize;

	sKeyInfo *keyInfoTmp	  = (sKeyInfo *)(ptb.buffer + pos); pos += sizeof( sKeyInfo );
//	tcEncBlock < dbHeader > *decryptedHeader = (tcEncBlock < dbHeader > *)(ptb.buffer + pos); pos += sizeof( *decryptedHeader );
	assert( ptb.bufferSize == pos );

	// Clears current busied map item ( it will be shredded later )
	int newPosNum = keyMapInfo->position;

	if ( ( keyMapInfo->position != -1 ) && ( keyMap != NULL ) )
		dbHeader::MarkEmpty( keyMap, keyMapInfo->position, keyMapInfo->quantity );

	int blocksRequired = dbKeyId::GetRequiredQ( kbSize );

	int curKeyMapQ = DATA_BLOCK::getKeyQ( *pDBSize );

	if ( bDataBlockIsDamaged )
	{
		if ( newPosNum >= DB_minKeyQ )
		{
			ODS_fls( L"Asertion! Damaged data block cannot be more then 4K" );
			return ERROR_NOT_ENOUGH_SPACE_FOR_KEY;
		}

		*pDBSize = DB_MinSize;
	}
	else
	{
		// Find space inside current block size 
		// See Note 1.
		newPosNum = dbHeader::findEmptySpace( keyMap, blocksRequired, 0, curKeyMapQ, true );
	}

	if ( newPosNum == -1 )
		return ERROR_NOT_ENOUGH_SPACE_FOR_KEY; 
    
	// Prepares encoded key
	if ( ( kbParams != NULL ) && ( hiddenKeyInfo != NULL  ) )
	{
		if ( !kbParams->encodeKey( hpKB, hiddenPartKBSize, hiddenKeyInfo, sizeof( *hiddenKeyInfo ) ) ) // hiddenKeyInfo->size ) )
		{
			assert( false );
			return ERROR_INTERNAL_PROBLEM;
		}

		if ( kbParams->kbType != kbtPKE )
			if ( !kbParams->decodeKey( keyInfoTmp, sizeof( sKeyInfo ), hpKB, hiddenPartKBSize ) )
			{
				assert( false );
				return ERROR_INTERNAL_PROBLEM;
			}

		//
		//	Encodes and checks the encoding of verification block
		//
		sVerificationBlock vb;
		vb.hashAlgId	= hiddenKeyInfo->hashAlgId;
		vb.keyAlgId		= hiddenKeyInfo->keyAlgId;
		vb.encModeId	= hiddenKeyInfo->encModeId;

		if ( !verifyBlock->encryptFromBody( &vb, hiddenKeyInfo, (*ppDB)->pool ) ||
			 !verifyBlock->decryptBodyTo( &vb, hiddenKeyInfo, (*ppDB)->pool ) 
				 )
		{
			assert( false );
			return ERROR_INTERNAL_PROBLEM;
		}

		memset( &vb, 0, sizeof( vb ) );
	}

	// Shreds old hidden part key blocks 
	if ( ( keyMapInfo != NULL ) && ( keyMapInfo->position != -1 ) )
	{
		dbHeader::freeKeyBlockChain( *ppDB, *pDBSize, keyMap, keyMapInfo );
	}

	//
	// Version 2.01.7.4 - 24.10.2018
	//		Checks if kbParams is not NULL.
	//
	ekbType kbType = kbParams == NULL ? kbtHP : kbParams->kbType;

	// Set new HP busied marks
	if ( keyMapInfo != NULL )
		keyMapInfo->setInfo( newPosNum, 0, kbType, kbSize ); // kbtHP

	dbHeader::MarkBusied( keyMap, newPosNum, blocksRequired, kbType, kbSize ); // kbtHP
	
	return DataBlockWriteToKeyTable( ppDB, pDBSize, newPosNum, hpKB, kbSize );
}

//
// DataBlockGetContainerId()
//		Container ID is initially randomized for main and hidden contInfo.
//		The info is not changed even if main contInfo is encrypted.
//
BOOL DataBlockGetContainerId( DATA_BLOCK *DataBlock, BYTE *ContainerId )
{
	memcpy( ContainerId, DataBlock->header.m_body.containerID, CONTAINER_ID_LENGTH );
	return true;
}

//
//	FindSamePassword
//		Tryes to decode every key block in the data block with different algorithms
//		Returns true if at least one key block is compatible, otherwise return false.
//
bool FindSamePassword(
		DATA_BLOCK	*pDB,
		DWORD		dbSize,
		cPassword	*pwd,
		ekbType		*pKBType, // = NULL
		sEncIDs		*encIds // = NULL
	)
{
	if ( pKBType != NULL ) *pKBType = kbtEmpty;

	cKeyBlockParams kbParams;

	DWORD errorCode = 0;
	DWORD flags = 0;

	//
	//	v. 1.05.5 - 28.06.2011 uses updated function from certificate.lib v. 1.03.4
	//
	kbParams.kbPBE.m_pwd.Set( pwd );

	cEnumAlgIDs algIDs( -1, false ); // ( bcsaRIJN );
	cEnumHashAlgIDs hashAlgs( bchaWhirlpool512, false );
	cEnumEncModes encModes( EX_ENCRYPT_XTS_MODE_FLAG );

	for( int i = algIDs.Q; i--; )
	{
		DWORD algId = algIDs.T[i];

		for ( int j = encModes.Q; j--; )
		{
			DWORD encModeId = encModes.T[j].id;

			for ( int k = hashAlgs.Q; k--; )
			{
				DWORD hashAlgId = hashAlgs.T[i];

				for( int l = maxLen( defIterationsT ); l-- ; )
				{
					int iterations = defIterationsT[l];

					// The tricks allow to use same code if algorithm id is defined ( the data block header is not encrypted )
					if ( !dbHeader::isEncrypted(pDB->header.m_body) )
					{
						i = j = k = l = 0;
						algId		= pDB->header.m_body.contAlgId;
						encModeId	= pDB->header.m_body.encModeId;
						hashAlgId	= pDB->header.m_body.hashAlgId;
						iterations	= dbHeader::getIterationCount(pDB->header.m_body);
					}
					else
					if ( encIds != NULL )
					{
						i = j = k = 0;
						algId		= encIds->m_algId;
						encModeId	= encIds->m_encModeId;
						hashAlgId	= encIds->m_hashAlgId;
					}

					kbParams.KeyBlockParamsInit( pDB, dbSize, algId, encModeId, hashAlgId, iterations, &errorCode, &flags, 0 );

					errorCode = DataBlockVerifyKeyBlockAndGetKey( pDB, dbSize, &kbParams ); // Parses first 4K

					if ( ( errorCode == ERROR_INCORRECT_PASSWORD ) && ( dbSize > _4K ) )
						errorCode = DataBlockVerifyKeyBlockAndGetKey( pDB, dbSize, &kbParams, 0,0,0,0,0,0,0, false ); // Parses last 16K

					if ( errorCode == ERROR_NO )
					{
						if ( pKBType != NULL ) *pKBType = kbParams.kbType;
						return true;
					}
				}
			}
		}
	}

	return false;
}

//
//	decryptHeaderEx
//		decrypts header by a key handle, the key handle may be for usual part only
//	
//		1. If dataBlock is not encrypted ( see DATA_BLOCK::header.isEncrypted ) it verifies header
//			and copies dataBlock to output buffer and returns true.
//		2. Tries to decrypt dataBlock::contInfo with the algoritm id and key handle.
//			If the contInfo is properly decrypted, the function decrypts header with keyMap
//			with same key handler, copies decrypted result to out buff and returns true.
//		3. Otherwise returns false.
//
bool
	decryptHeaderEx(
		DATA_BLOCK	*dataBlock,
		DWORD		contAlgId,	// Encryption Algorithm Identificator
		DWORD		encModeId,
		DWORD		*pHashAlgId,	//  = NULL, [Out] pointer to hash algorithm decrypted from encBlock tail
		DWORD		keyHandle,		// and Key Handle it may be key handle for hidden part
		tcEncBlock <dbHeader>
					*headerOut,		// = NULL, Decrypted header, caller procedure is responsible for secure erasing of the decrypted data
		DWORD		*pErrorCode		// = NULL, if the function returns FALSE, look on the Error code
	)
{
	DWORD err;
	if ( pErrorCode == NULL )
		pErrorCode = &err;
	
	// 1.
	DATA_BLOCK *db = ( DATA_BLOCK * )dataBlock;

	if ( db == NULL )
	{
		assert( false );
		return false;
	}

	bool isEncrypted = dbHeader::isEncrypted(db->header.m_body);

	DWORD hashAlgId = -1;

	if ( pHashAlgId == NULL )
		pHashAlgId = &hashAlgId;

	if ( !isEncrypted )
	{
		bool verified = db->header.verifyTail( contAlgId, encModeId, keyHandle, pHashAlgId );

		if ( verified )
		{
			copyEx < tcEncBlock < dbHeader >  >( headerOut, &dataBlock->header );

			*pErrorCode = ERROR_NO;
			return true;
		}
		else
		{
			// May be the key handle for hidden part
			// we can verify it later.
		}
	}

	// 2.
	cPlainTextBlock ptbDecrypted;
	if ( ptbDecrypted.Allocate( sizeof( DATA_BLOCK ) ) == NULL )
	{
		assert( false );
		*pErrorCode = ERROR_INTERNAL_PROBLEM;
		return false;
	}

	DATA_BLOCK *dbDecrypted = ( DATA_BLOCK *)ptbDecrypted.buffer;
	memcpy( dbDecrypted, db, sizeof( DATA_BLOCK ) );

	if ( !dbDecrypted->header.decryptFrom( &db->header, contAlgId, encModeId, keyHandle ) )
	{
		*pErrorCode = ERROR_INCORRECT_BLOCK_SIGNATURE;
		return false;
	}

	copyEx( headerOut, &(dbDecrypted->header) );

	if ( pHashAlgId != NULL )
		*pHashAlgId = dbDecrypted->header.m_body.hashAlgId;

	*pErrorCode = ERROR_NO;
	return true;
}

//
//	encryptHeaderEx
//		encrypts entire header or it's tail with digest by a key handle for usual part.
//
//		1. If dataBlock is encrypted ( see DATA_BLOCK::header.isEncrypted )
//			1.1 it checks if key handle is valid for the data block by decryptHeaderEx.
//			1.2 If is valid the function copies dataBlock to output buffer and returns true.
//			If is not valid the function returns false
//		2. Encrypts header by the key handle for usual part
//
bool
	encryptHeaderEx(
		DATA_BLOCK	*dataBlock,
		DWORD		algId,			// Encryption Algorithm Identificator
		DWORD		encModeId,		// Encryption Algorithm Identificator
		DWORD		hashAlgId,		// Hash algorithm is used for digest
		DWORD		keyHandle,		// and Key Handle it may be key handle for hidden part
		tcEncBlock <dbHeader>
					*dbOut,			// Entire encrypted block with tail
		DWORD		*pErrorCode,	// if the function returns FALSE, look on the Error code
		bool		bEntire			// = true, Defines if entire or tail only encryption is used
	)
{
	DATA_BLOCK *db = ( DATA_BLOCK * )dataBlock;

	cPlainTextBlock ptb;
	if ( ptb.Allocate( sizeof( DATA_BLOCK ) ) == NULL )
		return false;

	DATA_BLOCK *dbTmp = ( DATA_BLOCK *)ptb.buffer;
	memcpy( dbTmp, db, sizeof( DATA_BLOCK ) );

	// 1.

	*pErrorCode = ERROR_INCORRECT_BLOCK_SIGNATURE;

	if ( dbHeader::isEncrypted(db->header.m_body) )
	{
		// 1.1
		if ( !decryptHeaderEx( dataBlock, algId, encModeId, NULL, keyHandle, &dbTmp->header, pErrorCode ) )
			return false;

		// 1.2
		if ( bEntire )
		{
			if ( dbOut != &(dataBlock->header) )
				copyEx( dbOut, &(dataBlock->header) ); // , sizeof( *dbOut ) );
		}
		else
		{
			if ( !dbTmp->header.updateTail( hashAlgId, algId, encModeId, keyHandle ) )
				return false;

			copyEx< tcEncBlock < dbHeader> >( dbOut, &(dbTmp->header) );
		}

		*pErrorCode = ERROR_NO;
		return true;
	}

	// 2. 

	*pErrorCode = ERROR_INCORRECT_PASSWORD; // ERROR_INVALID_ALGORITHM;

	if ( bEntire )
	{
		if ( !db->header.encryptTo( &dbTmp->header, algId, encModeId, hashAlgId, keyHandle ) )
			return false;
	}
	else
	{
		if ( !dbTmp->header.updateTail( hashAlgId, algId, encModeId, keyHandle ) )
			return false;
	}

	*pErrorCode = ERROR_NO;
	copyEx< tcEncBlock < dbHeader> >( dbOut, &(dbTmp->header) );

#ifdef _DEBUG
	if ( bEntire )
		return dbOut->decryptBodyTo( &(dbTmp->header.m_body), algId, encModeId, keyHandle );
	else
		return dbOut->verifyTail( algId, encModeId, keyHandle, &hashAlgId );
#endif

	return true;
}

bool GetLocalPKID( BYTE *dbBuff, DWORD dbSize, int slot, wchar_t *bf, DWORD bfSize )
{
	static const int localKeyIDSize = 20; // SHA1 digest size.

	DATA_BLOCK *db = (DATA_BLOCK *)dbBuff;

	HWND hWnd = NULL;
	
	BYTE *keyIDSBuffer = &(db->keyTab[slot][0]); // db->keyTab[slot];

	cPTB ptb;
	type_kff kff;
	cWCStr wcsUserID;

	if ( KM_GetPKBlockByID( hWnd, keyIDSBuffer, pgp_keyIDSize, false, &ptb, &kff, &wcsUserID ) ||
		 KM_GetPKBlockByID( hWnd, keyIDSBuffer + pgp_keyIDSize, localKeyIDSize, false, &ptb, &kff, &wcsUserID ) )
	{
		wcsncpy( bf, wcsUserID.m_szBody, bfSize );
	}
	else
		bytes2string( bf, bfSize, keyIDSBuffer, pgp_keyIDSize );

	return true;
}

bool ClearSlots( BYTE *dbBuff, DWORD dbSize, int slotN, int slotQ )
{
	DATA_BLOCK *db = (DATA_BLOCK *)dbBuff;

	bool bIsEncrypted = dbHeader::isEncrypted(db->header.m_body);

	if ( slotQ == -1 )
	{	
		if ( bIsEncrypted )
			slotQ = 1;
		else
			slotQ = dbKeyId::GetRequiredQ( db->header.m_body.keyMap[slotN].kbSize );
	}

	int lastSlot = slotN + slotQ;

	DWORD maxSize = sizeofBefore( DATA_BLOCK, keyTab ) + DB_MinEncryptedKeyBlockSize * lastSlot ;

	if ( maxSize > dbSize )
	{ ODS_fls( _FormatT( L"Assertion! maxSize (%X) > dbSize (%X)", maxSize, dbSize ) ); return false; }

	if ( lastSlot > DB_maxKeyQ )
	{ ODS_fls( _FormatT( L"Assertion! lastSlot (%d) > DB_maxKeyQ (%d)", lastSlot, DB_maxKeyQ ) ); return false; }

	BYTE *bytePos = &(db->keyTab[ slotN ][0]);
	ShredData( bytePos, DB_MinEncryptedKeyBlockSize * slotQ );

	if ( bIsEncrypted )
	{ ODS_fls( L"Warning! Data block is encrypted!" ); return true; }

	return dbHeader::MarkEmpty( db->header.m_body.keyMap, slotN, slotQ );
}