/*******************************************************************\
*
*	Copyright (c) 2007-2016 Jetico Inc., Finland
*	All rights reserved.
*
*	File: kgGhost.cpp
*	Revision: 
*	Created:
*	Description: main file of the KGGhost.DLL with functions that 
*		may be called from external programs
*
\*******************************************************************/

#include <windows.h>
#include <stdio.h>
#include "kgGhost.h"
#include "KBlock.h"
#include "pwdlg.h"
#include "cKeyBlockParams.h"
#include "..\..\Comm_def\Kg_defs.h"

#include "cDlg.h"
#include "FileView.h"
#include "ODS.h"
#ifdef _ODS
#include "getFileVersion.h"
#endif

#include "alg.h"

#include "..\..\..\CertificateLib\symmetricAlgs\symmetricAlgs.h"
#include "..\..\..\CertificateLib\hashAlgs\hashAlgs.h"
#include "hdlg.h"
#include "resource.h"
#include "kgGhost.h"
//#include "commands.h"
#include "ShowKeyBlockInfo.h"
#include "cDlgEncHeaderParse.h"
#include "Translatedll.h"
#include "AutoDeallocate.h"
#include "..\..\Comm_def\GenerateSeed.h"
#include <time.h>

cTranslateDll translateDll;

extern "C" __declspec( dllexport )
	void LinkTranslateDll( cTranslateDll *pDll )
{
	SetTranslateDll( pDll );
}

cAppFlags
	appFlags(
		HKEY_CURRENT_USER,
		L"SOFTWARE\\Jetico\\BestCrypt",
		L"KG-GhostFlags" );

extern bool gOldModeCompatible;

extern "C"
	int APIENTRY
		DllMain( HINSTANCE hInstance, DWORD dwReason, LPVOID lpReserved )
{
    if (dwReason == DLL_PROCESS_ATTACH)
    {
		IH.SetHolder( hInstance );
		ODS_SetInstance( hInstance );

		ODSFuncS( L"DLL_PROCESS_ATTACH" );

#ifdef _ODS
		cWCStr wcs;
		cWCS wcsPath( MAX_PATH );
		GetModuleFileNameW( hInstance, wcsPath.m_szBody, wcsPath.m_len );
		ODS_fls( _FormatT( L"Module '%s' version: %s", wcsPath.m_szBody, getModuleVersion( wcs, hInstance ) ) );
#endif

		setDialogIds( IDD_RANDOMDLG, IDC_PROGRESS, IDS_szStop, IDI_BCDoc );

		cWCStr cmd;
		cmd.Set( GetCommandLineT( true ) );

		for ( int i = 0; i < cmd.m_len; i++ )
			cmd.m_szBody[i] = towlower( cmd.m_szBody[i] );

		if ( wcsstr( cmd.m_szBody, L"keygenold" ) != NULL )
			gOldModeCompatible = true;
    }
    else if (dwReason == DLL_PROCESS_DETACH)
    {
    }

    return 1;   // ok
}

#define DISPLAY_NAME_LENGTH 32
#define SERVICE_NAME_LENGTH 32
#define EYE_CATCH_STRING    "__BCKEYGEN__"

struct BC_KEYGEN_BLOCK
{
	char  eyeCatchString[16];
	DWORD keygenID;
	char  displayName[ DISPLAY_NAME_LENGTH ];
	char  serviceName[ SERVICE_NAME_LENGTH ];
	DWORD majorVersion;
	DWORD minorVersion;
};

BC_KEYGEN_BLOCK sig = 
	{
		EYE_CATCH_STRING, 
		KEYGEN_ID_V5, // (id == 5 ) Key generator for version 9, see http://nsk.jetico.com/redmine/projects/best-crypt-for-windo/wiki/Wiki
		DISPLAY_NAME, 
		SERVICE_NAME,
		MAJOR_VERSION,
		MINOR_VERSION 
	};

//
//	QueryFunction();
//		Checks if the function is available for the data block.
//
BOOL QueryFunction(
		const BYTE	**vDataBlock,	// block of data that is stored inside file-container in encrypted form
		const DWORD	dataSize,		// size of the block of data
		const DWORD	functionID		// identefier of function used as CerateKeyHandle(.., CreateFlag, ... )
		)
{ 
	switch ( functionID )
	{
		case CFLAG_CREATE_NEW_DATA_BLOCK:
		case CFLAG_VERIFY_AND_LOAD_KEY:
		case CFLAG_CHANGE_PASSWORD:
		case CFLAG_ADD_NEW_PASSWORD:
		case CFLAG_OLD_DATA_REENCRYPTED_CONTAINER:
		case CFLAG_REMOVE_ADDITIONAL_PASSWORD:
		case CFLAG_ShowKeyBlockInfo:
		case CFLAG_SupportMode_LRW:
		case CFLAG_SupportMode_XTS:
			return true;
		case CFLAG_EncryptKeyBlock:
		{
			DATA_BLOCK *db = (DATA_BLOCK *)vDataBlock;
			if ( db == NULL ) // Create new data block.
				return true;

			if ( db->header.m_body.wKeyGenId == KEYGEN_ID_V5 )
				return isDefaultIterations( db->header.m_body.iterations );

			return true;
		}
		default :
			return false;
	}
}

//
//	Key generator has uniquie identifier
//		But can suport different.
//		This query allows to get min and max supported identifiers,
//		if function is not available, the unique idetifier is ised only.
//
bool GetSupportedIDs(
		DWORD *MajorID, 
		DWORD *MinorID 
		)
{ 
	*MajorID = KEYGEN_ID_V5;
	*MinorID = KEYGEN_ID;
	return true; 
}

/*******************************************************
 *
 *	GetNameAndVersion()
 *
 *******************************************************/
BOOL GetNameAndVersion(
		char *Name, 
		UINT  nameSize, 
		DWORD *Major, 
		DWORD *Minor 
		)
{ 
	strncpy(Name, DISPLAY_NAME, nameSize - 1);
	Name[nameSize - 1] = 0;
	*Major = MAJOR_VERSION;
	*Minor = MINOR_VERSION;
	return TRUE; 
}

DWORD _getSeedValue( HWND hWnd, BYTE  *seed, DWORD seedLength, DWORD minSeedLength )
{
	//
	//	If password is specified in command line, so seed value is generated from the password.
	//	Note, it is not secure.
	//
	cPassword tmpPwd;
	if ( GetPasswordFromCommandLine( tmpPwd.wcBuffer, NULL ) )
	{
		DWORD seedCounter = 0;

		int len = (int)wcslen( tmpPwd.wcBuffer );

		for( ; seedCounter < seedLength; seedCounter++ )
		{
			BYTE value = (BYTE)seedCounter;
			wchar_t wData = tmpPwd.wcBuffer[ seedCounter % len ] ^ (wchar_t)clock();
			for( int i = sizeof( wData ); i--; )
				value ^= BYTE( wData >> ( i * 8 ) );

			seed[ seedCounter ++ ] = value;
		}

		return ERROR_NO;
	}

	return GetSeedValue( hWnd, seed, seedLength, minSeedLength );
}

////////////////////////////////////////////////
//
//	Main functions are called by CreateKeyHandleEx function
//		1. Call seed dialog to accumulates seed
//		2. Generates key to keyInfo->key
//	Note, the keyInfo's keyAlgId and hashAlgId have to be assigned before.
//
static BOOL 
	GetSeedAndGenerateKey(
		HWND hWnd,
		sKeyInfo *keyInfo,				// Key info to be filled.
		DWORD *pErrCode,
		BYTE *pAdditionalSeed = NULL,	// If additional seed is required for kbParams block, it may be generated
		int additionalSeedSize = 0		// togather with common group.
	)
{
	ODSFunc;

	int seedTotalSize = SEED_LENGTH + additionalSeedSize;

	cPlainTextBlock ptb;
	if ( ptb.Allocate(
			seedTotalSize +
			0 )
				== NULL
		)
	{
		assert( false );
		*pErrCode = ERROR_INTERNAL_PROBLEM;
		return FALSE;
	}

	int pos = 0;
	BYTE *seedTotal	= ptb.buffer + pos; pos += seedTotalSize;
	assert( pos == ptb.bufferSize );

	int keySize = keyInfo->getAlgKeySizeBytes(); // sKeyInfo::getAlgKeySizeBytes( keyInfo->keyAlgId );
	if ( keySize <= 0 )
		return FALSE;

	*pErrCode = _getSeedValue( hWnd, seedTotal, seedTotalSize, keySize * 4 );
	if ( *pErrCode != ERROR_NO ) return FALSE;

	if ( pAdditionalSeed != NULL )
		memcpy( pAdditionalSeed, seedTotal + seedTotalSize - additionalSeedSize, additionalSeedSize );

	*pErrCode = ERROR_INCORRECT_ENCR_ALGORITHM;

	// Generates key info block

	keyInfo->size = sKeyInfo::getBufferSize( keySize );

	// Generates cont key
	*pErrCode = GenerateKey( 
			keyInfo->key, keySize,
			NULL, 0,
			seedTotal, SEED_LENGTH,
			keyInfo->hashAlgId,
			true ); // generate key

	return *pErrCode == ERROR_NO;
}

static BOOL GetAlgorithmDescription( DWORD AlgorithmId, OUT WCHAR *AlgorithmDescription, DWORD DescriptionSizeMax )
{
	cSymAlgorithm symAlgorithm;

	if (!symAlgorithm.Init( AlgorithmId, 0 ))
		return FALSE;

	MultiByteToWideChar( CP_ACP, 0, symAlgorithm.m_dispName, -1, AlgorithmDescription, DescriptionSizeMax );

	return TRUE;
}

static const wchar_t *getContName( const wchar_t *Text )
{
	// Remove artefact - word "Container:  " before actual container file path
	for( DWORD i = 0; i < wcslen( Text ); i++ )
	{
		if ((Text[i] == L' ') && (Text[i+1] == L' '))
			return Text+i+2;
	}

	return Text;
}

static BOOL
	CreateNewDataBlock(
		HWND		hWnd,
		const wchar_t	*text,		// Text, containing filename of the container
		const wchar_t	*caption,	// Caption of the "get password" dialog window

		DATA_BLOCK	**ppDB,
		DWORD		*pDBSize,
		KEY_HANDLE	*pKeyHandle,

		DWORD		*pErrCode,
		DWORD		*pFlags,

		sCKHParams *ps
	)
{
	ODSFuncS( text );

	*pErrCode = ERROR_INTERNAL_PROBLEM;

	if ( !DataBlockAllocate( ppDB, pDBSize ) )
	{
		assert( false );
		ODS_fls( L"Assertion!" );
		return FALSE;
	}

	cKeyBlockParams kbParams;

#ifdef _ODS
	if ( ( kbParams.kbPBE.m_salt.buffer[0] == 0 ) && ( kbParams.kbPBE.m_salt.buffer[1] == 0 ) )
		ODS_fls( _FormatT( L"%X, Zero salt", &kbParams.kbPBE ) );
#endif

	if ( !DataBlockInit( *ppDB, *pDBSize, ps->contAlgId, ps->encModeId, ps->hashAlgId, kbParams, &ps->contInfo ) )
	{
		DataBlockFree( ppDB, pDBSize );
		*pDBSize = 0;
		assert( false );
		ODS_fls( L"Assertion!" );
		return FALSE;
	}

#ifdef _ODS
	if ( ( kbParams.kbPBE.m_salt.buffer[0] == 0 ) && ( kbParams.kbPBE.m_salt.buffer[1] == 0 ) )
		ODS_fls( _FormatT( L"%X, Zero salt", &kbParams.kbPBE ) );
#endif

	kbParams.KeyBlockParamsInit( *ppDB, *pDBSize, ps->contAlgId, ps->encModeId, ps->hashAlgId, dbHeader::getIterationCount((*ppDB)->header.m_body), pErrCode, pFlags, hWnd );

#ifdef _ODS
	if ( ( kbParams.kbPBE.m_salt.buffer[0] == 0 ) && ( kbParams.kbPBE.m_salt.buffer[1] == 0 ) )
		ODS_fls( _FormatT( L"%X, Zero salt", &kbParams.kbPBE ) );
#endif
	//
	//	Only at creating new data block the changes are allowied.
	//
	kbParams.m_allowChangeAlgorithms = true;

	DWORD maxKeyBuffSize = 0;
	*pErrCode = DataBlock_CheckFreeSpace( *ppDB, 0, &maxKeyBuffSize );
	if ( *pErrCode != ERROR_NO ) { ODS_fls( _FormatT( L"Error %x", *pErrCode ) ); return FALSE; }

	*pErrCode = GetKeyBlockParamsWithVerification( (*ppDB)->header.m_body, hWnd, text, caption, &kbParams, pFlags, maxKeyBuffSize );
	if ( *pErrCode != ERROR_NO ) { ODS_fls( _FormatT( L"Error %x", *pErrCode ) ); return FALSE; }

	//
	//	v.2.01.5.8 - 06.04.2017
	//	cPagePBE::InitV() does it but command line skiped the check.
	//
	if ( kbParams.kbPBE.m_keyFileList.hasFiles() || appFlags.GetFlag( efUseV9Default ) )
	{
		//
		//	Version 2.01.7.1 - 17.04.2018
		//		Bug #3869 Container with keyfiles always uses Whirpool 512.
		//		No it set PBE as V9 only if it was V8.
		//
		if ( !kbParams.kbPBE.isKeyGenV5() )
			kbParams.kbPBE.setAsBCV9();
	}

	// Algorithm iDs mabe changed
	ps->contAlgId = kbParams.kbPBE.m_algId;
	ps->encModeId = kbParams.kbPBE.m_encModeId;
	ps->hashAlgId = kbParams.kbPBE.m_hashAlgId;
	kbParams.kbSSS.version = -1; // Allow to init again.

	//
	//	Init once more, the parameters may be changed.
	//
	DataBlockInit( *ppDB, *pDBSize, ps->contAlgId, ps->encModeId, ps->hashAlgId, kbParams, &ps->contInfo );

	cPlainTextBlock ptbKey;
	sKeyInfo *keyInfo = (sKeyInfo *)(ptbKey.Allocate( sizeof( sKeyInfo ) ) );
	if ( keyInfo == NULL )
	{
		*pErrCode = ERROR_INTERNAL_PROBLEM;
		assert( false );
		ODS_fls( L"Assertion!" );
		return FALSE;
	}

	keyInfo->contInfo = ps->contInfo;

//	BYTE *decryptedKeyIVandDigest;

	//
	//	Key block generation process may require additional seed, so we generate it here in common procedure.
	//
	BYTE *additionalSeed = NULL;
	int additionalSeedSize = 0;

	if ( kbParams.kbType == kbtSSS )
	{
		additionalSeedSize = kbParams.kbSSS.getSeedSize( sizeof( sKeyInfo ) ); // keyInfo->size );

		additionalSeed = kbParams.kbSSS.ptbSeed.Allocate( additionalSeedSize );
		if ( additionalSeed == NULL )
		{
			ODS_fls( L"Assertion!" );
			return false;
		}
	}

	keyInfo->keyAlgId  = kbParams.kbPBE.m_algId;
	keyInfo->encModeId = kbParams.kbPBE.m_encModeId;
	keyInfo->hashAlgId = (BYTE)kbParams.kbPBE.m_hashAlgId;
	
	if ( !GetSeedAndGenerateKey(
			hWnd,
			keyInfo,
			pErrCode,
			additionalSeed,
			additionalSeedSize )
		)
	{
		ODS_fls( L"Assertion!" );
		return false;
	}

	(*ppDB)->header.m_body.contInfo = ps->contInfo;
	(*ppDB)->header.m_body.contAlgId = keyInfo->keyAlgId;
	(*ppDB)->header.m_body.encModeId = keyInfo->encModeId;
	(*ppDB)->header.m_body.hashAlgId = keyInfo->hashAlgId;

	(*ppDB)->header.updateTail( keyInfo, (*ppDB)->pool );

	ps->keyMapInfo.setUndefined();

	*pErrCode = DataBlockWriteKey( ppDB, pDBSize, ps->bEncrypted, keyInfo, &kbParams, &(ps->keyMapInfo), NULL, NULL );
	if ( *pErrCode != ERROR_NO )
	{
		DataBlockFree( ppDB, pDBSize );
		*pDBSize = 0;
		ODSFuncRes( false );
		ODS_fls( _FormatT( L"Error %x!", *pErrCode ) );
		return FALSE;
	}

	if ( kbParams.kbType == kbtPKE )
	{
		//  PKE cannot verify encryption, so it gets key handle only
		if ( pKeyHandle != NULL )
		{
			if ( !cSymAlgorithm::createKey( 
					pKeyHandle, 
					keyInfo->keyAlgId,
					keyInfo->encModeId,
					keyInfo->key,
					keyInfo->getKeySizeBytes(),
					(*ppDB)->pool )
				)
			{
				assert( false );
				ODS_fls( L"Assertion!" );
				*pErrCode = ERROR_INTERNAL_PROBLEM;
			}
		}
	}
	else
	{
		//  verify all that we did before and get key handle
	
		ps->keyMapInfo.setUndefined();

		*pErrCode =
			DataBlockVerifyKeyBlockAndGetKey( 
					*ppDB,
					*pDBSize,
					&kbParams,
					&(ps->keyMapInfo),
					pKeyHandle,
					keyInfo,
					&(ps->decryptedHeader),
					&(ps->contInfo )
				);
	}

	if ( *pErrCode != ERROR_NO )
	{
		ODS_fls( _FormatT( L"Error %x!", *pErrCode ) );
		DataBlockFree( ppDB, pDBSize );
		return FALSE;
	}
				
	updateKeyParams( ps, keyInfo );

	bool bEncryptHeader = !appFlags.GetFlag( efDoNotEncryptHeader );

	if ( bEncryptHeader )
	{
		//
		// Version 2.01.7.4 - 24.10.2018
		//	Even if pKeyHandle was not passed or filled it will be regenerated by cSymAlgorithm
		//
		cSymAlgorithm cAlg;

		KEY_HANDLE keyHandle;

		if ( pKeyHandle != NULL )
			keyHandle = *pKeyHandle;
		else
		{
			if ( !cAlg.createKey( keyInfo->keyAlgId, keyInfo->encModeId, keyInfo->key, keyInfo->getKeySizeBytes(), (*ppDB)->pool ) )
			{
				ODS_fls( L"Assertion! Error in kreating key!" );
				DataBlockFree( ppDB, pDBSize );
				*pErrCode = ERROR_INTERNAL_PROBLEM;
				return false;
			}

			keyHandle = cAlg.m_keyHandle;
		}

		if ( !encryptHeaderEx( 
				(DATA_BLOCK *)*ppDB,
				kbParams.kbPBE.m_algId, kbParams.kbPBE.m_encModeId, kbParams.kbPBE.m_hashAlgId, keyHandle, 
				(tcEncBlock < dbHeader > *)*ppDB,
				pErrCode,
				true ) )
		{
			DataBlockFree( ppDB, pDBSize );
			return false;
		}
	}

	*pErrCode = ERROR_NO;
	return true;
}

//
//	Data block header was encrypted, so key gen reads
//	maximum available space from data block from container.
//	Now we have properly decrypted header and can recover
//	original data block size
//
static bool 
	UpdateAfterDecryption(
		DATA_BLOCK	**ppDB,
		DWORD		*pDBSize,
		dbHeader	*decHeader
	)
{
	// Gets realy busied size
	DWORD realSize = decHeader->getRealSize();

	if ( realSize == DWORD(-1) )
	{
		assert( false );
		return false;
	}

	DWORD rounded_4K_Size = DRound( realSize, _4K );

	if ( rounded_4K_Size > *pDBSize )
	{
		assert( false );
		return false;
	}

	if ( *pDBSize > _4K )
	{
		// Data block has to be truncated
		// If some far keys have been deleted
		// So getRealSize returns size less then really busied
		// so really busied block may be lost 
		// Key gen read encrypted key block
		//		[ header 4K ] .... [       maximum 16K           ]
		// Real placement
		//		[ header 4K ] ....   [ real keys ][ deleted keys ]
		// This updating does
		//		[ header 4K ] ....         [ real size from tail ]

		BYTE *dbPtr = (BYTE*)*ppDB;
		DWORD tailSize = rounded_4K_Size - _4K;

		memmove( dbPtr + _4K, dbPtr + *pDBSize - tailSize, tailSize );

		// Variants to avoid the ambiguity
		// 1. Do not allow encrypt header if data block size more then 4K
		// 2. Checks and truncates size of data block before encryption, 
		//	and does not allow to modify encrypted key block
		// 3. Updates first 4K after decryption only.

		// The varionat 3 is implemented see flags
		//	FLAG_DATA_BLOCK_WAS_MODIFIED
		//	FLAG_DATA_BLOCK_4K_WAS_MODIFIED
	}

	*pDBSize = rounded_4K_Size;

	return true;
}

//
//	readAsKBB
//		original of -> BOOL CVerifyData::ReadFromSafeFile( const wchar_t *fileName )
//
static bool readAsKBB(
		DATA_BLOCK	**ppDB,
		DWORD		*pDBSize,
		HANDLE hFile
	)
{
	assert( *ppDB == NULL );
	*ppDB = NULL;
	*pDBSize = 0;

	if ( SetFilePointer( hFile, 0, NULL, FILE_BEGIN ) != 0 )
		return false;

	DWORD read;
	DWORD blockSize;

	//
	//	Read block size first...
	//
	if ( !ReadFile( hFile, &blockSize, sizeof( blockSize ), &read, NULL ) ||
			( sizeof( blockSize ) != read ) || 
			( blockSize > DB_MaxSize ) )
				return false;

	*pDBSize = blockSize;

	if ( !DataBlockAllocate( ppDB, pDBSize ) )
	{ assert( false ); return false; }

	//
	//	Read data block...
	//
	if ( !ReadFile( hFile, *ppDB, blockSize, &read, NULL ) ||
			blockSize != read )
				return true;

	//
	//	Read hidden sector...
	//		Really hidden sector is not required, but we can check KBB file validity.
	//
	HiddenSector hs;
	if ( !ReadFile( hFile, &hs, sizeof( HiddenSector ), &read, NULL ) ||
			sizeof( HiddenSector ) != read )
				return true;

	DWORDLONG curPos = SetFilePointerT( hFile, 0, FILE_CURRENT );
	DWORDLONG size = GetFileSizeT( hFile );

	return curPos == size;
}

//
//	readAsContainer
//		original of -> bool CKeyGenerator::ReadKeyBlock_V2(  BYTE **ppDataBlock, DWORD *pDBSize, HANDLE hFile )
//
static bool readAsContainer(
		DATA_BLOCK	**ppDB,
		DWORD		*pDBSize,
		HANDLE hFile
	)

{
	assert( *ppDB == NULL );
	*ppDB = NULL;
	*pDBSize = 0;

	if ( SetFilePointerT( hFile, 0, FILE_BEGIN ) == (DWORDLONG)-1 )
		return false;

	BYTE first4K[ _4K ];

	DWORD read;

	if ( !ReadFile( hFile, first4K, sizeof( first4K ), &read, NULL ) || ( read != sizeof( first4K ) ) )
		return false;

	DWORDLONG dwlSize = GetFileSizeT( hFile );
	if ( dwlSize == (DWORDLONG)-1 )
		return false;

	DWORDLONG dwlPos;

	dbHeader &dbh = *(dbHeader *)first4K;

	//
	//	Tries to read main block first
	//
	if ( dbHeader::isEncrypted( dbh ) )
	{
		*pDBSize = DB_MaxSize;

		if ( dwlSize < *pDBSize )
			*pDBSize = (DWORD)dwlSize;

		dwlPos = dwlSize - ( *pDBSize - _4K );
	}
	else
	{
		if ( !dbHeader::isValidFileSize( dwlSize, dbh ) )
		{
		//	return false;
			dwlPos = dwlSize;
		}
		else
		{
			dwlPos = dbh.contInfo.offset + dbh.contInfo.length;
		}

		*pDBSize = (DWORD)(dwlSize - dwlPos) + _4K;
	}

	
	if ( !DataBlockAllocate( ppDB, pDBSize ) )
	{ assert( false ); return false; }

	memcpy( *ppDB, first4K, _4K );

	if ( *pDBSize > _4K  )
	{
		bool setPosErr = false;

		if ( SetFilePointerT( hFile, dwlPos, FILE_BEGIN ) == (DWORDLONG)-1 )
		{	
			// A damaged container cannot set file pointer at specified size but
			// we may to open it in readonly mode
			//	return false;
			
			setPosErr = true;
			ODS_fls( _FormatT( L"Assertion! Error (%s) in reading key block.", cWCStr().GetSystemError() ) );
		}

		read = 0;
		if ( setPosErr || !ReadFile( hFile, *ppDB + _4K, *pDBSize - _4K, &read, NULL ) || ( _4K + read < *pDBSize ) )
			*pDBSize = _4K + read;
	}

	return true;
}

// Gets new data block from kbb file.
static
	bool
		updateKBParams(
			DATA_BLOCK	**ppDB,
			DWORD		*pDBSize,
			cKeyBlockParams &kbParams,
			const wchar_t *kbbPath
		)
{
	cAutoHandle ah;
	ah.m_handle = CreateFileT(
		kbbPath,
		GENERIC_READ,
		FILE_SHARE_READ,
		NULL,
		OPEN_EXISTING,
		FILE_ATTRIBUTE_NORMAL,
		NULL);

	if ( ah.m_handle == INVALID_HANDLE_VALUE )
		return false;

	//
	// Reallocation is not required!
	//	Reason:
	//		Data block may be minimal 4K if original container header is not encrypted and was not not extended.
	//		So kbb of this container has same size.
	//		 

	FreeDataBlock( (BYTE**)ppDB );

	if ( !readAsKBB( ppDB, pDBSize, ah.m_handle ) &&
		 !readAsContainer( ppDB, pDBSize, ah.m_handle ) )
			return false;

	kbParams.KeyBlockParamsInit( *ppDB, *pDBSize );

	return true;
}

static bool getAndCheckPassword(
		HWND hWnd, const wchar_t *Text, const wchar_t *Caption,

		DATA_BLOCK	**ppDB,
		DWORD		*pDBSize,

		DWORD		*pErrCode,
		DWORD		*pFlags,

		KEY_HANDLE	*pKeyHandle,

		cKeyBlockParams &kbParams, 
		sCKHParams *ps,
		cWCS &wcsKbb,

		sKeyInfo	*keyInfo,
		tcEncBlock < dbHeader > *decHeader,

		typeGetPasswordPWD *getPwdFunc
	)
{
	ODSFuncS( Text );

	while( true )
	{
		if ( getPwdFunc != NULL )
		{
			cPassword tmpPwd;
			*pErrCode = getPwdFunc( Text, Caption, kbParams.kbPBE.m_pwd.wcBuffer, tmpPwd.wcBufferSize, false, ps->getMountPoint(), MAX_PATH );
			kbParams.kbPBE.m_pwd.ReadFromBuffer( tmpPwd.wcBuffer );
		}
		else
			*pErrCode = GetKeyBlockParams( hWnd, Text, Caption, &kbParams, pFlags,	ps, &wcsKbb ); // mountPoint );

		if ( ( *pErrCode ) != ERROR_NO )
		{
			ODS_fls( _FormatT( L"Assertion! %X", *pErrCode ) );
			return false;
		}

		// Use key block backup file!
		if ( wcsKbb.isEmpty() )
			break;
		
		if ( pFlags == NULL )
		{
			assert( false );
			break;
		}
		else
		if ( ( *pFlags & FLAG_KBB_IS_USED ) == 0 ) // Was unchecked.
			break;

		// Gets new data block from kbb file.
		if ( !updateKBParams( ppDB, pDBSize, kbParams, wcsKbb.m_szBody ) )
		{	*pErrCode = ERROR_INTERNAL_PROBLEM;
			return false;
		}

		break;
	}

	bool bIsForUsual;

	if ( kbParams.algInfo.m_algId != -1 ) // ps->contAlgId != -1 )
	{
		*pErrCode =
			DataBlockVerifyKeyBlockAndGetKey(
				*ppDB,
				*pDBSize,
				&kbParams,
				&(ps->keyMapInfo),
				pKeyHandle,
				keyInfo,
				&(decHeader->m_body),
				0,
				0,
				NULL,
				true
			);
		
		bIsForUsual = !dbContInfo::isForHiddenPart( &(keyInfo->contInfo) );
	}
	else
	{
		*pErrCode = ERROR_INCORRECT_PASSWORD;

		cProgressDlg_EncHeaderParse dlg(
						pErrCode,
						*ppDB, *pDBSize,
						&kbParams,
						pKeyHandle,
						keyInfo,
						&(decHeader->m_body),
						&(ps->keyMapInfo)
					);

		dlg.DoModal( IH.m_hInst, IDD_WaitEncHeaderParseDlg, hWnd );

		bIsForUsual = !dbContInfo::isForHiddenPart( &(keyInfo->contInfo) );

		if ( ( *pErrCode == ERROR_NO ) && ( bIsForUsual ) )
		{
			//
			//	Data block header was encrypted, so key gen reads
			//	maximum available space from data block from container.
			//	Now we have properly decrypted header and can recover
			//	original data block size
			//
			if ( !UpdateAfterDecryption( ppDB, pDBSize, &(decHeader->m_body) ) )
			{
				assert( false );
				return ERROR_INTERNAL_PROBLEM;
			}

			memcpy( &(ps->decryptedHeader), decHeader, sizeof( dbHeader ) );
		}
	}

	return *pErrCode == ERROR_NO;
}

BOOL
	GetKeyHandle(
		HWND	hWnd,
		const wchar_t	*Text,
		const wchar_t	*Caption,

		DATA_BLOCK	**ppDB,
		DWORD		*pDBSize,

		DWORD		*pErrCode,
		DWORD		*pFlags,

		KEY_HANDLE	*pKeyHandle,

		sCKHParams	*ps,

		sKeyInfo	*pKeyInfo, // = NULL

		typeGetPasswordPWD *getPwdFunc
	)
{
	ODSFuncS( Text );

	DWORD offset = ERROR_NO;	// If we find out that user entered password for
								// hidden container, offset won't be ERROR_NO

	*pErrCode = DataBlockCheck(*ppDB, *pDBSize, ps->contAlgId );

	if ( *pErrCode == ERROR_DAMAGED_DATA_BLOCK )
	{
		if ( IDYES != MessageBoxT( hWnd, 
			_LoadIDS( IDS_KBDamagedQ ),
			_LoadIDS( IDS_KeyGen ),
			MB_YESNO ) )
		{
			ODS_fls( L"Assertion! ERROR_DAMAGED_DATA_BLOCK" );
			return FALSE;
		}
	}
	else
	if ( *pErrCode != ERROR_NO )
	{
		ODS_fls( _FormatT( L"Assertion! %X", *pErrCode ) );
		return FALSE;
	}

	cKeyBlockParams kbParams;

#ifdef _DEBUG
	static bool exit = false;
	if ( exit )
		return false;
#endif

	kbParams.KeyBlockParamsInit( *ppDB, *pDBSize, ps->contAlgId, ps->encModeId, ps->hashAlgId, sCKHParams::getIterationCount( ps ), pErrCode, pFlags, hWnd );

	cPlainTextBlock ptb;

	if ( ptb.Allocate(
			sizeof( sKeyInfo ) +
			sizeof( tcEncBlock < dbHeader > ) +
			0 )
				== NULL
		)
	{
		assert( false );
		*pErrCode = ERROR_INTERNAL_PROBLEM;
		ODS_fls( L"Assertion! ERROR_INTERNAL_PROBLEM" );
		return FALSE;
	}

	memset( ptb.buffer, 0, ptb.bufferSize );

	int pos = 0;

	sKeyInfo	*keyInfo		= (sKeyInfo		*)(ptb.buffer + pos); pos += sizeof( sKeyInfo );
	tcEncBlock < dbHeader > *decHeader	= (tcEncBlock < dbHeader > *)(ptb.buffer + pos); pos += sizeof( *decHeader );

	assert( pos == ptb.bufferSize );
	
	//
	//	Sets undefined key block type,
	//	it means that all available shemes PBE, PKE and SSS will be implemented to the DATA_BLOCK.
	//	See implementaion of the idea at DataBlockVerifyKeyBlockAndGetKey.
	//
	kbParams.kbType = kbtEmpty;

	cWCS wcsKbb;

	bool bIsEncrypted = dbHeader::isEncrypted((*ppDB)->header.m_body);

	while( true )
	{
		if ( getAndCheckPassword( 
				hWnd, Text, Caption, ppDB, pDBSize, pErrCode, pFlags, pKeyHandle,
				kbParams, ps, wcsKbb, keyInfo, decHeader, getPwdFunc )
			)
				break;


#ifdef _allowBadHeader
		if ( *pErrCode == ERROR_INCORRECT_BLOCK_SIGNATURE )
		{
			ODS_fls( L"Assertion! ERROR_INCORRECT_BLOCK_SIGNATURE" );
			break;
		}
#endif

		bool bIncorrectPassword = *pErrCode == ERROR_INCORRECT_PASSWORD;
		if ( bIncorrectPassword )
		{
			const wchar_t *contName = getContName( Text );
			const wchar_t *action = Caption;
			cWCStr wcs;
			if ( action == NULL )
				action = L"Check password";

			if ( ( *pFlags & FLAG_CMD_LINE_PASSWORD ) || ( getPwdFunc != NULL ) )
				break;
			else
			{
				MessageBoxT( hWnd,
					_LoadIDS( IDS_INCORRECT_PASSWORD ),
					_LoadIDS( IDS_WARNING ),
					MB_OK | MB_ICONWARNING );

				*pErrCode = ERROR_NO;
				continue;
			}
		}

		if ( *pErrCode != ERROR_NO )
		{
			ODS_fls( _FormatT( L"Assertion! %X", *pErrCode ) );
			return false;
		}
	}

	bool isForHiddenPart = dbContInfo::isForHiddenPart( &(keyInfo->contInfo) );

	if ( (*pErrCode) == ERROR_NO )
		updateKeyParams( ps, keyInfo, &decHeader->m_body );
	else
#ifdef _allowBadHeader
	if ( (*pErrCode) == ERROR_INCORRECT_BLOCK_SIGNATURE )
		updateKeyParams( ps, keyInfo, NULL );
	else
#endif
		return FALSE;

	if ( isForHiddenPart )
		HDLG_ShowInformationMessage( IH.m_hInst, hWnd );

	copyEx< sKeyInfo >( pKeyInfo, keyInfo );

	ODSFuncRes( _FormatT( L"%d, ErrCode %X", (int)((*pErrCode) == ERROR_NO), *pErrCode ) );
	return (*pErrCode) == ERROR_NO;
}								

//
//	ChangePassword
//
//	Note, it saves existing hidden parts.
//		1. If you enter new password but it is a password for hidden part
//		the function remembers position of the valid hidden part, it allows to save the valid HP when new key is added.
//		2. The key map at ps->decryptedHeader.keyMap is updated by known hidden parts
//
static BOOL
	ChangePassword(
		HWND		hWnd,
		const wchar_t	*Text,
		const wchar_t	*Caption,

		DATA_BLOCK	**ppDB,
		DWORD		*pDBSize,

		DWORD		*pErrCode,
		DWORD		*pFlags,

		sCKHParams	*ps
	)
{
	ODSFunc;

	*pErrCode = DataBlockCheck(	*ppDB, *pDBSize, ps->contAlgId );

	if ( *pErrCode	!= ERROR_NO )
			return FALSE;

	cKeyBlockParams kbParams;
	cKeyBlockParams kbParamsNew;

	int iterations = sCKHParams::getIterationCount( ps );

#ifdef _allowBadHeader
	if ( dbHeader::isEncrypted( ps->decryptedHeader ) && ( _lastIterations != -1 ) )
		iterations = _lastIterations;
#endif

	kbParams.KeyBlockParamsInit( *ppDB, *pDBSize, ps->contAlgId, ps->encModeId, ps->hashAlgId, iterations, pErrCode, pFlags, hWnd );
	kbParamsNew.KeyBlockParamsInit( *ppDB, *pDBSize, ps->contAlgId, ps->encModeId, ps->hashAlgId, iterations, pErrCode, pFlags, hWnd );

	if (((*pErrCode) =
			GetKeyBlockParams(
				hWnd, Text, _LoadIDS( IDS_EnterOldPassword ), 
				&kbParams, pFlags ))
					!= ERROR_NO)
			return FALSE;

	bool bHidPart = FALSE;
	DWORD offset = 0;

	//
	// Allocates space for decrypted header, cont and key info.
	//
	cPlainTextBlock ptb;
	if ( ptb.Allocate(
		sizeof( sKeyInfo )	 +
		sizeof( sKeyInfo )	 +
		sizeof( sKeyInfo )	 +
		sizeof( dbHeader )	 +
		0 ) == NULL )
	{
		assert( false );
		*pErrCode = ERROR_INTERNAL_PROBLEM;
		return FALSE;
	}

	int pos = 0;
	dbHeader	*decHeader		= (dbHeader	  *)(ptb.buffer + pos); pos += sizeof( dbHeader   );
	sKeyInfo	*keyInfo		= (sKeyInfo   *)(ptb.buffer + pos); pos += sizeof( sKeyInfo   );
	sKeyInfo	*tmpKeyInfo		= (sKeyInfo   *)(ptb.buffer + pos); pos += sizeof( sKeyInfo   );
	sKeyInfo	*mainKeyInfo	= (sKeyInfo   *)(ptb.buffer + pos); pos += sizeof( sKeyInfo   );
	assert( pos == ptb.bufferSize );
	
	*pErrCode = 
			DataBlockVerifyKeyBlockAndGetKey(
				*ppDB,
				*pDBSize,
				&kbParams,
				&(ps->keyMapInfo ),
				NULL, // pKeyHandle,
				keyInfo,
				decHeader
				);

	if ( ( *pErrCode != ERROR_NO )
#ifdef _allowBadHeader
		 && ( *pErrCode != ERROR_INCORRECT_BLOCK_SIGNATURE )
#endif
		 )
			return FALSE;

	bool bDataBlockIsDamaged = ( *pErrCode == ERROR_INCORRECT_BLOCK_SIGNATURE );

	bool bIsForUsual = !dbContInfo::isForHiddenPart( &(keyInfo->contInfo) );

	//
	// Every valid password you enter here leaves traces at the keyMap ( actual for hidden part passwords )
	//	so DataBlockWriteKey will saves the known password.
	//
	dbKeyId keyMapT[ DB_maxKeyQ ];
	dbHeader::updateKeyMap( keyMapT, decHeader->keyMap );
	dbHeader::updateKeyMap( keyMapT, ps->decryptedHeader.keyMap );

	if ( bIsForUsual )
	{
		bool bNewKeyIsPublicKey = kbParamsNew.kbPKE.pka != NULL;

		if ( bNewKeyIsPublicKey )
		{
			// skip check validity for public key
		}
		else
		{
			for(;;)
			{
				if (((*pErrCode) =
						GetKeyBlockParamsWithVerification(
							ps->decryptedHeader,
							hWnd, Text, _LoadIDS( IDS_EnterNewPassword ), 
							&kbParamsNew, pFlags, -1, bDataBlockIsDamaged ))
								!= ERROR_NO )
						return FALSE;

				// Check if new password is the compatible with any old password
				sKeyMapInfo kmInfo; // for valid hidden password places collecting

				if (((*pErrCode) = DataBlockVerifyKeyBlockAndGetKey(
										*ppDB,
										*pDBSize,
										&kbParamsNew,
										&kmInfo )
									) == ERROR_NO )
				{

					// Accumulates hidden keyMapT
					dbHeader::MarkBusied( keyMapT, &kmInfo );

					MessageBoxT(
							hWnd,
							_LoadIDS( IDS_PASSWORD_IN_USE ),
							_LoadIDS( IDS_WARNING ),
							MB_OK | MB_ICONWARNING );

					if ( bCmdLinePassword )
						return FALSE;
					else
						continue;
				}

				// Overwrites at found position saved at ps->keyMapInfo.
				if (((*pErrCode) =
					DataBlockWriteKey( ppDB, pDBSize, dbHeader::isEncrypted((*ppDB)->header.m_body), keyInfo, &kbParamsNew, &(ps->keyMapInfo ), keyMapT, decHeader, bDataBlockIsDamaged ) )
							!= ERROR_NO )
						return FALSE;

				break;
			}
		}
	}
	else
	{
		// We are going to change password for hidden part?
		int iRes = MessageBoxT(
				hWnd, 
				_LoadIDS( IDS_HidPasswordChangeWarning ),
				_LoadIDS( IDS_WARNING ),
				MB_YESNO | MB_ICONWARNING );

		if (iRes == IDNO)
		{
			*pErrCode = ERROR_CANCELLED_BY_USER;
			return FALSE;
		}

		// Check loop...
		for(;;)
		{
			if (((*pErrCode) = GetKeyBlockParamsWithVerification(
						ps->decryptedHeader,
						hWnd, Text, _LoadIDS( IDS_EnterNewHiddenPassword ), 
						&kbParamsNew, pFlags, -1, bDataBlockIsDamaged ))
							!= ERROR_NO)
				return FALSE;

			// Check if password for hidden part is the compatible with any old password
			sKeyMapInfo kmInfo; // for valid hidden password places collecting

			if (((*pErrCode) = DataBlockVerifyKeyBlockAndGetKey(
									*ppDB,
									*pDBSize,
									&kbParamsNew,
									&kmInfo,
									NULL,
									tmpKeyInfo )
								) == ERROR_NO )
			{
				// We have no yet main key info, so check the tmpKeyInfo
				// Accumulates hidden keyMapT and checks validity of main cont info
				if ( dbContInfo::isForHiddenPart( &(tmpKeyInfo->contInfo) ) )
				{
					dbHeader::MarkBusied( keyMapT, &kmInfo );
				}
				else
				if ( !mainKeyInfo->isValid() )
				{
					memcpy( mainKeyInfo, tmpKeyInfo, keyInfo->size );
				}
				else
				{
					assert( memcmp( mainKeyInfo, keyInfo, keyInfo->size ) );
				}

				MessageBoxT(
					hWnd,
					_LoadIDS( IDS_PASSWORD_IN_USE ),
					_LoadIDS( IDS_WARNING ),
					MB_OK | MB_ICONWARNING );

				if ( bCmdLinePassword )
					return FALSE;
				else
					continue;
			}

			*pErrCode =
				DataBlockWriteKey_Hidden(
						ppDB,
						pDBSize,
						keyInfo,
						&kbParamsNew,
						&(ps->keyMapInfo),
						keyMapT,
						bDataBlockIsDamaged
					);

			// break from the check loop...
			// v.1.03 - 24.04.08 the break was added
			break;
		}
	}

	if ( *pErrCode != ERROR_NO )
		return FALSE;

	updateKeyParams( ps, keyInfo, decHeader );

	return TRUE;
}

//
//	AddNewKey
//		Adds new key to data block.
// !!!	1. It has to every time ask password to decode the key info has to be saved wth additional password.
//			The password may be for hidden as well as for original parts.
//		2. It asks new key block parameters
//		3. Encodes key info by new key block parameters and writes encoded new key to data block
//
//	Notes.
//		1. If you enter new password but is is a password for hidden part
//		the function remembers position of the valid hidden part, it allows to save the valid HP when new key is added.
//		2. The key map at ps->decryptedHeader.keyMap is updated by known hidden parts
//
static BOOL
	AddNewKey(
		HWND		hWnd,
		const wchar_t	*Text,
		const wchar_t	*Caption,

		DATA_BLOCK	**ppDB,
		DWORD		*pDBSize,

		DWORD		*pErrCode,
		DWORD		*pFlags,

		sCKHParams	*ps
	)
{
	if (((*pErrCode) = DataBlockCheck(*ppDB, *pDBSize, ps->contAlgId )) != ERROR_NO)
		return FALSE;

	DWORD maxFreeSize = 0;

	//bool isEncrypted = (*ppDB)->header.m_body.IsEncrypted();

	////
	////	For encrypted it will be ckecked below, after decrypting
	////
	//if ( !isEncrypted )
	//{
	//	*pErrCode = dbHeader::checkFreeSpace( (*ppDB)->header.m_body.keyMap, DB_MinEncryptedKeyBlockSize, &maxFreeSize );
	//	if( *pErrCode != ERROR_NO )
	//		return false;
	//}

	bool isEncrypted = dbHeader::isEncrypted( ps->decryptedHeader );

	//
	//	For encrypted it will be ckecked below, after decrypting
	//
	if ( !isEncrypted )
	{
		*pErrCode = dbHeader::checkFreeSpace( ps->decryptedHeader.keyMap, DB_MinEncryptedKeyBlockSize, &maxFreeSize );
		if( *pErrCode != ERROR_NO )
			return false;
	}

	cKeyBlockParams kbParams;
	cKeyBlockParams kbParamsNew;

	kbParams.KeyBlockParamsInit( *ppDB, *pDBSize, ps->contAlgId, ps->encModeId, ps->hashAlgId, sCKHParams::getIterationCount( ps ), pErrCode, pFlags, hWnd );
	kbParamsNew.KeyBlockParamsInit( *ppDB, *pDBSize, ps->contAlgId, ps->encModeId, ps->hashAlgId, sCKHParams::getIterationCount( ps ), pErrCode, pFlags, hWnd );

	if (((*pErrCode) = GetKeyBlockParams(
				hWnd, Text, _LoadIDS( IDS_ENTER_EXISTING_PASSWORD ), 
				&kbParams, pFlags )) != ERROR_NO )
		return FALSE;

	BOOL bHidPart = FALSE;

	cPlainTextBlock ptb;

	if ( ptb.Allocate(
			sizeof( sKeyInfo ) +
			sizeof( dbHeader ) +
			0 )
				== NULL
		)
	{
		assert( false );
		*pErrCode = ERROR_INTERNAL_PROBLEM;
		return FALSE;
	}

	memset( ptb.buffer, 0, ptb.bufferSize );

	int pos = 0;

	sKeyInfo *keyInfo		= ( sKeyInfo   * )(ptb.buffer + pos); pos += sizeof( sKeyInfo   );
	dbHeader *decHeader		= ( dbHeader   * )(ptb.buffer + pos); pos += sizeof( dbHeader   );

	assert( pos == ptb.bufferSize );

	*pErrCode = 
			DataBlockVerifyKeyBlockAndGetKey(
					*ppDB,
					*pDBSize,				
					&kbParams,
					&(ps->keyMapInfo),
					NULL,
					keyInfo,
					decHeader
					);

	if ( *pErrCode != ERROR_NO )
		return FALSE;

	dbKeyId	keyMapT[ DB_maxKeyQ ];
	//decHeader->updateKeyMap( keyMapT, false );
	//ps->decryptedHeader.updateKeyMap( keyMapT, false );
	dbHeader::updateKeyMap( keyMapT, decHeader->keyMap );
	dbHeader::updateKeyMap( keyMapT, ps->decryptedHeader.keyMap );

	bool bIsForUsual = !dbContInfo::isForHiddenPart( &(keyInfo->contInfo) );

	if ( !bIsForUsual )
	{
		int iRes = MessageBoxT( hWnd,
						_LoadIDS( IDS_NewHidPasswordWarning ),
						_LoadIDS( IDS_WARNING ),
						MB_YESNO | MB_ICONWARNING );

		if (iRes == IDNO)
		{
			*pErrCode = ERROR_CANCELLED_BY_USER;
			return FALSE;
		}

		bHidPart = TRUE;
		dbHeader::MarkBusied( keyMapT, &ps->keyMapInfo );
	}
	
	//
	//	For not encrypted it was ckeck above, before password entering
	//
	if ( isEncrypted )
	{
		*pErrCode = dbHeader::checkFreeSpace( decHeader->keyMap, DB_MinEncryptedKeyBlockSize, &maxFreeSize );
		if( *pErrCode != ERROR_NO )
			return false;
	}

	kbParamsNew.kbType = kbtPBE; // use it as default for not advanced dialogue.

	for(;;)
	{
		if (((*pErrCode) =
					GetKeyBlockParamsWithVerification(
						ps->decryptedHeader,
						hWnd,
						Text, _LoadIDS( IDS_EnterAdditionalPassword ), 
						&kbParamsNew,
						pFlags,
						maxFreeSize )
							) != ERROR_NO )
		{
			break;
		}

		// Check if such new password was added earlier

		if ( ( kbParamsNew.kbType != kbtPKE ) && // PKE cannot check the key because secret key has to be activated first
			 ( kbParamsNew.kbType != kbtSSS )	// SSS checks every password at creating
		    )
		{

			sKeyMapInfo kmi;

			if(	DataBlockVerifyKeyBlockAndGetKey(
						*ppDB,
						*pDBSize,
						&kbParamsNew,
						&kmi ) 
					== ERROR_NO )
			{
				dbHeader::MarkBusied( keyMapT, &kmi );

				MessageBoxT( hWnd,
					_LoadIDS( IDS_PASSWORD_IN_USE ),
					_LoadIDS( IDS_WARNING ),
					MB_OK | MB_ICONWARNING );

				if ( bCmdLinePassword )
					return FALSE;
				else
					continue;
			}
		}

		break;
	}

	if (//((*pErrCode) == ERROR_NO) ||              // such password was added earliar - nothing to do
		((*pErrCode) == ERROR_CANCELLED_BY_USER)) // process is cancelled
	{
		return FALSE; // TRUE;
	}

	// It may be required for GenerateSeed dialogue as main wnd
	kbParamsNew.kbSSS.hWnd = hWnd;

	// If key map info passed to DataBlockWriteKey or DataBlockWriteKey_Hidden is not undefined the functions shreads existing.
	ps->keyMapInfo.setUndefined();

	if (!bHidPart)
	{
		*pErrCode = DataBlockWriteKey( ppDB, pDBSize, dbHeader::isEncrypted((*ppDB)->header.m_body), keyInfo, &kbParamsNew, &(ps->keyMapInfo), keyMapT, decHeader );
	}
	else
	{
		*pErrCode =
				DataBlockWriteKey_Hidden( 
					ppDB,
					pDBSize,
					keyInfo,
					&kbParamsNew,
					&(ps->keyMapInfo),
					keyMapT ); 
	}

	if ((*pErrCode) == ERROR_NOT_ENOUGH_SPACE_FOR_KEY )
	{
		MessageBoxT( hWnd,
			_LoadIDS( IDS_NoFreeSpaceToAddPassword ),
			_LoadIDS( IDS_ERROR ),
			MB_OK | MB_ICONWARNING );
		*pErrCode = ERROR_NO;
	}

	if ( *pErrCode != ERROR_NO )
		return FALSE;

	updateKeyParams( ps, keyInfo, decHeader );

	return TRUE;
}

//
//	RemovePassword
//
static BOOL
	RemovePassword(
		HWND		hWnd,
		const wchar_t	*Text,
		const wchar_t	*Caption,

		DATA_BLOCK	**ppDB,
		DWORD		*pDBSize,

		DWORD		*pErrCode,
		DWORD		*pFlags,
		sCKHParams	*ps
	)
{
	cKeyBlockParams kbParams;
	kbParams.KeyBlockParamsInit( *ppDB, *pDBSize, ps->contAlgId, ps->encModeId, ps->hashAlgId, sCKHParams::getIterationCount( ps ), pErrCode, pFlags, hWnd );

	cPlainTextBlock ptb;

	if ( ptb.Allocate(
			sizeof( sKeyInfo ) +
			sizeof( dbHeader ) +
			0 )
				== NULL
		)
	{
		assert( false );
		*pErrCode = ERROR_INTERNAL_PROBLEM;
		return FALSE;
	}

	memset( ptb.buffer, 0, ptb.bufferSize );

	int pos = 0;

	sKeyInfo *keyInfo		= ( sKeyInfo   * )(ptb.buffer + pos); pos += sizeof( sKeyInfo   );
	dbHeader *decHeader		= ( dbHeader   * )(ptb.buffer + pos); pos += sizeof( dbHeader   );

	assert( pos == ptb.bufferSize );

	if (((*pErrCode) = 
				DataBlockCheck(*ppDB, *pDBSize, ps->contAlgId ))
					!= ERROR_NO)
			return FALSE;

	for( int iAttempts = 10; iAttempts--;  )
	{
		if (((*pErrCode) = GetKeyBlockParams( 
					hWnd,
					Text,
					_LoadIDS( IDS_EnterAdditionalPasswordToDelete ),
					&kbParams, pFlags ))
						!= ERROR_NO)
		{	
			if ( *pErrCode != ERROR_CANCELLED_BY_USER )
				*pErrCode = ERROR_NO;

			return *pErrCode == ERROR_NO;
		}

		// Check if we have entered password for hidden part. 
		// If so, refuse to delete it

		*pErrCode = 
				DataBlockVerifyKeyBlockAndGetKey(
						*ppDB,
						*pDBSize,
						&kbParams,
						&(ps->keyMapInfo ),
						NULL,
						keyInfo,
						decHeader );

		if ((*pErrCode) == ERROR_INCORRECT_PASSWORD)
		{
			MessageBoxT( hWnd,
				_LoadIDS( IDS_INCORRECT_PASSWORD ),
				_LoadIDS( IDS_WARNING ),
				MB_OK | MB_ICONWARNING );
			*pErrCode = ERROR_NO;
			continue;
		}
		else
		if ((*pErrCode) != ERROR_NO )
			return FALSE;

		bool bIsForUsual = !dbContInfo::isForHiddenPart( &(keyInfo->contInfo) );

		if ( !bIsForUsual )
		{
			//MessageBoxT( hWnd,
			//	_LoadIDS( IDS_CannotDeleteHiddenPassword ),
			//	_LoadIDS( IDS_WARNING ),
			//	MB_OK | MB_ICONWARNING );

			//*pErrCode = ERROR_NO;
			//continue;

			int res = MessageBoxT( hWnd,
				_LoadIDS( IDS_DeleteHiddenPasswordQ ),
				_LoadIDS( IDS_INFORMATION ),
				MB_YESNO | MB_ICONQUESTION );

			if ( res == IDNO )
			{
				*pErrCode = ERROR_NO;
				continue;
			}
		}

		sKeyMapInfo kmi;
		kmi.setInfo( &(ps->keyMapInfo) );

		if ( bIsForUsual )
		{
			//
			//	Note - v.8
			//	BC v.8 allows to backup and restore key block so even the master password may be deleted.
			//	BC will notify before real key block updating.
			//
			//if ( ps->keyMapInfo.position == 0 )// == ERROR_MASTER_PASSWORD_ENTERED)
			//{
			//	MessageBoxT( hWnd, 
			//		_LoadIDS( IDS_CannotDeleteMasterPassword ),
			//		_LoadIDS( IDS_WARNING ),
			//		MB_OK | MB_ICONWARNING );
			//
			//	*pErrCode = ERROR_NO;
			//	continue;
			//}

			//
			//	Simple shreds found position by ps->keyMapInfo field without overwritting because kbParams is NULL
			//
			*pErrCode = DataBlockWriteKey( ppDB, pDBSize, dbHeader::isEncrypted((*ppDB)->header.m_body), keyInfo, NULL, &kmi, NULL, decHeader );
		}
		else
		{
			if ( dbHeader::shredKeyTab( *ppDB, *pDBSize, &kmi ) )
				*pErrCode = ERROR_NO;
			else
				*pErrCode = ERROR_INCORRECT_DATA_BLOCK_SIZE;
		}

		if ( (*pErrCode) != ERROR_NO )
			return FALSE;

		// BC v.8 will notify before key block updating. See Note - v.8 above
		//MessageBoxT( hWnd,
		//	_LoadIDS( IDS_SuccessInDeletePassword ),
		//	_LoadIDS( IDS_INFORMATION ),
		//	MB_OK | MB_ICONWARNING );

		break;
	}

	if ( (*pErrCode) != ERROR_NO )
		return FALSE;

	updateKeyParams( ps, keyInfo, decHeader );

	return TRUE;
}


extern "C" BOOL CreateKeyHandleEx(
		char	*AlgServiceName,// what Encryption Algorithm will use the encryption key
		DWORD	contAlgId,		// Encryption Algorithm Identificator
		DWORD	algKeyLength,	// encryption key length for the Encryption Algorithm
		char	*Text,			// Text, containing filename of the container
		char	*Caption,		// Caption of the "get password" dialog window
		DWORD	CreateFlag,		// whether to create new key or open existing
		BYTE	**vDataBlock,	// block of data that is stored inside file-container in encrypted form
		DWORD	*pDBSize,		// size of the block of data
		DWORD	*pKeyHandle,	// returned Key Handle
		DWORD	*pErrCode,		// if the function returns FALSE, look at the Error code
		HWND	hWnd,			// parent Window for any dialogs drawn by the DLL
		DWORD	*pFavoriteDrive,
		DWORD	*pFlags          // see section "Advanced flags" in kg_defs.h
		)
{
	assert( false );
	*pErrCode = ERROR_INTERNAL_PROBLEM;
	return false;
}

/* Users' rights */

#define RIGHT_INVALID                0xffffffff
#define RIGHT_DO_NOT_BACKUP_KEYS     0x00000001
#define RIGHT_BACKUP_FOR_MY_ACCOUNT  0x00000002
#define RIGHT_BACKUP_KEY_FOR_ADMIN   0x00000004
#define DEFAULT_USER_RIGHTS          0x00000008


extern "C" BOOL CreateKeyHandle_V2(
		HWND	hWnd,			// parent Window for any dialogs drawn by the DLL
		const wchar_t *text,	// Text, containing filename of the container
		const wchar_t *caption,	// Caption of the "get password" dialog window
		DWORD	contAlgId,		// Encryption Algorithm Identificator
		DWORD	CreateFlag,		// whether to create new key or open existing
		BYTE	**vDataBlock,	// block of data that is stored inside file-container in encrypted form
		DWORD	*pDBSize,		// size of the block of data
		DWORD	*pKeyHandle,	// returned Key Handle
		DWORD	*pErrCode,		// if the function returns FALSE, look at the Error code
		DWORD	*pFlags,        // see section "Advanced flags" in kg_defs.h
		sParams *pParams
	)
{
	ODSFuncS( _FormatT( L"%s, %X", text, CreateFlag ) );

	sCKHParams *params = ( sCKHParams *)pParams;

	if ( ( params == NULL ) || ( params->size < sizeof( sCKHParams ) ) )
	{
		*pErrCode = ERROR_INCORRECT_FORMAT_VERSION;
		return false;
	}

	typeGetPasswordPWD *getPwdFunc = NULL;

	ODS_fls( _FormatT( L"%d - sCKHParams_MP_AI %d", params->size, sizeof( sCKHParams_MP_AI )  ) );

	//
	//	Traveller has no registry information so it uses this.
	//
	if ( params->size >= sizeof( sCKHParams_MP_AI ) )
	{
		sCKHParams_MP_AI *p = (sCKHParams_MP_AI *)pParams;
		if ( p->symAlgQ != 0 )
			SetSymAlgInfo( p->symAlgT, p->symAlgQ );
	}

	params->contAlgId = contAlgId;

	BOOL	bResult;

	DATA_BLOCK **ppDB = (DATA_BLOCK **)vDataBlock;

	bResult = FALSE; // Assume error will occur

	// Test if given algorithm can be called
	if ( ( contAlgId != -1 ) && !Alg_GetAlgInfoByID( contAlgId, params->encModeId ) ) // if (!Alg_VerifyDriver( AlgServiceName )) 
	{
		*pErrCode = ERROR_INVALID_ALGORITHM;
		return FALSE;
	}

	if (!(CORRECT_CFLAG(CreateFlag)))
	{
		*pErrCode = ERROR_INCORRECT_CREATE_FLAG;
		ODS_fls( L"Assertion! ERROR_INCORRECT_CREATE_FLAG" );
		return FALSE;
	}

	switch (CreateFlag)
	{
		// **************  Creating new Data Block and return Key Handle ***************

		case CFLAG_CREATE_NEW_DATA_BLOCK:
		{
			if ( params->contInfo.length == 0 )
			{
				bResult = ERROR_INCORRECT_DATA_BLOCK_SIZE;
				break;
			}

			if ( params->contInfo.offset == 0 ) // Initial call, body size is defined here
			{
				params->contInfo.offset = DB_MinSize;
			}

			params->decryptedHeader.contInfo = params->contInfo;

			bResult = CreateNewDataBlock( hWnd, text, caption, ppDB, pDBSize, pKeyHandle, pErrCode, pFlags, params );
			break;
		}

		// **************  Verify password and return Key Handle ***************

		case CFLAG_VERIFY_AND_LOAD_KEY: // 1 - minimum number of keys inside the block
		{
			//bool bUseMountPoint = (
			//	( params->size == sizeof( sCKHParamsAndMountPoint ) ) ||
			//	( params->bUseMountPoint && ( params->size > sizeof( sCKHParamsAndMountPoint ) ) )
			//	);

			//wchar_t *mountPoint = ( bUseMountPoint ? params->mountPoint : NULL );

			params->bEncrypted = dbHeader::isEncrypted((*ppDB)->header.m_body);
			//dbHeaderBase *decryptedHeader = &params->decryptedHeader;

			bResult = GetKeyHandle( hWnd, text, caption, ppDB, pDBSize, pErrCode, pFlags, pKeyHandle, //mountPoint,
								params, NULL, getPwdFunc );
			
			const wchar_t *mountPoint = params->getMountPoint();
			
			bool bUseMountPoint =
				( (( bResult != FALSE ) && ( *pErrCode == ERROR_NO ))
#ifdef _allowBadHeader
				  || ( *pErrCode == ERROR_INCORRECT_BLOCK_SIGNATURE )
#endif
				)
				&& ( mountPoint != NULL );

			if ( bUseMountPoint )
			{
				wchar_t directory[ MAX_PATH ];

				int len = (int)wcslen( mountPoint );

				if ( len > MAX_PATH )
				{
					bResult = false;
					FreeKeyHandle( contAlgId, *pKeyHandle );
					*pErrCode = ERROR_MountPointIsInvalid;
					ODS_fls( L"Asserion! ERROR_MountPointIsInvalid" );
				}
				else
				if ( len > 3 ) // more then drive name
				{
					wcsncpy( directory, mountPoint, len );
					if ( directory[ len - 1 ] == L'\\' )
						len--;

					directory[ len ] = L'\0';

					if ( !FileExistsT( directory ) )
					{
						if ( !CreateDirectoryT( directory, NULL ) )
						{
							bResult = false;
							FreeKeyHandle( contAlgId, *pKeyHandle );
							*pErrCode = ERROR_MountPointIsInvalid;
							ODS_fls( L"Asserion! ERROR_MountPointIsInvalid" );
						}
					}
				}
			}

			break;
		}
		case CFLAG_CHANGE_PASSWORD:
			bResult = ChangePassword( hWnd, text, caption, ppDB, pDBSize, pErrCode, pFlags, params );
			break;
		case CFLAG_ADD_NEW_PASSWORD:
			bResult = AddNewKey( hWnd, text, caption, ppDB, pDBSize, pErrCode, pFlags, params );
			break;
		case CFLAG_REMOVE_ADDITIONAL_PASSWORD:
			bResult = RemovePassword( hWnd, text, caption, ppDB, pDBSize, pErrCode,	pFlags,	params );
			break;

		case CFLAG_ShowKeyBlockInfo:
		{
			bResult = ShowKeyBlockInfo( hWnd, text, (dbHeader *)*ppDB, *pDBSize );
			break;
		}

		case CFLAG_EncryptKeyBlock:
		{
			bResult = EncryptHeader( hWnd, text, &(*ppDB)->header, *pDBSize, pErrCode, pFlags, params );
			break;
		}
		case CFLAG_OLD_DATA_REENCRYPTED_CONTAINER:

			*pErrCode = ERROR_NO;
			bResult = TRUE;
			break;

		default:
			*pErrCode = ERROR_INCORRECT_CREATE_FLAG;
			bResult = FALSE;
			break;
	}
	ODSFuncRes( *pErrCode );
	return bResult;
}

extern "C" BOOL CreateKeyHandle(
		char		*AlgServiceName,// what Encryption Algorithm will use the encryption key
		DWORD		AlgId,			// Encryption Algorithm Identificator
		DWORD		AlgKeyLength,	// encryption key length for the Encryption Algorithm
		char		*Text,			// Text, containing filename of the container
		char		*Caption,		// Caption of the "get password" dialog window
		DWORD		CreateFlag,		// whether to create new key or open existing
		BYTE		**vDataBlock,	// block of data that is stored inside file-container in encrypted form
		DWORD		*pDBSize,		// size of the block of data
		DWORD		*KeyHandle,		// returned Key Handle
		DWORD		*pErrCode,		// if the function returns FALSE, look on the Error code
		HWND		hWnd			// parent Window for any dialogs drawn by the DLL
		)
{
	assert( false );
	*pErrCode = ERROR_INTERNAL_PROBLEM;
	return false;
}

/*******************************************************
 *
 *	FreeDataBlock()	 - frees the memory allocated when 
 *  user calls CreateKeyHandle() procedure with 
 *  CreateFlag equal to CFLAG_CREATE_NEW_DATA_BLOCK
 *
 *******************************************************/


DWORD FreeDataBlock( BYTE **vDataBlock )
{
	if (!DataBlockFree( (DATA_BLOCK **)vDataBlock, NULL ))
	{
		assert( false );
		return ERROR_INTERNAL_PROBLEM;
	}

	return ERROR_NO;
}


/*******************************************************
 *
 *	FreeKeyHandle() - frees the memory allocated for 
 *  encryption key inside algorithm driver when user calls 
 *  CreateKeyHandle() procedure with CreateFlag equal to 
 *  CFLAG_CREATE_NEW_DATA_BLOCK or
 *  CFLAG_VERIFY_AND_LOAD_KEY
 *
 *******************************************************/


extern "C" DWORD FreeKeyHandle( 
		char	*AlgServiceName,
		DWORD KeyHandle 
	)
{
	if (!Alg_FreeKeyHandler( AlgServiceName, KeyHandle ))
	{
		assert( false );
		return ERROR_INTERNAL_PROBLEM;
	}

	return ERROR_NO;
}

DWORD FreeKeyHandle( DWORD algId, DWORD KeyHandle )
{
	cSymAlgorithm cAlg;
	if ( !cAlg.Init( algId, 0 ) )
	{
		assert( false );
		return ERROR_INTERNAL_PROBLEM;
	}

	return FreeKeyHandle( cAlg.m_algServiceName, KeyHandle );
}

/***********************************************************
 *
 *	UpdateAdvancedSettings() - The procedure calls dialog 
 *  which allows user to set options specific for 
 *	the current Key Generation module. Other BestCrypt 
 *	modules know nothing about the settings, they just 
 *	use the settings (flags) when call other Key Generation
 *	functions.
 *
 ***********************************************************/

extern "C" DWORD UpdateAdvancedSettings( DWORD *Flags, HWND  hWnd )
{
	return GetAdvancedFlags( Flags, hWnd );
}


//
//	CreateHiddenPartEx
//	It does:
//		1. Gets new key block params for hidden part
//			1.1 Checks if same password already exists.
//		// See ( Description 1.1 ) below...	
//			1.2 If the kbParams is for hidden part, it asks confirmation to overwrite the existing.
//				So space at key map table may be freed.
//			1.3 It mart the field is busied and does not allow to overwrite when key key is written to data block.
//
//		2. Verifies all we did before and gets key handle
//		
//	Rules:
//		1. It does not modify dbHeader so main part key handle is not required.
//		2. It uses keyMapT to detect empty key places, and writes new key directly to data block 
//		3. It does not extend key block, so it has to treat as busied all key in map beyond real data block size. 
//
BOOL CreateHiddenPartEx(
		HWND		hWnd,
		wchar_t		*contFileName,

		DATA_BLOCK	*pDB,			// block of data that is stored inside file-container in encrypted form
		DWORD		dbSize,			// size of the block of data

		KEY_HANDLE	*pKeyHandleHP,	// hidden part key handle 
		sCKHParams	*psHP,			// parameters for created hidden part
		
		dbKeyId		*keyMapT,		// the information about busied key blocks, new hidden part block will be created at empty key places

		DWORD		*pErrCode,		// if the function returns FALSE, look at the Error code

		DWORD		flags
	)
{
	ODSFuncS( contFileName );

	*pErrCode = ERROR_INTERNAL_PROBLEM;

	int verifyBlockSize = sizeof( tcEncBlock < sVerificationBlock > );

	if ( ( keyMapT == NULL ) || ( psHP == NULL ) || ( psHP->size < sizeof( sCKHParams ) ) )
	{
		assert( false );
		*pErrCode = ERROR_INCORRECT_FORMAT_VERSION;
		return FALSE;
	}

	cSymAlgorithm alg;

	if ( !alg.Init( psHP->contAlgId, psHP->encModeId ) )
	{
		assert( false );
		*pErrCode = ERROR_INVALID_ALGORITHM;
		return FALSE;
	}

	int iterations = sCKHParams::getIterationCount( psHP );

	cKeyBlockParams kbParams;
	kbParams.KeyBlockParamsInit( pDB, dbSize, psHP->contAlgId, psHP->encModeId, psHP->hashAlgId, iterations, pErrCode, &flags, hWnd );

	*pErrCode = ERROR_INTERNAL_PROBLEM;

	cPlainTextBlock ptbDecrypted;
	if ( ptbDecrypted.Allocate(
			sizeof( sKeyInfo ) + 
			0 ) == NULL )
	{
		assert( false );
		return FALSE;
	}

	int pos = 0;
	sKeyInfo *keyInfo		= ( sKeyInfo *)			(ptbDecrypted.buffer + pos); pos += sizeof( sKeyInfo );
	assert( pos == ptbDecrypted.bufferSize );

	keyInfo->keyAlgId  = kbParams.kbPBE.m_algId;
	keyInfo->encModeId = kbParams.kbPBE.m_encModeId;
	keyInfo->hashAlgId = (BYTE)kbParams.kbPBE.m_hashAlgId;

	keyInfo->contInfo = psHP->contInfo;

	if ( !GetSeedAndGenerateKey( hWnd, keyInfo, pErrCode ) )
		return false;

	sKeyMapInfo kmiTmp;

	for(;;)
	{
		// 1. Gets new key block params for hidden part

		//DWORD maxFreeSize = DB_MinEncryptedKeyBlockSize;
		//dbHeader::checkFreeSpace( keyMap, DB_MinEncryptedKeyBlockSize, &maxFreeSize );

		//if ( maxFreeSize < DB_MinEncryptedKeyBlockSize ) // It may be freed see comments at 1.2
		//	maxFreeSize = DB_MinEncryptedKeyBlockSize;

		*pErrCode = 
			GetKeyBlockParamsWithVerification(
					psHP->decryptedHeader,
					hWnd, contFileName,
					_LoadIDS( IDS_HIDDEN_PASSWORD ),
					&kbParams, &flags ); // , maxFreeSize );

		if ( *pErrCode != ERROR_NO )
			return FALSE;

		// 1.1 Checks if same password already exists.
		sKeyMapInfo kmTmp;

// #define _allowToOverwriteExistingHiddenPartHere_
		//	Description 1.1 + 1.2
		//	The block 1.1 checks is same password already exists.
		//	But also it checks if the existing password is for hidden part and allows to overwrite the existing one.
		//	So we have to completely decode keyInfo in the case. It is checked immediately for PBE, but SSS requires 
		//	additional shared passwords ( note, we have been going to check password only ).
		//	
		//	Description 1.1 + 1.3 without 1.2
		//	It does not check hidden or usual part has the password. It just allow to save the existing key as all create 
		//	key functions ( Add or Change ) do.
		//
#ifdef _allowToOverwriteExistingHiddenPartHere_
		//dbContInfo contInfo;

		////
		////	Sets undefined key block type,
		////	it means that all available shemes PBE, PKE and SSS will be implemented to the DATA_BLOCK.
		////	See implementaion of the idea at DataBlockVerifyKeyBlockAndGetKey.
		////
		//ekbType kbTypeSave = kbParams.kbType;
		//kbParams.kbType = kbtEmpty;

		//*pErrCode = DataBlockVerifyKeyBlockAndGetKey( pDB, dbSize, &kbParams, &kmTmp, NULL, NULL, NULL, &contInfo );
		//
		//kbParams.kbType = kbTypeSave;

		//if ( *pErrCode == ERROR_NO )
		//{
		//	bool bIsForUsual = !contInfo.isForHiddenPart(); // bodyOffset == decHeader->contInfo.bodyOffset;

		//	if ( bIsForUsual )
		//	{
		//		MessageBoxT( hWnd,
		//			_LoadIDS( IDS_PASSWORDS_EQUAL ),
		//			_LoadIDS( IDS_INVALID_PWD ),
		//			MB_OK | MB_ICONWARNING );

		//		continue;
		//	}
		//	else
		//	{
		//		// 1.2 The kbParams is for hidden part.
		//		int res = MessageBoxT( hWnd,
		//				_LoadIDS( IDS_HidPasswordOverwrite ),
		//				_LoadIDS( IDS_INFORMATION ),
		//				MB_YESNO | MB_ICONQUESTION | MB_DEFBUTTON2 );

		//		if ( res != IDYES )
		//		{
		//			// 1.3 Marks the hideen part place as busied
		//			dbHeader::MarkBusied( keyMapT, &kmTmp );
		//			continue;
		//		}

		//		if ( kmTmp.position == -1 )
		//		{
		//			assert( false );
		//		}
		//		else
		//		{
		//			// Overwrites key table place were hidden part was found and frees keyMapT.
		//			//DataBlockWriteKey( &pDB, &dbSize, pDB->header.m_body.IsEncrypted(), NULL, NULL, &kmTmp, keyMapT );

		//			// Sheades key table only..
		//			// The hidden part creation may shred the key tab without header modification
		//			dbHeader::shredKeyTab( pDB, dbSize, &kmTmp );
		//			if ( keyMapT != NULL )
		//				dbHeader::MarkEmpty( keyMapT, kmTmp.position, kmTmp.quantity );

		//			kmTmp.setUndefined();
		//		}
		//	}
		//}
#else
		if ( kbParams.kbType == kbtPKE )
		{
			// does not nothing verification is impossible...
		}
		else
		if ( kbParams.kbType != kbtSSS ) // SSS checks try to implement every password at scheme creating, so it is not required to check it again.
		{
			*pErrCode = DataBlockVerifyKeyBlockAndGetKey( pDB, dbSize, &kbParams, &kmTmp ); //, NULL, NULL, NULL, NULL );
			
			if ( *pErrCode == ERROR_NO )
			{
				// 1.3 Accumulates hidden keyMapT
				dbHeader::MarkBusied( keyMapT, &kmTmp );

				MessageBoxT(
						hWnd,
						_LoadIDS( IDS_PASSWORD_IN_USE ),
						_LoadIDS( IDS_WARNING ),
						MB_OK | MB_ICONWARNING );

				if ( bCmdLinePassword )
					return FALSE;
				else
					continue;
			}
		}
#endif

		// 2. Writes hidden part key info to data block

		*pErrCode = DataBlockWriteKey_Hidden( &pDB, &dbSize, keyInfo, &kbParams, &kmiTmp, keyMapT );
		
		if ( *pErrCode != ERROR_NO )
		{	
			if ( (*pErrCode) == ERROR_NOT_ENOUGH_SPACE_FOR_KEY )
			{
				int res = MessageBoxT( 
					hWnd,
					_LoadIDS( IDS_NoFreeSpaceToAddPassword ),
					_LoadIDS( IDS_ERROR ),
					MB_ICONSTOP | MB_OKCANCEL );

				// The free space may be freed if an existing hidden part was overwritten see p 1.2
				if ( res == IDOK )
					continue;

			}

			if ( *pErrCode != ERROR_CANCELLED_BY_USER )
			{
				assert( false );
				*pErrCode = ERROR_CANCELLED_BY_USER;
			}

			return FALSE;
		}

		break;
	}

	// 2. verifies all that we did before and gets key handle
	if ( kbParams.kbType == kbtPKE )
	{
		// does not nothing verification is impossible...
		updateKeyParams( psHP, keyInfo );
		psHP->keyMapInfo = kmiTmp;
		if ( pKeyHandleHP != NULL )
		{
			if  (!cSymAlgorithm::createKey( pKeyHandleHP, keyInfo->keyAlgId, keyInfo->encModeId, keyInfo->key, keyInfo->getKeySizeBytes(), pDB->pool ) )
			{
				assert( false );
				*pErrCode = ERROR_INTERNAL_PROBLEM;
				return FALSE;
			}
		}
	}
	else
	{
		if (((*pErrCode) = 
				DataBlockVerifyKeyBlockAndGetKey(
						pDB, dbSize,
						&kbParams, &(psHP->keyMapInfo ),
						pKeyHandleHP,
						keyInfo,
						NULL,
						&(psHP->contInfo ) )
					) != ERROR_NO )
				return FALSE;

		updateKeyParams( psHP, keyInfo );
		
		if ( !dbContInfo::isForHiddenPart( &(psHP->contInfo) ) || 
			( kmiTmp.position != psHP->keyMapInfo.position ) ||
			( kmiTmp.quantity != psHP->keyMapInfo.quantity ) )

		{
			assert( false );

			*pErrCode = ERROR_INTERNAL_PROBLEM;

			if ( pKeyHandleHP != NULL )
				FreeKeyHandle( psHP->contAlgId, *pKeyHandleHP );

			return FALSE;
		}
	}
				
	*pErrCode = ERROR_NO;
	return TRUE;
}	

/***********************************************************************************************
 *
 * The function is needed if the Key Generator is developed for using with
 * some of the latest BestCrypt versions - because of some reason.
 *
 ***********************************************************************************************/

BOOL CheckBestCryptVersion(
		HWND    hWnd,
		DWORD   MajorVersion,
		DWORD   MinorVersion 
		)
{
	if ( MajorVersion < 8 )       return FALSE;
	
	return TRUE;
}

//
//	UpdateHeader
//		The function decryptes/encrypts data block to outBuff
//		Note, inBuff may be same as outBuff
//		If data block is encrypted so the header, contInfo and keyMap is encrypted
//
BOOL UpdateHeader(
	HWND	hWnd,			// parent Window for any dialogs drawn by the DLL
	BYTE	*dataBlock,
	DWORD	dataBlockSize,
	DWORD	algId,			// Encryption Algorithm Identificator
	DWORD	encModeId,		// Encryption mode
	DWORD	keyHandle,		// Key Handle
	BYTE	*outBuff,		// Caller procedure is responsible for secure erasing of the decrypted data
	DWORD	buffSize,
	DWORD	*pErrCode,		// if the function returns FALSE, look on the Error code
	sUBParam	*param		// = NULL
)
{
	DWORD errCode;
	if ( pErrCode == NULL )
		pErrCode = &errCode;
	
	//*pErrCode = verifyKeyHandle( (DATA_BLOCK *)dataBlock, keyHandle, algId, encModeId );
	//
	//if ( *pErrCode != ERROR_NO )
	//	return false;

	if ( param != NULL )
	{
		DWORD hashAlgId = 0;//bchaSHA256;

		sParamEncrypt *sp = (sParamEncrypt *)param;
		if ( param->size >= sizeof( sParamEncrypt ) )
			hashAlgId = sp->hashAlgId;

		if ( hashAlgId == 0 )
			hashAlgId = bchaSHA256;

		if (  ( param->flags == UBFlag_Encrypt ) ||
		      ( param->flags == UBFlag_Encrypt_Digest_Only ) )
		{
			if ( ( outBuff != NULL ) && ( buffSize < sizeof( tcEncBlock < dbHeader > ) ) )
			{
				*pErrCode = ERROR_BUFFER_TOO_SMALL;
				return false;
			}

			bool bEntire = ( param->flags == UBFlag_Encrypt );

			return
				encryptHeaderEx( 
						(DATA_BLOCK *)dataBlock,
						algId, encModeId, hashAlgId, keyHandle, 
						(tcEncBlock < dbHeader > *)outBuff,
						pErrCode,
						bEntire );
		}
		else
		if ( param->flags == UBFlag_Decrypt )
			return decryptHeaderEx( 
						(DATA_BLOCK *)dataBlock,
						algId, encModeId, &sp->hashAlgId, keyHandle,
						(tcEncBlock < dbHeader > *)outBuff,
						pErrCode );

		return false;
	}
	else
		return decryptHeaderEx(
						(DATA_BLOCK *)dataBlock,
						algId, encModeId, NULL, keyHandle,
						(tcEncBlock < dbHeader > *)outBuff,
						pErrCode );
}

int defIterationsT[2] = {
	DEFAULT_ITERATIONS_V5,
	DEFAULT_ITERATIONS_V4,
};

bool isDefaultIterations( int iterations )
{
	for( int i = maxLen( defIterationsT ); i--; )
	{
		if ( defIterationsT[i] == iterations )
			return true;
	}

	return false;
}

//static
int sCKHParams::getIterationCount( sCKHParams *ps )
{
	if ( ps == NULL )
		return DEFAULT_ITERATIONS_V5;

	return dbHeader::getIterationCount( ps->decryptedHeader );
}

#ifdef _allowBadHeader
int _lastIterations = -1;
#endif
