#include "cDlgKBParams.h"
#include "resource.h"
#include <windowsx.h>
#include "diskDriver.h"
#include "AutoDeallocate.h"
#include "cDlgKeyFileMan.h"
#include "..\..\..\Includes\OpenFiles.h"
#include "..\..\..\Includes\cArgs.h"

static void DrawFileIcon( HWND hDlg, HDC dc, HICON hIcon, int id )
{
	if ( !hIcon )
		return;

	HWND hWnd = GetDlgItem( hDlg, id );

	if ( hWnd == NULL )
		return;

	ShowWindow( hWnd, SW_HIDE );

	RECT r;
	GetWindowRect( hWnd, &r );
	MapWindowPoints( NULL, hDlg, (POINT*)&r, 2 );
	DrawIcon( dc, r.left, r.top, hIcon );
}

//
//	processContIcon
//		preprocesses dlg window procedure and exchange container icon to really extracted 
//		document from main exe instance.
//
//		use it just before main message switch
//
//static void processContIcon( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
//{
//	static HICON hContIcon = NULL;
//
//	switch (uMsg) 
//	{
//        case WM_INITDIALOG:
//			wchar_t path[ MAX_PATH ];
//			GetModuleFileNameT( NULL, path, maxLen( path ) );
//			hContIcon = ExtractIconT( (HINSTANCE)GetWindowLongPtrT( hwnd, GWLP_HINSTANCE ), path, 1 );
//			break;
//
//		case WM_PAINT :
//		{
//			HDC dc;
//			PAINTSTRUCT ps;
//
//			if ((HDC) wParam != NULL)
//				dc = (HDC) wParam;
//			else
//				dc = BeginPaint(hwnd,&ps);
//
//			DrawFileIcon( hwnd, dc, hContIcon, IDC_Icon );
//
//			if ((HDC) wParam == NULL)
//				EndPaint(hwnd,&ps);
//
//			break;
//		}
//
//		case WM_DESTROY:
//
//			if ( hContIcon != NULL )
//			{
//				DestroyIcon( hContIcon );
//				hContIcon = NULL;
//			}
//
//			break;
//
//		default :
//			break;
//	}
//}

static bool getRootDir( wchar_t *root, const wchar_t *path )
{
	// 'C:\qqq\"
	// \\PAV\c\qqq\

	if ( path[1] == L':' )
	{
		wcsncpy( root, path, 3 );
		root[3] = L'\0';
		return true;
	}
	else
	if ( path[0] == L'\\' )
	{
		int q = 4;
		int i = 0;

		for( ; path[i] != 0; )
		{
			root[i] = path[i];
			if ( path[i++] == L'\\' )
			{
				if ( --q == 0 )
					break;
			}
		}

		root[i] = L'\0';
		return true;
	}

	return false;
}

static bool isNTFS( const wchar_t *path, int *pErrIDS )
{
	wchar_t volume[ MAX_PATH ];
	DWORD dwSerNum, dwMaxComLen, dwFSFlags;
	wchar_t fileSystem[ 64 ] = L"";

	*pErrIDS = IDS_FolderInvalidPathName;

	wchar_t root[ MAX_PATH ];
	if ( !getRootDir( root, path ) )
		return false;

	if ( !GetVolumeInformationT(
			root,
			volume, maxLen( volume ),
			&dwSerNum,
			&dwMaxComLen,
			&dwFSFlags,
			fileSystem,
			maxLen( fileSystem ) )
		)
		return false;

	if ( wcsncmp( fileSystem, L"NTFS", 4 ) == 0 )
		return true;

	*pErrIDS = IDS_FolderNotNTFS;

	return false;
}

//
//	Checks if the folder may be used as mount point.
//

static bool checkFolder( wchar_t *folderName, int *pErrIDS )
{
	int errIDS;
	if ( pErrIDS == NULL )
		pErrIDS = &errIDS;

	*pErrIDS = IDS_FolderNameIsInvalid;

	wchar_t path[ MAX_PATH ];

	if ( !GetDriveRootDirT( path, folderName, true ) )
		return false;

	UINT drvType = GetDriveTypeT( path );

	if ( drvType == DRIVE_UNKNOWN )
	{
		*pErrIDS = IDS_DriveTypeIsUnknown;
		return false;
	}

	if ( drvType == DRIVE_NO_ROOT_DIR )
	{
		*pErrIDS = IDS_FolderIsNotRootDir;
		return false;
	}

	if ( drvType == DRIVE_REMOTE )
	{
		*pErrIDS = IDS_IsNetworkDrive;
		return false;
	}

	int len = (int)wcslen( folderName );

	if ( len > MAX_PATH )
		return false;

	if ( folderName[ len - 1 ] == L'\\' )
		len--;

	wchar_t oldName[ MAX_PATH ];
	wcsncpy( oldName, folderName, len );
	oldName[ len ] = L'\0';

	if ( !FileExistsT( oldName ) )
	{
		if ( CreateDirectoryT( oldName, NULL ) )
			return true;
	}

	*pErrIDS = IDS_FolderInvalidPathName;

	if ( len > MAX_PATH - 3 )
		return false;

	if ( !isNTFS( folderName, pErrIDS ) )
		return false;

	wcscpy_s( path, folderName );
	wcscat_s( path, L"\\*" );

	WIN32_FIND_DATAW findData;

	HANDLE hFind = FindFirstFileT( path, &findData );

	*pErrIDS = IDS_FolderDoesNotExistOrAccessIsDenied;

	if ( hFind == INVALID_HANDLE_VALUE )
		return false;

	*pErrIDS = IDS_NoError;

	bool isEmpty = true;

	do
	{
		if ( wcscmp( findData.cFileName, L".." ) == 0 ) continue;
		if ( wcscmp( findData.cFileName, L"." ) == 0 ) continue;

		*pErrIDS = IDS_FolderIsNotEmpty;

		isEmpty = false;

		break;
	}
	while( FindNextFileT( hFind, &findData ) );

	FindClose( hFind );

	if ( !isEmpty )
		return false;

	return true;
}

bool cDlgGetPassword::ShowKeyFilesButton()
{
	bool bHasKeyFiles = m_keyFileList.hasFiles();

	// Force advanced state if there are key files for simple view
	m_advancedState->Update( bHasKeyFiles );

	bool bShow = m_bAdvanced || bHasKeyFiles;

	ShowWindow( GetItem( IDC_btnKeyfiles ), bShow ? SW_SHOW : SW_HIDE );

	RECT rk;
	if ( !GetDlgItemRect( m_hDlg, IDC_btnKeyfiles, &rk ) )
		return false;

	cRect rp;
	if ( !GetDlgItemRect( m_hDlg, IDC_PASS, &rp ) )
		return false;

	rp.right = bShow ? rk.left - 2 : rk.right;
	SetWindowPos( GetItem( IDC_PASS ), NULL, 0, 0, rp.Width(), rp.Height(), SWP_SHOWWINDOW | SWP_NOZORDER | SWP_NOMOVE );

	return true;
}

struct sCtrl
{
	HWND m_hWnd;
	RECT m_rect;
};

//
//	class cAdvancedStateMP
//		Advanced styate for dialog with mount parameters
//
class cAdvancedStateMP :
	public cAdvancedState,
	public cEnumDlgItem
{
	sCtrl m_T[3];
	cADBlock < HWND > m_hiddenT;
	RECT m_rUpper;
	cRect m_rDlg;
	int m_dY;

public:

	cAdvancedStateMP( bool &bAdvanced, HWND hDlg )
		: cAdvancedState( bAdvanced, hDlg )
	{
	}

	virtual bool Init();
	virtual bool Update( bool bForceAdvanced = true );

	virtual bool childActor( HWND hWnd );
};

//
//	childActor
//		Is called by Walk function for all dialog items.
//		It callects items to be hidden at state is not advanced.
//
//virtual 
bool cAdvancedStateMP::childActor( HWND hWnd )
{
	RECT r;
	cDlg::GetDlgItemRect( m_hDlg, hWnd, &r );

	if ( r.top <= m_rUpper.bottom )
		return true;

	for( int i = maxLen( m_T ); i--; )
	{
		if ( m_T[i].m_hWnd == hWnd )
			return true;
	}

	m_hiddenT.Add( hWnd );
	return true;
}

bool cAdvancedStateMP::Init()
{
	cAdvancedState::Init();

	m_T[0].m_hWnd = m_hSysLink_Advanced;
	m_T[1].m_hWnd = ::GetDlgItem( m_hDlg, IDOK );
	m_T[2].m_hWnd = ::GetDlgItem( m_hDlg, IDCANCEL );

	for( int i = maxLen( m_T ); i--; )
		cDlg::GetDlgItemRect( m_hDlg, m_T[i].m_hWnd, &m_T[i].m_rect );

	HWND hUpper = ::GetDlgItem( m_hDlg, IDC_stGrp );

	if ( hUpper == NULL )
		return false;

	cDlg::GetDlgItemRect( m_hDlg, hUpper, &m_rUpper );
	GetWindowRect( m_hDlg, &m_rDlg );

	m_dY = m_T[0].m_rect.top - ( m_rUpper.bottom + 15 );

	return Walk( m_hDlg );
}

//virtual
bool cAdvancedState::Update( bool bForceAdvanced )
{
	if ( bForceAdvanced )
		m_bAdvanced = true;

	SetWindowTextT( m_hSysLink_Advanced, m_bAdvanced ? m_wcs[0].m_szBody : m_wcs[1].m_szBody );
	return true;
}

//virtual
bool cAdvancedState::Init()
{
	m_hSysLink_Advanced = ::GetDlgItem( m_hDlg, IDC_slAdvancedSettings );

	m_wcs[0].GetWindowTextT( m_hSysLink_Advanced );
	m_wcs[1].LoadIDS( IDS_szShowAdvancedSettings );

	return true;
}


bool cAdvancedStateMP::Update( bool bForceAdvanced )
{
	cAdvancedState::Update( false );

	int i;
	cRect r;

	for( i = m_hiddenT.m_Q; i--; )
		ShowWindow( m_hiddenT.m_T[i], m_bAdvanced ? SW_SHOW : SW_HIDE );

	ShowWindow( GetDlgItem( m_hDlg, IDC_chkMountAsRemovable ), ( m_bAdvanced && allowMountAsRemovable() ) ? SW_SHOW : SW_HIDE );

	for( i = maxLen( m_T ); i--; )
	{
		r.Set( m_T[i].m_rect );
		r.top -= m_bAdvanced ? 0 : m_dY;
		SetWindowPos( m_T[i].m_hWnd, NULL, r.left, r.top, 0, 0, SWP_SHOWWINDOW | SWP_NOZORDER | SWP_NOSIZE );
	}

	r = m_rDlg;
	r.bottom -= m_bAdvanced ? 0 : m_dY;
	SetWindowPos( m_hDlg, NULL, 0, 0, r.Width(), r.Height(), SWP_SHOWWINDOW | SWP_NOZORDER | SWP_NOMOVE );

	return true;
}

void cDlgGetPassword::UpdateAdvancedState( bool bInit )
{
	if ( m_advancedState == NULL )
	{
		if ( !bInit )
		{ assert( false ); return; }

		HWND hChk = GetItem( IDC_chkMountWithKBB );

		if ( hChk != NULL )
			m_advancedState = new cAdvancedStateMP( m_bAdvanced, m_hDlg );
		else
			m_advancedState = new cAdvancedState( m_bAdvanced, m_hDlg );
			
		if ( m_advancedState == NULL )
		{ assert( false ); return; }

		m_advancedState->Init();
	}

	ShowKeyFilesButton();
}

#define EndDialog_ODS( _hDlg, _dlgRes ) ODS_fls( _FormatT( L"EndDialog( %X )", _dlgRes) ); EndDialog( _hDlg, _dlgRes );

LRESULT cDlgGetAdvancedFlags::RealDlgProc( UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	switch (uMsg) 
	{ 
        case WM_INITDIALOG:
		{
			Button_SetCheck( GetDlgItem( m_hDlg, IDC_STYPING),  ( m_keyGenFlags & FLAG_ADVANCED_SECURE_TYPING) > 0);
			Button_SetCheck( GetDlgItem( m_hDlg, IDC_SHOWCHARS), ( m_keyGenFlags & FLAG_ADVANCED_SHOW_PWD_CHARS) > 0);

			//
			//	Set the flag, otherwise kbdEdit does not show valid characters
			//	They are read as buffer if OK is pressed.
			//
			m_keyGenFlags |= FLAG_ADVANCED_TEST; // FLAG_ADVANCED_SHOW_PWD_CHARS;

			m_useSecureTyping = ( Button_GetCheck( GetDlgItem( m_hDlg, IDC_STYPING) ) == BST_CHECKED );
			BOOL res = kbd.GenerateEncodingTable( TRUE ); //m_useSecureTyping );

			Edit_SetPasswordChar( GetDlgItem( m_hDlg, IDC_EDITSTRING ), 0 );

			m_edt.Attach( GetDlgItem( m_hDlg, IDC_EDITSTRING ) );

			SetFocus( m_edt.m_hEdit );

			return res;
		}
		case WM_DESTROY:
		{
			m_edt.Detach();
			kbd.FreeTables( TRUE );

			break;
		}

		case WM_BufferIsUpdated:
		{
			SetWindowTextT( GetItem( IDC_showInside ), m_edt.m_buffer );
			SetWindowTextT( GetItem( IDC_showOutside ), m_edt.m_rndBuffer );
			break;
		}

		case WM_COMMAND: 
		{
			int cmd = LOWORD( wParam );

			switch( cmd ) 
			{ 
				case IDC_STYPING:
				{
					m_useSecureTyping = ( Button_GetCheck( GetDlgItem( m_hDlg, IDC_STYPING) ) == BST_CHECKED );
					ShowWindow( GetDlgItem( m_hDlg, IDC_LINE ), m_useSecureTyping );
					break;
				}
				case IDC_SHOWCHARS:
			        if ( Button_GetCheck( GetDlgItem( m_hDlg, IDC_SHOWCHARS ) ) )
						 m_keyGenFlags |= FLAG_ADVANCED_SHOW_PWD_CHARS;
					else m_keyGenFlags &= ~FLAG_ADVANCED_SHOW_PWD_CHARS;

					break;
				case IDOK: 
 
			        if ( Button_GetCheck( GetDlgItem( m_hDlg, IDC_STYPING ) ) )
						 m_keyGenFlags |= FLAG_ADVANCED_SECURE_TYPING;
					else m_keyGenFlags &= ~FLAG_ADVANCED_SECURE_TYPING;

			        if ( Button_GetCheck( GetDlgItem( m_hDlg, IDC_SHOWCHARS ) ) )
						 m_keyGenFlags |= FLAG_ADVANCED_SHOW_PWD_CHARS;
					else m_keyGenFlags &= ~FLAG_ADVANCED_SHOW_PWD_CHARS;

					m_keyGenFlags &= ~FLAG_ADVANCED_TEST;

					EndDialog_ODS( m_hDlg, ERROR_NO);
					return TRUE; 
 
				case IDCANCEL: 
	
					m_keyGenFlags &= ~FLAG_ADVANCED_TEST;

					EndDialog_ODS( m_hDlg, (int)ERROR_CANCELLED_BY_USER); 
					return TRUE; 
			} 

            return 0; 
		}
    } 
    return FALSE; 

	UNREFERENCED_PARAMETER(lParam);
}

bool cDlgGetPasswordAndMount::checkEmptyFolder()
{
	int errIDS;

	if ( checkFolder( m_mountPoint, &errIDS ) )
		return true;

	cWCStr msg;
	
	msg.FormatT( _LoadIDS( IDS_InvalidMountPoint_S ), m_mountPoint );
	msg.Add( L"\n\n" );
	msg.Add( _LoadIDS( errIDS ) );

	DlgMessageBoxT( msg.m_szBody, MB_ICONSTOP, _LoadIDS( IDS_ERROR ) );

	return false;
}

bool cDlgGetPasswordAndMount::OnBrowseForMountPoint( HWND hCtrl )
{
	if ( m_mountPoint == NULL )
		return false;

	HWND cmbHwnd = GetDlgItem( m_hDlg, IDC_DRIVES );

	if ( hCtrl != cmbHwnd )
		return false;

	int curSel = ComboBox_GetCurSel( cmbHwnd );

	if ( curSel != m_posBrowseForMountPoint )
		return false;

	while( true ) 
	{
		UINT uiFlags = BIF_RETURNONLYFSDIRS | BIF_NEWDIALOGSTYLE;

		if ( !BrowseForFolderT( m_hInst, IDS_ChooseFolderForMounting, m_hDlg, m_mountPoint, uiFlags  ) )
		{
			ComboBox_SetCurSel( cmbHwnd, 0 );
			return false;
		}
		
		int len = (int)wcslen( m_mountPoint );
		if ( len == 0 )
			return false;

		if ( m_mountPoint[ len - 1 ] != L'\\' )
			wcsncpy( m_mountPoint + len, L"\\", MAX_PATH - len );

		if ( checkEmptyFolder() )
			break;
	} 

	SelectMountPoint( cmbHwnd );
	return true;
}

bool isDriveLetter( const wchar_t *mountPoint, int *pNum = NULL )
{
	if ( wcslen( mountPoint ) != 3 )
		return false;

	if ( wcscmp( mountPoint + 1, L":\\" ) != 0 )
		return false;

	if ( towupper( mountPoint[0] ) < L'A' )
		return false;

	if ( towupper( mountPoint[0] ) > L'Z' )
		return false;

	if ( pNum )
		*pNum = ( towupper( mountPoint[0] ) - L'A' );

	return true;
}

//
//	ComboBoxEx32 is exluded now ( it does not work for windows 98 ).
//	The combobox control works now perfectly for unicode texts
//
#define myComboBoxEx_GetItemTextT ComboBox_GetLBTextT
#define myComboBoxEx_InsertStringT ComboBox_InsertStringT
#define myComboBoxEx_SetItemTextT ComboBoxEx_SetItemTextT
#define myComboBoxEx_GetTextT GetWindowTextT

int cDlgGetPasswordAndMount::SelectMountPoint( HWND hCmb )
{
	if ( m_mountPoint == NULL )
		return 0;

	int count = ComboBox_GetCount( hCmb );

	// 1. Finds mount point name at available drives
	for (; count--; )
	{
		wchar_t text[ MAX_PATH ];
		myComboBoxEx_GetItemTextT( hCmb, count, text, maxLen( text ) );

		if ( _wcsicmp( text, m_mountPoint ) == 0 )
		{
			ComboBox_SetCurSel(  hCmb, count );
			return count;
		}
	}

	// 2. Checks if the mount point is a busied drive. If so it selects first available drive incombobox.

	if ( isDriveLetter( m_mountPoint ) )
	{
		ComboBox_SetCurSel( hCmb, 0 );
		return 0;
	}


	// 3. Adds mount point name  at the combobox tail and selects it
	int mpPos = m_posBrowseForMountPoint + 1;

//	myComboBoxEx_SetItemTextT( hCmb, mpPos, m_mountPoint );
	ComboBox_DeleteString( hCmb, mpPos );
	//mpPos = myComboBoxEx_InsertStringT( hCmb, mpPos, m_mountPoint );

	// See Note 1. 29.10.2009
	cWCStr wcs( m_mountPoint );
	mpPos = myComboBoxEx_InsertStringT( hCmb, mpPos, wcs.m_szBody );

	return ComboBox_SetCurSel(  hCmb, mpPos );
}

bool cDlgGetPasswordAndMount::updateFlags( int idc, DWORD mask )
{
	HWND hWnd = GetItem( idc );

	if ( hWnd == NULL )
		return false;

	if ( Button_GetCheck( hWnd ) == BST_CHECKED )
			m_keyGenFlags |= mask;
	else	m_keyGenFlags &= ~mask;

	return true;
}

bool cDlgGetPasswordAndMount::GetMountParams()
{
	if ( m_mountPoint == NULL )
		return true;

	int curSel = ComboBox_GetCurSel( GetDlgItem(m_hDlg, IDC_DRIVES) );
	if (  curSel == m_posBrowseForMountPoint )
	{
		assert( false );
		return false;
	}

	updateFlags( IDC_AUTOMOUNT, FLAG_CONTAINER_AUTO_MOUNT );
	updateFlags( IDC_READONLY, FLAG_CONTAINER_READ_ONLY );

	if ( m_hDlg == NULL )
	{
		assert( false );
		wcscpy_s( m_mountPoint, MAX_PATH, L"W:\\" );
		return true;
	}

	HWND cmbHwnd = GetDlgItem( m_hDlg, IDC_DRIVES );

	if ( cmbHwnd == NULL )
	{
		assert( false );
		wcscpy_s( m_mountPoint, MAX_PATH, L"Y:\\" );
		return true;
	}

	//
	//	Note 1. 29.10.2009
	//	The m_mountPoint has been allocated at BestCrypt.exe and pased to KGGhost.dll.
	//	The combobox function cannot read and write from the memory. 
	//	That is the result of strange behaivior, see below.
	//
	
	cWCStr mp( MAX_PATH );
	
	myComboBoxEx_GetTextT( cmbHwnd, mp.m_szBody, mp.m_len );

	wcsncpy( m_mountPoint, mp.m_szBody, MAX_PATH );

	return true;
}

struct sWinVer : public OSVERSIONINFO
{
	sWinVer()
	{
		dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
		GetVersionEx( this );
	}

	bool is2K()
	{
		return ( ( dwPlatformId == VER_PLATFORM_WIN32_NT ) && ( dwMajorVersion >= 5 ) );
	}
};

//virtual
HRESULT cDlgGetPasswordAndMount::Drop( cFilesList &ofl )
{
	while( ofl.getNext() )
		m_keyFileList.m_sa.addString( ofl.getCur() );

	ShowKeyFilesButton();
	m_keyFileList.updateWindowText( GetItem( IDC_btnKeyfiles ) );

	return S_OK;
}


bool cDlgGetPasswordAndMount::SetMountParams()
{
	Button_SetCheck( GetDlgItem(m_hDlg, IDC_AUTOMOUNT), ( m_keyGenFlags & FLAG_CONTAINER_AUTO_MOUNT) > 0);

	Button_SetCheck( GetDlgItem(m_hDlg, IDC_READONLY),  ( m_keyGenFlags & FLAG_CONTAINER_READ_ONLY) > 0);

	if ( m_keyGenFlags & FLAG_FORCE_READ_ONLY )
	{
		ODS_fls( L"m_keyGenFlags & FLAG_FORCE_READ_ONLY" );
		Button_SetCheck( GetDlgItem(m_hDlg, IDC_READONLY),  TRUE  );
		Button_Enable  ( GetDlgItem(m_hDlg, IDC_READONLY),  FALSE );
	}

	//
	//	08.10.2014
	//	Additional commands are blocked.
	//	So the FLAG_EXPIRED_DEMO has to work.
	//
	//if (  EvaluationVersionHasExpired ||			// Is assigned at GetAdditionalCommandNames but does not saved at time password is entered.
	if ( m_keyGenFlags & FLAG_EXPIRED_DEMO )
	{
		cWCStr wcs;
		wcs.GetWindowTextT( m_hDlg );
		wcs.AddF( L" (%s)", _LoadIDS( IDS_szExpiredCaption ) );
		SetWindowTextT( m_hDlg, wcs.m_szBody );
	}

	HWND cmbHwnd = GetDlgItem(m_hDlg, IDC_DRIVES);

	wchar_t  driveName[] = L"A:\\";

	if ( ( m_keyGenFlags & FLAG_MOUNT_ON_SINGLE_LETTER ) != 0 )
	{
		//
		// NT4 rudiment.
		//	if cont was not completely dismounted it will be mounted on the drive it was mounted before
		//

		assert( wcslen( m_mountPoint ) != 0 );

		// Spec hidden part creation function uses same flag for mounting before free cluster chain reading.
		EnableWindow( GetDlgItem( m_hDlg, IDC_AUTOMOUNT ), false );
		ODS_fls( L"m_keyGenFlags & FLAG_MOUNT_ON_SINGLE_LETTER" );
		EnableWindow( GetDlgItem( m_hDlg, IDC_READONLY ), false );
		m_posBrowseForMountPoint = -1;
	}
	else
	{
		DWORD curLogicalDisks = _GetLogicalDrives();

		for ( int i=2; i<26; i++ )
		{
			if ((curLogicalDisks & (1 << i)) == 0)
			{
				driveName[0] = L'A' + i;
				myComboBoxEx_InsertStringT( cmbHwnd, -1, driveName );
			}
		}

		int slotQ = ComboBox_GetCount( cmbHwnd );

		static sWinVer winVer;
		if ( !m_bTravellerMode && winVer.is2K() )
			m_posBrowseForMountPoint =
				myComboBoxEx_InsertStringT( cmbHwnd, -1, _LoadIDS( IDS_BrowseForMountPoint ) );
		else
		{
			if ( slotQ == 0 )
			{
				DlgMessageBoxT(
					L"There are no free slots for mounting!",
					MB_ICONSTOP,
					_LoadIDS( IDS_Attention ) );

				EndDialog_ODS( m_hDlg, ERROR_MountPointIsInvalid );
				return false;
			}

			m_posBrowseForMountPoint = -1;
		}
	}

	int len = (int)wcslen( m_mountPoint );
	if ( len != 0 )
		SelectMountPoint( cmbHwnd );
	else
		ComboBox_SetCurSel( cmbHwnd, 0 );

	return true;
}

static bool isFreeDrive( int drv )
{
	DWORD dwMask = GetLogicalDrives();
	return ( ( dwMask & ( 1 << drv ) ) == 0 );
}

bool cDlgGetPasswordAndMount::CheckValidMountPoint()
{
	int drv;

	if ( isDriveLetter( m_mountPoint, &drv ) )
	{
		if ( !isFreeDrive( drv ) )
		{
			DlgMessageBoxT(
				cWCStr().FormatT( _LoadIDS( IDS_Drive_S_isBusied ), m_mountPoint ),
				MB_ICONEXCLAMATION,
				_LoadIDS( IDS_WARNING ) );

			return false;
		}
	}
	else
	if ( !checkEmptyFolder() )
		return false;

	return true;
}

static void updatePath( HWND hWnd, cWCS *pOut = NULL )
{
	cWCStr wcs;
	wcs.GetWindowTextT( hWnd );
	if ( wcs.GetLength() == 0 )
		SetWindowTextT( hWnd, _LoadIDS( IDS_szEmpty ) );

	if ( pOut != NULL )
	{
		if ( wcs.equ( _LoadIDS( IDS_szEmpty ) ) )
			wcs.Set( L"" );

		pOut->Swap( &wcs );
	}
}

bool allowMountAsRemovable()
{
	cArgs args( GetCommandLineT( true ) );
	return args.ReadAndRemoveArgument( L"-navigant" ) && isAdministrator(); 
}

//
//	Keyborad filter problem!
//	If due to any problem WM_KILLFOCUS was not delivered to kbdEdit so we get problem with app what really get
//	focus it get random symbols instead. And later when we get focus again at kbdEdit we've got error here.
//
static 
DWORD WINAPI postCheckFocusMessage(
    LPVOID lpThreadParameter
    )
{

	HWND *pWnd = ( HWND* )lpThreadParameter;
	for( ;; )
	{
		Sleep( 100 );

		if ( *pWnd == NULL )
		{
			ODS_fls( L"Stop!" );	
			break;
		}

		ODS_fls( _FormatT( L"Post message to hWnd %X", *pWnd ) );
		PostMessage( *pWnd, WM_USER_CHECK_FOCUS, 0, 0 );
	}

	return 0;
}

bool cDlgGetPasswordAndMount::OnInitDialog()
{
	ODSFunc;

	if ( m_caption  != NULL ) SetWindowTextT( m_hDlg, m_caption );
	if ( m_fileName != NULL ) SetWindowTextT( GetDlgItem( m_hDlg, IDC_TextName), m_fileName );
	//if ( m_pathName != NULL ) SetWindowTextT( GetDlgItem( m_hDlg, IDC_TextPath), m_pathName );
	Button_SetCheck( GetItem( IDC_chkMountForCurrentUser ), m_bMountForCurrentUser ? BST_CHECKED : BST_UNCHECKED );
	Button_SetCheck( GetItem( IDC_chkMountAsRemovable ), m_bMountAsRemovable ? BST_CHECKED : BST_UNCHECKED );
	ShowWindow( GetItem( IDC_chkMountAsRemovable ), allowMountAsRemovable() ? SW_SHOW : SW_HIDE );
	
	bool bUseKbb = ( m_keyGenFlags & FLAG_KBB_IS_USED ) != 0;
	Button_SetCheck( GetItem( IDC_chkMountWithKBB ), bUseKbb ? BST_CHECKED : BST_UNCHECKED );
	SetWindowTextT( GetItem( IDC_edtKbb ), m_kbbPath.m_szBody );
	updatePath( GetItem( IDC_edtKbb ) );

	updateKBBItems();


	//if ( !m_useSecureTyping ) ShowWindow( GetDlgItem(m_hDlg, IDC_LINE), SW_HIDE );
	ShowWindow( GetDlgItem(m_hDlg, IDC_LINE), m_useSecureTyping? SW_SHOW : SW_HIDE ); // && !bDelayFiltering 

	// Shows password chars explicitly
	if ((m_keyGenFlags & FLAG_ADVANCED_SHOW_PWD_CHARS) > 0)
		Edit_SetPasswordChar(GetDlgItem(m_hDlg, IDC_PASS), 0);

	// Sets the default push button to "Cancel"
	SendMessage(m_hDlg, DM_SETDEFID, (WPARAM) IDCANCEL, (LPARAM) 0); 

	kbd.GenerateEncodingTable( m_useSecureTyping );

	m_edt.Attach( GetDlgItem( m_hDlg, IDC_PASS ), GetDlgItem(m_hDlg, IDC_LINE ) );

	//
	//	If the dialogue is used as mount dialog
	//
	if ( m_mountPoint != NULL )
	{
		if ( !SetMountParams() )
			return false;
	}

	m_keyFileList.updateWindowText( GetItem( IDC_btnKeyfiles ) );

	UpdateAdvancedState( true );

	return true;
}

bool cDlgGetPasswordAndMount::OnOK()
{
	ODSFunc;

	//
	//	Gets mount point and other mount attributes for mount dlg part.
	//
	if ( m_mountPoint != NULL )
	{
		if ( !GetMountParams() )
			return false;

		if ( !CheckValidMountPoint() )
		{
			SetFocus( GetDlgItem( m_hDlg, IDC_DRIVES ) );
			return false;
		}
	}

	m_edt.m_hEditPost = NULL;

	bool bUseKbb = Button_GetCheck( GetItem( IDC_chkMountWithKBB ) ) == BST_CHECKED;

	if ( bUseKbb )	m_keyGenFlags |= FLAG_KBB_IS_USED;
	else			m_keyGenFlags &= ~FLAG_KBB_IS_USED;		

	updatePath( GetItem( IDC_edtKbb ), &m_kbbPath );
	if ( bUseKbb && !FileExistsT( m_kbbPath.m_szBody ) )
	{
		DlgMessageBoxT(
			cWCSF().FormatT( _LoadIDS( IDS_KBBFileDoesNotExist_S ), m_kbbPath.m_szBody ),
			MB_ICONSTOP,
			_LoadIDS( IDS_ERROR )
		);
		SetFocus( GetItem( IDC_edtKbb ) );
		return false;
	}

	m_bMountForCurrentUser = Button_GetCheck( GetItem( IDC_chkMountForCurrentUser ) ) == BST_CHECKED;
	m_bMountAsRemovable = Button_GetCheck( GetItem( IDC_chkMountAsRemovable ) ) == BST_CHECKED;

	m_edt.CheckAndGetPassword( &m_pwd, &m_edt, true );

	if ( m_pwd.IsEmpty() )
	{
		MessageBoxT( m_edt.dlg->m_hDlg,
			_LoadIDS( IDS_PASSWORD_EMPTY ),
			_LoadIDS( IDS_INVALID_PWD ),
			MB_OK | MB_ICONWARNING );
		SetFocus(  m_edt.m_hEdit );
		return false;
	}

	EndDialog_ODS(m_hDlg, ERROR_NO);
	return TRUE; 
}

bool cDlgGetPasswordAndMount::OnBrowseKBB()
{
	cWCS wcs( 4 * 1024 );

	wchar_t szFilters[ 1024 ];
	LoadStringT( m_hInst, IDS_szFiltersKbb, szFilters, maxLen( szFilters ) );

	if ( !BrowseFiles( m_hDlg, wcs.m_szBody, wcs.m_allocatedLen, szFilters, _LoadIDS( IDS_szChooseKBBFile ), false ) )
		return false;

	SetWindowTextT( GetItem( IDC_edtKbb ), wcs.m_szBody );
	return true;
}

void cDlgGetPasswordAndMount::updateKBBItems()
{
	HWND hChk = GetItem( IDC_chkMountWithKBB );

	if ( hChk == NULL )
		return;

	int T[] = {
		IDC_btnBrowseKBB,
		IDC_stKBB,
		IDC_edtKbb
	};

	bool bEnable = ( Button_GetCheck( hChk ) == BST_CHECKED );

	for( int i = maxLen(T); i--; )
		EnableWindow( GetItem( T[i] ), bEnable );
}	

LRESULT cDlgGetPasswordAndMount::OnCommand( WPARAM wParam, LPARAM lParam )
{
	switch( LOWORD( wParam ) ) 
	{
		case IDC_chkMountWithKBB:
		{
			updateKBBItems();
			break;
		}
		case IDC_PASS:
		{
			// Set the default push button to "OK" when the user enters text
			if( HIWORD (wParam) == EN_CHANGE )
				SendMessage(m_hDlg, DM_SETDEFID, (WPARAM) IDOK, (LPARAM) 0); 

			break;
		}
		case IDC_DRIVES:
		{
			if ( HIWORD(wParam) == CBN_SELCHANGE )
				OnBrowseForMountPoint( (HWND)lParam );

			break;
		}
		case IDC_btnBrowseKBB:
		{
			OnBrowseKBB();
			break;
		}
		case IDC_slAdvancedSettings:
		{
			m_bAdvanced = !m_bAdvanced;
			UpdateAdvancedState( false );
			m_keyFileList.updateWindowText( GetItem( IDC_btnKeyfiles ) );
			break;
		}
		case IDC_btnKeyfiles:
		{
			cDlgKeyFileMan dlg( m_keyFileList );
			if ( dlg.DoModal( m_hInst, IDD_KeyFilesManager, m_hDlg ) != IDOK )
				return false;

			m_keyFileList.updateWindowText( GetItem( IDC_btnKeyfiles ) );
			return true;
		}
		case IDOK:
			return OnOK();

		case IDCANCEL: 
		{
			GetMountParams();

			//if ( Button_GetCheck( GetDlgItem(m_hDlg, IDC_AUTOMOUNT)))
			//		{ m_keyGenFlags |= FLAG_CONTAINER_AUTO_MOUNT; }
			//else	{ m_keyGenFlags &= ~FLAG_CONTAINER_AUTO_MOUNT; }

			//if ( Button_GetCheck( GetDlgItem(m_hDlg, IDC_READONLY)))
			//	m_keyGenFlags |= FLAG_CONTAINER_READ_ONLY;
			//else m_keyGenFlags &= ~FLAG_CONTAINER_READ_ONLY;

			EndDialog_ODS(m_hDlg, (int)ERROR_CANCELLED_BY_USER); 
			return TRUE; 
		}
	} 
	return 0; 
}


//
//	cDlgGetPasswordAndMount::RealDlgProc
//		has been generated from getPasswordDlgProc and adopted for getPasswordDlgProcSimple ( without m_mountPoint )
//
LRESULT cDlgGetPasswordAndMount::RealDlgProc( UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	//processContIcon( m_hDlg, uMsg, wParam, lParam );

	cDropTarget::DlgProc( m_hDlg, uMsg );
	cDlg::Redirect_ClickToCmd( m_hDlg, uMsg, IDC_slAdvancedSettings, lParam );

	switch (uMsg) 
	{ 
		case WM_INITDIALOG:
			return OnInitDialog();

		case WM_DESTROY:
		{
			m_edt.Detach();
			kbd.FreeTables( m_useSecureTyping );
			break;
		}
		case WM_COMMAND: 
			return OnCommand( wParam, lParam );

		default:
			break;
	} 

    return FALSE; 

	UNREFERENCED_PARAMETER(lParam);
}