#pragma once

#include "..\..\..\Includes\cDlg.h"
#include "..\..\..\Includes\cDropTarget.h"
#include "cKeyBlockParams.h"
#include "..\..\COMM_DEF\kbd.h"
#include "kbdEdit.h"
#include <assert.h>

class cDlgKBParams;

class cPageKBParams
{
	virtual bool InitV()		{ assert( false ); return false; }

public:
	cDlgKBParams *m_dlg;

	DWORD *m_MenuHelpIDs;
	const wchar_t *m_helpPath;

	HWND m_hDlg;
	bool m_useSecureTyping;

	cPageKBParams()
	{
		m_MenuHelpIDs = NULL;
		m_dlg = NULL;
		m_hDlg = NULL;
		m_useSecureTyping = false;
		m_helpPath = NULL;
	}

	bool Init( cDlgKBParams *dlg );

	virtual void ShowItems( bool show ) { assert( false ); }
	virtual void Destroy()	{ assert( false ); }
	virtual bool OnCommand( WPARAM wParam )	{ assert( false ); return false; }

	void UpdateRange( int id, int min, int pos, bool init );
};

// PBE function are implemented at cDlgKBParamsPBE

class cPagePBE : public cPageKBParams
{
	bool InitV();
	void initEditControls();
	void initKeyStrechingOptions();

	void writeSalt();
	bool readSaltAndIterations();
	bool getIterations( bool bSilent = true );
	void setIterations();
	bool isV8Compatible();
	void EnableWindowIteration();
	void OnCheckV8Compatibility( bool bInit = false );
	void OnCheckEncryptHeader( bool bInit = false );
	bool OnOK();
	bool checkUpdateIterations( bool bSilent = true );

	//
	//	Estimate time of password generation.
	//	1. The Start_TimeEstimationThread() is called at InitV and after hash algorithm changing.
	//	2. setEstimationTime is called after estimation tread is finished and after iteration count is changed,
	//		see EN_CHANGE with IDC_edtPBEIterations at cPagePBE::OnCommand
	//
	HANDLE m_hTread;

	void Start_TimeEstimationThread();
	void Wait_TimeEstimationThread();

	DWORDLONG m_timeFor256Iterations;
	DWORDLONG m_timeFor16KIterations;
	DWORD estimateTime();

	static DWORD WINAPI estimateTimeStatic( LPVOID param )
	{
		ODSFunc;
		cPagePBE *pThis = (cPagePBE * )param;
		if ( pThis == NULL )
			return (DWORD)-1;
		
		return pThis->estimateTime();
	}

	time_t setEstimationTime();

public:
	cKbdEdit m_edtA;
	cKbdEdit m_edtB;
	
	cPagePBE()
	{
		m_timeFor256Iterations = 0;
		m_timeFor16KIterations = 0;
		m_hTread = NULL;
	}

	~cPagePBE()
	{
		Wait_TimeEstimationThread();
	}

	void ShowItems( bool show );
	void Destroy();
	bool OnCommand( WPARAM wParam );
};

// PKE function are implemented at cDlgKBParamsPKE

class cPagePKE : public cPageKBParams
{
	cWCStr m_searchedUserID;
	bool InitV();
public:
	void ShowItems( bool show );
	void Destroy();
	bool OnCommand( WPARAM wParam );

	bool OnBrowse();
	bool OnSearch();
	bool UpdateButtons();
};

// SSS function are implemented at cDlgKBParamsSSS
class cPageSSS : public cPageKBParams
{
	cKbdEdit m_edtA;
	cKbdEdit m_edtB;

	bool InitV();
public:

	HWND m_hWndSave;

	void ShowItems( bool show );
	void Destroy();
	bool OnCommand( WPARAM wParam );
	void UpdateRanges( bool init = false );
//	void UpdateRange( int id, int min, int pos, bool init );
	int  Notify( NMHDR *phdr );
	bool AddShare();
	bool ResetAll();
	bool ViewShares();
};

class cAdvancedState
{
	cWCStr m_wcs[2];

public:
	// Use Redirect_ClickToCmd at DlgProc if control is not a button to process it as button click.
	HWND m_hSysLink_Advanced;
	bool &m_bAdvanced;

	HWND m_hDlg;
	cAdvancedState( bool &bAdvanced, HWND hDlg )
		: m_bAdvanced( bAdvanced )
	{
		m_hDlg = hDlg;
	}

	virtual bool Init();
	virtual bool Update( bool bForceAdvanced = true  );
};

class cDlgGetPassword :
	public cColoredDlg, // cDlg
	public cDropTarget
{
public:

	const wchar_t *m_caption;
	const wchar_t *m_fileName;
	const wchar_t *m_pathName;

	// Key board params 
	kbdParams kbd;

	cKeyFileList m_keyFileList;

	cDlgGetPassword(
			bool bColored = false,
			const wchar_t *caption = NULL,
			const wchar_t *fileName = NULL,
			const wchar_t *pathName = NULL
		)
	{
		m_changeBkColor = bColored;
		m_caption	= caption;
		m_fileName	= fileName;
		m_pathName	= pathName;
		m_useSecureTyping = false;
		m_keyGenFlags = 0;
		m_advancedState = NULL;
		m_bAdvanced = false;
	}
	
	~cDlgGetPassword()
	{
		if ( m_advancedState != NULL )
			delete m_advancedState;
	}
	
	DWORD m_keyGenFlags;
	bool m_useSecureTyping;

	bool m_bAdvanced;
	cAdvancedState *m_advancedState;
	void UpdateAdvancedState( bool bInit );
	bool ShowKeyFilesButton();
};


class cDlgGetAdvancedFlags : public cDlgGetPassword
{
public:

	cKbdEdit m_edt;

	cDlgGetAdvancedFlags( bool bColored ) : cDlgGetPassword( bColored )
	{
	}

	LRESULT RealDlgProc( UINT uMsg, WPARAM wParam, LPARAM lParam );
};

class cDlgGetPasswordAndMount : public cDlgGetPassword
{
public:

	cKbdEdit m_edt;
	cPassword m_pwd;

	// If the dialogue is used for mounting
	// the m_mountPoint is not null
	wchar_t *m_mountPoint;
	cWCS m_kbbPath;
	bool m_bMountForCurrentUser;
	bool m_bMountAsRemovable;

	int m_posBrowseForMountPoint;
	bool m_bTravellerMode;

	cDlgGetPasswordAndMount( 
			const wchar_t *caption = NULL,
			const wchar_t *fileName = NULL,
			const wchar_t *pathName = NULL
			) : cDlgGetPassword( false, caption, fileName, pathName )
	{
		cDropTarget::m_bSupportDrop = true;
		m_bTravellerMode = false;
		m_mountPoint = NULL;
		m_posBrowseForMountPoint = 0;
		m_bMountForCurrentUser = false;
		m_bMountAsRemovable = false;
	}

	virtual HRESULT Drop( cFilesList &ofl );// { return S_FALSE; }

	bool OnInitDialog();
	bool OnOK();
	bool OnBrowseKBB();

	LRESULT OnCommand( WPARAM wParam, LPARAM lParam );

	LRESULT RealDlgProc( UINT uMsg, WPARAM wParam, LPARAM lParam );

	bool SetMountParams();
	bool GetMountParams();
	int SelectMountPoint( HWND hCmb );
	bool OnBrowseForMountPoint( HWND hCtrl );
	bool CheckValidMountPoint();
	bool checkEmptyFolder();
	bool updateFlags( int idc, DWORD mask );
	void updateKBBItems();
};

class cDlgKBParams : public cDlgGetPassword
{
public:

	cKeyBlockParams *m_kbParams;

	cDlgKBParams( 
			cKeyBlockParams *kbParams,
			const wchar_t *caption = NULL,
			const wchar_t *fileName = NULL,
			const wchar_t *pathName = NULL
		) : cDlgGetPassword( false, caption, fileName, pathName )
	{
		m_kbParams = kbParams;
	}

	virtual LRESULT RealDlgProc( UINT uMsg, WPARAM wParam, LPARAM lParam );
	bool OnInitDialog();
	int OnNotify(  NMHDR *phdr );
	void ShowItems();
	BOOL OnCommand( WPARAM wParam, LPARAM lParam );

	virtual HRESULT Drop( cFilesList &ofl );

	cPagePBE PBE;
	cPagePKE PKE;
	cPageSSS SSS;

	cPageKBParams *CurPage();

	virtual void OnHelpInfo( HELPINFO *pHelpInfo = NULL )
	{
		if ( CurPage() == NULL )
			return;

		if ( CurPage()->m_helpPath != NULL )
			::HtmlHelpT( m_hDlg, _FormatT( L"%s::%s", GetHelpFileName(), CurPage()->m_helpPath ), HH_DISPLAY_TOPIC, NULL );

		//if ( CurPage()->m_MenuHelpIDs != NULL )
		//	OnHelpInfo( pHelpInfo, CurPage()->m_MenuHelpIDs, false );
	}
};

#define WM_userUpdateRanges ( WM_USER + 10 )
extern bool allowMountAsRemovable();