/*******************************************************************\
 *
 *	Copyright (c) 2015 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	File:        cDlgKeyFileMan.cpp
 *	Revision:  
 *	Created:     02-March-2015
 *  Author:      A. Pichuev
 *	Description: Dialog shows current list if key files and allows to modify it. 
 *
\*******************************************************************/

#include "cDlgKeyFileMan.h"
#include "..\..\..\Includes\OpenFiles.h"
#include "..\..\..\Includes\cFindFile.h"

static int ColumnLabelIDS[]	= { IDS_szPath,		IDS_szSize,		IDS_szFiles		};
static int _gnColumnWidth[]	= { 200,			50,				50				};
static int _gnColumnFmt[]	= { LVCFMT_LEFT,	LVCFMT_RIGHT,	LVCFMT_RIGHT	};

#ifdef _allowFolder
static const int NUM_COLUMNS = maxLen( ColumnLabelIDS );
#else
static const int NUM_COLUMNS = maxLen( ColumnLabelIDS ) - 1;
#endif

cDlgKeyFileMan::cDlgKeyFileMan( cKeyFileList &keyFileList )
	: m_keyFileList( keyFileList )
{
	m_iconCache = NULL;
	m_iDropTarget = NULL;
	m_filesQ = 0;
	m_ptbVals.Init( HKEY_CURRENT_USER, L"SOFTWARE\\Jetico\\BestCrypt", L"dlgKeyFileMan", sizeof( m_ptbVals ), NUM_COLUMNS );
	m_bSupportDrop = true;
}

static const wchar_t *szDirectory = L"*";

bool cDlgKeyFileMan::saveListView()
{
	m_keyFileList.Clear();

	for( int i = ListView_GetNextItem( m_hList, -1, LVNI_ALL ); i >= 0; i = ListView_GetNextItem( m_hList, i, LVNI_ALL ) )
	{
		cWCSF wcs( 1024 );
		int size = ListView_GetItemTextT( m_hList, i, 0, wcs.m_szBody, wcs.m_allocatedLen );
		if ( size == wcs.m_allocatedLen )
		{ assert( false ); return false; }


		if ( m_keyFileList.m_sa.addString( wcs.m_szBody ) == 0 )
		{ assert( false ); return false; }
	}

	return true;
}

bool cDlgKeyFileMan::addPath( const wchar_t *path )
{
	DWORDLONG dwlSize;

	int Q = cKeyFileList::getFileQ( path, &dwlSize );

	WIN32_FIND_DATAW fd;
	if ( ( Q <= 0 ) || !FileExistsT( path, &fd ) )
	{ assert( false );	return false; }

	LVITEMW lvi;
	memset( &lvi, 0, sizeof( lvi ) );

	lvi.mask = LVIF_TEXT | LVIF_PARAM | LVIF_IMAGE;
	lvi.iImage = m_iconCache->GetIconNum( path, NULL );
	lvi.iItem = 0x7fffFFFF;//-1;
	lvi.lParam = Q; // To be m_filesQ decreased at OnRemoveSelected
	lvi.pszText = (wchar_t *)path;

	int i = ListView_InsertItemT( m_hList, &lvi );
	if ( i == -1 )
	{ assert( false );	return false; }

	ListView_SetItemTextT( m_hList, i, 1, cWCStr().GetSizeText( dwlSize ) );
	ListView_SetItemTextT( m_hList, i, 2, _FormatT( L"%d", Q ) );
	
	m_filesQ += Q;
	SetWindowTextT( GetItem( IDC_totalFiles_D ), _FormatT( m_pattern.m_szBody, m_filesQ ) );

	return true;
}

void cDlgKeyFileMan::fillListView()
{
	m_filesQ = 0;

	for( int i = 0; i < m_keyFileList.m_sa.Q; i++ )
		addPath( m_keyFileList.m_sa.T[i] );

	SetWindowTextT( GetItem( IDC_totalFiles_D ), _FormatT( m_pattern.m_szBody, m_filesQ ) );
}

static bool initListViewColumns( HINSTANCE hInst, HWND hLView )
{
	int i;
	LVCOLUMNW lvc;

	memset( &lvc, 0, sizeof( lvc ) );

	lvc.mask = LVCF_WIDTH;

	//
	//	Check if already initialised...
	//
	if ( ListView_GetColumnT( hLView, 0, &lvc ) )
		return true;

	lvc.mask = LVCF_FMT | LVCF_WIDTH | LVCF_TEXT | LVCF_SUBITEM;

	wchar_t text[ 80 ];

	int res;

	for(i=0; i<NUM_COLUMNS; i++)
	{
		LoadStringT( hInst, ColumnLabelIDS[i], text, maxLen( text ) );

		lvc.iSubItem=i;
		lvc.pszText = text; // _gszColumnLabel[i];
		lvc.cx = _gnColumnWidth[i];
		lvc.fmt=_gnColumnFmt[i];

		res = ListView_InsertColumnT( hLView, i, &lvc );

		if ( res == -1 )
			return false;
	}

	ListView_SetExtendedListViewStyleEx( hLView, LVS_EX_FULLROWSELECT, LVS_EX_FULLROWSELECT );

	return true;
}


//virtual 
bool cDlgKeyFileMan::InitDialog()
{
	cWCStr wcs;
	wcs.GetWindowTextT( m_hDlg );
	SetWindowTextT( m_hDlg, _FormatT( wcs.m_szBody, m_keyFileList.getFileQ() ) );

	m_pattern.GetWindowTextT( GetItem( IDC_totalFiles_D ) );

	m_sizedCtrls.AddCtrl( IDOK, cSizedCtrls::ptPDR );
	m_sizedCtrls.AddCtrl( IDCANCEL, cSizedCtrls::ptPDR );
	m_sizedCtrls.AddCtrl( IDC_btnRemove, cSizedCtrls::ptPDR );
	m_sizedCtrls.AddCtrl( IDC_btnClearList, cSizedCtrls::ptPDR );

	m_sizedCtrls.AddCtrl( IDC_btnAddFiles, cSizedCtrls::ptPR );
	m_sizedCtrls.AddCtrl( IDC_btnAddFolders, cSizedCtrls::ptPR );

	m_sizedCtrls.AddCtrl( IDC_totalFiles_D, cSizedCtrls::ptPD );

	m_sizedCtrls.AddCtrl( IDC_List, cSizedCtrls::ptSDR );
	m_sizedCtrls.AddCtrl( IDC_grp, cSizedCtrls::ptSDR );

	m_hList = GetItem( IDC_List );

	initListViewColumns( m_hInst, m_hList );

	m_iconCache = new cIconCache( m_hInst, m_hList ); // IDI_Folder, IDI_Unknown );

	m_ptbVals.Restore( m_hDlg, m_hList );

	fillListView();

	initDropTarget();

	return true;
}

bool cDlgKeyFileMan::initDropTarget()
{
	int res = cDropTarget::Register( m_hDlg );

	switch( res )
	{
		case S_OK: break;
		case E_OUTOFMEMORY: 
			ODS_fls( L"Error In Allocating Memory" );
			return false;
		default:
			ODS_fls( L"Error In Registerring DragDrop" );
			return false;
	}

	return true;
}

//virtual
bool cDlgKeyFileMan::OnDestroy()
{
	if ( m_iconCache != NULL )
	{
		delete m_iconCache;
		m_iconCache = NULL;
	}

	if ( m_iDropTarget != NULL )
	{
		cDropTarget::Revoke();
		delete m_iDropTarget;
		m_iDropTarget = NULL;
	}

	return false;
}


//virtual
bool cDlgKeyFileMan::OnCommand( WORD ctrlId, HWND hCtrl, WORD isAccelerator )
{
	switch( ctrlId )
	{
		case IDC_btnAddFiles:		return OnAddFiles();
		case IDC_btnAddFolders:		return OnAddFolders();
		case IDC_btnRemove:			return OnRemoveSelected();
		case IDC_btnClearList:		return OnRemoveAll();

		case IDOK:
			if ( !saveListView() )
			{
				DlgMessageBoxT( L"Error in saving key file list!", MB_ICONSTOP );
				break;
			}

		case IDCANCEL:
			m_ptbVals.Save( m_hDlg, GetItem( IDC_List ) );
			EndDialog( m_hDlg, ctrlId );
			return true;
		default:
			break;
	}

	return false;
}

bool cDlgKeyFileMan::OnAddFiles()
{
	cWCS wcs( 4 * 1024 );
	if ( !BrowseFiles( m_hDlg, wcs.m_szBody, wcs.m_allocatedLen, NULL, _LoadIDS( IDS_szChooseKeyFiles ), true ) )
		return false;

	cOpenFilesList bfl( wcs.m_szBody, wcs.m_allocatedLen );

	while( bfl.getNext() != NULL )
	{
		if ( !addPath( bfl.getCur() ) )
		{ assert( false ); return false; }
	}
	
	return true;
}

bool cDlgKeyFileMan::OnAddFolders()
{
	cWCS wcs( MAX_PATH );
	if ( !BrowseForFolderT( _LoadIDS( IDS_ChooseKeyFileFolder ), m_hDlg, wcs.m_szBody ) )
		return false;
#ifdef _allowFolder	
	return addPath( wcs.m_szBody );
#else
	cFindFile ff( wcs.m_szBody );

	if ( !ff.findFirstFile() )
	{
		DlgMessageBoxT( L"Error in reading directory!", MB_ICONSTOP );
		return false;
	}

	int count = 0;

	do
	{
		if ( ff.isDir() )
			continue;

		if ( addPath( ff.getCurrentPath() ) )
			count++;
	}
	while( ff.findNextFile() );

	if ( count > 0 )
		return true;

	DlgMessageBoxT( _FormatT( L"Folder '%s'\nhas no files!", wcs.m_szBody ) , MB_ICONEXCLAMATION );

	return false;
#endif
}

bool cDlgKeyFileMan::OnRemoveSelected()
{
	while( true )
	{
		int i = ListView_GetNextItem( m_hList, -1, LVNI_SELECTED );
		if ( i == -1 )
			break;

		m_filesQ -= ListView_GetItemParam( m_hList, i );

		ListView_DeleteItem( m_hList, i );
	}

	SetWindowTextT( GetItem( IDC_totalFiles_D ), _FormatT( m_pattern.m_szBody, m_filesQ ) );
	checkButtonStates();
	return false;
}

bool cDlgKeyFileMan::OnRemoveAll()
{
	ListView_DeleteAllItems( m_hList );
	m_filesQ = 0;
	SetWindowTextT( GetItem( IDC_totalFiles_D ), _FormatT( m_pattern.m_szBody, m_filesQ ) );
	checkButtonStates();
	return true;
}

void cDlgKeyFileMan::checkButtonStates()
{
	bool bHasSelected = ( ListView_GetNextItem( m_hList, -1, LVNI_SELECTED ) != -1 );
	EnableWindow( GetItem( IDC_btnRemove ), bHasSelected );

	bool bHasItems = ListView_GetNextItem( m_hList, -1, LVNI_ALL ) != -1;
	EnableWindow( GetItem( IDC_btnClearList ), bHasItems );
}

//virtual 
LONG_PTR cDlgKeyFileMan::OnNotify( NMHDR *pnmh )
{
	if ( ( pnmh->code == LVN_ITEMCHANGED ) ||
		 ( pnmh->code == LVN_INSERTITEM ) || 
		 ( pnmh->code == LVN_DELETEITEM ) )
	{
		checkButtonStates();
	}

	return 0;
}

//virtual
HRESULT cDlgKeyFileMan::Drop( cFilesList &ofl )
{
	while( ofl.getNext() )
		addPath( ofl.getCur() );

	return S_FALSE;
}
