/*******************************************************************\
*
*	Copyright (c) 2006-2015 Jetico Inc., Finland
*	All rights reserved.
*
*	File: dbHeadeBase.h
*	Revision: 
*	Created:
*	Description: 
*		Base class for dbHeader, without functions.
*
\*******************************************************************/

#pragma once
#include "spaceUnitBase.h"

#pragma pack(1)

#define _4K ( 4 * 1024 )
#define DB_MinSize				   _4K


//
//	DB_MaxSize is completely calkulated at dbHeader size.
//
// Rest from all requred fields is busied by key map table 
// The size == 
//			max number of key map table items * DB_MinEncryptedKeyBlockSize 
//			DB_HeaderSizeWithKeyMap +
//			DB_PoolSize
//	= 5 * 4K = 20K
//#define DB_MaxSize	DRound( DB_maxKeyQ * DB_MinEncryptedKeyBlockSize + DB_HeaderSizeWithKeyMap + DB_PoolSize, _4K )
#define DB_MaxSize ( 20 * 1024 )

//#define DB_BusiedNameSize ( sizeof( BPB ) + 7 -  CONTAINER_ID_LENGTH ) // sizeof( BPB ) + 7 is space between signature and volume label in real BootRecord
//							   25         + 7  - 4 == 28
#define DB_BusiedNameSize 28

#define DB_KeyMapSize				512
#define DB_HeaderSizeWithKeyMap		1024
#define DB_PoolSize					512 

#define DB_maxKeyQ ( DB_KeyMapSize / sizeof( dbKeyId ) ) // 512 / 8 == 64	-> 64 minimal keys may be added to container ( maximal data block size is 5 * 4K ).
#define DB_minKeyQ (( DB_MinSize -  DB_HeaderSizeWithKeyMap - DB_PoolSize ) / DB_MinEncryptedKeyBlockSize ) // ( 4k - 3 * 512 ) / 256 == 10 -> 10 keys may be stored at minimal key block

#ifndef maxLen
#define maxLen( b ) ( sizeof( b ) / sizeof( *b ) )
#endif

#define CONTAINER_ID_LENGTH			4

#define DB_MinEncryptedKeyBlockSize	256 // Minimal step for key block ckecking

//
// Key block type enum
//
enum ekbType {
		kbtTail = -1,	// the block is busied, head is placed in a previous blocks 
		kbtEmpty = 0,
		kbtPBE = 1,		// Password Based Encrypted key block 
		kbtPKE = 2,		// Public Key Enccypted key block
		kbtSSS = 3,		// key block encrypted with Shamir Secret Shared scheme
		kbtHP = 4,		// Busied by hidden part, it used for temporary key maps only.
		kbtSalt = 5		// KeyGen V5 save salt at key map, see DATA_BLOCK::updateSalt
};

struct dbKeyId
{
	WORD	kbSize;	// size in bytes really busied by the key block. It may be one or more minimal blocks.
	WORD	kbType;	// see enum ekbType in kg_defs.h
	DWORD	keyParam;
		// kbtPBE - does not use the parameter
		// kbtPKE - saves here first 4 bytes of key id
		// kbtSSS - saves here number of required shares

	void Init()
	{
		memset( this, 0, sizeof( *this ) );
		kbType = kbtEmpty;
	}

	static int GetRequiredQ( WORD size )
	{
		return ( size + DB_MinEncryptedKeyBlockSize - 1 ) / DB_MinEncryptedKeyBlockSize;
	}
	
	dbKeyId()	{ Init(); }
	~dbKeyId()	{ Init(); }
};

//struct dbKeyId
//{
//	WORD	kbSize;	// size in bytes really busied by the key block. It may be one or more minimal blocks.
//	WORD	kbType;	// see enum ekbType in kg_defs.h
//	DWORD	keyParam;
//
//	dbKeyId()
//	{
//		memset( this, 0, sizeof( *this ) );
//	}
//
//	static int GetRequiredQ( WORD size )
//	{
//		return ( size + DB_MinEncryptedKeyBlockSize - 1 ) / DB_MinEncryptedKeyBlockSize;
//	}
//};

//
//	The structure is filled when key handle is been created
//
struct sKeyMapInfo
{
	sKeyMapInfo()	{ setUndefined(); }

	int position;	// Start position number in key map table of dbHeader
	int quantity;	// Quantity of minimal size blocks busized by the key
	WORD kbType;	// Real key block type used for encoding the key, same as dbKeyId::kbType
	WORD kbSize;	// encoded key size, same as dbKeyId::kbSize

	void setUndefined() { memset( this, 0, sizeof( *this ) ); position = -1; }
	bool setInfo( int pos, int q, WORD type, WORD size = 0 )
	{
		position = pos;
		kbType = type;
		kbSize = size;

		if ( size != 0 )
			q = dbKeyId::GetRequiredQ( size );

		quantity = q;

		return true;
	}

	bool setInfo( sKeyMapInfo *kmi )
	{
		return setInfo( kmi->position, kmi->quantity, kmi->kbType, kmi->kbSize );
	}

	~sKeyMapInfo()	{ setUndefined(); }
};

////
////	The structure is filled when key handle is been created
////
//struct sKeyMapInfo
//{
//	sKeyMapInfo()	{ setUndefined(); }
//
//	int position;	// Start position number in key map table of dbHeader
//	int quantity;	// Quantity of minimal size blocks busized by the key
//	WORD kbType;	// Real key block type used for encoding the key, same as dbKeyId::kbType
//	WORD kbSize;	// encoded key size, same as dbKeyId::kbSize
//
//	void setUndefined() { memset( this, 0, sizeof( *this ) ); position = -1; }
//
//	static const wchar_t *getTypeStr( ekbType type );
//	const wchar_t *getTypeStr() { return getTypeStr( (ekbType)kbType ); }
//
//	~sKeyMapInfo()	{ setUndefined(); }
//};
//


struct dbHeaderBase
{
	//
	//	Body.
	//

	enum { descriptionLength = 21, KEYGEN_ID = 4, KEYGEN_ID_V5 = 5 }; // sparsedPosition is added.

	BYTE	busiedFlag[ 3 ];	// BootRecord's jmpCode[3]
	char	signature[8];		// BootRecord's OEMid 'LOCOS94 ', required for guard utility compatibility.
	BYTE	containerID[ CONTAINER_ID_LENGTH ]; // See DataBlockGetContainerId
	char	busiedName[ 28 ];	//DB_BusiedNameSize ]; // 28 bytes
	char	volumeLabel[11];	// 'BC_KeyGenID'

	//DWORD	keyGenId;			// Key generator identifier...
	WORD	wKeyGenId;		// Key generator ID 5 identifier...
	WORD	wVersion;		// Key generator ID 5 version...

	union {
		DWORD	iterations;	// Key generator ID 5, allows iteration other then 256, but it does never uses the version field.
							// The version field is used by linux but in case to check does this container uses new suncro.
		DWORD	version;	// Container version ( BestyCrypt does never check this, but KG-Ghost checks this compatibility ).
	};
	wchar_t	description[ descriptionLength ];	// Same as old version descryption

	// !!! It is never used because sparsed files are procesed by ContainerProtection process.
	// Reserved for future.
	ULONGLONG sparsedPosition; // See Sparsed file service project. 

	//dbContInfo	contInfo;
	sSpaceUnitBase	contInfo;

	DWORD	contAlgId;	// The algorithm identifier is uesd for container encryption...
	DWORD	encModeId;	// Encryption mode see bc_ioctl.h EX_ENCRYPT_LRW_MODE_FLAG and EX_ENCRYPT_CBC_MODE_FLAG
	DWORD	hashAlgId;

	dbKeyId	keyMap[ DB_maxKeyQ ];
};

#pragma pack()
