/*******************************************************************
 *
 *	Copyright (c) 1994-2005 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	File: pwdlg.cpp
 *	Revision: 
 *	Created:
 *	Description: implementation of procedures that support 
 *		dialog boxes for accepting user's password and 
 *		dialog that supports generation of seed value for 
 *		random key generation
 *
 *******************************************************************/

#include <windows.h>
#include <windowsx.h>
#include <commctrl.h>
#include <string.h>

#include <stdio.h>
#include <assert.h>

#include "kg_defs.h"
#include "resource.h" 
#include "resrc1.h" 
#include "pwdlg.h"
#include "kbd.h"
#include "kg_misc.h" // For CenterPosition only...
#include "diskDriver.h" // For CmdGetVDiskMask, see description of _GetLogicalDrives()...
#include "..\comm_def\comm_def.h"
#include "dllMain.h"
#include "GenerateSeed.h"

#define MSG_SEED_READY		(WM_USER + 0x529)

typedef BOOL CALLBACK DlgProcType( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam );

/* Forward declarations of local procedures */

BOOL CALLBACK getPasswordVerifyDlgProc( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam );
BOOL CALLBACK getPasswordDlgProc( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam );
BOOL CALLBACK getPasswordDlgProcSimple( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam );
//BOOL CALLBACK getRandomDlgProc( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam );
BOOL CALLBACK getAdvancedFlagsDlgProc( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam );
//BOOL UpdateSeed( BYTE *Seed );
BOOL convertTextToFilenamePath( const wchar_t *Text );
static DWORD _GetLogicalDrives();

/* Global data */

#define BUFFER_SIZE 1024

// The problem with **** ||||| in password dialog is fixed 
// by changing font name in resource file.
const char *szPwdFontName = "MS Shell Dlg";
const bool changeFont = false;

char    *Password, *Buffer, *BufferB;
const wchar_t *Caption;
wchar_t FileName[128], PathName[128], NotFileName[256];
//BYTE    *Seed;
//DWORD   SeedCounter, SeedLength;
HMODULE HModule;

//
// See kbdParams structtre in KBD.H
//
//UINT   *VkOriginal,   *VkMessed;
//USHORT *ScodOriginal, *ScodMessed;
//BYTE   *ShiftTable;
//UINT   TableSizeNL, TableSizeSm, ShiftTableSize;

DWORD  KeyGenFlags;
DWORD  KeyGenFavoriteDrive;

BOOL   UseSecureTyping;

#ifdef STRICT  
//WNDPROC wpOrigWndProc;
WNDPROC wpEditOrigWndProc;
WNDPROC wpEditOrigWndProcA;
WNDPROC wpEditOrigWndProcB;
#else 
//FARPROC wpOrigWndProc;
FARPROC wpEditOrigWndProc;
FARPROC wpEditOrigWndProcA;
FARPROC wpEditOrigWndProcB;
#endif 

//LONG	dlgWndProc;
/* Exported procedures */
static BOOL getPasswordFromCommandLine(	char   *Password, int size,
										DWORD  *Flags,
										DWORD  *FavoriteDrive
									   )
{
	///// Try to get password from a command-line: BestCrypt.exe Open G:\30.jbc X: -p:=password=-p
	char *cmdLine, pwd[128];
	int i, j, k, cmdLineLength;
	BOOL bPwdFound = FALSE;

	memset( pwd, 0, 128 );
	cmdLine = GetCommandLine();
	cmdLineLength = strlen(cmdLine);

	for (i=0; i<cmdLineLength; i++)
	{
		if ((cmdLine[i] == '-') && (cmdLine[i+1] == 'p') && (cmdLine[i+2] == ':') && (cmdLine[i+3] == '='))
		{
			for (j=i+4, k=0; j<cmdLineLength; j++, k++)
			{
				if ((cmdLine[j] == '=') && (cmdLine[j+1] == '-') && (cmdLine[j+2] == 'p')) 
				{
					bPwdFound = TRUE;
					break;
				}

				pwd[k] = cmdLine[j];

				if (bPwdFound) break;
			}
		}
	}

	if (bPwdFound) 	memcpy( Password, pwd, size );
	else return FALSE;

	///// End of "Try to get password from a command-line"

	//// Get Logical drive letter

	DWORD curLogicalDisks = 0, firstAvailableDrive = 0;

	if (((*Flags) & FLAG_MOUNT_ON_SINGLE_LETTER) > 0)
	{
		// leave the drive letter as it was set
		return TRUE;
	}

	curLogicalDisks = _GetLogicalDrives();

	if (((*FavoriteDrive) > 0) && ((*FavoriteDrive) < 27))
	{
		if ((curLogicalDisks & (1 << (*FavoriteDrive))) == 0)
		{
			return TRUE; // FavoriteDrive has a reasonable setting and is not busied by other devices
		}
	}

	for (i=2; i<26; i++ )
	{
		if ((curLogicalDisks & (1 << i)) == 0)
		{
			firstAvailableDrive = i;
			break;
		}
	}

	if (!firstAvailableDrive) return FALSE;

	*FavoriteDrive = firstAvailableDrive;

	return TRUE;
}

DWORD GetPassword( 
				HWND   hWndParent,
				const wchar_t *text,	// Text, containing filename of the container
				const wchar_t *caption,	// Caption of the "get password" dialog window
				char   *password, 
				DWORD  *Flags,
				DWORD  *FavoriteDrive,
				BOOL   ContainsFileName
				)
{
	int result;

	Password = password;
	Caption  = caption;
	FileName[0] = '\0';
	PathName[0] = '\0';

	if (getPasswordFromCommandLine(	password, PASSWORD_MAXIMUM_LENGTH, Flags, FavoriteDrive ))
	{
		return ERROR_NO;
	}
	
	if (ContainsFileName) convertTextToFilenamePath( text );
	else
	{
		wcscpy_s( NotFileName, text );
		FileName[0] = L'\0';
		PathName[0] = L'\0';
	}

	Buffer = new char[ BUFFER_SIZE ];
	if (!Buffer) return ERROR_INTERNAL_PROBLEM;
	Buffer[0] = 0;

	if (Flags) KeyGenFlags  = *Flags;

	if ((KeyGenFlags & FLAG_ADVANCED_SECURE_TYPING) > 0)
			UseSecureTyping = TRUE;
	else	UseSecureTyping = FALSE;

	if (FavoriteDrive) KeyGenFavoriteDrive = *FavoriteDrive;
	else               KeyGenFavoriteDrive = 0xffff;

	HModule = hMainInst; // GetModuleHandle( KGSHA_MODULE_NAME );

	if (KeyGenFavoriteDrive <= 26)
	{
		try
		{
			result = DialogBox( HModule, MAKEINTRESOURCE(IDD_PasswdDlg), 
							hWndParent, (DLGPROC)getPasswordDlgProc );
		}
		catch(...)
		{
			assert( false );
		}
	}
	else
	{
		result = DialogBox( HModule, MAKEINTRESOURCE(IDD_PasswdDlgSimple), 
							hWndParent, (DLGPROC)getPasswordDlgProcSimple );
	}

	ShredData( (BYTE *)Buffer, BUFFER_SIZE * sizeof(char));
	delete Buffer;

	if (result == -1) /* system error */
		return ERROR_INTERNAL_PROBLEM;

	if (Flags) *Flags  = KeyGenFlags;

	if (FavoriteDrive) *FavoriteDrive = KeyGenFavoriteDrive;
	
	return (DWORD)result;
}


//DWORD GetPasswordWithVerification(
//				HWND   hWndParent,
//				LPCSTR mbsText,	// Text, containing filename of the container
//				LPCSTR mbsCaption,	// Caption of the "get password" dialog window
//				char   *password,
//				DWORD  *Flags
//				)
//{
//	//
//	//	Parameters are changed to wide char strings
//	//
//	wchar_t *text = NULL, *caption = NULL;
//
//	cWCStr wc[2];
//
//	if ( mbsText != NULL )
//	{
//		wc[0].Add( mbsText );
//		text = wc[0].m_szBody;
//	}
//
//	if ( mbsCaption != NULL )
//	{
//		wc[1].Add( mbsCaption );
//		caption = wc[1].m_szBody;
//	}
//
//	return GetPasswordWithVerification( hWndParent, text, caption, password, Flags );
//}

DWORD GetPasswordWithVerification(
				HWND   hWndParent,
				const wchar_t *text,	// Text, containing filename of the container
				const wchar_t *caption,	// Caption of the "get password" dialog window
				char   *password,
				DWORD  *Flags
				)
{
	int result;

	Password = password;
	Caption = caption;
	FileName[0] = L'\0';
	PathName[0] = L'\0';
	convertTextToFilenamePath( text );

	HModule = hMainInst; // GetModuleHandle( KGSHA_MODULE_NAME );

	KeyGenFlags  = *Flags;
	if ((KeyGenFlags & FLAG_ADVANCED_SECURE_TYPING) > 0)
			UseSecureTyping = TRUE;
	else	UseSecureTyping = FALSE;
	
	Buffer  = new char[ BUFFER_SIZE ];
	BufferB = new char[ BUFFER_SIZE ];

	if (!Buffer) return ERROR_INTERNAL_PROBLEM;
	if (!BufferB)
	{
		delete Buffer;
		return ERROR_INTERNAL_PROBLEM;
	}
	Buffer[0]  = 0;
	BufferB[0] = 0;

	result = DialogBox( HModule, MAKEINTRESOURCE(IDD_PasswdDlgV), 
						hWndParent, (DLGPROC)getPasswordVerifyDlgProc );

	ShredData( (BYTE *)Buffer,  BUFFER_SIZE * sizeof(char));
	ShredData( (BYTE *)BufferB, BUFFER_SIZE * sizeof(char));
	delete Buffer;
	delete BufferB;

	if (result == -1) /* system error */
		return ERROR_INTERNAL_PROBLEM;

	return (DWORD)result;
}


//DWORD GetSeedValue( 
//			HWND  hWndParent,
//			BYTE  *seed, 
//			DWORD seedLength )
//{
//	int result;
//	
//	if ((!seed) || (seedLength <= 0)) 
//		return ERROR_INTERNAL_PROBLEM;
//
//	Seed = seed;
//	SeedCounter = 0;
//	SeedLength = seedLength;
//
//	HModule = hMainInst; // GetModuleHandle( KGSHA_MODULE_NAME );
//
//	result = DialogBox( HModule, MAKEINTRESOURCE(IDD_RANDOMDLG), 
//		hWndParent, (DLGPROC)getRandomDlgProc );
//
//	if (result == -1) /* system error */
//		return ERROR_INTERNAL_PROBLEM;
//
//	return ERROR_NO;
//}

DWORD GetAdvancedFlags( 
			DWORD *Flags,
			HWND  hWndParent
			)
{
	int result;

	HModule = hMainInst; // GetModuleHandle( KGSHA_MODULE_NAME );

	KeyGenFlags  = *Flags;
	UseSecureTyping = TRUE;
	//UseSecureTyping = FALSE;

	result = DialogBox( HModule, MAKEINTRESOURCE(IDD_SetFlags), 
						hWndParent, (DLGPROC)getAdvancedFlagsDlgProc );

	if (result == -1) /* system error */
		return ERROR_INTERNAL_PROBLEM;

	*Flags = KeyGenFlags;
	
	return (DWORD)result;
}


/* Local procedures */

void insertCharToPwdBuffer( HWND Hwnd, char Letter, char *Buffer, BOOL IsLetter )
{
// 	assert( false );

	WORD  caretPos, pwdLength, i, j;
	/*
	SHORT state;
	state = GetKeyState( VK_CONTROL );
	if (((state >> 8) & 0x80) > 0 )
	{
	   return;
	}
	*/
	// Send the EM_GETSEL message to the edit control.
	// The low-order word of the return value is the character
	// position of the caret relative to the first character in the
	// edit control.

	caretPos = LOWORD((DWORD)SendMessage(Hwnd, EM_GETSEL, 0, 0L));

	// caretPos is numbered from 1, so we have to decrement it
	if ((Letter != 8) && ((Letter != 46) || IsLetter)) caretPos--;

	pwdLength = (WORD)SendMessage(Hwnd, EM_LINELENGTH, (WPARAM)0, (LPARAM) 0); 

	// Check if Letter is the BACKSPACE key
	if (Letter == 8)
	{
		// The original wnd proc was processed so we have to gues what chracters were removed.

		//
		//	The password state before original proc call
		//
		// [fffffssssslllll]
		//	fffff - firsat not selected 
		//	sssss - was selected 
		//	lllll - last not selected
		//
		//	After original proc call
		//
		//	[ffffflllll]
		//
		WORD pwdLengthOld = strlen( Buffer );
		WORD rCnt = pwdLengthOld - pwdLength;	// removed count...
		WORD xPos = caretPos;					// 
		WORD lPos = caretPos + rCnt;
		WORD tailCnt = pwdLengthOld - lPos;

		for( int i = 0; i < tailCnt; i++ )
		{
			Buffer[ xPos + i ] = Buffer[ lPos + i ];
		}

		//for (i=pwdLength-caretPos, j=caretPos; i; i--, j++)
		//{
		//	Buffer[j] = Buffer[j+1];
		//}
	}
	// Check if Letter is the DELETE key
	else if ((Letter == 46) && (!IsLetter)) 
	{
		if (caretPos < pwdLength)
		{
			for (i=pwdLength-caretPos, j=caretPos; i; i--, j++)
			{
				Buffer[j] = Buffer[j+1];
			}
		}
	}
	else
	{
		if (((caretPos + 1) != pwdLength) && (pwdLength > 0))
		{
			for( i=pwdLength-1; i>caretPos; i--)
			Buffer[i] = Buffer[i-1];
		}

		//
		//	19.01.2011
		//
		//	When dead char Letter == 0
		//
		if ( ( caretPos < 0xffff ) && ( Letter != 0 ) )
			Buffer[caretPos] = Letter;
	}

	Buffer[pwdLength] = 0;
}


////////////////////////////////////////

char getRandomLetter()
{
	BYTE       ticks = 0;
	SYSTEMTIME systemTime;

	GetSystemTime( &systemTime );
	ticks = (BYTE)(systemTime.wMilliseconds) & 0x3F;
	return (char)(20 + ticks);
}

////////////////////////////////////////

// Subclass procedure

//#define _Use_editSubclassProcA_

#ifdef _Use_editSubclassProcA_
#define editSubclassProc editSubclassProcA
#else

LRESULT APIENTRY editSubclassProc( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	char    letter;
	LRESULT lpRes;
	BOOL    isLetter;
	UINT    virtualKey, scanCode;
	DWORD   retCharsNum;
	char    retChar1, retChar2;

	isLetter = FALSE;
	wchar_t wcs[10];

	kbdParams *kbd = kbdParams::GetParams( GetParent(hwnd) );
	if ( kbd == NULL ) return 0;

	static int setQ = 0;

	switch (uMsg)
	{
		case WM_INPUTLANGCHANGE:
		{
			ODSF( L"WM_INPUTLANGCHANGE:" );

			if (UseSecureTyping)
			{
				HWND hParent = GetParent( hwnd );

				kbd->noSemaphore = true;

				kbd->ResetEncodingTable( UseSecureTyping, setQ );
				kbd->FreeTables( UseSecureTyping );

				//kbdParams::Free( kbd );
				//kbd = kbdParams::Allocate();

				if ( kbd != NULL )
				{
					kbdParams::SetParams( hParent, kbd );
					// SetWindowLong( hParent, DWL_USER, (LONG)kbd );
					
					kbd->noSemaphore = true;

					kbd->GenerateEncodingTable( UseSecureTyping );
					kbd->SetEncodingTable( UseSecureTyping, setQ );

					kbd->noSemaphore = false;
				}
				else
				{
					UseSecureTyping = false;
				}


			}

			break;
		}

		case WM_SETFOCUS:
			ODSF( L"WM_SETFOCUS:" );

			if (!kbd->SetEncodingTable( UseSecureTyping, setQ ))
			{	// send NULL string to delete all tables, because problems occured 
				kbd->FreeTables( UseSecureTyping );
				kbd->ResetEncodingTable( UseSecureTyping, setQ );
				UseSecureTyping = false;

				if ( !kbd->wasError )
				{
					kbd->wasError = true;
					MessageBox( GetParent(hwnd), "Error in setting encoding table", "KeyGen - editSubclassProc", MB_ICONSTOP );
				}

				EndDialog( GetParent(hwnd), ERROR_INTERNAL_PROBLEM );
			}
			break;

		case WM_KILLFOCUS:
			ODSF( L"WM_KILLFOCUS:" );

			if (!kbd->ResetEncodingTable( UseSecureTyping, setQ ))
			{	/* send NULL string to delete all tables, because problems occured */
				kbd->FreeTables( UseSecureTyping );
			}
			break;

		case WM_KEYDOWN:
			
			virtualKey = (UINT)wParam;
			scanCode = (UINT)((lParam >> 16) & 0xFF);

			kbd->Wm_KeyDown( virtualKey, UseSecureTyping, scanCode );
			kbd->IncrementCharacterOrderCNTRL( (int)wParam, UseSecureTyping );
			break;
		
		case WM_DEADCHAR:
			kbd->Wm_DeadChar();
			break;

		case WM_PASTE:
			lpRes = CallWindowProc(wpEditOrigWndProc, hwnd, uMsg, wParam, lParam);
			GetWindowText( hwnd, Buffer, BUFFER_SIZE );
			return lpRes;

		case WM_CHAR:
		{
			letter = (char)wParam;
			isLetter = TRUE;

			if ( kbd->Wm_Char( letter, UseSecureTyping, &retCharsNum, &retChar1, &retChar2, wcs ) )
			{
				//
				//	A service symbol is detected. Rather CTRL+V
				//
				if ( ( retCharsNum == 1 ) && ( wcs[0] == 0x16 ) )
				{
					wParam = wcs[0];
					lpRes = CallWindowProc(wpEditOrigWndProc, hwnd, uMsg, wParam, lParam);
					return lpRes;
				}

				if ((KeyGenFlags & FLAG_ADVANCED_SHOW_PWD_CHARS) > 0)
				{
					if ( retCharsNum == 2 )
					{
						wParam = (WPARAM)retChar2 & 0xFF;
						CallWindowProc( wpEditOrigWndProc, hwnd, uMsg, wParam, lParam);
						insertCharToPwdBuffer( hwnd, retChar2, Buffer, isLetter );
						wParam = (WPARAM)retChar1 & 0xFF;
						lpRes = CallWindowProc( wpEditOrigWndProc, hwnd, uMsg, wParam, lParam);
						insertCharToPwdBuffer( hwnd, retChar1, Buffer, isLetter );
					}
					else
					if ( retCharsNum == 1 )
					{
						wParam = (WPARAM)retChar2 & 0xFF;
						lpRes = CallWindowProc( wpEditOrigWndProc, hwnd, uMsg, wParam, lParam);
						insertCharToPwdBuffer( hwnd, retChar2, Buffer, isLetter );
					}
					else
					{
						ODS_fls( L"Assertion! return character number is zero!" );
					}
				}
				else if (UseSecureTyping)
				{
					if (retCharsNum == 2)
					{
						CallWindowProc( wpEditOrigWndProc, hwnd, uMsg, wParam, lParam);
						insertCharToPwdBuffer( hwnd, retChar2, Buffer, isLetter );
						letter = getRandomLetter();
						wParam = (WPARAM)letter & 0xFF;
						lpRes = CallWindowProc( wpEditOrigWndProc, hwnd, uMsg, wParam, lParam);
						insertCharToPwdBuffer( hwnd, retChar1, Buffer, isLetter );
					}
					else
					if ( retCharsNum == 1 )
					{
						lpRes = CallWindowProc( wpEditOrigWndProc, hwnd, uMsg, wParam, lParam);
						insertCharToPwdBuffer( hwnd, retChar2, Buffer, isLetter );
					}
					else
					{
						ODS_fls( L"Assertion! return character number is zero!" );
					}
				}
				else
				{
					lpRes = CallWindowProc(wpEditOrigWndProc, hwnd, uMsg, wParam, lParam);
					insertCharToPwdBuffer( hwnd, retChar2, Buffer, isLetter );
				}
			}
			else
			{
				lpRes = CallWindowProc(wpEditOrigWndProc, hwnd, uMsg, wParam, lParam);
				insertCharToPwdBuffer( hwnd, letter, Buffer, isLetter );

				////
				////	1. Uses startndard window procedure for filling edit box.
				////
				//cWCStr wcsOld;
				//wcsOld.GetWindowTextT( hwnd );
				//WORD caretPosOld = LOWORD( (DWORD)SendMessage( hwnd, EM_GETSEL, 0, 0L ) );

				//lpRes = CallWindowProc(wpEditOrigWndProc, hwnd, uMsg, wParam, lParam);

				////
				////	2. Copies added symbols to Buffer.
				////
				//if ( isLetter )
				//{
				//	cWCStr wcsNew;
				//	wcsNew.GetWindowTextT( hwnd );
				//	WORD caretPosNew = LOWORD( (DWORD)SendMessage( hwnd, EM_GETSEL, 0, 0L ) );

				//	int shift = caretPosNew - caretPosOld;

				//	int len = strlen( Buffer );

				//	for ( int i = len; i >= caretPosOld; i-- )
				//	{
				//		Buffer[ i + shift ] = Buffer[ i ];
				//	}

				//	cMBStr mbs( wcsNew.m_szBody, -1, false, CP_ACP );

				//	for( int i = caretPosOld; i < caretPosNew; i++ )
				//	{
				//		Buffer[ i ] = mbs.m_szBody[i];
				//	}
				//}
			}

			return lpRes;
		}

		case WM_SETTEXT:
		{
			char *s = (char *)lParam;

			ODSF( L"WM_SETTEXT '%s'", getWCStr( s ) );

			if ( strlen( Buffer ) == 0 )
			{
				ODSF( L"Assertion! Buffer is empty..." );
				//
				// DigitalPersona-Password-Manager sometimes does not send WM_CHAR message 
				// but always sends WM_SETTEXT
				//

				if ( strlen( s ) < maxLen( Buffer ) )
					strcpy_s( Buffer, maxLen( Buffer ), s );
				else
					ODSF( L"Assertion! Password length too large..." );
			}

			//for( int i = 0; s[i] != L'\0'; i++ )
			//{
			//	InsertCharToBufferW( s[i], true, 0 );
			//}
			break;
		}

		case WM_KEYUP:
			{
				kbd->IncrementCharacterOrder( (int)wParam, UseSecureTyping );
				lpRes = CallWindowProc(wpEditOrigWndProc, hwnd, uMsg, wParam, lParam);

				if	(((KeyGenFlags & FLAG_ADVANCED_SHOW_PWD_CHARS) == 0) &&
					 ( UseSecureTyping && (wParam == 46) ) // DELETE key
					)
				{
					isLetter = FALSE;
					letter = 46;
					insertCharToPwdBuffer( hwnd, letter, Buffer, isLetter );
				}
				
				return lpRes;
			}
	}

	lpRes = CallWindowProc(wpEditOrigWndProc, hwnd, uMsg, wParam, lParam);

	return lpRes;
}

#endif // _Use_editSubclassProcA_

LRESULT APIENTRY editSubclassProcA( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	char    letter;
	LRESULT lpRes;
	BOOL    isLetter;
	UINT    virtualKey, scanCode;
	DWORD   retCharsNum;
	char    retChar1, retChar2;

	isLetter = FALSE;
	
	kbdParams *kbd = kbdParams::GetParams( GetParent(hwnd) );
	if ( kbd == NULL ) return 0;

	static int setQ = 0;

	switch (uMsg)
	{
		case WM_INPUTLANGCHANGE:
		{
			ODSF( L"WM_INPUTLANGCHANGE:" );
			if (UseSecureTyping)
			{
				HWND hParent = GetParent( hwnd );

				kbd->noSemaphore = true;

				kbd->ResetEncodingTable( UseSecureTyping, setQ );
				kbd->FreeTables( UseSecureTyping );

				//kbdParams::Free( kbd );
				//kbd = kbdParams::Allocate();

				if ( kbd != NULL )
				{
					kbdParams::SetParams( hParent, kbd );
					//	SetWindowLong( hParent, DWL_USER, (LONG)kbd );

				kbd->noSemaphore = true;
					kbd->GenerateEncodingTable( UseSecureTyping );
					kbd->SetEncodingTable( UseSecureTyping, setQ );
				kbd->noSemaphore = false;
				}
				else
				{
					UseSecureTyping = false;
				}
			}

			break;
		}

		case WM_SETFOCUS:

			ODSF( L"WM_SETFOCUS:" );

			if (!kbd->SetEncodingTable( UseSecureTyping, setQ ))
			{	/* send NULL string to delete all tables, because problems occured */
				kbd->FreeTables( UseSecureTyping );
				kbd->ResetEncodingTable( UseSecureTyping, setQ );
				UseSecureTyping = false;

				if ( !kbd->wasError )
				{
					kbd->wasError = true;
					MessageBox( GetParent(hwnd), "Error in setting encoding table", "KeyGen - editSubclassProc", MB_ICONSTOP );
				}

				EndDialog( GetParent(hwnd), ERROR_INTERNAL_PROBLEM );
			}
			break;

		case WM_KILLFOCUS:

			ODSF( L"WM_KILLFOCUS:" );

			if (!kbd->ResetEncodingTable( UseSecureTyping, setQ ))
			{	/* send NULL string to delete all tables, because problems occured */
				kbd->FreeTables( UseSecureTyping );
			}
			break;
		case WM_KEYDOWN:
			
			virtualKey = (UINT)wParam;
			scanCode = (UINT)((lParam >> 16) & 0xFF);

			kbd->Wm_KeyDown( virtualKey, UseSecureTyping, scanCode );
			kbd->IncrementCharacterOrderCNTRL( (int)wParam, UseSecureTyping );
			break;
		
		case WM_DEADCHAR:
			kbd->Wm_DeadChar();
			break;

		case WM_CHAR:
		{
			letter = (char)wParam;
			isLetter = TRUE;

			if (kbd->Wm_Char( letter, UseSecureTyping, &retCharsNum, &retChar1, &retChar2 ))
			{
				if ((KeyGenFlags & FLAG_ADVANCED_SHOW_PWD_CHARS) > 0)
				{
					if (retCharsNum == 2)
					{
						wParam = (WPARAM)retChar2 & 0xFF;
						CallWindowProc( wpEditOrigWndProcA, hwnd, uMsg, wParam, lParam);
						insertCharToPwdBuffer( hwnd, retChar2, Buffer, isLetter );
						wParam = (WPARAM)retChar1 & 0xFF;
						lpRes = CallWindowProc( wpEditOrigWndProcA, hwnd, uMsg, wParam, lParam);
						insertCharToPwdBuffer( hwnd, retChar1, Buffer, isLetter );
					}
					else
					{
						wParam = (WPARAM)retChar2 & 0xFF;
						lpRes = CallWindowProc( wpEditOrigWndProcA, hwnd, uMsg, wParam, lParam);
						insertCharToPwdBuffer( hwnd, retChar2, Buffer, isLetter );
					}
				}
				else if (UseSecureTyping)
				{
					if (retCharsNum == 2)
					{
						CallWindowProc( wpEditOrigWndProcA, hwnd, uMsg, wParam, lParam);
						insertCharToPwdBuffer( hwnd, retChar2, Buffer, isLetter );
						letter = getRandomLetter();
						wParam = (WPARAM)letter & 0xFF;
						lpRes = CallWindowProc( wpEditOrigWndProcA, hwnd, uMsg, wParam, lParam);
						insertCharToPwdBuffer( hwnd, retChar1, Buffer, isLetter );
					}
					else
					{
						lpRes = CallWindowProc( wpEditOrigWndProcA, hwnd, uMsg, wParam, lParam);
						insertCharToPwdBuffer( hwnd, retChar2, Buffer, isLetter );
					}
				}
				else
				{
					lpRes = CallWindowProc(wpEditOrigWndProcA, hwnd, uMsg, wParam, lParam);
					insertCharToPwdBuffer( hwnd, retChar2, Buffer, isLetter );
				}
			}
			else
			{
				lpRes = CallWindowProc(wpEditOrigWndProcA, hwnd, uMsg, wParam, lParam);
				insertCharToPwdBuffer( hwnd, letter, Buffer, isLetter );
			}

			return lpRes;
		}

		case WM_KEYUP:
			{
				kbd->IncrementCharacterOrder( (int)wParam, UseSecureTyping );
				lpRes = CallWindowProc(wpEditOrigWndProcA, hwnd, uMsg, wParam, lParam);

				if	(((KeyGenFlags & FLAG_ADVANCED_SHOW_PWD_CHARS) == 0) &&
					 ( UseSecureTyping && (wParam == 46) ) // DELETE key
					)
				{
					isLetter = FALSE;
					letter = 46;
					insertCharToPwdBuffer( hwnd, letter, Buffer, isLetter );
				}
				
				return lpRes;
			}
	}

	lpRes = CallWindowProc(wpEditOrigWndProcA, hwnd, uMsg, wParam, lParam);

	return lpRes;
}


LRESULT APIENTRY editSubclassProcB( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	char    letter;
	LRESULT lpRes;
	BOOL    isLetter;
	UINT    virtualKey, scanCode;
	DWORD   retCharsNum;
	char    retChar1, retChar2;

	isLetter = FALSE;
	
	kbdParams *kbd = kbdParams::GetParams( GetParent(hwnd) );
	if ( kbd == NULL ) return 0;

	static int setQ = 0;

	switch (uMsg)
	{
		case WM_INPUTLANGCHANGE:
		{
			ODSF( L"WM_INPUTLANGCHANGE:" );

			if (UseSecureTyping)
			{
				HWND hParent = GetParent( hwnd );

				kbd->noSemaphore = true;

				kbd->ResetEncodingTable( UseSecureTyping, setQ );
				kbd->FreeTables( UseSecureTyping );

				//kbdParams::Free( kbd );
				//kbd = kbdParams::Allocate();

				if ( kbd != NULL )
				{
					kbdParams::SetParams( hParent, kbd );
					// 	SetWindowLong( hParent, DWL_USER, (LONG)kbd );

				kbd->noSemaphore = true;
					kbd->GenerateEncodingTable( UseSecureTyping );
					kbd->SetEncodingTable( UseSecureTyping, setQ );
				kbd->noSemaphore = false;
				}
				else
				{
					UseSecureTyping = false;
				}
			}

			break;
		}

		case WM_SETFOCUS:

			ODSF( L"WM_SETFOCUS:" );

			if (!kbd->SetEncodingTable( UseSecureTyping, setQ ))
			{	/* send NULL string to delete all tables, because problems occured */
				kbd->FreeTables( UseSecureTyping );
				kbd->ResetEncodingTable( UseSecureTyping, setQ );
				UseSecureTyping = false;
				if ( !kbd->wasError )
				{
					kbd->wasError = true;
					MessageBox( GetParent(hwnd), "Error in setting encoding table", "KeyGen - editSubclassProc", MB_ICONSTOP );
				}
				EndDialog( GetParent(hwnd), ERROR_INTERNAL_PROBLEM );
			}
			break;

		case WM_KILLFOCUS:

			ODSF( L"WM_KILLFOCUS:" );

			if (!kbd->ResetEncodingTable( UseSecureTyping, setQ ))
			{	/* send NULL string to delete all tables, because problems occured */
				kbd->FreeTables( UseSecureTyping );
			}
			break;

		case WM_KEYDOWN:
			
			virtualKey = (UINT)wParam;
			scanCode = (UINT)((lParam >> 16) & 0xFF);

			kbd->Wm_KeyDown( virtualKey, UseSecureTyping, scanCode );
			kbd->IncrementCharacterOrderCNTRL( (int)wParam, UseSecureTyping );
			break;
		
		case WM_DEADCHAR:
			kbd->Wm_DeadChar();
			break;

		case WM_CHAR:
			{
				letter = (char)wParam;
				isLetter = TRUE;

				if (kbd->Wm_Char( letter, UseSecureTyping, &retCharsNum, &retChar1, &retChar2 ))
				{
					if ((KeyGenFlags & FLAG_ADVANCED_SHOW_PWD_CHARS) > 0)
					{
						if (retCharsNum == 2)
						{
							wParam = (WPARAM)retChar2 & 0xFF;
							CallWindowProc( wpEditOrigWndProcB, hwnd, uMsg, wParam, lParam);
							insertCharToPwdBuffer( hwnd, retChar2, BufferB, isLetter );
							wParam = (WPARAM)retChar1 & 0xFF;
							lpRes = CallWindowProc( wpEditOrigWndProcB, hwnd, uMsg, wParam, lParam);
							insertCharToPwdBuffer( hwnd, retChar1, BufferB, isLetter );
						}
						else
						{
							wParam = (WPARAM)retChar2 & 0xFF;
							lpRes = CallWindowProc( wpEditOrigWndProcB, hwnd, uMsg, wParam, lParam);
							insertCharToPwdBuffer( hwnd, retChar2, BufferB, isLetter );
						}
					}
					else if (UseSecureTyping)
					{
						if (retCharsNum == 2)
						{
							CallWindowProc( wpEditOrigWndProcB, hwnd, uMsg, wParam, lParam);
							insertCharToPwdBuffer( hwnd, retChar2, BufferB, isLetter );
							letter = getRandomLetter();
							wParam = (WPARAM)letter & 0xFF;
							lpRes = CallWindowProc( wpEditOrigWndProcB, hwnd, uMsg, wParam, lParam);
							insertCharToPwdBuffer( hwnd, retChar1, BufferB, isLetter );
						}
						else
						{
							lpRes = CallWindowProc( wpEditOrigWndProcB, hwnd, uMsg, wParam, lParam);
							insertCharToPwdBuffer( hwnd, retChar2, BufferB, isLetter );
						}
					}
					else
					{
						lpRes = CallWindowProc(wpEditOrigWndProcB, hwnd, uMsg, wParam, lParam);
						insertCharToPwdBuffer( hwnd, retChar2, BufferB, isLetter );
					}
				}
				else
				{
					lpRes = CallWindowProc(wpEditOrigWndProcB, hwnd, uMsg, wParam, lParam);
					insertCharToPwdBuffer( hwnd, letter, BufferB, isLetter );
				}

				return lpRes;
			}

		case WM_KEYUP:
			{
				kbd->IncrementCharacterOrder( (int)wParam, UseSecureTyping );
				lpRes = CallWindowProc(wpEditOrigWndProcB, hwnd, uMsg, wParam, lParam);

				if	(((KeyGenFlags & FLAG_ADVANCED_SHOW_PWD_CHARS) == 0) &&
					 ( UseSecureTyping && (wParam == 46) ) // DELETE key
					)
				{
					isLetter = FALSE;
					letter = 46;
					insertCharToPwdBuffer( hwnd, letter, BufferB, isLetter );
				}
				
				return lpRes;
			}
	}

	lpRes = CallWindowProc(wpEditOrigWndProcB, hwnd, uMsg, wParam, lParam);

	return lpRes;
}

//
//	_GetLogicalDrives()
//	We have to ask BestCrypt about used mask before.
//	Reasons:
//	In Windows NT environment, if any BC's drive was not full dismount 
//	( only symbolik link removed ) the drive reported by system as free and
//	the container can be mounted again only on previous drive letter 
//	( see FLAG_MOUNT_ON_SINGLE_LETTER ).
//	But! If we mount other container the letter is to be band for mounting.
//	We have to ask BestCrypt driver for the really used virtual drive mask.
//
static DWORD _GetLogicalDrives()
{
	OSVERSIONINFO ver;
	DWORD mask;
	BOOL  isWin2000 = FALSE;

	mask = GetLogicalDrives();

	ver.dwOSVersionInfoSize = sizeof(ver);

	GetVersionEx( &ver );

	if ( ( ver.dwPlatformId == VER_PLATFORM_WIN32_NT ) &&
		 ( ver.dwMajorVersion == 5 ) && 
		 ( ver.dwMinorVersion >= 0 )
	   )
		isWin2000 = TRUE;
	else 
		isWin2000 = FALSE;
	
	if (!isWin2000)
	{
		DWORD bcMask = 0;

		if ( CmdGetVDiskMask( bcMask ) )
			mask |= bcMask;
	}

	return mask;
}

static void DrawFileIcon( HWND hDlg, HDC dc, HICON hIcon, int id )
{
	if ( hIcon == NULL )
		return;

	HWND hWnd = GetDlgItem( hDlg, id );

	if ( hWnd == NULL )
		return;

	ShowWindow( hWnd, SW_HIDE );

	RECT r;
	GetWindowRect( hWnd, &r );
	//POINT p = { r.left, r.top };
	//ScreenToClient( hDlg, &p );
	//DrawIcon( dc, p.x, p.y, hIcon );
	MapWindowPoints( NULL, hDlg, (POINT*)&r, 2 );
	DrawIcon( dc, r.left, r.top, hIcon );
}

//
//	processContIcon
//		preprocesses dlg window procedure.
//		use it just before main message switch
//
static void processContIcon( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	static HICON hContIcon = NULL;

	switch (uMsg) 
	{
        case WM_INITDIALOG:
		{
			ODSFuncS( L"WM_INITDIALOG" );
			wchar_t path[ MAX_PATH ];
			GetModuleFileNameT( NULL, path, maxLen( path ) );
			hContIcon = ExtractIconT( hMainInst, path, 1 );
			break;
		}
		case WM_PAINT :
		{
			HDC dc;
			PAINTSTRUCT ps;

			if ((HDC) wParam != NULL)
				dc = (HDC) wParam;
			else
				dc = BeginPaint(hwnd,&ps);

			DrawFileIcon( hwnd, dc, hContIcon, IDC_Icon );

			if ((HDC) wParam == NULL)
				EndPaint(hwnd,&ps);

			break;
		}

		case WM_DESTROY:

			if ( hContIcon != NULL )
			{
				DestroyIcon( hContIcon );
				hContIcon = NULL;
			}

			break;

		default :
			break;
	}
}

BOOL CALLBACK getPasswordDlgProc( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
    WORD  pwdLength;
	HWND  editHwnd, cmbHwnd;
	int   i, selDrive = 0, curSel;
	DWORD curLogicalDisks;
	char  driveName[] = "A:";

	static HFONT hFont = NULL;

	processContIcon( hwnd, uMsg, wParam, lParam );

	switch (uMsg) 
	{
		case WM_INITDIALOG:
		{
			ODSFuncS( L"WM_INITDIALOG" );
			if ( pTranslateDll != NULL )
				pTranslateDll->TranslateDialog( hwnd, false );

			if ( GetParent( hwnd ) == NULL )
				CenterPosition( hwnd );

			if (Caption)  SetWindowTextT( hwnd, Caption );
			if (FileName) SetWindowTextT( GetDlgItem(hwnd, IDC_TextName), FileName );
			if (PathName) SetWindowTextT( GetDlgItem(hwnd, IDC_TextPath), PathName );

			if (!UseSecureTyping) ShowWindow( GetDlgItem(hwnd, IDC_LINE), SW_HIDE );
			
			Button_SetCheck( GetDlgItem(hwnd, IDC_AUTOMOUNT), (KeyGenFlags & FLAG_CONTAINER_AUTO_MOUNT) > 0);
			Button_SetCheck( GetDlgItem(hwnd, IDC_READONLY),  (KeyGenFlags & FLAG_CONTAINER_READ_ONLY) > 0);

			if ( KeyGenFlags & FLAG_FORCE_READ_ONLY )
			{
				Button_SetCheck( GetDlgItem(hwnd, IDC_READONLY),  TRUE  );
				Button_Enable  ( GetDlgItem(hwnd, IDC_READONLY),  FALSE );
			}

			if ((KeyGenFlags & FLAG_ADVANCED_SHOW_PWD_CHARS) > 0)
			{
				Edit_SetPasswordChar(GetDlgItem(hwnd, IDC_PASS), 0);
			}
			else
			{
				if ( changeFont )
					hFont = CreateFont(0,0,0,0, FW_NORMAL, FALSE, FALSE, FALSE, 
							DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY,
							DEFAULT_PITCH, szPwdFontName );

				if ( hFont != NULL )
				{
					SendMessage( GetDlgItem( hwnd, IDC_PASS ), WM_SETFONT, (WPARAM)hFont, false );
				}
			}

			cmbHwnd = GetDlgItem(hwnd, IDC_DRIVES);

			if ( KeyGenFlags & FLAG_EXPIRED_DEMO )
			{
				cWCStr wcs;
				wcs.GetWindowTextT( hwnd );
				wcs.AddF( L" (%s)", _LoadIDS( IDS_szExpiredCaption ) );
				SetWindowTextT( hwnd, wcs.m_szBody );
			}

			if ((KeyGenFlags & FLAG_MOUNT_ON_SINGLE_LETTER) > 0)
			{
				driveName[0] = 'A' + (char)KeyGenFavoriteDrive;
				ComboBox_InsertString( cmbHwnd, 0, driveName);
				selDrive = 0;
			}
			else
			{
				curLogicalDisks = _GetLogicalDrives();

				for (i=2, curSel=0; i<26; i++ )
				{
					if ((curLogicalDisks & (1 << i)) == 0)
					{	driveName[0] = 'A' + i;
						ComboBox_InsertString( cmbHwnd, curSel, driveName);
						if (i == (int)KeyGenFavoriteDrive)
						{ selDrive = curSel; }

						curSel++;
					}
				}
			}
			ComboBox_SetCurSel( cmbHwnd, selDrive);

			/* Set the default push button to "Cancel" */ 
			SendMessage(hwnd, DM_SETDEFID, (WPARAM) IDCANCEL, (LPARAM) 0); 

			editHwnd = GetDlgItem(hwnd, IDC_PASS); 

#ifdef STRICT
			wpEditOrigWndProc = (WNDPROC) SetWindowLong(editHwnd,GWL_WNDPROC, (LONG) editSubclassProc);
#else
			wpEditOrigWndProc = (FARPROC) SetWindowLong(editHwnd,GWL_WNDPROC, (LONG) editSubclassProc);
#endif
//			kbdParams *kbd = new kbdParams;
			kbdParams *kbd = kbdParams::Allocate();
			if ( kbd == NULL )	{ assert( false ); return false; }

			kbdParams::SetParams( hwnd, kbd );
			//	SetWindowLong( hwnd, DWL_USER, (LONG)kbd );
			
			return (kbd->GenerateEncodingTable( UseSecureTyping ));
		}
		case WM_DESTROY:
		{
			if ( hFont != NULL )
			{
				DeleteObject( hFont );
				hFont = NULL;
			}

			// Retrieve the handle of the edit control
			editHwnd = GetDlgItem(hwnd, IDC_PASS);

			// Remove the subclass from the edit control.
			SetWindowLong(editHwnd, GWL_WNDPROC, (LONG) wpEditOrigWndProc); 

			kbdParams *kbd = kbdParams::GetParams( hwnd );

			if ( kbd == NULL )
				return 0;

			kbdParams::Free( kbd );

			break;
		}
		case WM_COMMAND: 
		{
			kbdParams *kbd = kbdParams::GetParams( hwnd );
			if ( kbd == NULL ) return 0;

			/* Set the default push button to "OK" when the user enters text */
			if((HIWORD (wParam) == EN_CHANGE) && (LOWORD(wParam) == IDC_PASS)) 
				SendMessage(hwnd, DM_SETDEFID, (WPARAM) IDOK, (LPARAM) 0); 
	
			switch(wParam) 
			{ 
				case IDOK:
				{
					/* Get number of characters. */ 
					pwdLength = (WORD) SendDlgItemMessage( hwnd, IDC_PASS,
									EM_LINELENGTH, (WPARAM) 0, (LPARAM) 0); 

					if (pwdLength == 0) 
					{

						//
						//	See commants at WM_SETTEXT
						//
						int bufferLen = strlen( Buffer );

						if ( bufferLen != 0 )
							pwdLength = bufferLen;
						else
						{
							wchar_t msgText[128];
							wchar_t captionText[32];
							LoadStringT( HModule, IDS_PASSWORD_EMPTY, msgText, maxLen(msgText) );
							LoadStringT( HModule, IDS_INVALID_PWD, captionText, maxLen(captionText) );
							MessageBoxT( hwnd, msgText, captionText, MB_OK | MB_ICONWARNING );
							return 0; /* Let user to continue to edit password */ 
						}
					} 
		            
					pwdLength = pwdLength < PASSWORD_MAXIMUM_LENGTH ? pwdLength : PASSWORD_MAXIMUM_LENGTH;

					//
					//	Note 1.
					// kbd->Wm_Char  uses _ToUnicodeEx instead of ToAsciiEx
					//	and translated character is always copied to Buffer
					//
					///**/
					//if ( ((KeyGenFlags & FLAG_ADVANCED_SHOW_PWD_CHARS) > 0) || (!UseSecureTyping) )
					//{
					//	// Put the number of characters into first word of buffer
					//	*((LPWORD)Password) = pwdLength; 
	
					//	// Get the characters.
					//	SendDlgItemMessage(hwnd, IDC_PASS, EM_GETLINE, (WPARAM)0, (LPARAM)Password);
	
					//	Password[pwdLength] = 0; // Null-terminate the string
					//}
					//else
					/**/
					{
						Buffer[pwdLength] = Password[pwdLength] = 0;
						memcpy(Password, Buffer, pwdLength);
					}

					kbd->FreeTables( UseSecureTyping );
	
					if ( Button_GetCheck( GetDlgItem(hwnd, IDC_AUTOMOUNT)))
							{ KeyGenFlags |= FLAG_CONTAINER_AUTO_MOUNT; }
					else	{ KeyGenFlags &= ~FLAG_CONTAINER_AUTO_MOUNT; }

					if ( Button_GetCheck( GetDlgItem(hwnd, IDC_READONLY)))
						KeyGenFlags |= FLAG_CONTAINER_READ_ONLY;
					else KeyGenFlags &= ~FLAG_CONTAINER_READ_ONLY;

					curSel = ComboBox_GetCurSel( GetDlgItem(hwnd, IDC_DRIVES) );
					ComboBox_GetText( GetDlgItem(hwnd, IDC_DRIVES), driveName, sizeof(driveName));
					KeyGenFavoriteDrive = (DWORD)(driveName[0] - 'A');
					
					EndDialog(hwnd, ERROR_NO);
					return TRUE; 
				}
				case IDCANCEL: 
				{
					if ( Button_GetCheck( GetDlgItem(hwnd, IDC_AUTOMOUNT)))
							{ KeyGenFlags |= FLAG_CONTAINER_AUTO_MOUNT; }
					else	{ KeyGenFlags &= ~FLAG_CONTAINER_AUTO_MOUNT; }

					if ( Button_GetCheck( GetDlgItem(hwnd, IDC_READONLY)))
						KeyGenFlags |= FLAG_CONTAINER_READ_ONLY;
					else KeyGenFlags &= ~FLAG_CONTAINER_READ_ONLY;

					SetDlgItemText( hwnd, IDC_PASS, "qwerty" );

					EndDialog(hwnd, (int)ERROR_CANCELLED_BY_USER); 

					kbd->FreeTables( UseSecureTyping );
					return TRUE; 
				}
			} 
			return 0; 
		}
	} 

    return FALSE; 

	UNREFERENCED_PARAMETER(lParam);
}

BOOL CALLBACK getPasswordDlgProcSimple( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
    WORD  pwdLength;
	HWND  editHwnd;
 
	static HFONT hFont = NULL;

	processContIcon( hwnd, uMsg, wParam, lParam );

	switch (uMsg) 
	{ 
        case WM_INITDIALOG:
		{
			ODSFuncS( L"WM_INITDIALOG" );

			if ( pTranslateDll != NULL )
				pTranslateDll->TranslateDialog( hwnd, false );

			if ( GetParent( hwnd ) == NULL )
				CenterPosition( hwnd );

			if ( wcslen( Caption )  )		SetWindowTextT( hwnd, Caption );
			if ( wcslen( FileName ) )		SetWindowTextT( GetDlgItem(hwnd, IDC_TextName), FileName );
			if ( wcslen( PathName ) )		SetWindowTextT( GetDlgItem(hwnd, IDC_TextPath), PathName );
			if ( wcslen( NotFileName ) )	SetWindowTextT( GetDlgItem(hwnd, IDC_NotFileName), NotFileName );

			if ( wcslen( PathName ) == 0 )
			{
				ShowWindow( GetDlgItem(hwnd, IDC_Text), SW_HIDE );
				ShowWindow( GetDlgItem(hwnd, IDC_TextName), SW_HIDE );
				ShowWindow( GetDlgItem(hwnd, IDC_TextPath), SW_HIDE );
				ShowWindow( GetDlgItem(hwnd, IDC_Location), SW_HIDE );
			}
			else
			{
				ShowWindow( GetDlgItem(hwnd, IDC_NotFileName), SW_HIDE );
			}

			if (!UseSecureTyping) ShowWindow( GetDlgItem(hwnd, IDC_LINE), SW_HIDE );

			if ((KeyGenFlags & FLAG_ADVANCED_SHOW_PWD_CHARS) > 0)
			{
				Edit_SetPasswordChar(GetDlgItem(hwnd, IDC_PASS), 0);
			}
			else
			{
				if ( changeFont )
					hFont = CreateFont(0,0,0,0, FW_NORMAL, FALSE, FALSE, FALSE, 
						DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY,
						DEFAULT_PITCH, szPwdFontName);

				if ( hFont != NULL )
				{
					SendMessage( GetDlgItem( hwnd, IDC_PASS ), WM_SETFONT, (WPARAM)hFont, false );
				}
			}

			/* Set the default push button to "Cancel" */ 
			SendMessage(hwnd, DM_SETDEFID, (WPARAM) IDCANCEL, (LPARAM) 0); 

			editHwnd = GetDlgItem(hwnd, IDC_PASS); 

#ifdef STRICT
			wpEditOrigWndProc = (WNDPROC) SetWindowLong(editHwnd,GWL_WNDPROC, (LONG) editSubclassProc);
#else
			wpEditOrigWndProc = (FARPROC) SetWindowLong(editHwnd,GWL_WNDPROC, (LONG) editSubclassProc);
#endif

			kbdParams *kbd = kbdParams::Allocate();
			if ( kbd == NULL )	{ assert( false ); return false; }
			kbdParams::SetParams( hwnd, kbd );
			// 	SetWindowLong( hwnd, DWL_USER, (LONG)kbd );

			return kbd->GenerateEncodingTable( UseSecureTyping );
		}

		case WM_DESTROY:
		{
			if ( hFont != NULL )
			{
				DeleteObject( hFont );
				hFont = NULL;
			}

			// Retrieve the handle of the edit control
			editHwnd = GetDlgItem(hwnd, IDC_PASS);

			// Remove the subclass from the edit control.
			SetWindowLong(editHwnd, GWL_WNDPROC, (LONG) wpEditOrigWndProc); 

			kbdParams *kbd = kbdParams::GetParams( hwnd );
			if ( kbd == NULL ) return 0;
			kbdParams::Free( kbd );

			break;
		}
        case WM_COMMAND: 
		{
			kbdParams *kbd = kbdParams::GetParams( hwnd );
			if ( kbd == NULL ) return 0;

			/* Set the default push button to "OK" when the user enters text */
			if((HIWORD (wParam) == EN_CHANGE) && (LOWORD(wParam) == IDC_PASS)) 
				SendMessage(hwnd, DM_SETDEFID, (WPARAM) IDOK, (LPARAM) 0); 
 
			switch(wParam) 
			{ 
				case IDOK: 
					/* Get number of characters. */ 
					pwdLength = (WORD) SendDlgItemMessage( hwnd, IDC_PASS,
									EM_LINELENGTH, (WPARAM) 0, (LPARAM) 0); 

                    if (pwdLength == 0) 
					{	wchar_t msgText[128];
						wchar_t captionText[32];
						LoadStringT( HModule, IDS_PASSWORD_EMPTY, msgText, maxLen( msgText ) );
						LoadStringT( HModule, IDS_INVALID_PWD, captionText, maxLen( captionText ) );
						MessageBoxT(hwnd, msgText, captionText, MB_OK | MB_ICONWARNING );
						return 0; /* Let user to continue to edit password */ 
                    } 
                    
					pwdLength = pwdLength < PASSWORD_MAXIMUM_LENGTH ? pwdLength : PASSWORD_MAXIMUM_LENGTH;

					//
					// See Note 1 above.
					//
					///**/
					//if ( ((KeyGenFlags & FLAG_ADVANCED_SHOW_PWD_CHARS) > 0) || (!UseSecureTyping) )
					//{
					//
					//	// Put the number of characters into first word of buffer
					//	*((LPWORD)Password) = pwdLength; 
 
					//	// Get the characters.
					//	SendDlgItemMessage(hwnd, IDC_PASS, EM_GETLINE,
					//				(WPARAM) 0, (LPARAM)Password);
 
					//	Password[pwdLength] = 0; // Null-terminate the string.
					//}
					//else
					///**/
					{
						Buffer[pwdLength] = Password[pwdLength] = 0;
						memcpy(Password, Buffer, pwdLength);
					}

					kbd->FreeTables( UseSecureTyping );

					SetDlgItemText( hwnd, IDC_PASS, "qwerty");
					EndDialog(hwnd, ERROR_NO);
					return TRUE; 
 
				case IDCANCEL: 
					SetDlgItemText( hwnd, IDC_PASS, "qwerty");
					EndDialog(hwnd, (int)ERROR_CANCELLED_BY_USER); 
					kbd->FreeTables( UseSecureTyping );
					return TRUE; 
            } 
            return 0; 
		}
    } 
    return FALSE; 

	UNREFERENCED_PARAMETER(lParam);
}

void ShowErrorMismatch(HWND hwnd )
{
	wchar_t caption[ 256 ];
	wchar_t message[ 256 ];

	HINSTANCE hInst = hMainInst; // GetModuleHandle( KGSHA_MODULE_NAME );

	LoadStringT( hInst, IDS_PASSWORD_NOT_SAME, message, maxLen( message ) );
	LoadStringT( hInst, IDS_PASSWORD_MISMATCH, caption, maxLen( caption ) );

	MessageBoxT( hwnd, message, caption, MB_ICONEXCLAMATION );
}


BOOL CALLBACK getPasswordVerifyDlgProc( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
    WORD pwdLength;
	char tmpStr[ PASSWORD_MAXIMUM_LENGTH + 4 ];
	WORD tmpStrLength;
	HWND editHwndA, editHwndB;

	static HFONT hFont = NULL;

	processContIcon( hwnd, uMsg, wParam, lParam );

	switch (uMsg) 
	{ 
        case WM_INITDIALOG: 
		{	
			ODSFuncS( L"WM_INITDIALOG" );

			if ( pTranslateDll != NULL )
				pTranslateDll->TranslateDialog( hwnd, false );

			if ( GetParent( hwnd ) == NULL )
				CenterPosition( hwnd );

			if (Caption)  SetWindowTextT( hwnd, Caption );
			if (FileName) SetWindowTextT( GetDlgItem(hwnd, IDC_TextName), FileName );
			if (PathName) SetWindowTextT( GetDlgItem(hwnd, IDC_TextPath), PathName );

			if (!UseSecureTyping) 
			{	ShowWindow( GetDlgItem(hwnd, IDC_LINEC), SW_HIDE );
				ShowWindow( GetDlgItem(hwnd, IDC_LINEB), SW_HIDE );
			}

			if ((KeyGenFlags & FLAG_ADVANCED_SHOW_PWD_CHARS) > 0)
			{
				Edit_SetPasswordChar(GetDlgItem(hwnd, IDC_PASSA), 0);
				Edit_SetPasswordChar(GetDlgItem(hwnd, IDC_PASSB), 0);
			}
			else
			{
				if ( changeFont )
					hFont = CreateFont(0,0,0,0, FW_NORMAL, FALSE, FALSE, FALSE, 
						DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY,
						DEFAULT_PITCH, szPwdFontName );

				if ( hFont != NULL )
				{
					SendMessage( GetDlgItem( hwnd, IDC_PASSA ), WM_SETFONT, (WPARAM)hFont, false );
					SendMessage( GetDlgItem( hwnd, IDC_PASSB ), WM_SETFONT, (WPARAM)hFont, false );
				}
			}

			/* Set the default push button to "Cancel" */ 
			SendMessage(hwnd, DM_SETDEFID, (WPARAM) IDCANCEL, (LPARAM) 0); 

			editHwndA = GetDlgItem(hwnd, IDC_PASSA); 
			editHwndB = GetDlgItem(hwnd, IDC_PASSB); 

#ifdef STRICT
			wpEditOrigWndProcA = (WNDPROC) SetWindowLong(editHwndA, GWL_WNDPROC, (LONG) editSubclassProcA);
			wpEditOrigWndProcB = (WNDPROC) SetWindowLong(editHwndB, GWL_WNDPROC, (LONG) editSubclassProcB);
#else
			wpEditOrigWndProcA = (FARPROC) SetWindowLong(editHwndA, GWL_WNDPROC, (LONG) editSubclassProcA);
			wpEditOrigWndProcB = (FARPROC) SetWindowLong(editHwndB, GWL_WNDPROC, (LONG) editSubclassProcB);
#endif
			kbdParams *kbd = kbdParams::Allocate();
			if ( kbd == NULL )	{ assert( false ); return false; }
			kbdParams::SetParams( hwnd, kbd );
			//	SetWindowLong( hwnd, DWL_USER, (LONG)kbd );

			return kbd->GenerateEncodingTable( UseSecureTyping );
		}

		case WM_DESTROY:
		{
			if ( hFont != NULL )
			{
				DeleteObject( hFont );
				hFont = NULL;
			}

			// Retrieve the handle of the edit control
			editHwndA = GetDlgItem(hwnd, IDC_PASSA);
			editHwndB = GetDlgItem(hwnd, IDC_PASSB);

			// Remove the subclass from the edit control.
			SetWindowLong(editHwndA, GWL_WNDPROC, (LONG) wpEditOrigWndProcA); 
			SetWindowLong(editHwndB, GWL_WNDPROC, (LONG) wpEditOrigWndProcB);

			kbdParams *kbd = kbdParams::GetParams( hwnd );
			if ( kbd == NULL ) return 0;
			kbdParams::Free( kbd );

			break;
		}

        case WM_COMMAND: 
		{
			kbdParams *kbd = kbdParams::GetParams( hwnd );
			if ( kbd == NULL ) return 0;

			/* Set the default push button to "OK" when the user enters text */
			if((HIWORD (wParam) == EN_CHANGE) && (LOWORD(wParam) == IDC_PASSA)) 
				SendMessage(hwnd, DM_SETDEFID, (WPARAM) IDOK, (LPARAM) 0); 
 
			switch(wParam) 
			{ 
				case IDOK: 
					/* Get number of characters. */ 
					pwdLength = (WORD) SendDlgItemMessage( hwnd, IDC_PASSA,
									EM_LINELENGTH, (WPARAM) 0, (LPARAM) 0); 

                    if (pwdLength < PASSWORD_MINIMUM_LENGTH) 
					{	wchar_t msgText[128];
						wchar_t captionText[32];
						LoadStringT( HModule, IDS_PWD_TOO_SHORT, msgText, maxLen(msgText) );
						LoadStringT( HModule, IDS_INVALID_PWD, captionText, maxLen(captionText) );
						MessageBoxT( hwnd, msgText, captionText, MB_OK | MB_ICONWARNING );
						SetFocus( GetDlgItem( hwnd, IDC_PASSA ));
						return 0; /* Let user to continue to edit password */ 
                    } 
                    
					pwdLength = pwdLength < PASSWORD_MAXIMUM_LENGTH ? pwdLength : PASSWORD_MAXIMUM_LENGTH;

					//
					// See Note 1 above.
					//
					///**/
					//if ( ((KeyGenFlags & FLAG_ADVANCED_SHOW_PWD_CHARS) > 0) || (!UseSecureTyping) )
					//{
					//	// Put the number of characters into first word of buffer
					//	*((LPWORD)Password) = pwdLength; 
 
					//	// Get the characters.
					//	SendDlgItemMessage(hwnd, IDC_PASSA, EM_GETLINE, (WPARAM) 0, (LPARAM)Password); 
 
					//	Password[pwdLength] = 0; // Null-terminate the string.
 
					//	// Get the password verification string
					//	tmpStrLength = (WORD) SendDlgItemMessage( hwnd, IDC_PASSB, 
					//					EM_LINELENGTH, (WPARAM) 0, (LPARAM) 0);
					//}
					//else
					///**/
					{
						Buffer[pwdLength] = Password[pwdLength] = 0;
						memcpy(Password, Buffer, pwdLength);
						tmpStrLength = strlen( BufferB );
					}

					if (tmpStrLength == 0)
					{	SetFocus( GetDlgItem( hwnd, IDC_PASSB ));
						return 0; /* let user to type the second copy of password */
					}
					else if (tmpStrLength != pwdLength)
					{
						ShowErrorMismatch( hwnd );
						SetFocus( GetDlgItem( hwnd, IDC_PASSB ));
						return 0;
					}

					//
					// See Note 1 above.
					//
					///**/
					//if ( ((KeyGenFlags & FLAG_ADVANCED_SHOW_PWD_CHARS) > 0) || (!UseSecureTyping) )
					//{
					//	*((LPWORD)tmpStr) = pwdLength; 
					//	SendDlgItemMessage(hwnd, IDC_PASSB, EM_GETLINE, (WPARAM) 0, (LPARAM)tmpStr); 
					//	tmpStr[pwdLength] = 0; // Null-terminate the string.
					//}
					//else
					///**/
					{
						BufferB[pwdLength] = tmpStr[pwdLength] = 0;
						memcpy(tmpStr, BufferB, pwdLength);
					}

					if (strcmp( tmpStr, Password ) != 0)
					{
						ShowErrorMismatch( hwnd );
						SetFocus( GetDlgItem( hwnd, IDC_PASSB ));
						return 0;
					}

					kbd->FreeTables( UseSecureTyping );
 
					SetDlgItemText( hwnd, IDC_PASSA, "qwerty");
					SetDlgItemText( hwnd, IDC_PASSB, "qwerty");
					EndDialog(hwnd, ERROR_NO); 
					return TRUE; 
 
				case IDCANCEL: 
					SetDlgItemText( hwnd, IDC_PASSA, "qwerty");
					SetDlgItemText( hwnd, IDC_PASSB, "qwerty");
					EndDialog(hwnd, (int)ERROR_CANCELLED_BY_USER); 
					kbd->FreeTables( UseSecureTyping );
					return TRUE; 
            } 

            return 0; 
		}
    } 
    return FALSE;

	UNREFERENCED_PARAMETER(lParam); 
}



//// Subclass procedure
//LRESULT APIENTRY progressSubclassProc( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
//{
//	HWND parentHwnd;
//
//	switch (uMsg)
//	{
//		case WM_GETDLGCODE:
//			return DLGC_WANTALLKEYS;
//
//		case WM_CHAR:
//		case WM_KEYUP:
//		case WM_KEYDOWN:
//		case WM_SYSKEYDOWN:
//			/* the code for generation Seed value */
//			if (!UpdateSeed( Seed ))
//			{	parentHwnd = GetParent( hwnd );
//				SendMessage( parentHwnd, MSG_SEED_READY, 0, 0);
//			}
//			else
//			{	SendMessage( hwnd, PBM_SETPOS, (WPARAM)(SeedCounter/2), 0);
//			}
//	}
//	return CallWindowProc(wpOrigWndProc, hwnd, uMsg, wParam, lParam);
//}
//


//BOOL CALLBACK getRandomDlgProc( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
//{
//	HWND   prHwnd;
//	
//	switch (uMsg) 
//	{ 
//        case WM_INITDIALOG:
//		{
//			ODSFuncS( L"WM_INITDIALOG" );
//
//			if ( pTranslateDll != NULL )
//				pTranslateDll->TranslateDialog( hwnd, false );
//
//			if ( GetParent( hwnd ) == NULL )
//				CenterPosition( hwnd );
//
//			// Retrieve the handle of the progress indicator control
//			prHwnd = GetDlgItem(hwnd, IDC_PROGRESS); 
//			SendMessage( prHwnd, PBM_SETRANGE, 0, MAKELPARAM(0, SeedLength));
// 
//			EnableWindow( GetDlgItem(hwnd, IDOK), FALSE); 
//
//			// Subclass the  progress indicator control
//#ifdef STRICT
//			wpOrigWndProc = (WNDPROC) SetWindowLong(prHwnd,
//					GWL_WNDPROC, (LONG) progressSubclassProc);
//#else
//			wpOrigWndProc = (FARPROC) SetWindowLong(prHwnd,
//					GWL_WNDPROC, (LONG) progressSubclassProc);
//#endif
//			return TRUE;
//		}
//		case WM_DESTROY:
//			// Retrieve the handle of the  progress indicator control
//			prHwnd = GetDlgItem(hwnd, IDC_PROGRESS);
//
//			// Remove the subclass from the progress indicator control.
//			SetWindowLong(prHwnd, GWL_WNDPROC, (LONG) wpOrigWndProc); 
//
//			break;
// 
//		case MSG_SEED_READY:
//			EnableWindow( GetDlgItem(hwnd, IDOK), TRUE); 
//
//			SetFocus( GetDlgItem(hwnd, IDOK) ); 
//			break;
//
//		case WM_COMMAND:
//			switch (wParam)
//			{
//				case IDOK: 
//					EndDialog(hwnd, (int)ERROR_CANCELLED_BY_USER); 
//					return TRUE; 
//			}
//			break;
//
//		default:
//			return 0;
//	}
//	return 0;
//}


//BOOL UpdateSeed( BYTE *Seed )
//{
//	BYTE          ticks = 0;
//	LARGE_INTEGER frequency;
//	SYSTEMTIME    systemTime;
//	DWORD         offset;
//
//	if (SeedCounter >= SeedLength*2) return FALSE;
//
//	/* Assume that we have no hardware for high-frequency 
//	performance counter */
//	GetSystemTime( &systemTime );
//	ticks = (BYTE)(systemTime.wMilliseconds);
//
//	/*	We try to get time with as much resolution as possible.
//		Intel processors supports Performance counter and
//		the count might be the cycle rate of the processor clock.
//		It means that on 100 MHz processor the counter updates 
//		~100.000.000 per second. We use only 4 low bits.
//	*/
//	if(QueryPerformanceFrequency(&frequency))
//	{	/* check if high performance counter updates less than 1 per millisec */
//		if ((frequency.LowPart > 1000) || (frequency.HighPart > 0))
//		{ 
//			QueryPerformanceCounter( &frequency );
//			ticks = (BYTE)(frequency.LowPart);
//		}
//	}
//
//	/* we will use only least 4 bites of ticks */
//	offset = SeedCounter;
//
//	if (offset & 1)	
//	{ Seed[offset/2] = (0x0f & Seed[offset/2]) | (0xf0 & ticks);
//	}
//	else
//	{ Seed[offset/2] = (0xf0 & Seed[offset/2]) | (0x0f & ticks);
//	}
//
//	SeedCounter++;
//	return TRUE;
//}


// Local: Get Advanced settings procedure

BOOL bDebug = FALSE;

LRESULT APIENTRY editSubclassProcAdvanced( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	UINT virtualKey, scanCode;
	char letter;

	//char dbgStr[64]; // for debugging

	kbdParams *kbd = kbdParams::GetParams( GetParent( hwnd ) );
	if ( kbd == NULL ) return 0;

	static int setQ = 0;

	switch (uMsg)
	{
		case WM_INPUTLANGCHANGE:
		{
			ODSF( L"WM_INPUTLANGCHANGE:" );

			if (UseSecureTyping)
			{
				HWND hParent = GetParent( hwnd );

				kbd->noSemaphore = true;

				kbd->ResetEncodingTable( UseSecureTyping, setQ );
				kbd->FreeTables( UseSecureTyping );

				kbdParams::Free( kbd );

				kbd = kbdParams::Allocate();
				if ( kbd != NULL )
				{
					kbdParams::SetParams( hParent, kbd );
					// 	SetWindowLong( hParent, DWL_USER, (LONG)kbd );
				kbd->noSemaphore = true;
					kbd->GenerateEncodingTable( UseSecureTyping );
					kbd->SetEncodingTable( UseSecureTyping, setQ );
				kbd->noSemaphore = false;
				}
				else
				{
					UseSecureTyping = false;
				}
			}

			break;
		}

		case WM_SETFOCUS:
			ODSF( L"WM_SETFOCUS:" );
			if (!kbd->SetEncodingTable( UseSecureTyping, setQ ))
			{	/* send NULL string to delete all tables, because problems occured */
				kbd->ResetEncodingTable( UseSecureTyping, setQ );
				kbd->FreeTables( UseSecureTyping );
				UseSecureTyping = false;
				if ( !kbd->wasError )
				{
					kbd->wasError = true;
					MessageBox( GetParent(hwnd), "Error in setting encoding table", "KeyGen - editSubclassProc", MB_ICONSTOP );
				}
				EndDialog( GetParent(hwnd), ERROR_INTERNAL_PROBLEM );
			}
			break;

		case WM_KILLFOCUS:
			ODSF( L"WM_KILLFOCUS:" );
			if (!kbd->ResetEncodingTable( TRUE, setQ ))
			{	/* send NULL string to delete all tables, because problems occured */
				kbd->FreeTables( TRUE );
			}
			break;
		
		case WM_KEYDOWN:
			virtualKey = (UINT)wParam;
			scanCode = (UINT)((lParam >> 16) & 0xFF);

			kbd->Wm_KeyDown( virtualKey, UseSecureTyping, scanCode );
			kbd->IncrementCharacterOrderCNTRL( (int)wParam, UseSecureTyping );
			break;
		
		case WM_DEADCHAR:
			kbd->Wm_DeadChar();
			break;

		case WM_CHAR:
			{
				DWORD retCharsNum;
				char  retChar1, retChar2;

				letter = (char)wParam;

				if (kbd->Wm_Char( letter, UseSecureTyping, &retCharsNum, &retChar1, &retChar2 ))
				{
					if (retCharsNum == 2)
					{
						wParam = (WPARAM)retChar2 & 0xFF;
						CallWindowProc(wpEditOrigWndProc, hwnd, uMsg, wParam, lParam);
						wParam = (WPARAM)retChar1 & 0xFF;
					}
					else
					{
						wParam = (WPARAM)retChar2 & 0xFF;
					}
				}

				break;
			}

		case WM_KEYUP:
			kbd->IncrementCharacterOrder( (int)wParam, UseSecureTyping );
			break;
	}

	return CallWindowProc(wpEditOrigWndProc, hwnd, uMsg, wParam, lParam);
}

//
//	Find SHA-256 in translated string and change to real name.
//
static void updateKGName( HWND hDlg )
{
#ifndef IDC_stFirst
	return;
#else
	HWND hWnd = GetDlgItem( hDlg, IDC_stFirst );

	if ( hWnd == NULL )	return;

	cWCStr wcsIn( 1024 );

	if ( wcsIn.m_szBody == NULL )
		return;

	GetWindowTextT( hWnd, wcsIn.m_szBody, wcsIn.m_len );

	const wchar_t *nameToFind = L"SHA-256";

	wchar_t *ptr = wcsstr( wcsIn.m_szBody, nameToFind );

	if ( ptr == NULL )
		return;

	int pos = (int)(ptr - wcsIn.m_szBody);

	cWCStr wcsOut;
	wcsOut.Add( wcsIn.m_szBody, pos );

	pos += wcslen( nameToFind );

	const wchar_t *nameToChange = _LoadIDST( IDS_kgName, false ); 

	if ( nameToChange == NULL ) return;
	
	int len = wcslen( nameToChange );

	if ( len == 0 ) return;

	wcsOut.Add( nameToChange );
	wcsOut.Add( wcsIn.m_szBody + pos );

	SetWindowTextT( hWnd, wcsOut.m_szBody );
#endif
}

BOOL CALLBACK getAdvancedFlagsDlgProc( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	HWND  editHwnd;
 
	switch (uMsg) 
	{
        case WM_INITDIALOG:
		{
			ODSFuncS( L"WM_INITDIALOG" );

			if ( pTranslateDll != NULL )
				pTranslateDll->TranslateDialog( hwnd, false );

			updateKGName( hwnd );

			if ( GetParent( hwnd ) == NULL ) CenterPosition( hwnd );

			Button_SetCheck( GetDlgItem(hwnd, IDC_STYPING),  (KeyGenFlags & FLAG_ADVANCED_SECURE_TYPING) > 0);
			Button_SetCheck( GetDlgItem(hwnd, IDC_SHOWCHARS),(KeyGenFlags & FLAG_ADVANCED_SHOW_PWD_CHARS) > 0);

			editHwnd = GetDlgItem(hwnd, IDC_EDITSTRING); 

#ifdef STRICT
			wpEditOrigWndProc = (WNDPROC) SetWindowLong(editHwnd,GWL_WNDPROC, (LONG) editSubclassProcAdvanced);
#else
			wpEditOrigWndProc = (FARPROC) SetWindowLong(editHwnd,GWL_WNDPROC, (LONG) editSubclassProcAdvanced);
#endif

			kbdParams *kbd = kbdParams::Allocate();
			if ( kbd == NULL )	{ assert( false ); return false; }
			kbdParams::SetParams( hwnd, kbd );
			//	SetWindowLong( hwnd, DWL_USER, (LONG)kbd );

			return (kbd->GenerateEncodingTable( TRUE ));
		}

		case WM_DESTROY:
		{
			// Retrieve the handle of the edit control
			editHwnd = GetDlgItem(hwnd, IDC_EDITSTRING);

			// Remove the subclass from the edit control.
			SetWindowLong(editHwnd, GWL_WNDPROC, (LONG) wpEditOrigWndProc); 

			kbdParams *kbd = kbdParams::GetParams( hwnd );
			if ( kbd == NULL ) return 0;
			kbdParams::Free( kbd );

			break;
		}
		case WM_COMMAND: 
		{
			kbdParams *kbd = kbdParams::GetParams( hwnd );
			if ( kbd == NULL ) return 0;

			switch(wParam) 
			{ 
				case IDOK: 
					kbd->FreeTables( UseSecureTyping );
 
			        if ( Button_GetCheck( GetDlgItem(hwnd, IDC_STYPING)))
						 KeyGenFlags |= FLAG_ADVANCED_SECURE_TYPING;
					else KeyGenFlags &= ~FLAG_ADVANCED_SECURE_TYPING;

			        if ( Button_GetCheck( GetDlgItem(hwnd, IDC_SHOWCHARS)))
						 KeyGenFlags |= FLAG_ADVANCED_SHOW_PWD_CHARS;
					else KeyGenFlags &= ~FLAG_ADVANCED_SHOW_PWD_CHARS;

					SetDlgItemText( hwnd, IDC_EDITSTRING, "qwerty");
					EndDialog(hwnd, ERROR_NO);
					return TRUE; 
 
				case IDCANCEL: 
					SetDlgItemText( hwnd, IDC_EDITSTRING, "qwerty");
					EndDialog(hwnd, (int)ERROR_CANCELLED_BY_USER); 
					kbd->FreeTables( UseSecureTyping );
					return TRUE; 
			} 
            return 0; 
		}
    } 
    return FALSE; 

	UNREFERENCED_PARAMETER(lParam);
}

BOOL convertTextToFilenamePath( const wchar_t *Text )
{
	// input Text has a form: "Container:  D:\path\container_name.jbc"
	// we return: ContainerName = "container_name.jbc", 
	//            ContainerPath = "D:\path\"

	int  i,j, pathLen, nameLen;
	const wchar_t *pathPtr = NULL, *namePtr = NULL;

	if (!Text) return FALSE;

	for( i=(int)wcslen(Text); i>0; i--)
	{
		if ((Text[i] == L' ') && (Text[i-1] == L':'))
		{
			pathPtr = &(Text[i]);
			j = i;
			while (Text[j++] == L' ') pathPtr = &(Text[j]);
			break;
		}
	}

	if (!pathPtr) pathPtr = Text;

	for( i=(int)wcslen(Text); i>0; i--)
	{
		if (Text[i] == L'\\')
		{
			namePtr = &(Text[i+1]);
			break;
		}
	}

	if (!namePtr) namePtr = Text;

	pathLen = (int)wcslen(pathPtr);
	nameLen = (int)wcslen(namePtr);

	if (pathLen > 128) pathLen = 128;
	if (nameLen > 128) nameLen = 128;

	wcsncpy_s( FileName, namePtr, nameLen );
	FileName[127] = L'\0';

	wcsncpy_s( PathName, pathPtr, pathLen );
	PathName[127] = L'\0';

	for( i= (int)wcslen(PathName); i>0; i--)
	{
		if (PathName[i] == L'\\')
		{
			PathName[i+1] = L'\0';
			break;
		}
	}

	return TRUE;
}
