/*++
Copyright:      (c) 1997-1998 Jetico, Inc.
Module Name:	bc_alg.h
Abstract:       This file includes data declarations for common code 
                for all encryption algorithm drivers in Windows NT.
Author:         Sergey A. Frolov
Environment:    Kernel mode only.
Notes:
Revision History:
--*/

#ifndef __BC_ALG__
#define __BC_ALG__

//#define DBG 1

// XTS Key2 starts from 32nd byte in a whole Key array ( Key = Key1 | Key2 )
#define XTS_KEY2_OFFSET 32 

typedef struct  _BC_ALG_EXTENSION
  { PDEVICE_OBJECT  DeviceObject;
    UNICODE_STRING  deviceLinkUString;
  } BCALG_EXTENSION, *PBCALG_EXTENSION;


typedef struct _KEY_ELEMENT 
  { 
       UCHAR *keyExtended;
       UCHAR *pool;
       UCHAR *lrwContext;
       UCHAR *xtsContext;
  } KEY_ELEMENT, *PKEY_ELEMENT;


#define DISPLAY_NAME_LENGTH 32
#define SERVICE_NAME_LENGTH 32

#define EYE_CATCH_STRING    "__BCALGDRV__"
#define EYE_CATCH_STRING_1  "__BCB_SIZE__"


// keyLengthLrw = keyLength + (key length required for LRW mode)
// For example, Blowfish-448 uses 448-bit key + 64 bit for LRW mode
// hence, keyLengthLrw must be 512 bites
//
// If keyLengthLrw == 0 or sizeof(BC_ALGID_BLOCK) does not include 
// keyLengthLrw field, hence, the algorithm module does not support LRW mode.

typedef struct _BC_ALGID_BLOCK
  { char  eyeCatchString[16];
    DWORD algorithmID;
    DWORD keyLength;
    char  displayName[ DISPLAY_NAME_LENGTH ];
    char  serviceName[ SERVICE_NAME_LENGTH ];
    char  eyeCatchString_1[16];
    DWORD structSize;
    DWORD blockSize;
    DWORD keyLengthLrw;
  } BC_ALGID_BLOCK, *PBC_ALGID_BLOCK;


#if DBG
#define BCDBUGCHECK            ((ULONG)0x80000000)
#define BCDDIAG1               ((ULONG)0x00000001)
#define BCDDIAG2               ((ULONG)0x00000002)
#define BCDERRORS              ((ULONG)0x00000004)

#define BCalgDump(LEVEL, STRING) \
        do { \
            if (BCDiskDebugLevel & LEVEL) { \
                DbgPrint STRING; \
            } \
            if (LEVEL == BCDBUGCHECK) { \
                ASSERT(FALSE); \
            } \
        } while (0)
#else
#define BCalgDump(LEVEL,STRING) do {NOTHING;} while (0)
#endif

int
sprintf(char *s, const char *format, ...);

//
// Device driver routine declarations.
//

NTSTATUS
DriverEntry(
    IN OUT PDRIVER_OBJECT   DriverObject,
    IN PUNICODE_STRING      RegistryPath
    );

NTSTATUS
BCalgInitializeDriver(
    IN PDRIVER_OBJECT       DriverObject
    );

NTSTATUS
BCalgCreateClose(
    IN PDEVICE_OBJECT       DeviceObject,
    IN PIRP                 Irp
    );

NTSTATUS
BCalgDeviceControl(
    IN PDEVICE_OBJECT       DeviceObject,
    IN PIRP                 Irp
    );

VOID
BCalgUnloadDriver(
    IN PDRIVER_OBJECT       DriverObject
    );

NTSTATUS
BCalgCreateKey(IN PCREATEKEY_BUFFER CkBuffer);

BOOL 
BC_ExtendKey( PUCHAR KeyFrom, PUCHAR KeyTo );


NTSTATUS 
BC_TestKey( KEY_HANDLE KeyHandle );

NTSTATUS 
BC_FreeKey( KEY_HANDLE KeyHandle );

NTSTATUS 
BC_Encrypt( KEY_HANDLE KeyHandle,
            PUCHAR     IVector,
            PUCHAR     InBuffer,
            PUCHAR     OutBuffer,
            DWORD      Length );       // in bytes

NTSTATUS 
BC_Decrypt( KEY_HANDLE KeyHandle,
            PUCHAR     IVector,
            PUCHAR     InBuffer,
            PUCHAR     OutBuffer,
            DWORD      Length );       // in bytes


// EX_ENCRYPT flags are defined in bc_ioctl.h
// #define EX_ENCRYPT_LRW_MODE_FLAG  0x00000001
// #define EX_ENCRYPT_CBC_MODE_FLAG  0x00000002

NTSTATUS 
BC_EncryptEx( KEY_HANDLE KeyHandle,
              PUCHAR     IVector,
              PUCHAR     InBuffer,
              PUCHAR     OutBuffer,
              DWORD      Length,       // in bytes
              DWORD      ExEncryptFlags);

NTSTATUS 
BC_DecryptEx( KEY_HANDLE KeyHandle,
              PUCHAR     IVector,
              PUCHAR     InBuffer,
              PUCHAR     OutBuffer,
              DWORD      Length,       // in bytes
              DWORD      ExEncryptFlags);




#endif
