#include <windows.h>

#ifndef __WIN95__
#include <windef.h>
#else
#include <basedef.h>
#endif

#include <stdio.h>
#include <string.h>
#include "..\bc_rc6.h"

/* RC6 test vectors from the RC6 paper. There are only two of them since the
   rest were made for keylengths different from 256-bits */
struct
{
	unsigned char key[32];
	unsigned char plain[16];
	unsigned char cipher[16];
	int length;
} testVectors[] =
{
	{
		{	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
			0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
			0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
			0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
		{	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
			0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
		{	0x8F, 0x5F, 0xBD, 0x05, 0x10, 0xD1, 0x5F, 0xA8,
			0x93, 0xFA, 0x3F, 0xDA, 0x6E, 0x85, 0x7E, 0xC2 }, 16
	},
	{
		{	0x01, 0x23, 0x45, 0x67, 0x89, 0xAB, 0xCD, 0xEF,
			0x01, 0x12, 0x23, 0x34, 0x45, 0x56, 0x67, 0x78,
			0x89, 0x9A, 0xAB, 0xBC, 0xCD, 0xDE, 0xEF, 0xF0,
			0x10, 0x32, 0x54, 0x76, 0x98, 0xBA, 0xDC, 0xFE },
		{	0x02, 0x13, 0x24, 0x35, 0x46, 0x57, 0x68, 0x79,
			0x8A, 0x9B, 0xAC, 0xBD, 0xCE, 0xDF, 0xE0, 0xF1 },
		{	0xC8, 0x24, 0x18, 0x16, 0xF0, 0xD7, 0xE4, 0x89,
			0x20, 0xAD, 0x16, 0xA1, 0x67, 0x4E, 0x5D, 0x48 }, 16
	},
	{
		{	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
			0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
			0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
			0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
		{	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
			0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
		{	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
			0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, 0
	}
};

int main() {
	int i;
	DWORD outputData[4];
	DWORD extendedKey[KEY_LENGTH_EXTENDED / 32];

	for (i=0; testVectors[i].length != 0; i++)
	{
		KeyExtend(testVectors[i].key, extendedKey);
	
		Test_RC6(extendedKey, (DWORD *)testVectors[i].plain, outputData, TRUE);
	
		if (memcmp(outputData, testVectors[i].cipher, testVectors[i].length) != 0)
		{
			printf("\n(ERROR) Test Vector %d: the RC6 implementation failed the test (encryption)!\n",i);
			return 0;
		}

		Test_RC6(extendedKey, (DWORD *)testVectors[i].cipher, outputData, FALSE);
	
		if (memcmp(outputData, testVectors[i].plain, testVectors[i].length) != 0)
		{
			printf("\n(ERROR) Test Vector %d: the RC6 implementation failed the test (decryption)!\n",i);
			return 0;
		}

		printf("Test Vector %d: the RC6 implementation passes the test successfully.\n",i);
	}
	return 0;
}
