#include <windows.h>

#ifndef __WIN95__
#include <windef.h>
#else
#include <basedef.h>
#endif

#include <stdio.h>
#include <string.h>
#include "..\bc_rijn.h"
#include "..\aes.h"

/* Rijndael test vectors at the 'vectors.txt' file, got from reference sources of 
   the developers of the Rijndael algorithm Joan Daemen and Vincent Rijmen,
   submitted to the National Institute of Standards and Technology (NIST) during 
   the Advanced Encryption Standard (AES) development.
 */
typedef struct 
    {    UCHAR key[ KEY_LENGTH / 8 ];
         UCHAR plain[16];
         UCHAR cipher[16];
		 DWORD dataLength;
		 int   ref_number;
	} TEST_VECTOR;


static unsigned char char2byte( const char *ch )
{
	char str[] = { ' ',' ', 0, 0 };
	unsigned char res;

	if (!ch) return 0;
	str[0] = ch[0]; str[1] = ch[1];
	sscanf(str,"%x",&res);
	return res;
}

static BOOL getTestVector( FILE *f, TEST_VECTOR *Vector )
{
	char str[128];
	DWORD  i,j;

	if (!f) return FALSE;

	Vector->dataLength = BLOCK_SIZE / 8;

	for(i=0; i<4; )
	{
		if (fscanf( f, "%s", str ) != 1) return FALSE;

		switch(str[0])
		{
			case 'I':
				if (sscanf( str, "I=%d", &(Vector->ref_number)) != 1) return FALSE;
				i++;
				break;

			case 'K':
				if (sscanf( str, "KEY=%s", str ) != 1) return FALSE;
				for (j=0; (j<(KEY_LENGTH/8)) && (j*2 < strlen(str)); j++)
				{
					Vector->key[j] = char2byte( &str[j*2] );
				}
				i++;
				break;

			case 'P':
				if (sscanf( str, "PT=%s", str ) != 1) return FALSE;
				for (j=0; (j<(BLOCK_SIZE/8)) && (j*2 < strlen(str)); j++)
				{
					Vector->plain[j] = char2byte( &str[j*2] );
				}
				i++;
				break;
			
			case 'C':
				if (sscanf( str, "CT=%s", str ) != 1) return FALSE;
				for (j=0; (j<(BLOCK_SIZE/8)) && (j*2 < strlen(str)); j++)
				{
					Vector->cipher[j] = char2byte( &str[j*2] );
				}
				i++;
				break;
		}
	}
	if (i==4) return TRUE; // all 4 test vectors elements were read (number, key, plain and cipher text)

	return FALSE;
}


int main()
{
  int   i;
  UCHAR outputData[16];
  BOOL  result;
  BC_AES_STRUCT extendedKey;
  FILE *f = NULL;
  TEST_VECTOR testVector;

  f = fopen("vectors.txt", "r");
  if (!f)
  {
	  printf("Error: Cannot open file with test vectors 'vectors.txt'.\n");
	  goto exit;
  }

  result = TRUE; // Assume everything will be Ok

  for (i=1; getTestVector( f, &testVector ) ; i++)
  {
    if (!KeyExtend( testVector.key, (DWORD *)(&extendedKey) ))
	{
		printf("\n(ERROR) Test Vector N %d: error to extend encryption key!\n",i);
		goto exit;
    }		
	  
	if (encrypt( testVector.plain, outputData, (const aes *)&extendedKey) != aes_good)
	{
		printf("\n(ERROR) Test Vector N %d: internal error (encryption)!\n",i);
		goto exit;
    }		

    if( memcmp( outputData, testVector.cipher, testVector.dataLength ) != 0)
	{
		printf("\n(ERROR) Test Vector N %d: the Rijndael implementation failed the test (encryption)!\n",i);
		goto exit;
	}

	if (decrypt( testVector.cipher, outputData, (const aes *)&extendedKey) != aes_good)
	{
		printf("\n(ERROR) Test Vector N %d: internal error (decryption)!\n",i);
		goto exit;
    }		

    if( memcmp( outputData, testVector.plain, testVector.dataLength ) != 0)
	{
		printf("\n(ERROR) Test Vector N %d: the  Rijndael implementation failed the test (decryption)!\n",i);
		goto exit;
	}

    printf("Test Vector N %3d: the  Rijndael implementation passes the test successfully.\n",i+1);
  }

exit:
  if (f) fclose(f);
  return 0;
}

