#include "UnicodeFuncs.h"
#include <io.h>
#include <assert.h>
#include <stdio.h>
#include <time.h>
#include "ODS.h"
#include "FormatT.h"

#ifdef _WIN32

#include <process.h>
#include <windowsx.h>
#include "token_info.h"
#include "CDll.h"
#include "CSysDll.h"
#include "shell32dll.h"
#include <wchar.h>
#include "cFindFile.h"
#include "cStreams.h"
#include "cMBStr.h"
#include "cAutoFile.h"
#include "kernelDll.h"
#include "cLongName.h"

#ifndef _NoTranslation_
#include "translateDll.h"
#endif

#ifndef _CRT_SECURE_NO_WARNINGS
#pragma warning(disable : 4996)
#endif

HINSTANCE cInstanceHolder::m_hInst = NULL;
HWND cInstanceHolder::m_hMainWnd = NULL;

DWORDLONG mSecClock() { return ( ((DWORDLONG)clock() * 1000 ) / CLOCKS_PER_SEC ); }

int bitcount(unsigned int a)
{
   a = (a & 0x55555555) + ((a >>  1) & 0x55555555); // max 2
   a = (a & 0x33333333) + ((a >>  2) & 0x33333333); // max 4
   a = (a + (a >> 4)) & 0x0f0f0f0f; // max 8 per 4, now 8 bits
   a = (a + (a >> 8)); // max 16 per 8 bits
   a = (a + (a >> 16)); // max 32 per 8 bits
   return a & 0xff;
}

//static
void cInstanceHolder::SetHolder( HINSTANCE hInst, HWND hWnd )
{
	if ( hInst != NULL ) // static m_hInst may be inited early.
		m_hInst = hInst;
	
	if ( hWnd != NULL )
		m_hMainWnd = hWnd;
}


//static
const wchar_t *cInstanceHolder::GetOSVersionStr()
{
	static cWCStr wcs;

	if ( wcs.m_len == 0 )
	{
		OSVERSIONINFO osvi;
		osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
		GetVersionEx ( &osvi );
		wcs.FormatT( L"OS: PlatformId %X, Major %d, Minor %d, Build %d ", osvi.dwPlatformId, osvi.dwMajorVersion, osvi.dwMinorVersion, osvi.dwBuildNumber );
	}

	return wcs.m_szBody;
}

//static 
bool cInstanceHolder::isWindows2K()
{
	OSVERSIONINFO osvi;
	osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
	GetVersionEx ( &osvi );
	bool bIsWindows2KorLater = (osvi.dwMajorVersion >= 5);
	return bIsWindows2KorLater;
}

//static 
bool cInstanceHolder::isWindowsXP()
{
	OSVERSIONINFO osvi;
	osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
	GetVersionEx ( &osvi );
	bool bIsWindowsXPorLater = 
	   ( (osvi.dwMajorVersion > 5) ||
	   ( (osvi.dwMajorVersion == 5) && (osvi.dwMinorVersion >= 1) ) );
	return bIsWindowsXPorLater;
}

//static 
bool cInstanceHolder::isWindowsVista()
{
	OSVERSIONINFO osvi;
	osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
	GetVersionEx ( &osvi );
	bool bIsWindowsVista = ( osvi.dwPlatformId == VER_PLATFORM_WIN32_NT ) && ( osvi.dwMajorVersion >= 6 );
	return bIsWindowsVista;
}

//static 
bool cInstanceHolder::isHomeVersion()
{
	OSVERSIONINFOEX osvi;
	osvi.dwOSVersionInfoSize = sizeof( osvi );
	GetVersionEx( (OSVERSIONINFO*)&osvi );
	bool bIsWindowsXPorLater = 
	   ( (osvi.dwMajorVersion > 5) ||
	   ( (osvi.dwMajorVersion == 5) && (osvi.dwMinorVersion >= 1) ) );

	if ( !bIsWindowsXPorLater )
		return false;

	bool bHome = ( ( osvi.wSuiteMask & VER_SUITE_PERSONAL ) != 0 );

	return bHome;
}

//static 
bool cInstanceHolder::isWindows98()
{
	OSVERSIONINFO vi;
	vi.dwOSVersionInfoSize = sizeof( vi );
	GetVersionEx( &vi );
/*
	dwMajorVersion	0x00000004
	dwMinorVersion	0x0000000a
	dwBuildNumber	0x040a07ce
	dwPlatformId	0x00000001
*/
	if ( (vi.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS) 
		&& ( vi.dwMajorVersion	>= 0x00000004 )
		&& ( vi.dwMinorVersion	>= 0x0000000a )
		&& ( vi.dwBuildNumber >= 0x040a07ce )
		)
		return true;

	return false;
}

//static 
bool cInstanceHolder::isWindowsME()
{
/*
dwMajorVersion = 0x00000004
dwMinorVersion = 0x0000005A
dwBuildNumber  = 0x045A0BB8
dwPlatformId   = 0x00000001
szCSDVersion   =  
*/
	OSVERSIONINFO vi;
	vi.dwOSVersionInfoSize = sizeof( vi );
	GetVersionEx( &vi );

	if ( (vi.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS) 
		&& ( vi.dwMajorVersion	>= 0x00000004 )
		&& ( vi.dwMinorVersion	>= 0x0000005a )
		&& ( vi.dwBuildNumber >= 0x045A0BB8 )
		)
		return true;

	return false;
}

//static 
bool cInstanceHolder::isW7()
{
	OSVERSIONINFO osvi;
	osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
	GetVersionEx ( &osvi );
	if ( osvi.dwPlatformId != VER_PLATFORM_WIN32_NT )
		return false;

	// Windows 7 beta returns 6.1

	if ( osvi.dwMajorVersion > 6 )
		return true;
	else
	if ( osvi.dwMajorVersion < 6 )
		return false;
	else // osvi.dwMajorVersion == 6
	if ( osvi.dwMinorVersion >= 1 )
		return true;

	return false;
}

//
//	IsWow64Process
//		returns true if 32-bit application is runned in 64-bit environment.
//
bool cInstanceHolder::IsWow64Process()
{
	static bool bIsWow64 = cKernelDll::getStatic().IsWow64Process();
	return bIsWow64;
}

const DWORD W10BuildNumber = 0x23f0;

//
//	Windows 10 with manifest
//<compatibility xmlns="urn:schemas-microsoft-com:compatibility.v1"> 
//    <application> 
//        <!-- Windows 10 --> 
//        <supportedOS Id="{8e0f7a12-bfb3-4fe8-b9a5-48fd50a15a9a}"/>
//        <!-- Windows 8.1 -->
//        <supportedOS Id="{1f676c76-80e1-4239-95bb-83d0f6d0da78}"/>
//        <!-- Windows Vista -->
//        <supportedOS Id="{e2011457-1546-43c5-a5fe-008deee3d3f0}"/> 
//        <!-- Windows 7 -->
//        <supportedOS Id="{35138b9a-5d96-4fbd-8e2d-a2440225f93a}"/>
//        <!-- Windows 8 -->
//        <supportedOS Id="{4a2f28e3-53b9-4441-ba9c-d69d4a4a6e38}"/>
//    </application> 
// OR
//  <compatibility xmlns="urn:schemas-microsoft-com:compatibility.v1">
//    <application>
//
//      <supportedOS Id="{8e0f7a12-bfb3-4fe8-b9a5-48fd50a15a9a}"></supportedOS>
//
//      <supportedOS Id="{1f676c76-80e1-4239-95bb-83d0f6d0da78}"></supportedOS>
//
//      <supportedOS Id="{e2011457-1546-43c5-a5fe-008deee3d3f0}"></supportedOS>
//
//      <supportedOS Id="{35138b9a-5d96-4fbd-8e2d-a2440225f93a}"></supportedOS>
//
//      <supportedOS Id="{4a2f28e3-53b9-4441-ba9c-d69d4a4a6e38}"></supportedOS>
//    </application>
//  </compatibility>
//
//	Reports following:
//
//	dwMajorVersion = 0x0000000A
//	dwMinorVersion = 0x00000000
//	dwBuildNumber  = 0x0000295A
//	dwPlatformId   = 0x00000002
//	szCSDVersion   = 
//	wSPMajor		= 0x0000
//	wSPMinor		= 0x0000
//	wSuiteMask		= 0x0100
//	wProductType   = 0x1
//

static bool findInMemory( const BYTE *buff, DWORD size, const char *str )
{
	DWORD checkLength = (DWORD)strlen( str );

	bool ok = FALSE;

	DWORD position = 0;

	for( int checkCount = 0 ;; (position)++ )
	{
		if ( position + checkLength > size )
			return false;

		const char *s = ( const char * )(buff + position);

		if ( memcmp( s, str, checkLength ) == 0 )
			return true;

	}

	return false;
}

static bool hasManifestString( const char *id )
{
	char str[ 256 ];
	sprintf_s( str, "<supportedOS Id=\"{%s}\"", id );

	wchar_t path[ 2 * MAX_PATH ];
	GetModuleFileNameT( NULL, path, MAX_PATH );

	cWCStr wcsManifest;
	wcsManifest.FormatT( L"%s.manifest", path );

	const wchar_t *lPath = path;

	if ( FileExistsT( wcsManifest.m_szBody ) )
		lPath = wcsManifest.m_szBody;

	HANDLE hFile = CreateFileT( lPath, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL  );
	if ( hFile == NULL )
		return false;

	DWORDLONG dwlSize = GetFileSizeT( hFile );

	const DWORD size1M = 1024 * 1024;
	DWORD btr = size1M;

	if ( dwlSize > size1M )
		SetFilePointerT( hFile, dwlSize - size1M, FILE_BEGIN );
	else
		btr = DWORD( dwlSize );

	BYTE *bf = new BYTE[ btr ];
	if ( bf == NULL )
	{
		CloseHandle( hFile );
		return false;
	}

	DWORD br = 0;
	bool ok = ReadFile( hFile, bf, btr, &br, NULL ) && ( btr == br );

	if ( ok )
		ok = findInMemory( bf, btr, str );

	CloseHandle( hFile );
	delete [] bf;

	return ok;
}

//static 
bool cInstanceHolder::isW10( bool bOrLater )
{
	OSVERSIONINFOEX osvi;
	osvi.dwOSVersionInfoSize = sizeof(osvi);
	GetVersionEx ( (OSVERSIONINFO*)&osvi );
	if ( osvi.dwPlatformId != VER_PLATFORM_WIN32_NT )
		return false;

	// Windows 8 and server 2012 returns 6.2
	// The difference is in W8 -> OSVERSIONINFOEX.wProductType == VER_NT_WORKSTATION 

	static bool bHasManifest = hasManifestString( "8e0f7a12-bfb3-4fe8-b9a5-48fd50a15a9a" );

	if ( bHasManifest )
	{
		if ( osvi.dwMajorVersion == 10 ) // Manifest version reports, see above.
			return true;
		else
		if ( osvi.dwMajorVersion > 10 )
			return bOrLater;

		return false;
	}

	if ( osvi.dwMajorVersion > 6 )
		return bOrLater;
	else
	if ( osvi.dwMajorVersion < 6 )
		return false;
	else // osvi.dwMajorVersion == 6
	if ( osvi.dwMinorVersion == 2 )
	{
		if ( osvi.dwBuildNumber < W10BuildNumber ) // Windows 8
			return false;
		else
		if ( osvi.dwBuildNumber == W10BuildNumber )
			return true;

		return bOrLater;
	}
	else
	if ( osvi.dwMinorVersion > 2 )
		return bOrLater;

	return false;
}

//
// W8 with manifest:
//
//	dwMajorVersion = 0x00000006
//	dwMinorVersion = 0x00000003
//	dwBuildNumber  = 0x00002580
//	dwPlatformId   = 0x00000002
//	szCSDVersion   =
//	wSPMajor                = 0x0000
//	wSPMinor                = 0x0000
//	wSuiteMask              = 0x0100
//	wProductType   = 0x1
//

//static 
bool cInstanceHolder::isW8( bool bOrLater )
{
	OSVERSIONINFO osvi;
	osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
	GetVersionEx ( &osvi );
	if ( osvi.dwPlatformId != VER_PLATFORM_WIN32_NT )
		return false;

	// Windows 8 and server 2012 returns 6.2
	// The difference is in W8 -> OSVERSIONINFOEX.wProductType == VER_NT_WORKSTATION 

	//  <!-- Windows 8.1 -->
	//  <supportedOS Id="{1f676c76-80e1-4239-95bb-83d0f6d0da78}"/>
	//	With manifest returns 6.3 build 0x00002580
	static bool bHasManifest = hasManifestString( "1f676c76-80e1-4239-95bb-83d0f6d0da78" );

	if ( osvi.dwMajorVersion > 6 )
		return bOrLater;
	else
	if ( osvi.dwMajorVersion < 6 )
		return false;
	else // osvi.dwMajorVersion == 6
	if ( osvi.dwMinorVersion == 2 )
	{
		if ( bHasManifest )
			return false;

		if ( osvi.dwBuildNumber < W10BuildNumber )
			return true;

		return bOrLater;
	}
	else
	if ( osvi.dwMinorVersion == 3 ) // Manifest version reports, see above.
	{
		if ( !bHasManifest )
			return bOrLater;

		if ( osvi.dwBuildNumber < 0x00002580 )
			return false;
		else
		if ( osvi.dwBuildNumber == 0x00002580 )
			return true;

		return bOrLater;
	}
	else
	if ( osvi.dwMinorVersion > 2 )
		return bOrLater;

	return false;
}

void cInstanceHolder::BeginCursor( ULONG_PTR idc )
{
	m_hSaveCursor = SetCursor( LoadCursor( NULL, MAKEINTRESOURCE(idc) ) );
}

void cInstanceHolder::EndCursor()
{
	SetCursor( m_hSaveCursor );
}

void cInstanceHolder::WaitLoop( DWORD mSec )
{
	DWORDLONG finTime = mSecClock() + mSec;

	HWND hWnd = NULL;

	if ( !IsWindow( m_hMainWnd ) )
		ODS_fls( L"Warning! Handle %X is not window. PeekMessage retrieves messages for any window that belongs to the current thread." );
	else
		hWnd = m_hMainWnd;

	while( true )
	{	
		MSG msg;

		while ( PeekMessage( &msg, hWnd,  0, 0, PM_REMOVE ) ) 
		{
			TranslateMessage( &msg ); 
			DispatchMessageT( &msg ); 
			
			if ( mSecClock() >= finTime )
				break;
		}

		if ( mSecClock() >= finTime )
			break;

		Sleep( 1 );
	}
}

class cUpdateWinVersion
{
public:
	OSVERSIONINFO vi;

	bool m_isWinNT;
	bool m_isWinXP;
	bool m_isWinVista;

	cUpdateWinVersion()
	{
		memset( &vi, 0, sizeof( vi ));
		vi.dwOSVersionInfoSize = sizeof( vi );
		GetVersionEx(&vi);

		m_isWinNT =( vi.dwPlatformId == VER_PLATFORM_WIN32_NT  );

		m_isWinXP = (
						( vi.dwPlatformId == VER_PLATFORM_WIN32_NT  )
						&& 
						(
							( vi.dwMajorVersion > 5 ) ||
							( vi.dwMajorVersion == 5 ) && ( vi.dwMinorVersion >= 1 )
						)
					);

		m_isWinVista = (
						( vi.dwPlatformId == VER_PLATFORM_WIN32_NT  )
						&& 
						( vi.dwMajorVersion > 6 )
					);
	}
};

bool isWinXP()
{
	static cUpdateWinVersion updVer;
	return updVer.m_isWinXP;
}

bool isWinVista()
{
	static cUpdateWinVersion updVer;
	return updVer.m_isWinVista;
}


UINT getDefaultCodePage()
{
	static cUpdateWinVersion windowsVersion;

	if ( ( windowsVersion.vi.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS  ) &&
		 ( windowsVersion.vi.dwMajorVersion == 4  ) &&
		 ( windowsVersion.vi.dwMinorVersion <= 10 ) )
		return CP_ACP;
	else
		return CP_UTF8;
}

//
//	convertToWchar
//		Converts the buffer from single byte ANSI char to wide char.
//		Reasons:
//			1. Windows 95 RegQueryValueExA returns ANSI buffer instead of RegQueryValueExT requires
//			2. Menu functions uses same structures excluding pointers to text buffers, so we may use the funcion for conversion
//			3. Some windowsx function uses same structures for ANSI and Unicode but the difference is int char ot wchar_t buffers are filled
//				the function helps in the case too.
//	
//		Note, the function may be used for copying from byte buffer bf to wide char buffer ws.
//		If ws is NULL it converts chars to wchars by overwriting on the byte buffer.
//
static int convertToWchar(	char *bf, int bfSize, bool multyString = false )
{
	if ( bf == NULL )
		return 0;

	int wsMaxLen = bfSize / sizeof( wchar_t );
	if ( wsMaxLen == 0 )
		return 0;

	cWCStr wcs( bf, -1, multyString );

	wchar_t *ws = (wchar_t *)bf;

//	wcsncpy_s( ws, wsMaxLen, wcs.m_szBody, _TRUNCATE );
	wcsncpy( ws, wcs.m_szBody, wsMaxLen - 1 );

	return wcs.m_len * sizeof( wchar_t );
}

static UINT codePage = CP_ACP;

UINT setCodePage( UINT newCP )
{
	UINT oldCP = codePage;
	
	codePage = newCP;

	return oldCP;
}

static UINT getCodePage()
{
	// windows 98 allows UTF8 encoding

	if ( codePage == -1  )
	{
		codePage = getDefaultCodePage();
	}

	return codePage;
}

static int __wcstombs( char *aStr, int aStrSize, const wchar_t *wStr, int wStrLen = -1, UINT codePage = -1 )
{
	return WideCharToMultiByte( ( codePage == -1 ? getCodePage() : codePage ), 0, wStr, wStrLen,  aStr, aStrSize, NULL, NULL );
}
static int __mbstowcs( wchar_t *wcs, int wcsMaxLen, const char *mbs, int mbsLen = -1 , UINT codePage = -1 )
{
	int res = MultiByteToWideChar( ( codePage == -1 ? getCodePage() : codePage ), 0, mbs, mbsLen, wcs, wcsMaxLen );

	return res;
}

int _mbstowcs( wchar_t *wcs, int wcsLen, const char *mbs, int mbsLen, bool doubleZeroTerminated, UINT codePage )
{
	int mbsPos = 0;
	int wcsPos = 0;

	int result = 0;

	if ( ( wcs == 0 ) && ( mbsLen != 0 ) )
	{
		return MultiByteToWideChar( ( codePage == -1 ? getCodePage() : codePage ), 0, mbs, mbsLen, NULL, 0 );
	}
	else
	if ( wcsLen == 0 ) 
	{
		// simple gets required length
		while( true )
		{
			int mbsMax = -1;

			if ( mbsLen != -1 )
				mbsMax = mbsLen - mbsPos;

			result += __mbstowcs( NULL, 0, mbs + mbsPos, mbsMax, codePage );

			if ( !doubleZeroTerminated )
				break;

			if ( mbs[ mbsPos ] == '\0' )
				break;

			mbsPos += (int)strlen( mbs + mbsPos ) + 1;
		}

		return result;
	}
	
	while( true )
	{
		int mbsMax = -1;

		if ( mbsLen != -1 )
			mbsMax = mbsLen - mbsPos;

		int curLen = __mbstowcs( wcs + wcsPos, wcsLen - wcsPos, mbs + mbsPos, mbsMax, codePage );

		if ( curLen == 0 )
			return 0;

		result += curLen;

		if ( !doubleZeroTerminated )
			break;

		if ( mbs[ mbsPos ] == L'\0' )
			break;

		wcsPos += ( mbsMax != -1 ? mbsMax : (int)wcslen( wcs + wcsPos ) ) + 1;
		mbsPos += curLen + 1; // (int)strlen( mbs + mbsPos ) + 1;
	}

	return result;
}

int _wcstombs( char *mbs, int mbsLen, const wchar_t *wcs, int wcsLen, bool doubleZeroTerminated, UINT codePage )
{
//	assert( len <= MAX_PATH * 10 );

	int mbsPos = 0;
	int wcsPos = 0;

	int result = 0;

	if ( mbsLen == 0 ) 
	{
		// simple gets required length
		while( true )
		{
			int wcsMax = -1;
			if ( wcsLen != -1 )
				wcsMax = wcsLen - wcsPos;

			result += __wcstombs( NULL, 0, wcs + wcsPos, wcsMax, codePage );

			if ( !doubleZeroTerminated )
				break;

			if ( wcs[ wcsPos ] == L'\0' )
				break;

			wcsPos += (int)wcslen( wcs + wcsPos ) + 1;
		}

		return result;
	}

	
	while( true )
	{
		int wcsMax = -1;
		if ( wcsLen != -1 )
			wcsMax = wcsLen - wcsPos;

		int curResult = __wcstombs( mbs + mbsPos, mbsLen - mbsPos, wcs + wcsPos, wcsMax, codePage );

		if ( curResult == 0 )
			return 0;

		result += curResult;

		if ( !doubleZeroTerminated )
			break;

		if ( wcs[ wcsPos ] == L'\0' )
			break;

		wcsPos += (int)wcslen( wcs + wcsPos ) + 1;
		mbsPos += (int)strlen( mbs + mbsPos ) + 1;
	}

	return result;
}

const wchar_t *wcsMove( wchar_t *dst, const wchar_t *src, int len )
{
	if ( len <= 0 )
	{
		ODS_fls( _FormatT( L"Assertion! len %X", len ) );
	}
	else
		memmove( dst, src, len * sizeof( wchar_t ) );

	return dst;
}

int wcsLen( const wchar_t *wcs, int maxLen )
{
	if ( wcs == NULL )
		return 0;

	if ( maxLen == -1 )
		return (int)wcslen( wcs );

	int i;

	for( i = 0; i < maxLen; i++ )
	{
		if ( wcs[i] == L'\0' )
			break;
	}

	return i;
}

int strLen( const char *str, int maxLen )
{
	if ( str == NULL )
		return 0;

	if ( maxLen == -1 )
		return (int)strlen( str );

	int i;

	for( i = 0; i < maxLen; i++ )
	{
		if ( str[i] == '\0' )
			break;
	}

	return i;
}

const wchar_t *getWCStr( DWORD dwRes )			{ return GetIntText( dwRes, 16 ); }
const wchar_t *getWCStr( int iRes, int radix )	{ return GetIntText( iRes, radix ); }
const wchar_t *getWCStr( bool bRes )			{ return bRes ? L"True" : L"False"; }
const wchar_t *statusStr( bool bRes )			{ return bRes ? L"Success" : L"Failed"; }

const wchar_t *getWCStr( const char *aStr, wchar_t *wStr, int wStrMaxLen, UINT cp, int aStrLen )
{
	if ( wStr == NULL )
	{
		static cAutoDeallocate < wchar_t > ad;

		int maxStrLen = 2048;

		if ( ad.Allocate( maxStrLen ) == NULL ) return L"<NULL>";

		wStr = ad.m_ptr;
		wStrMaxLen = maxStrLen;
	}

	if ( aStr == NULL )
	{
		wStr[ 0 ] = L'\0';
		return NULL;
	}

	int res = MultiByteToWideChar( cp, 0, aStr, aStrLen, wStr, wStrMaxLen );

	int wLast = min( res, wStrMaxLen - 1 );
	
	wStr[ wLast ] = L'\0';

	return wStr;
}

const char *getMBStr( const wchar_t *wStr, char *aStr, int aStrMaxLen, UINT cp )
{
	if ( aStr == NULL )
	{
		static cAutoDeallocate < char > ad;
		int maxStrLen = 2048;
		if ( ad.Allocate( maxStrLen ) == NULL ) return "<NULL>";
		aStr = ad.m_ptr;
		aStrMaxLen = maxStrLen;
	}

	if ( wStr == NULL )
	{
		aStr[ 0 ] = '\0';
		return NULL;
	}

	int res = WideCharToMultiByte( cp, 0, wStr, -1, aStr, aStrMaxLen, NULL, NULL );

	while( res < aStrMaxLen )
		aStr[res++] = '\0';

	return aStr;
}

cWCStr::cWCStr( const char *mbs, int len, bool doubleZeroTerminated, UINT codePage )
{
	//m_szBody = NULL;
	//m_len = 0;
	//m_allocatedLen = 0;

	if ( len == -1 )
	{
		if ( mbs == NULL )
			return;

		len = 0;

		while( true )
		{
			len += (int)strlen( mbs + len );

			if ( !doubleZeroTerminated )
				break;

			len++;

			if ( mbs[len] == '\0' )
				break;
		}
	}

	int lenNew = 0;
	
	if ( len > 0 )
		lenNew = _mbstowcs( NULL, 0, mbs, len, doubleZeroTerminated, codePage );

	if ( m_allocatedLen < lenNew + 1 )
	{
		freeStr( m_szBody, &m_allocatedLen );
		m_szBody = allocateStr( lenNew + 1, &m_allocatedLen );

		if ( m_szBody == NULL )
			return;
	}

	m_len = lenNew;

	if ( ( m_szBody != NULL ) && ( mbs != NULL ) )
	{
		_mbstowcs( m_szBody, m_len, mbs, len, doubleZeroTerminated, codePage );
		m_szBody[ m_len ] = L'\0';
	}
}

const wchar_t *GetIntText( int i, int radix )
{
	static cWCStr wcs;
	wcs.SetInt( i, radix );
	return wcs.m_szBody;
}

const wchar_t *cWCStr::Translate( const char *s )
{
#ifdef _NoTranslation_
	Set( s );
#else
	cWCStr bf;
	bool bFound = false;

	if  ( ( pTranslateDll != NULL ) &&
		  ( !pTranslateDll->error )  )
	{
		bf.SetLength( 1024 );
		pTranslateDll->Translate( bf.m_szBody, bf.m_len, s, 0, &bFound ); 
	}

	if ( bFound )
		Set( bf.m_szBody );
	else
		Set( s );
#endif
	return m_szBody;
}

//
//	GetSizeText
//		Creates size string with TB, GB, MB, KB prefix.
//		Use the GetSizeFromText to convert it back.
//
const wchar_t *cWCStr::GetSizeText( DWORDLONG size, bool sizeInBytes, const wchar_t *szZero )
{
	if ( SetLength( 80 ) == 0 )
		return L"Null";

	if ( ( size == 0 ) && ( szZero != NULL ) )
		return szZero;

	GetSizeTextT( m_szBody, m_len, size, sizeInBytes );
	Truncate();
	return m_szBody;
}

#ifdef _WIN32

bool isSystemRootDir( const wchar_t *dir, int len )
{
	static cWCStr root;

	if ( root.m_len == 0 )
	{
		wchar_t bf[ MAX_PATH ];
		GetWindowsDirectoryT( bf, maxLen( bf ) );

		root.Add( bf, 3 );
	}

	if ( len == -1 )
		len = (int)wcslen( dir );

	return ( _wcsnicmp( dir, root, len ) == 0 );
}

bool isProgramFilesDir( const wchar_t *dir )
{
	static cWCStr pfd;

	if ( pfd.m_len == 0 )
	{
		TCHAR bf[ MAX_PATH ];

		if ( SHGetSpecialFolderPath( 0, bf, CSIDL_PROGRAM_FILES, false ) )
			pfd.Set( bf ); 
	}

	return ( wcsncmp( dir, pfd, pfd.m_len ) == 0 );
}

const wchar_t *cWCStr::FormatT( HINSTANCE hInst, int idsPattern, ... )
{
	cWCStr pattern;
	pattern.LoadStringT( hInst, idsPattern, true, true );

	cAutoDeallocate < wchar_t > ad;

	if ( ad.Allocate( 4096 ) == NULL )
	{
		ODS_fls( L"Assertion!" );
		return L"<NULL>";
	}

	cVAList vl;
	va_start( vl.marker, idsPattern );

	::FormatT( ad.m_ptr, ad.m_len, &vl, pattern );

	Free();
	Add( ad.m_ptr );

	return m_szBody;
}

const wchar_t *cWCStr::LoadIDS( int ids, bool bTranslate, bool bCheckParams )
{
	return LoadStringT( IH.m_hInst, ids, bTranslate, bCheckParams );
}

const wchar_t *
	cWCStr::LoadStringT(
		HINSTANCE hInst,
		int ids,
		bool bTranslate,	// = true
		bool bCheckParams	// = true
	)
{
	if ( SetLength( 4096 ) == 0 )
	{
		ODS_fls( L"Assertion!" );
		Set( L"<NULL>" );
	}
	else
	{
		int len = ::LoadStringT( hInst, ids, m_szBody, m_allocatedLen, bTranslate, bCheckParams );

		if ( len <= 0 )
		{
			ODS_fls( _FormatT( L"Assertion! invalid IDS %d", ids ) );
			Set( L"<NULL>" );
		}
		else
			SetLength( len );
	}

	return m_szBody;
}

bool cWCStr::RegGetString( HKEY hRoot, LPCWSTR lpSubKey, LPCWSTR szValue )
{
	DWORD dwType = 0, cbData = 0;
	LONG lRes = RegGetValueT( hRoot, lpSubKey, szValue, 0, &dwType, NULL, &cbData );
	if ( lRes != ERROR_SUCCESS )
		return false;

	if ( ( dwType != REG_SZ ) &&
		 ( dwType != REG_MULTI_SZ ) &&
		 ( dwType != REG_EXPAND_SZ ) )
			return false;

	SetLength( cbData / sizeof( wchar_t ) );

	lRes = RegGetValueT( hRoot, lpSubKey, szValue, 0, &dwType, (void*)m_szBody, &cbData );

	return ( lRes == ERROR_SUCCESS );
}

//static
bool cWCStr::RegSetString( HKEY hRoot, LPCWSTR lpSubKey, LPCWSTR lpValue, const wchar_t *str )
{
	LONG lRes = RegSetValueT( hRoot, lpSubKey, lpValue, REG_SZ, (void*)str, (int)((wcslen( str ) + 1) * sizeof( DWORD )) );
	return ( lRes == ERROR_SUCCESS );
}

const wchar_t *cWCStr::ExpandEnvironmentString( const wchar_t *path, bool *pExpanded )
{
	cWCStr tmp( MAX_PATH );

	DWORD count = ExpandEnvironmentStringsW( path, tmp.m_szBody, tmp.m_allocatedLen );

	bool bExpanded = ( count > 0 );

	if ( bExpanded )
	{
		if ( tmp.getAt(0) == L'"' )
		{
			int len = tmp.GetLength();
			Set( tmp.m_szBody + 1, len - 2 );
		}
		else
			Swap( &tmp );
	}
	else
		Set( path );

	if ( pExpanded != NULL )
		*pExpanded = bExpanded;

	return m_szBody;
}

const wchar_t *cWCStr::GetWindowTextT( HWND hWnd )
{
	SetLength( 0 );
	GetWindowTextTI( hWnd );
	return m_szBody;
}

int cWCStr::GetWindowTextTI( HWND hWnd )
{
	int len = ::GetWindowTextLengthT( hWnd );

	if ( SetLength( len + 1 ) == 0 )
	{
		ODS_fls( L"Assertion!" );
		return 0;
	}

	if ( ::GetWindowTextT( hWnd, m_szBody, m_len ) == 0 )
	{
		m_len = 0;
		return 0;
	}

	// MSDN reports:
	//	Under certain conditions, the GetWindowTextLength function may return a 
	//	value that is larger than the actual length of the text.
	// But if you will to add text to the body the length has to be real size
	
	m_len = (int)wcslen( m_szBody );
	return m_len;
}

const wchar_t *cWCStr::GetSystemError( DWORD dwErr ) // = (DWORD)-1 )
{
	wchar_t *lpMsgBuf = NULL;

	if ( dwErr == (DWORD)-1 )
		dwErr = ::GetLastError(); 

	DWORD dwRes = FormatMessageW(
		FORMAT_MESSAGE_ALLOCATE_BUFFER | 
		FORMAT_MESSAGE_FROM_SYSTEM |
		FORMAT_MESSAGE_IGNORE_INSERTS,
		NULL,
		dwErr,
		MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
		(wchar_t *)&lpMsgBuf,
		0, NULL );

	if ( dwRes == 0 )
	{
		FormatT( L"FormatMessage error %d, Original error %d", ::GetLastError(), dwErr );
		//Set( getWCStr( dwErr ) );
	}
	else
		Set( lpMsgBuf );

	if ( lpMsgBuf != NULL )
		LocalFree( lpMsgBuf );

	int len = (int)wcslen( m_szBody );
	while ( ( len > 0 ) && 
		( ( m_szBody[ len -1 ] == L'\n' ) ||
		  ( m_szBody[ len -1 ] == L'\r' )
		)
	)
		len--;

	SetLength( len );

	AddF( L" (0x%X)", dwErr );

	return m_szBody;
}

void FileTimeToUnixTime( FILETIME ft, DWORDLONG *pMilliSeconds )
{
	ULARGE_INTEGER uli;
	uli.HighPart = ft.dwHighDateTime;
	uli.LowPart = ft.dwLowDateTime;

	uli.QuadPart -= 116444736000000000;

	*pMilliSeconds = uli.QuadPart / 10000;
}

void UnixTimeToFileTime( DWORDLONG milliSeconds, LPFILETIME pft)
{
	// utcTime in milli-secons from January 1, 1970

	//SYSTEMTIME st;
	//GetSystemTime( &st );
	//memset( &st, 0, sizeof( st ) );
	//st.wYear = 1970;
	//st.wMonth = 1;
	//st.wDay = 1;

	//FILETIME ft;
	//SystemTimeToFileTime( &st, &ft );

	//ULARGE_INTEGER uli;
	//uli.LowPart = ft.dwLowDateTime;
	//uli.HighPart = ft.dwHighDateTime;
	//uli.QuadPart == 116444736000000000

	// Note that LONGLONG is a 64-bit value
	LONGLONG ll;

	//ll = (LONGLONG)t * 10000000 + 116444736000000000;
	ll = (LONGLONG)milliSeconds * 10000;//000
	ll += 116444736000000000;
	pft->dwLowDateTime = (DWORD)ll;
	pft->dwHighDateTime = ll >> 32;
}

void UnixTimeToSystemTime( DWORDLONG milliSeconds, LPSYSTEMTIME pst )
{
	FILETIME ft;
	UnixTimeToFileTime( milliSeconds, &ft );

	FileTimeToSystemTime( &ft, pst );
}

void SystemTimeToUnixTime( LPSYSTEMTIME pst, DWORDLONG *pMilliSeconds )
{
	FILETIME ft;
	SystemTimeToFileTime( pst, &ft );

	FileTimeToUnixTime( ft, pMilliSeconds );
}

#endif // _WIN32


//
//	getNFSShareRootDir
//	v. 6.08.6.2 - 16.03.2016
//		Bug #2599 NFS share error
//	Error was in detecting root directory of NFS share
//		Sample path:	 "\\UMBRELLA\vol\testnfswipe\qqq\New Text Document.txt"
//		The really shared directory is '\\UMBRELLA\vol\testnfswipe\'
//		Prev version detects root as '\\UMBRELLA\vol\', both function GetVolumeInformationT and FileExistsT rise exception for this path ( we cannot catch it by try block )
//	Temporary solution:
//		Treat the first directory '\\UMBRELLA\vol\testnfswipe\qqq\' as root.
//		At least on Windows 7 environment GetVolumeInformationT returns true for any sub directory from NFS share and fills file system name as 'NFS'
//
static bool getNFSShareRootDir( const wchar_t *path, wchar_t *rootDir = NULL )
{
	//
	// NFS path sample: "\\UMBRELLA\vol\testnfswipe\qqq\New Text Document.txt"
	//
	if ( wcsncmp( path, L"\\\\", 2 ) != 0 )
		return false;

	WIN32_FIND_DATAW fd;

	if ( !FileExistsT( path, &fd ) )
		return false;

	cWCStr dir;

	if ( ( fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) == 0 )
	{
		int pos = splitPath( path );
		if ( pos <= 2 )
			return false;

		// Note, the last '\' is required for proper working of GetVolumeInformationT.
		path = dir.SetBody( path, pos );
	}

	cWCStr volume( MAX_PATH );
	DWORD dwSerNum, dwMaxCompLen, dwFSFlag;
	cWCStr fsName( MAX_PATH );

	if ( !GetVolumeInformationT( path, volume.m_szBody, volume.m_allocatedLen, &dwSerNum, &dwMaxCompLen, &dwFSFlag, fsName.m_szBody, fsName.m_allocatedLen, true ) )
	{
		int pos = splitPath( path );
		if ( pos <= 2 )
			return false;

		// Note, the last '\' is required for proper working of GetVolumeInformationT.
		path = dir.SetBody( path, pos );

		if ( !GetVolumeInformationT( path, volume.m_szBody, volume.m_allocatedLen, &dwSerNum, &dwMaxCompLen, &dwFSFlag, fsName.m_szBody, fsName.m_allocatedLen, true ) )
			return false;
	}

	if ( !fsName.equ( L"NFS" ) )
		return false;

	if ( rootDir != NULL )
		copyString( rootDir, path, MAX_PATH );

	ODS_fls( path );

	return true;
}

//
//	GetDriveRootDirT
//		Gets the file's/folder's drive or mount point root folder name by means of GetVolumeName.
//	We can get file mount point by the function so it rather may be called as GetFileMountPoint :)
//	Really it splits file path to several values:
//		file path	-> C:\tmp\text.txt
//		root dir	-> C:\
//		volume name	-> \\?\Volume{0cd30cef-bc98-11e0-ac71-806e6f6e6963}\
//		loca path	-> \tmp\text.txt
//
bool GetDriveRootDirT( wchar_t *rootDirOut, const wchar_t *path, bool silent, wchar_t *pVolumeName, const wchar_t **ppLocalPath )
{
	//
	//	It strange but the GetVolumePathNameT function does not return true if path is already root folder.
	//	more other it clears output value.
	//
	cWCStr bf;
	bf.Set( path );

	cWCStr wcsBf;

	if ( rootDirOut == NULL )
	{
		if ( wcsBf.SetLength( MAX_PATH ) == 0 )
			return false;

		rootDirOut = wcsBf.m_szBody;
	}

	if ( getNFSShareRootDir( path, rootDirOut ) )
		return true;

	if ( GetVolumePathNameT( path, rootDirOut, MAX_PATH ) )
	{
		if ( pVolumeName == NULL )
			return true;

		int len = (int)wcslen( rootDirOut );

		if ( rootDirOut[len-1] == L'\\' )
			rootDirOut[len-1] = L'\0';

		if ( !GetVolumeName( pVolumeName, MAX_PATH, rootDirOut ) )
			return false;

		copyString( rootDirOut + len, L"\\", MAX_PATH - len );
		
		if ( ppLocalPath != NULL )
			*ppLocalPath = path + len - 1;

		ODS_fls( _FormatT( L"%d, %s", len, rootDirOut ) );

		return true;
	}

	DWORD err = GetLastError();
	ODS_fls( _FormatT( L"GetVolumePathNameT( %s ) returns error %s", path, _GetSystemError( err ) ) );

	//
	// The GetVolumePathName returns error if file does not exist.
	//

	int len = (int)wcslen( bf.m_szBody );

	wchar_t lVolumeName[ MAX_PATH ] = {0};

	if ( pVolumeName == NULL )
		pVolumeName = lVolumeName;

	if ( !GetVolumeName( pVolumeName, MAX_PATH, bf.m_szBody, &err ) )
	{
		ODS_fls( _FormatT( L"GetVolumeName( %s ) returns error %s", bf.m_szBody, _GetSystemError( err ) ) );

		for(;; len-- )
		{
			if ( len == 0 )
			{
				cWCStr msg;
				msg.FormatT( L"Root folder is not found for path\n'%s'", path );

				ODS_fls( _FormatT( L"Assertion:: %s", msg.m_szBody ) );

				if ( !silent )
					MessageBoxT(
						NULL,
						msg.m_szBody,
						L"Get Disk Info",
						MB_ICONSTOP );

				return false;
			}

			if ( bf.m_szBody[ len ] == L'\\' )
			{
				bf.m_szBody[ len ] = L'\0';
				if ( GetVolumeName( pVolumeName, MAX_PATH, bf.m_szBody, &err ) )
				{
					bf.setAt( len, L'\\' );
					bf.setAt( len + 1, L'\0' );

					// root folder is found
					break;
				}
				else
					ODS_fls( _FormatT( L"GetVolumeName( %s ) returns error %s", bf.m_szBody, _GetSystemError( err ) ) );
			}
		}
	}

	//if ( bf.m_szBody[ len - 1 ] != L'\\' )
	//{
	//	bf.SetLength( len );
	//	bf.Add( L"\\" );
	//}

	if ( rootDirOut != NULL )
	{
		if ( bf.equ( L"\\\\" ) ) // GetVolumeName does not detect for network share root
		{
			ODS_fls( L"Error at GetVolumeName" );
			bf.Set( path );
			len = (int)wcslen( bf.m_szBody );
			if ( bf.m_szBody[ len - 1 ] != L'\\' )
			{
				bf.SetLength( len );
				bf.Add( L"\\" );
			}
		}
		copyString( rootDirOut, bf.m_szBody, MAX_PATH );
	}

	if ( ppLocalPath != NULL )
		*ppLocalPath = path + len;

	return true;
}

bool GetRootDirT( cWCS &out, const wchar_t *fileName )
{
	cWCS wcs( MAX_PATH );
	if ( wcs.m_allocatedLen < MAX_PATH )
		return false;

	if ( !GetRootDirT( wcs.m_szBody, fileName ) )
		return false;

	wcs.Truncate();
	wcs.Swap( &out );

	return true;
}


bool GetRootDirT( wchar_t *buff, const wchar_t *fileName )
{
	//buff[0] = L'\0';

	if ( fileName[0] == L'\\' ) // f.e. "\\server\share\subfolder." 
	{
		cWCStr wcs( MAX_PATH );

		// "\Device\HarddiskVolume0\1gb.jbc"
		if ( GetDOSNameForHarddiskVolume( wcs.m_szBody, wcs.m_len, fileName )  ) 
		{
			// Translated name has to be as sample 'C:\1gb.jbc"
			if ( wcs.m_szBody[0] ==  L'\\' )
			{
				ODS_fls( L"Assertion!" );
				return false;
			}

			return GetRootDirT( buff, wcs.m_szBody );
		}

		int count = 2; // Counts pair '\' for sever and share names

		for( int i = 2;; i++ )
		{
			if ( ( fileName[i] == L'\\' ) ||
				 ( ( fileName[ i ] == L'\0' ) && ( count == 1 ) ) // for case if name is "\\server\share"
				)
			{
				if ( --count )	// count the back slash.
					continue;

				//wcsncpy_s( buff, MAX_PATH, fileName, i );

				if ( i >= MAX_PATH )
					return false;

				wcsncpy( buff, fileName, i );

				buff[ i++ ] = L'\\';
				buff[ i++ ] = L'\0';
				return true;
			}

			if ( ( fileName[ i ] == L'\0' ) || ( i >= ( MAX_PATH - 2 ) ) )
				return false;
		}
	}
	else
	{
		if ( fileName[1] != L':' )
			return false;

		wchar_t ch = towupper( fileName[0] );
		wcscpy_s( buff, MAX_PATH, L"A:\\" );
		buff[0] = ch;
	}

	return true;
}

DWORD GetFullPathNameT( LPCWSTR lpFileName, cWCStr *wcs, LPWSTR *lpFilePart )
{
	DWORD len = GetFullPathNameT( lpFileName, 0, NULL, lpFilePart );

	if ( len == 0 )
		return 0;

	if ( wcs->SetLength( len ) == 0 )
		return 0;

	len = GetFullPathNameT( lpFileName, wcs->m_len, wcs->m_szBody, lpFilePart );

	wcs->SetLength( len );

	return len;
}

HANDLE CreateFileAndDirectory(
	LPCWSTR logFileName,
    DWORD dwDesiredAccess,
    DWORD dwShareMode,
    LPSECURITY_ATTRIBUTES lpSecurityAttributes,
    DWORD dwCreationDisposition,
    DWORD dwFlagsAndAttributes,
    HANDLE hTemplateFile,
	DWORD *pLastErr
 )
{
	HANDLE hFile = CreateFileT( logFileName, dwDesiredAccess, dwShareMode, lpSecurityAttributes, dwCreationDisposition, dwFlagsAndAttributes, hTemplateFile );

	if ( hFile != INVALID_HANDLE_VALUE )
		return hFile;

	DWORD err;

	if ( pLastErr == NULL )
		pLastErr = &err;
	
	*pLastErr = GetLastError();

	cWCStr dir;

	int pos = splitPath( logFileName, &dir );

	if ( ( pos != 0 ) && !FileExistsT( dir.m_szBody ) ) // Directory does not exist
	{
		if ( RecCreateDirectory( dir.m_szBody ) )
			hFile = CreateFileT( logFileName, dwDesiredAccess, dwShareMode, lpSecurityAttributes, dwCreationDisposition, dwFlagsAndAttributes, hTemplateFile );
	}
	
	*pLastErr = GetLastError();

	return hFile;
}

#ifndef _CreateHandle_ODS
HANDLE
CreateFileT(
    LPCWSTR lpFileName,
    DWORD dwDesiredAccess,
    DWORD dwShareMode,
    LPSECURITY_ATTRIBUTES lpSecurityAttributes,
    DWORD dwCreationDisposition,
    DWORD dwFlagsAndAttributes,
    HANDLE hTemplateFile
    )
{
#ifdef _DEBUG
	if ( lpFileName[0] == L'\0' )
		ODS_fls( L"Assertion!" );
#endif

	HANDLE hFile = CreateFileW( lpFileName, dwDesiredAccess, dwShareMode, lpSecurityAttributes, dwCreationDisposition, dwFlagsAndAttributes, hTemplateFile );

	if ( ( hFile == INVALID_HANDLE_VALUE ) && cLongName::isTooLongForShortName( lpFileName ) )
	{
		hFile = CreateFileW( cLongName().generate( lpFileName ), dwDesiredAccess, dwShareMode, lpSecurityAttributes, dwCreationDisposition, dwFlagsAndAttributes, hTemplateFile );
	}

	return hFile;
}
#endif

BOOL
CreateDirectoryT(
    LPCWSTR lpPathName,
    LPSECURITY_ATTRIBUTES lpSecurityAttributes,
	DWORD *pError
    )
{
	if ( !CreateDirectoryW( lpPathName, lpSecurityAttributes ) && cLongName::isTooLongForShortName( lpPathName ) )
		CreateDirectoryW( cLongName().generate( lpPathName ), lpSecurityAttributes );

	DWORD err = GetLastError();

	if ( pError != NULL )
		*pError = err;

	BOOL ok = FileExistsT( lpPathName );

	return ok;
}

BOOL
CreateDirectoryExT(
    IN LPCWSTR lpTemplateDirectory,
    IN LPCWSTR lpNewDirectory,
    IN LPSECURITY_ATTRIBUTES lpSecurityAttributes
    )
{
	CreateDirectoryExW( lpTemplateDirectory, lpNewDirectory, lpSecurityAttributes );

	if ( FileExistsT( lpNewDirectory ) )
		return true;

	if ( cLongName::isTooLongForShortName( lpTemplateDirectory ) ||
		 cLongName::isTooLongForShortName( lpNewDirectory ) 
		)
	{
		CreateDirectoryExW( cLongName().generate( lpTemplateDirectory ), cLongName().generate( lpNewDirectory ), lpSecurityAttributes );
		return FileExistsT( lpNewDirectory );
	}

	return false;
}

BOOL RecCreateDirectory( const wchar_t *dir )
{
	size_t len = wcslen( dir );

	if ( ( len == 2 ) && ( dir[1] == L':' ) )
		return true;

	if ( accessT( dir, 0 ) == 0 )
		return true;

	DWORD err = GetLastError();

	if ( err == ERROR_BAD_NETPATH )
		return false;
	else
	if ( err == ERROR_BAD_PATHNAME )
		return false;

	wchar_t parrent[ MAX_PATH ];

	len = splitPath( dir, parrent, NULL );

	if ( len == 0 )
	{
		SetLastError( ERROR_PATH_NOT_FOUND );
		return false;
	}

	RecCreateDirectory( parrent );

	CreateDirectoryT( dir, NULL );

	err = GetLastError();

	return accessT( dir, 0 ) == 0;
}

bool CheckFileDirectory( const wchar_t *filePath, bool bCreate )
{
	cWCStr dir;
	splitPath( filePath, &dir );

	if ( FileExistsT( dir.m_szBody ) )
		return true;

	return RecCreateDirectory( dir.m_szBody ) == TRUE;
}

int accessT( const wchar_t *lpPathName, int mode )
{
	if ( cLongName::isTooLongForShortName( lpPathName ) )
	{
		//ODSFuncS( lpPathName );

		// _waccess does not work with long name, we'll use FindFirstFileT and CreateFileT.
		//	return _waccess( cLongName().generate( lpPathName ), mode );

		WIN32_FIND_DATAW lfd;

		HANDLE hFind = FindFirstFileT( lpPathName, &lfd );

		if ( hFind == INVALID_HANDLE_VALUE )
		{
			//ODS_fl;
			return -1;
		}

		FindClose( hFind );

		if ( mode == 0 )
		{
			//ODS_fl;
			return 0;
		}

		if ( mode & 2 ) // Checks write permissions.
		{
			HANDLE hFile = CreateFileT( lpPathName, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );
			if ( hFile == INVALID_HANDLE_VALUE )
			{
				//ODS_fl;
				return -1;
			}

			CloseHandle( hFile );
		}
		else
		if ( mode & 4 ) // Checks read permissions ( but we checked read premitions if mode & 2 above, so we uses the 'else' ).
		{
			HANDLE hFile = CreateFileT( lpPathName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );
			if ( hFile == INVALID_HANDLE_VALUE )
			{
				//ODS_fl;
				return -1;
			}

			CloseHandle( hFile );
		}

		//ODS_fl;
		return 0;
	}

	return _waccess( lpPathName, mode );
}

//
//	FlushFileBuffersT
//		When we just rename file we have no file handle to use original FlushFileBuffers function.
//		But we need the flushing, for save the remanig result to disk.
//
bool FlushFileBuffersT( const wchar_t *path )
{
	HANDLE hFile = CreateFileT(
		path,
		GENERIC_READ | GENERIC_WRITE,
		FILE_SHARE_READ,
		NULL,
		OPEN_EXISTING,
		FILE_FLAG_WRITE_THROUGH, 
		NULL);

	if ( hFile == INVALID_HANDLE_VALUE )
		return false;

	FlushFileBuffers( hFile );

	CloseHandle( hFile );

	return true;
}

static DWORD WINAPI static_flushFileBuffers( LPVOID lpThreadParameter )
{
	HANDLE hFile = ( HANDLE )lpThreadParameter;
	return (DWORD)FlushFileBuffers( hFile );
}

DWORD FlushFileBuffersT( HANDLE hFile, DWORD timeout )
{
	HANDLE hThread = CreateThread( NULL, 0, static_flushFileBuffers, hFile, 0, NULL );

	if ( hThread == NULL )
		return (DWORD)-1;

	DWORD dwRes = WaitForSingleObject( hThread, timeout );

	if ( dwRes == WAIT_TIMEOUT )
		ODS_fls( L"Warning:: Timeout at FlushFileBuffers" );
	else
	if ( dwRes != WAIT_OBJECT_0 )
		ODS_fls( _FormatT( L"Assertion:: Error in waiting %x", dwRes ) );
	else
	if ( !GetExitCodeThread( hThread, &dwRes ) )
		ODS_fls( _FormatT( L"Assertion:: Error %d in GetExitCodeThread", GetLastError() ) );
	else
	if ( !dwRes )
		ODS_fls( _FormatT( L"Assertion:: Error %d in FlushFileBuffers(%04x)", GetLastError(), hFile ) );

	CloseHandle( hThread );

	return dwRes;
}

BOOL FileExistsT( const wchar_t *lpPathName )
{
	if ( lpPathName == NULL )
		return false;

	return accessT( lpPathName, 0 ) == 0;
}

BOOL DirectoryExistsT( const wchar_t *lpPathName )
{
	WIN32_FIND_DATAW lfd;

	if ( !FileExistsT( lpPathName, &lfd ) )
		return false;

	return ( lfd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) != 0;
}

BOOL FileExistsT( const wchar_t *lpPathName, WIN32_FIND_DATAW *pFindData, int *pLastError  )
{
	if ( lpPathName == NULL )
		return false;

	WIN32_FIND_DATAW lfd;

	if ( pFindData == NULL )
		pFindData = &lfd;

	HANDLE hFind = FindFirstFileT( lpPathName, pFindData );

	if ( hFind == INVALID_HANDLE_VALUE )
	{
		if ( pLastError != NULL )
			*pLastError = GetLastError();

		//
		//	Alternative stream cannot be enumerated by FindFirstFileT, but GetFileAttributesExW works fine starting from Win2K
		//
		WIN32_FILE_ATTRIBUTE_DATA fad;

		if ( GetFileAttributesExW( lpPathName, GetFileExInfoStandard, &fad ) )
		{
			memset( pFindData, 0, sizeof( WIN32_FIND_DATAW ) );
			pFindData->dwFileAttributes = fad.dwFileAttributes;
			pFindData->ftCreationTime = fad.ftCreationTime;
			pFindData->ftLastAccessTime = fad.ftLastAccessTime;
			pFindData->ftLastWriteTime = fad.ftLastWriteTime;
			pFindData->nFileSizeHigh = fad.nFileSizeHigh;
			pFindData->nFileSizeLow = fad.nFileSizeLow;

			return true;
		}

		if ( cLongName::isTooLongForShortName( lpPathName ) &&
			 GetFileAttributesExW( cLongName().generate( lpPathName ), GetFileExInfoStandard, &fad ) )
		{
			memset( pFindData, 0, sizeof( WIN32_FIND_DATAW ) );
			pFindData->dwFileAttributes = fad.dwFileAttributes;
			pFindData->ftCreationTime = fad.ftCreationTime;
			pFindData->ftLastAccessTime = fad.ftLastAccessTime;
			pFindData->ftLastWriteTime = fad.ftLastWriteTime;
			pFindData->nFileSizeHigh = fad.nFileSizeHigh;
			pFindData->nFileSizeLow = fad.nFileSizeLow;

			return true;
		}

		if ( !FileExistsT( lpPathName ) )
			return false;

		int len = (int)wcslen( lpPathName );

		if ( ( len == 0 ) || ( len >= MAX_PATH ) )
			return false;

		// It happens for directories with '\' at the end only
		if ( lpPathName[ len - 1 ] != L'\\' )
		{
			//ODS_fls( L"Assertion!" );
			return false;
		}

		//
		//	Note,
		//		FindFirstFileT return false for directory like 'C:\QQQ\',
		//		so we cut the last '\'.
		//		As result the function works as the accessT does.
		//
		wchar_t path[ MAX_PATH ];
		wcsncpy_s( path, lpPathName, len );

		path[ len - 1 ] = L'\0';
		
		hFind = FindFirstFileT( path, pFindData );

		if ( hFind == INVALID_HANDLE_VALUE )
		{
			// FindFirstFile with root directory returns error

			memset( pFindData, 0, sizeof( *pFindData ) );
			pFindData->dwFileAttributes = GetFileAttributesT( path );

			if ( pFindData->dwFileAttributes != DWORD(-1) )
				return true;

			return false;
		}
	}

	FindClose( hFind );

	return true;
}

BOOL MoveFileT( LPCWSTR lpExistingFileName, LPCWSTR lpNewFileName )
{
	return MoveFileExT( lpExistingFileName, lpNewFileName, 0 );
}

DWORD
GetFileAttributesT(
    LPCWSTR lpFileName
    )
{
	DWORD dwAttr = GetFileAttributesW( lpFileName );

	if ( ( dwAttr == INVALID_FILE_ATTRIBUTES ) && cLongName::isTooLongForShortName( lpFileName ) )
		dwAttr = GetFileAttributesW( cLongName().generate( lpFileName ) );

	return dwAttr;
}

DWORDLONG GetDirSizeT( const wchar_t *dir, int *pDirCount, int *pFileCount, bool bWalkStreams, bool bSkipReparsePoint )
{
	cFindFile cff( dir );

	if ( !cff.findFirstFile() )
		return 0;

	DWORDLONG size = 0;

	if ( pDirCount != NULL ) (*pDirCount)++;

	do
	{
		if ( cff.isDir() )
		{
			if ( ( ( cff.m_data.dwFileAttributes & FILE_ATTRIBUTE_REPARSE_POINT ) != 0 ) && bSkipReparsePoint )
			{
				ODS_fls( _FormatT( L"Reparse point '%s' - Skipped!", cff.getCurrentPath() ) );
				continue;
			}

			size += GetDirSizeT( cff.getCurrentPath(), pDirCount, pFileCount, bWalkStreams );
		}
		else
		if ( pFileCount != NULL )
			(*pFileCount)++;

		if ( bWalkStreams )
			size += GetFileSreamsAllocatedSizes( cff.getCurrentPath() );
		else
			size += cff.getSize();
	}
	while( cff.findNextFile() );

	return size;
}

DWORDLONG
GetFileSreamsAllocatedSizes( LPCWSTR lpFileName )
{
	cStreams cs( lpFileName );

	if ( cs.m_error )
		return GetFileSizeT( lpFileName );

	DWORDLONG size = 0;

	FILE_STREAM_INFORMATION streamInfo;

	while( cs.getNextName( NULL, false, &streamInfo ) )
		size += streamInfo.AllocationSize.QuadPart;

	return size;
}

DWORDLONG
GetFileSizeT( LPCWSTR lpFileName, DWORD *pErr )
{
	HANDLE hFile;
	WIN32_FIND_DATAW findFileData;
	hFile = FindFirstFileT( lpFileName, &findFileData );
	ULARGE_INTEGER uli;

	if ( hFile == INVALID_HANDLE_VALUE )
	{
		if ( pErr != NULL )
			*pErr = GetLastError();

		WIN32_FILE_ATTRIBUTE_DATA fad;

		if ( !GetFileAttributesExW( lpFileName, GetFileExInfoStandard, (void*)&fad ) )
			return (DWORDLONG)-1;

		uli.LowPart = fad.nFileSizeLow;
		uli.HighPart = fad.nFileSizeHigh;
	}
	else
	{
		FindClose( hFile );
		uli.LowPart = findFileData.nFileSizeLow;
		uli.HighPart = findFileData.nFileSizeHigh;
	}

	return uli.QuadPart;
}

DWORDLONG
GetFileSizeT( HANDLE hFile  )
{
	LARGE_INTEGER liSize;
	liSize.LowPart = GetFileSize( hFile , (DWORD *)&(liSize.HighPart) );

	if ( liSize.LowPart == INVALID_FILE_SIZE )
	{
		if ( GetLastError() != NO_ERROR )
			return (DWORDLONG)-1;
	}

	return liSize.QuadPart;
}

static void fda2fdw( LPWIN32_FIND_DATAW lpFindFileDataW, LPWIN32_FIND_DATAA lpFindFileDataA )
{
	memcpy( lpFindFileDataW, lpFindFileDataA, sizeof( WIN32_FIND_DATAA ) - sizeof( lpFindFileDataA->cFileName ) - sizeof( lpFindFileDataA->cAlternateFileName ) );
	size_t num;
	mbstowcs_s( &num, lpFindFileDataW->cFileName, lpFindFileDataA->cFileName, _TRUNCATE );
	mbstowcs_s( &num, lpFindFileDataW->cAlternateFileName, lpFindFileDataA->cAlternateFileName, _TRUNCATE );
}

HANDLE
FindFirstFileT(
    LPCWSTR lpFileName,
    LPWIN32_FIND_DATAW lpFindFileData,
	int *pErrCode // = NULL
    )
{
	HANDLE hFind;
	int err = 0;

	SetLastError( err );

	hFind = FindFirstFileW( lpFileName, lpFindFileData );
	err = GetLastError();

	if ( ( hFind == INVALID_HANDLE_VALUE ) && cLongName::isTooLongForShortName( lpFileName ) )
	{
		SetLastError( 0 );
		hFind = FindFirstFileW( cLongName().generate( lpFileName ), lpFindFileData );
		err = GetLastError();
	}

	if ( pErrCode != NULL )
		*pErrCode = err;

	//ODS_fls( _FormatT( L"%s, %X, %d", lpFileName, hFind, err ) );

	return hFind;
}

BOOL
CopyFileT(
    LPCWSTR lpExistingFileName,
    LPCWSTR lpNewFileName,
    BOOL bFailIfExists
    )
{
	BOOL ok = CopyFileW( lpExistingFileName, lpNewFileName, bFailIfExists );

	if ( !ok &&
		(
		 cLongName::isTooLongForShortName(lpExistingFileName) ||
		 cLongName::isTooLongForShortName(lpNewFileName)
		)
	   )
		   ok = CopyFileW( cLongName().generate( lpExistingFileName ), cLongName().generate( lpNewFileName ), bFailIfExists );

	return ok;
}

BOOL
CopyFileExT(
    IN LPCWSTR lpExistingFileName,
    IN LPCWSTR lpNewFileName,
    //IN LPPROGRESS_ROUTINE lpProgressRoutine OPTIONAL,
    IN LPVOID lpProgressRoutine OPTIONAL,
    IN LPVOID lpData OPTIONAL,
    IN LPBOOL pbCancel OPTIONAL,
    IN DWORD dwCopyFlags
    )
{
	static CDll kernel32dll( false, "KERNEL32.dll", "CopyFileExW", false );

	if ( kernel32dll.error )
		return CopyFileT( lpExistingFileName, lpNewFileName, dwCopyFlags == COPY_FILE_FAIL_IF_EXISTS );

	typedef WINBASEAPI BOOL WINAPI typeCopyFileExW(
		IN LPCWSTR lpExistingFileName,
		IN LPCWSTR lpNewFileName,
		IN LPPROGRESS_ROUTINE lpProgressRoutine OPTIONAL,
		IN LPVOID lpData OPTIONAL,
		IN LPBOOL pbCancel OPTIONAL,
		IN DWORD dwCopyFlags
		);

	return ((typeCopyFileExW *)kernel32dll.pProc)
				( lpExistingFileName, lpNewFileName, (LPPROGRESS_ROUTINE)lpProgressRoutine, lpData, pbCancel, dwCopyFlags );
}

BOOL
SetFileAttributesT(
    LPCWSTR lpFileName,
    DWORD dwFileAttributes
    )
{
	BOOL ok = SetFileAttributesW( lpFileName, dwFileAttributes );

	if ( !ok && cLongName::isTooLongForShortName( lpFileName ) )
	{
		ok =  SetFileAttributesW( cLongName().generate( lpFileName ), dwFileAttributes );
	}

	return ok;
}

BOOL
DeleteFileT(
    LPCWSTR lpFileName
    )
{
	// ODSFuncS( lpFileName );

	BOOL res = DeleteFileW( lpFileName );

	if ( !res && cLongName::isTooLongForShortName( lpFileName ) )
		res = DeleteFileW( cLongName().generate( lpFileName ) );

	//ODSFuncRes( res );

	return res;
}

BOOL
MoveFileExT(
    LPCWSTR lpExistingFileName,
    LPCWSTR lpNewFileName,
    DWORD dwFlags,
	DWORD *pError
    )
{
	// ODSFuncS( _FormatT( L"%s,%s, %X", lpExistingFileName, lpNewFileName, dwFlags ) );

	BOOL ok = MoveFileExW( lpExistingFileName, lpNewFileName, dwFlags );

	DWORD err = GetLastError();

	if ( !ok && 
		(
		 cLongName::isTooLongForShortName( lpExistingFileName ) ||
		 cLongName::isTooLongForShortName( lpNewFileName )
		)
	   )
	{
		cLongName lnExiting;
		cLongName lnNew;

		const wchar_t *szExisting = lnExiting.generate( lpExistingFileName );
		const wchar_t *szNew = lnNew.generate( lpNewFileName );

		ok = MoveFileExW( szExisting, szNew, dwFlags );

		err = GetLastError();

		if ( !ok )
		{
			ODS_fls( _FormatT( L"Error %d in moving long name file. '%s' -> '%s'", err, szExisting, szNew ) );
//1188:0F78 18:750            MoveFileExT:1763 Error 112 in moving long name file. '\\?\C:\~BCWipe.tmp\MFT_DIR\3\6\3\5\2XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX' -> '\\?\C:\~BCWipe.tmp\MFT_DIR\3\6\3\5\2XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX'
//\\?\C:\~BCWipe.tmp\MFT_DIR\3\6\3\5\2XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
//\\?\C:\~BCWipe.tmp\MFT_DIR\3\6\3\5\2XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
		}
	}

	if ( pError )
		*pError = err;

	return ok;
}

BOOL CopyUntilRebootT( const wchar_t *szSrcFile , const wchar_t *szDstFile )
{
	// 1) Create temporary file name....

	wchar_t tempName[ MAX_PATH ];

	GetTempPathT( MAX_PATH, tempName );

	GetTempFileNameT( tempName, L"BCS", 0, tempName );

	// 2) Copy file to the temporary file...

	if ( !CopyFileT( szSrcFile, tempName, FALSE ))
		return FALSE;

	// 3) Mark file for remove after reboot...

	MoveFileExT( szDstFile, NULL, MOVEFILE_DELAY_UNTIL_REBOOT );
	MoveFileExT( tempName,szDstFile,MOVEFILE_DELAY_UNTIL_REBOOT );

	return TRUE;
}

bool CreateShortestName( wchar_t *newName, const wchar_t *oldName, int diffLen, bool bForce )
{
//	ODS_fls( L"Assertion!" );

	static wchar_t symT[ 10 + 26 ] = L"";

	if ( symT[0] == L'\0' )
	{
		int i = 0;
		for( ; i < 10; i++ )
			symT[i] = L'0' + i;

		for( int j = 0; j <= 'Z' - 'A'; j++, i++ )
		{
			assert( i < maxLen( symT ) );
			symT[i] = L'A' + j;
		}

		assert( i == maxLen( symT ) );
	}

	int pos = splitPath( oldName );

	if ( pos == 0 )
		return false;

	copyString( newName, oldName, MAX_PATH );

	wchar_t *name = newName + pos;

	int len = (int)wcslen( name );

	//
	// If diffLen >= len, so it is not possible to rename just the file, we have to rename all directories in the pass.
	// But we'll try to create shortes name starting from 1 up to the the file name length
	//
	if ( diffLen >= len )
	{
		if ( !bForce )
			return false;

		for ( int i = len - 1; i; i-- )
		{
			if (  CreateShortestName( newName, oldName, i ) )
				return true;
		}

		return false;
	}
	else
		len -= diffLen;

	name[ len ] = L'\0';

	for( int i = len; i--; )
	{
		for( int j = maxLen( symT ); j--; )
		{
			//
			//	Renaming after checking allows to save existing name ( it will be just truncated ).
			//
			const bool bRenameBefore = false;
			
			if ( bRenameBefore || ( name[i] == L' ' ) ) 
				name[i] = symT[j];

			if ( !FileExistsT( newName ) )
				return true;

			if ( !bRenameBefore )
				name[i] = symT[j];
		}
	}

	return false;
}

bool RemoveDirectoryContent(
		LPCWSTR lpPathName,
		DWORD *pError
    )
{
	cWCStr dir;
	dir.Set( lpPathName );

	int len = (int)wcslen( dir.m_szBody );
	if ( dir.m_szBody[ len - 1 ] != L'\\' )
		dir.Add( L"\\" );

	cWCStr findPath;
	findPath.Set( dir.m_szBody );
	findPath.Add( L"*" );

	WIN32_FIND_DATAW findData;
	HANDLE hFind = FindFirstFileT( findPath.m_szBody, &findData );

	if ( hFind != INVALID_HANDLE_VALUE )
	{
		do
		{
			//int findPathLen = (int)wcslen( findData.cFileName ) + len;
			//wcscpy_s( findPath.m_szBody + len, findPath.m_len - len, findData.cFileName );
			findPath.Set( dir.m_szBody );
			findPath.Add( findData.cFileName );

			if ( ( findData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) != 0 )
			{
				if ( wcscmp( findData.cFileName, L"." ) == 0 ) continue;
				if ( wcscmp( findData.cFileName, L".." ) == 0 ) continue;
				
				RemoveDirectoryT( findPath.m_szBody, true, pError );
			}
			else
			{
				if ( ( findData.dwFileAttributes & FILE_ATTRIBUTE_READONLY ) != 0 )
					SetFileAttributesT( findPath.m_szBody, FILE_ATTRIBUTE_NORMAL );

				if ( !DeleteFileT( findPath.m_szBody ) )
				{
					if ( pError != NULL )
						*pError = GetLastError();

					return false;
				}
			}
		}
		while( FindNextFileT( hFind, &findData ) );

		FindClose( hFind );
	}

	return true;
}

BOOL
RemoveDirectoryT(
    LPCWSTR lpPathName,
	bool removeItems, // removes items and subfolders, default == true
	DWORD *pError
    )
{
	//ODSFuncS( lpPathName );

	if ( removeItems &&
			!RemoveDirectoryContent( lpPathName, pError ) )
					return false;

	BOOL bReadOnly = ( GetFileAttributesT( lpPathName ) & FILE_ATTRIBUTE_READONLY ) != 0;

	if ( bReadOnly )
		SetFileAttributesT( lpPathName, FILE_ATTRIBUTE_NORMAL );

	BOOL ok = false;
	DWORD err = 0;

	ok = RemoveDirectoryW( lpPathName );
	err = GetLastError();

	if ( !ok && cLongName::isTooLongForShortName( lpPathName ) )
	{
		ok = RemoveDirectoryW( cLongName().generate( lpPathName ) );
		err = GetLastError();
	}

	if ( !ok )
		ODS_fls( _FormatT( L"RemoveDirectory '%s' error %d", lpPathName, err ) );

	if ( FileExistsT( lpPathName, NULL ) )
	{
		ODS_fls( _FormatT( L"File exists" ) );
		if ( ok )
		{
			ok = false;
			err = ERROR_FILE_EXISTS;
		}
	}
	else
		ok = true;

	if ( pError != NULL )
		*pError = err;

	return ok;
}

wchar_t *getErrorString( int errNum, wchar_t *bf, int bfMaxLen, DWORD value )
{
	if ( errNum == eMemoryAllocation )
		return L"Error: Memory allocating";

	if ( ( bf == NULL ) || ( bfMaxLen == 0 ) )
		return L"Error: Invalid parameters";

	if ( errNum == eRCStringNotFound )
		return FormatT( bf, bfMaxLen, L"Error: Resource string 0x%04X is not found", value );
	else
	if ( errNum == eRCInstanceInvalid )
		return FormatT( bf, bfMaxLen, L"Error: Invalid resource instance 0x%04X", value );
	else
		return FormatT( bf, bfMaxLen, L"Error: %d\tValue: 0x%04X", errNum, value );
}


wchar_t *CreateNewString( HINSTANCE hInst, int ids )
{
	wchar_t ws[512] = L"";
	int len = LoadStringT( hInst, ids, ws, maxLen( ws ) );

	if ( len == -1 )
		return NULL;

	wchar_t *newWS = new wchar_t[ len + 1 ];

	if ( newWS != NULL )
	{
		wcsncpy_s( newWS, len, ws, _TRUNCATE );
		newWS[ len ] = L'\0';
	}

	return newWS;
}

int GetWindowTextT( HWND hWnd, LPWSTR lpString, int nMaxCount )
{
	return GetWindowTextW( hWnd, lpString, nMaxCount );
}

const wchar_t *GetWindowTextT( cWCStr *wcs, HWND hWnd )
{
	wcs->GetWindowTextT( hWnd );
	return wcs->m_szBody;
}

bool isRTLProcessLayout()
{
	DWORD layout;
	if ( !getProcessDefaultLayout( &layout ) )
		return false;

	return ( layout & LAYOUT_RTL ) != 0;
}

bool getProcessDefaultLayout( DWORD *pLayout )
{
	CDll dll( false, "user32.dll", "GetProcessDefaultLayout", false );

	if ( dll.error )
		return false;

	typedef WINUSERAPI BOOL WINAPI typeGetProcessDefaultLayout( DWORD *pdwDefaultLayout);

	return ((typeGetProcessDefaultLayout *)dll.pProc)( pLayout ) == TRUE;
}

bool setProcessDefaultLayout( bool rtlLayout )
{
	CDll dll( false, "user32.dll", "SetProcessDefaultLayout", false );

	if ( dll.error )
		return false;

	typedef WINUSERAPI BOOL WINAPI typeSetProcessDefaultLayout( DWORD dwDefaultLayout);

	BOOL res = ((typeSetProcessDefaultLayout *)dll.pProc)( rtlLayout ? LAYOUT_RTL : 0 );

	return res != FALSE;
}

HIMAGELIST ImageList_CreateT( int cx, int cy, UINT flags, int cInitial, int cGrow )
{
	static DWORD ver = (DWORD)-1;

	if ( ver == (DWORD)-1 )
	{
		if ( !cComCtl32Dll::getStatic().LoadOnce() )
			return NULL;

		ver = cComCtl32Dll::getStatic().version.dwMajorVersion;
	}

	if ( ( ver < 6 ) && ( ( flags & ILC_COLORDDB ) ==  ILC_COLOR32 ) )
	{
		flags = ( flags & ~ILC_COLORDDB ) | ILC_COLOR24;
	}

	//
	//	Icons in image list are automatically mirrored if process has the flag
	//	But it is not convenient for several icons, so the following flags disable mirroring at all.
	//
//	if ( ( pTranslateDll != NULL ) && ( pTranslateDll->m_flipDialog ) )

	if ( isRTLProcessLayout() )
		flags |= 0x00002000 | 0x00008000; // ILC_MIRROR | ILC_PERITEMMIRROR

	typedef HIMAGELIST WINAPI typeImageList_Create(int cx, int cy, UINT flags, int cInitial, int cGrow);

	static typeImageList_Create *proc = (typeImageList_Create *)GetProcAddress( cComCtl32Dll::getStatic().hInst,"ImageList_Create" );
	if ( proc == NULL )
		return NULL;

	return proc( cx, cy, flags, cInitial, cGrow );
}

int ImageList_GetImageCountT(HIMAGELIST himl)
{
	if ( !cComCtl32Dll::getStatic().LoadOnce(  ) )
		return 0;

	typedef int WINAPI typeImageList_GetImageCount(HIMAGELIST himl);

	static typeImageList_GetImageCount *proc = (typeImageList_GetImageCount *)GetProcAddress( cComCtl32Dll::getStatic().hInst,"ImageList_GetImageCount" );
	if ( proc == NULL )
		return 0;

	return proc( himl );
}

BOOL ImageList_GetIconSizeT(HIMAGELIST himl,int*cx,int*cy)
{
	if ( !cComCtl32Dll::getStatic().LoadOnce(  ) )
		return FALSE;

	typedef BOOL WINAPI typeImageList_GetIconSize(HIMAGELIST himl,int*cx,int*cy);

	static typeImageList_GetIconSize *proc = (typeImageList_GetIconSize *)GetProcAddress( cComCtl32Dll::getStatic().hInst,"ImageList_GetIconSize" );
	if ( proc == NULL )
		return FALSE;

	return proc( himl, cx, cy );
}

HICON ImageList_GetIconT(HIMAGELIST himl,int i,UINT flags)
{
	if ( !cComCtl32Dll::getStatic().LoadOnce(  ) )
		return NULL;

	typedef HICON WINAPI typeImageList_GetIcon(HIMAGELIST himl,int i,UINT flags);

	static typeImageList_GetIcon *proc = (typeImageList_GetIcon *)GetProcAddress( cComCtl32Dll::getStatic().hInst,"ImageList_GetIcon" );
	if ( proc == NULL )
		return NULL;

	return proc( himl, i, flags );
}

BOOL ImageList_DestroyT(HIMAGELIST himl)
{
	if ( !cComCtl32Dll::getStatic().LoadOnce(  ) )
		return false;

	typedef BOOL WINAPI typeImageList_Destroy(HIMAGELIST himl);

	static typeImageList_Destroy *proc = (typeImageList_Destroy *)GetProcAddress( cComCtl32Dll::getStatic().hInst,"ImageList_Destroy" );
	if ( proc == NULL )
		return false;

	return proc( himl );
}

int ImageList_ReplaceIconT(HIMAGELIST himl,int i,HICON hicon)
{
	if ( !cComCtl32Dll::getStatic().LoadOnce(  ) )
		return -1;

	typedef int WINAPI typeImageList_ReplaceIcon(HIMAGELIST himl,int i,HICON hicon);

	static typeImageList_ReplaceIcon *proc = (typeImageList_ReplaceIcon *)GetProcAddress( cComCtl32Dll::getStatic().hInst,"ImageList_ReplaceIcon" );
	if ( proc == NULL )
		return -1;

	return proc( himl, i, hicon );
}

HIMAGELIST ImageList_CreateCopyT( HIMAGELIST hImlOld )
{
	if ( !cComCtl32Dll::getStatic().LoadOnce(  ) )
		return NULL;

	if ( hImlOld == NULL )
		return NULL;

	HIMAGELIST hImlNew = NULL;

	int count = ImageList_GetImageCountT( hImlOld );
	int cx = 0, cy = 0;
	if ( count && ImageList_GetIconSizeT( hImlOld, &cx, &cy ) )
	{
		hImlNew = ImageList_CreateT( cx, cy, ILC_MASK | ILC_COLOR32, count, count );

		for ( int i = 0; i < count; i++ )
		{
			HICON hIcon = ImageList_GetIconT( hImlOld, i, ILD_NORMAL );

			if ( hIcon == NULL )
			{
				ImageList_DestroyT( hImlNew );
				hImlNew = NULL;
				break;
			}

			// ImageList_AddIcon( hImlNew, hIcon );
			ImageList_ReplaceIconT( hImlNew, -1, hIcon );
		}

	}

	return hImlNew;
}

int ListView_SetItemTextT( HWND hwndLV, int i, int iSubItem, const wchar_t *pszText )
{
	LV_ITEMW _ms_lvi;
	_ms_lvi.iSubItem = iSubItem;
	_ms_lvi.pszText = (wchar_t *)pszText;
	return (int)SendMessageW((hwndLV), LVM_SETITEMTEXTW, (WPARAM)i, (LPARAM)(LV_ITEM FAR *)&_ms_lvi);
}

//
//	It returns number of items copied to pszText_ without zero termination.
//
int ListView_GetItemTextT( HWND hwndLV, int i, int iSubItem_, wchar_t *pszText_, int cchTextMax_)
{
	int len;

	LV_ITEMW _ms_lvi;
	_ms_lvi.iSubItem = iSubItem_;
	_ms_lvi.cchTextMax = cchTextMax_;
	_ms_lvi.pszText = pszText_;

#pragma warning( disable : 4244 )
	len = SendMessageW((hwndLV), LVM_GETITEMTEXTW, (WPARAM)i, (LPARAM)&_ms_lvi);
#pragma warning( default : 4244 )

	return len;
}

BOOL ListView_SetItemData( HWND hwnd, int iItem, LPARAM data )
{
	LVITEMW lvi;
	memset( &lvi, 0, sizeof( lvi ) );
	lvi.iItem = iItem;
	lvi.mask = LVIF_PARAM;
	lvi.lParam = data;

	return ListView_SetItemT( hwnd, &lvi );
}

LPARAM ListView_GetItemData( HWND hwnd, int iItem )
{
	LVITEMW lvi;
	memset( &lvi, 0, sizeof( lvi ) );
	lvi.iItem = iItem;
	lvi.mask = LVIF_PARAM;

	if ( !ListView_GetItemT( hwnd, &lvi ) )
		return 0;

	return lvi.lParam;
}

#ifndef MIIM_STRING
#define MIIM_STRING      0x00000040
#define MIIM_FTYPE       0x00000100
#endif


BOOL
SetMenuTextT(
    IN HMENU hMenu,
    IN UINT uPosition,
	bool byPosition,
	const wchar_t *text 
	)
{
	UINT nID = uPosition; 
	if ( byPosition )
		nID = GetMenuItemID( hMenu, uPosition );

	UINT nState = GetMenuState( hMenu, nID, MF_BYCOMMAND );
	nState &= ~(MF_BITMAP|MF_OWNERDRAW|MF_SEPARATOR|MF_BYPOSITION);

	nState |= MF_STRING;

	// set menu text
	return ModifyMenuT( hMenu, nID, nState, nID, text );
}

BOOL
GetMenuTextT(
    IN HMENU hMenu,
    IN UINT uPosition,
	bool byPosition,
	wchar_t *text,
	int textMaxLen
	)
{
	MENUITEMINFOW mii;
	memset( &mii, 0, sizeof( mii ) );
	mii.cbSize = sizeof( mii );
	mii.fMask = MIIM_STRING;
	mii.dwTypeData = text;
	mii.cch = textMaxLen;

	return GetMenuItemInfoT( hMenu, uPosition, byPosition, &mii );
}

ULONG_PTR
	GetMenuItemParam(
		IN HMENU hMenu,
		IN UINT uPosition,
		bool byPosition
	)
{
	MENUITEMINFOW mii;
	memset( &mii, 0, sizeof( mii ) );
	mii.cbSize = sizeof( mii );
	mii.fMask = MIIM_DATA;

	GetMenuItemInfoT( hMenu, uPosition, byPosition, &mii );
	
	return (ULONG_PTR)mii.dwItemData;
}

BOOL ExtTextOutT( HDC hdc, int X, int Y, UINT fuOptions, CONST RECT* lprc, LPCWSTR lpString, UINT cbCount, CONST INT* lpDx )
{
	if ( cbCount == -1 )
		cbCount = (int)wcslen( lpString );

	return ExtTextOutW( hdc, X, Y, fuOptions, lprc, lpString, cbCount, lpDx );
}

int GetTextSizeT( HWND hWnd, wchar_t *ws )
{
	HDC hDC = GetDC( hWnd );
//	SIZE size = { 0, 0 };
//	GetTextExtentPoint32T( hDC, ws, wcslen( ws ), &size );
//	return size.x;

	RECT r = { 0,0,0,0 };
	DrawTextT( hDC, ws, (int)wcslen( ws ), &r, DT_CALCRECT );
	return r.right - r.left;
}

//
//	PrintClientToDC
//		Sends WM_PRINTCLIENT message to window.
//		The message draws window client aria to position 0,0 of device content.
//		But we need to draw on specified position, so:
//			1. new srcHDC is created
//			2. windowd client region is printed here.
//			3. BitBlt is used to copy srcHDC to dstHDC.
//
bool PrintClientToDC( HDC dstHDC, int dstX, int dstY, HWND srcHWnd, int srcX, int srcY, int cx, int cy )
{
	if ( ( srcX == 0 ) && ( srcY == 0 ) && ( cx == -1 ) && ( cy == -1 ) )
	{
		LRESULT lRes = SendMessage( srcHWnd, WM_PRINTCLIENT, (WPARAM)(dstHDC), (LPARAM)(PRF_ERASEBKGND | PRF_CLIENT | PRF_NONCLIENT));
		return lRes == 0;
	}

	// create hDCMem a copy of hDC
	HDC srcHDC = CreateCompatibleDC(dstHDC);
	if (srcHDC == NULL)
		return false;

#ifdef _UseTransparentBitmap_
	HBITMAP bitmap = LoadBitmap( IH.m_hInst, MAKEINTRESOURCE( IDB_transparent48x16 ) );
#else
	HBITMAP bitmap = CreateCompatibleBitmap( dstHDC, cx, cy );
#endif

	bool ok = false;

	if ( bitmap )
	{
		HBITMAP bitmapOld = reinterpret_cast<HBITMAP> ( SelectObject ( srcHDC, bitmap ) );
		LRESULT lRes = SendMessage( srcHWnd, WM_PRINTCLIENT, (WPARAM)(srcHDC), (LPARAM)(PRF_ERASEBKGND | PRF_CLIENT | PRF_NONCLIENT));
#ifdef _UseTransparentBitmap_
		COLORREF cr = GetPixel( srcHDC, 0, 0 );
		ok = TransparentBlt( dstHDC, dstX, dstY, cx, cy, srcHDC, srcX, srcY, cx, cy, cr ) == TRUE;
#else
		ok = BitBlt( dstHDC, dstX, dstY, cx, cy, srcHDC, srcX, srcY, SRCCOPY ) == TRUE;
#endif
		
		SelectObject( srcHDC, bitmapOld );
		DeleteObject( bitmap );
	}
	
	DeleteObject( srcHDC );

	return ok;
}

ULONG_PTR SetClassLongPtrT(
    HWND hWnd,
    int nIndex,
    LONG_PTR dwNewLong
)
{
#pragma warning( disable : 4244 )
	return SetClassLongPtrW(  hWnd, nIndex, dwNewLong );
#pragma warning( default : 4244 )
}


LONG_PTR SetWindowLongPtrT(
		HWND hWnd,
		int nIndex,
		LONG_PTR dwNewLong
)
{
#pragma warning( disable : 4244 )
	return SetWindowLongPtrW(  hWnd, nIndex, dwNewLong );
#pragma warning( default : 4244 )
}

LONG_PTR GetWindowLongPtrT( HWND hWnd, int nIndex )
{
#pragma warning( disable : 4244 )
	return GetWindowLongPtrW(  hWnd, nIndex );
#pragma warning( default : 4244 )
}


#ifndef _NoShell32Lib_

//#define saveConvert( lpA, lp ) _wcstombs( lpA, sizeof( lpA ), (lp != NULL ? lp : L"\0" ) )

bool ShellSelect( const wchar_t *path )
{
	if ( !isWinXP() )
		return false;

	if ( !FileExistsT( path ) )
		return false;

	wchar_t cmd[ MAX_PATH * 2  + 100 ];

	GetWindowsDirectoryT( cmd, maxLen( cmd ) );

//	cDisableWow64Redirection disableWow64( disableWow64Redirection );

	int len = (int)wcslen( cmd );

	FormatT( cmd + len, maxLen( cmd ) - len, L"\\Explorer.exe /select, %s", path );

	HANDLE hProcess = RunProcessWithWaitingT( cmd, false );
	CloseHandle( hProcess );

	return true;
}

static bool updateUnread( cWCStr *out, const wchar_t *text )
{
	if ( text )
	{
		for( ; *text != L'\0'; text ++ )
		{
			  wchar_t ch = *text;

			  if ( ( ( ch >= L'0' ) && ( ch <= L'9' ) ) ||
					   ( ( ch >= L'a' ) && ( ch <= L'z' ) ) ||
					   ( ( ch >= L'A' ) && ( ch <= L'Z' ) ) ||
					   ( ch == L'-' )
					  )
					  out->Add( &ch, 1 );
			  else
			  {
					  DWORD dwSym = BYTE( ch );

					  out->AddF( L"%%%02X", dwSym );
			  }
		}
	}

	return true;
}

HINSTANCE
ShellExecuteMailTo( HWND hWnd, const wchar_t *szAddress, const wchar_t *szSubject, const wchar_t *szNotes, const wchar_t *attacheFile )
{
	const wchar_t *sendMailPattern = L"mailto:%s?subject=%s&body=%s";

	//
	//	Historically, address in dialog already has 'mailto:' prefix,
	//	so we cut it off.
	//

	const wchar_t *szMailto = L"mailto:";
	int len = (int)wcslen( szMailto );
	
	if ( _wcsnicmp( szAddress, szMailto, len ) == 0 )
		szAddress += len;

	cWCStr updated;
	updateUnread( &updated, szNotes );

	cWCStr wcs;
	wcs.FormatT( sendMailPattern, szAddress, szSubject, updated.m_szBody );

	//
	//	02.09.2015 Does not work ((
	//	Use mapi_SendMail with attachement.
	//
	//if ( ( attacheFile != NULL ) && ( attacheFile[0] ) )
	////	wcs.AddF( L"&Att:filename=\"%s\"", attacheFile );
	////	wcs.AddF( L"&Attachment=\"%s\"", attacheFile );
	//	wcs.AddF( L"&attach=\"%s\"", attacheFile );

	if ( wcs.m_len > 1000 )
		wcs.SetLength( 1000 );

	HINSTANCE res = ShellExecuteT( hWnd, L"open", wcs.m_szBody, 0, 0, SW_SHOW );

	return res;
}

bool BrowseForFolderT( HINSTANCE hInst, int idsTitle, HWND hWnd, wchar_t *dir, UINT ulFlags, int clsIdRoot ) // ulFlags = BIF_RETURNONLYFSDIRS, clsIdRoot = CSIDL_DRIVES
{
	wchar_t szTitle[80];
	LoadStringT( hInst, idsTitle, szTitle, maxLen(szTitle));
	return BrowseForFolderT( szTitle, hWnd, dir, ulFlags, clsIdRoot );
}

bool BrowseForFolderT( const wchar_t *szTitle, HWND hWnd, wchar_t *dir, UINT ulFlags, int clsIdRoot ) // ulFlags = BIF_RETURNONLYFSDIRS, clsIdRoot = CSIDL_DRIVES
{
	BROWSEINFOW bi;
	bi.hwndOwner = hWnd;
	bi.pidlRoot = NULL;
	if ( clsIdRoot != 0 ) // CSIDL_DRIVES
		SHGetSpecialFolderLocation( hWnd, clsIdRoot, (LPITEMIDLIST *)&bi.pidlRoot );

	bi.pszDisplayName = NULL;
	bi.lpszTitle = szTitle;
	bi.ulFlags = ulFlags;
	bi.lpfn = NULL;

	LPITEMIDLIST pidl = SHBrowseForFolderT(&bi);

	BOOL bSuccess = false;

	if ( pidl != NULL )
		bSuccess = SHGetPathFromIDListT(pidl, dir );

	static LPMALLOC pMalloc = NULL;

	if ( pMalloc == NULL )
		SHGetMalloc( &pMalloc );

	if ( pMalloc != NULL )
	{
		if ( pidl != NULL )
			pMalloc->Free(pidl);

		if ( bi.pidlRoot != NULL )
			pMalloc->Free( (void *)bi.pidlRoot );
	}

	return bSuccess == TRUE;
}


//static 
FARPROC cCoInitialize::m_procCoInitialize = NULL;
//static 
FARPROC cCoInitialize::m_procCoUninitialize = NULL;

cCoInitialize::cCoInitialize( LPVOID pvReserved, bool bInitialize ) // = NULL, = true )
{
	m_hRes = S_FALSE;

	if ( !bInitialize )
		return;

	if ( m_procCoInitialize == NULL )
	{
		static CDll dll( false, L"Ole32.dll", "CoInitialize", false );
		if ( dll.error != CDll::errNo )
			return;

		m_procCoInitialize = (FARPROC)dll.pProc;
		m_procCoUninitialize = GetProcAddress( dll.hInst, "CoUninitialize" );

		if ( ( m_procCoInitialize == NULL ) || ( m_procCoUninitialize == NULL ) )
			return;
	}

	typedef HRESULT STDAPICALLTYPE typeCoInitialize( LPVOID pvReserved );

	m_hRes = ((typeCoInitialize *)m_procCoInitialize )( pvReserved ); // CoInitialize( pvReserved );
}

cCoInitialize::~cCoInitialize()
{
	if ( ( m_hRes == S_OK ) && ( m_procCoUninitialize != NULL ) )
	{
		typedef void STDAPICALLTYPE typeCoUninitialize( void );
		( (typeCoUninitialize*)m_procCoUninitialize)(); // CoUninitialize();
	}
}

bool SHGetFileInfoT( SHFILEINFOW &sfi, DWORD getFileInfo, DWORD clsid )
{
	LPITEMIDLIST pidl = NULL;

	HRESULT hr = SHGetFolderLocation(NULL, clsid, NULL, 0, &pidl);

	if ( !SUCCEEDED(hr) )
		return false;

	hr = (HRESULT)SHGetFileInfoW((LPCWSTR)pidl, -1, &sfi, sizeof(sfi), SHGFI_PIDL | getFileInfo );
	ILFree(pidl);

	return SUCCEEDED(hr);
}

DWORD* SHGetFileInfoT( LPCWSTR pszPath, DWORD dwFileAttributes, SHFILEINFOW *psfi, UINT cbFileInfo, UINT uFlags, bool bCoInitialize )
{
	//ODSFuncS( _FormatT( L"%s, %X, %X, %X, %s", pszPath, dwFileAttributes, cbFileInfo, uFlags, getWCStr( bCoInitialize ) ) );
	
	cCoInitialize cCOM( NULL, bCoInitialize );

	if ( bCoInitialize && ( cCOM.m_hRes != S_OK ) )
	{
		ODS_fls( L"Assertion! Error in CoInitialize" );
		return 0;
	}

	return (DWORD *)SHGetFileInfoW( pszPath, dwFileAttributes, psfi, cbFileInfo, uFlags );
}

#endif _NoShell32Lib_

DWORD GetTempPathT( cWCStr *path, bool doNotCreate )
{
	DWORD dwLen = GetTempPathT( 0, NULL, true );

	if ( path->SetLength( dwLen ) == 0 )
		return 0;

	dwLen = GetTempPathT( path->m_len, path->m_szBody, doNotCreate );

	path->SetLength( dwLen );

	return dwLen;
}

// GetTempPathIH is moved to GetTempPathIH.cpp

/////////////////

DWORD
GetTempPathT(
    DWORD nBufferLength,
    LPWSTR lpBuffer,
	bool doNotCreate
    )
{
	//
	//	If the function is called by impersonated thread it returns 'Windows\Temp' folder instead of 
	//	%userprofile%\AppData\Local\Temp\, so to avoid ambiguity we try to get temp folder from user profile first.
	//

	// V.3
	// use GetUserLocalAppDataPath_ByImpersonatedToken;

	//
	//	V.2
	//	GetEnvironmentVariableW gives same result as GetTempPathW for impersonated thread
	//	cWCStr wcsSrc( MAX_PATH );
	//	DWORD dwRes = GetEnvironmentVariableW( L"TEMP", wcsSrc.m_szBody, wcsSrc.m_len );
	//	wcsncpy( lpBuffer, wcsSrc.m_szBody, nBufferLength );
	//
	//
	//	V.1
	//	Windows XP does not link directories as Windows 7 does
	//		"C:\Users\Alexander V. Pichuev\AppData\Local\Temp\"
	//		"C:\Users\Alexander V. Pichuev\Local Settings\Temp\"
	//	Following code gets "C:\Users\Alexander V. Pichuev\AppData\Local\Temp\" as Temp
	//		but XP uses "C:\Users\Alexander V. Pichuev\Local Settings\Temp\"
	//
	//if ( SHGetSpecialFolderPathT( NULL, lpBuffer, CSIDL_PROFILE, false ) )
	//{
	//	SHGetSpecialFolderPathT( NULL, lpBuffer, CSIDL_LOCAL_APPDATA, false );
	//	
	//	wcscat_s( lpBuffer, nBufferLength, L"\\Temp\\" );
	//	if ( FileExistsT( lpBuffer ) )
	//		return (DWORD)wcslen( lpBuffer );
	//}

	DWORD dwRes = GetTempPathW( nBufferLength, lpBuffer );

	if ( dwRes == 0 )
	{
		dwRes = GetLastError();
		ODS_fls( _FormatT( L"Assertion! Error '%s' in GetTempPathW'", _GetSystemError( dwRes ) ) );
		return dwRes;
	}

	if ( !doNotCreate )
	{
		int res = _waccess( lpBuffer, 0 );

		if ( res != 0 )
		{
			BOOL bOk = CreateDirectoryW( lpBuffer, NULL );
			if ( !bOk )
			{
				dwRes = GetLastError();
				ODS_fls( _FormatT( L"Assertion! Error '%s' in creating directory '%s'", _GetSystemError( dwRes ), lpBuffer ) );
				return dwRes;
			}
		}
	}

	return dwRes;
}


BOOL
GetVolumeInformationT(
    IN LPCWSTR lpRootPathName,
    OUT LPWSTR lpVolumeNameBuffer,
    IN DWORD nVolumeNameSize,
    OUT LPDWORD lpVolumeSerialNumber,
    OUT LPDWORD lpMaximumComponentLength,
    OUT LPDWORD lpFileSystemFlags,
    OUT LPWSTR lpFileSystemNameBuffer,
    IN DWORD nFileSystemNameSize,
	bool silent
    )
{
	cSilentMode sm( silent );

	BOOL ok = false;

	ok = GetVolumeInformationW( 
				lpRootPathName,
				lpVolumeNameBuffer, nVolumeNameSize,
				lpVolumeSerialNumber,
				lpMaximumComponentLength,
				lpFileSystemFlags,
				lpFileSystemNameBuffer, nFileSystemNameSize
			);

	return ok;
}

//
//	CreateTempFileName
//		Works as GetTempFileName but allows to create handle to the temporary file.
//		GetTempFileName is insecure by 'Common weakness enumeration. CWE-377: Insecure Temporary File'.
//
UINT CreateTempFileName(
    IN LPCWSTR lpDirectory,
    IN LPCWSTR lpPrefixString,
    IN UINT uUnique,
    OUT LPWSTR lpTempFileName,
	HANDLE *pHFile,
	DWORD dwShareMode, // = FILE_SHARE_READ
	DWORD fileAtributes, // = FILE_ATTRIBUTE_NORMAL
	DWORD *pErr, // = NULL
	IN LPCWSTR lpPostfixString, // = L".tmp",
	IN LPCWSTR lpTemplateSrc // == NULL is used for directory while.
    )
{
	if ( uUnique == 0 )
		uUnique = (UINT)time(0) ^ (UINT)clock();

	cWCStr wcs;

	if ( lpDirectory != NULL )
		wcs.Set( lpDirectory );
	else
		GetTempPathT( &wcs );

	int len = wcs.GetLength();

	if ( len == 0 )
		return 0;

	if ( wcs.getAt( len - 1 ) != L'\\' )
	{
		wcs.Add( L"\\" );
		++len;
	}
	
	if ( lpPrefixString == NULL )
		lpPrefixString = L"";

	DWORD err = 0;

	for( ;;++uUnique)
	{
		wcs.SetLength( len );
		wcs.AddF( L"%s%08X%s", lpPrefixString, uUnique, lpPostfixString );

		if ( fileAtributes & FILE_ATTRIBUTE_DIRECTORY )
		{
			bool ok = false;

			if ( lpTemplateSrc != NULL )
				ok = CreateDirectoryExW( lpTemplateSrc, wcs.m_szBody, NULL ) == TRUE;
			else
				ok = CreateDirectoryW( wcs.m_szBody, NULL ) == TRUE;

			if ( ok )
			{
				*pHFile = CreateFileW( wcs.m_szBody, 
								GENERIC_READ,
								FILE_SHARE_READ,
								NULL,
								OPEN_EXISTING,
								FILE_ATTRIBUTE_NORMAL | FILE_FLAG_BACKUP_SEMANTICS,
								NULL );
			}
		}
		else
			*pHFile = CreateFileW( wcs.m_szBody, GENERIC_READ | GENERIC_WRITE, dwShareMode, NULL, CREATE_NEW, fileAtributes, NULL );

		err = GetLastError();

		if ( pErr != NULL )
			*pErr = err;

		if ( *pHFile != INVALID_HANDLE_VALUE )
		{
			if ( lpTempFileName != NULL )
				copyString( lpTempFileName, wcs.m_szBody, MAX_PATH );

			return uUnique;
		}

		if ( !FileExistsT( wcs.m_szBody ) )
			return 0;		
	}
}

UINT
GetTempFileNameT(
    IN LPCWSTR lpPathName,
    IN LPCWSTR lpPrefixString,
    IN UINT uUnique,
    OUT LPWSTR lpTempFileName,
	DWORD *pErr
  )
{
	// Veracode means the function insecure, because there is time before creating uniquie name and opening file.
	// See 'Common weakness enumeration. CWE-377: Insecure Temporary File'
	// Attacker may create link to a system file with the temporary name and privileged application overwrite this file.
	// It is recommended use CreateTempFileNameT instead.
	//	return GetTempFileNameW( lpPathName, lpPrefixString, uUnique, lpTempFileName );
	HANDLE hFile = INVALID_HANDLE_VALUE;

	UINT res = CreateTempFileName( lpPathName, lpPrefixString, uUnique, lpTempFileName, &hFile, FILE_SHARE_READ, FILE_ATTRIBUTE_NORMAL, pErr );

	if ( res != 0 )
		CloseHandle( hFile );

	return res;
}

BOOL GetProgramFilesDirT( wchar_t *maxPathBuff, bool bX64 )
{
	HKEY hKey;

	REGSAM samDesired = KEY_READ;
	
	if ( bX64 )
		samDesired |= KEY_WOW64_64KEY;

	if ( RegOpenKeyExT(
			HKEY_LOCAL_MACHINE,
			L"SOFTWARE\\Microsoft\\Windows\\CurrentVersion",
			0,
			samDesired,
			&hKey
		) != ERROR_SUCCESS )
			return FALSE;

	DWORD dwType = REG_EXPAND_SZ;

	DWORD cbData = MAX_PATH * sizeof( maxPathBuff[0] );

	if ( RegQueryValueExT(
			hKey,L"ProgramFilesDir",NULL,&dwType,(BYTE *)maxPathBuff,&cbData
			) != ERROR_SUCCESS) return FALSE;

	return (RegCloseKey(hKey) == ERROR_SUCCESS);
}

//BOOL GetProgramFilesDirT( wchar_t *maxPathBuff )
//{
//	HKEY hKey;
//
//	if (RegOpenKeyExT(
//			HKEY_LOCAL_MACHINE,
//			L"SOFTWARE\\Microsoft\\Windows\\CurrentVersion",
//			0,
//			KEY_READ,	// security access mask 
//			&hKey			// address of handle of open key 
//		) != ERROR_SUCCESS)
//		return FALSE;
//
//
//	DWORD dwType = REG_EXPAND_SZ;
//
//	DWORD cbData = MAX_PATH;
//
//	if ( RegQueryValueExT( hKey, L"ProgramFilesDir", NULL,&dwType,( BYTE *)maxPathBuff,&cbData )
//		!= ERROR_SUCCESS)
//			return FALSE;
//
//	return (RegCloseKey(hKey) == ERROR_SUCCESS);
//}

bool isSparseFile( const wchar_t *filePath )
{
	DWORDLONG dwlReal = GetFileSizeT( filePath );

	if ( dwlReal == (DWORDLONG)-1 )
	{ ODS_fls( L"Assertion!" );	return false; }

	ULARGE_INTEGER uli;
	uli.LowPart = GetCompressedFileSizeT( filePath, &( uli.HighPart ) );

	if ( ( uli.LowPart == INVALID_FILE_SIZE ) && ( GetLastError() != NO_ERROR ) )
	{ ODS_fls( L"Assertion!" );	return false; }

	if ( dwlReal == uli.QuadPart )
		return false;

	return true;
}

//
// CreateTempFilePath
//	Uses CreateTempFileName instead of GetTempFileName
//	GetTempFileName is insecure by 'Common weakness enumeration. CWE-377: Insecure Temporary File'.
//
bool CreateTempFilePath(
		wchar_t *buff,
		wchar_t *prefix,
		bool useShortName,
		HANDLE *phFile,
		DWORD dwShareMode // = FILE_SHARE_READ
   )
{
	wchar_t tmpDir[ MAX_PATH ];

	if ( GetTempPathT( maxLen( tmpDir ), tmpDir ) == 0 )
		return false;

	if ( !FileExistsT( tmpDir ) )
	{
		CreateDirectoryT( tmpDir, NULL );
		if ( !FileExistsT( tmpDir ) )
			return false;
	}

	if ( useShortName )
	{
		wchar_t shortPath[ MAX_PATH ];

		if ( GetShortPathNameT( tmpDir, shortPath, maxLen( shortPath ) ) == 0 )
			return false;

		copyString( tmpDir, shortPath );
	}

	if ( CreateTempFileName( tmpDir, prefix, 0, buff, phFile, dwShareMode ) == 0 )
		return false;

	return true;
}

bool GetTempFilePathT( wchar_t *buff, wchar_t *prefix, bool useShortName  )
{
	HANDLE hFile = INVALID_HANDLE_VALUE;
	bool ok = CreateTempFilePath( buff, prefix, useShortName, &hFile );

	if ( ok )
		CloseHandle( hFile );

	return ok;

	//wchar_t tmpDir[ MAX_PATH ];

	//if ( GetTempPathT( maxLen( tmpDir ), tmpDir ) == 0 )
	//	return false;

	//if ( !FileExistsT( tmpDir ) )
	//{
	//	CreateDirectoryT( tmpDir, NULL );
	//	if ( !FileExistsT( tmpDir ) )
	//		return false;
	//}

	//if ( useShortName )
	//{
	//	wchar_t shortPath[ MAX_PATH ];

	//	if ( GetShortPathNameT( tmpDir, shortPath, maxLen( shortPath ) ) == 0 )
	//		return false;

	//	copyString( tmpDir, shortPath );
	//}

	//if ( GetTempFileNameT( tmpDir, prefix, 0, buff ) == 0 )
	//	return false;

	//return true;
}

DWORD GetShortPathNameT( LPCWSTR lpszLongPath, cWCStr *wcs )
{
	DWORD len = GetShortPathNameT( lpszLongPath, NULL, 0 );

	if ( len == 0 )
		return 0;

	if ( wcs->SetLength( len + 1 ) == 0 )
		return 0;

	len = GetShortPathNameT( lpszLongPath, wcs->m_szBody, wcs->m_len );

	wcs->SetLength( len );

	return len;
}

DWORD
GetShortPathNameT(
    LPCWSTR lpszLongPath,
    LPWSTR  lpszShortPath,
    DWORD   cchBuffer
    )
{
	DWORD len = GetShortPathNameW( lpszLongPath, lpszShortPath, cchBuffer );

	if ( ( len == 0 ) && cLongName::isTooLongForShortName( lpszLongPath ) )
		len = GetShortPathNameW( cLongName().generate( lpszLongPath ), lpszShortPath, cchBuffer );

	return len;
}

DWORD GetLongPathNameT( LPCWSTR lpszShortPath, cWCStr *path )
{
	DWORD dwLen = GetLongPathNameT( lpszShortPath, NULL, 0 );

	if ( dwLen == 0 )
		return dwLen;

	if ( path->SetLength( dwLen ) == 0 )
		return 0;

	dwLen = GetLongPathNameT( lpszShortPath, path->m_szBody, path->m_len );
	
	path->SetLength( dwLen );

	return dwLen;
}

BOOL
GetDiskFreeSpaceExT(
    IN LPCWSTR lpDirectoryName,
    OUT PULARGE_INTEGER lpFreeBytesAvailableToCaller,
    OUT PULARGE_INTEGER lpTotalNumberOfBytes,
    OUT PULARGE_INTEGER lpTotalNumberOfFreeBytes
    )
{
	ULARGE_INTEGER TotalNumberOfBytes;
	if ( lpTotalNumberOfBytes == NULL )
		lpTotalNumberOfBytes = &TotalNumberOfBytes;

	ULARGE_INTEGER TotalNumberOfFreeBytes;
	if ( lpTotalNumberOfFreeBytes == NULL )
		lpTotalNumberOfFreeBytes = &TotalNumberOfFreeBytes;

	static CDll kernel32dll( false, "KERNEL32.dll", "GetDiskFreeSpaceExW", false );
	if ( kernel32dll.error )
	{
		ODS_fls( L"Assertion! Function GetDiskFreeSpaceExW is not fopund at KERNEL32.dll" );
		return false;
	}

	typedef WINBASEAPI BOOL WINAPI TypeGetDiskFreeSpaceExW(
		LPCWSTR lpDirectoryName,
		PULARGE_INTEGER lpFreeBytesAvailableToCaller,
		PULARGE_INTEGER lpTotalNumberOfBytes,
		PULARGE_INTEGER lpTotalNumberOfFreeBytes
	);

	return ((TypeGetDiskFreeSpaceExW *)kernel32dll.pProc)(
				lpDirectoryName,
				lpFreeBytesAvailableToCaller,
				lpTotalNumberOfBytes,
				lpTotalNumberOfFreeBytes );
}

//
//	Windows 95 GetModuleFileNameA returns path in upper case, it is not convenient.
//	The function slitly modify the text.
//
static bool updatePath( char* pathOut, const char *pathIn )
{
	char dir[ MAX_PATH ];

	int pos = splitPath( pathIn, dir, NULL );

	if ( pos == 0 )
	{
		strcpy_s( pathOut, MAX_PATH, pathIn );
		return true;
	}

	WIN32_FIND_DATAA findData;
	HANDLE hFind = FindFirstFileA( pathIn, &findData );

	if ( hFind == INVALID_HANDLE_VALUE )
		return false;

	FindClose( hFind );

	if ( !updatePath( pathOut, dir ) )
		return false;

	int len = (int)strlen( pathOut );

	pathOut[ len++ ] = '\\';
	strcpy_s( pathOut + len, MAX_PATH - len, findData.cFileName );

	return _access( pathOut, 0 ) == 0;
}

int splitPath( const char *path, char *dir, char *file )
{
	int len = (int)strlen( path );

	while( --len )
	{
		if ( path[ len ] == '\\' )
		{
			if ( dir != NULL )
			{
				strncpy( dir, path, len );
				dir[ len ] = '\0';
			}

			len++;

			if ( file != NULL )
			{
				strcpy_s( file, MAX_PATH, path + len );
			}

			return len;
		}
	}

	// ODS_fls( L"Assertion!" );

	if ( dir != NULL ) dir[0] = 0;
	if ( file != NULL ) strcpy_s( file, MAX_PATH, path );

	return 0;
}

//
//	isFileStream
//
//		Returns true if the path name contains stream name
//
//		Sample:
//			'QQQ.pdf:Zone.Identifier:$DATA'
//		stream pos -^
//
//		Return false for main stream
//		Samples:
//			'QQQ.pdf'
//			'QQQ.pdf::$DATA'
//
bool isFileStream( const wchar_t *path, int *pStreamPos, wchar_t *filePath, wchar_t *streamName )
{
	int startPos = splitPath( path );

	for ( int i = startPos; path[ i ] != L'\0'; i ++ )
	{
		if ( path[i] == L':' )
		{
			if ( pStreamPos != NULL )
				*pStreamPos = i;

			if ( filePath != NULL )
			{
				int len = i;

				if ( len > MAX_PATH )
				{
					ODS_fls( L"Assertion!" );
					len = MAX_PATH - 1;
				}

				wcsncpy( filePath, path, len );
				filePath[ len ] = L'\0';
			}

			if ( streamName != NULL )
			{
				wcsncpy( streamName, path + i, MAX_PATH - 1 );
			}

			if ( wcscmp( path + i, L"::$DATA" ) == 0 ) // Main stream
				return false;

			return true;
		}
	}

	return false;
}

//
//	splitPath
//	dir and file buffers may be NULL otherwise they are to be MAX_PATH buffers.
//
int splitPath( const wchar_t *path, wchar_t *dir, wchar_t *file )
{
	int len = (int)wcslen( path );

	while( len-- )
	{
		if ( path[ len ] == L'\\' )
		{
			if ( dir != NULL )
			{
				if ( len >= MAX_PATH - 1 )
				{
					ODS_fls( _FormatT( L"Assertion! Path name '%s' is too long %d", path, len ) );
					return 0;
				}

				wcsncpy( dir, path, len );
				dir[ len ] = L'\0';
			}

			len++;

			if ( file != NULL )
			{
				int tailLen = (int)wcslen( path + len );
				if ( tailLen >= MAX_PATH - 1  )
				{
					ODS_fls( _FormatT( L"File name '%s' is too long '%d'", path + len, tailLen ) );
					return 0;
				}

				//wcscpy_s( file, MAX_PATH, path + len );
				wcsncpy( file, path + len, MAX_PATH - 1 );
			}

			return len;
		}
	}

	// ODS_fls( L"Assertion!" );

	if ( dir != NULL ) dir[0] = 0;
	if ( file != NULL )
	{
	//	wcscpy_s( file, MAX_PATH, path );
		wcsncpy( file, path, MAX_PATH - 1 );
	}

	return 0;
}

int splitPath( const wchar_t *path, cWCS *dir, cWCS *file )
{
	int lenAll = (int)wcslen( path );

	int pos = lenAll;

	while( pos-- )
	{
		if ( path[ pos ] == L'\\' )
		{
			if ( dir != NULL )
				dir->Set( path, pos );

			pos++;

			if ( file != NULL )
				file->Set( path + pos, lenAll - pos );

			return pos;
		}
	}

	// ODS_fls( L"Assertion!" );

	if ( dir != NULL )
		dir->SetLength( 0 );

	if ( file != NULL )
		file->Set( path, lenAll );

	return 0;
}

void makePath( wchar_t *path, const wchar_t *drive, const wchar_t *dir, const wchar_t *fname, const wchar_t *ext )
{
	int pos = 0;
	memset( path, 0, MAX_PATH * sizeof( wchar_t ) );

	if ( drive	!= NULL ) { wcsncpy( path + pos, drive, MAX_PATH - 1 - pos ); pos = (int)wcslen( path ); if ( ( path[ pos - 1 ] != L'\\' ) && ( pos < MAX_PATH - 1 ) ) { wcsncpy( path + pos, L"\\", MAX_PATH - 1 - pos ); pos++; }}
	if ( dir	!= NULL ) { wcsncpy( path + pos, dir,	MAX_PATH - 1 - pos ); pos = (int)wcslen( path ); if ( ( path[ pos - 1 ] != L'\\' ) && ( pos < MAX_PATH - 1 ) ) { wcsncpy( path + pos, L"\\", MAX_PATH - 1 - pos ); pos++; } }
	if ( fname	!= NULL ) { wcsncpy( path + pos, fname, MAX_PATH - 1 - pos ); pos = (int)wcslen( path ); }
	if ( ext	!= NULL )
	{
		if ( ( path[ pos - 1 ] != L'.' ) && ( ext[0] != L'.' ) )
			{ wcsncpy( path + pos, L".",	MAX_PATH - 1 - pos ); pos = (int)wcslen( path ); }

		 { wcsncpy( path + pos, ext,	MAX_PATH - 1 - pos ); pos = (int)wcslen( path ); }
	}
//	ODS_fls( L"Assertion!" );
}

int SendStringMessage_Get( HWND hwndCtl, UINT uMsg, WPARAM wParam, wchar_t *lpsz, int size )
{
	return ((int)(DWORD)SendMessageW((hwndCtl), uMsg, wParam, (LPARAM)(lpsz)));
}

int SendStringMessage( HWND hwndCtl, UINT uMsg, WPARAM wParam, const wchar_t *lpsz )
{
	return ((int)(DWORD)SendMessageW((hwndCtl), uMsg, wParam, (LPARAM)(LPCWSTR)(lpsz)));
}

int SendStringMessage( HWND hwndCtl, UINT uMsgA, UINT uMsgW, WPARAM wParam, const wchar_t *lpsz )
{
	return ((int)(DWORD)SendMessageW((hwndCtl), uMsgW, wParam, (LPARAM)(LPCWSTR)(lpsz)));
}

BOOL TreeView_GetItemTextT( HWND hTree, HTREEITEM hItem, wchar_t *text, int textMaxLen )
{
	if ( hItem == TVI_ROOT )
	{
		wcsncpy( text, L"TVI_ROOT", textMaxLen );
		return true;
	}

	TV_ITEMW tvi;
	memset( &tvi, 0, sizeof( tvi ) );
	tvi.hItem = hItem;
	tvi.mask = TVIF_TEXT;
	tvi.pszText = text;
	tvi.cchTextMax = textMaxLen;

	return TreeView_GetItemT( hTree, &tvi );
}

BOOL TreeView_SetItemTextT( HWND hTree, HTREEITEM hItem, const wchar_t *text )
{
	TV_ITEMW tvi;
	memset( &tvi, 0, sizeof( tvi ) );
	tvi.hItem = hItem;
	tvi.mask = TVIF_TEXT;
	tvi.pszText = (wchar_t *)text;
	tvi.cchTextMax = (int)wcslen( text );

	return TreeView_SetItemT( hTree, &tvi );
}

LPARAM TreeView_GetItemData( HWND hTree, HTREEITEM hItem )
{
	TV_ITEM tvi;
	memset( &tvi, 0, sizeof( tvi ) );
	tvi.hItem = hItem;
	tvi.mask = TVIF_PARAM;

	if ( !TreeView_GetItem( hTree, &tvi ) )
	{
//		ODS_fls( L"Assertion!" );
		return 0;
	}

	return tvi.lParam;
}

BOOL TreeView_SetItemData(
		HWND hTree,
		HTREEITEM hItem,
		DWORD data,
		int iImage, // = -1,
		int iSelectedImage // = -1
	)
{
	TV_ITEM tvi;
	memset( &tvi, 0, sizeof( tvi ) );
	tvi.hItem = hItem;
	tvi.mask = TVIF_PARAM;
	tvi.lParam = data;

	if ( iImage != -1 )
	{
		tvi.mask |= TVIF_IMAGE;
		tvi.iImage = iImage;
	}

	if ( iSelectedImage != -1 )
	{
		tvi.mask |= TVIF_SELECTEDIMAGE;
		tvi.iSelectedImage = iSelectedImage;
	}

	return TreeView_SetItem( hTree, &tvi );
}


int ComboBoxEx_GetItemTextT( HWND hCmb, int i, wchar_t *wcs, int sizeInChar )
{
	COMBOBOXEXITEMW cbei;
	memset( &cbei, 0, sizeof( cbei ) );
	cbei.mask = CBEIF_TEXT;
	cbei.pszText = wcs;
	cbei.cchTextMax = sizeInChar;
	cbei.iItem = i;

	i = (int)SendMessageW( hCmb, CBEM_GETITEMW, 0, (LPARAM)&cbei );
	return i;
}

int ComboBoxEx_GetTextT( HWND hCmb, wchar_t *wcs, int sizeInChar )
{
	int curSel = ComboBox_GetCurSel( hCmb );
	return ComboBoxEx_GetItemTextT( hCmb, curSel, wcs, sizeInChar );
}

int ComboBoxEx_InsertStringT( HWND hCmb, int i, const wchar_t *wcs, int indent )
{
	COMBOBOXEXITEMW cbei;
	memset( &cbei, 0, sizeof( cbei ) );
	cbei.mask = CBEIF_TEXT;
	cbei.pszText = (wchar_t *)wcs;
	cbei.iItem = i;

	if ( indent != 0 )
	{
		cbei.mask |= CBEIF_INDENT;
		cbei.iIndent = indent;
	}

	i = (int)SendMessageW( hCmb, CBEM_INSERTITEMW, 0, (LPARAM)&cbei );

	return i;
}

int ComboBoxEx_SetItemTextT( HWND hCmb, int i, const wchar_t *wcs, int indent )
{
	int count = ComboBox_GetCount( hCmb );
	if ( count <= i )
		return ComboBoxEx_InsertStringT( hCmb, -1, wcs, indent );

	COMBOBOXEXITEMW cbei;
	memset( &cbei, 0, sizeof( cbei ) );
	cbei.mask = CBEIF_TEXT;
	cbei.pszText = (wchar_t *) wcs;
	cbei.iItem = i;
	if ( indent != 0 )
	{
		cbei.mask |= CBEIF_INDENT;
		cbei.iIndent = indent;
	}

	i = (int)SendMessageW( hCmb, CBEM_SETITEMW, 0, (LPARAM)&cbei );

	return i;
}

bool GetModuleFileNeighbour( cWCS &wcs, const wchar_t *fileName )
{
	if ( wcs.SetLength( MAX_PATH ) == 0 )
	{ ODS_fls( L"Assertion!" ); return false; }

	if ( !GetModuleFileNameT( NULL, wcs.m_szBody, wcs.m_len ) )
	{ ODS_fls( L"Assertion!" ); return false; }

	int pos = splitPath( wcs.m_szBody );
	if ( pos == 0 )
	{ ODS_fls( L"Assertion!" ); return false; }

	wcs.SetLength( pos );
	wcs.Add( fileName );

	return FileExistsT( wcs.m_szBody ) == TRUE;
}

int
MessageBoxT(
    IN HWND hWnd,
    IN LPCWSTR lpText,
    IN LPCWSTR lpCaption,
    IN UINT uType)
{
	ODSFuncS( lpText );

#ifndef _NoTranslation_
	if  ( ( pTranslateDll != NULL ) &&
		  ( !pTranslateDll->error ) &&
		  pTranslateDll->m_flipDialog
		  )
	{
		uType |= MB_RTLREADING | MB_RIGHT;
	}
#endif
		
	if ( ( hWnd != 0 ) && !IsWindow( hWnd ) )
	{
		ODS_fls( L"Assertion! Invalid parent window handle" );
		hWnd = GetDesktopWindow();
	}
	
	int res = MessageBoxW( hWnd, lpText, lpCaption, uType );

	ODSFuncRes(
		res == IDOK		? L"IDOK"		:
		res == IDCANCEL	? L"IDCANCEL"	:
		res == IDABORT	? L"IDABORT"	:
		res == IDRETRY	? L"IDRETRY"	:
		res == IDIGNORE	? L"IDIGNORE"	:
		res == IDYES	? L"IDYES"		:
		res == IDNO		? L"IDNO"		:
		_FormatT( L"0x%X", res )
	);

	return res;
}

DWORD GetCurrentDirectoryT( cWCStr *buffer )
{
	DWORD len = GetCurrentDirectoryT( 0, NULL );

	if ( len == 0 )
		return 0;

	if ( buffer->SetLength( len + 1 ) == 0 )
		return 0;

	len = GetCurrentDirectoryT( buffer->m_len, buffer->m_szBody );

	buffer->SetLength( len );

	return len;
}

static int getTimeDivider( time_t timeType )
{
	switch( timeType )
	{
		case timeSec:	return 60;
		case timeMin:	return 60;
		case timeHour:	return 24;
		case timeDay:	return 7;
		case timeWeek:	return 4;
		default:		return 12;
	}
}

static const int timeValMax = 5;
static const int timeValMaxLen = 20;
static const wchar_t timeName[ timeValMax ][ timeValMaxLen ] = { L"Sec", L"Min", L"Hour", L"Day", L"Week" };
static const wchar_t timeNameNumerous[ timeValMax ][ timeValMaxLen ] = { L"Secs", L"Mins", L"Hours", L"Days", L"Weeks" };
static const wchar_t *szLessTheOneSecond = L"Less than 1 Sec";

//
//	Note, the timeNameNumerousUnusual is used for Russian ( 0x419 ) locale only 
//	if corresponded translation is fund. See getNumerousType and getTimeName.
//	Russian lang file has to contain the additional string, see !Translation strategy.txt VI. Time value translation.
//	1 Secunda - none
//	2 Secundy - usual
//  5 Secund  - unusual
//
wchar_t timeNameNumerousUnusual[ timeValMax ][ timeValMaxLen ] = { L"SecS", L"MinS", L"HourS", L"DayS", L"WeekS" };

//#ifdef _DEBUG
//static bool bTimeNameAssigned = false;
//#endif

static int timeLocaleNumber = 0;

static eNumerousType getNumerousType( DWORD value, DWORD rest = 0 )
{

#ifndef _NoTranslation_
	if ( ( pTranslateDll == NULL ) || pTranslateDll->error )
		timeLocaleNumber = 0;
	else
		timeLocaleNumber = pTranslateDll->GetLocaleNumber();
#endif

	if ( timeLocaleNumber == 0x419 ) // Russian 
	{
		if ( value >= 20 )		// f.e. 22 chasa, 2 chasa, 12 chasov
			value = value % 10;

		if ( value > 4 )
			return entUnusual;	// 5-19.9 chasov

		if ( ( value == 0 ) && ( rest == 0 ) )
			return entUnusual; // 0.0, 20, 200, 1000 chasov

		if ( ( value == 1 ) && ( rest == 0 ) ) // 1 chas
			return entNone;

		return entUsual;
	}

	if ( ( value > 1 ) || ( ( value == 1 ) && ( rest > 0 ) ) || ( ( value == 0 ) && ( rest == 0 ) ) )
		return entUsual;
	
	return entNone;
}

static const wchar_t *translateStatic( const wchar_t *tName )
{
#ifdef _NoTranslation_
	return tName;
#else
	if ( ( pTranslateDll == NULL ) || ( pTranslateDll->error ) )
		return tName;

	static cADTable < cAutoDeallocate < wchar_t > > adT( 3 );
	static int pos = 0;

	pos = ( pos + 1 ) % adT.m_Q;

	if ( adT.m_T[ pos ].Allocate( 64 ) == NULL )
		return tName;

	bool found = false;

	pTranslateDll->Translate( adT.m_T[ pos ].m_ptr, adT.m_T[ pos ].m_len, getMBStr( tName ), 0, &found );

	// See notes at timeNameNumerousUnusual declaration
	if ( !found )
		return tName; // timeNameNumerous[ timeType ];

	return adT.m_T[ pos ].m_ptr;
#endif
}

static const wchar_t *getTimeName( int timeType, int numerous )
{
	if ( timeType > timeValMax )
		return L"XX";

	const wchar_t *tName = NULL;

	if ( numerous == entUnusual )
		tName =  timeNameNumerousUnusual[ timeType ];
	else
	if ( numerous == entUsual )
		tName = timeNameNumerous[ timeType ];
	else
		tName = timeName[ timeType ];

	return translateStatic( tName );
}

//static const wchar_t *getTimeNameOld( int timeType, int numerous )
//{
//	//
//	//	Simple check assigned translated values
//	//	see SetTimeTextValues
//	//
//#ifdef _DEBUG
//	assert( // ( pTranslateDll == NULL ) ||
//		  bTimeNameAssigned );
//#endif
//
//	if ( timeType > timeValMax )
//		return L"XX";
//
//	if ( numerous == entUnusual )
//		return timeNameNumerousUnusual[ timeType ];
//	if ( numerous == entUsual )
//		return timeNameNumerous[ timeType ];
//	else
//		return timeName[ timeType ];
//}

int StrFromTimeIntervalT(LPWSTR pszOut, UINT cchMax, DWORD dwTimeMS, int digits)
{
	static CDll dll( false, "shlwapi.dll", "StrFromTimeIntervalW", false );
	if ( dll.error )
		return 0;
	
	typedef int STDAPICALLTYPE typeStrFromTimeIntervalW(LPWSTR pszOut, UINT cchMax, DWORD dwTimeMS, int digits);
	return ((typeStrFromTimeIntervalW *)dll.pProc)( pszOut, cchMax, dwTimeMS, digits );
}

void GetTimeText( char *buff, int buffSize, DWORD timeInSec, int digits )
{
	cAutoDeallocate <wchar_t> ad;
	if ( ad.Allocate( buffSize ) == NULL )
		return;

	GetTimeTextT( ad.m_ptr, buffSize, timeInSec, digits );
	size_t size;
	wcstombs_s( &size, buff, buffSize, ad.m_ptr, _TRUNCATE );
}

const wchar_t *GetTimeText( time_t timeInSec, int digits )
{
	static cWCStr wcs( 64 );
	if ( wcs.m_szBody == NULL )
		return L"Null";

	GetTimeTextT( wcs.m_szBody, wcs.m_len, timeInSec, digits );
	return wcs.m_szBody;
}

const wchar_t *cWCStr::GetTimeText( time_t timeInSec, int digits )
{
	if ( SetLength( 64 ) == 0 )
		return L"Null";

	GetTimeTextT( m_szBody, m_len, timeInSec, digits );

	return m_szBody;
}

//
//	Sample:
//
//	SetTimeTextValues( false, timeSec,	m_hInst, IDS_szSec );
//	SetTimeTextValues( false, timeMin,	m_hInst, IDS_szMin );
//	SetTimeTextValues( false, timeHour, m_hInst, IDS_szHour );
//	SetTimeTextValues( false, timeDay,	m_hInst, IDS_szDay );
//	SetTimeTextValues( false, timeWeek,	m_hInst, IDS_szWeek );
//
//	SetTimeTextValues( true, timeSec,	m_hInst, IDS_szSecs );
//	SetTimeTextValues( true, timeMin,	m_hInst, IDS_szMins );
//	SetTimeTextValues( true, timeHour,	m_hInst, IDS_szHours );
//	SetTimeTextValues( true, timeDay,	m_hInst, IDS_szDays );
//	SetTimeTextValues( true, timeWeek,	m_hInst, IDS_szWeeks );
//

//void SetTimeTextValues( bool numerous, HINSTANCE hInst, int *idsT, int localeNumber )
//{
//	for( int ett = timeSec; ett <= timeLast; ett = ett + 1 )
//		SetTimeTextValues( numerous, eTimeType( ett ), hInst, idsT[ ett ], localeNumber );
//}
//
//void SetTimeTextValues( bool numerous, eTimeType timeType, wchar_t *szValue, int localeNumber )
//{
//#ifdef _DEBUG
//	bTimeNameAssigned = true;
//#endif
//
//	if ( numerous )
//	{
//		if ( localeNumber != 0 ) // 0x419;
//		{
//			timeLocaleNumber = localeNumber;
//			wcsncpy( timeNameNumerousUnusual[ timeType ], szValue,	timeValMaxLen );
//		}
//		else
//			wcsncpy( timeNameNumerous[ timeType ], szValue,	timeValMaxLen );
//	}
//	else
//		wcsncpy( timeName[ timeType ], szValue,	timeValMaxLen );
//}
//
//void SetTimeTextValues( bool numeroues, eTimeType timeType, HINSTANCE hInst, int idsValue, int localeNumber )
//{
//#ifdef _DEBUG
//	bTimeNameAssigned = true;
//#endif
//
//	if ( numeroues )
//	{
//		if ( localeNumber != 0 ) // 0x419;
//		{
//			timeLocaleNumber = localeNumber;
//			LoadStringT( hInst, idsValue, timeNameNumerousUnusual[ timeType ], timeValMaxLen );
//		}
//		else
//			LoadStringT( hInst, idsValue, timeNameNumerous[ timeType ], timeValMaxLen );
//	}
//	else
//		LoadStringT( hInst, idsValue, timeName[ timeType ], timeValMaxLen );
//
//}

static int getDigitsCount( time_t value )
{
	int i = 0;
	for(; value > 0; i++ )
		value /= 10;

	return i;
}

static time_t roundLast( time_t value, int digits )
{
	int i;

	for( i = digits; i--; )
	{
		int add = ( (value % 10) >= 5 );
		value = value / 10 + add;
	}

	for( i = digits; i--; )
		value *= 10;

	return value;
}


void GetTimeTextT( wchar_t *buff, int buffSize, time_t timeInSec, int digits )
{
	if ( timeInSec == 0 )
	{
		FormatT( buff, buffSize, translateStatic( szLessTheOneSecond ) );
		return;
	}
	//
	//	There are standard way to convert time values to string StrFromTimeInterval
	//	The method is used if time values are not transalted and time value is less the 24 hours.
	//	Because StrFromTimeInterval converts miliseconds to hours as max and maximal time valus are
	//	0x7fffFFFF msecs -> 586 hours
	//

	bool found = true;
	bool bRTL = false;

#ifndef _NoTranslation_
	if  ( ( pTranslateDll != NULL ) && ( !pTranslateDll->error ) )
	{
		bRTL = pTranslateDll->m_flipDialog;

		wchar_t bf[ 64 ];
		pTranslateDll->Translate( bf, maxLen( bf ), getMBStr( timeNameNumerous[4] ), 0, &found );
	}
#endif

	bool timeToLarge = timeInSec / 3600 > 24;

	if ( !timeToLarge && !found )
	{
		if ( digits < 0 )
			digits = 1;

		StrFromTimeIntervalT( buff, buffSize, DWORD(timeInSec * 1000), digits );
		return;
	}

	time_t timeType = timeSec;

	bool useRest = false;
	time_t rest = 0;

	time_t value = timeInSec;

	int timeDivider;

	while ( value >= (time_t)getTimeDivider( timeType ) )
	{
		if ( timeType == timeWeek )
			break;

		timeDivider = getTimeDivider( timeType );

		rest = (time_t)( value % timeDivider );

		value = value / timeDivider;

		timeType++;
	}

	if ( digits < 0 )
	{
		digits = ( value > 2 ? 1 : 2 );
	}

	int count = getDigitsCount( value );

	if ( count >= digits )
	{
		value = roundLast( value, count - digits );
		rest = 0;
	}
	else
	{
		digits = digits - count;
		
		count = getDigitsCount( rest );

		if ( count >= digits )
		{
			rest = roundLast( rest, count - digits  );
			if ( rest == timeDivider )
			{
				value++;
				rest = 0;
			}
			else
			if ( rest < 3 )
				rest = 0;
			else
			if ( rest > timeDivider - 3  )
			{
				value++;
				rest = 0;
			}
		}
	}
	
	useRest = ( rest != 0 ) && ( value < 5 );


	////
	////	Get localised decimal separator...
	////
	//static wchar_t separator = 0;

	//if ( separator == 0 )
	//	if ( GetLocaleInfoT(( LOCALE_USER_DEFAULT, LOCALE_SDECIMAL, &separator, 1 ) == 0 )
	//		separator = L'.';

	wchar_t szRest[ 100 ] = L"";
	wchar_t bf[100];

	const wchar_t *pattern = bRTL ? L"%2 %1" : L"%1 %2";

	if ( useRest )
	{
		pattern = bRTL ? L"%3 %2 %1" : L"%1 %2 %3";

		FormatT( szRest, maxLen( szRest ), pattern,
			_itow( (int)rest, bf, 10 ),
			getTimeName( (int)timeType - 1, getNumerousType( (DWORD)rest ) ),
			L""
			);
	}

	//
	//	Prepare string...
	//
	FormatT( buff, buffSize, pattern,
		_itow( (int)value, bf, 10 ),
		getTimeName( (int)timeType, getNumerousType( (DWORD)value ) ),
		szRest
		);
}

static char getSeparator()
{
	static char separator = 0;

	char localeStr[ 100 ];

	if ( separator == 0 )
	{
		int err = NO_ERROR;
		if ( GetLocaleInfoA( LOCALE_USER_DEFAULT, LOCALE_SDECIMAL, localeStr, maxLen( localeStr ) ) == 0 )
		{
			err = GetLastError();
		}

		if ( err != NO_ERROR )
		{
			separator = '.';
		}
		else
		{
			separator = localeStr[0];
		}
	}

	return separator;
}

char *GetSizeText( char *buff, int buffSize, DWORDLONG dwlSize, bool sizeInBytes )
{
	if ( buff == NULL ) return "NULL";

	DWORD rest = 0;

	enum { sizeBt = 0, sizeKB, sizeMB, sizeGB, sizeTB };

	int sizeType = sizeBt;

	BOOL useRest = FALSE;

	if ( !sizeInBytes )
	{
		while ( dwlSize >= 1024 )
		{
			if ( sizeType == sizeTB )
				break;

			rest = (DWORD)( dwlSize % 1024 );

			if ( !useRest )
				useRest = (rest != 0);

			dwlSize = dwlSize >> 10;

			rest = ( rest * 100 ) / 1024; // rest < 1024 ==>  ( rest * 100 / 1024 < 100 ) -> The range is 0..99

			// Rounds 0,05 to upper 0,1
			rest = ( rest + 5 ) / 10;

			sizeType++;
		}
	}

	if ( useRest && ( rest >= 10 ) )
	{
		dwlSize++;
		rest = 0;
	}

	//
	//	Get localised decimal separator...
	//
	char separator = getSeparator();

	//
	//	Prepare one figure decimal reminder...
	//
	char szRest[ 3 ] = ".0";

	if ( useRest )
	{
		szRest[ 0 ] = separator;
		szRest[ 1 ] = '0' + (int)rest;
	}
	else
		szRest[ 0 ] = 0;

	const char *typeStr = 
		sizeType == sizeBt ? "B" :
		sizeType == sizeKB ? "KB" :
		sizeType == sizeMB ? "MB" :
		sizeType == sizeGB ? "GB" :
		sizeType == sizeTB ? "TB" :
		"XX";

	//
	//	Prepare string...
	//	
	sprintf( buff, "%I64d%s %s", (__int64)dwlSize, szRest, typeStr );

	return buff;
}

bool isDigital( wchar_t wc, DWORD *pVal )
{
	if ( ( wc < L'0') || ( wc > L'9') )
		return false;

	if ( pVal != NULL )
		*pVal = wc - L'0';

	return true;
}

//
//	Do not return error if found probels at tail
//
static int defDimensionN = 0;
static DWORDLONG getDimensionSizeInBytes( const wchar_t *dimension )
{
	while( *dimension == L' ' ) dimension++;

	struct sDimensionSize { const wchar_t *pref; const wchar_t *translated; DWORDLONG size ; };

	sDimensionSize T[] = 
	{
		L"B",  L"", 1,
		L"KB", L"", 1024,
		L"MB", L"", 1024*1024,
		L"GB", L"", 1024*1024*1024,
		L"TB", L"", (DWORDLONG)1024*1024*1024*1024,
	};

	size_t len = wcslen( dimension );

	while( dimension[ len - 1 ] == L' ' ) len--;

	if ( len == 0 )
		return T[ defDimensionN ].size; // default dimension size

	//
	//	The KB,MB.. prefixes at size may be translated ( see GetSizeTextT - Translation note )
	//
	cAutoDeallocate < wchar_t > adT[ maxLen( T ) ];

#ifndef _NoTranslation_
	if ( pTranslateDll != NULL )
	{
		for( int i = 0; i < maxLen( T ); i++ )
		{
			if ( adT[i].Allocate( 64 ) == NULL )
				break;

			bool found = false;

			if ( pTranslateDll->Translate( adT[i].m_ptr, adT[i].m_len, getMBStr( T[i].pref ), 0, &found ) && found )
				T[i].translated = adT[i].m_ptr;
		}
	}
#endif

	for ( int i = maxLen( T ); i--; )
	{
		if ( _wcsnicmp( T[i].pref, dimension, len ) == 0 )
			return T[i].size;

		if ( _wcsnicmp( T[i].translated, dimension, len ) == 0 )
			return T[i].size;
	}

	return -1;
}

//
//	GetSizeFromText
//		Parses string like '512.5 KB' to DWORDLONG 512 * 1024 + 256
//		Allows probels between figures and dimension -> 1 512 MB
//		Return -1 if text was not properly parsed.
//
DWORDLONG GetSizeFromText( const wchar_t *text, DWORDLONG *pDimention )
{
	DWORDLONG val = 0;
	DWORDLONG dimensionByteSize = 1;
	DWORDLONG tailVal = 0;	// if dimension is MB, so value = value * 1024  + tailVal * 1024 / tailOrd;
	DWORDLONG tailOrd = 1;	// 

	int len = (int)wcslen( text );

	int pos = 0;

	const wchar_t *b = text;

	bool wasSeparator = false;

	for( ; pos < len; pos++ )
	{
		DWORD cur;

		if ( b[pos] == L' ' ) continue;

		if ( isDigital( b[pos], &cur  ) )
		{
			// Accumulates figure
			val = val * 10 + cur;
			continue;
		}
		else
		if ( ( b[pos] == L'.' ) || 
			 ( b[pos] == getSeparator() ) )
		{
			// Separator was already encountered.
			if ( wasSeparator )
				return -1;

			wasSeparator = true;

			// Accumulates tail after separator
			for( ; ++pos < len; )
			{
				if ( !isDigital( b[pos], &cur ) )
					break;

				tailOrd *= 10;
				tailVal = tailVal * 10 + cur;
			}
		}

		break;
	}

	while ( b[pos] == L' ' ) pos++;
	
	dimensionByteSize = getDimensionSizeInBytes( b + pos );

	if ( dimensionByteSize == -1 )
		return -1;

	tailVal = tailVal * dimensionByteSize / tailOrd;

	val = val * dimensionByteSize + tailVal;

	if ( ( dimensionByteSize > 1 ) && ( tailVal != 0 ) ) // Exludes error in rounding
		val += 1;

	if ( pDimention != NULL ) *pDimention = dimensionByteSize;

	return val;
}

char *GetSizeText( char *buff, int buffSize, DWORD sizeLow, DWORD sizeHigh )
{
	ULARGE_INTEGER ul;
	ul.LowPart = sizeLow;
	ul.HighPart = sizeHigh;
	return GetSizeText( buff, buffSize, ul.QuadPart );
}

wchar_t *GetSizeTextT( wchar_t *buff, int buffSize, DWORD sizeLow, DWORD sizeHigh )
{
	ULARGE_INTEGER uli;
	uli.LowPart = sizeLow;
	uli.HighPart = sizeHigh;
	return GetSizeTextT( buff, buffSize, uli.QuadPart );
}

//
//	GetSizeTextT - Translation note.
//	See !TranslationStrategy.txt VII.
//		The automatic right to left text transformation ( f.e. arabic ) adds KB,MB.. prefix 
//		to unpredictable place of common string. The prefix has to be translated for proper viewing.
//		F.e.  'min - 100 KB' looks '100 - aaa KB' where 'aaa' is arabic 'min', if we translate 'KB' 
//		to arabic 'bbb' it looks as 'bbb 100 - aaa'.
//
//	The GetSizeFromText function uses the translation feature too.
//
//	It works if following strings are added to langFile.txt and translated
//		"B"="B"
//		"KB"="KB"
//		"MB"="MB"
//		"GB"="GB"
//		"TB"="TB"
//
wchar_t *GetSizeTextT( wchar_t *buff, int buffSize, DWORDLONG size, bool sizeInBytes )
{
	if ( ( buffSize == 0 ) || ( buff == NULL ) )
	{
		ODS_fls( L"Assertion!" );
		return NULL;
	}

	buff[0] = L'\0';

	cAutoDeallocate < char > adbf;

	if ( adbf.Allocate( buffSize ) == NULL )
		return buff;
	
	GetSizeText( adbf.m_ptr, adbf.m_len, size, sizeInBytes );

#ifndef _NoTranslation_
	if ( pTranslateDll != NULL )
	{
		// sample buff == L"150 MB"
		char *ptr = adbf.m_ptr;
		for( ; *ptr; ptr++ )
		{
			if ( *ptr == ' ' ) { ptr++; break; }
			if ( ( *ptr >= '0' ) && ( *ptr <= '9' ) )
				continue;
		}

		if ( *ptr != L'\0' )
		{
			cWCStr wcs( buffSize );
			bool found = false;

			if ( pTranslateDll->Translate( wcs.m_szBody, wcs.m_len, ptr, 0, &found ) && found )
			{
				ptr[0] = '\0';
				size_t lsize;
				mbstowcs_s( &lsize, buff, buffSize, adbf.m_ptr, _TRUNCATE );
				wcscat_s( buff, buffSize, wcs.m_szBody );
				return buff;
			}
		}
	}
#endif

	size_t lsize;
	mbstowcs_s( &lsize, buff, buffSize, adbf.m_ptr, _TRUNCATE );

	return buff;
}

wchar_t *GetSizeTextOnceT( DWORDLONG size )
{
	static cAutoDeallocate < wchar_t > ad;

	int maxStrLen = 80;

	if ( ad.Allocate( maxStrLen ) == NULL ) return L"<NULL>";
	
	return GetSizeTextT( ad.m_ptr, maxStrLen, size );
}

char *getFileInfo( const char *fileName, char *dateStr, int dateStrSize, char *sizeStr, int sizeStrSize )
{
	WIN32_FIND_DATAA findData;
	HANDLE hFile = FindFirstFileA( fileName, &findData );

	if ( hFile == INVALID_HANDLE_VALUE )
	{
		MessageBoxT( NULL,
			_FormatT(
				L"File '%s'\n"
				L"does not exist or access is denied",
				_getWcs( fileName ) ),
			L"Error in reading file", MB_OK | MB_ICONINFORMATION | MB_TOPMOST );

		ODS_fls( L"Assertion!" );

		strncpy( dateStr, "Undefined", dateStrSize );

		if ( sizeStr )
			strncpy( sizeStr, "Undefined", sizeStrSize );

		return dateStr;
	}

	FindClose( hFile );

	FILETIME LastWriteTime	= findData.ftLastWriteTime;

	SYSTEMTIME sysTime;
	FILETIME localTime;

	FileTimeToLocalFileTime( &LastWriteTime, &localTime );
	FileTimeToSystemTime( &localTime, &sysTime );

	sprintf( dateStr, "%02d.%02d.%04d %02d:%02d",
		sysTime.wDay,
		sysTime.wMonth,
		sysTime.wYear,
		sysTime.wHour,
		sysTime.wMinute);

	if ( sizeStr )
		GetSizeText( sizeStr, sizeStrSize, findData.nFileSizeLow, findData.nFileSizeHigh );

	return dateStr;
}

wchar_t *getFileInfoT( const wchar_t *fileName, wchar_t *dateStr, int dateStrSize, wchar_t *sizeStr, int sizeStrSize )
{
	WIN32_FIND_DATAW findData;
	HANDLE hFile = FindFirstFileT( fileName, &findData );

	if ( hFile == INVALID_HANDLE_VALUE )
	{
		MessageBoxT( NULL, _FormatT( L"File '%s'\ndoes not exist or access is denied.", fileName ), L"Error in reading file", MB_OK | MB_ICONINFORMATION | MB_TOPMOST );

		ODS_fls( L"Assertion!" );

		copyString( dateStr, L"Undefined", dateStrSize );

		if ( sizeStr )
			copyString( sizeStr, L"Undefined", sizeStrSize );

		return dateStr;
	}

	FindClose( hFile );

	FILETIME LastWriteTime	= findData.ftLastWriteTime;

	SYSTEMTIME sysTime;
	FILETIME localTime;

	FileTimeToLocalFileTime( &LastWriteTime, &localTime );
	FileTimeToSystemTime( &localTime, &sysTime );

	char dateA[ 128 ];
	sprintf( dateA, "%02d.%02d.%04d %02d:%02d",
		sysTime.wDay,
		sysTime.wMonth,
		sysTime.wYear,
		sysTime.wHour,
		sysTime.wMinute);

	size_t size;
	mbstowcs_s( &size, dateStr, dateStrSize, dateA, _TRUNCATE );

	if ( sizeStr )
		GetSizeTextT( sizeStr, sizeStrSize, findData.nFileSizeLow, findData.nFileSizeHigh );

	return dateStr;
}

bool GetVolumeName( wchar_t *volumeName, int length, const wchar_t *mountPoint, DWORD *pErr )
{
	ODSFuncS( mountPoint );

	// assertion to check the proc only
	// It was checked but it does not work for bc mount point
	// We have to use CDiskDriver::GetVolumeName instead ( it uses  DeviceIoControl with FSCTL_GET_REPARSE_POINT )
//	ODS_fls( L"Assertion!" );

	DWORD err = -1;
	if ( pErr == NULL )
		pErr = &err;

	int len = (int)wcslen( mountPoint );

	if ( len == 0 )
	{
		*pErr = -1;
		return false;
	}

	if ( len >= MAX_PATH - 1 )
	{
		*pErr = -2;
		return false;
	}

	wchar_t mp[ MAX_PATH + 1 ];
	copyString( mp, mountPoint, MAX_PATH );

	if ( mp[len - 1] != L'\\' )
	{
		if ( len >= MAX_PATH - 2 )
		{
			*pErr = -3;
			return false;
		}

		copyString( mp + len, L"\\", MAX_PATH - len );
		len++;
	}

	if ( GetVolumeNameForVolumeMountPointT( mp, volumeName, length ) )
		return true;
	else
	{
		*pErr = GetLastError();
		ODS_fls( _FormatT( L"Assertion! GetVolumeNameForVolumeMountPointT( %s ) returns error %s", mp, _GetSystemError( *pErr ) ) );
	}

	if ( ( len != 3 ) || ( wcscmp( mp + 1, L":\\" ) != 0 ) )
		return false;

	if ( length < 6 )
		return false;

	// the name is usual drive name

	copyString( volumeName, L"\\\\.\\A:", MAX_PATH );
	volumeName[ 4 ] = towupper( mp[ 0 ] );

	if ( ( volumeName[ 4 ] > L'Z' ) || ( volumeName[ 4 ] < L'A' ) )
		return false;

	//ODSFuncRes( true );
	ODS_fls( _FormatT( L"Warning! Usual drive name is used. Volume name '%s'", volumeName ) );
					
	return true;
}


static char *getCharBuf( const wchar_t *wcs, char *s, int len )
{
	if ( wcs == NULL )	return NULL;
	size_t size;
	wcstombs_s( &size, s, len, wcs, _TRUNCATE );
	return s;
}


#ifndef _NoElevation
static bool runAsLimitedUser( const wchar_t *cmdLine, const wchar_t *taskName, DWORD timeout, DWORD *pExistCode )
{
	ODSFuncS( _FormatT( L"%s, %s", cmdLine, taskName ) );

	HANDLE hProcess;
	if ( !isWindowsVista() || isWindowsVistaLimitedProcess() ) // Windows XP and early version.
	{
		//
		//	The problem with windows XP ( see description below ) has simple solution.
		//	It is not required to use scheduled task facility to start limited user in XP at all.
		//	There are not limited users at XP.
		//
		//
		//	We just have to change single quotes (') to (") or remove escape symbol before double quotes (\") 
		//	SCHTASKS allows '%s' or \"%s\" for long name at task
		//
		cWCStr wcs( cmdLine );

		wchar_t *s = wcs.m_szBody;

		for ( int i = 0; s[i] != L'\0'; i++ )
		{
			if ( s[i] == L'\'' )
				s[i] = L'\"';
			else
			if ( ( s[i] == L'\\' ) && ( s[i+1] == L'\"' ) )
				memmove( s + i, s + i + 1, (wcs.m_len - i) * sizeof( wchar_t ) );
		}

		hProcess = RunProcessWithWaitingT( s, false, false );
		CloseHandle( hProcess );
		return true;
	}

	if ( taskName == NULL )
		taskName = L"BCTempTask";

	//
	//	Windows XP hangs in creating task :(
	//	We'll try to delete the task first.
	//	The deleting did not help it reported about timeout. See wise solution above :)
	//
	hProcess = RunProcessWithWaitingT( _FormatT( L"SCHTASKS /Delete /TN %s /F", taskName ), false, true );
	
	DWORD dwRes = WaitForSingleObject( hProcess, timeout );
	if ( dwRes == WAIT_TIMEOUT )
	{
		ODS_fls( L"Timeout in deleting task" );
		dwRes = WaitForSingleObject( hProcess, timeout );
		
		if ( dwRes != WAIT_TIMEOUT )
		{
			ODS_fls( _FormatT( L"Second time wait = %X", dwRes ) );
			timeout *= 2;
		}
	}

	CloseHandle( hProcess );

	cWCStr schCmd;
	schCmd.FormatT( L"SCHTASKS /Create /SC MONTHLY /TN %s /TR \"%s\" /F", taskName, cmdLine );

	hProcess = RunProcessWithWaitingT( schCmd.m_szBody , false, true );

	dwRes = WaitForSingleObject( hProcess, timeout );
	if ( dwRes == WAIT_TIMEOUT )
		ODS_fls( L"Timeout in creating task" );

	CloseHandle( hProcess );
	
	const wchar_t *szImmediatly = L"/I"; // Run task immeditly 
	
	// 13.08.2010
	if ( !IH.isW7() )
		szImmediatly = L""; // Vista SCHTASKS does not have the /I flag	:(

	hProcess = RunProcessWithWaitingT( _FormatT( L"SCHTASKS /Run %s /TN %s", szImmediatly, taskName ), false, true );

	//
	//	Symbolic waiting allows to start task before deleting.
	//
	dwRes = WaitForSingleObject( hProcess, timeout );

	bool ok = false;

	if ( dwRes == WAIT_TIMEOUT )
		ODS_fls( L"Timeout in running task" );
	else
	{
		DWORD exitCode;
		GetExitCodeProcess( hProcess, &exitCode );
		ODS_fls( _FormatT( L"%X, %X", dwRes, exitCode ) );
		ok = true;
		if ( pExistCode != NULL )
			*pExistCode = exitCode;
	}

	CloseHandle( hProcess );

	hProcess = RunProcessWithWaitingT( _FormatT( L"SCHTASKS /Delete /TN %s /F", taskName ), false, true );
	
	dwRes = WaitForSingleObject( hProcess, timeout );
	if ( dwRes == WAIT_TIMEOUT )
		ODS_fls( L"Timeout in deleting task" );

	CloseHandle( hProcess );

	return ok;
}

bool runAsLimitedUser( const wchar_t *cmdPath, const wchar_t *cmdArgs, const wchar_t *taskName, DWORD timeout, DWORD *pExistCode )
{
	cWCStr cmdLine;
	cmdLine.FormatT( L"\\\"%s\\\"", cmdPath );

	if ( ( cmdArgs != NULL ) && ( cmdArgs[0] != L'\0' ) )
		cmdLine.AddF( L" %s", cmdArgs );

	return runAsLimitedUser( cmdLine.m_szBody, taskName, timeout, pExistCode );
}
#endif

class cAdvAPI : public CDll
{
	SC_HANDLE m_hScManager;
public:

	cAdvAPI() : CDll( false, "advapi32.dll", NULL, false  )
	{
		//ODSFunc;

		m_hScManager = NULL;

		typedef WINADVAPI SC_HANDLE WINAPI typeOpenSCManager( LPCWSTR lpMachineName, LPCWSTR lpDatabaseName, DWORD dwDesiredAccess );

		if ( !GetProcAddress( "OpenSCManagerW", false, NULL ) )
			return;

		m_hScManager = ((typeOpenSCManager*)pProc)( 0, 0, SC_MANAGER_ALL_ACCESS );
	}
	
	bool isAdministrator()
	{
		return m_hScManager != NULL;
	}

	~cAdvAPI()
	{
		if ( m_hScManager == NULL )
			return;

		typedef WINADVAPI BOOL WINAPI typeCloseServiceHandle( SC_HANDLE hSCObject );

		if ( !GetProcAddress( "CloseServiceHandle", false, NULL ) )
			return;

		((typeCloseServiceHandle*)pProc)( m_hScManager );
	}
};

bool isAdministrator()
{
	static bool isNT = ( GetVersion() < 0x80000000 );
	if ( !isNT )
		return true;

	static cAdvAPI dll;
	return dll.isAdministrator();
}

//
//	Set compression of existion file...
//
bool SetCompression( HANDLE hFile, USHORT state  ) // = COMPRESSION_FORMAT_NONE )
{
	DWORD dwBytes = 0;

	BOOL ok = DeviceIoControl(
		hFile,
		FSCTL_SET_COMPRESSION,
		&state,
		sizeof( state ),
		NULL,
		0,
		&dwBytes,
		NULL);

	return ok == TRUE;
}

//
//	createNewRegularFileFile
//		If file is created at folder with ecrypted or compressed attributes so the attributes are automatically inherited.
//		So this function creates new file, checks if it is encrypted or compressed, resets both encryption and compression.
//
int createNewRegularFileFile( const wchar_t *filePath, HANDLE *pHFile ) // = NULL )
{
	WIN32_FIND_DATAW fd;

	if ( FileExistsT( filePath, &fd ) )
		return ERROR_FILE_EXISTS;

	HANDLE hFile =
		CreateFileT( 
			filePath, 
			GENERIC_WRITE | GENERIC_READ,
			FILE_SHARE_READ | FILE_SHARE_WRITE,
			NULL,
			CREATE_NEW,
			FILE_FLAG_NO_BUFFERING,
			NULL);

	if ( hFile == INVALID_HANDLE_VALUE )
		return GetLastError();

    BY_HANDLE_FILE_INFORMATION bhfi;
    if ( !GetFileInformationByHandle(hFile, &bhfi) )
	{
		CloseHandle( hFile );
		return GetLastError();
	}

	if ( bhfi.dwFileAttributes & FILE_ATTRIBUTE_COMPRESSED ) 
	{
		bool ok = ResetCompression( hFile );

		if ( !ok )
		{
			CloseHandle( hFile );
			return GetLastError();
		}
	}

	if ( bhfi.dwFileAttributes & FILE_ATTRIBUTE_ENCRYPTED )
	{
		CloseHandle( hFile );

		if ( !DecryptFileW( filePath, 0 ) )
			return GetLastError();

		if ( pHFile != NULL )
			*pHFile = CreateFileT( 
			filePath, 
			GENERIC_WRITE | GENERIC_READ,
			FILE_SHARE_READ | FILE_SHARE_WRITE,
			NULL,
			OPEN_EXISTING,
			FILE_FLAG_NO_BUFFERING,
			NULL);
	}
	else
	if ( pHFile != NULL )
	{
		*pHFile = hFile;
	}
	else
		CloseHandle( hFile );

	return ERROR_SUCCESS;
}

//
//	createNonCompressedFile
//		Create file and reset compressed attributes if required.
//
HANDLE createNonCompressedFile( const wchar_t *filePath, DWORD dwDisp ) // = OPEN_ALWAYS )
{
	DWORD dwAttr = GetFileAttributesT( filePath );

	if ( dwAttr & FILE_ATTRIBUTE_ENCRYPTED )
	{
		BOOL res = DecryptFileW( filePath, 0 );
		if ( res ) 
			ODS_fls( _FormatT( L"Encrypted file '%s' is decrypted", filePath ) );
		else
			ODS_fls( _FormatT( L"Encrypted file '%s' is NOTE decrypted, err %s", filePath, cWCStr().GetSystemError() ) );
	}

	HANDLE hFile =
		CreateFileT( 
			filePath, 
			GENERIC_WRITE | GENERIC_READ,
			FILE_SHARE_READ | FILE_SHARE_WRITE,
			NULL,
			dwDisp,
			FILE_FLAG_NO_BUFFERING,
			NULL);

	if ( hFile == INVALID_HANDLE_VALUE )
	{
		int err = GetLastError();
		ODS_fls( _FormatT( L"Assertion! Error 0x%X in creating file '%s'", err, filePath ) );
		return NULL;
	}

    BY_HANDLE_FILE_INFORMATION bhfi;
    if ( GetFileInformationByHandle(hFile, &bhfi) )
		dwAttr = bhfi.dwFileAttributes;

	if ( dwAttr & FILE_ATTRIBUTE_COMPRESSED ) 
	{
		bool ok = ResetCompression( hFile );

		if ( !ok )
		{
			CloseHandle( hFile );
			int err = GetLastError();
			ODS_fls( _FormatT( L"Assertion! Error 0x%X in reseting compression on file '%s'", err, filePath ) );
			return NULL;
		}
	}

	return hFile;
}


#ifndef _NoElevation

static BOOL ShellExec(	HWND hwnd, 
				const wchar_t * pszVerb, 
				const wchar_t * pszPath, 
				const wchar_t * pszParameters = NULL,
				const wchar_t * pszDirectory = NULL,
				HANDLE *pHProcess = NULL,
				HINSTANCE *pError = NULL )
{
	ODSFuncS( pszPath );

	SHELLEXECUTEINFOW shex;

	memset( &shex, 0, sizeof( shex) );

	shex.cbSize			= sizeof( shex ); 
	shex.fMask			= ( pHProcess != NULL  ? SEE_MASK_NOCLOSEPROCESS : 0 ); 
	shex.hwnd			= hwnd;
	shex.lpVerb			= pszVerb; 
	shex.lpFile			= pszPath; 
	shex.lpParameters	= pszParameters; 
	shex.lpDirectory	= pszDirectory; 
	shex.nShow			= SW_NORMAL; 
 
	if ( !ShellExecuteExW( &shex ) )
	{
		if ( pError != NULL )
			*pError = (HINSTANCE)(DWORDLONG)GetLastError();

		ODS_fls( L"ShellExecuteExW returns false" );
		return false;
	}

	if ( pHProcess != NULL )
		*pHProcess = shex.hProcess;
	
	if ( pError != NULL )
		*pError = shex.hInstApp;

	//ODSFuncRes( (__int64)shex.hInstApp );

	if ( shex.hInstApp <= (HINSTANCE)32 )
	{
		ODS_fls( _FormatT( L"ShellExecuteExW error %d", shex.hInstApp ) );
		return false;
	}

	return true;
}

bool runElevated( 
	HWND	hwnd, 
	const wchar_t *pszPath, 
	const wchar_t *pszParameters, //	= NULL, 
	const wchar_t *pszDirectory, //	= NULL );
	HANDLE *pHProcess, // = NULL
	int *pError ) // = NULL
{
	ODSFuncS( pszPath );

	//if ( !isWindowsVista() )
	//	return false;

	int res = ShellExec( hwnd, L"runas", pszPath, pszParameters, pszDirectory, pHProcess, (HINSTANCE *)pError );
	
	int err = GetLastError();
	ODS_fls( _FormatT( L"ShellExec = %d, last error %d", res, err ) );
	
	return res == TRUE;
}
#endif // #ifndef _NoElevation

bool userAccountControlIsEnabled( bool bForced )
{
	//
	//	UAC is enabled if 
	//	HKLM\SOFTWARE\Microsoft\Windows\CurrentVersion\Policies\System Value EnableLUA == 1
	//
	static DWORD dwRes = -2;

	if ( bForced || ( dwRes == -2 ) )
	{
		DWORD dwSize = sizeof( dwRes );

		if ( ERROR_SUCCESS != 
				RegGetValueT( HKEY_LOCAL_MACHINE, L"SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Policies\\System", L"EnableLUA", 0, NULL, &dwRes, &dwSize ) )
			dwRes = -1;
	}

	return dwRes == 1;
}

//
//	ReadFromFile
//		Reads bytes from opened file.
//
bool ReadFromFile( HANDLE hFile, cADBlock < BYTE > &adb )
{
	const int BUFSIZE = 1024;
	DWORD size = BUFSIZE, rd = 0, attempt = 0, err = 0;

	for (;;)
	{
		BYTE text[BUFSIZE];
		BOOL ok = ReadFile( hFile, text, sizeof( text ), &rd, NULL);

		if (ok)
			adb.Add( text, rd );

		if ( !ok || ( 0 == rd ) )
		{
			if ( !adb.isEmpty() )
				break; 

			err = GetLastError();

			ODS_fls( _GetSystemError( err ) );

			if ( err == ERROR_NO_DATA ) // ERROR_NO_DATA 232  The pipe is being closed.
 				break;

			Sleep(100);
			if (attempt++ > 128)
				break;
		}
	} 

	return !adb.isEmpty();
}

//
//	GetProcessOutput
//		Calls RunProcessWithWaitingT, Note argumments are completely same excluding the first 'out'
//		Before calling creates an anonymous pipe and pass writting handle to RunProcessWithWaitingT as outFile
//		After process completes reads from the pipe by ReadToMbs.
//
HANDLE GetProcessOutput(
		cADBlock < BYTE > &out,
		const wchar_t *cmdLine,
		int waitType, bool hidden,
		const wchar_t *curDir,
		int *pRetCode, bool bRunAs,
		HWND hWnd, HANDLE hToken, HANDLE *pHPrimaryTread, 
		DWORD dwFlags,
		DWORD *pProcessId, DWORD *pThreadId
	)
{
	if ( !waitType || bRunAs )
	{
		ODS_fls( L"Assertion! Parameters are invalid!" );
		return HANDLE(-2);
	}

	cAutoHandle ahRead;
	cAutoHandle ahWrite;

	DWORD errCode = 0;

	SECURITY_ATTRIBUTES	sa;
	ZeroMemory(&sa, sizeof(sa));
	sa.bInheritHandle = TRUE;
	sa.nLength = sizeof(SECURITY_ATTRIBUTES);

	BOOL bOk = CreatePipe( &ahRead.m_handle, &ahWrite.m_handle, &sa, 0);

	if ( !bOk )
	{
		errCode = GetLastError();
		ODS_fls( _FormatT( L"Assertion! Error %s in creating pipe.", _GetSystemError( errCode ) ) );
		return HANDLE(-3);
	}

	bOk = SetHandleInformation( ahRead.m_handle, HANDLE_FLAG_INHERIT, 0 );

	if ( !bOk )
	{
		errCode = GetLastError();
		ODS_fls( _FormatT( L"Assertion! Error %s in SetHandleInformation.", _GetSystemError( errCode ) ) );
		return HANDLE(-4);
	}

	DWORD m = PIPE_READMODE_BYTE | PIPE_NOWAIT;
	bOk = SetNamedPipeHandleState(ahRead.m_handle, &m, NULL, NULL);
	if ( !bOk )
	{
		errCode = GetLastError();
		ODS_fls( _FormatT( L"Assertion! Error %s in SetNamedPipeHandleState.", _GetSystemError( errCode ) ) );
		return HANDLE(-5);
	}

	HANDLE hProcess = RunProcessWithWaitingT( cmdLine, waitType, hidden, curDir, pRetCode, bRunAs, hWnd, hToken, pHPrimaryTread, dwFlags, pProcessId, pThreadId, ahWrite.m_handle );

	if ( hProcess > 0 )
	{
		out.Free();
		ReadFromFile( ahRead.m_handle, out );
	}

	return hProcess;
}

static bool isMouseClick( UINT uMsg )
{
	return 
	 ( uMsg == WM_NCLBUTTONDOWN ) ||
	 ( uMsg == WM_NCLBUTTONDBLCLK ) ||
	 ( uMsg == WM_NCRBUTTONDOWN ) ||
	 ( uMsg == WM_NCRBUTTONDBLCLK ) ||
	 ( uMsg == WM_RBUTTONDOWN ) ||
	 ( uMsg == WM_RBUTTONDBLCLK ) ||
	 ( uMsg == WM_LBUTTONDOWN ) ||
	 ( uMsg == WM_LBUTTONDBLCLK );
}

struct sParam
{
	HWND hWnd;
	DWORD id;
};

static BOOL CALLBACK findByProcessID( HWND hWnd, LPARAM lParam )
{
	wchar_t bf[256];
	GetWindowTextW( hWnd, bf, maxLen( bf ) );

	sParam *ps = (sParam*)lParam;
	DWORD id;

	GetWindowThreadProcessId( hWnd, &id );
	if ( id != ps->id )
		return true;

	ps->hWnd = hWnd;
	return false;
}

HWND FindWindowByProcessId( DWORD id )
{
	ODSFuncS( getWCStr( id ) );

	if ( id == 0 )
	{
		ODS_fls( L"Assertion! Process identifier is NULL!" );
		return NULL;
	}

	sParam s = { 0 };
	s.id = id;
	EnumWindows( findByProcessID, LPARAM(&s) );

	return s.hWnd;
}

#ifdef _X86_MACHINE
/* CAUTION: ONLY x86 TESTED
 * get the thread id of the main thread of a target process
 *
 * params:
 *     DWORD dwPid  process id of the target process
 *
 * return:
 *     Success      thread id
 *     Error        NULL
 */
DWORD GetProcessMainThreadId(DWORD dwPid)
{
	SYSTEM_INFO siSysInfo = {0};
	GetSystemInfo(&siSysInfo);
	if ( ( siSysInfo.dwProcessorType != PROCESSOR_INTEL_PENTIUM ) &&
		 ( siSysInfo.dwProcessorType != PROCESSOR_INTEL_486 ) &&
		 ( siSysInfo.dwProcessorType != PROCESSOR_INTEL_386 ) 
		)
		return NULL;
 
    LPVOID lpTid;

    _asm
    {
        mov eax, fs:[18h]
        add eax, 36
        mov [lpTid], eax
    }

    HANDLE hProcess = OpenProcess(PROCESS_VM_READ, FALSE, dwPid);
    if(hProcess == NULL)
        return NULL;

    DWORD dwTid;
    if(ReadProcessMemory(hProcess, lpTid, &dwTid, sizeof(dwTid), NULL) == FALSE)
    {
        CloseHandle(hProcess);
        return NULL;
    }

    CloseHandle(hProcess);

    return dwTid;
}
#endif

#if(WINVER >= 0x0500) // FLASHWINFO is not defined for early versions

bool WaitProcessAsModalDialogue( HANDLE hProcess, DWORD processId )
{
	ODSFuncS( getWCStr( processId ) );
	MSG msg = {0};

	HWND hProcessMainWind = NULL;
	
	if ( processId == 0 )
	{
		processId = GetProcessId( hProcess );
		if ( processId == 0 )
			ODS_fls( _FormatT( L"GetProcessId(0x%X) error %s", _GetSystemError(-1) ) );
	}

	while( true )
	{
		DWORD dwRes = WaitForSingleObject( hProcess, 1000 );
		if ( dwRes == WAIT_OBJECT_0 )
			break;

		if ( dwRes != WAIT_TIMEOUT )
		{
			ODS_fls( _FormatT( L"hProcess %04X(0x04X), error in waiting", hProcess, processId ) );
			break;
		}

		bool bWasMessage = false;

		static __int64 pickQ_WM_PAINT = 0;

		while ( PeekMessage( &msg, NULL,  0, 0, PM_REMOVE ) ) 
		{
			//if ( msg.message == WM_PAINT )
			//	pickQ_WM_PAINT++;
			//else
			//	ODS_fls( _FormatT( L"PeekMessage 0x%X", msg.message ) );

			if ( msg.message == WM_QUIT )
				break;

			if ( ( msg.message == WM_SETFOCUS ) 
				|| isMouseClick( msg.message )
//				|| ( msg.message == 0xC03E ) 
				)
			{
				bWasMessage = true;
			}

			//
			//	25.09.2017
			//		It happens on win8 a lot of WM_PAINT messages get, so this loop did never stop
			//		But the process we wait was already closed. So we just check it.
			//
			DWORD dwExitCode = 0;
			if ( GetExitCodeProcess( hProcess, &dwExitCode ) &&
				 ( dwExitCode != STILL_ACTIVE ) )
			{
				ODS_fls( _FormatT( L"Process exit code 0x%X", dwExitCode ) );
				return true;
			}
		}
		
		if ( msg.message == WM_QUIT )
			break;

		// Post FlashWindow message
		if ( bWasMessage )
		{
			if ( hProcessMainWind == NULL )
				hProcessMainWind = FindWindowByProcessId( processId );

			if ( hProcessMainWind != NULL )
			{
				SetWindowPos( hProcessMainWind, HWND_TOP, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_SHOWWINDOW );
				FLASHWINFO fi = {0};
				fi.cbSize = sizeof( fi );
				fi.dwTimeout = 300;
				fi.hwnd = hProcessMainWind;
				fi.dwFlags = FLASHW_ALL;
				fi.uCount = 3;
				FlashWindowEx( &fi );
			}
		}
	}

	if ( msg.message != WM_QUIT )
		return true;

	if ( hProcessMainWind != NULL )
		PostMessage( hProcessMainWind, msg.message, msg.wParam, msg.lParam );

	ODS_fls( L"PostQuitMessage" );
	PostQuitMessage( (int)msg.wParam );

	return false;
}

#endif // WINVER >= 0x0500

//
//	Returns:
//		0 if process was not started.
//		-1 if error was encountered.
//		process handle if successfully starded
//	Note, if the procedure does not wait started process so the process handle has to be closed by called process
//	29.06.2011
//		The bRunElevated flag is renamed to bRunAs. It checks adminitrator privileges as well as limited user.
//		As result it asks user name and password for non-admintrative user.
//		Limited user is elevated ( as earlier )
//
HANDLE RunProcessWithWaitingT(
		const wchar_t *cmdLine,
		int waitType, bool hidden,
		const wchar_t *curDir,
		int *pRetCode, bool bRunAs,
		HWND hWnd, HANDLE hToken, HANDLE *pHPrimaryTread, 
		DWORD dwFlags,
		DWORD *pProcessId, DWORD *pThreadId,
		HANDLE hOutFile,
		HANDLE hInFile
	)
{
	ODSFuncS( _FormatT( L"%s, %s, %s, %s, %X, %s, %X, %X", cmdLine, getWCStr( waitType ), getWCStr( hidden ), curDir, pRetCode, getWCStr( bRunAs ), hWnd, hToken ) );

	bool bRunElevated = false;

	HANDLE hProcess = 0;
	BOOL ok;

	if ( pHPrimaryTread != NULL )
		*pHPrimaryTread = NULL;

	int res = -1;

	DWORD processId = 0;

#ifdef _NoElevation
	if ( bRunAs )
	{
		ODS_fls( L"Assertion: _NoElevation flag is assigned but bRunAs is true" );
		ODS_fls( L"Assertion!" );
		return 0;
	}
#else
	//
	//	31.05.2013
	//		If user account control is off so runElevated ( ShellExec - "runas" ) will start process with same credentials.
	//		We get a problem here. Like this:
	//		A non privileged user start process and this program desides to elevate you self.
	//		It calls this function with bRunAs = true and waits result code.
	//		New process starts with same credentials and  deside to elevate again.
	//		Loop :(
	//		So we will start process this way is user account control is enabled.
	//		Called process has to check if user account control is enabled before this automatic elevation. ( See userAccountControlIsEnabled() )
	//		Otherwise it just suggest to restart this process as admionistrator.
	//
	//if ( bRunAs && ( isWindowsVistaLimitedProcess() || !isAdministrator() ) )
	if ( bRunAs && !isAdministrator() )
	{
		if ( !userAccountControlIsEnabled() )
		{
			// See comments above.
			ODS_fls( L"Assertion! User Account Control is disabled" );

			if ( pRetCode == NULL )
				*pRetCode = -2;

			return 0;
		}

		wchar_t wcFilePath[ MAX_PATH ];

		wchar_t separator = L' ';
		if ( *cmdLine == L'\"' )
		{
			separator = *cmdLine;
			cmdLine++;
		}

		for( int i = 0;; i++ )
		{
			if ( i >= maxLen( wcFilePath ) )
			{
				ODS_fls( L"Assertion!" );
				return (HANDLE)-1;
			}

			if ( *cmdLine == separator )
			{
				// if ( separator == L'\"' )
				{
					wcFilePath[i] = L'\0';
					cmdLine++;
					break;
				}
			}

			if ( *cmdLine == L'\0' )
			{
				if ( separator == L'\"' )
				{
					ODS_fls( L"Assertion!" );
					return (HANDLE)-1;
				}

				wcFilePath[i] = L'\0';
				break;
			}

			wcFilePath[i] = *(cmdLine++);
		}

		while( *cmdLine == L' ' )
			cmdLine++;

		const wchar_t *wcParams = NULL;

		if ( *cmdLine != L'\0' )
			wcParams = cmdLine;

		ok = runElevated( hWnd, wcFilePath, wcParams, curDir, &hProcess, &res );

		bRunElevated = true;
	}
#endif // _NoElevation
	else
	{
		STARTUPINFOW stInfo;
		memset(&stInfo,0,sizeof(stInfo));
		stInfo.cb = sizeof(stInfo);

		if ( hidden )
		{
			stInfo.dwFlags = STARTF_USESHOWWINDOW;
			stInfo.wShowWindow = SW_HIDE;
		}
		else
		{
			stInfo.dwFlags = STARTF_USESHOWWINDOW;
			stInfo.wShowWindow = SW_SHOW;
		}

		bool bInheritHandlers = false;

		if ( hOutFile != NULL )
		{
			stInfo.hStdError = hOutFile;
			stInfo.hStdOutput = hOutFile;
			stInfo.dwFlags |= STARTF_USESTDHANDLES;

			if ( !SetHandleInformation(hOutFile, HANDLE_FLAG_INHERIT, HANDLE_FLAG_INHERIT ) )
				ODS_fls( L"Assertion!" );
			
			bInheritHandlers = true;
		}

		if ( hInFile != NULL )
		{
			stInfo.hStdInput = hInFile;
			stInfo.dwFlags |= STARTF_USESTDHANDLES;

			if ( !SetHandleInformation( hInFile, HANDLE_FLAG_INHERIT, HANDLE_FLAG_INHERIT ) )
				ODS_fls( L"Assertion!" );

			bInheritHandlers = true;
		}

		PROCESS_INFORMATION prInfo = {0};

		DWORD dwCreationFlag = _isWinNT ? CREATE_SHARED_WOW_VDM : CREATE_NEW_CONSOLE ; // DETACHED_PROCESS; //NORMAL_PRIORITY_CLASS;
		
		dwCreationFlag |= dwFlags;

		//
		//	Impersonated token.
		//	If process has been created with impersonation, it's thread token is not inherited.
		//	The new process thread token rather is inherited from process token. So we suspend 
		//	the thread ater process creating, assign the token and resume thread.
		//
		if ( hToken != NULL )
			dwCreationFlag |= CREATE_SUSPENDED;
		
		ok = CreateProcessT(
			NULL,
			(wchar_t *)cmdLine,
			NULL, // Security process attribute...
			NULL, // Security thread attribute...
			bInheritHandlers, // Inherit handlers
			dwCreationFlag,
			NULL, // Environment
			curDir, // Current directory
			&stInfo,
			&prInfo
			);

		res = GetLastError();

		ODSF( L"CreateProcessT = %d, last error %d", (int)ok, res );

		if ( ok )
			processId = prInfo.dwProcessId;
		
		//
		//	See notes above.
		//
		if ( ok && ( hToken != NULL ) )
		{
			ok = SetThreadToken( &prInfo.hThread, hToken );	// impersonation token

			ODS_f( _FormatT( L"SetThreadToken( 0x%X ) = %s", hToken, getWCStr( ok ) ) );

			if ( ok )
			{
				if ( ( dwFlags & CREATE_SUSPENDED ) == 0 ) // Otherwise caller is responsible for resuming...
					ResumeThread( prInfo.hThread );
			}
			else
			{
				CloseHandle( prInfo.hProcess );
				CloseHandle( prInfo.hThread );
			}
		}

		if ( ok )
		{
			hProcess = prInfo.hProcess;

			if ( pProcessId != NULL )
				*pProcessId = prInfo.dwProcessId;

			if ( pHPrimaryTread != NULL )
				*pHPrimaryTread = prInfo.hThread;

			if ( pThreadId != NULL )
				*pThreadId = prInfo.dwThreadId;
		}
	}

	if ( pRetCode == NULL )
		pRetCode = &res;

	if ( ok )
	{
		if ( waitType != ewptNoWait )
		{
			if ( waitType == ewptWaitAsModalDialogue )
#if ( WINVER >= 0x0500 )
				WaitProcessAsModalDialogue( hProcess, processId );
			else
#else
				ODS_fls( L"Assertion! WaitProcessAsModalDialogue does not work for this version of windows" );
#endif
			if ( WaitForSingleObject( hProcess, INFINITE ) != WAIT_OBJECT_0 )
				ODS_fls( _FormatT( L"hProcess %04X(0x04X), error in waiting", hProcess, processId ) );

			BOOL ok = GetExitCodeProcess( hProcess, (DWORD*)pRetCode );
			if ( !ok )
				ODS_fls( _FormatT( L"GetExitCodeProcess( %04X ) error %s", hProcess, _GetSystemError( -1 ) ) );

			ODS_fls( _FormatT( L"hProcess %04X, exit code %d", hProcess, *pRetCode ) );

			if ( bRunElevated && ( *pRetCode == STATUS_ACCESS_VIOLATION ) ) // 0xC0000005 ) )
			{
				//
				//	18.10.2018
				//		GetExitCodeProcess above works without error but fill pRetCode by STATUS_ACCESS_VIOLATION.
				//		The error was detected here with BCWipe.exe, but real result of elevated process was 0.
				//		Caller function may skip the error.
				//
				ODS_fls( _FormatT( L"Warning! Invalid elevated process exit code detected! The STATUS_ACCESS_VIOLATION error may be skipped!" ) );
			}

			CloseHandle( hProcess );
		}
		else
		{
			ODSFuncRes( (__int64)hProcess );
		}
	}
	else
	{
		ODSF( L"Error %d in creating process", res );
		hProcess = (HANDLE)-1;
		*pRetCode = res;
	}

	return hProcess;
}

LSTATUS
RegDeleteKeyValueT (
	HKEY     hKey,
	LPCWSTR lpSubKey,
	LPCWSTR lpValueName
)
{
	CDll advapiDll;

	if ( advapiDll.LoadProcOnce( false, "advapi32.dll", "RegDeleteKeyValueW", true ) )
	{
		typedef WINADVAPI LSTATUS APIENTRY typeRegDeleteKeyValueW(
			HKEY     hKey,
			LPCWSTR lpSubKey,
			LPCWSTR lpValueName
		);

		LSTATUS res = ((typeRegDeleteKeyValueW*)advapiDll.pProc )( hKey, lpSubKey, lpValueName );
		return res;
	}

	HKEY hRes;
	LSTATUS res = RegCreateKeyT( hKey, lpSubKey, &hRes );
	if ( res != ERROR_SUCCESS )
		return res;

	res = RegDeleteValueT( hRes, lpValueName );

	RegCloseKey( hRes );

	return res;
}

LONG
RegGetValueT(
	HKEY    hRoot,
	LPCWSTR  lpSubKey,
	LPCWSTR  lpValue,
	DWORD    dwFlags,
	LPDWORD pdwType,
	PVOID   pvData,
	LPDWORD pcbData 
    )
{
	LONG lRes;

	//bool bIsNT = isWinNT;
	//
	//if ( bIsNT )
	//	lRes = RegGetValueW( hRoot, lpSubKey, lpValue, dwFlags, pdwType, pvData, pcbData );
	//else
	//	lRes = RegGetValueA( hRoot, cMBStr( lpSubKey ).m_szBody, cMBStr( lpValue ).m_szBody, dwFlags, pdwType, pvData, pcbData );

	//if ( lRes == ERROR_SUCCESS )
	//	return lRes;

	HKEY hKey;
	lRes = RegOpenKeyExT( hRoot, lpSubKey, 0, KEY_QUERY_VALUE, &hKey );

	if ( lRes != ERROR_SUCCESS )
		return lRes;

	lRes = RegQueryValueExT( hKey, lpValue, NULL, pdwType, (BYTE*)pvData, pcbData );

	RegCloseKey( hKey );

	return lRes;
}

LONG
RegSetValueT( HKEY root, const wchar_t *szSection, const wchar_t *szValue, DWORD dwType, void *bf, int bfSize )
{
	HKEY hKey;

	LONG lRes = RegOpenKeyExT( root, szSection, 0, KEY_WRITE, &hKey );

	if ( lRes != ERROR_SUCCESS )
		return lRes;

	if ( bf == NULL )
		lRes = RegDeleteValueT( hKey, szValue );
	else
	{
		DWORD cbData = bfSize;

		lRes = RegSetValueExT( hKey, szValue, NULL, dwType, (const BYTE *)bf, cbData );
	}
	
	RegCloseKey(hKey);

	return lRes;
}

LONG
RegQueryValueExT(
    IN HKEY hKey,
    IN LPCWSTR lpValueName,
    IN LPDWORD lpReserved,
    OUT LPDWORD lpType,
    IN OUT LPBYTE lpData,
    IN OUT LPDWORD lpcbData
    )
{
	DWORD dwType;
	if ( lpType == NULL )
		lpType = &dwType;

	return RegQueryValueExW( hKey, lpValueName, lpReserved, lpType, lpData, lpcbData );
}

LONG
RegDeleteEntireKeyT(
    IN HKEY hKey,
    IN LPCWSTR lpSubKey
    )
{
	typedef DWORD STDAPICALLTYPE typeSHDeleteKeyW(HKEY hkey, LPCWSTR pszSubKey);

	static CDll dll( false, L"shlwapi.dll", "SHDeleteKeyW", false );

	if ( dll.error )
	{
		RegDeleteSubKeysAndValuesT( hKey, lpSubKey );
		return RegDeleteKeyT( hKey, lpSubKey );
	}
	else
		return ((typeSHDeleteKeyW*)dll.pProc)( hKey, lpSubKey );
}


LONG RegDeleteSubKeysAndValuesT(
	HKEY hRoot,
	LPCWSTR szKeyName
	)
{
	ODSFunc;

	HKEY hKey;
	LONG res = RegOpenKeyExT( hRoot, szKeyName, 0, KEY_ALL_ACCESS, &hKey );

	if ( res != ERROR_SUCCESS )
		return -1;

	DWORD subKeysQ = 0;
	DWORD valuesQ = 0;

	res = RegQueryInfoKey( hKey, 0, 0, 0, &subKeysQ, 0, 0, &valuesQ, 0, 0, 0, 0 );

	if ( res != ERROR_SUCCESS )
	{
		RegCloseKey( hKey );
		return -1;
	}

	cWCStr name( 512 );

	// Enum subkeys

	int i;

	for ( i = subKeysQ; i--; )
	{
		ULONG nameSize = name.m_len;

		res = RegEnumKeyExT( hKey, i, name.m_szBody, &nameSize, 0, NULL, NULL, NULL );

		if ( res == ERROR_SUCCESS )
		{
			res = RegDeleteEntireKeyT( hKey, name.m_szBody );
			if ( res != ERROR_SUCCESS )
				ODS_fls( _FormatT( L"Assertion! Error %d in deleting key", res ) );
		}
		else
			ODS_fls( _FormatT( L"Assertion! Error %d in key enumeration", res ) );
	}

	// Enum values

	for ( i = valuesQ; i--; )
	{
		ULONG nameSize = name.m_len;

		DWORD type;

		res = RegEnumValueT( hKey, i, name.m_szBody, &nameSize, NULL, &type, NULL, NULL );

		if ( res == ERROR_SUCCESS )
		{
			res = RegDeleteValueT( hKey, name.m_szBody );
			if ( res != ERROR_SUCCESS )
				ODS_fls( _FormatT( L"Assertion! Error %d in deleting value", res ) );
		}
		else
			ODS_fls( _FormatT( L"Assertion! Error %d in key enumeration", res ) );

	}

	res = RegQueryInfoKey( hKey, 0, 0, 0, &subKeysQ, 0, 0, &valuesQ, 0, 0, 0, 0 );

	RegCloseKey( hKey );

	if ( res )
	{
		if ( ( subKeysQ != 0 ) || ( valuesQ != 0 ) )
			res = -1;
	}

	return res;
}

////
//
//	Instance and DlgFuncs handlers
//
bool DlgCheckButton( HWND hDlg, int idBtn, bool bCheck )
{
	HWND hBtn = GetDlgItem( hDlg, idBtn );

	if ( hBtn == NULL )
		return false;

	Button_SetCheck( hBtn, bCheck ? BST_CHECKED : BST_UNCHECKED );

	return true;
}

int DlgMessageBox( HINSTANCE hInst, HWND hDlg, int IDS, int MB_ID, const char *pCaption )
{
	wchar_t text[ MAX_PATH * 2 ];

	if ( LoadStringT( hInst, IDS, text, maxLen( text ) ) == 0 )
	{
		getErrorString( eRCStringNotFound, text, maxLen( text ), IDS );
		return MessageBoxT( hDlg, text, L"DlgMessageBox", MB_ICONSTOP );
	}

	const wchar_t *pCaptionW = NULL;

	if ( pCaption != NULL )
		pCaptionW = getWCStr( pCaption );

	return DlgMessageBoxT( hDlg, text, MB_ID, pCaptionW );
}

int DlgMessageBox( HWND hDlg, const char *text, int MB_ID, const char *pCaption )
{
	char caption[ MAX_PATH ];

	if ( pCaption == NULL )
	{
		GetWindowTextA( hDlg, caption, maxLen( caption ) );
		pCaption = caption;
	}

	return MessageBoxA( hDlg, text, pCaption, MB_ID );
}

const wchar_t *getParentCaption( HWND hWnd )
{
	HWND hParent = GetParent( hWnd );
	if ( hParent == NULL )
		return NULL;

	cWCStr &wcs = IH.getStaticCaption();
	//cWCStr &wcs = m_wcsDefCaption;

	wcs.GetWindowTextT( hParent );
	if ( wcs.isEmpty() )
		return NULL;
	
	return wcs.m_szBody;
}

int DlgMessageBoxT( HINSTANCE hInst, HWND hDlg, int IDS, int MB_ID, const wchar_t *pCaption )
{
	wchar_t text[ MAX_PATH * 2 ];

	if ( LoadStringT( hInst, IDS, text, maxLen( text ) ) == 0 )
	{
		getErrorString( eRCStringNotFound, text, maxLen( text ), IDS );
		return MessageBoxT( hDlg, text, L"DlgMessageBox", MB_ICONSTOP );
	}

	return DlgMessageBoxT( hDlg, text, MB_ID, pCaption );
}

int DlgMessageBoxT( HWND hDlg, const wchar_t *text, int MB_ID, const wchar_t *pCaption )
{
	wchar_t caption[ MAX_PATH ];

	if ( pCaption == NULL )
	{
		if ( hDlg != NULL )
		{
			GetWindowTextT( hDlg, caption, maxLen( caption ) );
			pCaption = caption;
		}
		else
			pCaption = L"Attention!";
	}

	return MessageBoxT( hDlg, text, pCaption, MB_ID );
}

int GetDlgItemTextT( HWND hDlg, int nIDDlgItem, wchar_t *lpString, int nMaxCount )
{
	if ( nMaxCount == 0 )
		return 0;

	int res = GetDlgItemTextW( hDlg, nIDDlgItem, lpString, nMaxCount );
	lpString[ --nMaxCount ] = L'\0';
	return res;
}

void cDlgFuncsHolder::SetHolder( HINSTANCE hInst, HWND hDlg, int idsDefCaption )
{
	cInstanceHolder::SetHolder( hInst );
	m_hDlg = hDlg;

	cWCStr &wcs = getStaticCaption();
	//cWCStr &wcs = m_wcsDefCaption;

	if ( idsDefCaption != 0 )
		wcs.LoadStringT( m_hInst, idsDefCaption );
	else
	if ( m_hDlg != NULL )
		wcs.GetWindowTextT( hDlg );

	m_defCaption = wcs.m_szBody;
}

bool cDlgFuncsHolder::IsChecked( int nIDDlgItem )
{
	return Button_GetCheck( GetItem( nIDDlgItem ) ) == BST_CHECKED;
}

void cDlgFuncsHolder::UpdateDlgItemTextT( HWND hDlg, int idText, const wchar_t *txt )
{
	wchar_t pattern[ 2048 ];
	int len = ::GetDlgItemTextT( hDlg, idText, pattern, maxLen( pattern ) );

	cWCStr wcs;
	if ( wcschr( pattern, L'%' ) != NULL )
	{
		wcs.FormatT( pattern, txt );
		txt = wcs.m_szBody;
	}

	::SetDlgItemTextT( hDlg, idText, txt );
}

void cDlgFuncsHolder::UpdateDlgItemText( HWND hDlg, int idText, const char *txt )
{
	UpdateDlgItemTextT( hDlg, idText, cWCStr( txt ).m_szBody );
}


//static 
int cInstanceHolder::cxIcon = 16;
int cInstanceHolder::cyIcon = 16;

//static 
HICON cInstanceHolder::LoadIconT( HINSTANCE hInst, int idi, int cx, int cy )
{
	if ( cx == 0 ) cx = cInstanceHolder::cxIcon;
	if ( cy == 0 ) cy = cInstanceHolder::cyIcon;

	HICON hIcon = (HICON)LoadImage( hInst, MAKEINTRESOURCE( idi ), IMAGE_ICON, cx, cy, LR_DEFAULTCOLOR ); 
	return hIcon;
}

HICON cInstanceHolder::LoadIconT( int idi, int cx, int cy )
{
	return cInstanceHolder::LoadIconT( m_hInst, idi, cx, cy );
}



#ifndef _NoComdlg32Lib_

static BOOL GetFileNameT( LPOPENFILENAMEW ofnW, bool open )
{
	if ( open )
		return GetOpenFileNameW( ofnW );
	else // save
		return GetSaveFileNameW( ofnW );
}

BOOL GetOpenFileNameT( LPOPENFILENAMEW ofnW ) { return GetFileNameT( ofnW, true  ); }
BOOL GetSaveFileNameT( LPOPENFILENAMEW ofnW ) { return GetFileNameT( ofnW, false ); }

#endif // _NoComdlg32Lib_

//
//	GetCommandLineT
//		If excludeProgramName is true the command line looks like WinMain command line parameter
//		otherwise it contains module file path before parameters 
//
const wchar_t *GetCommandLineT( bool excludeProgramName )
{
	static cAutoDeallocate < wchar_t > adCmdLine;

	wchar_t *cmdLine;

	cmdLine = adCmdLine.AllocateAndCopy( GetCommandLineW() );

	if ( cmdLine == NULL )
		return L"\0";

	if ( !excludeProgramName )
		return cmdLine;

	wchar_t separator = L' ';
	if ( cmdLine[0] == L'\"' )
		separator = *(cmdLine++);

	while( *cmdLine != 0    ) if ( *(cmdLine++) == separator ) break;
	while( *cmdLine == L' ' ) if ( *(cmdLine++) == 0 ) return L"\0";

	return cmdLine;
}

bool GetUserNameExT( int nameFormat, LPWSTR lpNameBuffer, PULONG nSize )
{
	static CDll dll( false, "secur32.dll", "GetUserNameExW", false );

	if ( dll.error )
		return GetUserNameT( lpNameBuffer, nSize ) != 0;

	int NameDnsDomain = 12;
	ULONG lSize = *nSize;
	typedef BOOLEAN WINAPI typeGetUserNameExW( int nameFormat, LPWSTR lpNameBuffer, PULONG nSize  );
	
	if ( nameFormat <= 0 )
		nameFormat = NameDnsDomain;

	BOOLEAN bRes = ((typeGetUserNameExW *)dll.pProc)( nameFormat, lpNameBuffer, &lSize );

	if ( bRes != 0 )
	{
		*nSize = lSize;
		return true;
	}

	if ( *nSize == 0 )
	{
		*nSize = lSize;
		return false;
	}
		
	return GetUserNameT( lpNameBuffer, nSize ) != 0;
}


wchar_t *GetFileLocalTime( wchar_t *bf, int bfLen, const wchar_t *filePath )
{
	if ( bfLen != 0 )
		bf[0] = L'\\';

	HANDLE hFind;
	WIN32_FIND_DATAW findFileData;
	hFind = FindFirstFileT( filePath, &findFileData );
	if ( hFind == INVALID_HANDLE_VALUE )
		return bf;

	FindClose( hFind );

	return GetFileLocalTime( bf, bfLen, &findFileData.ftLastWriteTime );
}

wchar_t *getSystemTimeStr( wchar_t *str, int strMaxLen, SYSTEMTIME *systime )
{
	static cWCS wcs( 256 );
	if ( str == NULL )
	{
		str = wcs.m_szBody;
		strMaxLen = wcs.m_allocatedLen;
	}

	SYSTEMTIME _systime;

	if ( systime == NULL )
	{
		systime = &_systime;
		GetLocalTime( systime );
	}

	FormatT( str, strMaxLen,
		L"%u/%u/%u %02u:%02u:%02u:%03u",
		systime->wYear, systime->wMonth, systime->wDay,
		systime->wHour, systime->wMinute, systime->wSecond, systime->wMilliseconds);

	return str;
}

wchar_t *GetSystemTimeStr( wchar_t *bf, int bfLen, const SYSTEMTIME *pSysTime )
{
	SYSTEMTIME sysTime;

	if ( pSysTime == NULL )
		GetLocalTime( &sysTime ); // GetSystemTime( &sysTime );
	else
		sysTime = *pSysTime;

	wchar_t shortDate[ 100 ] = L"";
	wchar_t shortTime[ 100 ] = L"";

	GetDateFormatT( LOCALE_USER_DEFAULT, DATE_SHORTDATE, &sysTime, NULL, shortDate, maxLen( shortDate ) ); 
	GetTimeFormatT( LOCALE_USER_DEFAULT, TIME_NOSECONDS, &sysTime, NULL, shortTime, maxLen( shortTime ) ); 

	return FormatT( bf, bfLen, L"%s %s", shortDate, shortTime );
}

const wchar_t *GetSystemTimeStr( const SYSTEMTIME *pSysTime )
{
	static cWCStr wcs( 256 );
	if ( wcs.m_szBody == NULL )
		return L"Null";

	return GetSystemTimeStr( wcs.m_szBody, wcs.m_len, pSysTime );
}

wchar_t *GetFileLocalTime( wchar_t *bf, int bfLen, const FILETIME *fileTime )
{
	SYSTEMTIME sysTime;
	FILETIME localTime;

	FileTimeToLocalFileTime( fileTime, &localTime );
	FileTimeToSystemTime( &localTime, &sysTime );

	return GetSystemTimeStr( bf, bfLen, &sysTime );

	//return FormatT( bf, bfLen, L"%02d.%02d.%04d %02d:%02d",
	//	sysTime.wDay,
	//	sysTime.wMonth,
	//	sysTime.wYear,
	//	sysTime.wHour,
	//	sysTime.wMinute);
}

HPROPSHEETPAGE CreatePropertySheetPageT( LPCPROPSHEETPAGEW psp )
{
	// HPROPSHEETPAGE hpage = CreatePropertySheetPageW( psp );

#ifdef ISOLATION_AWARE_ENABLED
	return CreatePropertySheetPageW( psp );
#else
	// See notes at cComctl32Dll
	if ( !cComCtl32Dll::getStatic().LoadOnce() )
		return NULL;

	typedef HPROPSHEETPAGE WINAPI typeCreatePropertySheetPageW( LPCPROPSHEETPAGEW constPropSheetPagePointer );
	static typeCreatePropertySheetPageW *proc = (typeCreatePropertySheetPageW *)GetProcAddress( cComCtl32Dll::getStatic().hInst, "CreatePropertySheetPageW" );
	if ( proc == NULL )
		return NULL;

	return proc( psp);
#endif
}

BOOL DestroyPropertySheetPageT( HPROPSHEETPAGE hpsp )
{
#ifdef ISOLATION_AWARE_ENABLED
	return DestroyPropertySheetPage( hpsp );
#else
	if ( !cComCtl32Dll::getStatic().LoadOnce() )
		return NULL;

	typedef BOOL WINAPI typeDestroyPropertySheetPage(HPROPSHEETPAGE unnamed1);
	static typeDestroyPropertySheetPage *proc = (typeDestroyPropertySheetPage *)GetProcAddress( cComCtl32Dll::getStatic().hInst, "DestroyPropertySheetPage" );
	if ( proc == NULL )
		return NULL;

	return proc( hpsp );
#endif
}

BOOL
GetUserNameTA (
    OUT LPSTR lpBuffer,
    IN OUT LPDWORD nSize
    )
{
	CDll advapiDll;

	if ( !advapiDll.LoadProcOnce( false, "advapi32.dll", "GetUserNameA", true ) )
		return false;

	typedef BOOL WINAPI typeGetUserNameA (	LPSTR lpBuffer, LPDWORD nSize );

	BOOL res = ((typeGetUserNameA*)advapiDll.pProc )( lpBuffer, nSize );

	return res;
}

BOOL
GetUserNameT (
    OUT LPWSTR lpBuffer,
    IN OUT LPDWORD nSize
    )
{
	CDll advapiDll;

	if ( !advapiDll.LoadProcOnce( false, "advapi32.dll", "GetUserNameW", true ) )
		return false;

	typedef BOOL WINAPI typeGetUserNameW (	LPWSTR lpBuffer, LPDWORD nSize );

	return ((typeGetUserNameW*)advapiDll.pProc)( lpBuffer, nSize );
}

wchar_t *allocateWCS( const wchar_t *bf, int len )
{
	return newString( bf, len );
}

wchar_t *allocateWCS( const char *bf, int len )
{
	cWCStr wcs( bf, len );
	return allocateWCS( wcs.m_szBody );
}

char *allocateMBS( const char *bf, int len )
{
	if ( len == -1 )
		len = (int)strlen( bf );

	char *mbs = new char[ len + 1 ];

	if ( mbs == NULL )
		return NULL;

	mbs[len] = 0;
	return strncpy( mbs, bf, len );
}

char *allocateMBS( const wchar_t *bf, int len )
{
	cMBStr mbs( bf, len );
	return allocateMBS( mbs.m_szBody );
}

HWND CreateStatusWindowT(LONG style, LPCWSTR lpszText, HWND hwndParent, UINT wID)
{
#ifdef ISOLATION_AWARE_ENABLED
	return CreateStatusWindowW( style, lpszText, hwndParent, wID );
#else
	// See notes at cComctl32Dll
	//if ( !cComCtl32Dll::getStatic().LoadOnce() )
	//	return NULL;
	
	typedef WINCOMMCTRLAPI HWND WINAPI typeCreateStatusWindowW(LONG style, LPCWSTR lpszText, HWND hwndParent, UINT wID);

	static typeCreateStatusWindowW *proc = (typeCreateStatusWindowW *)GetProcAddress( cComCtl32Dll::getStatic().hInst, "CreateStatusWindowW" );
	if ( proc == NULL )
		return NULL;

	return proc( style, lpszText, hwndParent, wID );
#endif
}

DWORD
QueryDosDeviceT(
		LPCWSTR lpDeviceName,
		LPWSTR lpTargetPath,
		DWORD ucchMax
	)
{
	static CDll kernel32dll;

	if ( !kernel32dll.LoadProcOnce( false, "KERNEL32.dll", "QueryDosDeviceW", false ) )
		return 0;

	typedef WINBASEAPI DWORD WINAPI typeQueryDosDeviceW(
			LPCWSTR lpDeviceName,
			LPWSTR lpTargetPath,
			DWORD ucchMax
		);

	return ((typeQueryDosDeviceW*)kernel32dll.pProc)( lpDeviceName, lpTargetPath, ucchMax );
}

HANDLE
FindFirstVolumeT(
    LPWSTR lpszVolumeName,
    DWORD cchBufferLength
    )
{
	static CDll kernel32dll( false, "KERNEL32.dll", "FindFirstVolumeW", false );

	if ( kernel32dll.error != CDll::errNo )
		return false;

	typedef WINBASEAPI HANDLE WINAPI typeFindFirstVolumeW( LPWSTR lpszVolumeName, DWORD cchBufferLength );

	HANDLE hFind = ((typeFindFirstVolumeW*)kernel32dll.pProc)( lpszVolumeName, cchBufferLength );

	return hFind;
}

BOOL
FindNextVolumeT(
    HANDLE hFindVolume,
    LPWSTR lpszVolumeName,
    DWORD cchBufferLength
    )
{
	static CDll kernel32dll( false, "KERNEL32.dll", "FindNextVolumeW", false );

	if ( kernel32dll.error != CDll::errNo )
		return false;

	typedef WINBASEAPI BOOL WINAPI typeFindNextVolumeW( HANDLE hFindVolume, LPWSTR lpszVolumeName, DWORD cchBufferLength );

	return ((typeFindNextVolumeW*)kernel32dll.pProc)( hFindVolume, lpszVolumeName, cchBufferLength );
}

BOOL
FindVolumeCloseT(
    HANDLE hFindVolume
    )
{
	static CDll kernel32dll( false, "KERNEL32.dll", "FindVolumeClose", false ) ;

	if ( kernel32dll.error != CDll::errNo )
		return false;

	typedef WINBASEAPI BOOL WINAPI typeFindVolumeClose( HANDLE hFindVolume );

	return ((typeFindVolumeClose*)kernel32dll.pProc)( hFindVolume );
}

HANDLE
FindFirstVolumeMountPointT(
    LPCWSTR lpszRootPathName,
    LPWSTR lpszVolumeMountPoint,
    DWORD cchBufferLength
    )
{
	static CDll kernel32dll( false, "KERNEL32.dll", "FindFirstVolumeMountPointW", false );

	if ( kernel32dll.error != CDll::errNo )
		return false;

	typedef WINBASEAPI HANDLE WINAPI typeFindFirstVolumeMountPointW( LPCWSTR lpszRootPathName, LPWSTR lpszVolumeMountPoint, DWORD cchBufferLength );

	return ((typeFindFirstVolumeMountPointW*)kernel32dll.pProc)( lpszRootPathName, lpszVolumeMountPoint, cchBufferLength );
}

BOOL
GetVolumePathNameT(
    LPCWSTR lpszFileName,
    LPWSTR lpszVolumePathName,
    DWORD cchBufferLength
    )
{
	static CDll kernel32dll( false, "KERNEL32.dll", "GetVolumePathNameW", false );

	if ( kernel32dll.error != CDll::errNo )
		return false;

	typedef WINBASEAPI BOOL WINAPI typeGetVolumePathNameW( LPCWSTR lpszFileName, LPWSTR lpszVolumePathName, DWORD cchBufferLength );


	return ((typeGetVolumePathNameW*)kernel32dll.pProc)( lpszFileName, lpszVolumePathName, cchBufferLength );
}


BOOL
FindNextVolumeMountPointT(
    HANDLE hFindVolumeMountPoint,
    LPWSTR lpszVolumeMountPoint,
    DWORD cchBufferLength
    )
{
	static CDll kernel32dll( false, "KERNEL32.dll", "FindNextVolumeMountPointW", false );

	if ( kernel32dll.error != CDll::errNo )
		return false;

	typedef WINBASEAPI BOOL WINAPI typeFindNextVolumeMountPointW( HANDLE hFindVolumeMountPoint, LPWSTR lpszVolumeMountPoint, DWORD cchBufferLength );

	return ((typeFindNextVolumeMountPointW*)kernel32dll.pProc)( hFindVolumeMountPoint, lpszVolumeMountPoint, cchBufferLength );
}

BOOL
FindVolumeMountPointCloseT(
    HANDLE hFindVolumeMountPoint
    )
{
	static CDll kernel32dll( false, "KERNEL32.dll", "FindVolumeMountPointClose", false );

	if ( kernel32dll.error != CDll::errNo )
		return false;

	typedef WINBASEAPI BOOL WINAPI typeFindVolumeMountPointClose( HANDLE hFindVolumeMountPoint );

	return ((typeFindVolumeMountPointClose*)kernel32dll.pProc)( hFindVolumeMountPoint );
}

BOOL GetVolumeNameForVolumeMountPointT(
	LPCWSTR lpszVolumeMountPoint,
	LPWSTR lpszVolumeName,
	DWORD cchBufferLength
)
{
	//ODSFuncS( lpszVolumeMountPoint );

	typedef BOOL WINAPI typeGetVolumeNameForVolumeMountPoint(
		const wchar_t *lpszVolumeMountPoint,
		wchar_t *lpszVolumeName,
		DWORD cchBufferLength
	);

	static 
		typeGetVolumeNameForVolumeMountPoint
			*procGetVolumeNameForVolumeMountPoint = 
				(typeGetVolumeNameForVolumeMountPoint *)
					GetProcAddress(
						GetModuleHandleA("kernel32.dll"),
						"GetVolumeNameForVolumeMountPointW" );
	
	if ( procGetVolumeNameForVolumeMountPoint == NULL )
	{
		//ODSFuncRes( L"NULL" );
		return false;
	}
	
	BOOL ok = procGetVolumeNameForVolumeMountPoint( lpszVolumeMountPoint, lpszVolumeName, cchBufferLength );

	return ok;
}

//
// class cLogicalDriveVolumeName
//	is used for Windows 2K because it does not have the GetVolumePathNamesForVolumeNameT
//	but it has GetVolumeNameForVolumeMountPointT.
//	It create table of logical drives with theirs volume names.
//
class cLogicalDriveVolumeName
{
	wchar_t *T[ 'Z' - 'A' ];
	
	bool m_bInited;

public:

	cLogicalDriveVolumeName()
	{
		memset( T, 0, sizeof( T ) );
		m_bInited = false;
	}
	
	~cLogicalDriveVolumeName()
	{
		int i = maxLen( T );

		while ( i-- )
		{
			if ( T[i] == NULL )
				continue;

			deleteString( &T[i] );
		}
	}

	bool Init()
	{
		m_bInited = true;

		cWCStr wcs( 1024 );

		wchar_t drv[4] = L"A:\\";

		DWORD dwMask = GetLogicalDrives();

		for ( int i = 0; i < maxLen( T ); i++ )
		{
			if ( ( dwMask & ( 1 << i ) ) == 0 )
				continue;

			drv[0] = L'A' + i;

			// Look up each device name
			if ( GetVolumeNameForVolumeMountPointT( drv, wcs.m_szBody, wcs.m_len ) )
			{
				size_t uNameLen = wcslen( wcs.m_szBody );

				if ( uNameLen < MAX_PATH ) 
				{
					T[i] = newString( wcs.m_szBody );
				}
			}
		}
 
		return true;
	}

	int FindDrive( const wchar_t *volumeName )
	{
		//ODSFuncS( volumeName );

		int i = maxLen( T );

		while ( i-- )
		{
			if ( T[i] == NULL )
				continue;

			if ( _wcsicmp( T[i], volumeName ) == 0 )
				break;
		}

		//ODSFuncRes( i );

		return i;
	}

	bool GetDriveName( LPCWSTR lpszVolumeName,  LPWSTR lpszVolumePathNames,  DWORD cchBufferLength )
	{
		//ODSFuncS( lpszVolumeName );

		if ( !m_bInited )
			Init();

		int drvNum = FindDrive( lpszVolumeName );

		if ( drvNum == -1 )
			return false;
	
		FormatT( lpszVolumePathNames, cchBufferLength, L"%c:\\", L'A' + drvNum );

		//ODSFuncRes( lpszVolumePathNames );

		return true;
	}

};

BOOL
W2K_GetVolumePathNamesForVolumeNameT(
    LPCWSTR lpszVolumeName,
    LPWSTR lpszVolumePathNames,
    DWORD cchBufferLength,
	PDWORD lpcchReturnLength
	)
{
	static cLogicalDriveVolumeName ldNames;
	BOOL ok = ldNames.GetDriveName( lpszVolumeName, lpszVolumePathNames, cchBufferLength );

	if ( ok && ( lpcchReturnLength != NULL ) )
		*lpcchReturnLength = (DWORD)wcslen( lpszVolumePathNames );

	return ok;
}

HANDLE
FindFirstFileNameT(
	LPCWSTR lpFileName,
	DWORD dwFlags,
	LPDWORD StringLength,
	PWCHAR LinkName
)
{
	typedef HANDLE WINAPI typeFindFirstFileNameW(
		__in    LPCWSTR lpFileName,
		__in    DWORD dwFlags,
		__inout LPDWORD StringLength,
		__inout_ecount(*StringLength) PWCHAR LinkName
    );

	static CDll kernel32dll( false, "KERNEL32.dll", "FindFirstFileNameW", false );

	if ( kernel32dll.error )
		return false;


	return ((typeFindFirstFileNameW*)kernel32dll.pProc)( lpFileName, dwFlags, StringLength, LinkName );
}

BOOL
FindNextFileNameT(
	HANDLE hFindStream,
	LPDWORD StringLength,
	PWCHAR LinkName
)
{
	typedef BOOL APIENTRY typeFindNextFileNameW(
		HANDLE hFindStream,
		LPDWORD StringLength,
		PWCHAR LinkName
	);

	static CDll kernel32dll( false, "KERNEL32.dll", "FindNextFileNameW", false );

	if ( kernel32dll.error )
		return false;


	return ((typeFindNextFileNameW*)kernel32dll.pProc)( hFindStream, StringLength, LinkName );
}


BOOL
GetVolumePathNamesForVolumeNameT(
    LPCWSTR lpszVolumeName,
    LPWSTR lpszVolumePathNames,
    DWORD cchBufferLength,
    PDWORD lpcchReturnLength
    )
{
	//ODSFuncS( lpszVolumeName );

	//
	// Strange but GetVolumePathNamesForVolumeNameW does not work without '\\' at the end.
	//	The path works fine -> "\\?\Volume{87ae22ee-99b0-11de-99ed-806e6f6e6963}\"
	//
	size_t len = wcslen( lpszVolumeName );
	cWCStr wcs;
	if ( ( len > 0 ) && ( lpszVolumeName[len-1] != L'\\' ) )
	{
		lpszVolumeName = wcs.FormatT( L"%s\\", lpszVolumeName );
	}

	typedef WINBASEAPI BOOL WINAPI typeGetVolumePathNamesForVolumeNameW( LPCWSTR lpszVolumeName, LPWSTR lpszVolumePathNames, DWORD cchBufferLength, PDWORD lpcchReturnLength );

	BOOL ok;
	
	static void *proc = GetProcAddress( cKernelDll::getStatic().getInstance(), "GetVolumePathNamesForVolumeNameW" );

	//
	//	Win 2K does not have the function.
	//
	if ( proc == NULL )
	{
		 ok = W2K_GetVolumePathNamesForVolumeNameT( lpszVolumeName, lpszVolumePathNames, cchBufferLength );

		 if ( ok && ( lpcchReturnLength != NULL ) )
			 *lpcchReturnLength = (DWORD)wcslen( lpszVolumePathNames );

		 return ok;
	}

	ok = ((typeGetVolumePathNamesForVolumeNameW*)proc)( lpszVolumeName, lpszVolumePathNames, cchBufferLength, lpcchReturnLength );

	return ok;
}

UINT ExtractIconExT(
		LPCWSTR lpszFile,
		int nIconIndex,
		HICON *phiconLarge,
		HICON *phiconSmall,
		UINT nIcons
	)
{
	//static cShell32Dll shell32dll;
	//static cShell32Dll &shell32dll = cShell32Dll::getStatic();

	if ( cShell32Dll::getStatic().error )
		return 0;

	if ( cShell32Dll::getStatic().m_ExtractIconExW == NULL )
		return 0;

	return cShell32Dll::getStatic().m_ExtractIconExW( lpszFile, nIconIndex, phiconLarge, phiconSmall, nIcons );
}

HICON ExtractIconT(HINSTANCE hInst, LPCWSTR lpszExeFileName, UINT nIconIndex)
{
	if ( cShell32Dll::getStatic().error )
		return NULL;

	return cShell32Dll::getStatic().m_ExtractIconW( hInst, lpszExeFileName, nIconIndex );
}


int
LoadStringTA(
    HINSTANCE hInstance,
    UINT uID,
    LPSTR lpBuffer,
    int nBufferMax,
	bool translate )
{
	if ( hInstance == NULL )
	{
		ODS_fls( L"Assertion!" );
		cAutoDeallocate < wchar_t > ad;
		if ( ad.Allocate( nBufferMax ) == NULL )
			_wcstombs( lpBuffer, nBufferMax, getErrorString( eMemoryAllocation ) );
		else
			_wcstombs( lpBuffer, nBufferMax, getErrorString( eRCInstanceInvalid, ad.m_ptr, ad.m_len, (DWORD)0 ) );

		return (int)strlen( lpBuffer );
	}

#ifndef _NoTranslation_
	if ( pTranslateDll != NULL )
	{
		int res = pTranslateDll->Translate( lpBuffer, nBufferMax, hInstance, uID );

		bool isTranslated = res != 0;

		if ( isTranslated )
			return res;
	}
#endif

	int res = LoadStringA( hInstance, uID, lpBuffer, nBufferMax );

	if ( res <= 0 )
	{
		cAutoDeallocate < wchar_t > ad;
		if ( ad.Allocate( nBufferMax ) == NULL )
			_wcstombs( lpBuffer, nBufferMax, getErrorString( eMemoryAllocation ) );
		else
			_wcstombs( lpBuffer, nBufferMax, getErrorString( eRCStringNotFound, ad.m_ptr, ad.m_len, uID ) );

		return (int)strlen( lpBuffer );
	}

	return res;
}

int TranslateT( wchar_t *lpBuffer, int nBufferMax, const char *mbs )
{
#ifndef _NoTranslation_
	if ( pTranslateDll != NULL )
		return pTranslateDll->Translate( lpBuffer, nBufferMax, mbs );
#endif

	return _mbstowcs( lpBuffer, nBufferMax, mbs );
}

int
LoadStringT(
    HINSTANCE hInstance,
    UINT uID,
    LPWSTR lpBuffer,
    int nBufferMax,
	bool bTranslate,
	bool bCheckParams )
{
	if ( hInstance == NULL )
	{
		ODS_fls( L"Assertion!" );
		return (int)wcslen( copyString( lpBuffer, L"Resource Instance is not initialized", nBufferMax ) );
	}

#ifndef _NoTranslation_
	if ( ( pTranslateDll != NULL ) && bTranslate )
	{
		int res = pTranslateDll->Translate( lpBuffer, nBufferMax, hInstance, uID, bCheckParams );

		bool isTranslated = res != 0;

		if ( isTranslated )
			return res;
	}
#endif

	//
	//	Note! LoadString ( A and W ) does not add last zero for double zero terminated strings
	//	as result the output multy sz string is not really double terminated.
	//

	int res = LoadStringW( hInstance, uID, lpBuffer, nBufferMax );

	if ( res <= 0 )
	{
#ifdef _DEBUG
		return (int)wcslen( getErrorString( eRCStringNotFound, lpBuffer, nBufferMax, uID ) );
#else
		return res;
#endif
	}

	bool doubleZeroTerminated = res > (int)wcslen( lpBuffer );

	if ( doubleZeroTerminated && ( ++res < nBufferMax ) )
		lpBuffer[ res ] = L'\0';

	return res;
}

int EnumWindowsByName( const wchar_t *name, typeEnumWindowActor *enumWindowActor, void *param, LPTSTR wndClass ) // = WC_DIALOG )
{
	// WC_DIALOG #32770 (Dialog)

	HWND hChild = NULL;

	cWCStr wcName( name );
	int len = (int)wcslen( wcName.m_szBody );

	cWCStr wndName( len + 20 );

	int count = 0;

	while( hChild = FindWindowEx( NULL, hChild, wndClass, NULL ) )
	{
		GetWindowTextT( hChild, wndName.m_szBody, wndName.m_len );

		if ( wcsncmp( wndName.m_szBody, wcName.m_szBody, len ) != 0 ) 
			continue;

		if ( ( enumWindowActor != NULL ) && !enumWindowActor( hChild, param ) )
			return -1;

		count++;
	}

	return count;
}

wchar_t *makeValidFileName( wchar_t *name )
{
	int pos = 0;

	for( ; name[pos] != L'\0'; pos++ )
	{
		wchar_t ch = name[pos];
		if ( ( ch == L'\\' ) || ( ch == L'/' ) )
			ch = L'_';
		else
		if ( ch == L':' )
			ch = L'$';
		else
			continue;

		name[pos] = ch;
	}

	return name;
}

char *newString( const char *str, int len )
{
	if ( str == NULL )
		return NULL;

	if ( len == -1 )
		len = (int)strlen( str );

	char *newStr = new char [ len + 1 ];

	if ( newStr != NULL )
	{
		memcpy( newStr, str, len );
		newStr[ len ] = '\0';
	}

	return newStr;
}

wchar_t *copyString( wchar_t *dst, const wchar_t *src, int dstMaxLen, bool setLastZero )
{
	if ( dstMaxLen <= 0 )
	{
		ODSF( L"copyString: Assertion! Destination string length == %d, src = '%s'", dstMaxLen, src );
	}
	else
	{
		int srcLen = (int)wcslen( src );

		if ( ( srcLen >= dstMaxLen ) && !setLastZero )
			ODSF( L"copyString: Warning! Source string length %d >= destination buffer size %d, src = '%s'", srcLen, dstMaxLen, src );

		size_t size = (size_t)dstMaxLen;

		wcsncpy( dst, src, size );

		if ( setLastZero )
			dst[ size - 1 ] = L'\0';
	}

	return dst;
}

wchar_t *catString( wchar_t *dst, const wchar_t *src, int dstMaxLen, bool setLastZero )
{
	int dstLen = (int)wcslen( dst );
	if ( dstLen >= dstMaxLen )
		ODSF( L"catString: Assertion! String length %d >= %d, dst = '%s', src = '%s'", dstLen, dstMaxLen, dst, src );
	else
		copyString( dst + dstLen, src, dstMaxLen - dstLen, setLastZero );

	return dst;
}

wchar_t *newString( const wchar_t *str, int len )
{
	if ( ( str == NULL ) && ( len == -1 ) )
		return NULL;

	if ( len == -1 )
		len = (int)wcslen( str );

	wchar_t *newStr = new wchar_t [ len + 1 ];

	if ( newStr == NULL )
		return NULL;

	if ( str != NULL )
	{
		memcpy( newStr, str, len * sizeof( wchar_t ) );
		newStr[ len ] = L'\0';
	}
	else
		memset( newStr, 0, ( len + 1 ) * sizeof( wchar_t ) );

	return newStr;
}

bool updateString( wchar_t **ptr, const char *val, int len )
{
	cWCStr wcs( val, len );
	return updateString( ptr, wcs.m_szBody );
}

bool updateString( wchar_t **ptr, const wchar_t *val, int len )
{
	return updateString( (const wchar_t **)ptr, val, len );
}

bool updateString( char **ptr, const char *val, int len )
{
	deleteString( ptr );

	if ( val == NULL )
		return true;

	*ptr = newString( val, len );

	return ( *ptr != NULL );
}

bool updateString( const wchar_t **ptr, const wchar_t *val, int len )
{
	deleteString( ptr );

	if ( ( val == NULL ) && ( len == -1 ) )
		return true;

	*ptr = newString( val, len );

	return ( *ptr != NULL );
}

void deleteString( const wchar_t **pValue )
{
	if ( pValue == NULL )
	{
		ODS_fls( L"Assertion!" );
		assert ( false );
		return;
	}

	if ( *pValue == NULL )
		return;

	delete [] *pValue;
	*pValue = NULL;
}

void deleteString( wchar_t **pValue )
{
	return deleteString( (const wchar_t **)pValue );
}

void deleteString( char **pValue )
{
	return deleteString( (const char **)pValue );
}

void deleteString( const char **pValue )
{
	if ( pValue == NULL )
	{
		ODS_fls( L"Assertion!" );
		assert ( false );
		return;
	}

	if ( *pValue == NULL )
		return;

	delete [] *pValue;
	*pValue = NULL;
}

struct sDecodeSymbol
{
	char ch;
	const char *str;
};

sDecodeSymbol urlCodesT[] = {
	' ', "%20",
	'+', "%2B",
	'-', "%2D",
	'/', "%2F",
	'=', "%3D",
	'\r', "",
	'\n', "",
};

bool urlEncode( cWCStr *wcs, const char *bf, int bfSize )
{
	int i;

	for( i = 0; i < bfSize; i++ )
	{
		const char *s = bf + i;
		int l = 1;

		for( int j = maxLen( urlCodesT ); j--; )
		{
			if ( *s == urlCodesT[j].ch )
			{
				s = urlCodesT[j].str;
				l = (int)strlen( s );
				break;
			}
		}

		wcs->Add( s, l );
	}
	
	wcs->Add( "", 1 );

	return true;
}

bool ReleaseSemaphoreT( const wchar_t *semaphoreName, LONG lReleaseCount, LPLONG lpPreviousCount )
{
	//ODSFuncS( semaphoreName );

	HANDLE hSemaphore = OpenSemaphoreT( SEMAPHORE_MODIFY_STATE, false, semaphoreName );

	int err = GetLastError();

	if ( hSemaphore == NULL )
	{
		ODSF( L"Error %x in opening semaphore %s", err, semaphoreName );
		return false;
	}

	LONG prev = -1;
	
	if ( lpPreviousCount == NULL )
		lpPreviousCount = &prev;
	
	BOOL bRes = ReleaseSemaphore( hSemaphore, 1, lpPreviousCount );

	err = GetLastError();

	if ( !bRes )
	{
		ODSF( L"ReleaseSemaphore returns false, err %d", err );
	}

	CloseHandle( hSemaphore );

	//ODSFuncRes( *lpPreviousCount );

	return bRes != 0; //*lpPreviousCount == 0;
}

//
//	TimetToFileTime
//		The FILETIME structure is a 64-bit value representing the number of 100-nanosecond intervals since January 1, 1601 (UTC).
//		time() returns the time as seconds elapsed since midnight, January 1, 1970.
//
void TimetToFileTime( FILETIME *pft, time_t t )
{
	// 21.05.2013 Prev version used signed multiplication here Int32x32To64
    LONGLONG ll = UInt32x32To64(t, 10000000ULL) + 116444736000000000ULL;
    pft->dwLowDateTime = (DWORD) ll;
    pft->dwHighDateTime = (DWORD)(ll >>32);
}

time_t FileTimeToTimet( const FILETIME *ft )
{
   ULARGE_INTEGER ull;
   ull.LowPart = ft->dwLowDateTime;
   ull.HighPart = ft->dwHighDateTime;

   return time_t( ull.QuadPart / 10000000ULL - 11644473600ULL );
}

time_t SystemTimeToTimet( const SYSTEMTIME * pTime )
{
	struct tm tm;
	memset(&tm, 0, sizeof(tm));

	tm.tm_year = pTime->wYear;
	tm.tm_mon = pTime->wMonth - 1;
	tm.tm_mday = pTime->wDay;

	tm.tm_hour = pTime->wHour;
	tm.tm_min = pTime->wMinute;
	tm.tm_sec = pTime->wSecond;

	return (time_t)_mktime64( &tm );
}

const wchar_t *GetReparsePointFinalPathName( cWCS &finaleFileName, wchar_t *reparsePointName, DWORD *pErr )
{
	DWORD err = 0;

	if ( pErr == NULL )
		pErr = &err;

	//
	//	02.06.2016 Bug #2736 WFS - false reparse points
	//		1. The READ_CONTROL right is enought to open link for firther query by GetFinalPathNameByHandleT
	//		The GENERIC_READ gets error here for most of reparse points.
	//		2. OPEN_EXISTING flag get error ERROR_FILE_NOT_FOUND if finale name does not exist. So OPEN_ALWAYS works better.
	//
	HANDLE hFile = CreateFileT( reparsePointName, READ_CONTROL, FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,  NULL,  OPEN_ALWAYS, FILE_FLAG_BACKUP_SEMANTICS, NULL);

	if ( hFile == INVALID_HANDLE_VALUE )
	{
		*pErr = GetLastError();
		ODS_fls( _FormatT( L"Assertion! Error %s in opening file '%s'", _GetSystemError(*pErr), reparsePointName ) );
		return NULL;
	}
	
	DWORD len = GetFinalPathNameByHandleT( hFile, NULL, 0 );

	if ( len == 0 )
	{
		CloseHandle( hFile );
		*pErr = GetLastError();
		ODS_fls( _FormatT( L"Assertion! Error %s in GetFinalPathNameByHandleT for file '%s'", _GetSystemError(*pErr), reparsePointName ) );
		return NULL;
	}

	wchar_t *pos = finaleFileName.AddLength( len + 1 );

	if ( GetFinalPathNameByHandleT( hFile, pos, len ) == 0 )
		*pErr = GetLastError();

	CloseHandle( hFile );

	return pos;
}

DWORD GetFinalPathNameByHandleT( HANDLE hFile, LPWSTR lpszFilePath, DWORD cchFilePath, DWORD dwFlags ) // dwFlags = VOLUME_NAME_DOS
{
	//
	//	Vista has GetFinalPathNameByHandle function, tries it fist.
	//
	typedef DWORD WINAPI typeGetFinalPathNameByHandle(
	  __in   HANDLE hFile,
	  __out  LPWSTR lpszFilePath,
	  __in   DWORD cchFilePath,
	  __in   DWORD dwFlags
	);

	static CDll kernel32dll( false, "KERNEL32.dll", "GetFinalPathNameByHandleW", false );

	if ( kernel32dll.error )
		return 0;

	#ifndef VOLUME_NAME_DOS
	#define VOLUME_NAME_DOS 0x0		// returns '\\?\I:\50M.jbc'
	#define VOLUME_NAME_GUID 0x1	// returns '\\?\Volume{fe79354d-81f8-11e0-b27b-806e6f6e6963}\q1.jbc'
	#define VOLUME_NAME_NONE 0x4	// returns '\50M.jbc'
	#define VOLUME_NAME_NT 0x2		// returns '\Device\HarddiskVolume5\50M.jbc'
	#endif

	DWORD len = ((typeGetFinalPathNameByHandle*)kernel32dll.pProc)( hFile, lpszFilePath, cchFilePath, dwFlags );

	return len;
}

BOOL GetFileNameFromHandle(HANDLE hFile, wchar_t *pszFileName, DWORD size, bool bFinaleNameOnly ) // = MAX_PATH, = false
{
	//ODSFunc;

	if ( GetFinalPathNameByHandleT( hFile, pszFileName, size, VOLUME_NAME_DOS ) > 0 )
		return true;

	if ( bFinaleNameOnly )
		return false;

	HANDLE hFileMap;

	static CDll dll( false, L"Psapi.dll", "GetMappedFileNameW", false );

	typedef DWORD WINAPI typeGetMappedFileNameW( 
		HANDLE hProcess,
		LPVOID lpv,
		LPWSTR lpFilename,
		DWORD nSize
		);

	if ( dll.error )
	{	ODS_fl; return false; 	}

	const int BUFSIZE = 512;

	// Get the file size.
	DWORD dwFileSizeHi = 0;
	DWORD dwFileSizeLo = GetFileSize(hFile, &dwFileSizeHi); 

	if( dwFileSizeLo == 0 && dwFileSizeHi == 0 )
	{
		ODS_fls( L"Assertion! Cannot map a file with a length of zero.\n" );
//		return FALSE;
	}

	bool bSuccess = false;

	// Create a file mapping object.
	hFileMap = CreateFileMapping(hFile, 
		NULL, 
		PAGE_READONLY,
		0, 
		1,
		NULL);

	if (hFileMap) 
	{
		// Create a file mapping to get the file name.
		void* pMem = MapViewOfFile(hFileMap, FILE_MAP_READ, 0, 0, 1);

		if (pMem) 
		{
			if ( (( typeGetMappedFileNameW *)( dll.pProc ))( GetCurrentProcess(), pMem,	pszFileName, MAX_PATH ) ) 
				bSuccess = true;
		
			UnmapViewOfFile(pMem);
		} 

		CloseHandle(hFileMap);
	}

	int err = -1;

	if ( bSuccess )
	{
		// Sample:
		// GetMappedFileName returns '\Device\HarddiskVolume5\50M.jbc'
		// QueryDosDeviceT( C: ) returns '\Device\HarddiskVolume2'

		cWCStr dosName( MAX_PATH );
		
		if ( GetDOSNameForHarddiskVolume( dosName.m_szBody, dosName.m_len, pszFileName ) )
			FormatT( pszFileName, MAX_PATH, dosName.m_szBody );
	}

	return bSuccess;
}

//
//	GetDOSNameForHarddiskVolume
//		Input name is '\Device\HarddiskVolumeNN\filePath'
//		Enumerates logical drives GetLogicalDrives 
//		Queries dos device name by QueryDosDeviceT ( X:\ ) -> '\Device\HarddiskVolumeNN'
//		Formats output dos name as 'X:\filePath'
//		Returns true if path is translated.
//
bool GetDOSNameForHarddiskVolume( wchar_t *bf, int bfLen, const wchar_t *hardDiskFilePath )
{
	// Sample:
	// GetMappedFileName returns '\Device\HarddiskVolume5\50M.jbc'
	// QueryDosDeviceT( C: ) returns '\Device\HarddiskVolume2'

	const wchar_t *szDeviceHarddiskVolume = L"\\Device\\HarddiskVolume";
	size_t len = wcslen( szDeviceHarddiskVolume );

	if ( _wcsnicmp( hardDiskFilePath, szDeviceHarddiskVolume, len ) != 0 )
		return false;

	for( ; hardDiskFilePath[ len ] != L'\\'; len++ )
	{
		if ( hardDiskFilePath[ len ] == L'\0' )
		{
			ODS_fls( L"Assertion!" );
			return false;
		}
	}

	DWORD dwMask = GetLogicalDrives();
	wchar_t drv[4] = L"A:";
	cWCStr dosName( MAX_PATH );

	for( int driveN = 2; driveN < 26; driveN++ )
	{
		if ( ( ( 1 << driveN ) & dwMask ) == 0 )
			continue;

		drv[0] = L'A' + driveN;

		if ( !QueryDosDeviceT( drv, dosName.m_szBody, dosName.m_len ) )
		{
			ODS_fls( L"Assertion!" );
			continue;
		}

		if ( dosName.equ( hardDiskFilePath, (int)len ) )
		{
			FormatT( bf, bfLen, L"%c:%s", drv[0], hardDiskFilePath + len );
			return true;
		}
	}

	return false;
}

#ifndef _NoShell32Lib_

void FreeIDList( ITEMIDLIST *pidl )
{
	if ( pidl == NULL )
		return;

	//	CoTaskMemFree( pidl );

	LPMALLOC pMalloc = NULL;

	HRESULT hr = SHGetMalloc(&pMalloc);

	if ( hr != S_OK )
		return;

	pMalloc->Free(pidl);
	pMalloc->Release();
}

//
//	GetIDListFromPath
//		Creates PID list from specified path. Remember fre the created list by FreeIDList
//
ITEMIDLIST *GetIDListFromPath( LPCWSTR path, bool bCoInitialize ) // = true
{
	//typedef HRESULT STDAPICALLTYPE typeCoInitialize( LPVOID pvReserved );
	//typedef void STDAPICALLTYPE typeCoUninitialize( void );

	//static CDll dll( false, L"Ole32.dll", "CoInitialize", false );
	//if ( dll.error != CDll::errNo )
	//	return NULL;

	//static typeCoInitialize *procCoInitialize = ( typeCoInitialize * )dll.pProc;
	//static typeCoUninitialize *procCoUninitialize = ( typeCoUninitialize * )GetProcAddress( dll.hInst, "CoUninitialize" );

	//if ( ( procCoInitialize == NULL ) || ( procCoUninitialize == NULL ) )
	//	return NULL;

	//HRESULT hrCo;
	//
	////	hrCo = CoInitialize(NULL); // return S_OK If the COM library was initialized successfully on this thread, so we'll free this below.
	//hrCo = procCoInitialize( NULL );

	cCoInitialize cCOM( NULL, bCoInitialize );

	if ( bCoInitialize && ( cCOM.m_hRes == NULL ) )
	{
		ODS_fls( L"Assertion! Error in CoInitialize" );
		return NULL;
	}

	LPSHELLFOLDER psfDesktop = NULL;
	HRESULT hr = SHGetDesktopFolder( &psfDesktop );
	
	ITEMIDLIST *pidl = NULL;

	if ( ( hr == NOERROR ) && ( psfDesktop != NULL ) )
	{
		hr = psfDesktop->ParseDisplayName(NULL, NULL, (LPOLESTR)path, NULL, &pidl, NULL);
		psfDesktop->Release();
	}

	//if ( hrCo == S_OK ) 
	//{
	//	//CoUninitialize();
	//	procCoUninitialize();
	//}

	return pidl;
}

#endif // _NoShell32Lib_

//
//	SetFilePointerT
//		Return new file position if success otherwise (DWORDLONG)-1
//
DWORDLONG SetFilePointerT( HANDLE hFile, LONGLONG llDistanceToMove, DWORD dwMoveMethod )
{
	//ODSFuncS( _FormatT( L"%X, %I64X, %X", hFile, llDistanceToMove, dwMoveMethod ) );

	ULARGE_INTEGER uli;
	uli.QuadPart = llDistanceToMove;

	uli.LowPart = SetFilePointer( hFile, uli.LowPart, (LONG*)&uli.HighPart, dwMoveMethod );

	bool ok = ( ( uli.LowPart != INVALID_SET_FILE_POINTER ) || ( GetLastError() == NO_ERROR ) );

	if ( !ok )
		return (DWORDLONG)-1;

	//ODSFuncRes( (__int64)uli.QuadPart );

	return uli.QuadPart;
}

bool SetFileSize( HANDLE hFile, DWORDLONG fileSize, DWORD *pErr )
{
	DWORDLONG res = SetFilePointerT( hFile, (LONGLONG)fileSize, FILE_BEGIN );

	DWORD err;
	if ( pErr == NULL )
		pErr = &err;

	if ( res == (DWORDLONG)-1 )
	{
		*pErr = GetLastError();
		ODS_fls( _FormatT( L"Assertion! SetFilePointerT( %X, %I64X, FILE_BEGIN ) error %X", hFile, (LONGLONG)fileSize, *pErr ) );
		return false;
	}

	bool ok = SetEndOfFile( hFile ) == TRUE;

	if ( !ok )
	{
		*pErr = GetLastError();
		ODS_fls( _FormatT( L"Assertion! SetEndOfFile( %X ) at %I64X, error %X", hFile, (LONGLONG)fileSize, *pErr ) );
	}
	
	return ok;
}

bool SetFileSize( const wchar_t *filePath, DWORDLONG fileSize, bool bFlushBuffers )
{
	bool ok = true;

	HANDLE hFile = CreateFileT( filePath, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_ALWAYS, 0, NULL );

	if ( hFile == INVALID_HANDLE_VALUE )
	{
		ODS_fls( _FormatT( L"Assertion! CreateFileT return %d", GetLastError() ) );
		ok = false;
	}

	if ( ok )
		ok = SetFileSize( hFile, fileSize );

	if ( ok && bFlushBuffers )
		FlushFileBuffers( hFile );

	CloseHandle( hFile );

	return ok;
}

//
// The WNetGetConnection function retrieves the name of the network resource associated with a local device.
//	Key words: 
//		UNC name
//		mapped netword drive.
//
DWORD WNetGetConnectionT( LPCWSTR lpLocalName, LPWSTR  lpRemoteName, LPDWORD lpnLength )
{
	static CDll dll( false, L"Mpr.dll", "WNetGetConnectionW", false );

	typedef DWORD APIENTRY typeWNetGetConnectionW( LPCWSTR lpLocalName, LPWSTR  lpRemoteName, LPDWORD lpnLength );

	if ( dll.error != CDll::errNo )
		return -1;

	return ((typeWNetGetConnectionW*)dll.pProc)( lpLocalName, lpRemoteName, lpnLength );
}

DWORD WaitThreadAndTerminate_SA( const wchar_t *szFunction, int iLine, typeTerminateThreadActor *actor, void *ctx, HANDLE hThread, DWORD dwMilliseconds, DWORD exitCode )
{
	DWORD waitRes = WAIT_TIMEOUT;
	
	if ( dwMilliseconds != 0 )
		waitRes = WaitForSingleObject( hThread, dwMilliseconds );

	if ( waitRes == WAIT_TIMEOUT )
	{
		ODS( szFunction, iLine, _FormatT( L"Wait thread timeout (%d)!", dwMilliseconds ) );

		// Terminate thread is dangerous function so it is excluded.
		//TerminateThread( hThread, exitCode );
		if ( actor != NULL )
			waitRes = actor( ctx, _FormatT( L"%s:%d", szFunction, iLine ), hThread, exitCode );
	}
	else
	if ( waitRes != WAIT_OBJECT_0 )
		ODS( szFunction, iLine, _FormatT( L"Wait thread result (%X)!", waitRes ) );

	return waitRes;
}

BOOL SetFileTimeT(
    const wchar_t *fileName,
    CONST FILETIME *lpCreationTime,
    CONST FILETIME *lpLastAccessTime,
    CONST FILETIME *lpLastWriteTime
    )
{
	cAutoFile chf( fileName, cAutoFile::openMode_ReadWrite, cAutoFile::shareMode_All, NULL, OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS );

	if ( chf.m_hFile == INVALID_HANDLE_VALUE )
	{
		ODS_fls( _FormatT( L"Assertion! Error %d in opening file '%s'", chf.m_error, fileName ) );
		return false;
	}

	return SetFileTime( chf.m_hFile, lpCreationTime, lpLastAccessTime, lpLastWriteTime );
}

BOOL SafeGetFileTime(
	const wchar_t *fileName,
	LPFILETIME lpCreationTime,
	LPFILETIME lpLastAccessTime, 
	LPFILETIME lpLastWriteTime,
	LPDWORD lpDWFileAttributes,
	DWORDLONG *lpSizeSize
	) 
{
	WIN32_FIND_DATAW findData;

	HANDLE findFile = FindFirstFileT( fileName, &findData );

	if ( findFile == INVALID_HANDLE_VALUE )
		return false;

	if (lpCreationTime)		*lpCreationTime		= findData.ftCreationTime;
	if (lpLastAccessTime)	*lpLastAccessTime	= findData.ftLastAccessTime;
	if (lpLastWriteTime)	*lpLastWriteTime	= findData.ftLastWriteTime;
	if (lpDWFileAttributes)	*lpDWFileAttributes = findData.dwFileAttributes;
	if (lpSizeSize)
	{
		ULARGE_INTEGER uli;
		uli.LowPart = findData.nFileSizeLow;
		uli.HighPart = findData.nFileSizeHigh;
		*lpSizeSize = uli.QuadPart;
	}

	FindClose( findFile );

	return true;
}

#ifndef FileBasicInfo
#define FILE_INFO_BY_HANDLE_CLASS int
#define FileBasicInfo 0
#endif

bool SetFileInformationT( const wchar_t *fileName, FILE_INFO_BY_HANDLE_CLASS FileInformationClass, LPVOID lpFileInformation, DWORD dwBufferSize )
{
	typedef BOOL WINAPI typeSetFileInformationByHandle( HANDLE hFile, FILE_INFO_BY_HANDLE_CLASS FileInformationClass, LPVOID lpFileInformation, DWORD dwBufferSize );

	static CDll kernel32dll( false, "KERNEL32.dll", "SetFileInformationByHandle", false );

	if ( kernel32dll.error )
		return false;

	cAutoHandle ah;
	ah.m_handle = CreateFileT( fileName, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE, NULL, OPEN_EXISTING, 0, NULL );

	if ( ah.m_handle == INVALID_HANDLE_VALUE )
		return false;

	return ((typeSetFileInformationByHandle *)kernel32dll.pProc)( ah.m_handle, FileInformationClass, lpFileInformation, dwBufferSize ) == TRUE;
}

bool GetFileInformationT( const wchar_t *fileName, FILE_INFO_BY_HANDLE_CLASS FileInformationClass, LPVOID lpFileInformation, DWORD dwBufferSize )
{
	typedef BOOL WINAPI typeGetFileInformationByHandleEx( HANDLE hFile, FILE_INFO_BY_HANDLE_CLASS FileInformationClass, LPVOID lpFileInformation, DWORD dwBufferSize );

	static CDll kernel32dll( false, "KERNEL32.dll", "GetFileInformationByHandleEx", false );

	if ( kernel32dll.error )
		return false;

	cAutoHandle ah;
	ah.m_handle = CreateFileT( fileName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL );

	if ( ah.m_handle == INVALID_HANDLE_VALUE )
		return false;

	return ((typeGetFileInformationByHandleEx *)kernel32dll.pProc)( ah.m_handle, FileInformationClass, lpFileInformation, dwBufferSize ) == TRUE;
}

BOOL cFileAttributes::Get( const wchar_t *fileName )
{
	m_biOk = GetFileInformationT( fileName, FileBasicInfo, &m_bi, sizeof( m_bi ) );
	if ( !m_biOk )
		ODS_fls( _FormatT( L"Warning! GetFileInformationT( %s ) returns error %s", _GetLastSystemError() ) );

	BOOL ok = SafeGetFileTime( fileName, &creationTime, &lastAccessTime, &lastWriteTime, &attributes );

	m_err = GetLastError();

	if ( !ok )
		ODS_fls( _FormatT( L"Warning! SafeGetFileTime( %s ) returns error %s", _GetSystemError( m_err ) ) );

	return ok;
}

BOOL cFileAttributes::Set( const wchar_t *fileName )
{
	//
	//	Version 6.08.4.1 - 14.12.2015
	//		SetFileInformationT calls SetFileInformationByHandle, the function works with FILE_BASIC_INFO structure.
	//		It has four times including ChangeTime. Windows shows not shows three time only:
	//		CreationTime, LastAccessTime, LastWriteTime. But anoer application ( like FAR ) shows the ChangeTime.
	//
	if ( m_biOk && SetFileInformationT( fileName, FileBasicInfo, &m_bi, sizeof( m_bi ) ) )
		return true;

	//
	//	Version 6.08.4.1 - 11.12.2015
	//	Bug #2426 Timestamp changed by Wipe Free Space
	//		Read only attributes was restored before times, so function SetFileTimeT was failed.
	//		Function SetFileAttributes does not enfluent on usual timestamp.
	//		But both the functions modify 'Change time' field shown by FAR.
	//
	if ( lastWriteTime.dwHighDateTime == 0 )
		ODS_fls( _FormatT( L"Warning! File '%s' - time is zero", fileName ) );

	BOOL ok = SetFileTimeT( fileName, &creationTime, &lastAccessTime, &lastWriteTime );
	SetFileAttributesT( fileName, attributes );
	return ok;
}

bool PostCloseAppMessage( const wchar_t *wndClass, const wchar_t *caption ) // caption = NULL
{
	ODSFuncS( _FormatT( L"%s, %s", wndClass, caption ) );

	HWND hWnd = FindWindowT( wndClass, caption );

	if ( hWnd == NULL )
		return false;
	else
	{
		PostMessage( hWnd, WM_CLOSE, 0 , 0 );
		hWnd = FindWindowT( wndClass, caption );
	}

	if ( hWnd )
	{
		PostMessage( hWnd, WM_QUIT, 0 , 0 );
		hWnd = FindWindowT( wndClass, caption );
	}

	if ( hWnd )
	{
		PostMessage( hWnd, WM_QUERYENDSESSION, 0, ENDSESSION_LOGOFF );
		hWnd = FindWindowT( wndClass, caption );
	}

	if ( hWnd )
	{
		//
		//	EndTask rudimentary function, but it allows forcely close application window.
		//	If fForce is false it just send WM_CLOSE.
		//
		typedef BOOL WINAPI typeEndTask( HWND hWnd, BOOL fShutDown, BOOL fForce );

		static CDll dll( false, L"user32.dll", "EndTask", false );

		if ( !dll.error )
		{
			((typeEndTask*)dll.pProc)( hWnd, FALSE, true );
			hWnd = FindWindowT( wndClass, caption );
		}
	}

	return hWnd == NULL;
}
#endif // #ifdef _WIN32